program ex1_bd_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines BD_CMP and ORTHO_GEN_BD
!   in module SVD_Procedures .
!                                                                              
! LATEST REVISION : 23/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, bd_cmp,         &
                         ortho_gen_bd, norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, m=3000, n=3000, nm=min(n,m)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of bd_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, tauq, taup
    real(stnd), allocatable, dimension(:,:) :: a, a2, resid, bd, p
!
    integer(i4b) :: l
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : BIDIAGONAL REDUCTION OF A m-by-n REAL MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), p(n,nm), d(nm), e(nm),     &
              tauq(nm), taup(nm), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM DATA MATRIX .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), bd(nm,nm), resid(nm,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST, CALL bd_cmp TO REDUCE THE MATRIX a TO BIDIAGONAL FORM
!
!                      a = Q*BD*P**(t)
!
!   WHERE Q AND P ARE ORTHOGONAL AND BD IS AN UPPER OR LOWER BIDIAGONAL MATRIX.
!
    call bd_cmp( a, d, e, tauq, taup )
!
!   ON OUTPUT OF bd_cmp:
!
!       a, tauq AND taup CONTAINS THE ELEMENTARY REFLECTORS
!       DEFINING Q AND P IN PACKED FORM.
!
!       d AND e CONTAINS RESPECTIVELY, THE DIAGONAL AND
!       SUBDIAGONAL OF THE BIDIAGONAL MATRIX BD.
!
!   SECOND, CALL ortho_gen_bd TO GENERATE Q AND P.
!
    call ortho_gen_bd( a, tauq, taup, p )
!
!   ON OUTPUT OF ortho_gen_bd, a CONTAINS THE FIRST min(n,m) COLUMNS OF Q
!   AND p CONTAINS THE ORTHOGONAL MATRIX P.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION Q**(t)*a - BD*P**(t),
!
        bd(:nm,:nm) = zero
!
        if ( m>=n ) then
!
!           BD IS UPPER BIDIAGONAL.
!
            do l = 1_i4b, nm-1_i4b
                bd(l,l)       = d(l)
                bd(l,l+1_i4b) = e(l+1_i4b)
            end do
!
            bd(nm,nm) = d(nm)
!
        else
!
!           BD IS LOWER BIDIAGONAL.
!
            bd(1_i4b,1_i4b) = d(1_i4b)
!
            do l = 2_i4b, nm
                bd(l,l-1_i4b) = e(l)
                bd(l,l)       = d(l)
            end do
!
        endif
!
        resid(:nm,:n) = matmul( transpose(a(:m,:nm)), a2(:m,:n) )           &
                        - matmul( bd(:nm,:nm), transpose(p(:n,:nm )) )
!
        bd(:nm,1_i4b) = norm( resid(:nm,:n), dim=1_i4b )
        err1 =  maxval( bd(:nm,1_i4b) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
        call unit_matrix( a2(:nm,:nm) )
!
        resid(:nm,:nm) = abs( a2(:nm,:nm) - matmul( transpose(a(:m,:nm )), a(:m,:nm ) ) )
        err2 = maxval( resid(:nm,:nm) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - P**(t)*P.
!
        resid(:nm,:nm) = abs( a2(:nm,:nm) - matmul( transpose(p(:n,:nm )), p(:n,:nm ) ) )
        err3 = maxval( resid(:nm,:nm) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, bd, resid )
!
    endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, p, d, e, tauq, taup )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed bidiagonal decomposition a = Q*BD*P**(t) = ', err1
        write (prtunit,*) 'Orthogonality of the computed Q orthogonal matrix                 = ', err2
        write (prtunit,*) 'Orthogonality of the computed P orthogonal matrix                 = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the bidiagonal reduction of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_cmp
! =========================
!
end program ex1_bd_cmp
