program ex1_bd_coef
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines/functions BD_COEF,
!   FREQ_FUNC and SYMLIN_FILTER in module Time_Series_Procedures.
!                                                                              
! LATEST REVISION : 12/11/2007
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, zero, half, one, two, pi, true, false, merror, allocate_error,   &
                         bd_coef, freq_func, symlin_filter, init_fft, fft, end_fft
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE LENGTH OF THE TIME SERIES .
!
    integer(i4b), parameter :: prtunit=6, n=2001
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err, tmp, fch, fcl
    real(stnd), dimension(n)              :: y, y2, y3, freqr
    real(stnd), dimension(:), allocatable :: coef
!
    complex(stnd), dimension(n)           :: yc
!
    integer(i4b) :: i, k, k1, k2, pl, ph, nfilt, khalf, kmid
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of bd_coef'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM REAL NUMBER SEQUENCE OF LENGTH n .
!
    call random_number( y(:n) )
!
!   SAVE THE REAL RANDOM NUMBER ARRAY.
!
    y2(:n) = y(:n)
!
!   pl IS THE MINIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   ph IS THE MAXIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   pl AND ph ARE EXPRESSED IN NUMBER OF OBSERVATIONS, I.E. pl=6(18) and ph=32(96) SELECT
!   PERIODS BETWEEN 1.5 YRS AND 8 YRS FOR QUARTERLY (MONTHLY) DATA.
!
    pl  = 35
    ph  = 96
!
!   COMPUTE THE CORRESPONDING CUTOFF FREQUENCIES.
!
    fch = one/real( ph, stnd )
    fcl = one/real( pl, stnd )
!
!   NOW SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS k FOR THE LANCZOS FILTER.
!
    k1  = ceiling( one/(half-fcl) )
    k2  = ceiling(  2.6/(fcl-fch) )
    k   = max( k1, k2, ph+1 )
    if ( (k/2)*2==k  ) k = k + 1
!
!   ALLOCATE WORK ARRAY FOR THE FILTER COEFFICIENTS.
!
    allocate( coef(k), stat=iok )
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FUNCTION bd_coef COMPUTES THE K-TERM LEAST SQUARES APPROXIMATION TO AN
!   -IDEAL- BAND PASS FILTER WITH CUTOFF PERIODS PL AND PH (EG CUTOFF FREQUENCIES 1/PL AND 1/PH).
!
    coef(:k) = bd_coef( PL=pl, PH=ph, K=k )
!
!   SUBROUTINE symlin_filter PERFORMS A SYMMETRIC FILTERING OPERATION ON AN IMPUT
!   TIME SERIES (THE ARGUMENT VEC).
!
!   HERE symlin_filter FILTERS THE TIME SERIES, KEEPING THE PERIODS BETWEEN pl AND ph .
!   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE LOST FROM EACH END OF THE SERIES SO THAT
!   NFILT (NFILT= size(VEC) - size(COEF) + 1) TIME OBSERVATIONS ARE RETURNED IN VEC(:NFILT)
!   AND THE REMAINING OBSERVATIONS ARE SET TO ZERO.
!
    call symlin_filter( VEC=y2(:n), COEF=coef(:k),  NFILT=nfilt )
!
!   NOW COMPUTE THE TRANSFERT FUNCTION freqr(:) OF THE LANCZOS FILTER AT THE FOURIER FREQUENCIES.
!
    call freq_func( NFREQ=n, COEF=coef(:k), FREQR=freqr(:n), FOUR_FREQ=true )
!
!   NOW, APPLY THE FILTER USING THE FFT AND THE CONVOLUTION THEOREM.
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    call init_fft( n )
!
!   TRANSFORM THE TIME SERIES.
!
    yc(1:n) = cmplx( y(1:n), zero, kind=stnd )
!
    call fft( yc(:n), forward=true  )
!
!   MULTIPLY THE FOURIER TRANSFORM OF THE TIME SERIES
!   BY THE TRANSFERT FUNCTION OF THE FILTER.
!
    yc(:n) = yc(:n)*freqr(:n)
!
!   INVERT THE SEQUENCE BACK TO GET THE FILTERED TIME SERIES.
!
    call fft( yc(:n), forward=false )
!
    y3(:n) = real( yc(:n),  kind=stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft()
!
    deallocate( coef )
!
!   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
    kmid  = ( k + 1 )/2
    khalf = ( k - 1 )/2
    k1    = kmid
    k2    = n - khalf
!
    err = maxval(abs(y3(k1:k2)-y2(:nfilt)))/maxval(abs(y(k1:k2)))
!
    if ( err<=sqrt(epsilon(err))  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_bd_coef
! ==========================
!
end program ex1_bd_coef
