program ex1_bd_deflate
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_DEFLATE
!   in module SVD_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_SINGVAL in module SVD_Procedures.
!
! LATEST REVISION : 27/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, c50, bd_deflate,    &
                         bd_singval, unit_matrix, norm, lamch, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, nsing=100
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of bd_deflate'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, abstol, normbd, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diag, sup, singval, resid2
    real(stnd), allocatable, dimension(:,:) :: leftvec, rightvec, resid
!
    integer(i4b) :: max_qr_steps, nsing2
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure1, failure2, bd_is_upper, ortho, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL SVD OF A REAL BIDIAGONAL MATRIX BD USING A BISECTION ALGORITHM
!               FOR SINGULAR VALUES AND THE GODUNOV DEFLATION TECHNIQUE FOR SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    abstol = sqrt(lamch('s'))
    eps    = fudge*epsilon( err )
    err    = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( diag(n), sup(n), singval(n),                    &
              leftvec(n,nsing), rightvec(n,nsing), stat=iok   )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE AN UPPER BIDIAGONAL MATRIX BD.
!   THE DIAGONAL ELEMENTS ARE STORED IN diag.
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN sup.
!
    bd_is_upper  = true
!
    sup(1_i4b)   = zero
!
    diag(:n)     = half
    sup(2_i4b:n) = half
!
!    call random_number( diag(:n) )
!    call random_number( sup(2_i4b:n) )
!
!    diag(1_i4b)   = 1._stnd
!    diag(2_i4b:n) = 200._stnd*epsilon( err )
!    sup(2_i4b:n)  = 200._stnd*epsilon( err )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST COMPUTE THE LARGEST nsing SINGULAR VALUES OF THE BIDIAGONAL MATRIX BD.
!
!   ON ENTRY OF bd_singval :
!
!         diag(:n) MUST CONTAIN THE DIAGONAL ELEMENTS OF THE BIDIAGONAL MATRIX.
!         sup(:n)  MUST CONTAIN THE OFF-DIAGONAL ELEMENTS OF THE BIDIAGONAL MATRIX (sup(1) IS ARBITRARY).
!
!         THE OPTIONAL ARGUMENT abstol OF bd_singval MAY BE USED TO INDICATE THE REQUIRED PRECISION FOR THE
!         SINGULAR VALUES. A SINGULAR VALUE IS CONSIDERED TO BE LOCATED IF IT HAS BEEN DETERMINED TO LIE IN
!         AN INTERVAL WHOSE WIDTH IS EQUAL OR LESS THAN abstol. SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY
!         WHEN abstol IS SET TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (EG sqrt(LAMCH('S')) ).
!
!         THE OPTIONAL ARGUMENT ls MAY BE USED TO INDICATE THE NUMBER OF SINGULAR VALUES TO BE COMPUTED.
!
    call bd_singval( diag(:n), sup(:n), nsing2, singval(:n), failure=failure1, sort=sort, abstol=abstol, ls=nsing  )
!
!   ON EXIT OF bd_singval :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE BISECTION ALGORITHM.
!
!         THE COMPUTED SINGULAR VALUES ARE STORED IN THE FIRST nsing2 ELEMENTS OF THE ARRAY singval IN
!         DECREASING (sort='d') OR ASCENDING (sort='a') ORDER. NOTE THAT nsing2 MAY BE GREATER THAN
!         ARGUMENT ls IN CASE OF MULTIPLE SINGUAR VALUES.
!
!   NEXT COMPUTE THE FIRST nsing SINGULAR VECTORS OF THE BIDIAGONAL MATRIX BD
!   BY A DEFLATION TECHNIQUE WITH SUBROUTINE bd_deflate.
!
!   ON ENTRY OF bd_deflate:
!
!         bd_is_upper INDICATES IF THE BIDIAGONAL MATRIX IS UPPER OR LOWER BIDIAGONAL.
!
!         diag(:n) MUST CONTAIN THE DIAGONAL ELEMENTS OF THE BIDIAGONAL MATRIX.
!         sup(:n)  MUST CONTAIN THE OFF-DIAGONAL ELEMENTS OF THE BIDIAGONAL MATRIX (sup(1) IS ARBITRARY).
!
!         PARAMETER singval CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX bd.
!         THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
!         OPTIONAL PARAMETER ortho DETERMINES IF DEFLATION IS APPLIED SEQUENTIALLY TO ALL
!         COMPUTED SINGULAR VECTORS (ortho=true) OR ONLY IN EACH CLUSTER OF SINGULAR VALUES
!         (ortho=false). IN THE LATTER CASE, ORTHOGONALITY OF THE SINGULAR VECTORS MAY BE
!         SLIGHTLY DEGRADED, BUT COMPUTATIONS WILL BE MUCH FASTER.
!
    ortho        = true
    max_qr_steps = 10_i4b
!
    call bd_deflate( bd_is_upper, diag(:n), sup(:n), singval(:nsing), leftvec(:n,:nsing), rightvec(:n,:nsing),  &
                     failure=failure2, ortho=ortho, max_qr_steps=max_qr_steps )
!
!   ON EXIT OF bd_deflate :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE DEFLATION ALGORITHM.
!
!         leftvec AND rightvec CONTAIN, RESPECTIVELY, THE nsing LEFT AND RIGHT
!         SINGULAR VECTORS OF THE BIDIAGONAL MATRIX BD ASSOCIATED WITH THE SINGULAR VALUES singval(:nsing).
!
!         bd_deflate MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER singval ARE NEARLY
!         IDENTICAL FOR SOME PATHOLOGICAL BIDIAGONAL MATRICES.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( resid2(nsing), resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION BD*rightvec - leftvec*diag(singval(:nsing)),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX BD.
!
        if ( bd_is_upper ) then
!
            resid(:n,:nsing) = spread( diag(:n),        dim=2, ncopies=nsing )*rightvec                 + &
                               eoshift( spread(sup(:n), dim=2,ncopies=nsing)*rightvec, shift=1 )        - &
                               spread( singval(:nsing), dim=1, ncopies=n )*leftvec

        else
!
            resid(:n,:nsing) = spread( diag(:n),        dim=2, ncopies=nsing )*leftvec                   + &
                               eoshift( spread(sup(:n), dim=2,ncopies=nsing)*leftvec, shift=1 )          - &
                               spread( singval(:nsing), dim=1, ncopies=n )*rightvec
!
        end if
!
        resid2(:nsing) = norm( resid(:n,:nsing), dim=2_i4b )
!
        normbd = sum( diag(:n)*diag(:n) + sup(2_i4b:n)*sup(2_i4b:n) )
!
        err1 = maxval( resid2(:nsing) )/( normbd*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - leftvec**(t)*leftvec
!       WHERE leftvec ARE THE LEFT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        call unit_matrix( resid(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( resid(:nsing,:nsing) - matmul( transpose( leftvec ), leftvec ) )
        err2 = maxval(resid(:nsing,:nsing))/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - rightvec(t)*rightvec
!       WHERE rightvec ARE THE RIGHT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        call unit_matrix( resid(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( resid(:nsing,:nsing) - matmul( transpose( rightvec ), rightvec ) )
        err3 = maxval(resid(:nsing,:nsing))/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( resid, resid2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( diag, sup, singval, leftvec, rightvec )
!
!   PRINT RESULTS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', nsing, ' singular values and vectors of a ', &
       n, ' by ', n,' real bidiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_deflate
! =============================
!
end program ex1_bd_deflate
