program ex1_bd_inviter
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_INVITER
!   in module SVD_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_SVD in module SVD_Procedures.
!
! LATEST REVISION : 27/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, c100, bd_inviter, bd_svd, &
                         unit_matrix, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, nsing=100
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='Example 1 of bd_inviter'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diag, sup, sup2, singval
    real(stnd), allocatable, dimension(:,:) :: id, resid, leftvec, rightvec
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=2
!
    logical(lgl) :: failure1, failure2, bd_is_upper, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL SVD OF A REAL n-BY-n BIDIAGONAL MATRIX BD USING THE GOLUB-REINSCH ALGORITHM
!               FOR ALL SINGULAR VALUES AND THE INVERSE ITERATION TECHNIQUE FOR SELECTED SINGULAR
!               VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( diag(n), sup(n), sup2(n), singval(n),           &
              leftvec(n,nsing), rightvec(n,nsing), stat=iok   )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE AN UPPER BIDIAGONAL MATRIX bd.
!   THE DIAGONAL ELEMENTS ARE STORED IN diag .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN sup .
!
    bd_is_upper   = true
!
    sup(1_i4b)   = zero
!
!    diag(:n)     = half
!    sup(2_i4b:n) = half
!
    call random_number( diag(:n) )
    call random_number( sup(2_i4b:n) )
!
!    diag(1_i4b)   = 1._stnd
!    diag(2_i4b:n) = 200._stnd*epsilon( err )
!    sup(2_i4b:n)  = 200._stnd*epsilon( err )
!
!   MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
    singval(:n) = diag(:n)
    sup2(:n)    = sup(:n)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE ALL SINGULAR VALUES OF THE BIDIAGONAL MATRIX BD .
!   THE SINGULAR VALUES ARE STORED IN singval IN DECREASING ORDER (sort='d').
!
    call bd_svd( bd_is_upper, singval(:n), sup2(:n), failure=failure1, sort=sort  )
!
!   ON EXIT OF bd_svd :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE BIDIAGONAL QR ALGORITHM.
!
!   COMPUTE THE FIRST nsing SINGULAR VECTORS OF BD BY maxiter INVERSE ITERATIONS WITH
!   SUBROUTINE bd_inviter. THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    call bd_inviter( bd_is_upper, diag(:n), sup(:n), singval(:nsing), leftvec(:n,:nsing), rightvec(:n,:nsing),  &
                     failure=failure2, maxiter=maxiter )
!
!   ON EXIT OF bd_inviter :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM.
!
!         leftvec AND rightvec CONTAIN, RESPECTIVELY, THE nsing LEFT AND RIGHT SINGULAR
!         VECTORS OF THE BIDIAGONAL MATRIX BD ASSOCIATED WITH THE SINGULAR VALUES singval(:nsing).
!
!   bd_inviter MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER singval ARE NEARLY
!   IDENTICAL FOR SOME PATHOLOGICAL BIDIAGONAL MATRICES.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( id(nsing,nsing), resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION bd*rightvec - leftvec*diag(singval(:nsing)),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        if ( bd_is_upper ) then
!
            resid(:n,:nsing) = spread( diag(:n),        dim=2, ncopies=nsing )*rightvec                 + &
                               eoshift( spread(sup(:n), dim=2,ncopies=nsing)*rightvec, shift=1 )        - &
                               spread( singval(:nsing), dim=1, ncopies=n )*leftvec

        else
!
            resid(:n,:nsing) = spread( diag(:n),        dim=2, ncopies=nsing )*leftvec                   + &
                               eoshift( spread(sup(:n), dim=2,ncopies=nsing)*leftvec, shift=1 )          - &
                               spread( singval(:nsing), dim=1, ncopies=n )*rightvec
!
        end if
!
        id(:nsing,1_i4b) = norm( resid(:n,:nsing), dim=2_i4b )
        err1 = maxval( id(:nsing,1_i4b) )/( sum( singval(:nsing) )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - leftvec**(t)*leftvec
!       WHERE leftvec ARE THE LEFT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        call unit_matrix( id )
!
        resid(:nsing,:nsing) = abs( id - matmul( transpose( leftvec ), leftvec ) )
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - rightvec**(t)*rightvec
!       WHERE rightvec ARE THE RIGHT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        resid(:nsing,:nsing) = abs( id - matmul( transpose( rightvec ), rightvec ) )
        err3 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( id, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( diag, sup, sup2, singval, leftvec, rightvec )
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a ', &
       n, ' by ', n,' real bidiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_inviter
! =============================
!
end program ex1_bd_inviter
