program ex1_bd_singval2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_SINGVAL2
!   in module SVD_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_INVITER in module SVD_Procedures.
!
! LATEST REVISION : 08/01/2016
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, c50,        &
                         bd_inviter, bd_singval2, unit_matrix, norm, geop, lamch, &
                         merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=50, ls=20
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, abstol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, s
    real(stnd), allocatable, dimension(:,:) :: id, resid, leftvec, rightvec
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=2, nsing
!
    logical(lgl) :: failure1, failure2, a_is_upper, do_test
!   
    character    :: sort='d'
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of bd_singval2'
!   
    real(stnd), parameter  :: fudge=c50
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : ls SINGULAR VALUES AND VECTORS OF A n-BY-n REAL BIDIAGONAL MATRIX bd
!               BY THE BISECTION-INVERSE ITERATION METHOD (eg PARTIAL SVD DECOMPOSITION).
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    abstol = sqrt( lamch('S') )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( d(n), e(n), s(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE AN UPPER BIDIAGONAL GRADED MATRIX bd.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
    a_is_upper   = true
!
    d(:n)      = geop( one, two, n )
    e(1_i4b)   = zero
    e(2_i4b:n) = d(:n-1_i4b)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-n BIDIAGONAL MATRIX bd
!   IS WRITTEN
!
!                       bd = u * s * v**(t)
!
!   WHERE s IS AN n-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   n DIAGONAL ELEMENTS, u AND v ARE n-BY-n ORTHOGONAL MATRICES.
!   THE DIAGONAL ELEMENTS OF s ARE THE SINGULAR VALUES OF bd; THEY ARE
!   REAL AND NON-NEGATIVE.
!   THE COLUMNS OF u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF bd.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE THE ls LARGEST SINGULAR VALUES OF bd AND
!   THE ASSOCIATED LEFT AND RIGHT SINGULAR VECTORS (EG A PARTIAL SVD DECOMPOSITION OF bd)
!   IN TWO STEPS:
!
!   STEP1 : COMPUTE THE ls LARGEST SINGULAR VALUES OF bd BY BISECTION
!   WITH SUBROUTINE bd_singval2 :
!
    call bd_singval2( d(:n), e(:n), nsing, s(:n), failure=failure1, sort=sort, abstol=abstol, ls=ls )
!
!   ON EXIT OF bd_singval2 :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT
!   failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                      THAT FULL ACCURACY WAS NOT ATTAINED IN THE
!                      COMPUTATION OF THE SINGULAR VALUES OF bd.
!
!   nsing GIVES THE NUMBER OF SINGULAR VALUES REALLY COMPUTED BY bd_singval2. nsing MAY BE GREATER
!   THAN ls IF CLUSTERS OF SINGULAR VALUES ARE PRESENT.
!   THE nsing LARGEST SINGULAR VALUES ARE STORED IN THE FIRST nsing POSITIONS OF s IN DECREASING ORDER IF sort='d'.
!
!   STEP2 : COMPUTE THE nsing ASSOCIATED SINGULAR VALUES OF bd BY INVERSE ITERATION WITH SUBROUTINE
!   bd_inviter :
!
    if ( nsing>0 ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE SINGULAR VECTORS.
!
        allocate( leftvec(n,nsing),    &
                  rightvec(n,nsing),   &
                  stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE FIRST nsing SINGULAR VECTORS OF a BY maxiter INVERSE ITERATIONS WITH
!       SUBROUTINE bd_inviter .
!
        call bd_inviter( a_is_upper, d(:n), e(:n), s(:nsing), leftvec(:n,:nsing), rightvec(:n,:nsing),  &
                         failure=failure2, maxiter=maxiter )
!
!       ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX bd .
!       THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
!       ON EXIT OF bd_inviter :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE
!
!       THE SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd ARE COMPUTED USING maxiter INVERSE ITERATIONS
!       FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF bd ARE THEN ORTHOGONALIZED BY 
!       THE MODIFIED GRAM-SCHMIDT ALGORITHM IF THE SINGULAR VALUES ARE NOT WELL SEPARATED.
!       ON EXIT, leftvec AND rightvec CONTAIN THE FIRST nsing SINGULAR VECTORS OF bd .
!
!       bd_inviter MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!       IDENTICAL.
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. nsing>0 ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( id(nsing,nsing), resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION bd*v - u*singval(:nsing),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF bd .
!
        if ( a_is_upper ) then
!
            resid(:n,:nsing) = spread( d(:n),         dim=2, ncopies=nsing )*rightvec                 + &
                               eoshift( spread(e(:n), dim=2,ncopies=nsing)*rightvec, shift=1 )        - &
                               spread( s(:nsing),     dim=1, ncopies=n )*leftvec 

        else
!
            resid(:n,:nsing) = spread( d(:n),         dim=2, ncopies=nsing )*leftvec                   + &
                               eoshift( spread(e(:n), dim=2,ncopies=nsing)*leftvec, shift=1 )          - &
                               spread( s(:nsing),     dim=1, ncopies=n )*rightvec
!
        end if
!
        id(:nsing,1_i4b) = norm( resid(:n,:nsing), dim=2_i4b )
        err1 = maxval( id(:nsing,1_i4b) )/( sum( s(:nsing) )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are LEFT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX a.
!
        call unit_matrix( id )
!
        resid(:nsing,:nsing) = abs( id - matmul( transpose( leftvec ), leftvec ) )
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - rightvec**(t)*rightvec
!       WHERE rightvec ARE THE RIGHT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        resid(:nsing,:nsing) = abs( id - matmul( transpose( rightvec ), rightvec ) )
        err3 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( id, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( nsing>0 ) then
        deallocate( d, e, s, leftvec, rightvec )
    else
        deallocate( d, e, s )
    end if
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. nsing>0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', nsing, ' singular values and vectors of a ', &
       n, ' by ', n,' real bidiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_singval2
! ==============================
!
end program ex1_bd_singval2
