program ex1_bd_svd2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_SVD2
!   in module SVD_Procedures .
!                                                                              
!                                                     
! LATEST REVISION : 23/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, one, c1_m2, c900, c50, &
                         bd_svd2, unit_matrix, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of bd_svd2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diag, sup, sup2, singval
    real(stnd), allocatable, dimension(:,:) :: leftvec, rightvec, id, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, bd_is_upper, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FULL SVD OF A REAL BIDIAGONAL MATRIX BD USING THE GOLUB-REINSCH ALGORITHM
!               WITH LAPACK STYLE CONVENTION FOR STORING THE SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( diag(n), sup(n), sup2(n), singval(n),   &
              leftvec(n,n), rightvec(n,n), stat=iok   )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE AN UPPER BIDIAGONAL MATRIX BD.
!   THE DIAGONAL ELEMENTS ARE STORED IN diag .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN sup .
!
    bd_is_upper   = true
!
    sup(1_i4b)   = zero
!
!    diag(:n)     = half
!    sup(2_i4b:n) = one
!
!    diag(:n)     = c1_m2
!    sup(2_i4b:n) = c900
!
    call random_number( diag(:n) )
    call random_number( sup(2_i4b:n) )
!
!   MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
    singval(:n) = diag(:n)
    sup2(:n)    = sup(:n)
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE ALL SINGULAR VALUES AND VECTORS OF THE BIDIAGONAL MATRIX BD .
!
!   FIST INITALIZED THE LEFT AND RIGHT SINGULAR VECTORS TO THE IDENTITY MATRIX OF ORDER n.
!
    call unit_matrix( leftvec(:n,:n) )
    call unit_matrix( rightvec(:n,:n) )
!
!   bd_svd2 COMPUTES THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL
!   n-BY-n (UPPER OR LOWER) BIDIAGONAL MATRIX BD. THE SVD IS WRITTEN
!
!                       BD = U * S * V**(t)
!
!   WHERE S IS AN n-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN n-BY-n ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF BD; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF BD.
!
    call bd_svd2( bd_is_upper, singval(:n), sup2(:n), failure, leftvec(:n,:n), &
                  rightvec(:n,:n), sort=sort )
!
!   THE ROUTINE RETURNS THE SINGULAR VALUES, THE LEFT AND RIGHT
!   SINGULAR VECTORS. THE LEFT SINGULAR VECTORS ARE RETURNED COLUMNWISE,
!   BUT THE RIGHT SINGULAR VECTORS ARE RETURNED ROWWISE. THIS IS THE LAPACK
!   CONVENTION AND IS THE ONLY DIFFERENCE BETWEEN bd_svd2 AND bd_svd.
!
!   ON EXIT OF bd_svd2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE BIDIAGONAL
!                         SVD OF BD.
!
!       singval  IS OVERWRITTEN WITH THE SINGULAR VALUES OF BD.
!       leftvec  IS OVERWRITTEN WITH THE LEFT SINGULAR VECTORS OF BD  IF leftvec IS THE IDENTITY ON ENTRY.
!       rightvec IS OVERWRITTEN WITH THE RIGHT SINGULAR VECTORS OF BD IF rightvec IS THE IDENTITY ON ENTRY.
!
!   IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!   THE SINGULAR VECTORS ARE REARRANGED ACCORDINGLY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( id(n,n), resid(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE THE RIGHT SINGULAR VECTORS SO THAT THEY ARE STORED COLUMNWISE.
!
        resid(:n,:n) = transpose( rightvec(:n,:n) )
        rightvec(:n,:n) = resid(:n,:n)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION bd*rightvec - leftvec*singval(:n),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS.
!
        if ( bd_is_upper ) then
!
            resid(:n,:n) = spread( diag(:n),        dim=2, ncopies=n )*rightvec                 + &
                           eoshift( spread(sup(:n), dim=2,ncopies=n)*rightvec, shift=1 )        - &
                           spread( singval(:n), dim=1, ncopies=n )*leftvec 

        else
!
            resid(:n,:n) = spread( diag(:n),        dim=2, ncopies=n )*leftvec                   + &
                           eoshift( spread(sup(:n), dim=2,ncopies=n)*leftvec, shift=1 )          - &
                           spread( singval(:n), dim=1, ncopies=n )*rightvec
!
        end if
!
        id(:n,1_i4b) = norm( resid(:n,:n), dim=2_i4b )
        err1 = maxval( id(:n,1_i4b) )/( sum( singval(:n) )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are LEFT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        call unit_matrix( id )
!
        resid = abs( id - matmul( transpose( leftvec ), leftvec ) )
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!       WHERE v are RIGHT SINGULAR VECTORS OF THE BIDIAGONAL MATRIX bd.
!
        resid = abs( id - matmul( transpose( rightvec ), rightvec ) )
        err3 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( id, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( diag, sup, sup2, singval, leftvec, rightvec )
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and singular vectors of a ', &
       n, ' by ', n,' real bidiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_svd2
! ==========================
!
end program ex1_bd_svd2
