program ex1_chol_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines CHOL_CMP and CHOL_SOLVE
!   in module Lin_Procedures. 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, true, false, chol_cmp,   &
                         chol_solve, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=8000, m=n+10
!
    character(len=*), parameter :: name_proc='Example 1 of chol_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, d1, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, c
    real(stnd), dimension(:),   allocatable :: invdiag, b, x, res
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: do_test, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING A LINEAR SYSTEM WITH A REAL n-BY-n SYMMETRIC DEFINITE POSITIVE MATRIX
!               AND ONE RIGHT HAND-SIDE WITH THE CHOLESKY DECOMPOSITION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
    err = zero
!
    do_test = false
    upper   = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( c(m,n), a(n,n), b(n), x(n), invdiag(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM SYMMETRIC POSITIVE DEFINITE MATRIX a .
!
    call random_number( c )
!
    a = matmul( transpose(c), c )
!
!   GENERATE A n RANDOM SOLUTION VECTOR x .
!
    call random_number( x )
!
!   COMPUTE THE MATRIX-VECTOR PRODUCT b = a*x .
!
    b = matmul( a, x )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE SOLUTION VECTOR FOR SYMMETRIC POSITIVE DEFINITE
!   SYSTEM
!
!            a*x = b .
!
!   BY COMPUTING THE CHOLESKY DECOMPOSITION OF MATRIX a .
!   IF ON OUTPUT OF chol_cmp d1 IS DIFFERENT FROM ZERO
!   THEN THE SYMMETRIC LINEAR SYSTEM IS NOT SINGULAR
!   AND CAN BE SOLVED BY SUBROUTINE chol_solve.
!
    call chol_cmp( a, invdiag, d1, upper=upper )
!
    if ( d1==zero ) then
!
!       ANORMAL EXIT IN chol_cmp SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in exit of CHOL_CMP subroutine, d1=', d1
        write (prtunit,*)
!
    else
!
!       SOLVE THE SYMMETRIC LINEAR SYSTEM.
!
        call chol_solve( a, invdiag, b, upper=upper )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( d1/=zero .and. do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n) = b(:n) - x(:n)
        err     = norm(res) /( real(n,stnd)*norm(x) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, c, x, invdiag, res )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, c, x, invdiag )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1/=zero ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a positive definite symmetric linear system of size ', &
       n, ' by ', n, ' is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_chol_cmp
! ===========================
!
end program ex1_chol_cmp
