program ex1_gchol_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines GCHOL_CMP and CHOL_SOLVE
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, c10, true, false, gchol_cmp, &
                         chol_solve, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000, m=n-10
!
    real(stnd), parameter :: fudge=c10
!
    character(len=*), parameter :: name_proc='Example 1 of gchol_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, tol, d1, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, c
    real(stnd), dimension(:),   allocatable :: invdiag, b, b2, d, res
!
    integer(i4b) :: krank
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: do_test, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING A LINEAR SYSTEM WITH A REAL n-BY-n SYMMETRIC SEMI-DEFINITE POSITIVE
!               MATRIX AND ONE RIGHT HAND-SIDE WITH THE CHOLESKY DECOMPOSITION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    tol = sqrt( epsilon( err ) )
    eps = fudge*tol
    err = zero
!
    do_test = true
    upper   = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( c(m,n), a(n,n), b(n), invdiag(n), d(m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM SYMMETRIC POSITIVE SEMIDEFINITE MATRIX a .
!
    call random_number( c )
!
    a = matmul( transpose(c), c )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b BELONGING TO THE RANGE OF a.
!
    call random_number( d )
!
    b = matmul( transpose(c), d )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS .
!
        allocate(  a2(n,n), b2(n), res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE COEFFICIENT MATRIX AND RIGHT HAND-SIDE VECTOR .
!
        a2(:n,:n) = a(:n,:n)
        b2(:n)    = b(:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE SOLUTION VECTOR FOR SYMMETRIC POSITIVE SEMIDEFINITE
!   SYSTEM
!
!                     a*x = b .
!
!   BY COMPUTING THE CHOLESKY DECOMPOSITION OF MATRIX a .
!   IF ON OUTPUT OF gchol_cmp d1 IS GREATER OR EQUAL TO ZERO
!   THEN THE SYMMETRIC LINEAR SYSTEM CAN BE SOLVED BY 
!   SUBROUTINE chol_solve.
!
    call gchol_cmp( a, invdiag, krank, d1, tol=tol, upper=upper )
!
    if ( d1<zero ) then
!
!       ANORMAL EXIT FROM gchol_cmp SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in exit of GCHOL_CMP subroutine, d1=', d1
        write (prtunit,*)
!
    else
!
!       SOLVE THE SYMMETRIC LINEAR SYSTEM.
!
        call chol_solve( a, invdiag, b, upper=upper )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( d1>=zero .and. do_test ) then
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n) = b2(:n) - matmul( a2, b(:n) )
        err = norm(res) / ( real(n,stnd)*( norm(a2) + norm(b2) ) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, b, c, d, invdiag, a2, b2, res )
    else
        deallocate( a, b, c, d, invdiag )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1>=zero ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a positive semi-definite symmetric system of size ', &
       n, ' is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_gchol_cmp
! ============================
!
end program ex1_gchol_cmp
