program ex1_hp_coef
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines/functions HP_COEF
!   and SYMLIN_FILTER in module Time_Series_Procedures.
!                                                                              
! LATEST REVISION : 30/11/2007
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, zero, half, one, two, pi, true, false, merror, allocate_error, arth,   &
                         hp_coef, symlin_filter, init_fft, fft, end_fft
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE LENGTH OF THE TIME SERIES .
!
    integer(i4b), parameter :: prtunit=6, n=2000
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err, tmp, fch
    real(stnd), dimension(n)              :: y, y2, y3, freq, freqr, tmpvec
    real(stnd), dimension(:), allocatable :: coef
!
    complex(stnd), dimension(n)           :: yc
!
    integer(i4b) :: i, k, k1, k2, ph, nfilt, khalf, kmid
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of hp_coef'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM REAL NUMBER SEQUENCE OF LENGTH n .
!
    call random_number( y(:n) )
!
!   SAVE THE REAL RANDOM NUMBER ARRAY.
!
    y2(:n) = y(:n)
!
!   ph IS THE MAXIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   ph IS EXPRESSED IN NUMBER OF OBSERVATIONS, I.E. pc==32(96) SELECT
!   A PERIOD OF 8 YRS FOR QUARTERLY (MONTHLY) DATA.
!
    ph  = 32
!
!   COMPUTE THE CORRESPONDING CUTOFF FREQUENCY fch .
!
    fch = one/real( ph, stnd )
!
!   NOW SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS k FOR THE LANCZOS FILTER.
!
    i = ceiling( one/(half-fch) )
    k = max( i, ph+1 )
!
!   CHECK IF k IS ODD.
!
    if ( (k/2)*2==k  ) k = k + 1
!
!   ALLOCATE WORK ARRAY FOR THE FILTER COEFFICIENTS.
!
    allocate( coef(k), stat=iok )
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FUNCTION hp_coef COMPUTES THE K-TERM LEAST SQUARES APPROXIMATION TO AN
!   -IDEAL- HIGH-PASS FILTER WITH CUTOFF PERIODS PH (EG CUTOFF FREQUENCY 1/PH).
!
    coef(:k) = hp_coef( PH=ph, K=k )
!
!   SUBROUTINE symlin_filter PERFORMS A SYMMETRIC FILTERING OPERATION ON AN IMPUT
!   TIME SERIES (THE ARGUMENT VEC).
!
!   HERE symlin_filter FILTERS THE TIME SERIES .
!   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE LOST FROM EACH END OF THE SERIES SO THAT
!   NFILT (NFILT= size(VEC) - size(COEF) + 1) TIME OBSERVATIONS ARE RETURNED IN VEC(:NFILT)
!   AND THE REMAINING OBSERVATIONS ARE SET TO ZERO.
!
    call symlin_filter( VEC=y2(:n), COEF=coef(:k),  NFILT=nfilt )
!
!   NOW COMPUTE THE TRANSFERT FUNCTION freqr(:) OF THE LANCZOS FILTER AT THE FOURIER FREQUENCIES.
!
    kmid  = ( k + 1 )/2
    khalf = ( k - 1 )/2
!
    freqr(:n) = coef(kmid)
!
    tmp        = (two*pi)/real( n, stnd )
    freq(:n)   = arth( zero, tmp, n )
    tmpvec(:n) = zero
!
    do i = 1, khalf
        tmpvec(:n) = tmpvec(:n) + freq(:n)
        freqr(:n)  = freqr(:n)  + two*coef(kmid+i)*cos( tmpvec(:n) )
    end do
!
!   NOW, APPLY THE FILTER USING THE FFT AND THE CONVOLUTION THEOREM.
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    call init_fft( n )
!
!   TRANSFORM THE TIME SERIES.
!
    yc(:n) = cmplx( y(1:n), zero, kind=stnd )
!
    call fft( yc(:n), forward=true  )
!
!   MULTIPLY THE FOURIER TRANSFORM OF THE TIME SERIES
!   BY THE TRANSFERT FUNCTION OF THE FILTER.
!
    yc(:n) = yc(:n)*freqr(:n)
!
!   INVERT THE SEQUENCE BACK TO GET THE FILTERED TIME SERIES.
!
    call fft( yc(:n), forward=false )
!
    y3(:n) = real( yc(:n),  kind=stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft()
!
    deallocate( coef )
!
!   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
    k1    = kmid
    k2    = n - khalf
!
    err = maxval(abs(y3(k1:k2)-y2(:nfilt)))/maxval(abs(y(k1:k2)))
!
    if ( err<=sqrt(epsilon(err))  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_hp_coef
! ==========================
!
end program ex1_hp_coef
