program ex1_lapack_gels
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine GELS
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 24/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gels
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=2000, n=1000, p=min(m,n), k=max(m,n), nrhs=400
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of gels'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)              :: err, eps, elapsed_time, work2(1)
    real(stnd), allocatable :: work(:), a(:,:), a2(:,:), b(:,:), b2(:,:), res(:,:)
!
    integer      :: info, lwork, iok, istart, iend, irate, imax, itime, mrhs
!
    logical(lgl)   :: do_test
!
    character :: trans
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FULL RANK REAL LINEAR LEAST SQUARES PROBLEM AND SEVERAL RIGHT HAND-SIDES
!               USING THE QR OR QL DECOMPOSITION OF THE COEFFICENT MATRIX a . IT IS
!               ASSUMED THAT a IS OF FULL RANK.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   If trans='N' THE LINEAR SYSTEM INVOLVES a AND If trans='T' THE LINEAR
!   SYSTEM INVOLVES a**(t). IN BOTH CASES, THE RANK OF a IS ASSUMED TO BE
!   min(m,n) .
!
    trans = 'N'
!
    mrhs = merge( m, n, trans=='N' )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(k,nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM COEFFICIENT MATRIX a .
!
    call random_number( a(:m,:n) )
!
!   GENERATE A m-by-nrhs (trans='N') OR n-by-nrhs (trans='T')
!   REAL RANDOM RIGHT HAND-SIDE MATRIX b .
!
    call random_number( b(:mrhs,:nrhs) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), b2(mrhs,nrhs), res(mrhs,nrhs), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
!       SAVE RIGHT HAND SIDE MATRIX .
!
        b2(:mrhs,:nrhs) = b(:mrhs,:nrhs)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gels SUBROUTINE.
!
    lwork = -1
!
    call gels( trans, m, n, nrhs, a, m, b, k, work2, lwork, info )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1))
!
!       ALLOCATE WORK VARIABLE NEEDED BY gels SUBROUTINE.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE MINIMAL NORM SOLUTION TO THE LINEAR LEAST SQUARES PROBLEM .
!
        call gels( trans, m, n, nrhs, a, m, b, k, work, lwork, info )
!
!       THE ROUTINE RETURNS THE MINIMUM NORM SOLUTION OF THE LINEAR
!       LEAST SQUARE PROBLEM. HOWEVER, THE COEFFICIENT MATRIX a IS
!       ASSUMED OF FULL RANK.
!
!       ON EXIT OF gels :
!
!         info=     0 :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT the iTH ARGUMENT HAS AN ILLEGAL VALUE
!         info=i  > 0 :  INDICATES THAT the iTH DIAGONAL ELEMENT OF THE TRIANGULAR
!                        FACTOR OF a IS ZERO AND a DOES NOT HAVE FULL RANK.
!
!       ON EXIT OF gels :
!
!         a IS OVERWRITTEN BY DETAILS OF ITS QR (m>=n) OR LQ (m<n) FACTORIZATIONS.
!
!         b IS OVERWRITTEN BY THE n-by-nrhs (trans='N') OR
!           m-by-nrhs (trans='T') SOLUTION.
!
!           If m >= n AND trans='N', THE RESIDUAL SUM-OF-SQUARES
!           FOR THE SOLUTION IN THE i-th COLUMN IS GIVEN BY THE
!           SUM OF SQUARES OF ELEMENTS n+1:m IN THAT COLUMN.
!
!           If m < n AND trans='T', THE RESIDUAL SUM-OF-SQUARES
!           FOR THE SOLUTION IN THE i-th COLUMN IS GIVEN BY THE
!           SUM OF SQUARES OF ELEMENTS m+1:n IN THAT COLUMN.
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM gels SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GELS subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COLUMN (trans='N')
!       OR ROW (trans='T') VECTORS OF a .
!
        if ( trans=='N' ) then
            res = b2 - matmul( a2, b(:n,:nrhs) )
            err = maxval( sum(abs(matmul(transpose(a2),res)),   dim=1) )/ sum( abs(a2) )
        else
            res = b2 - matmul( transpose(a2), b(:m,:nrhs) )
            err = maxval( sum(abs(matmul(a2,res)),   dim=1) )/ sum( abs(a2) )
        end if
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, b, a2, b2, res )
    else
        deallocate( a, b )
    end if
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*)   &
          'Orthogonality of the residuals against the columns/rows of the coefficient matrix = ', err
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the solution of a linear least squares real problem of size ', &
       m, ' by ', n, ' with ', nrhs,' right hand sides is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_gels
! ==============================
!
end program ex1_lapack_gels
