program ex1_lapack_gesvdx
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of the driver subroutine GESVDX
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 04/12/2016
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gesvdx
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, unit_matrix, merror,  &
                         allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=2000, n=1000, k=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of gesvdx'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, elapsed_time, vl, vu
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: s, work
    real(stnd), allocatable, dimension(:,:) :: a, a2, u, vt
!
    integer                            :: info, lwork, il, iu, ns, iok,     &
                                          istart, iend, irate, imax, itime
    integer, allocatable, dimension(:) :: iwork
!
    logical(lgl)   :: do_test
!
    character      :: jobu, jobvt, range
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FULL OR PARTIAL SVD OF A REAL MATRIX USING AN ASSOCIATED EIGENVALUE PROBLEM
!               AND THE BISECTION AND INVERSE ITERATION METHODS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), u(m,k), vt(k,n), s(k), iwork(12*k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM DATA MATRIX .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(m,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   DETERMINE THE WORK TO DO AND HOW MANY SINGULAR TRIPLETS
!   WILL BE COMPUTED AND STORED.
!
!   IF jobu  = 'V', LEFT SINGULAR VECTORS WILL BE COMPUTED.
!   IF jobu  = 'N', LEFT SINGULAR VECTORS WILL NOT BE COMPUTED.
!   IF jobvt = 'V', RIGHT SINGULAR VECTORS WILL BE COMPUTED.
!   IF jobvt = 'N', RIGHT SINGULAR VECTORS WILL NOT BE COMPUTED.
!
    jobu  = 'V'
    jobvt = 'V'
!
!   range DETERMINED HOW MANY SINGULAR VALUES AND VECTORS WILL BE COMPUTED.
!
!   IF range='A' ALL SINGULAR VALUES WILL BE FOUND,
!   IF range='V' ALL SINGULAR VALUES IN THE HALF OPEN INTERVAL (vl,vu] WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH SINGULAR VALUES IN ASCENDING ORDER WILL BE FOUND.
!
!   IN EACH CASE, THE ASSOCIATED SINGULAR VECTORS WILL BE ALSO COMPUTED.
!
!    range = 'A'
    range = 'I'
!
    vl = zero
    vu = zero
!
    il = k - 100 + 1
    iu = k
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gesvdx SUBROUTINE.
!
    lwork = -1
!
    call gesvdx( jobu, jobvt, range, m, n, a, m, vl, vu, il, iu, ns,   &
                 s, u, m, vt, k, work2, lwork, iwork, info )
!
    if ( info==0 ) then
!
        lwork = int(work2(1))
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY gesvdx SUBROUTINE.
!
        allocate( work(lwork), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       gesvdx COMPUTES THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL
!       m-BY-n MATRIX a BY SOLVING AN ASSOCIATE EIGENVALUE PROBLEM. THE SVD
!       IS WRITTEN
!
!                       a = U * S * V**(t)
!
!       WHERE S IS AN m-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!       min(m,n) DIAGONAL ELEMENTS, U IS AN m-BY-m ORTHOGONAL MATRIX, AND
!       V IS AN n-BY-n ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!       ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!       THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!       THE ADVANTAGE OF gesvdx SUBROUTINE COMPARED TO OTHER SVD SUBROUTINES
!       AVAILABLE IN LAPACK (E.G. gesvd OR gesdd) IS THAT gesvdx IS ABLE TO COMPUTE
!       A PARTIAL SVD DECOMPOSITION OR SELECTED SINGULAR TRIPLETS OF THE INPUT
!       REAL MATRIX WHILE THE OTHER SUBROUTINES DO NOT OFFER THIS POSSIBILITY.
!       HOWEVER, BEWARE THAT THE gesvdx SUBROUTINE IS AVAILABLE ONLY IN LAPACK 3.6.0
!       AND ABOVE.
!
        call gesvdx( jobu, jobvt, range, m, n, a, m, vl, vu, il, iu, ns,   &
                     s, u, m, vt, k, work, lwork, iwork, info )
!
!       THE ROUTINE RETURNS THE SINGULAR VALUES, THE LEFT AND RIGHT
!       SINGULAR VECTORS. THE RIGHT SINGULAR VECTORS ARE RETURNED ROWWISE.
!
!       ON EXIT OF gesvdx :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!         info=i  > 0 :  INDICATES THAT i EIVENVECTORS FAILED TO CONVERGE.
!         info=n*2+1  :  INDICATES THAT AN INTERNAL ERROR OCCURRED.
!
!       ON EXIT OF gesvdx WITH THE VALUES SPECIFIED ABOVE FOR jobu ('V'),
!       jobvt ('V') AND range ('A'):
!
!         u IS OVERWRITTEN WITH THE FIRST  min(m,n)
!         COLUMNS OF U (THE LEFT SINGULAR VECTORS,
!         STORED COLUMNWISE);
!
!         vt IS OVERWRITTEN WITH THE FIRST min(m,n)
!         ROWS OF V**(t) (THE RIGHT SINGULAR VECTORS,
!         STORED ROWWISE);
!
!         s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a IN DECREASING ORDER.
!
!       FOR range='I' OR range='V', ns IS THE NUMBER OF SINGULAR TRIPLETS
!       COMPUTED AND STORED BY THE SUBROUTINE IN ARGUMENTS u, vt AND s.
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
 !
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM gesvdx SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GESVDX subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:ns) - U(:m,:ns)*S(:ns,:ns).
!
        a(:m,:ns)     = matmul(a2(:m,:n),transpose(vt(:ns,:n))) - u(:m,:ns)*spread(s(:ns),dim=1,ncopies=m)
        a2(:ns,1_i4b) = norm( a(:m,:ns), dim=2_i4b )
!
        err1          =  maxval( a2(:ns,1_i4b) )/( sum( abs(s(:ns)) )*real(m,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a2(:ns,:ns) )
!
        a(:ns,:ns) = abs( a2(:ns,:ns) - matmul( transpose(u(:m,:ns)), u(:m,:ns) ) )
!
        err2       = maxval( a(:ns,:ns) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
        a(:ns,:ns) = abs( a2(:ns,:ns) - matmul( vt(:ns,:n), transpose(vt(:ns,:n)) ) )
!
        err3 = maxval( a(:ns,:ns) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, s, u, vt, iwork, a2 )
    else
        deallocate( a, s, u, vt, iwork )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', ns,' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_gesvdx
! ================================
!
end program ex1_lapack_gesvdx
