program ex1_lapack_orgtr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines SYTRD and ORGTR
!   in LAPACK software . 
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine SYMTRID_QRI2 in module EIG_Procedures.
!                                                                              
! LATEST REVISION : 20/03/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : sytrd, orgtr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm,    &
                         unit_matrix, symtrid_qri2, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=2000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of orgtr'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: d, e, tau, work, resid2
    real(stnd), allocatable, dimension(:,:) :: a, a2, resid
!
    integer      :: info_trd, info_q, lwork, lwork_trd, lwork_q,   &
                    iok, istart, iend, irate, imax, itime
    integer(i4b) :: l
!
    logical(lgl) :: failure, do_test
!   
    character    :: uplo, sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND EIGENVECTORS OF A REAL SYMMETRIC
!               MATRIX USING THE QR METHOD BY COMBINING LAPACK AND
!               STATPACK SUBROUTINES.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), d(n), e(n), tau(n-1_i4b), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX AND
!   FROM IT A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid(n,n), resid2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE SYMMETRIC MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE EIGENVALUES AND EIGENVECTORS OF THE SELF-ADJOINT MATRIX a
!   WITH SUBROUTINES sytrd AND orgtr FROM LAPACK AND symtrid_qri2 FROM
!   STATPACK.
!
!   THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!   IS WRITTEN
!
!                       a = U * D * U**(t)
!
!   WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!   THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!   THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
!   FIRST REDUCE THE SYMMETRIC MATRX a TO TRIDIAGONAL FORM WITH
!   SUBROUTINES sytrd AND orgtr FROM LAPACK.
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo = 'U'
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR sytrd AND orgtr SUBROUTINES.
!
    lwork = -1
!
    call sytrd( uplo, n, a, n, d, e(:n-1_i4b), tau, work2, lwork, info=info_trd )
!
    lwork_trd = int(work2(1))
!
    call orgtr( uplo, n, a, n, tau, work2, lwork, info=info_q )
!
    lwork_q = int(work2(1))
!
    if ( min(info_trd,info_q)==0 ) then
!
        lwork = max( lwork_trd, lwork_q )
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY sytrd AND orgtr SUBROUTINES.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CALL sytrd AND orgtr TO REDUCE THE MATRIX a TO TRIDIAGONAL FORM
!
!                      a = Q*TRID*Q**(t)
!
!       WHERE Q IS ORTHOGONAL AND TRID IS A SYMMETRIC TRIDIAGONAL MATRIX.
!
        call sytrd( uplo, n, a, n, d, e(:n-1_i4b), tau, work(:lwork_trd), lwork_trd, info=info_trd )
!
!       ON EXIT OF sytrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE SYMMETRIC TRIDIAGONAL MATRIX, RESPECTIVELY.
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q IN ITS UPPER TRIANGLE IF uplo='U'.
!         tau CONTAINS THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q.
!
        call orgtr( uplo, n, a, n, tau, work(:lwork_q), lwork_q, info=info_q )
!
!       ON EXIT OF THIS orgtr CALL:
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         a IS OVERWRITTEN WITH THE n-BY-n ORTHOGONAL MATRIX Q.
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
!       NEXT COMPUTE ALL EIGENVALUES AND EIGENVECTORS OF THE REAL SYMMETRIC MATRIX a WITH
!       SUBROUTINE symtrid_qri2 FROM STATPACK, USING TRID AND Q COMPUTED FROM LAPACK SUBROUTINES.
!       THE EIGENVALUES OF a ARE ALSO THE EIGENVALUES OR THE TRIDIAGONAL MATRIX TRID.
!
        call symtrid_qri2( d(:n), e(:n), failure, a(:n,:n), sort=sort )
!
!       THE ROUTINE RETURNS THE EIGENVALUES AND THE EIGENVECTORS OF a.
!
!       ON EXIT OF symtrid_qri2:
!
!           failure= false :  INDICATES SUCCESSFUL EXIT.
!           failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                             THAT FULL ACCURACY WAS NOT ATTAINED IN THE EIGENDECOMPOSITION 
!                             OF THE INTERMEDIATE TRIDIAGONAL FORM TRID OF a.
!
!           a IS OVERWRITTEN WITH THE EIGENVECTORS, STORED COLUMNWISE;
!
!           d IS OVERWRITTEN WITH THE EIGENVALUES OF a.
!
!           IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!           IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!           THE EIGENVECTORS ARE REARRANGED ACCORDINGLY.
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( min(info_trd,info_q)/=0 ) then
!
!       ANORMAL EXIT FROM sytrd AND orgtr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the calls to SYTRD or ORGTR subroutines, Info=', min(info_trd,info_q)
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        resid(:n,:n) = matmul(a2(:n,:n),a(:n,:n)) - a(:n,:n)*spread(d,1,n)
        resid2(:n) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( resid2(:n) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a2(:n,:n) )
!
        resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose(a(:n,:n )), a(:n,:n) ) )
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, a2, resid, resid2, d, e, tau )
    else
        deallocate( a, d, e, tau )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. min(info_trd,info_q)==0 .and. .not.failure  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. min(info_trd,info_q)==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', n, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_orgtr
! ===============================
!
end program ex1_lapack_orgtr
