program ex1_lapack_spevx
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SPEVX
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 24/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : spevx
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, two, c50, norm,   &
                         merror, allocate_error, lamch, unit_matrix, triangle
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit = 6, n=3000, p=(n*(n+1))/2
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of spevx'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, abstol, vl, vu, elapsed_time
    real(stnd), dimension(:),   allocatable :: a_packed, w, work, resid2
    real(stnd), dimension(:,:), allocatable :: a, z, resid
!
    integer(i4b)                            :: info, i, il, iu, m, iok, istart, iend, irate, imax, itime
    integer(i4b), dimension(:), allocatable :: iwork, ifail
!
    logical(lgl)   :: do_test
!
    character  :: jobz, uplo, range
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF A REAL SYMMETRIC
!               MATRIX IN PACKED STORAGE USING THE BISECTION AND INVERSE ITERATION METHODS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    abstol = two*lamch('s')
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a_packed(p), z(n,n), w(n), work(8*n), iwork(5*n), ifail(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM SELF-ADJOINT MATRIX a IN PACKED FORM.
!
    call random_number( a_packed(:p) )
!
!   MAKE a POSITIVE DEFINITE ASSUMING THAT THE UPPER TRIANGLE OF
!   THE SELF-ADJOINT MATRIX IS STORED IN PACKED FORM.
!
    do i = 1_i4b, n
        a_packed(i+((i-1_i4b)*i/2_i4b)) = real( n, stnd )
    end do
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(n,n), resid(n,n), resid2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       UNPACKED AND SAVE RANDOM SELF-ADJOINT MATRIX a 
!       ASSUMING THAT THE UPPER TRIANGLE OF THE SELF-ADJOINT
!       MATRIX IS STORED IN PACKED FORM.
!
        a(:n,:n) = unpack( a_packed(:p), mask=triangle( true, n, n, extra=1_i4b), field=zero )
!
        do i = 1_i4b, n
             a(i+1_i4b:n,i) = a(i,i+1_i4b:n)
        end do
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST DETERMINE THE WORK TO DO.
!
!   IF jobz='N' COMPUTE EIGENVALUES ONLY,
!   IF jobz='V' COMPUTE EIGENVALUES AND EIGENVECTORS.
!
    jobz  = 'V'
!
!   IF range='A' ALL EIGENVALUES WILL BE FOUND,
!   IF range='V' ALL EIGENVALUES IN THE HALF OPEN INTERVAL (vl,vu] WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH EIGENVALUES IN ASCENDING ORDER WILL BE FOUND.
!
    range = 'A'
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo  = 'U'
!
    vl = zero
    vu = zero
!
    il = 1
    iu = n
!
!   COMPUTE EIGENVALUES AND EIGENVECTORS OF THE SELF-ADJOINT MATRIX a
!   IN PACKED STORAGE WITH SUBROUTINE spevx FROM LAPACK.
!   THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!   IS WRITTEN
!
!                       a = U * D * U**(t)
!
!   WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!   THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!   THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
    call spevx( jobz, range, uplo, n, a_packed, vl, vu, il, iu, abstol,   &
                m, w, z, n, work, iwork, ifail, info                      )
!
!   ON EXIT OF spevx :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!         info=i  > 0 :  INDICATES THAT i EIGENVECTORS FAILED TO CONVERGE AND
!                        THEIR INDICES ARE STORED IN ARRAY ifail.
!
!   ON EXIT OF spevx :
!
!         a_packed IS OVERWITTEN BY DETAILS OF ITS TRIDIAGONAL FACTORIZATION.
!
!         m GIVES THE TOTAL NUMBER OF EIGENVALUES/EIGENVECTORS FOUND.
!
!         z IS OVERWRITTEN WITH THE EIGENVECTORS
!         OF a (THE EIGENVECTORS ARE STORED COLUMNWISE).
!
!         w IS OVERWRITTEN WITH THE EIGENVALUES OF a IN ASCENDING ORDER,
!         CORRESPONDING TO THE EIGENVECTORS IN z .
!
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
   if ( info/=0 ) then
!
!       ANORMAL EXIT FROM spevx SUBROUTINE, PRINT A WARNING.
!
       write (prtunit,*) 'Error in the call to SPEVX subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        resid(:n,:n) = matmul(a(:n,:n),z(:n,:n)) - z(:n,:n)*spread(w,1,n)
        resid2(:n) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( resid2(:n) )/( norm( a )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a(:n,:n) )
!
        resid(:n,:n) = abs( a(:n,:n) - matmul( transpose(z(:n,:n)), z(:n,:n) ) )
        err2 = maxval( resid(:n,:n) )/real( n, stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a_packed, z, w, work, iwork, ifail, a, resid, resid2 )
    else
       deallocate( a_packed, z, w, work, iwork, ifail )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', n, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix in packed storage is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_spevx
! ===============================
!
end program ex1_lapack_spevx
