program ex1_lapack_syev
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYEV
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 21/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : syev
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm,   &
                         merror, allocate_error, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of syev'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), dimension(:),   allocatable :: w, work, resid2
    real(stnd), dimension(:,:), allocatable :: a, a2, resid
!
    integer        :: info, lwork, iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: do_test
!
    character      :: jobz, uplo
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF A REAL SYMMETRIC
!               MATRIX USING THE QR IMPLICIT METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS .
!
    allocate( a(n,n), w(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid(n,n), resid2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM SELF-ADJOINT MATRIX a .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   IF jobz='N' COMPUTE EIGENVALUES ONLY,
!   IF jobz='V' COMPUTE EIGENVALUES AND EIGENVECTORS.
!
    jobz = 'V'
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo = 'U'
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR syev SUBROUTINE.
!
    lwork = -1
!
    call syev( jobz, uplo, n, a, n, w, work2, lwork, info )
!
    if ( info==0 ) then
!
        lwork = int(work2(1))
!
!       ALLOCATE WORK VARIABLE NEEDED BY syev SUBROUTINE.
!
        allocate( work(lwork), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE EIGENVALUES AND EIGENVECTORS OF THE SELF-ADJOINT MATRIX a
!       WITH SUBROUTINE syev FROM LAPACK.
!       THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!       IS WRITTEN
!
!                       a = U * D * U**(t)
!
!       WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!       THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!       THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
        call syev( jobz, uplo, n, a, n, w, work, lwork, info )
!
!       ON EXIT OF syev :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!         info=i  > 0 :  INDICATES THAT THE ALGORITHM FAILED TO CONVERGE
!                        AND i OFF-DIAGONAL ELEMENTS OF AN INTERMEDIATE
!                        TRIDIAGONAL FORM OF a DID NOT CONVERGE TO ZERO.
!
!       ON EXIT OF syev :
!
!         a IS OVERWRITTEN WITH THE EIGENVECTORS OF a IF 
!         jobz='V' (THE EIGENVECTORS ARE STORED COLUMNWISE).
!
!         w IS OVERWRITTEN WITH THE EIGENVALUES OF a IN ASCENDING ORDER.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM syev SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to SYEV subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        resid(:n,:n) = matmul(a2(:n,:n),a(:n,:n)) - a(:n,:n)*spread(w,1,n)
        resid2(:n) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( resid2(:n) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a2(:n,:n) )
!
        resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose(a(:n,:n)), a(:n,:n) ) )
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, w, a2, resid, resid2 )
    else
        deallocate( a, w )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', n, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_syev
! ==============================
!
end program ex1_lapack_syev
