program ex1_lapack_sytrd
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines SYTRD and ORGTR
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 27/04/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : sytrd, orgtr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm,    &
                         unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of sytrd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: d, e, tau, work
    real(stnd), allocatable, dimension(:,:) :: a, a2, resid, trid
!
    integer      :: info_trd, info_q, lwork, lwork_trd, lwork_q,   &
                    iok, istart, iend, irate, imax, itime
    integer(i4b) :: l
!
    logical(lgl) :: do_test
!   
    character    :: uplo
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : TRIDIAGONAL REDUCTION OF A REAL SYMMETRIC MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), d(n), e(n-1_i4b), tau(n-1_i4b), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX AND
!   FROM IT A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), trid(n,n), resid(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE SYMMETRIC MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo = 'U'
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR sytrd AND orgtr SUBROUTINES.
!
    lwork = -1
!
    call sytrd( uplo, n, a, n, d, e, tau, work2, lwork, info=info_trd )
!
    lwork_trd = int(work2(1))
!
    call orgtr( uplo, n, a, n, tau, work2, lwork, info=info_q )
!
    lwork_q = int(work2(1))
!
    if ( min(info_trd,info_q)==0 ) then
!
        lwork = max( lwork_trd, lwork_q )
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY sytrd AND orgtr SUBROUTINES.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CALL sytrd AND orgtr TO REDUCE THE MATRIX a TO TRIDIAGONAL FORM
!
!                      a = Q*TRID*Q**(t)
!
!       WHERE Q IS ORTHOGONAL AND TRID IS A SYMMETRIC TRIDIAGONAL MATRIX.
!
        call sytrd( uplo, n, a, n, d, e, tau, work(:lwork_trd), lwork_trd, info=info_trd )
!
!       ON EXIT OF sytrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE SYMMETRIC TRIDIAGONAL MATRIX, RESPECTIVELY.
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q IN ITS UPPER TRIANGLE IF uplo='U'.
!         tau CONTAINS THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q.
!
        call orgtr( uplo, n, a, n, tau, work(:lwork_q), lwork_q, info=info_q )
!
!       ON EXIT OF THIS orgtr CALL:
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         a IS OVERWRITTEN WITH THE n-BY-n ORTHOGONAL MATRIX Q.
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( min(info_trd,info_q)/=0 ) then
!
!       ANORMAL EXIT FROM sytrd AND orgtr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the calls to SYTRD or ORGTR subroutines, Info=', min(info_trd,info_q)
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*Q - Q*TRID.
!
        trid(:n,:n) = zero
!
        do l = 1_i4b, n-1_i4b
            trid(l,l)       = d(l)
            trid(l,l+1_i4b) = e(l)
            trid(l+1_i4b,l) = e(l)
        end do
!
        trid(n,n) = d(n)
!
        resid(:n,:n) = matmul( a2(:n,:n), a(:n,:n)  )           &
                       - matmul( a(:n,:n), trid(:n,:n) )
!
        trid(:n,1_i4b) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( trid(:n,1_i4b) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
        call unit_matrix( a2(:n,:n) )
!
        resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose(a(:n,:n )), a(:n,:n) ) )
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, a2, trid, resid, d, e, tau )
    else
        deallocate( a, d, e, tau )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. min(info_trd,info_q)==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. min(info_trd,info_q)==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed tridiagonal decomposition a = Q*TRD*Q**(t) = ', err1
        write (prtunit,*) 'Orthogonality of the computed Q orthogonal matrix                   = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the tridiagonal reduction of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_sytrd
! ===============================
!
end program ex1_lapack_sytrd
