program ex1_lp_coef2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines/functions LP_COEF2
!   and SYMLIN_FILTER2 in module Time_Series_Procedures.
!                                                                              
! LATEST REVISION : 12/11/2007
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, true, merror, allocate_error,   &
                         lp_coef2, symlin_filter2
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE LENGTH OF THE TIME SERIES .
!
    integer(i4b), parameter :: prtunit=6, n=2001
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err
    real(stnd), dimension(n)              :: y, y2, y3
    real(stnd), dimension(:), allocatable :: coef
!
    integer(i4b) :: k, k1, k2, pl, khalf, kmid
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of lp_coef2'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM REAL NUMBER SEQUENCE OF LENGTH n .
!
    call random_number( y(:n) )
!
!   SAVE THE REAL RANDOM NUMBER ARRAY.
!
    y2(:n) = y(:n)
    y3(:n) = y(:n)
!
!   pl IS THE MINIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   pl IS EXPRESSED IN NUMBER OF OBSERVATIONS, I.E. pc==32(96) SELECT
!   A PERIOD OF 8 YRS FOR QUARTERLY (MONTHLY) DATA.
!
    pl  = 32
!
!   NOW SELECT THE NUMBER OF FILTER COEFFICIENTS k FOR THE HAMMING FILTER.
!
    k = pl + 1
    if ( (k/2)*2==k )  k = k + 1
!
!   ALLOCATE WORK ARRAY FOR THE FILTER COEFFICIENTS.
!
    allocate( coef(k), stat=iok )
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FUNCTION lp_coef2 COMPUTES THE K-TERM LEAST SQUARES APPROXIMATION TO AN
!   -IDEAL- LOW-PASS FILTER WITH CUTOFF PERIODS PL (EG CUTOFF FREQUENCY 1/PL).
!
    coef(:k) = lp_coef2( PL=pl, K=k )
!
!   SUBROUTINE symlin_filter2 PERFORMS A SYMMETRIC FILTERING OPERATION ON AN IMPUT
!   TIME SERIES (THE ARGUMENT VEC).
!
!   HERE symlin_filter2 FILTERS THE TIME SERIES .
!   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE AFFECTED BY END EFFECTS  FROM EACH END OF THE SERIES.
!
    call symlin_filter2( VEC=y2(:n), COEF=coef(:k) )
!
!   FILTER AGAIN THE TIME SERIES, BUT COMPUTE THE VALUES AT THE ENDS OF THE OUTPUT
!   BY ASSUMING THAT THE INPUT IS PART OF A PERIODIC SEQUENCE OF PERIOD n .
!
    call symlin_filter2( VEC=y3(:n), COEF=coef(:k), USEFFT=true )
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( coef )
!
!   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
    kmid  = ( k + 1 )/2
    khalf = ( k - 1 )/2
!
    k1    = kmid
    k2    = n - khalf
!
    err = maxval(abs(y3(k1:k2)-y2(k1:k2)))/maxval(abs(y(k1:k2)))
!
    if ( err<=sqrt(epsilon(err))  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_lp_coef2
! ===========================
!
end program ex1_lp_coef2
