program ex1_lu_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines LU_CMP and LU_SOLVE
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, true, false, lu_cmp, lu_solve, &
                         norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000
!
    character(len=*), parameter :: name_proc='Example 1 of lu_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, d1, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a
    real(stnd), dimension(:),  allocatable  :: b, x, res
!
    integer(i4b), dimension(:), allocatable :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLVING A LINEAR SYSTEM WITH A REAL n-BY-n MATRIX
!               AND ONE RIGHT HAND-SIDE WITH THE LU DECOMPOSITION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), b(n), x(n), ip(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
!   GENERATE A n RANDOM SOLUTION VECTOR x .
!
    call random_number( x )
!
!   COMPUTE THE MATRIX-VECTOR PRODUCT b = a*x .
!
    b = matmul( a, x )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE SOLUTION VECTOR FOR THE LINEAR SYSTEM
!
!            a*x = b .
!
!   BY COMPUTING THE LU DECOMPOSITION WITH PARTIAL PIVOTING AND
!   IMPLICIT ROW SCALING OF MATRIX a. IF ON OUTPUT OF lu_cmp
!   d1 IS DIFFERENT FROM ZERO THEN THE LINEAR SYSTEM IS NOT
!   SINGULAR AND CAN BE SOLVED BY SUBROUTINE lu_solve.
!
    call lu_cmp( a, ip, d1 )
!
    if ( d1==zero ) then
!
!       ANORMAL EXIT FROM lu_cmp SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in exit of LU_CMP subroutine, d1=', d1
        write (prtunit,*)
!
    else
!
!       SOLVE THE LINEAR SYSTEM.
!
        call lu_solve( a, ip, b )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( d1/=zero .and. do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( res(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n) = b(:n) - x(:n)
        err     = norm(res) /( real(n,stnd)*norm(x) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, x, ip, res )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, x, ip )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1/=zero ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the solution of a linear real system of size ', &
       n, ' by ', n, ' is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lu_cmp
! =========================
!
end program ex1_lu_cmp
