program ex1_lu_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines LU_CMP2
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, one, true, false,   &
                         lu_cmp2, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000
!
    character(len=*), parameter :: name_proc='Example 1 of lu_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, d1, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, ainv, res
!
    integer(i4b)                            :: j
    integer(i4b), dimension(:), allocatable :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE THE INVERSE OF A REAL n-BY-n MATRIX
!               BY USING THE LU DECOMPOSITION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), ainv(n,n), ip(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), res(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE A COPY OF THE MATRIX a.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE INVERSE OF A SQUARE REAL MATRIX BY COMPUTING
!   THE LU DECOMPOSITION WITH PARTIAL PIVOTING AND
!   IMPLICIT ROW SCALING OF MATRIX a. IF ON OUTPUT OF lu_cmp2
!   d1 IS DIFFERENT FROM ZERO THEN THE MATRIX IS NOT SINGULAR
!   AND THE INVERSE OF a HAS BEEN COMPUTED.
!
    call lu_cmp2( a, ip, d1, matinv=ainv )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( d1==zero ) then
!
!       ANORMAL EXIT FROM lu_cmp2 SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in exit of LU_CMP2 subroutine, d1=', d1
        write (prtunit,*)
!
    else if ( do_test ) then
!
!       COMPUTE a TIMES ITS INVERSE - IDENTITY.
!
        res = matmul( a2, ainv )
!
        do j = 1_i4b, n
            res(j,j) = res(j,j) - one
        end do
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        err = norm( res(:n,:n) ) /( real(n,stnd)*norm(a2(:n,:n)) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, ainv, ip, a2, res )
    else
        deallocate( a, ainv, ip )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1/=zero ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the inverse of a  real matrix of size ', &
       n, ' by ', n, ' is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lu_cmp2
! ==========================
!
end program ex1_lu_cmp2
