program ex1_phase_scramble_cor
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines COMP_COR and PHASE_SCRAMBLE_COR
!   in module Mul_Stat_Procedures .
!                                                                              
! LATEST REVISION : 26/06/2006
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, comp_cor, phase_scramble_cor, pinvn
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
! p       IS THE NUMBER OF OBSERVATIONS OF THE TIME SERIES VECTORS
! nrep    IS THE NUMBER OF SHUFFLES FOR THE PHASE-SCRAMBLED BOOTSTRAP TEST
! nsample IS THE NUMBER OF SAMPLES TO EVALUATE THE REJECTION RATE OF THE TEST
!
    integer(i4b), parameter :: prtunit=6, p=50, nrep=99, nsample=2000
!
! sign_level IS THE SIGNIFICANCE LEVEL OF PHASE-SCRAMBLED BOOTSTRAP TEST
! eps        IS THE ALLOWED RELATIVE ERROR FOR THE REJECTION RATE
! b          IS THE LAG-1 AUTOCORRELATION FOR THE AR(1) MODEL USED
!            TO GENERATE THE TIME SERIES
!
    real(stnd),  parameter :: sign_level=0.05, eps=0.2, b=0.2
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                      :: xyn, xycor, prob, err_prob
    real(stnd), dimension(2)        :: xstat, ystat
    real(stnd), dimension(p)        :: x, y, e
!
    integer(i4b) :: i, j, rej_rate
!
    logical(lgl) :: first, last
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of phase_scramble_cor'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   INITIALIZE THE RANDOM GENERATOR.
!
    call random_seed()
!
!   INITIALIZE THE REJECTION RATE FOR THE PERMUTATION TEST.
!
    rej_rate = 0
    first    = true
    last     = true
!
!   GENERATE A TIME SERIES USING AN AR(1) MODEL OF THE FORM
!
!               y(i+1) = b*y(i) + e(i)
!
!     WHERE b IS THE SPECIFIED LAG-1 AUTOCORRELATION AND e(I)
!     IS A NORMALLY DISTRIBUTED RANDOM VARIABLE WITH A 0 MEAN
!     AND A VARIANCE OF 1.
!
    call random_number( y(:p) )
    e(:p) = pinvn( y(:p) )
!
    y(1) = e(1)
    do j=2, p
        y(j) = b*y(j-1) + e(j)
    end do
!
    do i=1, nsample
!
!       GENERATE ANOTHER INDEPENDENT TIME SERIES FROM THE SAME AR(1) MODEL .
!
        call random_number( x(:p) )
        e(:p) = pinvn( x(:p) )
!
        x(1) = e(1)
        do j=2, p
            x(j) = b*x(j-1) + e(j)
        end do
!
!       COMPUTE THE CORRELATIONS BETWEEN x AND y
!       FOR THE p OBSERVATIONS .
!
        call comp_cor( x(:p), y(:p), first, last, xstat(:2), ystat(:2), xycor, xyn )
!
!       ON EXIT OF COMP_COR WHEN last=true :
!
!      xstat(1)     CONTAINS THE MEAN VALUE OF THE OBSERVATION VECTOR x(:p).
!
!      xstat(2)     CONTAINS THE VARIANCE OF THE OBSERVATION VECTOR x(:p).
!
!      ystat(1)     CONTAINS THE MEAN VALUE OF THE OBSERVATION VECTOR  y(:p).
!
!      ystat(2)     CONTAINS THE VARIANCE OF THE OBSERVATION VECTOR  y(:p).
!
!      xycor        CONTAINS THE CORRELATION COEFFICIENT BETWEEN x(:p) AND y(:p).
!
!      xyn          CONTAINS THE NUMBER OF OBSERVATIONS IN THE DATA VECTORS
!                   x(:p) AND  y(:p) (xyn=real(p,stnd) ).
!
!
!       NOW COMPUTE A THE PHASE-SCRAMBLED BOOTSTRAP TEST OF THE CORRELATION
!       BETWEEN x AND y WITH SUBROUTINE phase_scramble_cor WITH nrep SHUFFLES .
!
        call phase_scramble_cor( x(:p), y(:p), xstat(:2), ystat(:2), xycor, prob, nrep=nrep )
!
!       EVALUATE THE REJECTION RATE .
!
        if ( prob<=sign_level ) rej_rate = rej_rate + 1
!
    end do
!
!   NOW COMPUTE THE REJECTION RATE OF THE TEST. IDEALLY FOR THE sign_level
!   SIGNIFICANCE LEVEL, WE WOULD REJECT ONLY THE NULL HYPOTHESIS sign_level %
!   OF THE TIME. IN OTHER WORDS, THE REJECTION RATE MUST BE APPROXIMATELY EQUAL
!   TO THE SIGNIFICANCE LEVEL sign_level .
!
    prob = real( rej_rate, stnd )/real( nsample, stnd )
!
!   COMPUTE THE RELATIVE ERROR OF THE REJECTION RATE.
!
    err_prob = abs( (prob-sign_level)/sign_level )
!
!   CHECK THAT THE RELATIVE ERROR OF THE REJECTION RATE IS LESS THAN eps .
!
    if ( err_prob<=eps ) then
        write (prtunit,*) 'Example 1 of PHASE_SCRAMBLE_COR is correct'
    else
        write (prtunit,*) 'Example 1 of PHASE_SCRAMBLE_COR is incorrect'
    end if
!
!
! END OF PROGRAM ex1_phase_scramble_cor
! =====================================
!
end program ex1_phase_scramble_cor
