program ex1_real_fft
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine REAL_FFT
!   in module FFT_Procedures .
!                                                                              
! LATEST REVISION : 15/06/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, true, false, init_fft, fft, real_fft, end_fft, &
                         merror, allocate_error
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS AN EVEN POSITIVE INTEGER.
!
    integer(i4b), parameter :: prtunit=6, n=100000
!
    character(len=*), parameter :: name_proc='Example 1 of real_fft'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err, eps, elapsed_time
    real(stnd), dimension(:), allocatable :: y, y2
!
    complex(stnd), dimension(:), allocatable :: yt
!
    integer(i4b) :: nd2
    integer      :: iok, istart, iend, irate, imax, itime
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FORWARD FFT OF A REAL SEQUENCE.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( y(n), y2(n), yt(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM REAL NUMBER SEQUENCE OF EVEN LENGTH n .
!
    call random_number( y(:n) )
!
!   INITIALIZE THE REAL_FFT SUBROUTINE.
!
    nd2   = n/2
!
    call init_fft( nd2 )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   TRANSFORM THE REAL SEQUENCE.
!
    call real_fft( y(:n), yt(:nd2+1), forward=true  )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    call end_fft()
!
!   COMPUTE THE REMAINING VALUES OF THE FOURIER TRANSFORM.
!
    yt(n:nd2+2:-1) = conjg( yt(2:nd2) )
!
!   INITIALIZE THE FFT SUBROUTINE FOR THE BACKWARD TRANSFORM.
!
    call init_fft( n )
!
!   INVERT THE SEQUENCE BACK.
!
    call fft( yt(:n), forward=false )
!
    call end_fft()
!
!   CHECK THAT INVERSE(TRANSFORM(SEQUENCE)) = SEQUENCE.
!
    y2(:n) = real( yt(:n) )
!
    err = maxval(abs(y(:n)-y2(:n)))/maxval(abs(y(:n)))
!
!   DEALLOCATE ARRAYS.
!
    deallocate( y, y2, yt )
!
    if ( err<=eps  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (*,'(a,i8,a,0pd12.4,a)')    &
      'The elapsed time for computing the forward FFT of a real sequence of size ', &
       n, ' is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_real_fft
! ===========================
!
end program ex1_real_fft
