program ex1_select_eigval_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SELECT_EIGVAL_CMP
!   in module Eig_Procedures.
!                                                                              
! LATEST REVISION : 15/06/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, select_eigval_cmp, &
                         trid_inviter, merror, allocate_error, norm, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m IS THE NUMBER OF THE COMPUTED EIGENVALUES/EIGENVECTORS
!
    integer(i4b), parameter :: prtunit=6, n=2000, m=100
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of select_eigval_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, elapsed_time
    real(stnd), dimension(:), allocatable   :: d, res
    real(stnd), dimension(:,:), allocatable :: a, a2, eigvec, d_e
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=4
!
    logical(lgl) :: failure, failure2, do_test, small
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL EIGENVALUES/EIGENVECTORS DECOMPOSITION OF A n-BY-n REAL
!               SYMMETRIC MATRIX USING A RATIONAL QR ALGORITHM FOR THE EIGENVALUES
!               AND THE INVERSE ITERATION TECHNIQUE FOR THE EIGENVECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   DETERMINE IF YOU WANT TO COMPUTE THE m SMALLEST OR LARGEST EIGENVALUES.
!
    small = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), eigvec(n,m), d(m), d_e(n,2), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
        allocate( a2(n,n), res(m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE FIRST m LARGEST OR SMALLEST EIGENVALUES OF THE SELF-ADJOINT MATRIX a
!   AND SAVE THE INTERMEDIATE TRIDIAGONAL MATRIX IN PARAMETER d_e .
!
    call select_eigval_cmp( a(:n,:n), d(:m), small, failure, d_e=d_e )
!
    if ( .not. failure ) then
!
!       COMPUTE THE ASSOCIATED m EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!       maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION.
!
        call trid_inviter( d_e(:n,1), d_e(:n,2), d(:m), eigvec(:n,:m), failure2,     &
                           mat=a(:n,:n), maxiter=maxiter )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. .not.failure ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec - eigvec*d
!       WHERE d ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF a.
!
        a(:n,:m) = matmul( a2(:n,:n), eigvec(:n,:m)) - eigvec(:n,:m)*spread( d(:m), dim=1, ncopies=n)
        res(:m) = norm( a(:n,:m), dim=2_i4b )
!
        err1 = maxval( res(:m) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec.
!
        call unit_matrix( a(:m,:m) )
!
        a2(:m,:m) = abs( a(:m,:m) - matmul( transpose(eigvec(:n,:m)), eigvec(:n,:m) ) )
!
        err2 = maxval( a2(:m,:m) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        deallocate( a, eigvec, d_e, d, a2, res )
!
    else
!
        deallocate( a, eigvec, d_e, d )
!
    end if
!
    if ( err<=eps .and. .not.failure  .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test  .and. .not.failure ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues/eigenvectors couplets = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors                 = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', m, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_select_eigval_cmp
! ====================================
!
end program ex1_select_eigval_cmp
