program ex1_svd_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SVD_CMP 
!   in module SVD_Procedures . 
!                                                                              
! LATEST REVISION : 01/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, svd_cmp, norm, &
                         unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit = 6, m=3000, n=3000, k=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of svd_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, ulp, elapsed_time
    real(stnd), allocatable, dimension(:)   :: s
    real(stnd), allocatable, dimension(:,:) :: a, u, v, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: failure, do_test
!   
    character      :: sort='a'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : FULL SVD OF A REAL MATRIX USING THE BIDIAGONAL QR IMPLICIT METHOD WITH
!               A WAVE-FRONT ALGORITHM FOR APPLYING GIVENS ROTATIONS IN THE BIDIAGONAL
!               QR ALGORITHM AND, OPTIONALLY, A PERFECT SHIFT FOR THE SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( u(m,n), v(n,k), s(k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n RANDOM DATA MATRIX .
!
    call random_number( u )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(m,n), resid(m,k), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        a(:m,:n) = u(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   svd_cmp COMPUTES THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL
!   m-BY-n MATRIX a. THE SVD IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN m-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN m-BY-m ORTHOGONAL MATRIX, AND
!   V IS AN n-BY-n ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   BY DEFAULT, svd_cmp USES A PERFECT SHIFT STRATEGY FOR THE SINGULAR VECTORS
!   SINCE IT IS USUALLY FASTER.
!   IF YOU DONT WANT TO USE THIS STRATEGY, YOU CAN SPECIFY THE OPTIONAL LOGICAL
!   PARAMETER perfect_shift WITH THE VALUE false.
!
    call svd_cmp( u, s, failure, v=v, sort=sort, max_francis_steps=10_i4b )
!
!   THE ROUTINE RETURNS THE SINGULAR VALUES AND THE FIRST min(m,n) LEFT
!   AND RIGHT SINGULAR VECTORS.
!
!   ON EXIT OF svd_cmp :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE BIDIAGONAL
!                         SVD OF AN INTERMEDIATE BIDIAGONAL FORM B OF a.
!
!       u IS OVERWRITTEN WITH THE FIRST min(m,n) LEFT SINGULAR VECTORS,
!       STORED COLUMNWISE;
!   
!       v CONTAINS THE FIRST min(m,n) RIGHT SINGULAR VECTORS,
!       STORED COLUMNWISE;
!
!       s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a.
!
!   IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!   THE SINGULAR VECTORS ARE REARRANGED ACCORDINGLY.
!                
!   IF THE PARAMETER v IS ABSENT, svd_cmp COMPUTES ONLY THE SINGULAR VALUES OF a             
!   AND, OPTIONALLY, STORES THE INTERMEDIATE BIDIAGONAL FORM OF a AND THE ORTHOGONAL
!   MATRICES USED TO REDUCE a TO BIDIAGONAL FORM. SEE EXAMPLES ex2_svd_cmp.f90 OR 
!   ex1_bd_inviter2.f90, WHICH SHOW HOW TO COMPUTE A PARTIAL SVD, FOR MORE DETAILS.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!                
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:,:k) - U(:,:k)*S(:k,:k).
!
        resid(:m,:k) = matmul(a,v) - u(:,:k)*spread(s,dim=1,ncopies=m)
        a(:k,1_i4b) = norm( resid(:m,:k), dim=2_i4b )
        err1 =  maxval( a(:k,1_i4b) )/( sum( abs(s) )*real(m,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a(:k,:k) )
!
        resid(:k,:k) = abs( a(:k,:k) - matmul( transpose(u(:m,:k)), u(:m,:k) ) )
        err2 = maxval( resid(:k,:k) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
        resid(:k,:k) = abs( a(:k,:k) - matmul( transpose(v(:n,:k)), v(:n,:k) ) )
        err3 = maxval( resid(:k,:k) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( u, v, s )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_svd_cmp
! ==========================
!
end program ex1_svd_cmp
