program ex2_bd_deflate2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_DEFLATE2
!   in module SVD_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_CMP2 and BD_SINGVAL2 in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 09/07/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, lamch, bd_cmp2, bd_singval2,  &
                         bd_deflate2, norm, unit_matrix, c50, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, m=3000, nsing=100
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of bd_deflate2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, abstol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: s, d, e
    real(stnd), allocatable, dimension(:,:) :: a, a2, p, leftvec, rightvec, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: ns, max_qr_steps
!
    logical(lgl) :: failure1, failure2, failure3, ortho, do_test, gen_p
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : PARTIAL SVD OF A n-BY-m REAL MATRIX WITH n>=m USING
!               THE RALHA-BARLOW ONE-SIDED BIDIAGONALISATION ALGORITHM, 
!               A BISECTION ALGORITHM FOR SINGULAR VALUES AND THE GODUNOV
!               DEFLATION TECHNIQUE FOR SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    abstol = sqrt( lamch('S') )
    err    = zero
!
    gen_p = false
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,m), p(m,m), leftvec(n,nsing), rightvec(m,nsing),      &
              s(m), d(m), e(m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-m RANDOM REAL MATRIX a.
!
    call random_number( a )
!
    if ( do_test ) then
!
        allocate( a2(n,m), resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        a2(:n,:m) = a(:n,:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-m MATRIX a
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN n-BY-m MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX. THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE ALL THE SINGULAR VALUES
!   OF a AND nsing LEFT AND RIGHT SINGULAR VECTORS OF a IN THREE STEPS:
!
!   STEP1 : CALL bd_cmp2 TO REDUCE THE MATRIX a TO BIDIAGONAL FORM
!
!               a = Q*BD*P**(t)
!
!   WHERE Q AND P ARE ORTHOGONAL AND BD IS AN UPPER BIDIAGONAL MATRIX.
!   bd_cmp2 USES THE ONE-SIDED RALHA-BARLOW ALGORITM AND IS FASTER THAN
!   THE STANDARD BIDIAGONALIZATION ALGORITHM USED IN bd_cmp. HOWEVER, A
!   A SLIGHT LOSS OF ORTHOGONALITY CAN BE EXPECTED FOR Q WHEN a IS NEARLY
!   SINGULAR SINCE Q IS COMPUTED FROM A RECURRENCE RELATIONSHIP.
!
    call bd_cmp2( a(:n,:m), d(:m), e(:m), p(:m,:m), failure=failure1, gen_p=gen_p )
!
!   ON OUTPUT OF bd_cmp2:
!
!       a CONTAINS THE FIRST min(m,n) COLUMNS OF Q
!       AND p CONTAINS THE ORTHOGONAL MATRIX P IN FACTORED FORM IF gen_p=false
!       OR IN EXPLICIT FORM IF gen_p=true.
!
!       d AND e CONTAINS RESPECTIVELY, THE DIAGONAL AND
!       SUBDIAGONAL OF THE BIDIAGONAL MATRIX BD.
!
!       failure= false :  INDICATES THAT MAXIMUM ACCURACY WAS OBTAINED.
!       failure= true  :  INDICATES THAT a IS NEARLY SINGULAR AND SOME LOSS
!                         OF ORTHOGONALITY FOR Q CAN BE EXPECTED.
!
!   STEP2 : COMPUTE SINGULAR VALUES OF THE BIDIAGONAL MATRIX (WHICH ARE ALSO THE
!   SINGULAR VALUES OF a) WITH SUBROUTINE bd_singval2 TO HIGH RELATIVE PRECISION.
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d').
!
!   THE OPTIONAL ARGUMENT abstol OF bd_singval2 MAY BE USED TO INDICATE THE REQUIRED
!   PRECISION FOR THE SINGULAR VALUES. A SINGULAR VALUE IS CONSIDERED TO BE LOCATED
!   IF IT HAS BEEN DETERMINED TO LIE IN AN INTERVAL WHOSE WIDTH IS EQUAL OR LESS
!   THAN abstol. SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (EG sqrt(LAMCH('S')) ).
!
    call bd_singval2( d(:m), e(:m), ns, s(:m), failure=failure2, sort=sort, abstol=abstol )
!
!   ON EXIT OF bd_singval2 :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT
!         failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND THAT
!                           FULL  ACCURACY WAS NOT ATTAINED IN COMPUTING THE SINGULAR
!                           VALUES OF AN INTERMEDIATE BIDIAGONAL FORM BD OF a.
!
!         s IS OVERWRITTEN WITH THE SINGULAR VALUES OF THE BIDIAGONAL FORM BD OF a.
!
!         IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!         IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!
!   HERE, SORT = 'd' IS USED. THIS IS REQUIRED FOR THE USE OF bd_inviter2.
!
!   STEP3 : COMPUTE nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY A DEFLATION TECHNIQUE
!   APPLIED TO THE INTERMEDIATE BIDIAGONAL MATRIX AND BACK-TRANSFORMATION WITH SUBROUTINE
!   bd_deflate2.
!
!   ON ENTRY OF bd_deflate2: PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL
!   MATRIX (OR OF a). THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
!   OPTIONAL PARAMETER ortho DETERMINES IF DEFLATION IS APPLIED SEQUENTIALLY TO ALL
!   COMPUTED SINGULAR VECTORS (ortho=true) OR ONLY IN EACH CLUSTER OF SINGULAR VALUES
!   (ortho=false). IN THE LATTER CASE, ORTHOGONALITY OF THE SINGULAR VECTORS MAY BE
!   SLIGHTLY DEGRADED, BUT COMPUTATIONS WILL BE MUCH FASTER.
!
    ortho = false
    max_qr_steps = 4_i4b
!
    call bd_deflate2( a(:n,:m), p(:m,:m), d(:m), e(:m), s(:nsing),                 &
                      leftvec(:n,:nsing), rightvec(:m,:nsing), failure=failure3,   &
                      ortho=ortho, max_qr_steps=max_qr_steps                       )
!
!   ON EXIT OF bd_deflate2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE
!                         DEFLATION ALGORITHM.

!       leftvec AND rightvec CONTAIN, RESPECTIVELY, THE FIRST nsing LEFT AND RIGHT
!       SINGULAR VECTORS OF a .
!
!   bd_deflate2 MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!   IDENTICAL FOR SOME PATHOLOGICAL MATRICES.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
!
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:m,:nsing) - U(:n,:nsing)*S(:nsing,:nsing),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:n,:nsing) = matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n)
        a2(:nsing,1_i4b) = norm( resid(:n,:nsing), dim=2_i4b )
!
        err1 = maxval( a2(:nsing,1_i4b) )/ ( sum( abs(s(:m)) )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U(:n,:nsing)**(t)*U(:n,:nsing).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(leftvec), leftvec ) )
!
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V(:m,:nsing)**(t)*V(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, p, leftvec, rightvec, s, d, e )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure2 .and. .not.failure3 ) then
!    if ( err<=eps .and. .not.failure1 .and. .not.failure2 .and. .not.failure3 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_bd_deflate2
! ==============================
!
end program ex2_bd_deflate2
