program ex2_bd_inviter
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_INVITER
!   in module SVD_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_CMP2 and BD_SVD in module SVD_Procedures
!    and subroutine APPLY_Q_QR in module QR_Procedures.
!
! LATEST REVISION : 10/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, one, c100, bd_inviter, bd_svd, &
                         bd_cmp2, unit_matrix, norm, merror, allocate_error, apply_q_qr
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, m=3000, n=2000, nsing=1000
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='Example 2 of bd_inviter'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, sup, singval
    real(stnd), allocatable, dimension(:,:) :: a, a2, p, resid, leftvec, rightvec, leftvec0, rightvec0
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=2
!
    logical(lgl) :: failure1, failure2, failure3, bd_is_upper, gen_p, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : PARTIAL SVD OF A REAL m-BY-n MATRIX USING THE Ralha-Barlow ONE_SIDED ALGORITHM,
!               THE GOLUB-REINSCH ALGORITHM FOR ALL SINGULAR VALUES AND THE INVERSE ITERATION
!               TECHNIQUE FOR SELECTED SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    bd_is_upper = true
    gen_p       = false
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    if ( gen_p ) then
!
        allocate( a(m,n), p(n,n), d(n), e(n), singval(n),           &
                  sup(n), leftvec(m,nsing), rightvec(n,nsing),      &
                  leftvec0(n,nsing), rightvec0(n,nsing), stat=iok   )
!
    else
!
        allocate( a(m,n), p(n,n), d(n), e(n), singval(n),           &
                  sup(n), leftvec(m,nsing), rightvec(n,nsing),      &
                  leftvec0(n,nsing), stat=iok   )
!
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX .
!
    call random_number( a )
!
    if ( do_test ) then
!
        allocate( a2(m,n), resid(m,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST CALL bd_cmp2 TO REDUCE THE MATRIX a TO BIDIAGONAL FORM
!
!               a = Q*BD*P**(t)
!
!   WHERE Q AND P ARE ORTHOGONAL AND BD IS AN UPPER BIDIAGONAL MATRIX.
!   bd_cmp2 USES THE ONE_SIDED RHALA-BARLOW ALGORITM AND IS FASTER THAN
!   THE STANDARD BIDIAGONALIZATION ALGORITHM USED IN bd_cmp. HOWEVER, A
!   A SLIGHT LOSS OF ORTHOGONALITY CAN BE EXPECTED FOR Q SINCE Q IS
!   COMPUTED FROM A RECURRENCE RELATIONSHIP.
!
    call bd_cmp2( a, d, e, p, failure=failure1, gen_p=gen_p )
!
!   ON OUTPUT OF bd_cmp2:
!
!       a CONTAINS THE FIRST min(n,m) COLUMNS OF Q
!       AND p CONTAINS THE ORTHOGONAL MATRIX P IN FACTORED FORM IF gen_p=false
!       OR IN EXPLICIT FORM IF gen_p=true.
!
!       d AND e CONTAINS RESPECTIVELY, THE DIAGONAL AND
!       SUBDIAGONAL OF THE BIDIAGONAL MATRIX BD.
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT a IS NEARLY SINGULAR AND SOME LOSS
!                         OF ORTHOGONALITY FOR Q CAN BE EXPECTED.
!
!   NEXT COMPUTE ALL SINGULAR VALUES OF THE BIDIAGONAL MATRIX BD .
!   THE SINGULAR VALUES ARE STORED IN singval IN DECREASING ORDER (sort='d').
!
    singval(:n) = d(:n)
    sup(:n)     = e(:n)
!
    call bd_svd( bd_is_upper, singval(:n), sup(:n), failure=failure2, sort=sort  )
!
!   ON EXIT OF bd_svd :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE BIDIAGONAL QR ALGORITHM.
!
!   NOW COMPUTE THE FIRST nsing SINGULAR VECTORS OF BD BY maxiter INVERSE ITERATIONS WITH
!   SUBROUTINE bd_inviter. THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    if ( gen_p ) then
!
        call bd_inviter( bd_is_upper, d(:n), e(:n), singval(:nsing), leftvec0(:n,:nsing), rightvec0(:n,:nsing),  &
                         failure=failure3, maxiter=maxiter )
!
    else
!
        call bd_inviter( bd_is_upper, d(:n), e(:n), singval(:nsing), leftvec0(:n,:nsing), rightvec(:n,:nsing),  &
                         failure=failure3, maxiter=maxiter )
!
    end if
!
!   ON EXIT OF bd_inviter :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM.
!
!         leftvec AND rightvec CONTAIN, RESPECTIVELY, THE nsing LEFT AND RIGHT SINGULAR
!         VECTORS OF THE BIDIAGONAL MATRIX BD ASSOCIATED WITH THE SINGULAR VALUES singval(:nsing).
!
!   bd_inviter MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER singval ARE NEARLY
!   IDENTICAL FOR SOME PATHOLOGICAL BIDIAGONAL MATRICES.
!
!   FINALLY COMPUTE SINGULAR VECTORS OF THE ORIGINAL MATRIX BY MULTIPLICATION OR BACK-TRANSFORMATION.
!
    if ( gen_p ) then
!
        rightvec(:n,:nsing) = matmul( p(:n,:n), rightvec0(:n,:nsing) )
!
    else
!
        call apply_q_qr( p(2_i4b:n,2_i4b:n), p(2_i4b:n,1_i4b), rightvec(2_i4b:n,:nsing),  &
                         left=true, trans=false  )
!
    end if
!
    leftvec(:m,:nsing) = matmul( a(:m,:n), leftvec0(:n,:nsing) )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:n,:nsing) - U(:m,:nsing)*S(:nsing,:nsing),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:m,:nsing) = matmul(a2,rightvec) - leftvec*spread(singval(:nsing),dim=1,ncopies=m)
        a2(:nsing,1_i4b) = norm( resid(:m,:nsing), dim=2_i4b )
!
        err1 = maxval( a2(:nsing,1_i4b) )/( sum( abs(singval(:n)) )*real(m,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U(:n,:nsing)**(t)*U(:n,:nsing).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(leftvec), leftvec ) )
!
        err2 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V(:m,:nsing)**(t)*V(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( gen_p ) then
!
        deallocate( a, p, leftvec, rightvec, leftvec0, rightvec0, singval, d, e, sup )
!
    else
!
        deallocate( a, p, leftvec, rightvec, leftvec0, singval, d, e, sup )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure2 .and. .not.failure3 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_bd_inviter
! =============================
!
end program ex2_bd_inviter
