program ex2_bd_inviter2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_INVITER2
!   in module SVD_Procedures.
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_CMP2 and BD_SVD in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 06/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, true, false, bd_inviter2, bd_cmp2, bd_svd, &
                         norm, unit_matrix, c50, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, m=2000, nsing=100
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of bd_inviter2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: s, d, e, e2
    real(stnd), allocatable, dimension(:,:) :: a, a2, p, leftvec, rightvec, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=2
!
    logical(lgl) :: failure1, failure2, failure3, do_test, gen_p
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : ALL SINGULAR VALUES AND nsing SINGULAR VECTORS OF A n-BY-m REAL MATRIX WITH n>=m
!               USING THE RALHA-BARLOW ONE-SIDED BIDIAGONALISATION ALGORITHM, THE BIDIAGONAL QR
!               ALGORITHM FOR THE SINGULAR VALUES AND THE INVERSE ITERATION METHOD FOR THE 
!               SINGULAR VECTORS (EG PARTIAL SVD DECOMPOSITION).
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    gen_p = false
!
    do_test = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,m), p(m,m), leftvec(n,nsing), rightvec(m,nsing),      &
              s(m), d(m), e(m), e2(m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-m RANDOM REAL MATRIX a.
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,m), resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:m) = a(:n,:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-m MATRIX a
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN n-BY-m MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE ALL THE SINGULAR VALUES OF a AND
!   ONLY nsing LEFT AND RIGHT SINGULAR VECTORS OF a (EG A PARTIAL SVD DECOMPOSITION
!   OF a) IN THREE STEPS:
!
!   STEP1 : CALL bd_cmp2 TO REDUCE THE MATRIX a TO BIDIAGONAL FORM
!
!               a = Q*BD*P**(t)
!
!   WHERE Q AND P ARE ORTHOGONAL AND BD IS AN UPPER BIDIAGONAL MATRIX.
!   bd_cmp2 USES THE ONE-SIDED RHALA-BARLOW ALGORITM AND IS FASTER THAN
!   THE STANDARD BIDIAGONALIZATION ALGORITHM USED IN bd_cmp. HOWEVER, A
!   A SLIGHT LOSS OF ORTHOGONALITY CAN BE EXPECTED FOR Q WHEN a IS NEARLY
!   SINGULAR SINCE Q IS COMPUTED FROM A RECURRENCE RELATIONSHIP.
!
    call bd_cmp2( a(:n,:m), d(:m), e(:m), p(:m,:m), failure=failure1, gen_p=gen_p )
!
!   ON OUTPUT OF bd_cmp2:
!
!       a CONTAINS THE FIRST min(m,n) COLUMNS OF Q
!       AND p CONTAINS THE ORTHOGONAL MATRIX P IN FACTORED FORM IF gen_p=false
!       OR IN EXPLICIT FORM IF gen_p=true.
!
!       d AND e CONTAINS RESPECTIVELY, THE DIAGONAL AND
!       SUBDIAGONAL OF THE BIDIAGONAL MATRIX BD.
!
!       failure= false :  INDICATES THAT MAXIMUM ACCURACY WAS OBTAINED.
!       failure= true  :  INDICATES THAT a IS NEARLY SINGULAR AND SOME LOSS
!                         OF ORTHOGONALITY FOR Q CAN BE EXPECTED.
!
!   STEP2 : COMPUTE ALL SINGULAR VALUES OF THE BIDIAGONAL MATRIX BD WITH SUBROUTINE bd_svd.
!
!   FIRST MAKE A COPY OF THE BIDIAGONAL MATRIX BD FOR LATER USE WITH bd_inviter2 SUBROUTINE.
!
    s(:m)  = d(:m)
    e2(:m) = e(:m)
!
    call bd_svd( true, s(:m), e2(:m), failure=failure2, sort=sort  )
!
!   ON EXIT OF bd_svd :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT
!         failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                           THAT FULL ACCURACY WAS NOT ATTAINED IN THE BIDIAGONAL
!                           SVD OF AN INTERMEDIATE BIDIAGONAL FORM BD OF a.
!
!         s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a.
!
!         IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!         IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!
!   HERE, SORT = 'd' IS USED. THIS IS REQUIRED FOR THE USE OF bd_inviter2.
!
!   STEP3 : COMPUTE THE FIRST nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY maxiter
!   INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX AND BACK-TRANSFORMATION
!   WITH SUBROUTINE bd_inviter2 .
!
!   ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX d_e (OR a).
!   THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    call bd_inviter2( a, p, d, e, s(:nsing), leftvec, rightvec, failure=failure3, maxiter=maxiter )
!
!   ON EXIT OF bd_inviter2 :
!
!         failure= false :  INDICATES SUCCESSFUL EXIT.
!         failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE FOR SOME SINGULAR VECTORS.
!
!   THE SINGULAR VECTORS OF THE BIDIAGONAL FORM ARE COMPUTED USING maxiter INVERSE ITERATIONS
!   FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF ARE THEN ORTHOGONALIZED BY 
!   THE MODIFIED GRAM-SCHMIDT ALGORITHM IF NECESSARY. THE SINGULAR VECTORS OF a ARE FINALLY COMPUTED 
!   BY BACK TRANSFORMATIONS. ON EXIT, leftvec AND rightvec CONTAIN THE FIRST nsing LEFT AND RIGHT
!   SINGULAR VECTORS OF a, RESPECTIVELY.
!
!   NOTE THAT bd_inviter2 MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!   IDENTICAL FOR SOME PATHOLOGICAL MATRICES.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:m,:nsing) - U(:n,:nsing)*S(:nsing,:nsing),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:n,:nsing) = matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n)
        a2(:nsing,1_i4b) = norm( resid(:n,:nsing), dim=2_i4b )
        err1 = maxval( a2(:nsing,1_i4b) )/( sum( abs(s(:n)) )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U(:n,:nsing)**(t)*U(:n,:nsing).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(leftvec), leftvec ) )
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V(:m,:nsing)**(t)*V(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
        err3 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, p, leftvec, rightvec, s, d, e, e2 )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure2 .and. .not.failure3 ) then
!    if ( err<=eps .and. .not.failure1 .and. .not.failure2 .and. .not.failure3 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_bd_inviter2
! ==============================
!
end program ex2_bd_inviter2
