program ex2_comp_triang_inv
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine COMP_TRIANG_INV
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, one, true, false, allocate_error,   &
                         triangle, norm, comp_triang_inv, merror
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000, p=n*(n+1)/2
!
    character(len=*), parameter :: name_proc='Example 2 of comp_triang_inv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, res
    real(stnd), dimension(:),   allocatable :: ap
!
    integer(i4b) :: j
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : COMPUTE IN PLACE THE INVERSE OF A REAL n-BY-n TRIANGULAR MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
    err = zero
!
    do_test = true
    upper   = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), ap(p), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM TRIANGULAR MATRIX IN PACKED FORM ap .
!
    call random_number( ap )
!
!   MAKE SURE THAT TRIANGULAR MATRIX IS NOT SINGULAR.
!
    ap = ap + real( n, stnd )
!
!   UNPACK THE TRIANGULAR MATRIX a .
!
    a = unpack( ap, mask=triangle(upper,n,n,extra=1_i4b), field=zero )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), res(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE TRIANGULAR MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE IN PLACE THE INVERSE OF a WITH SUBROUTINE comp_triang_inv.
!   THE INPUT ARGUMENT IS OVERWRITTEN.
!
    call comp_triang_inv( a, upper=upper )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       COMPUTE a TIMES ITS INVERSE - IDENTITY.
!
        res(:n,:n) = matmul( a(:n,:n), a2(:n,:n) )
!
        do j = 1_i4b, n
            res(j,j) = res(j,j) - one
        end do
!
        err = norm( res(:n,:n) ) /( real(n,stnd)*norm(a2(:n,:n)) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, ap, a2, res )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, ap )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing in place the inverse of a real triangular matrix of size ', &
       n, ' by ', n, ' is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_comp_triang_inv
! ==================================
!
end program ex2_comp_triang_inv
