program ex2_gchol_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines GCHOL_CMP and CHOL_SOLVE
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 18/09/2014
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, c10, true, false, gchol_cmp, chol_solve, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=1000, m=n-1, nrhs=100
!
    real(stnd), parameter :: fudge=c10
!
    character(len=*), parameter :: name_proc='Example 2 of gchol_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, tol, d1, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, c, b, b2, d, res
    real(stnd), dimension(:), allocatable   :: invdiag
!
    integer(i4b) :: krank
    integer      :: iok, istart, iend, irate
!
    logical(lgl)   :: do_test, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : REAL SYMMETRIC DEFINITE POSITIVE MATRIX AND SEVERAL RIGHT HAND-SIDES.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    tol = sqrt( epsilon( err ) )
    eps = fudge*tol
    err = zero
!
    do_test = true
    upper   = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( c(m,n), a(n,n), b(n,nrhs), invdiag(n), d(m,nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM SYMMETRIC POSITIVE SEMIDEFINITE MATRIX a .
!
    call random_number( c )
!
    a = matmul( transpose(c), c )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE MATRIX b BELONGING TO THE RANGE OF a.
!
    call random_number( d )
!
    b = matmul( transpose(c), d )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS .
!
        allocate(  a2(n,n), b2(n,nrhs), res(n,nrhs), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF COEFFICIENT MATRIX AND RIGHT HAND-SIDE MATRIX .
!
        a2 = a
        b2 = b
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count=istart, count_rate=irate )
!
!   COMPUTE THE SOLUTION MATRIX FOR SYMMETRIC POSITIVE SEMIDEFINITE
!   SYSTEM
!
!                     a*x = b .
!
!   BY COMPUTING THE CHOLESKY DECOMPOSITION OF MATRIX a .
!   IF ON OUTPUT OF gchol_cmp d1 IS GREATER OR EQUAL TO ZERO
!   THEN THE SYMMETRIC LINEAR SYSTEM CAN BE SOLVED BY 
!   SUBROUTINE chol_solve.
!
    call gchol_cmp( a, invdiag, krank, d1, tol=tol, upper=upper )
!
    if ( d1<zero ) then
!
!       ANORMAL EXIT FROM gchol_cmp SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GCHOL_CMP subroutine, d1=', d1
!
    else
!
        call chol_solve( a, invdiag, b, upper=upper )
!
        if ( do_test ) then
!
!           CHECK THE RESULTS FOR SMALL RESIDUALS.
!
            res(:n,:nrhs) = b2(:n,:nrhs) - matmul( a2, b(:n,:nrhs) )
            err = maxval( sum( abs(res), dim=1 ) / ( sum(abs(a2)) + sum(abs(b2), dim=1) ) )
!
        end if
!
    end if
!
    call system_clock( count=iend )
!
    elapsed_time = real( iend - istart, stnd )/real( irate, stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, b, c, d, invdiag, a2, b2, res )
    else
        deallocate( a, b, c, d, invdiag )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1>=zero ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a positive semidefinite symmetric system of size ', &
       n, ' with', nrhs, ' right hand side vectors is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_gchol_cmp
! ============================
!
end program ex2_gchol_cmp
