program ex2_lapack_gelsd
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine GELSD
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 24/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gelsd
!
    use Statpack, only : lgl, stnd, true, false, zero, c50, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=4000, n=200, p=min(m,n), k=max(m,n), nrhs=100
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of gelsd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)              :: err, eps, rcond, elapsed_time, rwork2(1)
    real(stnd)              :: s(p)
    real(stnd), allocatable :: ra(:,:), ia(:,:), rb(:,:), ib(:,:), rwork(:)
!
    complex(stnd)              :: work2(1)
    complex(stnd), allocatable :: work(:), a(:,:), a2(:,:), b(:,:), b2(:,:), res(:,:)
!
    integer              :: info, lwork, liwork, lrwork, rank, iok, iwork2(1), &
                            istart, iend, irate, imax, itime
    integer, allocatable :: iwork(:)
!
    logical(lgl)   :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : MINIMUM-NORM SOLUTION OF A COMPLEX LINEAR LEAST SQUARES PROBLEM
!               WITH SEVERAL RIGHT HAND-SIDES USING A SVD DECOMPOSITION OF THE
!               COEFFICIENT MATRIX. THE SVD IS COMPUTED WITH THE DIVIDE AND CONQUER
!               METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), ra(m,n), ia(m,n),                   &
              b(k,nrhs), rb(m,nrhs), ib(m,nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n COMPLEX RANDOM DATA MATRIX a .
!
    call random_number( ra(:m,:n) )
    call random_number( ia(:m,:n) )
!
    a(:m,:n) = cmplx( ra(:m,:n), ia(:m,:n), stnd )
!
!   GENERATE A m-by-nrhs COMPLEX RANDOM RIGHT HAND-SIDE MATRIX b .
!
    call random_number( rb(:m,:nrhs) )
    call random_number( ib(:m,:nrhs) )
!
    b(:m,:nrhs) = cmplx( rb(:m,:nrhs), ib(:m,:nrhs), stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate ( ra, ia, rb, ib )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), b2(m,nrhs), res(m,nrhs), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
!       SAVE RIGHT HAND SIDE MATRIX .
!
        b2(:m,:nrhs) = b(:m,:nrhs)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE EFFECTIVE RANK OF a, rank, IS DETERMINED BY TREATING AS ZERO THOSE
!   SINGULAR VALUES WHICH ARE LESS THAN rcond TIMES THE LARGEST SINGULAR VALUE.
!
    rcond = 0.0000001_stnd
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gelsd SUBROUTINE.
!
    lwork = -1
!
    call gelsd( m, n, nrhs, a, m, b, k, s, rcond, rank, work2, lwork, rwork2, iwork2, info )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1)) + 10
        lrwork = int(rwork2(1))
        liwork = iwork2(1)
!
!       ALLOCATE WORK VARIABLES NEEDED BY gelsd SUBROUTINE.
!
        allocate( work(lwork), rwork(lrwork), iwork(liwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE MINIMAL NORM SOLUTION TO THE LINEAR LEAST SQUARES PROBLEM .
!
        call gelsd( m, n, nrhs, a, m, b, k, s, rcond, rank, work, lwork, rwork, iwork, info )
!
!       THE ROUTINE RETURNS THE SOLUTIONS AND SINGULAR VALUES OF THE COMPLEX LINEAR
!       LEAST SQUARE PROBLEM.
!
!       ON EXIT OF gelsd :
!
!         info=     0 :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT the iTH ARGUMENT HAS AN ILLEGAL VALUE
!         info=i  > 0 :  INDICATES THAT the SVD ALGORITHM FAILED TO CONVERGE
!                        AND i SUBDIAGONAL ELEMENTS OF AN INTERMEDIATE
!                        BIDIAGONAL MATRIX DID NOT CONVERGE TO ZERO.
!
!       ON EXIT OF gelsd :
!
!         a IS DESTROYED.
!
!         b IS OVERWRITTEN BY THE n-by-nrhs SOLUTION.
!           If m >= n AND rank = n, THE RESIDUAL SUM-OF-SQUARES
!           FOR THE SOLUTION IN THE i-th COLUMN IS GIVEN BY THE
!           SUM OF SQUARES OF THE MODULUS OF ELEMENTS n+1:m IN
!           THAT COLUMN.
!
!         s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a
!           IN DECREASING ORDER.
!
!         rank IS OVERWRITTEN BY THE EFFECTIVE RANK OF a, i.e.
!              THE NUMBER OF SINGULAR VALUES WHICH ARE GREATER
!              THAN rcond*s(1) .
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate ( work, iwork, rwork )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM gelsd SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GELSD subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COLUMN VECTORS OF a .
!
        res = b2 - matmul( a2, b(:n,:nrhs) )
        err = maxval( sum(abs(matmul(transpose(conjg(a2)),res)),   dim=1) )/ sum( abs(a2) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, b, a2, b2, res )
    else
        deallocate( a, b )
    end if
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*)   &
          'Orthogonality of the residuals against the columns of the coefficient matrix = ', err
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the solution of a linear least squares complex problem of size ', &
       m, ' by ', n, ' with ', nrhs,' right hand sides is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_lapack_gelsd
! ===============================
!
end program ex2_lapack_gelsd
