program ex2_lapack_ormbr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines GEBRD and ORMBR
!   in LAPACK software . 
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_SVD and BD_INVITER
!    in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 08/05/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gebrd, ormbr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, unit_matrix,  &
                         bd_svd, bd_inviter, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=3000, n=3000, nvec=100, k=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of ormbr'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, singval, e2
    real(stnd), allocatable, dimension(:,:) :: ra, ia, bd_leftvec, bd_rightvec, resid_ortho
!
    complex(stnd)                              :: work2(1)
    complex(stnd), allocatable, dimension(:)   :: tauq, taup, work
    complex(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, rightvec
!
    integer      :: info_bd, info_q, info_p, lwork, lwork_bd, lwork_q, lwork_p,  &
                    iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=4, l
!
    logical(lgl) :: failure, bd_is_upper, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : BIDIAGONAL REDUCTION AND PARTIAL SVD DECOMPOSITION
!               OF A COMPLEX MATRIX USING AN INVERSE ITERATION METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    bd_is_upper = m>=n
    do_test     = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), ra(m,n), ia(m,n),                         &
              bd_leftvec(k,nvec), bd_rightvec(k,nvec),          &
              leftvec(m,nvec), rightvec(n,nvec), d(k), e(k),    &
              tauq(k), taup(k), singval(k), e2(k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n COMPLEX RANDOM DATA MATRIX .
!
    call random_number( ra(:m,:n) )
    call random_number( ia(:m,:n) )
!
    a(:m,:n) = cmplx( ra(:m,:n), ia(:m,:n), stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate ( ra, ia )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), resid_ortho(nvec,nvec), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gebrd AND orgbr SUBROUTINES.
!
    lwork = -1
!
    call gebrd( m, n, a, m, d(:k), e(2_i4b:k), tauq, taup, work2, lwork, info=info_bd )
!
    lwork_bd = int(work2(1))
!
    call ormbr( 'Q', 'L', 'N', m, nvec, n, a, m, tauq, leftvec, m, work2, lwork, info=info_q )
!
    lwork_q = int(work2(1))
!
    call ormbr( 'P', 'L', 'N', n, nvec, m, a, m, taup, rightvec, n, work2, lwork, info=info_p )
!
    lwork_p = int(work2(1))
!
    if ( min(info_bd,info_q,info_p)==0 ) then
!
        lwork = max( lwork_bd, lwork_q, lwork_p )
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY gebrd AND ormbr SUBROUTINES.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE BIDIAGONAL REDUCTION OF RANDOM DATA MATRIX a = Q*BD*P**(h).
!       WHERE Q AND P ARE UNITARY MATRICES AND BD IS A REAL BIDIAGONAL MATRIX
!
        call gebrd( m, n, a, m, d(:k), e(2_i4b:k), tauq, taup, work(:lwork_bd), lwork_bd, info=info_bd )
!
!       ON EXIT OF gebrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE BIDAGONAL MATRIX BD, RESPECTIVELY.
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q AND P.
!         taup AND tauq ARE THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q AND P, RESPECTIVELY.
!
!       MAKE A COPY OF THE BIDIAGONAL MATRIX.
!
        singval(:k) = d(:k)
        e2(:k)      = e(:k)
!
!       COMPUTE ALL SINGULAR VALUES OF THE REAL BIDIAGONAL MATRIX BD.
!
        call bd_svd( bd_is_upper, singval(:k), e2(:k), failure, sort=sort  )
!
        if ( .not. failure ) then
!
!           COMPUTE THE FIRST nvec SINGULAR VECTORS OF THE BIDIAGONAL MATRIX BY maxiter INVERSE ITERATIONS.
!
            call bd_inviter( bd_is_upper, d(:k), e(:k), singval(:nvec), bd_leftvec(:k,:nvec),   &
                             bd_rightvec(:k,:nvec), failure, maxiter=maxiter )
!
!           NOW COMPUTE THE FIRST nvec SINGULAR VECTORS OF THE FULL m-BY-n MATRIX a BY BACK-TRANSFORMATION WITH
!           LAPACK SUBROUTINE ormbr.
!
            leftvec(:k,:nvec) = cmplx( bd_leftvec(:k,:nvec), zero, stnd )
            rightvec(:k,:nvec) = cmplx( bd_rightvec(:k,:nvec), zero, stnd )
!
            if ( bd_is_upper ) then
                leftvec(k+1_i4b:m,:nvec) = cmplx( zero, zero, stnd )
            else
                rightvec(k+1_i4b:n,:nvec) = cmplx( zero, zero, stnd )
            end if
!
            call ormbr( 'Q', 'L', 'N', m, nvec, n, a, m, tauq, leftvec, m, work(:lwork_q), lwork_q, info=info_q )
!
            call ormbr( 'P', 'L', 'N', n, nvec, m, a, m, taup, rightvec, n, work(:lwork_p), lwork_p, info=info_p )
!
!           ON EXIT OF THESE ormbr CALLS:
!
!           info= 0     :  INDICATES SUCCESSFUL EXIT.
!           info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!
!           leftvec  IS OVERWRITTEN WITH THE FIRST nvec LEFT SINGULAR VECTORS of a STORED COLUMNWISE.
!           rightvec IS OVERWRITTEN WITH THE FIRST nvec RIGHT SINGULAR VECTORS of a STORED COLUMNWISE.
!
        end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( min(info_bd,info_q,info_p)/=0 ) then
!
!       ANORMAL EXIT FROM gebrd OR ormbr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the calls to GEBRD or ORMBR subroutines, Info=', min(info_bd,info_q,info_p)
!                
    else if ( do_test .and. .not.failure ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*rightvec - leftvec*diag(singval(:nvec)),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        a(:m,:nvec) =  matmul(a2,rightvec) - leftvec*spread( singval(:nvec),dim=1,ncopies=m )
!
        resid_ortho(:nvec,1_i4b) = norm( a(:m,:nvec), dim=2_i4b )
        err1 =  maxval( resid_ortho(:nvec,1_i4b) )/( sum( abs(singval(:k)) )*real(m,stnd) )

        err1 = norm( a(:m,:nvec) )/( sum( abs(singval(:k)) )*real(nvec,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - leftvec**(h)*leftvec.
!
        call unit_matrix( a2(:nvec,:nvec) )
!
        resid_ortho(:nvec,:nvec) = abs( a2(:nvec,:nvec) - matmul( transpose(conjg(leftvec(:m,:nvec))), leftvec(:m,:nvec) ) )
        err2 = maxval( resid_ortho(:nvec,:nvec) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - rightvec**(h)*rightvec.
!
        resid_ortho(:nvec,:nvec) = abs( a2(:nvec,:nvec) - matmul( transpose(conjg(rightvec(:n,:nvec))), rightvec(:n,:nvec) ) )
        err3 = maxval( resid_ortho(:nvec,:nvec) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, a2, bd_leftvec, bd_rightvec, leftvec, rightvec,     &
                    d, e, tauq, taup, singval, e2, resid_ortho )
    else
        deallocate( a, bd_leftvec, bd_rightvec, leftvec, rightvec, d, e,   &
                    tauq, taup, singval, e2 )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. min(info_bd,info_q,info_p)==0 .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. min(info_bd,info_q,info_p)==0 .and. .not.failure ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and', nvec, ' singular vectors of a', &
       m, ' by', n,' complex matrix is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_lapack_ormbr
! ===============================
!
end program ex2_lapack_ormbr
