program ex2_lapack_sysv
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of the driver subroutine SYSV
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 21/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : sysv
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=1000, nrhs=400
!
    character(len=*), parameter :: name_proc='Example 2 of sysv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, elapsed_time
    real(stnd), dimension(:,:), allocatable :: ra, ia, rx, ix
!
    complex(stnd)                              :: work2(1)
    complex(stnd), dimension(:),   allocatable :: work
    complex(stnd), dimension(:,:), allocatable :: a, b, x, res
!
    integer(i4b), dimension(:), allocatable :: ipiv
    integer                                 :: info, lwork, iok, istart, &
                                               iend, irate, imax, itime
!
    logical(lgl)   :: do_test
!
    character      :: uplo
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : SOLUTION OF A LINEAR SYSTEM WITH A COMPLEX SYMMETRIC COEFFICIENT
!               MATRIX AND SEVERAL RIGHT HAND-SIDES.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), ra(n,n), ia(n,n), b(n,nrhs), x(n,nrhs),      &
              rx(n,nrhs), ix(n,nrhs), ipiv(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n COMPLEX RANDOM SYMMETRIC MATRIX a .
!
    call random_number( ra(:n,:n) )
    call random_number( ia(:n,:n) )
!
    a(:n,:n) = cmplx( ra(:n,:n), ia(:n,:n), stnd )
!
    a = a + transpose( a ) 
!
!   GENERATE A n-by-nrhs COMPLEX RANDOM SOLUTION MATRIX x .
!
    call random_number( rx(:n,:nrhs) )
    call random_number( ix(:n,:nrhs) )
!
    x(:n,:nrhs) = cmplx( rx(:n,:nrhs), ix(:n,:nrhs), stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate ( ra, ia, rx, ix )
!
!   COMPUTE THE MATRIX-MATRIX PRODUCT b = a*x .
!
    b = matmul( a, x )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo = 'U'
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR sysv SUBROUTINE.
!
    lwork = -1
!
    call sysv( uplo, n, nrhs, a, n, ipiv, b, n, work2, lwork, info )
!
    if ( info==0 ) then
!
        lwork = int(work2(1))
!
!       ALLOCATE WORK VARIABLE NEEDED BY sysv SUBROUTINE.
!
        allocate( work(lwork), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE SOLUTION MATRIX FOR COMPLEX SYMMETRIC LINEAR SYSTEM
!
!                a*x = b .
!
        call sysv( uplo, n, nrhs, a, n, ipiv, b, n, work, lwork, info )
!
!       ON EXIT OF sysv :
!
!          info    = 0 :  INDICATES SUCCESSFUL EXIT
!          info=-i < 0 :  INDICATES THAT the iTH ARGUMENT HAS AN ILLEGAL VALUE
!          info=i  > 0 :  INDICATES THAT THE SOLUTION CAN NOT BE COMPUTED BECAUSE
!                         THE MATRIX IS SINGULAR
!
!       THE ROUTINE RETURNS THE SOLUTION VECTORS IN b.
!       DETAILS OF THE U*D*U**(T) (if uplo='U') OR THE L*D*L**(T) (if uplo='L')
!       FACTORIZATION OF a ARE RETURNED IN ARGUMENTS ipiv AND a .
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM sysv SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to SYSV subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!
    else if ( do_test ) then
!
!       ALLOCATE WORK ARRAY .
!
        allocate( res(n,nrhs), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n,:nrhs) = b(:n,:nrhs) - x(:n,:nrhs)
        err = maxval( sum( abs(res), dim=1 ) /    &
                      sum(abs(x),    dim=1 )      )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( res )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x, ipiv )
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed solutions = ', err
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the solution of a linear symmetric complex system of size ', &
       n, ' with ', nrhs,' right hand sides is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_lapack_sysv
! ==============================
!
end program ex2_lapack_sysv
