program ex2_permute_cor
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines COMP_COR and PERMUTE_COR
!   in module Mul_Stat_Procedures .
!                                                                              
! LATEST REVISION : 06/07/2006
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, comp_cor, permute_cor, random_seed_, random_number_
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
! p       IS THE NUMBER OF OBSERVATIONS OF THE RANDOM VECTORS
! nrep    IS THE NUMBER OF SHUFFLES FOR THE PERMUTATION TEST
! nsample IS THE NUMBER OF SAMPLES TO EVALUATE THE REJECTION RATE OF THE TEST
!
    integer(i4b), parameter :: prtunit=6, p=50, nrep=99, nsample=3000
!
! sign_level IS THE SIGNIFICANCE LEVEL OF THE PERMUTATION TEST
! eps        IS THE ALLOWED RELATIVE ERROR FOR THE REJECTION RATE
!
    real(stnd),  parameter :: sign_level=0.05, eps=0.2
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                       :: xyn, err_prob, prob_rej_rate
    real(stnd), dimension(nsample)   :: xycor, prob
    real(stnd), dimension(nsample,2) :: xstat
    real(stnd), dimension(2)         :: ystat
    real(stnd), dimension(nsample,p) :: x
    real(stnd), dimension(p)         :: y
!
    integer(i4b) :: rej_rate
!
    logical(lgl) :: first, last
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 2 of permute_cor'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   INITIALIZE THE RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE A RANDOM UNIFORM OBSERVATION VECTOR y .
!
    call random_number_( y(:p) )
!
!   GENERATE A RANDOM UNIFORM OBSERVATION ARRAY x .
!
    call random_number_( x(:nsample,:p) )
!
!   COMPUTE THE CORRELATIONS BETWEEN x AND y
!   FOR THE p OBSERVATIONS .
!
    first = true
    last  = true
    call comp_cor( x(:nsample,:p), y(:p), first, last, xstat(:nsample,:2), ystat(:2),    &
                   xycor(:nsample), xyn )
!
!   ON EXIT OF COMP_COR WHEN last=true :
!
!      xstat(:nsample,1)     CONTAINS THE MEAN VALUES OF THE OBSERVATION ARRAY x(:nsample,:p).
!
!      xstat(:nsample,2)     CONTAINS THE VARIANCES OF THE OBSERVATION ARRAY x(:nsample,:p).
!
!      ystat(1)              CONTAINS THE MEAN VALUE OF THE OBSERVATION VECTOR  y(:p).
!
!      ystat(2)              CONTAINS THE VARIANCE OF THE OBSERVATION VECTOR  y(:p).
!
!      xycor(:nsample)       CONTAINS THE CORRELATION COEFFICIENTS BETWEEN x(:nsample,:p) AND y(:p).
!
!      xyn                   CONTAINS THE NUMBER OF OBSERVATIONS IN THE DATA ARRAYS
!                            x(:nsample,:p) AND  y(:p) (xyn=real(p,stnd) ).
!
!
!   NOW COMPUTE A PERMUTATION TEST OF THE CORRELATION BETWEEN x AND y WITH
!   SUBROUTINE permute_cor WITH nrep SHUFFLES .
!
    call permute_cor( x(:nsample,:p), y(:p), xstat(:nsample,:2), ystat(:2),   &
                      xycor(:nsample), prob(:nsample), nrep=nrep )
!
!   NOW COMPUTE THE REJECTION RATE OF THE TEST. IDEALLY, FOR THE sign_level
!   SIGNIFICANCE LEVEL, WE WOULD REJECT ONLY THE NULL HYPOTHESIS sign_level %
!   OF THE TIME. IN OTHER WORDS, THE REJECTION RATE MUST BE APPROXIMATELY EQUAL
!   TO THE SIGNIFICANCE LEVEL sign_level .
!
    rej_rate      = count( prob(:nsample)<=sign_level )
    prob_rej_rate = real( rej_rate, stnd )/real( nsample, stnd )
!
!   COMPUTE THE RELATIVE ERROR OF THE REJECTION RATE.
!
    err_prob = abs( (prob_rej_rate-sign_level)/sign_level )
!
!   CHECK THAT THE RELATIVE ERROR OF THE REJECTION RATE IS LESS THAN eps .
!
    if ( err_prob<=eps ) then
        write (prtunit,*) 'Example 2 of PERMUTE_COR is correct'
    else
        write (prtunit,*) 'Example 2 of PERMUTE_COR is incorrect'
    end if
!
!
! END OF PROGRAM ex2_permute_cor
! ==============================
!
end program ex2_permute_cor
