program ex2_qr_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines QR_CMP and QR_SOLVE
!   in modules QR_Procedures and LLSQ_Procedures .
!    
! LATEST REVISION : 06/07/2006                                                                             
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, true, qr_cmp, qr_solve
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, m=1000, n=500
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                 :: err
    real(stnd), dimension(m,n) :: a, a2
    real(stnd), dimension(n)   :: x, diagr, beta
    real(stnd), dimension(m)   :: b
!
    character(len=*), parameter :: name_proc='Example 2 of qr_cmp'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a .
!
    call random_number( a )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b .
!
    call random_number( b )
!
!   MAKE A COPY OF COEFFICIENT MATRIX a(:m,:n) .
!
    a2(:m,:n) = a(:m,:n)
!
!   EXAMPLE 2 : LINEAR LEAST SQUARES SYSTEM.
!
!   COMPUTE SOLUTION VECTOR FOR LINEAR LEAST SQUARES SYSTEM
!
!           a(:m,:n)*x(:n)=b(:m) .
!
!
!   FIRST COMPUTE QR FACTORIZATION OF a .
!
    call qr_cmp( a2, diagr, beta )
!
!   qr_cmp COMPUTES AN ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX
!   a . a IS ASSUMED OF FULL RANK. THE ROUTINE COMPUTES A QR FACTORIZATION
!   a :
!
!                     a = q * r 
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS ORTHOGONAL
!   FACTORIZATION. 
!
!   THE MATRIX q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            q = h(1)*h(2)* ... *h(k), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + BETA * ( V * V' ) ,
!                      
!   WHERE BETA IS A REAL SCALAR AND V IS A REAL M-ELEMENT VECTOR WITH V(1:i-1) = 0.
!   V(i:m) IS STORED ON EXIT IN a(i:m,i) AND BETA IN beta(i).
!                      
!   THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!   CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX r. THE ELEMENTS
!   OF THE DIAGONAL OF r ARE STORED IN THE ARRAY diagr. 
!
!
!   NOW, COMPUTE SOLUTION AND RESIDUAL VECTOR FOR LINEAR LEAST SQUARES SYSTEM
!
!             a(:m,:n)*x(:n)=b(:m) .
!
    call qr_solve( a2, diagr, beta, b, x, comp_resid=true )
!
!   qr_solve SOLVES OVERDETERMINED OR UNDERDETERMINED REAL LINEAR SYSTEMS
!
!                               a*x = b          
!
!   WITH AN m-BY-n MATRIX a, USING AN ORTHOGONAL FACTORIZATION OF a, AS 
!   COMPUTED BY qr_cmp. m>=n OR n>m IS PERMITTED, BUT a IS ASSUMED OF FULL RANK.
!
!   b IS A m RIGHT HAND SIDE VECTOR AND x IS A n SOLUTION VECTOR. SEVERAL RIGHT
!   HAND SIDES AND SOLUTION VECTORS MAY BE HANDLE IN A SINGLE CALL. IN THIS CASE,
!   b IS AN m-BY-l MATRIX AND x is AN n-BY-l MATRIX.
!   
!   IT IS ASSUMED THAT qr_cmp  HAS BEEN USED TO COMPUTE THE ORTHOGONAL 
!   FACTORIZATION OF a BEFORE CALLING qr_solve.
!
!   ON EXIT, IF comp_resid IS PRESENT AND IS EQUAL true,
!   THE RESIDUAL VECTOR b - a*x OVERWRITES b.
!
!   THE 2-NORM OF THE RESIDUAL VECTOR FOR THE SOLUTION VECTOR x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL REAL PARAMETER rnorm IN THE CALL OF qr_solve .
!
!   CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COLUMN VECTORS OF a .
!
    err = sum( abs( matmul( b, a ) ) )/ sum( abs(a) )
!
!   PRINT THE RESULT OF THE TEST.
!
    if ( err<=sqrt(epsilon(err)) ) then
        write (prtunit,*) 'Example 2 of QR_CMP is correct'
    else
        write (prtunit,*) 'Example 2 of QR_CMP is incorrect'
    end if
!
!
! END OF PROGRAM ex2_qr_cmp
! =========================
!
end program ex2_qr_cmp
