program ex2_symtrid_bisect
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYMTRID_BISECT
!   in module Eig_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine TRID_DEFLATE in module EIG_Procedures.
!                                                                              
! LATEST REVISION : 08/05/2016
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, symtrid_cmp, symtrid_bisect, &
                         trid_deflate, lamch, norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, neig=100
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of symtrid_bisect'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, abstol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: eigval, resid2, d, e
    real(stnd), allocatable, dimension(:,:) :: a, eigvec, a2
!
    integer       :: iok, istart, iend, irate, imax, itime
    integer(i4b)  :: max_qr_steps, neig2
!
    logical(lgl)  :: failure1, failure2, do_test, ortho
!   
    character     :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : FIRST neig EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF
!               A REAL SYMMETRIC MATRIX USING BISECTION AND A DEFLATION METHOD
!               FOR THE EIGENVECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    abstol = sqrt( lamch('s') )
    err    = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), eigvec(n,neig), eigval(n), d(n), e(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid2(neig), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE RANDOM SELF-ADJOINT MATRIX a .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!   IS WRITTEN
!
!                       a = U * D * U**(t)
!
!   WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!   THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!   THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
!   FIRST, REDUCE TO TRIDIAGONAL FORM THE SELF-ADJOINT MATRIX a AND SAVE
!   THE INTERMEDIATE TRIDIAGONAL MATRIX IN PARAMETERS d AND e WITH
!   SUBROUTINE symtrid_cmp.
!
    call symtrid_cmp( a, d, e, store_q=true )
!
!   ON EXIT OF symtrid_cmp, THE UPPER TRIANGLE OF a IS OVERWRITTEN WITH THE HOUSEHOLDER TRANSFORMATIONS
!   USED TO REDUCE a TO TRIDIAGONAL FORM IF THE PARAMETER store_q IS PRESENT AND SET TO true, OTHERWISE
!   THE UPPER TRIANGLE OF a IS DESTROYED AND THE HOUSEHOLDER TRANSFORMATIONS ARE NOT SAVED.
!   PARAMETERS d AND e STORE, RESPECTIVELY, THE DIAGONAL AND THE OFF-DIAGONAL ELEMENTS OF THE
!   INTERMEDIATE TRIDIAGONAL FORM OF a .
!
!   NEXT, COMPUTE neig EIGENVALUES OF THE INTERMEDIATE TRIDIAGONAL MATRIX BY A BISECTION METHOD
!   WITH HIGH ACCURACY. THE EIGENVALUES ARE COMPUTED WITH MAXIMUM ACCURACY WHEN OPTIONAL PARAMETER
!   abstol IS SET TO sqrt( lamch('s') ), WHICH IS THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    call symtrid_bisect( d, e, neig2, eigval, failure=failure1, sort=sort, le=neig, abstol=abstol )
!
!   ON EXIT OF symtrid_bisect:
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE BISECTION ALGORITHM. 
!                         APPLIED TO THE INTERMEDIATE TRIDIAGONAL FORM OF a.
!
!       eigval IS OVERWRITTEN WITH THE REQUESTED EIGENVALUES OF a.
!       IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!       IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!
!   FINALLY, COMPUTE THE FIRST neig EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY A
!   DEFLATION TECHNIQUE APPLIED ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!   d_e AND BACK-TRANSFORMATION.
!
!   ON ENTRY OF SUBROUTINE trid_deflate, PARAMETER eigval CONTAINS SELECTED
!   EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!   THE EIGENVALUES VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    ortho = false
    max_qr_steps = 4_i4b
!
    call trid_deflate( d(:n), e(:n), eigval(:neig), eigvec(:n,:neig), failure=failure2,     &
                       mat=a, ortho=ortho, max_qr_steps=max_qr_steps )
!
!   ON EXIT OF trid_deflate :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM.
!
!       eigvec CONTAINS THE neig EIGENVECTORS OF a ASSOCIATED WITH THE EIGENVALUES eigval(:neig).
!
!   trid_deflate MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!   IDENTICAL AND IF THESE EIGENVALUES ARE NOT COMPUTED WITH HIGH ACCURACY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        a(:n,:neig) = matmul( a2, eigvec ) - eigvec*spread( eigval(:neig), 1, n )
        resid2(:neig) = norm( a(:n,:neig), dim=2_i4b )
        err1 =  maxval( resid2(:neig) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a(:neig,:neig) )
!
        a2(:neig,:neig) = abs( a(:neig,:neig) - matmul( transpose( eigvec ), eigvec ) )
        err2 = maxval(a2(:neig,:neig))/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, a2, eigvec, eigval, d, e, resid2 )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, eigvec, eigval, d, e )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', neig, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_symtrid_bisect
! =================================
!
end program ex2_symtrid_bisect
