program ex2_symtrid_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines SYMTRID_CMP and
!   ORTHO_GEN_SYMTRID in module EIG_Procedures .
!                                                                              
! LATEST REVISION : 28/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, symtrid_cmp,    &
                         ortho_gen_symtrid, triangle, norm, unit_matrix,         &
                         merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, p=n*(n+1)/2
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of symtrid_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, a_packed
    real(stnd), allocatable, dimension(:,:) :: a, q, resid, trid
!
    integer(i4b) :: l
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : TRIDIAGONAL REDUCTION OF A REAL SYMMETRIC
!   MATRIX STORED IN PACKED FORM.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a_packed(p), q(n,n), d(n), e(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM SELF-ADJOINT MATRIX a IN PACKED FORM.
!
    call random_number( a_packed(:p) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       UNPACK AND SAVE RANDOM SELF-ADJOINT MATRIX a .
!
        a(:n,:n) = unpack( a_packed(:p), mask=triangle( true, n, n, extra=1_i4b), field=zero )
!
        do l = 1_i4b, n-1_i4b
             a(l+1_i4b:n,l) = a(l,l+1_i4b:n)
        end do
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   CALL symtrid_cmp AND ortho_gen_symtrid_cmp TO REDUCE THE MATRIX a
!   IN PACKED FORM TO TRIDIAGONAL FORM
!
!                      a = Q*TRID*Q**(t)
!
!   WHERE Q IS ORTHOGONAL AND TRID IS A SYMMETRIC TRIDIAGONAL MATRIX.
!
!   ON ENTRY OF symtrid_cmp, a_packed MUST CONTAINS THE LEADING n-BY-n UPPER TRIANGULAR PART
!   OF THE MATRIX TO BE REDUCED IN PACKED FORMAT.
!
    call symtrid_cmp( a_packed(:p), d, e, store_q=true )
!
!   ON EXIT OF symtrid_cmp:
!
!         ARGUMENTS d AND e CONTAIN, RESPECTIVELY, THE DIAGONAL AND
!         OFF-DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX TRID.
!
!         IF THE OPTIONAL ARGUMENT store_q IS PRESENT AND SET TO TRUE,
!         THE LINEAR ARRAY a_packed IS OVERWRITTEN BY THE MATRIX Q
!         AS A PRODUCT OF ELEMENTARY REFLECTORS.
!
!   UNPACKED THE MATRIX IN ORDER TO GENERATE Q BY A CALL TO ortho_gen_symtrid.
!
    q(:n,:n) = unpack( a_packed(:p), mask=triangle( true, n, n, extra=1_i4b), field=zero )
!
!   ortho_gen_symtrid GENERATES THE MATRIX Q STORED AS A PRODUCT OF
!   ELEMENTARY REFLECTORS AFTER A CALL TO symtrid_cmp WITH store_q=true.
!
    call ortho_gen_symtrid( q )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( trid(n,n), resid(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*Q - Q*TRID
!
        trid(:n,:n) = zero
!
        do l = 1_i4b, n-1_i4b
            trid(l,l)       = d(l)
            trid(l,l+1_i4b) = e(l)
            trid(l+1_i4b,l) = e(l)
        end do
!
        trid(n,n) = d(n)
!
        resid(:n,:n) = matmul( a(:n,:n), q(:n,:n)  )           &
                       - matmul( q(:n,:n), trid(:n,:n) )
!
        trid(:n,1_i4b) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( trid(:n,1_i4b) )/( norm( a )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
        call unit_matrix( a(:n,:n) )
!
        resid(:n,:n) = abs( a(:n,:n) - matmul( transpose(q(:n,:n )), q(:n,:n) ) )
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a_packed, q, d, e, a, trid, resid )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a_packed, q, d, e )
!
    endif
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed tridiagonal decomposition a = Q*TRD*Q**(t) = ', err1
        write (prtunit,*) 'Orthogonality of the computed Q orthogonal matrix                   = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the tridiagonal reduction of a ', &
       n, ' by ', n,' real symmetric matrix in packed form is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_symtrid_cmp
! ==============================
!
end program ex2_symtrid_cmp
