program ex3_qr_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines QR_CMP2 and QR_SOLVE2
!   in modules QR_Procedures and LLSQ_Procedures .
!                                                                              
! LATEST REVISION : 06/07/2006
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, true, qr_cmp2, qr_solve2
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, m=1000, n=500, k=min(m,n)
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                 :: err, tol
    real(stnd), dimension(m,n) :: a, a2
    real(stnd), dimension(n)   :: x, diagr, beta
    real(stnd), dimension(m)   :: b
    real(stnd), dimension(k)   :: tau
!
    integer(i4b)               :: krank
    integer(i4b), dimension(n) :: ip
!
    character(len=*), parameter :: name_proc='Example 3 of qr_cmp2'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a .
!
    call random_number( a )
!
!   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) .
!
    a(:m,5) = a(:m,10) + a(:m,11)
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b .
!
    call random_number( b )
!
!   MAKE A COPY OF COEFFICIENT MATRIX a(:m,:n) .
!
    a2(:m,:n) = a(:m,:n)
!
!   EXAMPLE 3 : LINEAR LEAST SQUARES SYSTEM WITH A MINIMAL 2-NORM SOLUTION.
!
!   COMPUTE MINIMAL 2-NORM SOLUTION VECTOR FOR LINEAR LEAST SQUARES SYSTEM
!
!                          a(:m,:n)*x(:n)=b(:m) .
!
!
!   SET TOLERANCE .
!
    tol = 0.00001_stnd
!
    krank = 0
!
!   COMPUTE COMPLETE QR DECOMPOSITION OF RANDOM DATA MATRIX a .
!
    call qr_cmp2( a2(:m,:n), diagr(:n), beta(:n), ip(:n), krank,    &
                  tol=tol, tau=tau(:k) )
!
!   qr_cmp2 COMPUTES A (COMPLETE) ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX
!   a . a MAY BE RANK-DEFICIENT. THE ROUTINE FIRST COMPUTES A QR FACTORIZATION
!   WITH COLUMN PIVOTING OF a :
!
!                     a * p = q * r = q * [ r11 r12 ]
!                                         [  0  r22 ]
!
!   WITH r11 DEFINED AS THE LARGEST LEADING SUBMATRIX WHOSE ESTIMATED CONDITION
!   NUMBER, IN THE 1-NORM, IS LESS THAN 1/tol OR SUCH THAT ABS(r11[j,j])>0 IF 
!   tol IS ABSENT. THE ORDER OF r11, krank, IS THE EFFECTIVE RANK OF a. 
!
!   ON INPUT, krank=k, IMPLIES THAT THE FIRST k COLUMNS OF A ARE
!   TO BE FORCED INTO THE BASIS. PIVOTING IS PERFORMED ON THE LAST n-k
!   COLUMNS OF a.
!   WHEN krank>=min(m,n) IS USED, PIVOTING IS NOT PERFORMED. THIS IS
!   APPROPRIATE WHEN a IS KNOWN TO BE FULL RANK. krank<=0 CAN BE USED
!   WHEN PIVOTING IS DESIRED ON ALL COLUMNS OF a.
!   
!   IF THE OPTIONAL PARAMETER tau IS PRESENT, THEN r22 IS CONSIDERED TO BE NEGLIGIBLE
!   AND r12 IS ANNIHILATED BY ORTHOGONAL TRANSFORMATIONS FROM THE RIGHT,
!   ARRIVING AT THE COMPLETE ORTHOGONAL FACTORIZATION:
!  
!                     a * p = q * [ t11 0 ] * z
!                                 [  0  0 ]
!   
!   p IS A n-BY-n PERMUTATION MATRIX, q IS A m-BY-m ORTHOGONAL MATRIX,
!   r IS A m-BY-n UPPER TRIANGULAR MATRIX, t11 IS A krank-BY-krank UPPER
!   TRIANGULAR MATRIX AND z IS A n-BY-n ORTHOGONAL MATRIX.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS (COMPLETE) 
!   ORTHOGONAL FACTORIZATION. 
!
!   THE MATRIX q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            q = h(1)*h(2)* ... *h(k), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + BETA * ( V * V' ) ,
!                      
!   WHERE BETA IS A REAL SCALAR AND V IS A REAL M-ELEMENT VECTOR WITH V(1:i-1) = 0.
!   V(i:m) IS STORED ON EXIT IN a(i:m,i) AND BETA IN beta(i).
!                      
!   IF THE OPTIONAL PARAMETER tau IS ABSENT :         
!                      
!      qr_cmp2 COMPUTES ONLY A QR FACTORIZATION WITH COLUMN PIVOTING OF a.
!                      
!      THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!      CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX r. THE ELEMENTS
!      OF THE DIAGONAL OF r ARE STORED IN THE ARRAY diagr. 
!
!      ON EXIT, IF ip(j)=k, THEN THE jTH COLUMN OF a*p WAS 
!      THE kTH COLUMN OF a. THE MATRIX p IS REPRESENTED IN THE ARRAY 
!      ip AS FOLLOWS: IF ip(j) = i THEN THE jTH COLUMN OF p IS 
!      THE iTH CANONICAL UNIT VECTOR.
!                      
!      ON EXIT, krank CONTAINS THE EFFECTIVE RANK OF a, I.E., THE ORDER OF
!      THE SUBMATRIX r11.
!                      
!   IF THE OPTIONAL PARAMETER tau IS PRESENT :         
!                      
!      qr_cmp2 COMPUTES A COMPLETE ORTHOGONAL FACTORIZATION OF a FROM THE QR 
!      FACTORIZATION WITH COLUMN PIVOTING OF a .
!                      
!      THE FACTORIZATION IS OBTAINED BY HOUSEHOLDER'S METHOD. THE kTH TRANSFORMATION
!      MATRIX, z(k), WHICH IS USED TO INTRODUCE ZEROS INTO THE kTH ROW OF r,
!      IS GIVEN IN THE FORM
!
!             z(k) = ( I    0  ),
!                    ( 0  T(k) )
!
!      WHERE
!
!             T(k) = I + TAU * ( U(k) * U(k)' ) , U(k) = (   1  )
!                                                        (   0  )
!                                                        ( L(k) )
!
!      TAU IS A SCALAR AND L(k) IS AN (n-krank) ELEMENT VECTOR. TAU and L(k) ARE CHOSEN
!      TO ANNIHILATE THE ELEMENTS OF THE kTH ROW OF r12.
!
!      ON EXIT, THE SCALAR TAU IS RETURNED IN THE kTH ELEMENT OF tau AND THE VECTOR U(K)
!      IN THE kTH ROW OF a, SUCH THAT THE ELEMENTS OF L(k) ARE IN a(k,krank+1:n).
!
!      THE z n-BY-n ORTHOGONAL MATRIX WHICH IS APPLIED FROM THE RIGHT TO R IS 
!      GIVEN BY THE PRODUCT
!
!             z = z(1) * z(2) * ... * z(krank)
!
!      ON EXIT, THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY SECTION a(1:krank,1:krank)
!      CONTAIN THE CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX t11. THE ELEMENTS OF
!      THE DIAGONAL OF t11 ARE STORED IN THE ARRAY SECTION diagr(1:krank).
!                      
!      ON EXIT, IF ip(j)=k, THEN THE jTH COLUMN OF a*p WAS 
!      THE kTH COLUMN OF a. THE MATRIX p IS REPRESENTED IN THE ARRAY 
!      ip AS FOLLOWS: IF ip(j) = i THEN THE jTH COLUMN OF p IS 
!      THE iTH CANONICAL UNIT VECTOR.
!                      
!      ON EXIT, krank CONTAINS THE EFFECTIVE RANK OF a, I.E., THE ORDER OF
!      THE SUBMATRIX t11.
!                      
!   IF tol IS PRESENT AND IS IN [0,1[, THEN :
!       ON ENTRY, THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF a
!       ARE PERFORMED. THEN, tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF a, WHICH IS DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX r11 IN THE QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER IS LESS THAN 1/tol. IF tol=0 IS SPECIFIED
!       THE NUMERICAL RANK OF a IS DETERMINED.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF a ARE NOT
!       PERFORMED AND CRUDE TESTS ON r(j,j) ARE DONE TO DETERMINE
!       THE NUMERICAL RANK OF a.
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE DETERMINED BY USING krank=0  AND tol=RELATIVE PRECISION OF THE ELEMENTS
!   IN a. IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD
!   BE USED. ALSO, IT MAY BE HELPFUL TO SCALE THE COLUMNS OF a SO THAT ALL ELEMENTS 
!   ARE ABOUT THE SAME ORDER OF MAGNITUDE.
!
!
!   NOW, COMPUTE  MINIMAL 2-NORM SOLUTION VECTOR AND RESIDUAL VECTOR
!   FOR LINEAR LEAST SQUARES SYSTEM
!
!                     a(:m,:n)*x(:n)=b(:m) .
!
    call qr_solve2( a2(:m,:n), diagr(:n), beta(:n), ip(:n), krank, b(:m), x(:n),    &
                    comp_resid=true, tau=tau(:k) )
!
!   qr_solve2 SOLVES OVERDETERMINED OR UNDERDETERMINED REAL LINEAR SYSTEMS
!
!                               a*x = b          
!
!   WITH AN m-BY-n MATRIX a, USING A (COMPLETE) ORTHOGONAL FACTORIZATION OF a, AS 
!   COMPUTED BY qr_cmp2. m>=n OR n>m IS PERMITTED AND a MAY BE RANK-DEFICIENT.
!
!   b IS A m RIGHT HAND SIDE VECTOR AND x IS A n SOLUTION VECTOR. SEVERAL RIGHT
!   HAND SIDES AND SOLUTION VECTORS MAY BE HANDLE IN A SINGLE CALL. IN THIS CASE,
!   b IS AN m-BY-l MATRIX AND x is AN n-BY-l MATRIX.
!   
!   IT IS ASSUMED THAT qr_cmp2  HAS BEEN USED TO COMPUTE THE (COMPLETE) ORTHOGONAL 
!   FACTORIZATION OF a BEFORE qr_solve2.
!
!   ON EXIT, IF comp_resid IS PRESENT AND IS EQUAL TO true,
!   THE RESIDUAL VECTOR b - a*x OVERWRITES b.
!
!   THE 2-NORM OF THE RESIDUAL VECTOR FOR THE SOLUTION VECTOR x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL REAL PARAMETER rnorm IN THE CALL OF qr_solve2 .
!
!   THE MINIMUN 2-NORM SOLUTION IS COMPUTED IF THE OPTIONAL PARAMETER tau IS PRESENT.
!   OTHERWISE, A SOLUTION IS COMPUTED SUCH THAT IF THE jTH COLUMN OF a
!   IS OMITTED FROM THE BASIS, x[j] IS SET TO ZERO.
!
!
!   CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COLUMN VECTORS OF a .
!
    err = sum( abs( matmul( b, a ) ) )/ sum( abs(a) )
!
!   PRINT THE RESULT OF THE TEST.
!
    if ( err<=sqrt(epsilon(err)) ) then
        write (prtunit,*) 'Example 3 of QR_CMP2 is correct'
    else
        write (prtunit,*) 'Example 23 of QR_CMP2 is incorrect'
    end if
!
!
!
! END OF PROGRAM ex3_qr_cmp2
! ==========================
!
end program ex3_qr_cmp2
