program ex3_trid_inviter
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine TRID_INVITER
!   in module EIG_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine SYMTRID_BISECT in module EIG_Procedures.
!
! LATEST REVISION : 23/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, one, two, c50, trid_inviter,    &
                         symtrid_bisect, unit_matrix, norm, lamch, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, neig=1000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 3 of trid_inviter'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, abstol, &
                                               normr, normt, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diag, sup, sup2, eigval
    real(stnd), allocatable, dimension(:,:) :: resid, eigvec
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: neig2, maxiter=2
!
    logical(lgl) :: failure, do_test
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 3 : SELECTED EIGENVALUES AND EIGENVECTORS OF
!               A REAL SYMMETRIC TRIDIAGONAL MATRIX USING
!               BISECTION FOR THE EIGENVALUES AND INVERSE
!               ITERATION METHOD FOR THE EIGENVECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    abstol = sqrt( lamch('s') )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( diag(n), sup(n), eigval(n), eigvec(n,neig), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A (1-2-1) TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN diag .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN sup .
!
!    diag(:n) = two
!    sup(:n)  = one
!
!    diag(:n) = 0.5_stnd
!    sup(:n)  = 0.5_stnd
!
!   GENERATE A RANDOM TRIDIAGONAL MATRIX.
!
    call random_number( diag(:n) )
    call random_number( sup(:n) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( resid(n,neig), sup2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE SELECTED EIGENVALUES OF THE TRIDIAGONAL MATRIX BY BISECTION.
!
    call symtrid_bisect( diag, sup, neig2, eigval, failure, sort=sort, le=neig, abstol=abstol )
!
    if ( .not. failure ) then
!
!       COMPUTE THE FIRST neig EIGENVECTORS OF THE TRIDIAGONAL MATRIX BY maxiter INVERSE ITERATIONS
!       WITH SUBROUTINE trid_inviter.
!
!       ON ENTRY, PARAMETER eigval CONTAINS SELECTED EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!       THE EIGENVALUES VALUES MUST BE GIVEN IN DECREASING ORDER.
!
        call trid_inviter( diag(:n), sup(:n), eigval(:neig), eigvec(:n,:neig), failure, maxiter=maxiter )
!
!       ON EXIT OF trid_inviter :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM

!       eigvec CONTAINS THE neig EIGENVECTORS ASSOCIATED WITH THE EIGENVALUES eigval(:neig).
!
!       trid_inviter MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!       IDENTICAL.
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s,
!       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX .
!
        sup2(:n) = eoshift( sup(:n), -1 )
        resid(:n,:neig) = spread( diag(:n),         dim=2, ncopies=neig )*eigvec                                + &
                          spread( sup2(:n),         dim=2, ncopies=neig )*eoshift( eigvec, shift=-1, dim=1 )    + &
                          eoshift( spread(sup2(:n), dim=2, ncopies=neig)*eigvec, shift=1 )                      - &
                          spread( eigval(:neig),    dim=1, ncopies=n )*eigvec
!
        sup(:neig) = norm( resid(:n,:neig), dim=2_i4b )
!
        normr = maxval( sup(:neig) )
        normt = sqrt( sum( diag(1_i4b:n)**2 ) + sum( sup(1_i4b:n-1_i4b)**2 ) )
!
        err1  = normr/( normt*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX .
!
        call unit_matrix( resid(:neig,:neig) )
!
        resid(:neig,:neig) = abs( resid(:neig,:neig) - matmul( transpose( eigvec ), eigvec ) )
!
        normr = maxval( resid(:neig,:neig) )
!
        err2  = normr/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( diag, sup, eigval, eigvec, resid, sup2 )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( diag, sup, eigval, eigvec )
!
    end if
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', neig,' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric tridiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex3_trid_inviter
! ===============================
!
end program ex3_trid_inviter
