module BLAS_interfaces
! ####################################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING GENERIC INTERFACES FOR SELECTED SUBROUTINES AND FUNCTIONS IN THE  *
! BLAS LIBRARY.                                                                      *
!                                                                                    *
! THIS INTERFACE MODULE ENSURES THAT CALLS TO BLAS ROUTINES ARE CORRECT, WHEN USED   *
! WITH STATPACK.                                                                     *
!                                                                                    *
! GENERIC INTERFACES ARE PRESENTLY PROVIDED FOR THE FOLLOWING BLAS ROUTINES:         *
!                                                                                    *
!    Xaxpy, Xcopy, Xdot, Xdotu, Xrot, Xswap, Xscal, Xnrm2, Xgemv, Xger, Xgeru,       *
!    Xtrsv, Xgemm                                                                    *
!                                                                                    *
! WHERE X CAN BE s, d, c AND z. THE GENERIC INTERFACES HAVE THE FORM:                *
!                                                                                    *
!    axpy, copy, dot, dotu, rot, swap, scal, nrm2, gemv, ger, geru, trsv, gemm       *
!                                                                                    * 
! LATEST REVISION : 16/05/2018                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! 
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public :: axpy, copy, dot, dotu, rot, swap, scal, nrm2, gemv, ger, geru, trsv, gemm
!
! GENERIC INTERFACES FOR BLAS ROUTINES WITH OVERLOADED VERSIONS
! =============================================================
!
!   Interface for AXPY subroutines in BLAS
!
    interface axpy
!
        subroutine saxpy( N, SA, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           REAL             SA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine daxpy( N, DA, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           DOUBLE PRECISION DA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine caxpy( N, CA, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           COMPLEX          CA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zaxpy( N, ZA, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           COMPLEX*16       ZA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for COPY subroutines in BLAS
!
    interface copy
!
        subroutine scopy( N, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine dcopy( N, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine ccopy( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zcopy( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for DOT functions in BLAS
!
    interface dot
!
        REAL function sdot( N, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end function
!
        DOUBLE PRECISION function ddot( N, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end function
!
        COMPLEX function cdotc( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end function
!
        COMPLEX*16 function zdotc( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * ), ZY( * )
        end function
!
    end interface
!
!   Interface for DOTU functions in BLAS
!
    interface dotu
!
        COMPLEX function cdotu( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end function
!
        COMPLEX*16 function zdotu( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * ), ZY( * )
        end function
!
    end interface
!
!   Interface for ROT subroutines in BLAS
!
    interface rot
!
        subroutine srot( N, SX, INCX, SY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           REAL             C, S
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine drot( N, DX, INCX, DY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           DOUBLE PRECISION C, S
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine csrot( N, CX, INCX, CY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           REAL             C, S
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zdrot( N, ZX, INCX, ZY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           DOUBLE PRECISION C, S
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for SWAP subroutines in BLAS
!
    interface swap
!
        subroutine sswap( N, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine dswap( N, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine cswap( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zswap( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for SCAL subroutines in BLAS
!
    interface scal
!
        subroutine sscal( N, SA, SX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           REAL             SA
!     ..
!     .. Array Arguments ..
           REAL             SX( * )
        end subroutine
!
        subroutine dscal( N, DA, DX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           DOUBLE PRECISION DA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * )
        end subroutine
!
        subroutine cscal( N, CA, CX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           COMPLEX          CA
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * )
        end subroutine
!
        subroutine zscal( N, ZA, ZX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           COMPLEX*16       ZA
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * )
        end subroutine
!
        subroutine csscal( N, SA, CX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           REAL             SA
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * )
        end subroutine
!
        subroutine zdscal( N, DA, ZX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           DOUBLE PRECISION DA
!     ..
!     .. Array Arguments ..
           COMPLEX*16       ZX( * )
        end subroutine
!
    end interface
!
!   Interface for NRM2 subroutines in BLAS
!
    interface nrm2
!
        REAL function snrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           REAL             X( * )
        end function
!
        DOUBLE PRECISION function dnrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION X( * )
        end function
!
        REAL function scnrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           COMPLEX          X( * )
        end function
!
        DOUBLE PRECISION function dznrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       X( * )
        end function
!
    end interface
!
!   Interface for GEMV subroutines in BLAS
!
    interface gemv
!
        subroutine sgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           REAL             ALPHA, BETA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine dgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           DOUBLE PRECISION ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine cgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX          ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine zgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX*16       ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for GER subroutines in BLAS
!
    interface ger
!
        subroutine sger( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           REAL             ALPHA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine dger( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           DOUBLE PRECISION ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine cgerc( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX          ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine zgerc( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX*16       ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for GERU subroutines in BLAS
!
    interface geru
!
        subroutine cgeru( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX          ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine zgeru( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX*16       ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for TRSV subroutines in BLAS
!
    interface trsv
!
        subroutine strsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), X( * )
        end subroutine
!
        subroutine dtrsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), X( * )
        end subroutine
!
        subroutine ctrsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * )
        end subroutine
!
        subroutine ztrsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), X( * )
        end subroutine
!
    end interface
!
!   Interface for GEMM subroutines in BLAS
!
    interface gemm
!
        subroutine sgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           REAL             ALPHA, BETA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine dgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           DOUBLE PRECISION ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine cgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           COMPLEX          ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine zgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           COMPLEX*16       ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!
! *******************************
! END OF MODULE BLAS_interfaces *
! *******************************
!
end module BLAS_interfaces
