module Hous_Procedures
! ##############################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ##############################################################################
!                                                                              *
! ******************************************************************************
! MODULE EXPORTING HOUSEHOLDER REFLECTIONS.                                    *
!                                                                              *
! LATEST REVISION : 27/09/2018                                                 *
!                                                                              *
! ##############################################################################
!                                                                              *
! ******************************************************************************
!                                                                              *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                              *
!                                                                              *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                   *
!  _DOT_PRODUCT FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH STATPACK *
!               dot_product2 FUNCTION                                          *
!  _MATMUL      FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK      *
!               matmul2 FUNCTION                                               *
!                                                                              *
! ******************************************************************************
!
!
! USED MODULES 
! ============
!
    use Select_Parameters,  only : lgl, i4b, stnd
#ifdef _MATMUL
    use Utilities,          only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,          only : dot_product=>dot_product2
#endif
#ifdef _BLAS
    use BLAS_interfaces,    only : axpy
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!
!   use Reals_Constants
!   use Utilities
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::                                                                 &
              hous1, apply_hous1, hous2, apply_hous2,                         &
              h1, apply_h1, h2, apply_h2
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface hous1
        module procedure    hous1_a, hous1_b
    end interface
!
    interface apply_hous1
        module procedure    apply_hous1_rv, apply_hous1_rm
    end interface
!
    interface apply_hous2
        module procedure    apply_hous2_rv, apply_hous2_rm
    end interface
!
    interface h1
        module procedure    h1_r, h1_rv, h1_rm
    end interface
!
    interface h2
        module procedure    h2_r, h2_rv, h2_rm
    end interface
!
    interface apply_h1
        module procedure    apply_h1_rv, apply_h1_rm
    end interface
!
    interface apply_h2
        module procedure    apply_h2_rv, apply_h2_rm
    end interface
!
!
! =========================================================================================
!
!
                                  contains
!                                 ========
!
! =========================================================================================
!                                HOUSEHOLDER TRANSFORMATIONS
! =========================================================================================
!
!
    subroutine hous1_a( u, tau )
!
! Purpose
! _______
!
!   HOUS1 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I and D' = (  beta  0  )
!
!   where beta is scalar and X is an n-element real vector.
!
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!
!   where tau is a real scalar and v is an n-element real vector with v(1) = 1.
!
!   If the elements of X(2:n) are all zero or size(X)=1, then tau = 0 
!   and H is taken to be the unit matrix.
!
!   Otherwise  1 <= tau <= 2.
!
!
! Arguments
! _________
!
!   U        (INPUT/OUTPUT) real(stnd), dimension(:) 
!            On entry, the vector X.
!
!            On exit, it is overwritten with the vector [beta v(2:n) ].
!
!   TAU      (OUTPUT) real(stnd)
!            On exit, the value tau.
!
!
! Further Details
! _______________
!
!   This subroutine is based on the routine  DLARFG in LAPACK77 (version 3) 
!   with improvements suggested by E. Anderson and M. Fahey. See,
!
!   (1)  Anderson, E., and Fahey, M., 1997: Performance improvements to LAPACK for 
!        the Cray Scientific Library. LAPACK Working Note No 126.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout) :: u(:)
    real(stnd),  intent(out)   :: tau
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, tmp
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( u )
!
!   QUICK RETURN IF POSSIBLE.
!
    tau = zero
    if ( n<=1_i4b ) return
!
    tmp = maxval( abs( u(2_i4b:n) ) )
!
    if ( tmp>zero ) then
!
!       COMPUTE THE 2-NORM OF u.
!
        tmp = max( tmp, abs( u(1_i4b) ) )
        cl = tmp*sqrt( sum( (u(:n)/tmp)**2 ) ) 
!
!       GENERATE ELEMENTARY REFLECTOR H.
!
        if ( cl>zero ) then
            cl = sign( cl, u(1_i4b) )
            tmp = u(1_i4b) + cl
            tau = -(tmp/cl)
!
            u(2_i4b:n) = ( one/tmp )*u(2_i4b:n)
            u(1_i4b) = -cl
        end if
!
    end if
!
!
! END OF SUBROUTINE hous1_a
! _________________________
!
    end subroutine hous1_a
!
! =========================================================================================
!
    subroutine hous1_b( u, tau, beta )
!
! Purpose
! _______
!
!   HOUS1 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I and D' = (  beta  0  )
!
!   where beta is scalar and X is an n-element real vector.
!
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!
!   where tau is a real scalar and v is an n-element real vector with v(1) = 1.
!
!   If the elements of X(2:n) are all zero or size(X)=1, then tau = 0 
!   and H is taken to be the unit matrix.
!
!   Otherwise  1 <= tau <= 2.
!
!
! Arguments
! _________
!
!   U        (INPUT/OUTPUT) real(stnd), dimension(:) 
!            On entry, the vector x.
!
!            On exit, it is overwritten with the vector [1 v(2:n) ].
!
!   TAU      (OUTPUT) real(stnd)
!            On exit, the value tau.
!
!   BETA     (OUTPUT) real(stnd)
!            On exit, the value beta.
!
!
! Further Details
! _______________
!
!   This subroutine is based on the routine  DLARFG in LAPACK77 (version 3) 
!   with improvements suggested by E. Anderson and M. Fahey. See,
!
!   (1)  Anderson, E., and Fahey, M., 1997: Performance improvements to LAPACK for 
!        the Cray Scientific Library. LAPACK Working Note No 126.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout) :: u(:)
    real(stnd),  intent(out)   :: tau, beta
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, tmp
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( u )
!
!   QUICK RETURN IF POSSIBLE.
!
    tau = zero
    if ( n<=0_i4b ) return
!
    beta = u(1_i4b)
!
    tmp = maxval( abs( u(2_i4b:n) ) )
!
    if ( tmp>zero ) then
!
!       COMPUTE THE 2-NORM OF u.
!
        tmp = max( tmp, abs( beta ) )
        cl = tmp*sqrt( sum( (u(:n)/tmp)**2 ) ) 
!
!       GENERATE ELEMENTARY REFLECTOR H.
!
        if ( cl>zero ) then
            cl = sign( cl, beta )
            tmp = beta + cl
            tau = -(tmp/cl)
!
            u(2_i4b:n) = ( one/tmp )*u(2_i4b:n)
            beta = -cl
        end if
!
    end if
!
    u(1_i4b) = one
!
!
! END OF SUBROUTINE hous1_b
! _________________________
!
    end subroutine hous1_b
!
! =========================================================================================
!
    subroutine hous2( pivot, u, tau )
!
! Purpose
! _______
!
!   HOUS2 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I , X' = (  alpha  x  ) and D' = (  beta  0  )
!
!   where alpha and beta are scalars, and x is an (n-1)-element real vector.
!
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!
!   where tau is a real scalar and v is an n-element real vector with v(1) = 1.
!
!   If the elements of x are all zero, then tau = 0 and H is taken to be
!   the unit matrix.
!
!   Otherwise  1 <= tau <= 2.
!
!
! Arguments
! _________
!
!   PIVOT    (INPUT/OUTPUT) real(stnd) 
!            On entry, the value alpha.
!            On exit, it is overwritten with the value beta.
!
!   U        (INPUT/OUTPUT) real(stnd), dimension(:) 
!            On entry, the (n-1)-element vector x.
!
!            On exit, it is overwritten with the vector v(2:n) .
!
!   TAU      (OUTPUT) real(stnd)
!            On exit, the value tau.
!
!
! Further Details
! _______________
!
!   This subroutine is based on the routine  DLARFG in LAPACK77 (version 3) 
!   with improvements suggested by E. Anderson and M. Fahey. See,
!
!   (1)  Anderson, E., and Fahey, M., 1997: Performance improvements to LAPACK for 
!        the Cray Scientific Library. LAPACK Working Note No 126.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Utilities,         only : pythag
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout) :: pivot, u(:)
    real(stnd),  intent(out)   :: tau
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, tmp
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( u )
!
!   QUICK RETURN IF POSSIBLE.
!
    tau = zero
    if ( n<=0_i4b ) return
!
    tmp = maxval( abs( u(:n) ) )
!
    if ( tmp>zero ) then
!
!       COMPUTE THE 2-NORM OF u.
!
        cl = tmp*sqrt( sum( (u(:n)/tmp)**2 ) ) 
!
!       COMPUTE THE HOUSEHOLDER REFLECTOR IF THE 2-NORM OF u 
!       IS GREATER THAN zero.
!
        if ( cl>zero ) then
            cl = sign( pythag(pivot,cl), pivot )
            tmp = pivot + cl
            tau = -(tmp/cl)
!
            u(:n) = ( one/tmp )*u(:n)
            pivot = -cl
        end if
!
    end if
!
!
! END OF SUBROUTINE hous2
! _______________________
!
    end subroutine hous2
!
! =========================================================================================
!
    subroutine apply_hous1_rv( u, tau, vec )
!
! Purpose
! _______
!
!   APPLY_HOUS1 applies a real elementary reflector H generated by HOUS1
!   to a real vector C. H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!
!   where tau is a real scalar and v is an n-element real vector with v(1) = 1.
!
!   If tau = 0, then H is taken to be the unit matrix.
!
!
! Arguments
! _________
!
!   U      (INPUT/OUTPUT) real(stnd), dimension(:) 
!          On entry, U(2:) contains the vector v(2:) in the representation 
!          of H as output by HOUS1. U is not used if tau = 0.
!
!          U is restored on exit.
!
!   TAU    (INPUT) real(stnd)
!          The value tau in the representation of H as output by HOUS1.
!
!   VEC    (INPUT/OUTPUT) real(stnd), dimension(:) 
!          On entry, the real vector C.
!
!          On exit, C is overwritten by the vector H * C
!
!
! Further Details
! _______________
!
!   It is assumed that size(VEC)>=size(U) .
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(inout), dimension(:) :: vec, u
    real(stnd),  intent(in)                  :: tau
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    REAL(stnd)   :: pivot, prod 
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau==zero ) return
!
    pivot = u(1_i4b)
    u(1_i4b) = one
!
    n = size( u )
!
    prod = tau*dot_product( u(:n), vec(:n) )
!
#ifdef _BLAS
    call axpy( n, prod, u(:n), 1_i4b, vec(:n), 1_i4b )
#else
    vec(:n) = vec(:n) + prod*u(:n)
#endif
!
    u(1_i4b) = pivot
!
!
! END OF SUBROUTINE apply_hous1_rv
! ________________________________
!
    end subroutine apply_hous1_rv
!
! =========================================================================================
!
    subroutine apply_hous1_rm( u, tau, mat, left )
!
! Purpose
! _______
!
!   APPLY_HOUS1 applies a real elementary reflector H generated by HOUS1
!   to a real n by m or m by n matrix, C, from the left or the right.
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!
!   where tau is a real scalar and v is an n-element real vector with v(1) = 1.
!
!   If tau = 0, then H is taken to be the unit matrix.
!
!
! Arguments
! _________
!
!   U        (INPUT/OUTPUT) real(stnd), dimension(:) 
!            On entry, U(2:) contains the vector v(2:) in the representation 
!            of H as output by HOUS1. U is not used if tau = 0.
!
!            U is restored on exit.
!
!   TAU      (INPUT) real(stnd)
!            The value tau in the representation of H as output by HOUS1.
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:) 
!            On entry, the n by m or m by n matrix C.
!
!            On exit, C is overwritten by the matrix H * C (LEFT=true) or
!            C * H (LEFT=false).
!
!   LEFT     (INPUT) logical(lgl)
!            If:
!
!            - LEFT=true,  H is  applied to the real matrix C from the left
!            - LEFT=false, H is  applied to the real matrix C from the right.
!
!
! Further Details
! _______________
!
!   It is assumed that:
!
!   - size(MAT,1)>=size(U) if LEFT=true
!   - size(MAT,2)>=size(U) if LEFT=false.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: mat(:,:), u(:)
    real(stnd),   intent(in)    :: tau
!
    logical(lgl), intent(in)    :: left
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: pivot, tmp( (size(mat,1)+size(mat,2))-size(u) )
!
    integer(i4b) :: n, m, i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau==zero ) return
!
    pivot = u(1_i4b)
    u(1_i4b) = one
!
    n = size( u )
!
    if ( left ) then
!
        m = size( mat, 2 )
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO mat(:,:) FROM THE LEFT.
!
        tmp(:m) = matmul( u(:n), mat(:n,:m) )
        tmp(:m) = tmp(:m)*tau
!
        do i=1_i4b, m
#ifdef _BLAS
            call axpy( n, tmp(i), u(:n), 1_i4b, mat(:n,i), 1_i4b )
#else
            mat(:n,i) = mat(:n,i) + tmp(i)*u(:n)
#endif
        end do
!
    else
!
        m = size( mat, 1 )
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO mat(:,:) FROM THE RIGHT.
!
        tmp(:m) = matmul( mat(:m,:n), u(:n) )
        tmp(:m) = tmp(:m)*tau
!
        do i=1_i4b, n
#ifdef _BLAS
            call axpy( m, u(i), tmp(:m), 1_i4b, mat(:m,i), 1_i4b )
#else
            mat(:m,i) = mat(:m,i) + u(i)*tmp(:m)
#endif
        end do
!
    end if
!
    u(1_i4b) = pivot
!
!
! END OF SUBROUTINE apply_hous1_rm
! ________________________________
!
    end subroutine apply_hous1_rm
!
! =========================================================================================
!
    subroutine apply_hous2_rv( u, tau, piv, vec )
!
! Purpose
! _______
!
!   APPLY_HOUS2 applies a real n-by-n elementary reflector H generated
!   by HOUS2 to a real vector C. H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!
!   where tau is a real scalar and v is an n-element real vector with v(1) = 1.
!
!   If tau = 0, then H is taken to be the unit matrix.
!
!
! Arguments
! _________
!
!   U        (INPUT) real(stnd), dimension(:) 
!            On entry, U contains the vector v(2:n) in the representation 
!            of H as output by HOUS2. U is not used if tau = 0.
!
!   TAU      (INPUT) real(stnd)
!            The value tau in the representation of H as output by HOUS2.
!
!   PIV      (INPUT/OUTPUT) real(stnd)
!            On entry, the scalar C[1].
!
!            On exit, PIV is overwritten by the scalar (H * C)[1].
!
!   VEC      (INPUT/OUTPUT) real(stnd), dimension(:) 
!            On entry, the real vector  C[2:].
!
!            On exit, C is overwritten by the vector (H * C)[2:].
!
!
! Further Details
! _______________
!
!   It is assumed that size(VEC)>=size(U).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)     :: u(:), tau
    real(stnd),  intent(inout)  :: piv, vec(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau==zero ) return
!
    n = size( u )
!
    tmp = dot_product( u(:n), vec(:n) )
    tmp = ( piv + tmp )*tau
    piv = piv + tmp
!
#ifdef _BLAS
    call axpy( n, tmp, u(:n), 1_i4b, vec(:n), 1_i4b )
#else
    vec(:n) = vec(:n) + tmp*u(:n)
#endif
!
!
! END OF SUBROUTINE apply_hous2_rv
! ________________________________
!
    end subroutine apply_hous2_rv
!
! =========================================================================================
!
    subroutine apply_hous2_rm( u, tau, vec_piv, mat, left )
!
! Purpose
! _______
!
!   APPLY_HOUS2 applies a real n-by-n elementary reflector H generated
!   by HOUS2 to a real n by m or m by n matrix, C, from the left or the right.
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!
!   where tau is a real scalar and v is an n-element real vector with v(1) = 1.
!
!   If tau = 0, then H is taken to be the unit matrix.
!
!
! Arguments
! _________
!
!   U            (INPUT) real(stnd), dimension(:) 
!                On entry, U contains the vector v(2:n) in the representation 
!                of H as output by HOUS2. U is not used if tau = 0.
!
!   TAU          (INPUT) real(stnd)
!                The value tau in the representation of H as output by HOUS2.
!
!   VEC_PIV      (INPUT/OUTPUT) real(stnd), dimension(:)
!                If LEFT=true:
!
!                - On entry, the row_vector C[1,:]
!                - On exit, VEC_PIV is overwritten by the vector (H * C)[1,:].
!
!                If LEFT=false:
!
!                - On entry, the column_vector C[:,1]
!                - On exit, VEC_PIV is overwritten by the vector (C * H)[:,1].
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:) 
!                If LEFT=true:
!
!                - On entry, the n-1 by m matrix C[2:,:]
!                - On exit, MAT is overwritten by the matrix  (H * C)[2:,:].
!
!                If LEFT=false:
!
!                - On entry, the m by n-1 matrix C[:,2:].
!                - On exit, MAT is overwritten by the matrix (C * H)[:,2:].
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT=true,  H is  applied to the real matrix C from the left
!                - LEFT=false, H is  applied to the real matrix C from the right.
!
!
! Further Details
! _______________
!
!   It is assumed that:
!
!   - size(MAT,1)>=size(U)  and  size(MAT,2)>=size(VEC_PIV)  if LEFT=true
!   - size(MAT,2)>=size(U)  and  size(MAT,1)>=size(VEC_PIV)  if LEFT=false.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)     :: u(:), tau
    real(stnd),  intent(inout)  :: vec_piv(:), mat(:,:)
!
    logical(lgl), intent(in)    :: left
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp(size(vec_piv))
!
    integer(i4b) :: n, m, i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau==zero ) return
!
    n = size( u )
    m = size( vec_piv )
!
    if ( left ) then
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO C FROM THE LEFT.
!    
        tmp(:m) = matmul( u(:n), mat(:n,:m) )
        tmp(:m) = ( vec_piv(:m) + tmp(:m) )*tau
!
        do i=1_i4b, m
#ifdef _BLAS
            call axpy( n, tmp(i), u(:n), 1_i4b, mat(:n,i), 1_i4b )
#else
            mat(:n,i) = mat(:n,i) + tmp(i)*u(:n)
#endif
        end do
!
    else
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO C FROM THE RIGHT.
!
        tmp(:m) = matmul( mat(:m,:n), u(:n) )
        tmp(:m) = ( vec_piv(:m) + tmp(:m) )*tau
!
        do i=1_i4b, n
#ifdef _BLAS
            call axpy( m, u(i), tmp(:m), 1_i4b, mat(:m,i), 1_i4b )
#else
            mat(:m,i) = mat(:m,i) + u(i)*tmp(:m)
#endif
        end do
!
    end if
!
    vec_piv(:m) = vec_piv(:m) + tmp(:m)
!
!
! END OF SUBROUTINE apply_hous2_rm
! ________________________________
!
    end subroutine apply_hous2_rm
!
! =========================================================================================
!
    subroutine h1_r( u, beta, tau )
!
! Purpose
! _______
!
!   H1 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I and D' = (  beta  0  )
!
!   where beta is scalar and X is an n-element real vector.

!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!
! Arguments
! _________
!
!   U       (INPUT/OUTPUT) real(stnd), dimension(:) 
!           On entry, U contains the pivot vector X.
!
!           On exit, U contains the vector v of the Householder reflector.
!
!   BETA    (OUTPUT) real(stnd)
!           On exit, the value beta.
!
!   TAU     (OUTPUT) real(stnd)
!           On exit, the value tau.
!
!
! Further Details
! _______________
!
!   On output,  H is the identity matrix if TAU = 0.
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: u(:)
    real(stnd),   intent(out)   :: beta, tau
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, smax
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n = size( u )
!
    smax = maxval( abs( u(2_i4b:n) ) )
!
    if ( smax>zero ) then
!    
!       COMPUTE THE HOUSEHOLDER REFLECTOR IF u(2:n) 
!       IS NOT THE NULL VECTOR.
!
!       COMPUTE THE 2-NORM OF u.
!
        smax = max( smax, abs( u(1_i4b) ) )
        u(:n) = u(:n)/smax
        cl = sqrt( dot_product( u(:n), u(:n) ) ) 
!    
!       COMPUTE THE HOUSEHOLDER TRANSFORMATION H .
!    
        cl = -sign( cl, u(1_i4b) )
        u(1_i4b) = u(1_i4b) - cl
        tau = one/( cl*u(1_i4b) )
!    
!       COMPUTE THE SCALAR beta .
!    
        beta = cl*smax
!
    else
!    
!       H IS THE IDENTITY MATRIX.
!    
        tau = zero
        if ( n>=1_i4b ) then
            beta = u(1_i4b)
            u(1_i4b) =zero
        else
            beta = zero
        end if
!
    end if
!
!
! END OF SUBROUTINE h1_r
! ______________________
!
    end subroutine h1_r
!
! =========================================================================================
!
    subroutine h1_rv( u, beta, tau, vec )
!
! Purpose
! _______
!
!   H1 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I and D' = (  beta  0  )
!
!   where beta is scalar and X is an n-element real vector.

!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!   The real elementary reflector H is then applied to a real vector C .
!
!
! Arguments
! _________
!
!   U       (INPUT/OUTPUT) real(stnd), dimension(:) 
!           On entry, U contains the pivot vector X.
!
!           On exit, U contains the vector v of the Householder reflector.
!
!   BETA    (OUTPUT) real(stnd)
!           On exit, the value beta.
!
!   TAU     (OUTPUT) real(stnd)
!           On exit, the value tau.
!
!   VEC     (INPUT/OUTPUT) real(stnd), dimension(:) 
!           On entry, the real vector  C .
!
!           On exit, C is overwritten by the vector H * C .
!
!
! Further Details
! _______________
!
!   On output,  H is the identity matrix if TAU = 0.
!
!   It is assumed that size(VEC) >= size(U) = n .
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:) :: u, vec
    real(stnd),   intent(out)                 :: beta, tau
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, smax
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n = size( u )
!
    smax = maxval( abs( u(2_i4b:n) ) )
!
    if ( smax>zero ) then
!    
!       COMPUTE THE HOUSEHOLDER REFLECTOR IF u(2:n) 
!       IS NOT THE NULL VECTOR.
!
!       COMPUTE THE 2-NORM OF u.
!
        smax = max( smax, abs( u(1_i4b) ) )
        u(:n) = u(:n)/smax
        cl = sqrt( dot_product( u(:n), u(:n) ) ) 
!    
!       COMPUTE THE HOUSEHOLDER TRANSFORMATION H .
!    
        cl = -sign( cl, u(1_i4b) )
        u(1_i4b) = u(1_i4b) - cl
        tau = one/( cl*u(1_i4b) )
!    
!       COMPUTE THE SCALAR beta .
!    
        beta = cl*smax
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO THE VECTOR C .
!    
        cl = tau*dot_product( u(:n), vec(:n) )
!
#ifdef _BLAS
        call axpy( n, cl, u(:n), 1_i4b, vec(:n), 1_i4b )
#else
        vec(:n) = vec(:n) + cl*u(:n)
#endif
!
    else
!    
!       H IS THE IDENTITY MATRIX.
!    
        tau = zero
        if ( n>=1_i4b ) then
            beta = u(1_i4b)
            u(1_i4b) =zero
        else
            beta = zero
        end if
!
    end if
!
!
! END OF SUBROUTINE h1_rv
! _______________________
!
    end subroutine h1_rv
!
! =========================================================================================
!
    subroutine h1_rm( u, beta, tau, mat, left )
!
! Purpose
! _______
!
!   H1 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I and D' = (  beta  0  )
!
!   where beta is scalar and X is an n-element real vector.
!
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!   The real elementary reflector H is then applied to a real matrix C from
!   the left or the right.
!
!
! Arguments
! _________
!
!   U       (INPUT/OUTPUT) real(stnd), dimension(:) 
!           On entry, U contains the pivot vector X.
!
!           On exit, U contains the vector v of the Householder reflector.
!
!   BETA    (OUTPUT) real(stnd)
!           On exit, the value beta.
!
!   TAU     (OUTPUT) real(stnd)
!           On exit, the value tau.
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:) 
!           On entry, the n by m or m by n real matrix C.
!
!           On exit, C is overwritten by the matrix H * C (LEFT=true)
!           or C * H (LEFT=false).
!
!   LEFT    (INPUT) logical(lgl)
!           If:
!
!           - LEFT=true,  H is  applied to the real matrix C from the left
!           - LEFT=false, H is  applied to the real matrix C from the right.
!
!
! Further Details
! _______________
!
!   On output,  H is the identity matrix if TAU = 0.
!
!   It is assumed that:
!
!   - size(MAT,1) >= size(U) if LEFT=true
!   - size(MAT,2) >= size(U) if LEFT=false.
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: u(:), mat(:,:)
    real(stnd),   intent(out)   :: beta, tau
!
    logical(lgl), intent(in)    :: left
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, smax, tmp( (size(mat,1)+size(mat,2))-size(u) )
!
    integer(i4b) :: n, m, i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n = size( u )
!
    smax = maxval( abs( u(2_i4b:n) ) )
!
    if ( smax>zero ) then
!    
!       COMPUTE THE HOUSEHOLDER REFLECTOR IF u(2:n) 
!       IS NOT THE NULL VECTOR.
!
!       COMPUTE THE 2-NORM OF u.
!
        smax = max( smax, abs( u(1_i4b) ) )
        u(:n) = u(:n)/smax
        cl = sqrt( dot_product( u(:n), u(:n) ) ) 
!    
!       COMPUTE THE HOUSEHOLDER TRANSFORMATION H .
!    
        cl = -sign( cl, u(1_i4b) )
        u(1_i4b) = u(1_i4b) - cl
        tau = one/( cl*u(1_i4b) )
!    
!       COMPUTE THE SCALAR beta .
!    
        beta = cl*smax
!
        if ( left ) then
!
            m = size( mat, 2 )
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION H TO mat(:,:) FROM THE LEFT.
!
            tmp(:m) = matmul( u(:n), mat(:n,:m) )
            tmp(:m) = tmp(:m)*tau
!
            do i=1_i4b, m
#ifdef _BLAS
                call axpy( n, tmp(i), u(:n), 1_i4b, mat(:n,i), 1_i4b )
#else
                mat(:n,i) = mat(:n,i) + tmp(i)*u(:n)
#endif
            end do
!
        else
!
            m = size( mat, 1 )
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION H TO mat(:,:) FROM THE RIGHT.
!
            tmp(:m) = matmul( mat(:m,:n), u(:n) )
            tmp(:m) = tmp(:m)*tau
!
            do i=1_i4b, n
#ifdef _BLAS
                call axpy( m, u(i), tmp(:m), 1_i4b, mat(:m,i), 1_i4b )
#else
                mat(:m,i) = mat(:m,i) + u(i)*tmp(:m)
#endif
            end do
!
        end if
!
    else
!    
!       H IS THE IDENTITY MATRIX.
!    
        tau = zero
        if ( n>=1_i4b ) then
            beta = u(1_i4b)
            u(1_i4b) =zero
        else
            beta = zero
        end if
!
    end if
!
!
! END OF SUBROUTINE h1_rm
! _______________________
!
    end subroutine h1_rm
!
! =========================================================================================
!
    subroutine h2_r( beta, u, up, tau )
!
! Purpose
! _______
!
!   H2 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I , X' = (  alpha  x  ) and D' = (  beta  0  )
!
!   where alpha and beta are scalars, and x is an (n-1)-element real vector.
!
!   H is represented in the form
!
!            H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!
! Arguments
! _________
!
!   BETA    (INPUT/OUTPUT) real(stnd) 
!           On entry, the value alpha.
!
!           On exit, it is overwritten with the value beta.
!
!   U       (INPUT/OUTPUT) real(stnd), dimension(:) 
!           On entry, U contains the pivot (n-1)-element vector x.
!
!           On exit, U contains the vector v(2:) of the Householder reflector.
!
!   UP      (OUTPUT) real(stnd)
!           On exit, UP contains the value v(1) of the Householder reflector.
!
!   TAU     (OUTPUT) real(stnd)
!           On exit, TAU contains the value tau of the Householder reflector.
!
!
! Further Details
! _______________
!
!   On output,  H is the identity matrix if TAU = 0.
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: beta, u(:)
    real(stnd),   intent(out)   :: up, tau
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, smax
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n = size( u )
!
    smax = maxval( abs( u(:n) ) )
!
    if ( smax>zero ) then
!    
!       COMPUTE THE HOUSEHOLDER REFLECTOR IF u
!       IS NOT THE NULL VECTOR.
!
!       COMPUTE THE 2-NORM OF u.
!
        smax = max( smax, abs( beta ) )
        u(:n) = u(:n)/smax
        beta = beta/smax
        cl = sqrt( dot_product( u(:n), u(:n) ) + beta*beta ) 
!    
!       COMPUTE THE HOUSEHOLDER TRANSFORMATION H .
!    
        cl = -sign( cl, beta )
        up = beta - cl
        tau = one/( up*cl )
!    
!       COMPUTE THE SCALAR beta .
!    
        beta = cl*smax
!
    else
!    
!       H IS THE IDENTITY MATRIX.
!    
        tau = zero
        up = zero
!
    end if
!
!
! END OF SUBROUTINE h2_r
! ______________________
!
    end subroutine h2_r
!
! =========================================================================================
!
    subroutine h2_rv( beta, u, up, tau, piv, vec )
!
! Purpose
! _______
!
!   H2 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I , X' = (  alpha  x  ) and D' = (  beta  0  )
!
!   where alpha and beta are scalars, and x is an (n-1)-element real vector.
!
!   H is represented in the form
!
!            H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!   The real elementary reflector H is then applied to a real vector  C  .
!
!
! Arguments
! _________
!
!   BETA        (INPUT/OUTPUT) real(stnd) 
!               On entry, the value alpha.
!
!               On exit, it is overwritten with the value beta.
!
!   U           (INPUT/OUTPUT) real(stnd), dimension(:) 
!               On entry, U contains the pivot (n-1)-element vector x.
!               On exit, U contains the vector v(2:) of the Householder reflector.
!
!   UP          (OUTPUT) real(stnd)
!               On exit, UP contains the value v(1) of the Householder reflector.
!
!   TAU         (OUTPUT) real(stnd)
!               On exit, TAU contains the value tau of the Householder reflector.
!
!   PIV         (INPUT/OUTPUT) real(stnd)
!               On entry, the scalar C[1].
!
!               On exit, PIV is overwritten by the scalar (H * C)[1].
!
!   VEC         (INPUT/OUTPUT) real(stnd), dimension(:) 
!               On entry, the real vector  C[2:].
!
!               On exit, C is overwritten by the vector (H * C)[2:].
!
!
! Further Details
! _______________
!
!   On output,  H is the identity matrix if TAU = 0.
!
!   It is assumed that size(VEC) >= size(U) .
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: beta, u(:), piv, vec(:)
    real(stnd),   intent(out)   :: up, tau
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl, smax
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n = size( u )
!
    smax = maxval( abs( u(:n) ) )
!
    if ( smax>zero ) then
!    
!       COMPUTE THE HOUSEHOLDER REFLECTOR IF u
!       IS NOT THE NULL VECTOR.
!
!       COMPUTE THE 2-NORM OF u.
!
        smax = max( smax, abs( beta ) )
        u(:n) = u(:n)/smax
        beta = beta/smax
        cl = sqrt( dot_product( u(:n), u(:n) ) + beta*beta ) 
!    
!       COMPUTE THE HOUSEHOLDER TRANSFORMATION H .
!    
        cl = -sign( cl, beta )
        up = beta - cl
        tau = one/( up*cl )
!    
!       COMPUTE THE SCALAR beta .
!    
        beta = cl*smax
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO THE VECTOR C .
!    
        cl = dot_product( u(:n), vec(:n) )
        cl = ( piv*up + cl )*tau
!
#ifdef _BLAS
        call axpy( n, cl, u(:n), 1_i4b, vec(:n), 1_i4b )
#else
        vec(:n) = vec(:n) + cl*u(:n)
#endif
!
        piv = piv + up*cl
!
    else
!    
!       H IS THE IDENTITY MATRIX.
!    
        tau = zero
        up = zero
!
    end if
!
!
! END OF SUBROUTINE h2_rv
! _______________________
!
    end subroutine h2_rv
!
! =========================================================================================
!
    subroutine h2_rm( beta, u, up, tau, vec_piv, mat, left )
!
! Purpose
! _______
!
!   H2 generates a real elementary reflector H of order n, such that
!
!           H * X = D , with  H' * H = I , X' = (  alpha  x  ) and D' = (  beta  0  )
!
!   where alpha and beta are scalars, and x is an (n-1)-element real vector.
!
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!   The real elementary reflector H is then applied to a real matrix C from
!   the left or the right.
!
!
! Arguments
! _________
!
!   BETA        (INPUT/OUTPUT) real(stnd) 
!               On entry, the value alpha.
!
!               On exit, it is overwritten with the value beta.
!
!   U           (INPUT/OUTPUT) real(stnd), dimension(:) 
!               On entry, U contains the pivot (n-1)-element vector x.
!               On exit, U contains the vector v(2:) of the Householder reflector.
!
!   UP          (OUTPUT) real(stnd)
!               On exit, UP contains the value v(1) of the Householder reflector.
!
!   TAU         (OUTPUT) real(stnd)
!               On exit, TAU contains the value tau of the Householder reflector.
!
!   VEC_PIV     (INPUT/OUTPUT) real(stnd), dimension(:)
!               If LEFT=true:
!
!               - On entry, the row_vector C[1,:]
!               - On exit, VEC_PIV is overwritten by the vector (H * C)[1,:].
!
!               If LEFT=false:
!
!               - On entry, the column_vector C[:,1]
!               - On exit, VEC_PIV is overwritten by the vector (C * H)[:,1].
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:) 
!               If LEFT=true:
!
!               - On entry, the n-1 by m matrix C[2:,:]
!               - On exit, MAT is overwritten by the matrix  (H * C)[2:,:] .
!
!               If LEFT=false:
!
!               - On entry, the m by n-1 matrix C[:,2:]
!               - On exit, MAT is overwritten by the matrix (C * H)[:,2:].
!
!   LEFT        (INPUT) logical(lgl)
!               If:
!
!               - LEFT=true,  H is  applied to the real matrix C from the left
!               - LEFT=false, H is  applied to the real matrix C from the right.
!
!
! Further Details
! _______________
!
!   On output,  H is the identity matrix if TAU = 0.
!
!   It is assumed that:
!
!   - size(MAT,1) >= size(U)  and  size(MAT,2) >= size(VEC_PIV) if LEFT=true.
!   - size(MAT,2) >= size(U)  and  size(MAT,1) >= size(VEC_PIV) if LEFT=false.
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: beta, u(:), vec_piv(:), mat(:,:)
    real(stnd),   intent(out)   :: up, tau
!
    logical(lgl), intent(in)    :: left
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   ::  cl, smax, tmp(size(vec_piv))
!
    integer(i4b) :: n, m, i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n = size( u )
!
    smax = maxval( abs( u(:n) ) )
!
    if ( smax>zero ) then
!    
!       COMPUTE THE HOUSEHOLDER REFLECTOR IF u
!       IS NOT THE NULL VECTOR.
!
!       COMPUTE THE 2-NORM OF u.
!
        smax = max( smax, abs( beta ) )
        u(:n) = u(:n)/smax
        beta = beta/smax
        cl = sqrt( dot_product( u(:n), u(:n) ) + beta*beta )
!    
!       COMPUTE THE HOUSEHOLDER TRANSFORMATION H .
!    
        cl = -sign( cl, beta )
        up = beta - cl
        tau = one/( up*cl )
!    
!       COMPUTE THE SCALAR beta .
!    
        beta = cl*smax
!    
        m = size( vec_piv )
!
        if ( left ) then
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION H TO C FROM THE LEFT.
!    
            tmp(:m) = matmul( u(:n), mat(:n,:m) )
            tmp(:m) = ( vec_piv(:m)*up + tmp(:m) )*tau
!
            do i=1_i4b, m
#ifdef _BLAS
                call axpy( n, tmp(i), u(:n), 1_i4b, mat(:n,i), 1_i4b )
#else
                mat(:n,i) = mat(:n,i) + tmp(i)*u(:n)
#endif
            end do
!
        else
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION H TO C FROM THE RIGHT.
!
            tmp(:m) = matmul( mat(:m,:n), u(:n) )
            tmp(:m) = ( vec_piv(:m)*up + tmp(:m) )*tau
!
            do i=1_i4b, n
#ifdef _BLAS
                call axpy( m, u(i), tmp(:m), 1_i4b, mat(:m,i), 1_i4b )
#else
                mat(:m,i) = mat(:m,i) + u(i)*tmp(:m)
#endif
            end do
!
        end if
!
        vec_piv(:m) = vec_piv(:m) + up*tmp(:m)
!
    else
!    
!       H IS THE IDENTITY MATRIX.
!    
        tau = zero
        up = zero
!
    end if
!
!
! END OF SUBROUTINE h2_rm
! _______________________
!
    end subroutine h2_rm
!
! =========================================================================================
!
    subroutine apply_h1_rv( u, tau, vec )
!
! Purpose
! _______
!
!   APPLY_H1 applies a real elementary reflector H generated by H1
!   to a real vector  C  . H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!
! Arguments
! _________
!
!   U        (INPUT) real(stnd), dimension(:) 
!            On entry, U contains the vector v of the Householder reflector,
!            as generated by H1.
!
!   TAU      (INPUT) real(stnd)
!            On entry, the scalar tau of the Householder reflector,
!            as generated by H1.
!
!   VEC      (INPUT/OUTPUT) real(stnd), dimension(:) 
!            On entry, the real vector  C .
!
!            On exit, C is overwritten by the vector H * C .
!
!
! Further Details
! _______________
!
!   It is assumed that size(VEC) >= size(U) .
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)    :: u(:), tau
    real(stnd), intent(inout) :: vec(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau>=zero ) return
!
    n = size( u )
!
    tmp = tau*dot_product( u(:n), vec(:n) )
!
#ifdef _BLAS
    call axpy( n, tmp, u(:n), 1_i4b, vec(:n), 1_i4b )
#else
    vec(:n) = vec(:n) + tmp*u(:n)
#endif
!
!
! END OF SUBROUTINE apply_h1_rv
! _____________________________
!
    end subroutine apply_h1_rv
!
! =========================================================================================
!
    subroutine apply_h1_rm( u, tau, mat, left )
!
! Purpose
! _______
!
!   APPLY_H1 applies a real elementary reflector H generated by H1
!   to a real n by m or m by n matrix, C, from the left or the right.
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!
! Arguments
! _________
!
!   U       (INPUT) real(stnd), dimension(:) 
!           On entry, U contains the vector v of the Householder reflector,
!           as generated by H1.
!
!   TAU     (INPUT) real(stnd)
!           On entry, the scalar tau of the Householder reflector,
!           as generated by H1.
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:) 
!           On entry, the n by m or m by n matrix C.
!
!           On exit, C is overwritten by the matrix H * C (LEFT=true) or
!           C * H (LEFT=false).
!
!   LEFT    (INPUT) logical(lgl)
!           If:
!
!           - LEFT=true,  H is  applied to the real matrix C from the left
!           - LEFT=false, H is  applied to the real matrix C from the right.
!
!
! Further Details
! _______________
!
!   It is assumed that:
!
!   - size(MAT,1) >= size(U) if LEFT=true
!   - size(MAT,2) >= size(U) if LEFT=false.
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)    :: u(:), tau
    real(stnd), intent(inout) :: mat(:,:)
!
    logical(lgl), intent(in)  :: left
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp( (size(mat,1)+size(mat,2))-size(u) )
!
    integer(i4b) :: n, m, i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau>=zero ) return
!
    n = size( u )
!
    if ( left ) then
!
        m = size( mat, 2 )
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO mat(:,:) FROM THE LEFT.
!
        tmp(:m) = matmul( u(:n), mat(:n,:m) )
        tmp(:m) = tmp(:m)*tau
!
        do i=1_i4b, m
#ifdef _BLAS
            call axpy( n, tmp(i), u(:n), 1_i4b, mat(:n,i), 1_i4b )
#else
            mat(:n,i) = mat(:n,i) + tmp(i)*u(:n)
#endif
        end do
!
    else
!
        m = size( mat, 1 )
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO mat(:,:) FROM THE RIGHT.
!
        tmp(:m) = matmul( mat(:m,:n), u(:n) )
        tmp(:m) = tmp(:m)*tau
!
        do i=1_i4b, n
#ifdef _BLAS
            call axpy( m, u(i), tmp(:m), 1_i4b, mat(:m,i), 1_i4b )
#else
            mat(:m,i) = mat(:m,i) + u(i)*tmp(:m)
#endif
        end do
!
    end if
!
!
! END OF SUBROUTINE apply_h1_rm
! _____________________________
!
    end subroutine apply_h1_rm
!
! =========================================================================================
!
    subroutine apply_h2_rv( u, up, tau, piv, vec )
!
! Purpose
! _______
!
!   APPLY_H2 applies a real elementary reflector H generated by H2
!   to a real vector C . H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!
! Arguments
! _________
!
!   U      (INPUT) real(stnd), dimension(:) 
!          On entry, U contains the vector v(2:) of the Householder reflector,
!          as generated by H2.
!
!   UP     (INPUT) real(stnd) 
!          On entry, the value v(1) of the Householder reflector,
!          as generated by H2.
!
!   TAU    (INPUT) real(stnd)
!          On entry, the scalar tau of the Householder reflector,
!          as generated by H2.
!
!   PIV    (INPUT/OUTPUT) real(stnd)
!          On entry, the scalar C[1].
!
!          On exit, PIV is overwritten by the scalar (H * C)[1].
!
!   VEC    (INPUT/OUTPUT) real(stnd), dimension(:) 
!          On entry, the real vector  C[2:].
!
!          On exit, C is overwritten by the vector (H * C)[2:].
!
!
! Further Details
! _______________
!
!   It is assumed that size(VEC) >= size(U) .
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)     :: u(:), up, tau
    real(stnd),  intent(inout)  :: piv, vec(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau>=zero ) return
!
    n = size( u )
!
    tmp = dot_product( u(:n), vec(:n) )
    tmp = ( piv*up + tmp )*tau
!
#ifdef _BLAS
    call axpy( n, tmp, u(:n), 1_i4b, vec(:n), 1_i4b )
#else
    vec(:n) = vec(:n) + tmp*u(:n)
#endif
!
    piv = piv + up*tmp
!
!
! END OF SUBROUTINE apply_h2_rv
! _____________________________
!
    end subroutine apply_h2_rv
!
! =========================================================================================
!
    subroutine apply_h2_rm( u, up, tau, vec_piv, mat, left )
!
! Purpose
! _______
!
!   APPLY_H2 applies a real elementary reflector H generated by H2
!   to a real n by m or m by n matrix, C, from the left or the right.
!   H is represented in the form
!
!           H = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector.
!
!
! Arguments
! _________
!
!   U          (INPUT) real(stnd), dimension(:) 
!              On entry, U contains the vector v(2:) of the Householder reflector,
!              as generated by H2.
!
!   UP         (INPUT) real(stnd) 
!              On entry, the value v(1) of the Householder reflector,
!              as generated by H2.
!
!   TAU        (INPUT) real(stnd)
!              On entry, the scalar tau of the Householder reflector,
!              as generated by H2.
!
!   VEC_PIV    (INPUT/OUTPUT) real(stnd), dimension(:)
!              If LEFT=true:
!
!              - On entry, the row_vector C[1,:]
!              - On exit, VEC_PIV is overwritten by the vector (H * C)[1,:].
!
!              If LEFT=false:
!
!              - On entry, the column_vector C[:,1]
!              - On exit, VEC_PIV is overwritten by the vector (C * H)[:,1].
!
!   MAT        (INPUT/OUTPUT) real(stnd), dimension(:,:) 
!              If LEFT=true:
!
!              - On entry, the n-1 by m matrix C[2:,:]
!              - On exit, MAT is overwritten by the matrix  (H * C)[2:,:] .
!
!              If LEFT=false:
!
!              - On entry, the m by n-1 matrix C[:,2:]
!              - On exit, MAT is overwritten by the matrix (C * H)[:,2:].
!
!   LEFT       (INPUT) logical(lgl)
!              If:
!
!              - LEFT=true,  H is  applied to the real matrix C from the left
!              - LEFT=false, H is  applied to the real matrix C from the right.
!
!
! Further Details
! _______________
!
!   It is assumed that:
!
!   - size(MAT,1) >= size(U) and  size(MAT,2) >= size(VEC_PIV) if LEFT=true
!   - size(MAT,2) >= size(U) and  size(MAT,1) >= size(VEC_PIV) if LEFT=false.
!
!   This subroutine is adapted from
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)     :: u(:), up, tau
    real(stnd),  intent(inout)  :: vec_piv(:), mat(:,:)
!
    logical(lgl), intent(in)    :: left
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp(size(vec_piv))
!
    integer(i4b) :: n, m, i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( tau>=zero ) return
!
    n = size( u )
    m = size( vec_piv )
!
    if ( left ) then
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO C FROM THE LEFT.
!    
        tmp(:m) = matmul( u(:n), mat(:n,:m) )
        tmp(:m) = ( vec_piv(:m)*up + tmp(:m) )*tau
!
        do i=1_i4b, m
#ifdef _BLAS
            call axpy( n, tmp(i), u(:n), 1_i4b, mat(:n,i), 1_i4b )
#else
            mat(:n,i) = mat(:n,i) + tmp(i)*u(:n)
#endif
        end do
!
    else
!    
!       APPLY THE HOUSEHOLDER TRANSFORMATION H TO C FROM THE RIGHT.
!
        tmp(:m) = matmul( mat(:m,:n), u(:n) )
        tmp(:m) = ( vec_piv(:m)*up + tmp(:m) )*tau
!
        do i=1_i4b, n
#ifdef _BLAS
            call axpy( m, u(i), tmp(:m), 1_i4b, mat(:m,i), 1_i4b )
#else
            mat(:m,i) = mat(:m,i) + u(i)*tmp(:m)
#endif
        end do
!
    end if
!
    vec_piv(:m) = vec_piv(:m) + up*tmp(:m)
!
!
! END OF SUBROUTINE apply_h2_rm
! _____________________________
!
    end subroutine apply_h2_rm
!
! =========================================================================================
!
!
! *******************************
! END OF MODULE Hous_Procedures *
! *******************************
!    
end module Hous_Procedures
