module QR_Procedures
! ###################################################################################
!
! Copyright 2021 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR COMPUTING QR AND LQ DECOMPOSITIONS. *
!                                                                                    *
! LATEST REVISION : 20/12/2021                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!                                                                                    *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                                    *
!                                                                                    *
!  _OPENMP      FOR ACTIVATING OPENMP PARALLELIZATION                                *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                         *
!  _DOT_PRODUCT FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH STATPACK       *
!               dot_product2 FUNCTION                                                *
!  _MATMUL      FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK            *
!               matmul2 FUNCTION                                                     *
!  _TRANSPOSE   FOR REPLACING THE transpose INTRINSIC FUNCTION WITH STATPACK         *
!               transpose2 FUNCTION                                                  *
!  _WHERE       FOR REPLACING SOME where CONSTRUCTS BY do LOOPS INSIDE SOME OPENMP   *
!               DIRECTIVES (USEFUL FOR SOME COMPILERS LIKE IFORT WHEN OPENMP IS      *
!               USED)                                                                *
!  _USE_INTEL   FOR SIGNALING THAT THE INTEL IFORT COMPILER IS USED.                 *
!               HERE, THIS INCLUDES ONLY THE ACTIVATION OF THE _WHERE MACRO          *
!                                                                                    *
! ************************************************************************************
!
#ifdef _USE_INTEL
#define _WHERE
#endif
!
!
! USED MODULES
! ============
!   
    use Select_Parameters,     only : lgl, i4b, stnd
    use Reals_Constants,       only : zero, one, ten
#ifdef _MATMUL
    use Utilities,             only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,             only : dot_product=>dot_product2
#endif
#ifdef _TRANSPOSE
    use Utilities,             only : transpose=>transpose2
#endif
#ifdef _OPENMP
    use Select_Parameters,     only : omp_limit, omp_limit2
    use omp_lib,               only : omp_lock_kind, omp_get_max_threads, omp_in_parallel, omp_get_nested,  &
                                      omp_get_num_threads, omp_get_thread_num, omp_init_lock, omp_set_lock, &
                                      omp_unset_lock, omp_destroy_lock
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : axpy
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!   
!   use Logical_Constants
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use Hous_Procedures
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   lq_cmp, qr_cmp, qrfac, qr_cmp2,                                  &
                ortho_gen_lq, ortho_gen_qr, apply_q_lq, apply_q_qr
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
!
    interface apply_q_qr
        module procedure    apply_q_qr_rv, apply_q_qr_rm
    end interface
!
    interface apply_q_lq
        module procedure    apply_q_lq_rv, apply_q_lq_rm
    end interface
!
!   
! MODULE PARAMETERS 
! =================
!
! BLOCK SIZE FOR APPLYING HOUSEHOLDER TRANSFORMATIONS
! IN lq_cmp, qr_cmp, apply_q_lq, apply_q_qr, ortho_gen_lq AND ortho_gen_qr SUBROUTINES.
!
    integer(i4b), parameter :: blksz=20_i4b
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
!
! =========================================================================================
!                              LQ DECOMPOSITION
! =========================================================================================
!
!
    subroutine lq_cmp( mat, diagl, tau, use_qr )
!
! Purpose
! _______
!
!   LQ_CMP computes a LQ factorization of a real m-by-n matrix MAT :
!
!                       MAT = L * Q
!
!   where Q is orthogonal and L is lower trapezoidal (lower triangular if m<=n).
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real matrix to be decomposed.
!
!            On exit, the elements below the diagonal of the array contain  
!            the corresponding elements of L. The elements on and above the 
!            diagonal, with the array TAU, represent the orthogonal matrix Q
!            as a product of elementary reflectors, see Further Details.
!
!   DIAGL    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix L.
!
!            The size of DIAGL must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAU      (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors.
!            see Further Details.
!
!            The size of TAU must be min( size(MAT,1) , size(MAT,2) ).
!
!   USE_QR   (INPUT, OPTIONAL) logical(lgl)
!            If the optional argument USE_QR is set to true, the input matrix is
!            transposed, a fast QR decomposition is used for computing the LQ
!            decomposition and the results are transposed again in the input matrix.
!            The default is true.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!           Q = H(k) * . . . * H(2) * H(1), where k = min( size(MAT,1) , size(MAT,2) )
!
!   Each H(i) has the form
!
!            H(i) = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector with v(1:i-1) = 0.
!   v(i:n) is stored on exit in MAT(i,i:n) and tau in TAU(i).

!   A blocked algorithm is used for computing the LQ factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm used here , see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!            Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!            The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989: Block
!           reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988: Implementation of the GMRES method using Householder
!           transformations. Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert_eq, merror
    use Hous_Procedures,  only : hous1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagl, tau
!
    logical(lgl), intent(in), optional :: use_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                               :: prod, tau1
    real(stnd), dimension(:,:), allocatable  :: lmat, umat, x
!
    integer(i4b) :: m, n, k, k2, i, i1, i2, j, l, l1, p, q, nblk, rem
    integer      :: iok
!
    logical(lgl) :: useqr
!
#ifdef _OPENMP
!
    real(stnd), dimension(:),   allocatable :: x2
    real(stnd), dimension(:,:), allocatable :: umat2, mat2
!
    integer(i4b) :: iblk, nblk2, j0, j1, j2, q2, num_threads, thread_num
!
    logical      :: test_par, ompparallel, ompnested
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lq_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),              &
                   int(size(tau),i4b),    &
                   int(size(diagl),i4b),  &
                   name_proc              )
!
    useqr = true
!
    if ( present(use_qr) ) then
        useqr = use_qr
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    if ( useqr ) then
!
!       ALLOCATE WORK VARIABLE FOR A FAST QR ALGORITHM IF POSSIBLE.
!
        allocate( x(n,m), stat = iok )
!
    else
!
!       USE AN LQ ALGORITHM AUTOMATICALLY.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST QR DECOMPOSITION.
!
        x(:n,:m) = transpose( mat(:m,:n) )
!
        call qr_cmp( x(:n,:m), diagl, tau )
!
        mat(:m,:n) = transpose( x(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x )
!
    else
!
!       USE A SLOWER LQ ALGORITHM IF THE INPUT MATRIX IS TOO BIG.
!
!       COMPUTE THE NUMBER OF BLOCKS.
!
        nblk = k/blksz
        rem  = mod( k, blksz )
!
        if ( rem/=0_i4b ) then
!
            nblk = nblk + 1_i4b
            k2   = k - rem
!
        else
!
            k2   = k - blksz
!
        end if
!
#ifdef _OPENMP
!
!       COMPUTE THE NUMBER OF ROW BLOCKS.
!
        nblk2 = m/blksz
        rem  = mod( m, blksz )
!
        if ( rem/=0_i4b ) nblk2 = nblk2 + 1_i4b
!
        i           = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*m>=omp_limit                      .and.      &
                      k>blksz                             .and.      &
                      nblk2>1_i4b                         .and.      &
                      i>1_i4b
!
        if ( test_par ) then
!
!           ALLOCATE SHARED WORK VARIABLES.
!
            allocate( umat(n,blksz), umat2(n,blksz), lmat(blksz,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
!           PARALLEL AND BLOCKED LQ DECOMPOSITION OF mat.
!
!$OMP PARALLEL PRIVATE(iok,i,i1,i2,iblk,j,j0,j1,j2,l,l1,p,q,q2,num_threads,thread_num,tau1,prod,x2,mat2)
!
!           ALLOCATE PRIVATE WORK VARIABLES.
!
            allocate( mat2(n,blksz), x2(blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
            if ( thread_num==0_i4b ) then
!
!               COMPUTE LQ FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
                umat(1_i4b:n,1_i4b:blksz) = transpose( mat(1_i4b:blksz,1_i4b:n) )
!
                do l = 1_i4b, blksz
!    
!                   GENERATE ELEMENTARY REFLECTOR H(l) TO ANNIHILATE mat(l,l+1:n) .
!    
                    call hous1( umat(l:n,l), tau(l), diagl(l) )
!    
!                   APPLY H(l) TO mat(l+1:blksz,l:n) FROM THE RIGHT.
!
                    tau1 = tau(l)
!
                    do j = l + 1_i4b, blksz
!
                        prod        = tau1*dot_product( umat(l:n,l), umat(l:n,j) )
#ifdef _BLAS
                        call axpy( n-l+1_i4b, prod, umat(l:n,l), 1_i4b, umat(l:n,j), 1_i4b )
#else
                        umat(l:n,j) = umat(l:n,j) + prod*umat(l:n,l)
#endif
!    
                    end do
!    
                end do
!
                mat(1_i4b:blksz,1_i4b:n) = transpose( umat(1_i4b:n,1_i4b:blksz) )
!
            end if
!
!$OMP BARRIER
!
!           START THE BLOCKED LQ FACTORIZATION.
!
            do i = 0_i4b, k2, blksz
!    
                i2 = i + blksz
!
                if ( m>i2 ) then
!    
                    i1 = i + 1_i4b
                    q  = min( i2, k ) - i
!    
!                   APPLY H(i+1)* ...*H(i+blksz) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!                   WITH A BLOCKED ALGORITHM AS
!
!                mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
!                   First, COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, q
!
                        l1 = i + l
!
                        umat2(l1:n,l) = umat(l1:n,l)                      
!
                        do j = l + 1_i4b, q
!
                            l1 = i + j
!
                            lmat(j,l) = -tau(l1)*dot_product( umat(l1:n,l), umat(l1:n,j) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
!                   COMPUTE THE PREVIOUS BLOCK NUMBER.
!
                    iblk = i/blksz + 1_i4b
!
!                   EACH THREAD WILL TOUCH EXACTLY THE SAME ROWS BLOCKS DURING THE ITERATIONS.
!
                    p = ( (iblk/num_threads)*num_threads ) + thread_num
                    if ( thread_num<mod(iblk,num_threads) ) p = p + num_threads
!
                    do j = p, nblk2-1_i4b, num_threads
!
!                       COMPUTE INDICES FOR THE FIRST AND LAST ROWS OF THE j+1 BLOCK.
!
                        j0 = j*blksz
                        j1 = j0 + 1_i4b
                        j2 = min( j0+blksz, m )
!
                        q2 = j2 - j0
!
                        mat2(i1:n,1_i4b:q2) = transpose( mat(j1:j2,i1:n) )
!
                        do p = 1_i4b, q2
!
!                           COMPUTE mat(p,i+1:n)*U(i+1:n,:blksz).
!
                            do l = 1_i4b, q
!
                                l1 = i + l
!
                                x2(l) = tau(l1)*dot_product( umat2(l1:n,l), mat2(l1:n,p) )
!                    
                            end do
!
!                           SOLVE x2(:blksz)*lmat(:blksz,:blksz) = mat(p,i+1:n)*U(i+1:n,:blksz) AND
!                           COMPUTE mat(p,i+1:n) = mat(p,i+1:n) + x2(:blksz)*U(i+1:n,:blksz)' .
!
                            do l = 1_i4b, q
!
                                x2(l+1_i4b:q) = x2(l+1_i4b:q) - x2(l)*lmat(l+1_i4b:q,l)
!
                                l1 = i + l
!
#ifdef _BLAS
                                call axpy( n-l1+1_i4b, x2(l), umat2(l1:n,l), 1_i4b, mat2(l1:n,p), 1_i4b )
#else
                                mat2(l1:n,p) = mat2(l1:n,p) + x2(l)*umat2(l1:n,l)
#endif
!
                            end do
!
                        end do
!
                        if ( j==iblk .and. j1<=k ) then
!
!                           COMPUTE THE LQ FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(i+blksz+1:i+2*blksz,i+blksz+1:n), IN PLACE.
!
                            do l = 1_i4b, min(q2,k-j0)
!    
                                l1 = j0 + l
!    
!                               GENERATE ELEMENTARY REFLECTOR H(l1) TO ANNIHILATE mat(l1,l1+1:n) .
!                    
                                call hous1( mat2(l1:n,l), tau(l1), diagl(l1) )
!
                                umat(l1:n,l) = mat2(l1:n,l)
!    
!                               APPLY H(l1) TO mat(l1+1:i+2*blksz,l1:n) FROM THE RIGHT.
!
                                tau1 = tau(l1)
!
                                do p = l + 1_i4b, q2
!
                                    prod         = tau1*dot_product( mat2(l1:n,l), mat2(l1:n,p) )
#ifdef _BLAS
                                    call axpy( n-l1+1_i4b, prod, mat2(l1:n,l), 1_i4b, mat2(l1:n,p), 1_i4b )
#else
                                    mat2(l1:n,p) = mat2(l1:n,p) + prod*mat2(l1:n,l)
#endif
!
                                end do
!
                            end do
!
                        end if
!
                        mat(j1:j2,i1:n) = transpose( mat2(i1:n,1_i4b:q2) )
!
                    end do
!
                end if
!
!$OMP BARRIER
!
            end do
!
!           DEALLOCATE PRIVATE WORK VARIABLES.
!
            deallocate( x2, mat2 )
!
!$OMP END PARALLEL
!
!           DEALLOCATE SHARED WORK VARIABLES.
!
            deallocate( lmat, umat, umat2 )
!
        else
!
#endif
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( lmat(blksz,blksz), umat(n,blksz), x(m,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
!           BLOCKED LQ DECOMPOSITION OF mat.
!
            do i = 0_i4b, k2, blksz
!
                i1 = i + 1_i4b
                i2 = min( i + blksz, k )
!
                q = i2 - i
!
                umat(i1:n,1_i4b:q) = transpose( mat(i1:i2,i1:n) )
!
                do l = 1_i4b, q
!    
                    l1 = i + l
!    
!                   GENERATE ELEMENTARY REFLECTOR H(i+l) TO ANNIHILATE mat(i+l,i+l+1:n) .
!    
                    call hous1( umat(l1:n,l), tau(l1), diagl(l1) )
!    
!                   APPLY H(i+l) TO mat(i+l+1:i+blksz,i+l:n) FROM THE RIGHT.
!
                    tau1 = tau(l1)
!
                    do j = l + 1_i4b, q
!
                        prod         = tau1*dot_product( umat(l1:n,l), umat(l1:n,j) )
#ifdef _BLAS
                        call axpy( n-l1+1_i4b, prod, umat(l1:n,l), 1_i4b, umat(l1:n,j), 1_i4b )
#else
                        umat(l1:n,j) = umat(l1:n,j) + prod*umat(l1:n,l)
#endif
!    
                    end do
!    
                end do
!
                mat(i1:i2,i1:n) = transpose( umat(i1:n,1_i4b:q) )
!    
                if ( m>i2 ) then
!    
!                   APPLY H(i+1)* ...*H(i+blksz) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!                   WITH A BLOCKED ALGORITHM AS
!
!              mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
!                   FIRST, COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                    do l = 1_i4b, q - 1_i4b
!
                        do j = l + 1_i4b, q
!
                            l1 = i + j
!
                            lmat(j,l) = -tau(l1)*dot_product( umat(l1:n,l), umat(l1:n,j) )
!
                        end do
!                    
                    end do
!
                    j = i2 + 1_i4b
!
!                   COMPUTE mat(j:m,i+1:n)*U(i+1:n,:blksz).
!
                    do l = 1_i4b, q
!
                        l1 = i + l
!
                        x(j:m,l) = tau(l1)*matmul( mat(j:m,l1:n), umat(l1:n,l) )
!                    
                    end do
!
!                   SOLVE x(:blksz,:blksz)*lmat(:blksz,:blksz) = mat(j:m,i+1:n)*U(i+1:n,:blksz) AND
!                   COMPUTE mat(j:m,i+1:n) = mat(j:m,i+1:n) + x(:blksz,:blksz)*U(i+1:n,:blksz)' .
!
                    do l = 1_i4b, q
!
                        do p = l + 1_i4b, q
!
#ifdef _BLAS
                            call axpy( m-j+1_i4b, -lmat(p,l), x(j:m,l), 1_i4b, x(j:m,p), 1_i4b )
#else
                            x(j:m,p) = x(j:m,p) - x(j:m,l)*lmat(p,l)
#endif
!
                            l1 = i + p - 1_i4b
!
#ifdef _BLAS
                            call axpy( m-j+1_i4b, umat(l1,l), x(j:m,l), 1_i4b, mat(j:m,l1), 1_i4b )
#else
                            mat(j:m,l1) = mat(j:m,l1) + x(j:m,l)*umat(l1,l)
#endif
!
                        end do
!
                    end do
!
                    do p = i + q, n
!
                        do l = 1_i4b, q
!
#ifdef _BLAS
                            call axpy( m-j+1_i4b, umat(p,l), x(j:m,l), 1_i4b, mat(j:m,p), 1_i4b )
#else
                            mat(j:m,p) = mat(j:m,p) + x(j:m,l)*umat(p,l)
#endif
!        
                        end do
!                    
                    end do
!
                end if
!
            end do
!
!           DEALLOCATE WORK VARIABLES.
!
            deallocate( lmat, umat, x )
!
#ifdef _OPENMP
        end if
#endif
!
    end if
!
!
! END OF SUBROUTINE lq_cmp
! ________________________
!
    end subroutine lq_cmp
!
! =========================================================================================
!
    subroutine ortho_gen_lq( mat, tau, use_qr )
!
! Purpose
! _______
!
!   ORTHO_GEN_LQ generates an m-by-n real matrix with orthonormal rows, which is
!   defined as the first m rows of a product of k elementary reflectors of order n
!
!           Q  =  H(k) * . . . * H(2) * H(1)
!
!   as returned by LQ_CMP.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the i-th row must contain the vector which
!            defines the elementary reflector H(i), for i = 1,2,...,k,
!            as returned by LQ_CMP in the first k rows of its array
!            argument MAT.
!
!            On exit, the first m rows of Q.
!
!            The shape of MAT must verify size( MAT, 1 ) <= size( MAT, 2 ).
!
!   TAU      (INPUT) real(stnd), dimension(:)
!            TAU(i) must contain the scalar factor of the elementary
!            reflector H(i), as returned by LQ_CMP.
!            The size of TAU determines the number k of elementary reflectors
!            whose product defines the matrix Q.
!
!            The size of TAU must verify   size( TAU )  <= ( MAT, 1 ) .
!
!   USE_QR   (INPUT, OPTIONAL) logical(lgl)
!            If the optional argument USE_QR is set to true, the input matrix is
!            transposed, a fast QR type algorithm is used for computing the first m
!            rows of Q and the results are transposed again in the input matrix.
!            The default is true.
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the upper triangle
!   of MAT and generating the orthogonal matrix Q of the LQ factorization.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!          Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!           The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989: Block
!          reduction of matrices to condensed form for eigenvalue computations. 
!          J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988: Implementation of the GMRES method using Householder
!          transformations. Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tau
!
    logical(lgl), intent(in), optional :: use_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: prod, tau1
    real(stnd), dimension(:),   allocatable :: x2
    real(stnd), dimension(:,:), allocatable :: lmat, umat, x, tmp
!
    integer(i4b) :: m, n, k, i, i0, i1, i2, l, l2, j, blksz2, p
    integer      :: iok
!
    logical(lgl) :: useqr
!
#ifdef _OPENMP
!
    integer(i4b) :: j1, j2, blksz3
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_lq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
    k = size( tau )
!
    call assert( logical( (k<=m) .and. ( m<=n), lgl ), name_proc )
!
    useqr = true
!
    if ( present(use_qr) ) then
        useqr = use_qr
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    if ( useqr ) then
!
!       ALLOCATE WORK VARIABLE FOR A FAST QR-TYPE ALGORITHM IF POSSIBLE.
!
        allocate( x(n,m), stat = iok )
!
    else
!
!       USE AN LQ-TYPE ALGORITHM AUTOMATICALLY.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST QR-TYPE ALGORITHM.
!
        x(:n,:m) = transpose( mat(:m,:n) )
!
        call ortho_gen_qr( x(:n,:m), tau )
!
        mat(:m,:n) = transpose( x(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x )
!
    else
!
!       USE A SLOWER LQ-TYPE ALGORITHM IF THE INPUT MATRIX IS TOO BIG.
!
!       INITIALIZE ROWS k+1:m OF mat TO ROWS OF THE UNIT MATRIX.
!
        mat(k+1_i4b:m,1_i4b:n) = zero
!
        do i = k+1_i4b, min( n, m )
            mat(i,i) = one
        end do
!
#ifdef _OPENMP
        i = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*m>=omp_limit                      .and.      &
                      m/i>=omp_limit2                     .and.      &
                      i>1_i4b
!
        if ( test_par ) then
!
!           ALLOCATE SHARED WORK VARIABLES.
!
            allocate( umat(n,blksz), lmat(blksz,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
!           PARALLEL AND BLOCKED LQ-TYPE ALGORITHM.
!
!$OMP PARALLEL PRIVATE(iok,i,i0,i1,i2,l,l2,j,j1,j2,p,blksz2,blksz3,tau1,prod,x2,tmp)
!
!           ALLOCATE PRIVATE WORK VARIABLES.
!
            allocate( tmp(n,blksz), x2(blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            do i = k - blksz, -blksz+1_i4b, -blksz
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!               WITH A BLOCKED ALGORITHM AS
!
!           mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                blksz2 = min( blksz, i+blksz )
                i0     = max( 0_i4b, i )
!
!               SAVE H(i+1) ...H(i+blksz) FOR LATER USE.
!    
!$OMP SINGLE
!
                umat(i0+1_i4b:n,1_i4b:blksz2) = transpose( mat(i0+1_i4b:i0+blksz2,i0+1_i4b:n) )
!
!$OMP END SINGLE
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 2_i4b, blksz2
!
                    i1 = i0 + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i0 + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:n,l), umat(i1:n,l2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC)
!
                do j = i0+blksz2, m-1_i4b, blksz
!
                    j1     = j + 1_i4b
                    j2     = min( j + blksz, m)
                    blksz3 = j2-j1+1_i4b
!
                    tmp(i0+1_i4b:n,1_i4b:blksz3) = transpose( mat(j1:j2,i0+1_i4b:n) )
!
                    do p = 1_i4b, blksz3
!
!                       COMPUTE mat(j+p,i+1:n)*U(i+1:n,:blksz) .
!
                        do l = 1_i4b, blksz2
!
                            i1 = i0 + l
!
                            x2(l) = tau(i1)*dot_product( umat(i1:n,l), tmp(i1:n,p) )
!                    
                        end do
!
!                       SOLVE x2(:blksz)*lmat(:blksz,:blksz) = mat(j+p,i+1:n)*U(i+1:n,:blksz) AND
!                       COMPUTE mat(j+p,i+1:n) = mat(j+p,i+1:n) + x2(:blksz)*U(i+1:n,:blksz)' .
!
                        do l = blksz2, 1_i4b, -1_i4b
!
                            x2(1_i4b:l-1_i4b) = x2(1_i4b:l-1_i4b) - x2(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i0 + l
!
#ifdef _BLAS
                            call axpy( n-i1+1_i4b, x2(l), umat(i1:n,l), 1_i4b, tmp(i1:n,p), 1_i4b )
#else
                            tmp(i1:n,p) = tmp(i1:n,p) + x2(l)*umat(i1:n,l)
#endif
!
                        end do
!
                    end do
!
                    mat(j1:j2,i0+1_i4b:n) = transpose( tmp(i0+1_i4b:n,1_i4b:blksz3) )
!                    
                end do
!
!$OMP END DO NOWAIT
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+1:i+blksz,i+1:n) FROM THE RIGHT
!               WITH A SERIAL ALGORITHM.
!    
!$OMP SINGLE
!
                do l = blksz2, 1_i4b, -1_i4b
!
                    i1 = i0 + l
!
                    tau1 = tau(i1)
!
!                   APPLY H(i1) TO mat(i1+1:i+blksz,i1:n) FROM THE RIGHT.
!
                    do l2 = l + 1_i4b, blksz2
!
                        prod = tau1*dot_product( umat(i1+1_i4b:n,l), tmp(i1+1_i4b:n,l2) )
#ifdef _BLAS
                        call axpy( n-i1+1_i4b, prod, umat(i1:n,l), 1_i4b, tmp(i1:n,l2), 1_i4b )
#else
                        tmp(i1:n,l2) = tmp(i1:n,l2) + prod*umat(i1:n,l)
#endif
!    
                    end do
!
!                   APPLY H(i1) TO mat(i1,i1:n) FROM THE RIGHT.
!
                    tmp(1_i4b:i1-1_i4b,l) = zero
                    tmp(i1,l)             = one + tau1
                    tmp(i1+1_i4b:n,l)     = tau1*umat(i1+1_i4b:n,l)
!    
                end do
!    
                mat(i0+1_i4b:i0+blksz2,1_i4b:n) = transpose( tmp(1_i4b:n,1_i4b:blksz2) )
!    
!$OMP END SINGLE
!
            end do
!
!           DEALLOCATE PRIVATE WORK VARIABLES.
!
            deallocate( x2, tmp )
!
!$OMP END PARALLEL
!
!           DEALLOCATE SHARED WORK VARIABLES.
!
            deallocate( lmat, umat )
!
        else
#endif
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( umat(n,blksz), tmp(n,blksz), lmat(blksz,blksz), x(m,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
            tmp(1_i4b:n,1_i4b:blksz) = zero
!
!           BLOCKED LQ-TYPE ALGORITHM.
!
            do i = k - blksz, -blksz+1_i4b, -blksz
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!               WITH A BLOCKED ALGORITHM AS
!
!           mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                blksz2 = min( blksz, i+blksz )
                i0     = max( 0_i4b, i )
!
!               SAVE H(i+1) ...H(i+blksz) FOR LATER USE.
!
                umat(i0+1_i4b:n,1_i4b:blksz2) = transpose( mat(i0+1_i4b:i0+blksz2,i0+1_i4b:n) )
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz2
!
                    i1 = i0 + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i0 + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:n,l), umat(i1:n,l2) )
!
                    end do
!                    
                end do
!
                p = i0 + blksz2 + 1_i4b
!
!               COMPUTE mat(p:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz2
!
                    i1 = i0 + l
!
                    x(p:m,l) = tau(i1)*matmul( mat(p:m,i1:n), umat(i1:n,l) )
!                    
                end do
!
!               SOLVE x(p:m,:blksz)*lmat(:blksz,:blksz) = mat(p:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE mat(p:m,i+blksz-1:n) = mat(p:m,i+blksz-1:n) + x(p:m,:blksz)*U(i+blksz-1:n,:blksz)' .
!
                do l = blksz2, 1_i4b, -1_i4b
!
                    do l2 = 1_i4b, l-1_i4b
!
#ifdef _BLAS
                        call axpy( m-p+1_i4b, -lmat(l2,l), x(p:m,l), 1_i4b, x(p:m,l2), 1_i4b )
#else
                        x(p:m,l2) = x(p:m,l2) - x(p:m,l)*lmat(l2,l)
#endif
!
                    end do
!
                    do l2 = l, blksz2-1_i4b
!
                        i2 = i0 + l2
!
#ifdef _BLAS
                        call axpy( m-p+1_i4b, umat(i2,l), x(p:m,l), 1_i4b, mat(p:m,i2), 1_i4b )
#else
                        mat(p:m,i2) = mat(p:m,i2) + x(p:m,l)*umat(i2,l)
#endif
!
                    end do
!
                end do
!
!               COMPUTE mat(p:m,i+blksz:n) = mat(p:m,i+blksz:n) + x(p:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i0 + blksz2, n
!
                    do l = 1_i4b, blksz2
!
#ifdef _BLAS
                        call axpy( m-p+1_i4b, umat(j,l), x(p:m,l), 1_i4b, mat(p:m,j), 1_i4b )
#else
                        mat(p:m,j) = mat(p:m,j) + x(p:m,l)*umat(j,l)
#endif
!                    
                    end do
!                    
                end do
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+1:i+blksz,i+1:n) FROM THE RIGHT
!               WITH A SERIAL ALGORITHM.
!
                do l = blksz2, 1_i4b, -1_i4b
!
                    i1 = i0 + l
!
                    tau1 = tau(i1)
!
!                   APPLY H(i1) TO mat(i1+1:i+blksz,i1:n) FROM THE RIGHT.
!
                    do l2 = l + 1_i4b, blksz2
!
                        prod = tau1*dot_product( umat(i1+1_i4b:n,l), tmp(i1+1_i4b:n,l2) )
#ifdef _BLAS
                        call axpy( n-i1+1_i4b, prod, umat(i1:n,l), 1_i4b, tmp(i1:n,l2), 1_i4b )
#else
                        tmp(i1:n,l2) = tmp(i1:n,l2) + prod*umat(i1:n,l)
#endif
!    
                    end do
!
!                   APPLY H(i1) TO mat(i1,i1:n) FROM THE RIGHT.
!
                    tmp(i1,l)         = one + tau1
                    tmp(i1+1_i4b:n,l) = tau1*umat(i1+1_i4b:n,l)
!    
                end do
!    
                mat(i0+1_i4b:i0+blksz2,1_i4b:n) = transpose( tmp(1_i4b:n,1_i4b:blksz2) )
!
            end do
!
!           ALLOCATE WORK VARIABLES.
!
            deallocate( umat, tmp, lmat, x, stat = iok )
!
#ifdef _OPENMP
        end if
#endif
!
    end if
!
!
! END OF SUBROUTINE ortho_gen_lq
! ______________________________
!
    end subroutine ortho_gen_lq
!
! =========================================================================================
!
    subroutine apply_q_lq_rm( mat, tau, c, left, trans )
!
! Purpose
! _______
!
!   APPLY_Q_LQ overwrites the general real m-by-n matrix C with
!
!           Q * C   if LEFT = true  and TRANS = false, or
!
!           Q' * C  if LEFT = true  and TRANS = true,  or
!
!           C * Q   if LEFT = false and TRANS = false, or
!
!           C * Q'  if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                   Q = H(k) * . . . * H(2) * H(1)
!
!   as returned by LQ_CMP. Q is of order m if LEFT = true and of order n
!   if LEFT = false.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th row must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by LQ_CMP in the first k rows of its array
!                argument MAT. MAT is not modified by the routine.
!
!   TAU          (INPUT) real(stnd), dimension(:)
!                TAU(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by LQ_CMP.
!                The size of TAU determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of TAU must verify :
!
!                        size( TAU ) <= min( size( MAT, 1 ), ( MAT, 2 ) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify :
!
!                -  if LEFT = true,  size( C, 1 ) = size( MAT, 2 )
!                -  if LEFT = false, size( C, 2 ) = size( MAT, 2 ) .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left
!                - LEFT = false : apply Q or Q' from the right
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORML2 in LAPACK.
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the upper triangle
!   of MAT and applying the orthogonal matrix Q of the LQ factorization to the
!   real m-by-n matrix C.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm used here, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!          Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!          The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989: Block
!           reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988: Implementation of the GMRES method using Householder
!           transformations. Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tau
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                           :: tau1, prod
    real(stnd), dimension( blksz )                       :: x
    real(stnd), dimension( max( size(c,1), size(c,2) ) ) :: tmp
    real(stnd), dimension( blksz, blksz )                :: lmat
    real(stnd), dimension( size(mat,2), blksz )          :: umat
    real(stnd), dimension( :, : ), allocatable           :: x2
!
    integer(i4b) :: m, n, nm, nq, k, i, i1, i2, j, j1, j2, j3, l, l2, p
    integer      :: iok
!
    logical(lgl) :: descending
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical      :: test_par, ompparallel, ompnested
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_lq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c, 1 )
    n = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
    k = size( tau )
!
    call assert( logical( nq==size(mat,2), lgl ),             &
                 logical( k<=min(nq,size(mat,1)), lgl ),      &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  nm<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  (nm/nq)/i>=omp_limit2               .and.      &
                  i>1_i4b
#endif
!
    descending = (left.and.trans) .or.  .not.(left.or.trans)
!
!$OMP PARALLEL IF(test_par)                                               &
!$OMP         ,PRIVATE(iok,i,i1,i2,l,l2,j,j1,j2,j3,jinc,jrem,joffset,     &
!$OMP                  num_threads,thread_num,tau1,prod,x,umat,tmp,lmat,x2)
!
#ifdef _OPENMP
    if ( test_par ) then
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem        = mod( p, num_threads )
        jinc        = ( p - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
    else
#endif
        j1 = 1_i4b
        j2 = p
#ifdef _OPENMP
    end if
#endif
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
    if ( left ) then
!
        if ( descending ) then
!
!           COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
            do i = k - blksz, 0_i4b, -blksz
!    
!               APPLY H(i+1)* ...*H(i+blksz) TO c(i+1:m,1:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!          ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,1:n) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:m,1_i4b:blksz) = transpose( mat(i1:i2,i1:m) )
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz
!
                    i1 = i + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:m,l), umat(i1:m,l2) )
!
                    end do
!                    
                end do
!
                do j = j1, j2
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j) .
!
                    tmp(i+1_i4b:m) = c(i+1_i4b:m,j)
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(l) = tau(i1)*dot_product( umat(i1:m,l), tmp(i1:m) )
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*c(i+1:m,j)  AND
!                   COMPUTE c(i+1:m,j) = c(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
                    do l = blksz, 1_i4b, -1_i4b
!
                        x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                        i1 = i + l
!
#ifdef _BLAS
                        call axpy( m-i1+1_i4b, x(l), umat(i1:m,l), 1_i4b, tmp(i1:m), 1_i4b )
#else
                        tmp(i1:m) = tmp(i1:m) + x(l)*umat(i1:m,l)
#endif
!
                    end do
!
                    c(i+1_i4b:m,j) = tmp(i+1_i4b:m)
!
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                tmp(i:m) = mat(i,i:m)
!
                do j = j1, j2
!
                    prod = dot_product( tmp(i:m), c(i:m,j) )*tau1
#ifdef _BLAS
                    call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m,j), 1_i4b )
#else
                    c(i:m,j) = c(i:m,j) + prod*tmp(i:m)               
#endif
!
                end do
!
            end do            
!
        else
!
!           COMPUTE H(k)* . . .*H(2)*H(1)*c(:m,:n) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+blksz)* ...*H(i+1) TO c(i+1:m,1:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!          ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,1:n) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:m,1_i4b:blksz) = transpose( mat(i1:i2,i1:m) )
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 1_i4b, blksz - 1_i4b
!
                    i1 = i + l
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i2:m,l), umat(i2:m,l2) )
!
                    end do
!                    
                end do
!
                do j = j1, j2
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j) .
!
                    tmp(i+1_i4b:m) = c(i+1_i4b:m,j)
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(l) = tau(i1)*dot_product( umat(i1:m,l), tmp(i1:m) )
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*c(i+1:m,j)  AND
!                   COMPUTE c(i+1:m,j) = c(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
                    do l = 1_i4b, blksz
!
                        x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                        i1 = i + l
!
#ifdef _BLAS
                        call axpy( m-i1+1_i4b, x(l), umat(i1:m,l), 1_i4b, tmp(i1:m), 1_i4b )
#else
                        tmp(i1:m) = tmp(i1:m) + x(l)*umat(i1:m,l)
#endif
!
                    end do
!
                    c(i+1_i4b:m,j) = tmp(i+1_i4b:m)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                tmp(i:m) = mat(i,i:m)
!
                do j = j1, j2
!
                    prod     = dot_product( tmp(i:m), c(i:m,j) )*tau1
#ifdef _BLAS
                    call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m,j), 1_i4b )
#else
                    c(i:m,j) = c(i:m,j) + prod*tmp(i:m)               
#endif
!
                end do            
!
            end do
!
        end if
!
    else
!
        j3 = j2 - j1 + 1_i4b
!
!       ALLOCATE WORK VARIABLE.
!
        allocate( x2(j3,blksz), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( descending ) then
!
!           COMPUTE c(:m,:n)*H(k)* . . .*H(2)*H(1) .
!
            do i = k - blksz, 0_i4b, -blksz
!
!               APPLY H(i+blksz)* . . .*H(i+1) TO c(1:m,i+1:n) FROM THE RIGHT AS
!
!             c(1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:n,1_i4b:blksz) = transpose( mat(i1:i2,i1:n) )
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz
!
                    i1 = i + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:n,l), umat(i1:n,l2) )
!
                    end do
!                    
                end do
!
!               COMPUTE c(1:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
                    x2(1_i4b:j3,l) = tau(i1)*matmul( c(j1:j2,i1:n), umat(i1:n,l) )
!                    
                end do
!
!               SOLVE x(1:m,:blksz)*lmat(:blksz,:blksz) = c(1:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE c(1:m,i+1:i+blksz-1) = c(1:m,i+1:i+blksz-1) + x(1:m,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                do l = blksz, 1_i4b, -1_i4b
!
                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = 1_i4b, l-1_i4b
!
#ifdef _BLAS
                        call axpy( j3, -lmat(l2,l), tmp(1_i4b:j3), 1_i4b, x2(1_i4b:j3,l2), 1_i4b )
#else
                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - tmp(1_i4b:j3)*lmat(l2,l)
#endif
!
                    end do
!
                    do l2 = l, blksz-1_i4b
!
                        i2 = i + l2
!
#ifdef _BLAS
                        call axpy( j3, umat(i2,l), tmp(1_i4b:j3), 1_i4b, c(j1:j2,i2), 1_i4b )
#else
                        c(j1:j2,i2) = c(j1:j2,i2) + tmp(1_i4b:j3)*umat(i2,l)
#endif
!
                    end do
!
                end do
!
!               COMPUTE c(1:m,i+blksz:n) = c(1:m,i+blksz:n) + x(1:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i + blksz, n
!
                    tmp(1_i4b:j3) = c(j1:j2,j)
!
                    do l = 1_i4b, blksz
!
#ifdef _BLAS
                        call axpy( j3, umat(j,l), x2(1_i4b:j3,l), 1_i4b, tmp(1_i4b:j3), 1_i4b )
#else
                        tmp(1_i4b:j3) = tmp(1_i4b:j3) + x2(1_i4b:j3,l)*umat(j,l)
#endif
!                    
                    end do
!
                    c(j1:j2,j) = tmp(1_i4b:j3)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(:m,i:n) FROM THE RIGHT.
!
                tmp(i:n) = mat(i,i:n)
!
                do j = j1, j2
!
                    prod     = dot_product( tmp(i:n), c(j,i:n) )*tau1
                    c(j,i:n) = c(j,i:n) + prod*tmp(i:n)               
!
                end do            
!
            end do            
!
        else
!
!           COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(k) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+1)* . . .*H(i+blksz) TO c(:m,i+1:n) FROM THE RIGHT AS
!
!             c(:m,i+1:n)*( I + U(i+1i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:n,1_i4b:blksz) = transpose( mat(i1:i2,i1:n) )
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 1_i4b, blksz - 1_i4b
!
                    i1 = i + l
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i2:n,l), umat(i2:n,l2) )
!
                    end do
!                    
                end do
!
!               COMPUTE c(1:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
                    x2(1_i4b:j3,l) = tau(i1)*matmul( c(j1:j2,i1:n), umat(i1:n,l) )
!                    
                end do
!
!               SOLVE x(1:m,:blksz)*lmat(:blksz,:blksz) = c(1:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE c(1:m,i+1:i+blksz-1) = c(1:m,i+1:i+blksz-1) + x(1:m,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                do l = 1_i4b, blksz
!
                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = l + 1_i4b, blksz
!
#ifdef _BLAS
                        call axpy( j3, -lmat(l2,l), tmp(1_i4b:j3), 1_i4b, x2(1_i4b:j3,l2), 1_i4b )
#else
                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - tmp(1_i4b:j3)*lmat(l2,l)
#endif
!
                        i2 = i + l2 - 1_i4b
!
#ifdef _BLAS
                        call axpy( j3, umat(i2,l), tmp(1_i4b:j3), 1_i4b, c(j1:j2,i2), 1_i4b )
#else
                        c(j1:j2,i2) = c(j1:j2,i2) + tmp(1_i4b:j3)*umat(i2,l)
#endif
!
                    end do
!
                end do
!
!               COMPUTE c(1:m,i+blksz:n) = c(1:m,i+blksz:n) + x(1:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i + blksz, n
!
                    tmp(1_i4b:j3) = c(j1:j2,j)
!
                    do l = 1_i4b, blksz
!
#ifdef _BLAS
                        call axpy( j3, umat(j,l), x2(1_i4b:j3,l), 1_i4b, tmp(1_i4b:j3), 1_i4b )
#else
                        tmp(1_i4b:j3) = tmp(1_i4b:j3) + x2(1_i4b:j3,l)*umat(j,l)
#endif
!                    
                    end do
!
                    c(j1:j2,j) = tmp(1_i4b:j3)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(:m,i:n) FROM THE RIGHT.
!
                tmp(i:n) = mat(i,i:n)
!
                do j = j1, j2
!
                    prod     = dot_product( tmp(i:n), c(j,i:n) )*tau1
                    c(j,i:n) = c(j,i:n) + prod*tmp(i:n)     
!
                end do            
!
            end do
!
        end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x2 )
!
    end if
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE apply_q_lq_rm
! _______________________________
!
    end subroutine apply_q_lq_rm
!
! =========================================================================================
!
    subroutine apply_q_lq_rv( mat, tau, c, trans )
!
! Purpose
! _______
!
!   APPLY_Q_LQ overwrites the general real m vector C with
!
!           Q * C   if TRANS = false, or
!
!           Q' * C  if TRANS = true,
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                   Q = H(k) * . . . * H(2) * H(1)
!
!   as returned by LQ_CMP. Q is of order m.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th row must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by LQ_CMP in the first k rows of its array
!                argument MAT. MAT is not modified by the routine.
!
!   TAU          (INPUT) real(stnd), dimension(:)
!                TAU(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by LQ_CMP.
!                The size of TAU determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of TAU must verify :
!
!                        size( TAU ) <= min( size( MAT, 1 ), ( MAT, 2 ) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m vector C.
!
!                On exit, C is overwritten by Q * C or Q' * C.
!
!                The shape of C must verify size( C ) = size( MAT, 2 ).
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORML2 in LAPACK.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm used here, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!           The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tau
    real(stnd),   intent(inout), dimension(:)   :: c
!
    logical(lgl), intent(in)    :: trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: tau1, prod
    real(stnd), dimension( size(c) ) :: tmp
!
    integer(i4b) :: m, k, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_lq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c )
    k = size( tau )
!
    call assert( logical( m==size(mat,2), lgl ),            &
                 logical( k<=min(m,size(mat,1)), lgl ),     &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  m<=0_i4b ) return
!
    if ( trans ) then
!
!       COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
        do i = k, 1_i4b, -1_i4b
!
            tau1 = tau(i)
!
            if ( tau1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i,i:m)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*tau1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)
#endif
!
        end do            
!
    else
!
!       COMPUTE H(k)* . . .*H(2)*H(1)*c(:m) .
!
        do i = 1_i4b, k
!
            tau1 = tau(i)
!
            if ( tau1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i,i:m)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*tau1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)     
#endif
!
        end do
!
    end if
!
!
! END OF SUBROUTINE apply_q_lq_rv
! _______________________________
!
    end subroutine apply_q_lq_rv
!
!
! =========================================================================================
!                            QR DECOMPOSITION
! =========================================================================================
!
!
    subroutine qr_cmp( mat, diagr, beta )
!
! Purpose
! _______
!
!   QR_CMP computes a QR factorization of a real m-by-n matrix MAT :
!
!                        MAT = Q * R
!
!   where Q is orthogonal and R is upper trapezoidal (upper triangular if m>=n).
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real matrix to be decomposed.
!
!            On exit, the elements above the diagonal of the array contain  
!            the corresponding elements of R. The elements on and below the 
!            diagonal, with the array BETA, represent the orthogonal matrix Q
!            as a product of elementary reflectors, see Further Details.
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R.
!
!            The size of DIAGR must verify :
!
!                         size(DIAGR) <= min( size(MAT,1) , size(MAT,2) )
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors.
!
!            The size of BETA must verify :
!
!                         size(BETA) = size(DIAGR) <= min( size(MAT,1) , size(MAT,2) )
!
!            See Further Details.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = size(BETA) = size(DIAGR)
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!   A blocked algorithm is used for computing the QR factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm used here, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!             Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!             The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989: Block
!            reduction of matrices to condensed form for eigenvalue computations. 
!            J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988: Implementation of the GMRES method using Householder
!             transformations. Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Utilities,          only : assert_eq, assert
    use Hous_Procedures,    only : hous1
#ifdef _OPENMP
    use Char_Constants,     only : allocate_error
    use Utilities,          only : merror
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: prod, beta1
    real(stnd), dimension( blksz )        :: x
    real(stnd), dimension( size(mat,1) )  :: tmp
    real(stnd), dimension( blksz, blksz ) :: lmat
!
    integer(i4b) :: m, n, k, k2, i, i1, i2, j, q, q1, q2, nblk, rem
!
#ifdef _OPENMP
    integer       :: iok
    integer(i4b)  :: iblk, nblk2, l, l1, l2, p, num_threads, thread_num
    integer(kind=omp_lock_kind), dimension(:), allocatable :: lck
!
    logical(lgl) :: do_lock
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qr_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( int(size(beta),i4b),   &
                   int(size(diagr),i4b),  &
                   name_proc              )
!
    call assert( logical( k<=min(m,n), lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE THE NUMBER OF BLOCKS.
!
    nblk = k/blksz
    rem  = mod( k, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        k2   = k - rem
!
    else
!
        k2   = k - blksz
!
    end if
!
#ifdef _OPENMP
!
!   COMPUTE THE NUMBER OF COLUMN BLOCKS.
!
    nblk2 = n/blksz
    rem   = mod( n, blksz )
!
    if ( rem/=0_i4b ) nblk2 = nblk2 + 1_i4b
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  nblk2>1_i4b                         .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!       PARALLEL AND BLOCKED QR DECOMPOSITION OF mat.
!
!       COMPUTE QR FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
        do j = 1_i4b, blksz
!    
!           GENERATE ELEMENTARY REFLECTOR H(j) TO ANNIHILATE mat(j+1:m,j) .
!    
            call hous1( mat(j:m,j), beta(j), diagr(j) )
!    
!           APPLY H(j) TO mat(j:m,j+1:blksz) FROM THE LEFT.
!
            beta1 = beta(j)
!
            do q = j + 1_i4b, blksz
!
                prod = beta1*dot_product( mat(j:m,j), mat(j:m,q) )
#ifdef _BLAS
                call axpy( m-j+1_i4b, prod, mat(j:m,j), 1_i4b, mat(j:m,q), 1_i4b )
#else
                mat(j:m,q) = mat(j:m,q) + prod*mat(j:m,j)
#endif
!    
            end do
!
        end do
!
!       ALLOCATE WORK ARRAY FOR STORING THE LOCKS FOR EACH BLOCK.
!
        allocate( lck(nblk), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SET UP LOCKS FOR EACH BLOCK.
!
        do i = 1_i4b, nblk
            call omp_init_lock( lck(i) )
        end do
!
!$OMP PARALLEL PRIVATE(i,i1,i2,iblk,j,l,l1,l2,p,q,q1,q2,num_threads,thread_num,do_lock,beta1,prod,tmp,x,lmat)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       SET THE LOCKS FOR THE BLOCKS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do i = 1_i4b, nblk
            call omp_set_lock( lck(i) )
        end do
!
!$OMP END DO
!
!       INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!       USED IN THE BLOCKED ALGORITHM.
!
        lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
        do i = 1_i4b, blksz
            lmat(i,i) = one
        end do
!
        if ( thread_num==0_i4b ) then
!
!           RELEASE THE FIRST LOCK.
!
            call omp_unset_lock( lck(1) )     
!
            do_lock = false
!
        else
!
            do_lock = true
!
        end if
!
!       START THE BLOCKED QR FACTORIZATION.
!
        do i = 0_i4b, k2, blksz
!
!           COMPUTE THE PREVIOUS BLOCK NUMBER.
!
            iblk = i/blksz + 1_i4b
!
!           EACH THREAD WILL TOUCH EXACTLY THE SAME COLUMNS BLOCKS DURING THE ITERATIONS.
!
            p = ( (iblk/num_threads)*num_threads ) + thread_num
            if ( thread_num<mod(iblk,num_threads) ) p = p + num_threads
!    
            if ( p<nblk2 ) then
!
                if ( do_lock ) then
!
                    call omp_set_lock( lck(iblk) )
!
! $OMP FLUSH (mat,beta)
!
                    call omp_unset_lock( lck(iblk) )
!
                else
!
                    do_lock = true
!
                end if
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+1:m,i+blksz+1:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!        ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*mat(i+1:m,i+blksz+1:n) .
!
                i1 = i + 1_i4b
                i2 = min( i + blksz, k )
!
                q2 = i2 - i
!
!               FIRST, COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do j = i1, i2 - 1_i4b
!
                    q1 = j - i
!
                    do q = j + 1_i4b, i2
!
                        lmat(q-i,q1) = -beta(q)*dot_product( mat(q:m,j), mat(q:m,q) )
!
                    end do
!                    
                end do
!
                do l = p, nblk2-1_i4b, num_threads
!
!                   COMPUTE INDICES FOR THE FIRST AND LAST COLUMNS OF THE l+1 BLOCK.
!
                    l1 = l*blksz + 1_i4b
                    l2 = min( (l+1_i4b)*blksz, n )
!
                    do j = l1, l2
!
!                       COMPUTE U(i+1:m,:blksz)'*mat(i+1:m,j) .
!
                        tmp(i1:m) = mat(i1:m,j)
!
                        do q = i1, i2
!
                            x(q-i) = beta(q)*dot_product( mat(q:m,q), tmp(q:m) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*mat(i+1:m,j) AND
!                       COMPUTE mat(i+1:m,j) = mat(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .

                        do q = i1, i2
!
                            q1 = q - i
!
                            x(q1+1_i4b:q2) = x(q1+1_i4b:q2) - x(q1)*lmat(q1+1_i4b:q2,q1)
!
#ifdef _BLAS
                            call axpy( m-q+1_i4b, x(q1), mat(q:m,q), 1_i4b, tmp(q:m), 1_i4b )
#else
                            tmp(q:m) = tmp(q:m) + x(q1)*mat(q:m,q)
#endif
!
                        end do
!
                        mat(i1:m,j) = tmp(i1:m)
!
                    end do
!
                    if ( l==iblk .and. l1<=k ) then
!
!                       COMPUTE THE QR FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(i2+1:m,i2+1:i2+blksz), IN PLACE.
!
                        do j = l1, min(l2,k)
!    
!                           GENERATE ELEMENTARY REFLECTOR H(j) TO ANNIHILATE mat(j+1:m,j) .
!    
                            call hous1( mat(j:m,j), beta(j), diagr(j) )
!    
!                           APPLY H(j) TO mat(j:m,j+1:i+2*blksz) FROM THE LEFT.
!
                            beta1 = beta(j)
!
                            do q = j + 1_i4b, l2
!
                                prod = beta1*dot_product( mat(j:m,j), mat(j:m,q) )
#ifdef _BLAS
                                call axpy( m-j+1_i4b, prod, mat(j:m,j), 1_i4b, mat(j:m,q), 1_i4b )
#else
                                mat(j:m,q) = mat(j:m,q) + prod*mat(j:m,j)
#endif
!    
                            end do
!
                        end do
!
!                       RELEASE THE iblk+1_i4b LOCK.
!
! $OMP FLUSH (mat,beta)
!
                        call omp_unset_lock( lck(iblk+1_i4b) )                 
!
                        do_lock = false
!
                    end if
!
                end do
!
            end if
!
        end do
!
!$OMP END PARALLEL
!
!       DESTROY THE LOCKS.
!
        do i = 1_i4b, nblk
            call omp_destroy_lock( lck(i) )
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( lck )
!
    else
!
#endif
!
!       BLOCKED QR DECOMPOSITION OF mat.
!
!       INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!       USED IN THE BLOCKED ALGORITHM.
!
        lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
        do i = 1_i4b, blksz
            lmat(i,i) = one
        end do
!
        do i = 0_i4b, k2, blksz
!
            i1 = i + 1_i4b
            i2 = min( i + blksz, k )
!
            do j = i1, i2
!    
!               GENERATE ELEMENTARY REFLECTOR H(j) TO ANNIHILATE mat(j+1:m,j) .
!    
                call hous1( mat(j:m,j), beta(j), diagr(j) )
!    
!               APPLY H(j) TO mat(j:m,j+1:i+blksz) FROM THE LEFT.
!
                beta1 = beta(j)
!
                do q = j + 1_i4b, i2
!
                    prod = beta1*dot_product( mat(j:m,j), mat(j:m,q) )
#ifdef _BLAS
                    call axpy( m-j+1_i4b, prod, mat(j:m,j), 1_i4b, mat(j:m,q), 1_i4b )
#else
                    mat(j:m,q) = mat(j:m,q) + prod*mat(j:m,j)
#endif
!    
                end do
!
            end do
!    
            if ( i2<n ) then
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+1:m,i+blksz+1:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!        ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*mat(i+1:m,i+blksz+1:n) .
!
!               FIRST, COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do j = i1, i2 - 1_i4b
!
                    q1 = j - i
!
                    do q = j + 1_i4b, i2
!
                        lmat(q-i,q1) = -beta(q)*dot_product( mat(q:m,j), mat(q:m,q) )
!
                    end do
!                    
                end do
!
                q2 = i2 - i
!
                do j = i2+1_i4b, n
!
!                   COMPUTE U(i+1:m,:blksz)'*mat(i+1:m,j) .
!
                    tmp(i1:m) = mat(i1:m,j)
!
                    do q = i1, i2
!
                        x(q-i) = beta(q)*dot_product( mat(q:m,q), tmp(q:m) )
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*mat(i+1:m,j) AND
!                   COMPUTE mat(i+1:m,j) = mat(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
                    do q = i1, i2
!
                        q1 = q - i
!
                        x(q1+1_i4b:q2) = x(q1+1_i4b:q2) - x(q1)*lmat(q1+1_i4b:q2,q1)
!
#ifdef _BLAS
                        call axpy( m-q+1_i4b, x(q1), mat(q:m,q), 1_i4b, tmp(q:m), 1_i4b )
#else
                        tmp(q:m) = tmp(q:m) + x(q1)*mat(q:m,q)
#endif
!
                    end do
!
                    mat(i1:m,j) = tmp(i1:m)
!                    
                end do
!
            end if
!
        end do
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE qr_cmp
! ________________________
!
    end subroutine qr_cmp
!
! =========================================================================================
!
    subroutine qrfac( name_proc, syst, kfix, krank, min_norm, diagr, beta, h, tol, ip )
!
! Purpose
! _______
!
!   QRFAC is a low level subroutine for computing a (complete) orthogonal factorization
!   of the array section SYST(1:m,1:n) where n<=size(SYST,2) and m=size(SYST,1).
!
!   The routine first computes a QR factorization with column pivoting :
!
!                     SYST(1:m,1:n) * P = Q * R
!
!   , where P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!   The orthogonal transformation Q is then applied to SYST(1:m,n+1:) :
!
!                     SYST(1:m,n+1:) = Q * B
!
!   Then, the rank of SYST(1:m,1:n) is determined by finding the submatrix R11 of R
!   which is defined as the largest leading submatrix whose estimated condition
!   number, in the 1-norm, is less than 1/TOL or such that abs(R11[j,j])>0 if TOL
!   is absent. The order of R11, KRANK, is the effective rank of SYST(1:m,1:n).
!
!   This leads to the following partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R21 is zero by construction (since R is an upper triangular or trapezoidal matrix)
!   and R22 is considered to be negligible.
!
!   If MIN_NORM=true, R22 is considered to be negligible and R12 is annihilated by orthogonal
!   transformations from the right, arriving at the complete orthogonal factorization:
!
!                       SYST(1:m,1:n) * P = Q * T * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-m orthogonal matrix,
!   Z is a n-by-n orthogonal matrix and T is a m-by-n matrix and has the form:
!
!   [  T11  T12  ]
!
!   [  T21  T22  ]
!
!   , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a KRANK-by-KRANK upper
!   triangular matrix.
!   
!
! Arguments
! _________
!
!   NAME_PROC    (INPUT) character(len=*)
!                Name of the subroutine calling QRFAC.
!
!   SYST         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the real m-by-n matrix to be decomposed and, eventually, the
!                the right hand side matrix of an associated least squares problem in
!                the array section SYST(:,n+1:).
!
!                On exit, SYST(:,1:n) has been overwritten by details of its (complete) 
!                orthogonal factorization and B is stored in SYST(:,n+1:) .
!                See Further Details.
!
!   KFIX         (INPUT) integer(i4b)
!                KFIX=k, implies that the first k columns of SYST(:,1:n) are
!                to be forced into the basis. Pivoting is performed on the last n-k
!                columns of SYST(:,1:n). KFIX<=0 can be used when pivoting is desired
!                on all columns of SYST(:,1:n). If KFIX<min(m,n) then the optional
!                argument IP must be present to store the permutation matrix P.
!                When KFIX>=min(m,n) is used, pivoting is not performed. This is
!                appropriate when SYST(:,1:n) is known to be full rank.
!            
!   KRANK        (INPUT/OUTPUT) integer(i4b)
!                On entry, KRANK=n . KRANK must verify  KRANK <=size(SYST,2).
!            
!                On exit, KRANK contains the effective rank of SYST(:,1:n), i.e.,
!                the order of the submatrix R11. This is the same as the order of
!                the submatrix T11 in the complete orthogonal factorization of SYST(:,1:n).
!
!   MIN_NORM     (INPUT) logical(lgl)
!                MIN_NORM=true indicates that a complete orthogonal factorization
!                of SYST(:,1:n) must be computed.
!                MIN_NORM=false indicates that only a QR factorization (with column
!                pivoting) of SYST(:,1:n) must be computed.
!
!   DIAGR        (OUTPUT) real(stnd), dimension(:) 
!                On exit, the diagonal elements of the matrix R if MIN_NORM=false or
!                the diagonal elements of the matrix T11 if MIN_NORM=true. The diagonal
!                elements of T11 are stored in DIAGR(1:KRANK).
!                see Further Details.
!
!                The size of DIAGR must verify: 
!            
!                               size( DIAGR ) >= min( size(SYST,1) , KRANK ) .
!
!   BETA         (OUTPUT) real(stnd), dimension(:)
!                On exit, the scalars factors of the elementary reflectors defining Q.
!                see Further Details.
!
!                The size of BETA must  verify:
!            
!                               size( BETA ) >= min( size(SYST,1) , KRANK ) .
!
!   H            (OUTPUT) real(stnd), dimension(:)
!                On exit, if MIN_NORM=true, the scalars factors of the elementary
!                reflectors defining Z are stored in H(1:KRANK). see Further Details.
!
!                The size of H must verify  size(H) >= KRANK  .
!
!   TOL          (INPUT/OUTPUT, OPTIONAL) real(stnd)
!                If TOL is present and is in [0,1[, then:
!
!                - On entry, the calculations to determine the condition number of
!                  of SYST(:,1:n) are performed. Then, TOL is used to determine the
!                  effective rank of SYST(:,1:n), which is defined as the order of 
!                  the largest leading triangular submatrix R11 in the QR factorization
!                  with pivoting of SYST(:,1:n), whose estimated condition number < 1/TOL.
!                  If TOL=0 is specified the numerical rank of SYST(:,1:n) is determined.
!                - On exit, the reciprocal of the condition number is returned in TOL.
!
!                If TOL is not specified or is outside [0,1[ :
!
!                - The calculations to determine the condition number of MAT are not
!                  performed and crude tests on R(j,j) are done to determine
!                  the rank of MAT. If TOL is present, it is not changed.
!
!   IP           (OUTPUT, OPTIONAL) integer(i4b), dimension(:) 
!                On exit, if IP(j)=k, then the j-th column of SYST(:,1:n) * P was 
!                the k-th column of SYST(:,1:n). 
!                IP must be present if KFIX<min( SYST,1), KRANK )=min( m, n).
!
!                The size of IP must verify  size(IP) >= KRANK.
!
!
! Further Details
! _______________
!
!   QRFAC is called by the subroutines QR_CMP2, LLSQ_QR_SOLVE, LLSQ_QR_SOLVE2 and
!   SOLVE_LLSQ in modules QR_Procedures and LLSQ_Procedures.
!
!   Since QRFAC is a low level subroutine, no checking of the correctness of the
!   dimensions of the array arguments is performed inside of the subroutine and
!   such checking must be done before calling QRFAC.
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = min( m , n )
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in SYST(i:m,i) and beta in BETA(i).
!
!   The matrix P is represented in the array IP (if present) as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   If MIN_NORM=false, a QR factorization with column pivoting of SYST(:,1:n) is computed
!   and, on exit, the elements above the diagonal of the array SYST(:,1:n) contain the 
!   corresponding elements of the triangular matrix R. The elements of the diagonal
!   of R are stored in the array DIAGR. 
!
!   If MIN_NORM=true, a complete orthogonal factorization of SYST(:,1:n) is computed.
!   The factorization is obtained by Householder's method. The kth transformation
!   matrix, Z(k), which is used to introduce zeros into the kth row of R,
!   is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!         T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar and z(k) is an (n-KRANK) element vector. tau and z(k) are chosen
!   to annihilate the elements of the kth row of R12.
!
!   The Z n-by-n orthogonal matrix is given by
!
!         Z = Z(1) * Z(2) * ... * Z(KRANK)
!
!   On exit, the scalar tau is returned in the kth element of TAU and the vector u(k)
!   in the kth row of SYST, such that the elements of z(k) are in SYST(k,KRANK+1:n).
!
!   On exit, the elements above the diagonal of the array section SYST(1:KRANK,1:KRANK)
!   contain the corresponding elements of the triangular matrix T11. The elements of
!   the diagonal of T11 are stored in the array section DIAGR(1:KRANK).
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!            Prentice-Hall.
!
!   (2) G.H. Golub and C.F. Van Loan, 1996: Matrix Computations. The Johns
!            Hopkins University Press, Baltimore, Maryland.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants,   only : machhuge, lamch
    use Char_Constants,  only : qr_error1
    use Utilities,       only : merror, imaxloc, arth, norm
    use Hous_Procedures, only : hous1, hous2, apply_hous2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: syst
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta, h
    real(stnd),   intent(inout), optional       :: tol
!
    integer(i4b), intent(in)                          :: kfix
    integer(i4b), intent(inout)                       :: krank
    integer(i4b), intent(out), dimension(:), optional :: ip
!
    logical(lgl), intent(in) :: min_norm
!
    character(len=*), intent(in) :: name_proc
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: xnorm, anorm, rcond, rloss, safmin, cl, tau
    real(stnd), dimension(max(size(syst,1),size(syst,2))) :: tmp
!
    integer(i4b) :: n, m, ndiag, i, j, j1, j2, jmax, ncol, k, kp1
!
    logical(lgl) :: test, notfirst
!
#ifdef _OPENMP
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   INITIALIZE WORK VARIABLES AND PERMUTATION VECTOR FOR THE FACTORIZATION.
!    
    n     = krank
    m     = size( syst, 1 )
    ndiag = min( n, m )
    ncol  = size( syst, 2 )
    i     = min( ndiag, kfix )
    k     = max( i, 0_i4b )
    kp1   = k + 1_i4b
!    
    if ( k<ndiag ) then
        notfirst = false
        test     = true
    end if
!
    if ( present(ip) )   ip(:n) = arth( 1_i4b, 1_i4b, n )
    beta(:ndiag) = zero
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF syst(:,:n) .
!    
    if ( k>0_i4b ) then
!    
        call qr_cmp( syst(:m,:ncol), diagr(:k), beta(:k) )
!    
    end if
!    
    if ( k<ndiag ) then
!
#ifdef _OPENMP
        i = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      m*(ncol-k)>=omp_limit               .and.      &
                      (ncol-k)/i>=omp_limit2              .and.      &
                      i>1_i4b
!
!$OMP PARALLEL IF(test_par)                                                 &
!$OMP         ,PRIVATE(i,j,j1,j2,jinc,jrem,joffset,num_threads,thread_num,cl)
!
        if ( test_par ) then
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
        else
#endif
            j2 = ncol
#ifdef _OPENMP
        end if
#endif
!
        do j = kp1, ndiag
!    
!$OMP SINGLE
!    
!           DO COLUMN PIVOTING OF syst(:,j:n) .
!    
!           UPDATE  COLUMN LENGTHS OF syst(:,j:n) .
!    
            if ( notfirst ) then
!
#ifdef _WHERE
                do i = j, n
!
                    if ( h(i)/=zero ) then
                        tmp(i) = h(i)
                        h(i)   = h(i)*sqrt( max(one-(syst(j-1_i4b,i)/h(i))**2,zero) )
                        tmp(i) = h(i)/tmp(i)
                    else
                        tmp(i) = machhuge
                    end if
!
                end do
#else
                where ( h(j:n)/=zero )
                    tmp(j:n) = h(j:n)
                    h(j:n)   = h(j:n)*sqrt( max(one-(syst(j-1_i4b,j:n)/h(j:n))**2,zero) )
                    tmp(j:n) = h(j:n)/tmp(j:n)
                elsewhere
                    tmp(j:n) = machhuge
                end where
#endif
!
                anorm = minval( tmp(j:n) )
                rloss = rloss*merge( anorm, one, anorm/=machhuge )
                anorm = rloss + ten
                test  = anorm <= ten
!
            end if
!    
!           COMPUTE COLUMN LENGTHS OF syst(:,:n) .
!    
            if ( test ) then
!
                h(j:n) = norm( syst(j:m,j:n), dim=2_i4b )
!
!BUG: For compiling STATPACK on linux machines with OPENMP and the older pgf90 compiler,
!     replace the preceding line with
!
!                h(j:n) = sqrt( sum( syst(j:m,j:n)**2, dim=1_i4b ) )
!
                rloss    = one
                notfirst = true
!
            end if
!    
!           FIND jmax .                
!    
            jmax = imaxloc( h(j:n) ) + j - 1_i4b
!
            tmp(:m) = syst(:m,jmax)
!    
!           DO COLUMN INTERCHANGES IF NEEDED.
!    
            if ( jmax/=j ) then
!    
                krank    = ip(j)
                ip(j)    = ip(jmax)
                ip(jmax) = krank
!    
                syst(:m,jmax) = syst(:m,j)
!    
                h(jmax) = h(j)
!    
            end if
!    
!           COMPUTE THE J-TH TRANSFORMATION.
!    
            call hous1( tmp(j:m), tau, diagr(j) )
!
!$OMP END SINGLE
!
            if ( tau/=zero ) then
!    
!               APPLY THE J-TH TRANSFORMATION TO syst(j:,j+1:) FROM THE LEFT.
!    
#ifdef _OPENMP
                if ( test_par ) then
!
                    i    = ncol - j
                    jrem = mod( i, num_threads )
                    jinc = ( i - jrem )/num_threads
!
                    if ( thread_num<jrem ) then
                        jinc    = jinc + 1_i4b
                        joffset = j + 1_i4b
                    else
                        joffset = j + jrem + 1_i4b
                    end if
!
                    j1 = thread_num*jinc + joffset
                    j2 = j1 + jinc - 1_i4b
!
                else
#endif
!    
                    j1 = j + 1_i4b
!
#ifdef _OPENMP
                end if
#endif
!
                do i= j1, j2
!
                    cl = dot_product( tmp(j:m), syst(j:m,i) )*tau
#ifdef _BLAS
                    call axpy( m-j+1_i4b, cl, tmp(j:m), 1_i4b, syst(j:m,i), 1_i4b )
#else
                    syst(j:m,i) = syst(j:m,i) + cl*tmp(j:m)
#endif
!
                end do
!
            end if
!    
!           SAVE THE J-TH TRANSFORMATION.
!
!$OMP SINGLE
!
            syst(:m,j) = tmp(:m)
            beta(j)    = tau
!
!$OMP END SINGLE
!
        end do
!
!$OMP END PARALLEL
!
    end if
!    
!   TEST TOLERANCE.
!    
    if ( present(tol) ) then
        test = zero<=tol .and. tol<one
    else
        test = false
    end if
!    
    tmp(:ndiag) = abs( diagr(:ndiag) )
!    
    if ( test ) then
!
!       FIND safmin, THE SAFE MINIMUM, SUCH
!       THAT  1/safmin  DOES NOT OVERFLOW.
!
        safmin = lamch( 's' )
!    
!       DETERMINE THE PSEUDO-RANK USING THE TOLERANCE, tol .
!    
        anorm = zero
        xnorm = zero
        rcond = zero
!
        do j = 1_i4b, ndiag
!
            if ( tmp(j)<safmin ) exit
!
            anorm  = max( anorm, sum(abs(syst(1_i4b:j-1_i4b,j)))+tmp(j) )
            tmp(j) = one/diagr(j)
!
            do i = j-1_i4b, 1_i4b, -1_i4b
                tmp(i) = -dot_product( tmp(i+1_i4b:j), syst(i,i+1_i4b:j) )/diagr(i)
            end do
!
            xnorm = max( xnorm, sum(abs(tmp(1_i4b:j))) )
            cl    = one/( anorm*xnorm )
!
            if ( tol>=cl ) exit
!
            rcond = cl
!
        end do
!
        tol = rcond
!
    else
!    
!       DETERMINE THE NUMERICAL RANK.
!    
        do j = 1_i4b, ndiag
            if ( tmp(j)<=zero ) exit
        end do
!
    end if
!    
!   OUTPUT RANK OF syst(:,:n) .
!    
    krank = j - 1_i4b
!    
!   STOP THE PROGRAM IF syst(:,:n) IS A DEFICIENT MATRIX WITH krank < min(kfix,ndiag) .
!    
    if ( krank<min(kfix,ndiag) ) call merror( name_proc//qr_error1 )
!    
!   IF THE PSEUDORANK krank IS LESS THAN n AND THE MINIMUM_NORM SOLUTION IS REQUIRED,
!   COMPUTE HOUSEHOLDER DECOMPOSITION OF syst(1:krank,:n) .
!    
    if ( krank/=n .and. min_norm ) then
!    
        h(1_i4b:krank) = zero
!    
        do i = krank, 1_i4b, -1_i4b
!    
!           COMPUTE THE I-TH HOUSEHOLDER TRANSFORMATION.
!    
            call hous2( diagr(i), syst(i,j:n), h(i) )
!    
!           APPLY THE I-TH HOUSEHOLDER TRANSFORMATION FROM THE RIGHT.
!    
            call apply_hous2( syst(i,j:n), h(i),                                    &
                              syst(1_i4b:i-1_i4b,i), syst(1_i4b:i-1_i4b,j:n), false )
        end do
!    
    end if
!
!
! END OF SUBROUTINE qrfac
! _______________________
!
    end subroutine qrfac
!
! =========================================================================================
!
    subroutine qr_cmp2( mat, diagr, beta, ip, krank, tol, tau )
!
! Purpose
! _______
!
!   QR_CMP2 computes a (complete) orthogonal factorization of a real m-by-n matrix
!   MAT. MAT may be rank-deficient. The routine first computes a QR factorization
!   with column pivoting of MAT:
!
!                     MAT * P = Q * R
!
!   , here P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!   R can then be partioned by defining R11 as the largest leading submatrix of R whose
!   estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!   if TOL is absent). The order of R11, KRANK, is the effective rank of MAT.
!
!   This leads to the following partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R21 is zero by construction (since R is an upper triangular or trapezoidal matrix).
!
!   If TAU is present, R22 is considered to be negligible and R12 is annihilated 
!   by orthogonal transformations from the right, arriving at the complete
!   orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-m orthogonal matrix,
!   Z is a n-by-n orthogonal matrix and T is a m-by-n matrix and has the form:
!
!   [  T11  T12  ]
!
!   [  T21  T22  ]
!
!   , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a KRANK-by-KRANK upper
!   triangular matrix.
!   
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real m-by-n matrix to be decomposed.
!
!            On exit, MAT has been overwritten by details of its (complete) 
!            orthogonal factorization. See Further Details.
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R if TAU is absent or
!            the diagonal elements of the matrix T11 if TAU is present. The
!            diagonal elements of T11 are stored in DIAGR(1:KRANK).
!
!            The size of DIAGR must be min( size(MAT,1) , size(MAT,2) ).
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors defining Q.
!            See Further Details.
!
!            The size of BETA must be min( size(MAT,1) , size(MAT,2) ).
!
!   IP       (OUTPUT) integer(i4b), dimension(:) 
!            On exit, if IP(j)=k, then the j-th column of MAT*P was 
!            the k-th column of MAT.
!            See Further Details.
!
!            The size of IP must be size(MAT,2)=n.
!
!   KRANK    (INPUT/OUTPUT) integer(i4b)
!            On input, KRANK=k, implies that the first k columns of MAT are
!            to be forced into the basis. Pivoting is performed on the last n-k
!            columns of MAT.
!            When KRANK>=min(m,n) is used, pivoting is not performed. This is
!            appropriate when MAT is known to be full rank. KRANK<=0 can be used
!            when pivoting is desired on all columns of MAT.
!
!            On exit, KRANK contains the effective rank of MAT, i.e., the order of
!            the submatrix R11. This is the same as the order of the submatrix T11
!            in the complete orthogonal factorization of MAT.
!
!   TOL      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            If TOL is present and is in [0,1[, then :
!
!            - On entry, the calculations to determine the condition number of MAT
!              are performed. Then, TOL is used to determine the effective rank
!              of MAT, which is defined as the order of the largest leading
!              triangular submatrix R11 in the QR factorization with pivoting of MAT,
!              whose estimated condition number < 1/TOL. If TOL=0 is specified the
!              numerical rank of MAT is determined.
!            - On exit, the reciprocal of the condition number is returned in TOL.
!
!            If TOL is not specified or is outside [0,1[ :
!
!            - The calculations to determine the condition number of MAT are not
!              performed and crude tests on R(j,j) are done to determine
!              the numerical rank of MAT. If TOL is present, it is not changed.
!
!   TAU     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!           On entry, if TAU is present, a complete orthogonal factorization of
!           MAT is computed. Otherwise, only a QR factorization with column pivoting
!           of MAT is computed.
!
!           On exit, the scalars factors of the elementary reflectors defining Z.
!           See Further Details.
!
!           The size of TAU must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   1) If it is possible that MAT may not be full rank (i.e., certain columns of MAT are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be determined by using KRANK=0  and TOL=relative precision of the elements
!      in MAT. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!      are about the same order of magnitude.
!
!
!   2) The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = min( size(MAT,1) , size(MAT,2) )
!
!      Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!      v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!
!   3) The matrix P is represented in the array IP as follows:
!      If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!
!   4) If TAU is absent, a QR factorization with column pivoting of MAT is computed
!      and, on exit, the elements above the diagonal of the array MAT contain the 
!      corresponding elements of the triangular matrix R. The elements of the diagonal
!      of R are stored in the array DIAGR. 
!
!
!   5) If TAU is present, a complete orthogonal factorization of MAT is computed.
!      The factorization is obtained by Householder's method. The kth transformation
!      matrix, Z(k), which is used to introduce zeros into the kth row of R (e.g. in
!      R12), is given in the form
!
!      [  I  0    ]
!
!      [  0  T(k) ]
!
!      where
!
!           T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!      tau is a scalar and z(k) is an (n-KRANK) element vector. tau and z(k) are chosen
!      to annihilate the elements of the kth row of R12.
!
!      The Z n-by-n orthogonal matrix is given by
!
!            Z = Z(1) * Z(2) * ... * Z(KRANK)
!
!      On exit, the scalar tau is returned in the kth element of TAU and the vector u(k)
!      in the kth row of MAT, such that the elements of z(k) are in MAT(k,KRANK+1:n).
!
!      On exit, the elements above the diagonal of the array section MAT(1:KRANK,1:KRANK)
!      contain the corresponding elements of the triangular matrix T11. The elements of
!      the diagonal of T11 are stored in the array section DIAGR(1:KRANK).
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!          Prentice-Hall.
!
!   (2) G.H. Golub and C.F. Van Loan, 1996: Matrix Computations. The Johns
!          Hopkins University Press, Baltimore, Maryland.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants, only : machhuge
    use Utilities,     only : assert_eq, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
    real(stnd),   intent(inout),                 optional   :: tol
    real(stnd),   intent(out),   dimension(:),   optional   :: tau
!
    integer(i4b), intent(inout)               :: krank
    integer(i4b), intent(out),   dimension(:) :: ip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tau_tmp(size(ip))
!
    integer(i4b) :: nlig, ncol, ndiag, kfix 
!
    logical(lgl) :: comp_min_norm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qr_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    nlig = size(mat,1)
!
    ncol =  assert_eq( int(size(mat,2),i4b) ,   &
                       int(size(ip),i4b)    ,   &
                       name_proc )
!
    if ( present(tau) ) then
!
        ndiag =  assert_eq( min(ncol,nlig)       ,   &
                            int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            int(size(tau),i4b)   ,   &
                            name_proc )
!
        comp_min_norm = true
        tau(:ndiag) = zero
!
    else
!
        ndiag =  assert_eq( min(ncol,nlig)       ,   &
                            int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            name_proc )
!
        comp_min_norm = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b .or. all(mat(:nlig,:ncol)==zero) ) then
!
        krank = 0_i4b
        diagr(:ndiag) = zero
        beta(:ndiag)  = zero
        ip(:ncol) = arth( 1_i4b, 1_i4b, ncol )
!
        if ( present(tol) ) tol = machhuge
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF MAT .
!    
    kfix = max( krank, 0_i4b )
    krank = ncol
!    
    call qrfac( name_proc, mat, kfix, krank, comp_min_norm, diagr, beta, tau_tmp,     &
                tol=tol, ip=ip )
!
!   OUTPUT OPTIONAL ARGUMENT tau ON EXIT, IF REQUIRED.
!
    if ( comp_min_norm ) tau(:krank) = tau_tmp(:krank)
!
!
! END OF SUBROUTINE qr_cmp2
! _________________________
!
    end subroutine qr_cmp2
!
! =========================================================================================
!
    subroutine ortho_gen_qr( mat, beta )
!
! Purpose
! _______
!
!   ORTHO_GEN_QR generates an m-by-n real matrix with orthonormal columns, which is
!   defined as the first n columns of a product of k elementary reflectors of order m
!
!            Q = H(1) * H(2) * ... * H(k)
!
!   as returned by QR_CMP or QR_CMP2.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the i-th column must contain the vector which
!            defines the elementary reflector H(i), for i = 1,2,...,k,
!            as returned by QR_CMP (or QR_CMP2) in the first k columns
!            of its array argument MAT.
!
!            On exit, the first n columns of Q.
!
!            The shape of MAT must verify size( MAT, 2 ) <= size( MAT, 1 ).
!
!   BETA     (INPUT) real(stnd), dimension(:)
!            BETA(i) must contain the scalar factor of the elementary
!            reflector H(i), as returned by QR_CMP (or QR_CMP2).
!            The size of BETA determines the number k of elementary reflectors
!            whose product defines the matrix Q.
!
!            The size of BETA must verify size size( BETA ) <= ( MAT, 2 ) .
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the lower triangle
!   of MAT and generating the orthogonal matrix Q of the QR factorization.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!           The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989: Block
!           reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988: Implementation of the GMRES method using Householder
!           transformations. Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,        only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: beta
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: prod, beta1
    real(stnd), dimension( blksz )              :: x
    real(stnd), dimension( size(mat,1) )        :: tmp
    real(stnd), dimension( size(mat,1), blksz ) :: umat
    real(stnd), dimension( blksz, blksz )       :: lmat
!
    integer(i4b) :: m, n, k, i, i0, i1, i2, l, l2, j, blksz2
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
    k = size( beta )
!
    call assert( logical( (k<=n) .and. (m>=n), lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   INITIALIZE COLUMNS k+1:n OF mat TO COLUMNS OF THE UNIT MATRIX.
!
    mat(:m,k+1_i4b:n) = zero
!
    do i = k+1_i4b, n
        mat(i,i) = one
    end do
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/i>=omp_limit2                     .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                               &
!$OMP         ,PRIVATE(i,i0,i1,i2,l,l2,j,blksz2,beta1,prod,x,tmp)
!
    do i = k - blksz, -blksz+1_i4b, -blksz
!    
!       APPLY H(i+1)* ...*H(i+blksz) TO mat(i+1:m,i+blksz+1:n) FROM THE LEFT
!       WITH A BLOCKED ALGORITHM AS
!
!    ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*mat(i+1:m,i+blksz+1:n) .
!
        blksz2 = min( blksz, i+blksz )
        i0     = max( 0_i4b, i )
!
!       SAVE H(i+1) ...H(i+blksz) FOR LATER USE AND COMPUTE THE blksz BY blksz UNIT UPPER
!       TRIANGULAR MATRIX lmat(:blksz,:blksz) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do l = 1_i4b, blksz2
!
            i1 = i0 + l
!
            umat(i1:m,l) = mat(i1:m,i1)
!
            do l2 = 1_i4b, l - 1_i4b
!
                i2 = i0 + l2
!
                lmat(l2,l) = -beta(i2)*dot_product( mat(i1:m,i1), mat(i1:m,i2) )
!
            end do
!                    
        end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do j = i0+blksz2+1_i4b, n
!
!           COMPUTE U(i+1:m,:blksz)'*mat(i+1:m,j) .
!
            tmp(i0+1_i4b:m) = mat(i0+1_i4b:m,j)
!
            do l = 1_i4b, blksz2
!
                i1 = i0 + l
!
                x(l) = beta(i1)*dot_product( umat(i1:m,l), tmp(i1:m) )
!                    
            end do
!
!           SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*mat(i+1:m,j)  AND
!           COMPUTE mat(i+1:m,j) = mat(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
            do l = blksz2, 1_i4b, -1_i4b
!
                x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                i1 = i0 + l
#ifdef _BLAS
                call axpy( m-i1+1_i4b, x(l), umat(i1:m,l), 1_i4b, tmp(i1:m), 1_i4b )
#else
                tmp(i1:m) = tmp(i1:m) + x(l)*umat(i1:m,l)
#endif
!
            end do
!
            mat(i0+1_i4b:m,j) = tmp(i0+1_i4b:m)
!                    
        end do
!
!$OMP END DO NOWAIT
!    
!       APPLY H(i+1)* ...*H(i+blksz) TO mat(i+1:m,i+1:i+blksz) FROM THE LEFT
!       WITH A SERIAL ALGORITHM.
!
!$OMP SINGLE
!
        do l = blksz2, 1_i4b, -1_i4b
!
            i1 = i0 + l
!
            beta1 = beta(i1)
!
!           APPLY H(i1) TO mat(i1:m,i1+1:i+blksz) FROM THE LEFT.
!
            do l2 = l + 1_i4b, blksz2
!
                i2 = i0 + l2
!
                prod = beta1*dot_product( umat(i1+1_i4b:m,l), mat(i1+1_i4b:m,i2) )
#ifdef _BLAS
                call axpy( m-i1+1_i4b, prod, umat(i1:m,l), 1_i4b, mat(i1:m,i2), 1_i4b )
#else
                mat(i1:m,i2) = mat(i1:m,i2) + prod*umat(i1:m,l)
#endif
!    
            end do
!
!           APPLY H(i1) TO mat(i1:m,i1) FROM THE LEFT.
!
            mat(1_i4b:i1-1_i4b,i1) = zero
            mat(i1,i1)             = one + beta1
            mat(i1+1_i4b:m,i1)     = beta1*umat(i1+1_i4b:m,l)
!    
        end do
!    
!$OMP END SINGLE
!
    end do
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE ortho_gen_qr
! ______________________________
!
    end subroutine ortho_gen_qr
!
! =========================================================================================
!
    subroutine apply_q_qr_rm( mat, beta, c, left, trans  )
!
! Purpose
! _______
!
!   APPLY_Q_QR overwrites the general real m-by-n matrix C with
!
!            Q * C   if LEFT = true  and TRANS = false, or
!
!            Q' * C  if LEFT = true  and TRANS = true,  or
!
!            C * Q   if LEFT = false and TRANS = false, or
!
!            C * Q'  if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                    Q = H(1) * H(2) * . . . * H(k)
!
!   as returned by QR_CMP or QR_CMP2. Q is of order m if LEFT = true and of order n
!   if LEFT = false.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th column must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by QR_CMP (or QR_CMP2) in the first k columns of its array
!                argument MAT. MAT is not modified by the routine.
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                BETA(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by QR_CMP (or QR_CMP2).
!                The size of BETA determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of BETA must verify :
!
!                        size( BETA ) <= min( size( MAT, 1 ), ( MAT, 2 ) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify :
!
!                - if LEFT = true,  size( C, 1 ) = size( MAT, 1 )
!                - if LEFT = false, size( C, 2 ) = size( MAT, 1 ) .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left
!                - LEFT = false : apply Q or Q' from the right
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine DORM2R in LAPACK.
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the lower triangle
!   of MAT and applying the orthogonal matrix Q of the QR factorization to the
!   real m-by-n matrix C.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm used here, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!           The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989: Block
!           reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988: Implementation of the GMRES method using Householder
!           transformations. Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: beta
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                           :: beta1, prod
    real(stnd), dimension( blksz )                       :: x
    real(stnd), dimension( max( size(c,1), size(c,2) ) ) :: tmp
    real(stnd), dimension( blksz, blksz )                :: lmat
    real(stnd), dimension( :, : ), allocatable           :: x2
!
    integer(i4b) :: m, n, nm, nq, k, i, i1, i2, j, j1, j2, j3, l, l2, p
    integer      :: iok
!
    logical(lgl) :: ascending
!
#ifdef _OPENMP
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c, 1 )
    n = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
    k = size( beta )
!
    call assert( logical( nq==size(mat,1), lgl ),             &
                 logical( k<=min(nq,size(mat,2)), lgl ),      &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  nm<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  (nm/nq)/i>=omp_limit2               .and.      &
                  i>1_i4b
#endif
!
    ascending = (left.and.trans) .or. .not.(left.or.trans)
!
!$OMP PARALLEL IF(test_par)                                           &
!$OMP         ,PRIVATE(iok,i,i1,i2,l,l2,j,j1,j2,j3,jinc,jrem,joffset, &
!$OMP                  num_threads,thread_num,beta1,prod,x,x2,tmp,lmat)
!
#ifdef _OPENMP
    if ( test_par ) then
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem        = mod( p, num_threads )
        jinc        = ( p - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
    else
#endif
        j1 = 1_i4b
        j2 = p
#ifdef _OPENMP
    end if
#endif
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
    if ( left ) then
!
        if ( ascending ) then
!
!           COMPUTE H(k)* . . .*H(2)*H(1)*c(:m,:n) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+blksz)* . . . *H(i+1) TO c(i+1:m,:n) FROM THE LEFT AS
!
!             ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,:n) .
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 1_i4b, blksz - 1_i4b
!
                    i1 = i + l
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i2:m,i1), mat(i2:m,i2) )
!
                    end do
!                    
                end do
!
                do j = j1, j2
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j) .
!
                    tmp(i+1_i4b:m) = c(i+1_i4b:m,j)
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(l) = beta(i1)*dot_product( mat(i1:m,i1), tmp(i1:m) )
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*c(i+1:m,j)  AND
!                   COMPUTE c(i+1:m,j) = c(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
                    do l = 1_i4b, blksz
!
                        x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                        i1 = i + l
!
#ifdef _BLAS
                        call axpy( m-i1+1_i4b, x(l), mat(i1:m,i1), 1_i4b, tmp(i1:m), 1_i4b )
#else
                        tmp(i1:m) = tmp(i1:m) + x(l)*mat(i1:m,i1)
#endif
!
                    end do
!
                    c(i+1_i4b:m,j) = tmp(i+1_i4b:m)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                tmp(i:m) = mat(i:m,i)
!
                do j = j1, j2
!
                    prod = dot_product( tmp(i:m), c(i:m,j) )*beta1
#ifdef _BLAS
                    call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m,j), 1_i4b )
#else
                    c(i:m,j) = c(i:m,j) + prod*tmp(i:m)          
#endif
!
                end do            
!
            end do
!
        else
!
!           COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
            do i = k - blksz, 0_i4b, -blksz
!
!               APPLY H(i+1)* . . . *H(i+blksz) TO c(i+1:m,:n) FROM THE LEFT AS
!
!             ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,:n) .
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz
!
                    i1 = i + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i1:m,i1), mat(i1:m,i2) )
!
                    end do
!                    
                end do
!
                do j = j1, j2
!
                    tmp(i+1_i4b:m) = c(i+1_i4b:m,j)
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j) .
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(l) = beta(i1)*dot_product( mat(i1:m,i1), tmp(i1:m) )
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*c(i+1:m,j)  AND
!                   COMPUTE c(i+1:m,j) = c(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
                    do l = blksz, 1_i4b, -1_i4b
!
                        x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                        i1 = i + l
!
#ifdef _BLAS
                        call axpy( m-i1+1_i4b, x(l), mat(i1:m,i1), 1_i4b, tmp(i1:m), 1_i4b )
#else
                        tmp(i1:m) = tmp(i1:m) + x(l)*mat(i1:m,i1)
#endif
!
                    end do
!
                    c(i+1_i4b:m,j) = tmp(i+1_i4b:m)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                tmp(i:m) = mat(i:m,i)
!
                do j = j1, j2
!
                    prod = dot_product( tmp(i:m), c(i:m,j) )*beta1
#ifdef _BLAS
                    call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m,j), 1_i4b )
#else
                    c(i:m,j) = c(i:m,j) + prod*tmp(i:m)          
#endif
!
                end do            
!
            end do
!
        end if
!
    else
!
        j3 = j2 - j1 + 1_i4b
!
!       ALLOCATE WORK VARIABLES.
!
        allocate( x2(j3,blksz), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( ascending ) then
!
!           COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(k) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+1)* . . .*H(i+blksz) TO c(:m,i+1:n) FROM THE RIGHT AS
!
!             c(:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 1_i4b, blksz - 1_i4b
!
                    i1 = i + l
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i2:n,i1), mat(i2:n,i2) )
!
                    end do
!                    
                end do
!
!               COMPUTE c(1:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
                    x2(1_i4b:j3,l) = beta(i1)*matmul( c(j1:j2,i1:n), mat(i1:n,i1) )
!                    
                end do
!
!               SOLVE x(1:m,:blksz)*lmat(:blksz,:blksz) = c(1:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE c(1:m,i+1:i+blksz-1) = c(1:m,i+1:i+blksz-1) + x(1:m,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = l + 1_i4b, blksz
!
#ifdef _BLAS
                        call axpy( j3, -lmat(l2,l), tmp(1_i4b:j3), 1_i4b, x2(1_i4b:j3,l2), 1_i4b )
#else
                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - tmp(1_i4b:j3)*lmat(l2,l)
#endif
!
                        i2 = i + l2 - 1_i4b
!
#ifdef _BLAS
                        call axpy( j3, mat(i2,i1), tmp(1_i4b:j3), 1_i4b, c(j1:j2,i2), 1_i4b )
#else
                        c(j1:j2,i2) = c(j1:j2,i2) + tmp(1_i4b:j3)*mat(i2,i1)
#endif
!
                    end do
!
                end do
!
!               COMPUTE c(1:m,i+blksz:n) = c(1:m,i+blksz:n) + x(1:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i + blksz, n
!
                    tmp(1_i4b:j3) = c(j1:j2,j)
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
#ifdef _BLAS
                        call axpy( j3, mat(j,i1), x2(1_i4b:j3,l), 1_i4b, tmp(1_i4b:j3), 1_i4b )
#else
                        tmp(1_i4b:j3) = tmp(1_i4b:j3) + x2(1_i4b:j3,l)*mat(j,i1)
#endif
!                    
                    end do
!
                    c(j1:j2,j) = tmp(1_i4b:j3)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(:m,i:n) FROM THE RIGHT.
!
                tmp(i:n) = mat(i:n,i)
!
                do j = j1, j2
!
                    prod     = dot_product( tmp(i:n), c(j,i:n) )*beta1
                    c(j,i:n) = c(j,i:n) + prod*tmp(i:n)       
!
                end do            
!
            end do
!
        else
!
!           COMPUTE c(:m,:n)*H(k)* . . .*H(2)*H(1) .
!
            do i = k - blksz, 0_i4b, -blksz
!
!               APPLY H(i+blksz)* . . .*H(i+1) TO c(1:m,i+1:n) FROM THE RIGHT AS
!
!             c(1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz
!
                    i1 = i + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i1:n,i1), mat(i1:n,i2) )
!
                    end do
!                    
                end do
!
!               COMPUTE c(1:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
                    x2(1_i4b:j3,l) = beta(i1)*matmul( c(j1:j2,i1:n), mat(i1:n,i1) )
!                    
                end do
!
!               SOLVE x(1:m,:blksz)*lmat(:blksz,:blksz) = c(1:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE c(1:m,i+1:i+blksz-1) = c(1:m,i+1:i+blksz-1) + x(1:m,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                do l = blksz, 1_i4b, -1_i4b
!
                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = 1_i4b, l-1_i4b
!
#ifdef _BLAS
                        call axpy( j3, -lmat(l2,l), tmp(1_i4b:j3), 1_i4b, x2(1_i4b:j3,l2), 1_i4b )
#else
                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - tmp(1_i4b:j3)*lmat(l2,l)
#endif
!
                    end do
!
                    i1 = i + l
!
                    do l2 = l, blksz-1_i4b
!
                        i2 = i + l2
!
#ifdef _BLAS
                        call axpy( j3, mat(i2,i1), tmp(1_i4b:j3), 1_i4b, c(j1:j2,i2), 1_i4b )
#else
                        c(j1:j2,i2) = c(j1:j2,i2) + tmp(1_i4b:j3)*mat(i2,i1)
#endif
!
                    end do
!
                end do
!
!               COMPUTE c(1:m,i+blksz:n) = c(1:m,i+blksz:n) + x(1:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i + blksz, n
!
                    tmp(1_i4b:j3) = c(j1:j2,j)
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
#ifdef _BLAS
                        call axpy( j3, mat(j,i1), x2(1_i4b:j3,l), 1_i4b, tmp(1_i4b:j3), 1_i4b )
#else
                        tmp(1_i4b:j3) = tmp(1_i4b:j3) + x2(1_i4b:j3,l)*mat(j,i1)
#endif
!                    
                    end do
!
                    c(j1:j2,j) = tmp(1_i4b:j3)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(:m,i:n) FROM THE RIGHT.
!
                tmp(i:n) = mat(i:n,i)
!
                do j = j1, j2
!
                    prod     = dot_product( tmp(i:n), c(j,i:n) )*beta1
                    c(j,i:n) = c(j,i:n) + prod*tmp(i:n)             
!
                end do            
!
            end do
!
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( x2 )
!
!
    end if
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE apply_q_qr_rm
! _______________________________
!
    end subroutine apply_q_qr_rm
!
! =========================================================================================
!
    subroutine apply_q_qr_rv( mat, beta, c, trans  )
!
! Purpose
! _______
!
!   APPLY_Q_QR overwrites the real m vector C with
!
!            Q * C   if TRANS = false, or
!
!            Q' * C  if TRANS = true
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                    Q = H(1) * H(2) * . . . * H(k)
!
!   as returned by QR_CMP or QR_CMP2. Q is of order m .
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th column must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by QR_CMP (or QR_CMP2) in the first k columns of its array
!                argument MAT. MAT is not modified by the routine.
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                BETA(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by QR_CMP (or QR_CMP2).
!                The size of BETA determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of BETA must verify :
!
!                        size( BETA ) <= min( size( MAT, 1 ), ( MAT, 2 ) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m vector C.
!
!                On exit, C is overwritten by Q * C or Q' * C .
!
!                The shape of C must verify size( C ) = size( MAT, 1 ) .
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine DORM2R in LAPACK.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm used here, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974: Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and van Loan, C.F., 1996: Matrix Computations, 3rd ed.
!           The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: beta
    real(stnd),   intent(inout), dimension(:)   :: c
!
    logical(lgl), intent(in) :: trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                        :: beta1, prod
    real(stnd), dimension( size(c) )  :: tmp
!
    integer(i4b) :: m, k, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c )
    k = size( beta )
!
    call assert( logical( m==size(mat,1), lgl ),            &
                 logical( k<=min(m,size(mat,2)), lgl ),     &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  m<=0_i4b ) return
!
    if ( trans ) then
!
!       COMPUTE H(k)* . . .*H(2)*H(1)*c(:m) .
!
        do i = 1_i4b, k
!
            beta1 = beta(i)
!
            if ( beta1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i:m,i)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*beta1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)          
#endif
!
        end do
!
    else
!
!       COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
        do i = k, 1_i4b, -1_i4b
!
            beta1 = beta(i)
!
            if ( beta1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i:m,i)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*beta1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)             
#endif
!
        end do
!
    end if
!
!
! END OF SUBROUTINE apply_q_qr_rv
! _______________________________
!
    end subroutine apply_q_qr_rv
!
!
! =========================================================================================
!
!
! *****************************
! END OF MODULE QR_Procedures *
! *****************************
!    
end module QR_Procedures
