module Sort_Procedures
! ####################################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING SORTING UTILITIES.                                                *
!                                                                                    * 
! LATEST REVISION : 16/04/2018                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!
!
! USED MODULES 
! ============
!
    use Select_Parameters, only : lgl, i4b, stnd
!
! OTHER MODULES USED IN SPECIFIC SUBROUTINES
! ==========================================
!
!   use Logical_Constants
!   use Utilities
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES
! ===============
!    
    private
    public :: tri_insert, quick_sort, do_index, rank, reorder
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface tri_insert
        module procedure    tri_insert_rv,     tri_insert_iv,            &
                            tri_insert_rv_ord, tri_insert_iv_ord
    end interface
!
    interface quick_sort
        module procedure    quick_sort_rv,     quick_sort_iv,            &
                            quick_sort_rv_ord, quick_sort_iv_ord
    end interface
!
    interface do_index
        module procedure    index_iv, index_rv
    end interface
!
    interface reorder
        module procedure    reorder_rv, reorder_rm,              &
                            reorder_iv, reorder_im,              &
                            reorder_cv, reorder_cm
    end interface
!   
! MODULE PARAMETERS 
! =================
!
    integer(i4b), parameter  :: mini = 9 ! MAXIMAL SIZE TO USE STRAIGHT INSERTION
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!
    subroutine tri_insert_iv( list )
!
! Purpose
! _______
!
!   Sort the integer array LIST into ascending numerical order, by straight insertion.
!   LIST is replaced on output by its sorted rearrangement.
!
!
! Arguments
! _________
!
!   LIST    (INPUT/OUTPUT) integer(i4b), dimension(:)
!           The integer vector to sort.
!
!
! ___________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(inout) :: list(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: k, j, k_elt
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    do k = 2_i4b, size( list )
!
        k_elt = list( k )
!
        do j = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(j)<k_elt ) exit
!
        end do
!
        list(j+2_i4b:k) = list(j+1_i4b:k-1_i4b)
        list(j+1_i4b) = k_elt
!
    end do
!
!
! END OF SUBROUTINE tri_insert_iv
! _______________________________
!
    end subroutine tri_insert_iv
!
! =========================================================================================
!
    subroutine tri_insert_iv_ord( list, order )
!
! Purpose
! _______
!
!   Sort the integer array LIST into ascending numerical order, by straight insertion.
!   LIST is replaced on output by its sorted rearrangement.
!
!   ORDER is an associated integer array which gives the positions 
!   of the elements in the original order.
!
!
! Arguments
! _________
!
!   LIST    (INPUT/OUTPUT) integer(i4b), dimension(:)
!           The integer vector to sort.
!
!   ORDER   (OUTPUT) integer(i4b), dimension(:)
!           Array which gives the positions of the elements
!           in the original order.
!
!
! Further Details
! _______________
!
!   The size of LIST and ORDER must match.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(inout) :: list(:)
    integer(i4b), intent(out)   :: order(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, k, j, k_elt
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( list )
    j = size( order )
!
    n = assert_eq( k, j, 'tri_insert' )
!
    if ( n==0_i4b ) return
!
    order( 1 ) = 1_i4b
!
    do k = 2_i4b, n
!
        k_elt = list( k )
!
        do j = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(j)<k_elt ) exit
!
        end do
!
        list(j+2_i4b:k) = list(j+1_i4b:k-1_i4b)
        list(j+1_i4b) = k_elt
!
        order(j+2_i4b:k) = order(j+1_i4b:k-1_i4b)
        order(j+1_i4b) = k
!
    end do
!
!
! END OF SUBROUTINE tri_insert_iv_ord
! ___________________________________
!
    end subroutine tri_insert_iv_ord
!
! =========================================================================================
!
    subroutine tri_insert_rv( list )
!
! Purpose
! _______
!
!   Sort the real array LIST into ascending numerical order, by straight insertion.
!   LIST is replaced on output by its sorted rearrangement.
!
!
! Arguments
! _________
!
!   LIST    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The real vector to sort.
!
!
! ___________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: list(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: k_elt
!
    integer(i4b) :: k, j
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    do k = 2_i4b, size( list )
!
        k_elt = list(k)
!
        do j = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(j)<k_elt ) exit
!
        end do
!
        list(j+2_i4b:k) = list(j+1_i4b:k-1_i4b)
        list(j+1_i4b)=k_elt
!
    end do
!
!
! END OF SUBROUTINE tri_insert_rv
! _______________________________
!
   end subroutine tri_insert_rv
!
! =========================================================================================
!
    subroutine tri_insert_rv_ord( list, order )
!
! Purpose
! _______
!
!   Sort the real array LIST into ascending numerical order, by straight insertion.
!   LIST is replaced on output by its sorted rearrangement.
!
!   ORDER is an associated integer array which gives the positions 
!   of the elements in the original order.
!
!
! Arguments
! _________
!
!   LIST    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The real vector to sort.
!
!   ORDER   (OUTPUT) integer(i4b), dimension(:)
!           Array which gives the positions of the elements
!           in the original order.
!
!
! Further Details
! _______________
!
!   The size of LIST and ORDER must match.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: list(:)
!
    integer(i4b), intent(out)   :: order(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: k_elt
!
    integer(i4b) :: n, k, j
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( list )
    j = size( order )
!
    n = assert_eq( k, j, 'tri_insert' )
!
    if ( n==0_i4b ) return
!
    order( 1 ) = 1_i4b
!
    do k = 2_i4b, n
!
        k_elt = list(k)
!
        do j = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(j)<k_elt ) exit
!
        end do
!
        list(j+2_i4b:k) = list(j+1_i4b:k-1_i4b)
        list(j+1_i4b)=k_elt
!
        order(j+2_i4b:k) = order(j+1_i4b:k-1_i4b)
        order(j+1_i4b) = k
!
    end do
!
!
! END OF SUBROUTINE tri_insert_rv_ord
! ___________________________________
!
   end subroutine tri_insert_rv_ord
!
! =========================================================================================
!
    subroutine quick_sort_iv( list, ascending )
!
! Purpose
! _______
!
!   Sort an integer array LIST into ascending or descending order using the QuickSort algorithm.
!   LIST is replaced on output by its sorted rearrangement.
!
!
! Arguments
! _________
!
!   LIST        (INPUT/OUTPUT) integer(i4b), dimension(:)
!               The integer vector to sort.
!
!   ASCENDING   (INPUT, OPTIONAL) logical(lgl)
!               Sort the array list into ascending order if ASCENDING = true,
!               or in descending order if ASCENDING = false. The default is true.
!
!
! Further Details
! _______________
!
!   Quick sort routine adapted (and modified to reverse order) from:
!
!   (1) Brainerd, W.S., Goldberg, C.H., and Adams, J.C., 1990: Programmer's Guide to
!              Fortran 90, McGraw-Hill,  ISBN 0-07-000248-7, pages 149-150.
!
!
! ___________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(inout)          :: list(:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, temp, reference
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( list )
!
    if ( n<=1_i4b ) return
!
!   USE PARTITION ("QUICK") SORT.
!
    call quick_sort_1( 1_i4b, n )
!
!   USE INTERCHANGE SORT FOR SMALL LISTS.
!
    call tri_insert_iv( list )
!
!   DESCENDING SORT.
!
    if ( present(ascending) ) then
        if ( .not.ascending ) list = list(n:1_i4b:-1_i4b)
    end if
!                                                                              
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        recursive subroutine quick_sort_1( left_end, right_end )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: left_end, right_end
!
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   ----------------------------------
!
        integer(i4b) :: left, right, i, j
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        left = left_end
        right = right_end
!
        do
!
            if ( right<left+mini ) exit
!
!           CHOOSE MEDIAN OF LEFT, CENTER AND RIGHT ELEMENTS 
!           AS PARTITIONING ELEMENT reference.
!
            if ( list(right)<list(left) ) then
                i = right ; j = left
            else
                i = left ; j =right
            end if
!
            reference = list( (left + right)/2_i4b )
!
            if ( reference<list(j) ) then
                if ( reference<list(i) ) reference = list(i)
            else
                reference = list(j)
            end if
!
            i = left - 1_i4b
            j = right + 1_i4b
!
            do
!
!               SCAN list FROM LEFT END UNTIL ELEMENT >= reference IS FOUND.
!
                do
                    i = i + 1_i4b
                    if ( list(i)>=reference ) exit
                end do
!
!               SCAN list FROM RIGHT END UNTIL ELEMENT <= reference IS FOUND.
!
                do
                    j = j - 1_i4b
                    if ( list(j)<=reference ) exit
                end do
!
                if ( i<j ) then
!
!                   SWAP TWO OUT-OF-ORDER ELEMENTS.
!
                    temp = list(i); list(i) = list(j); list(j) = temp
!
                else if ( i==j ) then
!
                    i = i + 1_i4b
!
                    exit
!
                else
!
                    exit
!
                end if
!
            end do
!
!           SORT THE TWO PARTITIONS.
!
            if ( j-left>right-i ) then
                call quick_sort_1( i, right )
                right = j
            else
                call quick_sort_1( left, j )
                left = i
            end if
!
        end do
!
!
!   END OF SUBROUTINE quick_sort_1
!   ------------------------------
!
        end subroutine quick_sort_1
!
! ___________________________________________________________________________________________
!
!
! END OF SUBROUTINE quick_sort_iv
! _______________________________
!
    end subroutine quick_sort_iv
!
! =========================================================================================
!
    subroutine quick_sort_rv( list, ascending )
!
! Purpose
! _______
!
!   Sort a real array LIST into ascending or descending order using the QuickSort algorithm.
!   LIST is replaced on output by its sorted rearrangement.
!
!
! Arguments
! _________
!
!   LIST        (INPUT/OUTPUT) real(stnd), dimension(:)
!               The real vector to sort.
!
!   ASCENDING   (INPUT, OPTIONAL) logical(lgl)
!               Sort the array list into ascending order if ASCENDING = true,
!               or in descending order if ASCENDING = false. The default is true.
!
!
! Further Details
! _______________
!
!   Quick sort routine adapted (and modified to reverse order) from:
!
!   (1) Brainerd, W.S., Goldberg, C.H., and Adams, J.C., 1990: Programmer's Guide to
!              Fortran 90, McGraw-Hill,  ISBN 0-07-000248-7, pages 149-150.
!
!
! ___________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)          :: list(:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n
!
    real(stnd)   :: temp, reference
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( list )
!
    if ( n<=1_i4b ) return
!
!   USE PARTITION ("QUICK") SORT.
!
    call quick_sort_1( 1_i4b, n )
!
!   USE INTERCHANGE SORT FOR SMALL LISTS.
!
    call tri_insert_rv( list )
!
!   DESCENDING SORT.
!
    if ( present(ascending) ) then
        if ( .not.ascending ) list = list(n:1_i4b:-1_i4b)
    end if
!                                                                              
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        recursive subroutine quick_sort_1( left_end, right_end )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: left_end, right_end
!
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   ----------------------------------
!
        integer(i4b) :: left, right, i, j
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        left = left_end
        right = right_end
!
        do
!
            if ( right<left+mini ) exit
!
!           CHOOSE MEDIAN OF LEFT, CENTER AND RIGHT ELEMENTS 
!           AS PARTITIONING ELEMENT reference.
!
            if ( list(right)<list(left) ) then
                i = right ; j = left
            else
                i = left ; j =right
            end if
!
            reference = list( (left + right)/2_i4b )
!
            if ( reference<list(j) ) then
                if ( reference<list(i) ) reference = list(i)
            else
                reference = list(j)
            end if
!
            i = left - 1_i4b
            j = right + 1_i4b
!
            do
!
!               SCAN list FROM LEFT END UNTIL ELEMENT >= reference IS FOUND.
!
                do
                    i = i + 1_i4b
                    if ( list(i)>=reference ) exit
                end do
!
!               SCAN list FROM RIGHT END UNTIL ELEMENT <= reference IS FOUND.
!
                do
                    j = j - 1_i4b
                    if ( list(j)<=reference ) exit
                end do
!
                if ( i<j ) then
!
!                   SWAP TWO OUT-OF-ORDER ELEMENTS.
!
                    temp = list(i); list(i) = list(j); list(j) = temp
!
                else if ( i==j ) then
!
                    i = i + 1_i4b
!
                    exit
!
                else
!
                    exit
!
                end if
!
            end do
!
!           SORT THE TWO PARTITIONS.
!
            if ( j-left>right-i ) then
                call quick_sort_1( i, right )
                right = j
            else
                call quick_sort_1( left, j )
                left = i
            end if
!
        end do
!
!
!   END OF SUBROUTINE quick_sort_1
!   ------------------------------
!
        end subroutine quick_sort_1
!
! ___________________________________________________________________________________________
!
!
! END OF SUBROUTINE quick_sort_rv
! _______________________________
!
    end subroutine quick_sort_rv
!
! =========================================================================================
!
    subroutine quick_sort_iv_ord( list, order, ascending )
!
! Purpose
! _______
!
!   Sort an integer array LIST into ascending or descending order using the QuickSort algorithm.
!   LIST is replaced on output by its sorted rearrangement.
!
!   ORDER is an associated integer array which gives the positions 
!   of the elements in the original order.
!
!
! Arguments
! _________
!
!   LIST        (INPUT/OUTPUT) integer(i4b), dimension(:)
!               The integer vector to sort.
!
!   ORDER       (OUTPUT) integer(i4b), dimension(:)
!               Array which gives the positions of the elements
!               in the original order.
!
!   ASCENDING   (INPUT, OPTIONAL) logical(lgl)
!               Sort the array list into ascending order if ASCENDING = true,
!               or in descending order if ASCENDING = false. The default is true.
!
!
! Further Details
! _______________
!
!   Quick sort routine adapted from reference (1), modified to include an
!   associated integer array, which gives the positions of the elements
!   in the original order, and also modified to reverse order. 
!
!   The size of LIST and ORDER must match.
!
!   For further details, see:
!
!   (1) Brainerd, W.S., Goldberg, C.H., and Adams, J.C., 1990: Programmer's Guide to
!              Fortran 90, McGraw-Hill,  ISBN 0-07-000248-7, pages 149-150.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(inout)          :: list(:)
    integer(i4b), intent(out)            :: order(:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: k, l, n, temp, reference
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( list )
    l = size( order )
!
    n = assert_eq( k, l, 'quick_sort' )
!
    order = arth( 1_i4b, 1_i4b, n )
!
!   USE PARTITION ("QUICK") SORT.
!
    call quick_sort_1( 1_i4b, n )
!
!   USE INTERCHANGE SORT FOR SMALL LISTS.
!
    do k = 2_i4b, n
!
        temp = list( k )
!
        do l = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(l)<temp ) exit
!
        end do
!
        list(l+2_i4b:k) = list(l+1_i4b:k-1_i4b)
        list(l+1_i4b) = temp
!
        temp = order(k)
        order(l+2_i4b:k) = order(l+1_i4b:k-1_i4b)
        order(l+1_i4b) = temp
!
    end do
!
!   DESCENDING SORT.
!
    if ( present(ascending) ) then
!
        if ( .not.ascending ) then
            list = list(n:1_i4b:-1_i4b)
            order = order(n:1_i4b:-1_i4b)
        end if
!
    end if
!                                                                              
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        recursive subroutine quick_sort_1( left_end, right_end )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: left_end, right_end
!
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   ----------------------------------
!
        integer(i4b) :: left, right, i, j
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        left = left_end
        right = right_end
!
        do
!
            if ( right<left+mini ) exit
!
!           CHOOSE MEDIAN OF LEFT, CENTER AND RIGHT ELEMENTS 
!           AS PARTITIONING ELEMENT reference.
!
            if ( list(right)<list(left) ) then
                i = right ; j = left
            else
                i = left ; j =right
            end if
!
            reference = list( (left + right)/2_i4b )
!
            if ( reference<list(j) ) then
                if ( reference<list(i) ) reference = list(i)
            else
                reference = list(j)
            end if
!
            i = left - 1_i4b
            j = right + 1_i4b
!
            do
!
!               SCAN list FROM LEFT END UNTIL ELEMENT >= reference IS FOUND.
!
                do
                    i = i + 1_i4b
                    if ( list(i)>=reference ) exit
                end do
!
!               SCAN list FROM RIGHT END UNTIL ELEMENT <= reference IS FOUND.
!
                do
                    j = j - 1_i4b
                    if ( list(j)<=reference ) exit
                end do
!
                if ( i<j ) then
!
!                   SWAP TWO OUT-OF-ORDER ELEMENTS.
!
                    temp = list(i); list(i) = list(j); list(j) = temp
                    temp = order(i); order(i) = order(j); order(j) = temp
!
                else if ( i==j ) then
!
                    i = i + 1_i4b
!
                    exit
!
                else
!
                    exit
!
                end if
!
            end do
!
!           SORT THE TWO PARTITIONS.
!
            if ( j-left>right-i ) then
                call quick_sort_1( i, right )
                right = j
            else
                call quick_sort_1( left, j )
                left = i
            end if
!
        end do
!
!
!   END OF SUBROUTINE quick_sort_1
!   ------------------------------
!
        end subroutine quick_sort_1
!
! ___________________________________________________________________________________________
!
!
! END OF SUBROUTINE quick_sort_iv_ord
! ___________________________________
!
    end subroutine quick_sort_iv_ord
!
! =========================================================================================
!
    subroutine quick_sort_rv_ord( list, order, ascending )
!
! Purpose
! _______
!
!   Sort a real array LIST into ascending or descending order using the QuickSort algorithm.
!   LIST is replaced on output by its sorted rearrangement.
!
!   ORDER is an associated integer array which gives the positions 
!   of the elements in the original order.
!
!
! Arguments
! _________
!
!   LIST        (INPUT/OUTPUT) real(stnd), dimension(:)
!               The real vector to sort.
!
!   ORDER       (OUTPUT) integer(i4b), dimension(:)
!               Array which gives the positions of the elements
!               in the original order.
!
!   ASCENDING   (INPUT, OPTIONAL) logical(lgl)
!               Sort the array list into ascending order if ASCENDING = true,
!               or in descending order if ASCENDING = false. The default is true.
!
!
! Further Details
! _______________
!
!   Quick sort routine adapted from reference (1), modified to include an
!   associated integer array, which gives the positions of the elements
!   in the original order, and also modified to reverse order. 
!
!   The size of LIST and ORDER must match.
!
!   For further details, see:
!
!   (1) Brainerd, W.S., Goldberg, C.H., and Adams, J.C., 1990: Programmer's Guide to
!              Fortran 90, McGraw-Hill,  ISBN 0-07-000248-7, pages 149-150.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)          :: list(:)
!
    integer(i4b), intent(out)            :: order(:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: temp, reference
!
    integer(i4b) :: k, l, n, itemp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( list )
    l = size( order )
!
    n = assert_eq( k, l, 'quick_sort' )
!
    order = arth( 1_i4b, 1_i4b, n )
!
!   USE PARTITION ("QUICK") SORT.
!
    call quick_sort_1( 1_i4b, n )
!
!   USE INTERCHANGE SORT FOR SMALL LISTS.
!
    do k = 2_i4b, n
!
        temp = list( k )
!
        do l = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(l)<temp ) exit
!
        end do
!
        list(l+2_i4b:k) = list(l+1_i4b:k-1_i4b)
        list(l+1_i4b) = temp
!
        itemp = order(k)
        order(l+2_i4b:k) = order(l+1_i4b:k-1_i4b)
        order(l+1_i4b) = itemp
!
    end do
!
!   DESCENDING SORT.
!
    if ( present(ascending) ) then
!
        if ( .not.ascending ) then
            list = list(n:1_i4b:-1_i4b)
            order = order(n:1_i4b:-1_i4b)
        end if
!
    end if
!                                                                              
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        recursive subroutine quick_sort_1( left_end, right_end )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: left_end, right_end
!
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   ----------------------------------
!
        integer(i4b) :: left, right, i, j
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        left = left_end
        right = right_end
!
        do
!
            if ( right<left+mini ) exit
!
!           CHOOSE MEDIAN OF LEFT, CENTER AND RIGHT ELEMENTS 
!           AS PARTITIONING ELEMENT reference.
!
            if ( list(right)<list(left) ) then
                i = right ; j = left
            else
                i = left ; j =right
            end if
!
            reference = list( (left + right)/2_i4b )
!
            if ( reference<list(j) ) then
                if ( reference<list(i) ) reference = list(i)
            else
                reference = list(j)
            end if
!
            i = left - 1_i4b
            j = right + 1_i4b
!
            do
!
!               SCAN list FROM LEFT END UNTIL ELEMENT >= reference IS FOUND.
!
                do
                    i = i + 1_i4b
                    if ( list(i)>=reference ) exit
                end do
!
!               SCAN list FROM RIGHT END UNTIL ELEMENT <= reference IS FOUND.
!
                do
                    j = j - 1_i4b
                    if ( list(j)<=reference ) exit
                end do
!
                if ( i<j ) then
!
!                   SWAP TWO OUT-OF-ORDER ELEMENTS.
!
                    temp = list(i); list(i) = list(j); list(j) = temp
                    itemp = order(i); order(i) = order(j); order(j) = itemp
!
                else if ( i==j ) then
!
                    i = i + 1_i4b
!
                    exit
!
                else
!
                    exit
!
                end if
!
            end do
!
!           SORT THE TWO PARTITIONS.
!
            if ( j-left>right-i ) then
                call quick_sort_1( i, right )
                right = j
            else
                call quick_sort_1( left, j )
                left = i
            end if
!
        end do
!
!
!   END OF SUBROUTINE quick_sort_1
!   ------------------------------
!
        end subroutine quick_sort_1
!
! ___________________________________________________________________________________________

!
!
! END OF SUBROUTINE quick_sort_rv_ord
! ___________________________________
!
    end subroutine quick_sort_rv_ord
!
! =========================================================================================
!
    subroutine index_iv( list, index )
!
! Purpose
! _______
!
!   This subroutine indexes an integer array LIST, i.e., outputs the array index
!   of length N such that LIST(INDEX(j)) is in ascending order for j=1,
!   2, ..., N. The input quantity LIST is not changed.
!
!
! Arguments
! _________
!
!   LIST    (INPUT) integer(i4b), dimension(:)
!           The integer vector to index.
!
!   INDEX   (OUTPUT) integer(i4b), dimension(:)
!           The index array.
!
!
! Further Details
! _______________
!
!   The size of LIST and INDEX must match.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: list(:)
    integer(i4b), intent(out) :: index(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, k, l, temp, reference
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( list )
    l = size( index )
!
    n = assert_eq( k, l, 'do_index' )
!
    index = arth( 1_i4b, 1_i4b, n )
!
!   USE PARTITION ("QUICK") SORT.
!
    call quick_sort_1( 1_i4b, n )
!
!   USE INTERCHANGE SORT FOR SMALL LISTS.
!
    do k = 2_i4b, n
!
        temp = index(k)
        reference = list(temp)
!
        do l = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(index(l))<=reference ) exit
!
        end do
!
        index(l+2_i4b:k) = index(l+1_i4b:k-1_i4b)
        index(l+1_i4b) = temp
!
    end do
!                                                                              
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        recursive subroutine quick_sort_1( left_end, right_end )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: left_end, right_end
!
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   ----------------------------------
!
        integer(i4b)      :: left, right, i, j
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        left = left_end
        right = right_end
!
        do
!
            if ( right<left+mini ) exit
!
!           CHOOSE MEDIAN OF LEFT, CENTER AND RIGHT ELEMENTS 
!           AS PARTITIONING ELEMENT reference.
!
            i = index(left) ; j = index(right)
!
            if ( list(j)<list(i) ) then
                temp = i ; i = j ; j = temp
            end if
!
            reference = list(index((left + right)/2_i4b))
!
            if ( reference<list(j) ) then
                if ( reference<list(i) ) reference = list(i)
            else
                reference = list(j)
            end if
!
            i = left - 1_i4b
            j = right + 1_i4b
!
            do
!
!               SCAN list FROM LEFT END UNTIL ELEMENT >= reference IS FOUND.
!
                do
                    i = i + 1_i4b
                    if ( list(index(i))>=reference ) exit
                end do
!
!               SCAN list FROM RIGHT END UNTIL ELEMENT <= reference IS FOUND.
!
                do
                    j = j - 1_i4b
                    if ( list(index(j))<=reference ) exit
                end do
!
                if ( i<j ) then
!
!                   SWAP TWO OUT-OF-ORDER ELEMENTS.
!
                    temp = index(i); index(i) = index(j); index(j) = temp
!
                else if ( i==j ) then
!
                    i = i + 1_i4b
!
                    exit
!
                else
!
                    exit
!
                end if
!
            end do
!
!           SORT THE TWO PARTITIONS.
!
            if ( j-left>right-i ) then
                call quick_sort_1( i, right )
                right = j
            else
                call quick_sort_1( left, j )
                left = i
            end if
!
        end do
!
!
!   END OF SUBROUTINE quick_sort_1
!   ------------------------------
!
        end subroutine quick_sort_1
!
! ___________________________________________________________________________________________
!
!
! END OF SUBROUTINE index_iv
! __________________________
!
    end subroutine index_iv
!
! ===========================================================================================
!
    subroutine index_rv( list, index )
!
! Purpose
! _______
!
!   This subroutine indexes a real array LIST, i.e., outputs the array index
!   of length N such that LIST(INDEX(j)) is in ascending order for j=1,
!   2, ..., N. The input quantity LIST is not changed.
!
!
! Arguments
! _________
!
!   LIST    (INPUT) real(stnd), dimension(:)
!           The real vector to index.
!
!   INDEX   (OUTPUT) integer(i4b), dimension(:)
!           The index array.
!
!
! Further Details
! _______________
!
!   The size of LIST and INDEX must match.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)  :: list(:)
!
    integer(i4b), intent(out) :: index(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: reference
!
    integer(i4b) :: n, k, l, itemp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( list )
    l = size( index )
!
    n = assert_eq( k, l, 'do_index' )
!
    index = arth( 1_i4b, 1_i4b, n )
!
!   USE PARTITION ("QUICK") SORT.
!
    call quick_sort_1( 1_i4b, n )
!
!   USE INTERCHANGE SORT FOR SMALL LISTS.
!
    do k = 2_i4b, n
!
        itemp = index(k)
        reference = list(itemp)
!
        do l = k-1_i4b, 1_i4b, -1_i4b
!
            if ( list(index(l))<=reference ) exit
!
        end do
!
        index(l+2_i4b:k) = index(l+1_i4b:k-1_i4b)
        index(l+1_i4b) = itemp
!
    end do
!                                                                              
! ___________________________________________________________________________________________
!
                             contains
!                            ________
!
! ___________________________________________________________________________________________
!
        recursive subroutine quick_sort_1( left_end, right_end )
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ----------------------------
!
        integer(i4b), intent(in) :: left_end, right_end
!
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   ----------------------------------
!
        integer(i4b) :: left, right, i, j
!
!
!   EXECUTABLE STATEMENTS
!   ---------------------
!
!
        left = left_end
        right = right_end
!
        do
            if ( right<left+mini ) exit
!
!           CHOOSE MEDIAN OF LEFT, CENTER AND RIGHT ELEMENTS 
!           AS PARTITIONING ELEMENT reference.
!
            i = index(left) ; j = index(right)
!
            if ( list(j)<list(i) ) then
                itemp = i ; i = j ; j = itemp
            end if
!
            reference = list(index((left + right)/2_i4b))
!
            if ( reference<list(j) ) then
                if ( reference<list(i) ) reference = list(i)
            else
                reference = list(j)
            end if
!
            i = left - 1_i4b
            j = right + 1_i4b
!
            do
!
!               SCAN list FROM LEFT END UNTIL ELEMENT >= reference IS FOUND.
!
                do
                    i = i + 1_i4b
                    if ( list(index(i))>=reference ) exit
                end do
!
!               SCAN list FROM RIGHT END UNTIL ELEMENT <= reference IS FOUND.
!
                do
                    j = j - 1_i4b
                    if ( list(index(j))<=reference ) exit
                end do
!
                if ( i<j ) then
!
!                   SWAP TWO OUT-OF-ORDER ELEMENTS.
!
                    itemp = index(i); index(i) = index(j); index(j) = itemp
!
                else if ( i==j ) then
!
                    i = i + 1_i4b
!
                    exit
!
                else
!
                    exit
!
                end if
!
            end do
!
!           SORT THE TWO PARTITIONS.
!
            if ( j-left>right-i ) then
                call quick_sort_1( i, right )
                right = j
             else
                call quick_sort_1( left, j )
                left = i
            end if
!
        end do
!
!
!   END OF SUBROUTINE quick_sort_1
!   ------------------------------
!
        end subroutine quick_sort_1
!
! ___________________________________________________________________________________________
!
!
! END OF SUBROUTINE index_rv
! __________________________
!
    end subroutine index_rv
!
! ===========================================================================================
!
    function rank( index )
!
! Purpose
! _______
!
!   Given INDEX as output from the routine DO_INDEX, this routine returns
!   a same-size array RANK, the corresponding table of ranks.
!
!
! Arguments
! _________
!
!   INDEX   (INPUT) integer(i4b), dimension(:)
!           The index.
!
!
! Further Details
! _______________
!
!   This function is adapted from Numerical Recipes.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: index(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b), dimension(size(index)) :: rank
    integer(i4b)  :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( index )
!
    if ( n/=0_i4b ) rank(index(:)) = arth( 1_i4b, 1_i4b, n )
!
!
! END OF FUNCTION rank
! ____________________
!
   end function rank
!
! =========================================================================================
!
    subroutine reorder_iv( index, slave, ascending )
!
! Purpose
! _______
!
!   Given INDEX as output from the routine DO_INDEX, this routine makes
!   the corresponding rearrangement of the same-size integer array SLAVE.
!   The rearrangement is performed by means of the index array.
!
!
! Arguments
! _________
!
!   INDEX           (INPUT) integer(i4b), dimension(:)
!                   The index vector.
!
!   SLAVE           (INPUT/OUTPUT) integer(i4b), dimension(:)
!                   Integer vector to rearrange according to INDEX.
!
!   ASCENDING       (INPUT, OPTIONAL) logical(lgl)
!                   Rearrange SLAVE according to ascending order if ASCENDING = true,
!                   or to descending order if ASCENDING = false. The default is true.
!
!
! Further Details
! _______________
!
!   The size of SLAVE and INDEX must match.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Utilities,         only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)    :: index(:)
    integer(i4b), intent(inout) :: slave(:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)                         :: k, l, n
    integer(i4b), dimension(size(slave)) :: tmp
!
    logical(lgl) :: test
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( slave )
    l = size( index )
!
    n = assert_eq( k, l, 'reorder' )
!
    test = false
    if ( present(ascending) ) test = .not. ascending
!
!   REARRANGE slave IN ASCENDING ORDER.
!
    tmp(:n) = slave(index)
!
    if ( test ) then
!
!       REARRANGE slave IN DESCENDING ORDER.
!
        slave(:n) = tmp(n:1_i4b:-1_i4b)
!
    else
!
!       REARRANGE slave IN ASCENDING ORDER.
!
        slave(:n) = tmp(:n)
!
    end if
!
!
! END OF SUBROUTINE reorder_iv
! ____________________________
!
    end subroutine reorder_iv
!
! ===========================================================================================
!
    subroutine reorder_rv( index, slave, ascending )
!
! Purpose
! _______
!
!   Given INDEX as output from the routine DO_INDEX, this routine makes
!   the corresponding rearrangement of the same-size real array SLAVE.
!   The rearrangement is performed by means of the index array.
!
!
! Arguments
! _________
!
!   INDEX           (INPUT) integer(i4b), dimension(:)
!                   The index vector.
!
!   SLAVE           (INPUT/OUTPUT) real(stnd), dimension(:)
!                   Real vector to rearrange according to INDEX.
!
!   ASCENDING       (INPUT, OPTIONAL) logical(lgl)
!                   Rearrange SLAVE according to ascending order if ASCENDING = true,
!                   or to descending order if ASCENDING = false. The default is true.
!
!
! Further Details
! _______________
!
!   The size of SLAVE and INDEX must match.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Utilities,         only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)    :: index(:)
!
    real(stnd),   intent(inout) :: slave(:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(slave)) :: tmp
!
    integer(i4b) :: k, l, n
!
    logical(lgl) :: test
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( slave )
    l = size( index )
!
    n = assert_eq( k, l, 'reorder' )
!
    test = false
    if ( present(ascending) ) test = .not. ascending
!
!   REARRANGE slave IN ASCENDING ORDER.
!
    tmp(:n) = slave(index)
!
    if ( test ) then
!
!       REARRANGE slave IN DESCENDING ORDER.
!
        slave(:n) = tmp(n:1_i4b:-1_i4b)
!
    else
!
!       REARRANGE slave IN ASCENDING ORDER.
!
        slave(:n) = tmp(:n)
!
    end if
!
!
! END OF SUBROUTINE reorder_rv
! ____________________________
!
    end subroutine reorder_rv
!
! ===========================================================================================
!
    subroutine reorder_cv( index, slave, ascending )
!
! Purpose
! _______
!
!   Given INDEX as output from the routine DO_INDEX, this routine makes
!   the corresponding rearrangement of the same-size complex array SLAVE.
!   The rearrangement is performed by means of the index array.
!
!
! Arguments
! _________
!
!   INDEX           (INPUT) integer(i4b), dimension(:)
!                   The index vector.
!
!   SLAVE           (INPUT/OUTPUT) complex(stnd), dimension(:)
!                   Complex vector to rearrange according to INDEX.
!
!   ASCENDING       (INPUT, OPTIONAL) logical(lgl)
!                   Rearrange SLAVE according to ascending order if ASCENDING = true,
!                   or to descending order if ASCENDING = false. The default is true.
!
!
! Further Details
! _______________
!
!   The size of SLAVE and INDEX must match.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Utilities,         only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)       :: index(:)
!
    complex(stnd),   intent(inout) :: slave(:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(size(slave)) :: tmp
!
    integer(i4b) :: k, l, n
!
    logical(lgl) :: test
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( slave )
    l = size( index )
!
    n = assert_eq( k, l, 'reorder' )
!
    test = false
    if ( present(ascending) ) test = .not. ascending
!
!   REARRANGE slave IN ASCENDING ORDER.
!
    tmp(:n) = slave(index)
!
    if ( test ) then
!
!       REARRANGE slave IN DESCENDING ORDER.
!
        slave(:n) = tmp(n:1_i4b:-1_i4b)
!
    else
!
!       REARRANGE slave IN ASCENDING ORDER.
!
        slave(:n) = tmp(:n)
!
    end if
!
!
! END OF SUBROUTINE reorder_cv
! ____________________________
!
    end subroutine reorder_cv
!
! ===========================================================================================
!
    subroutine reorder_im( index, slave, ascending )
!
! Purpose
! _______
!
!   Given INDEX as output from the routine DO_INDEX, this routine makes
!   the corresponding rearrangement of the columns of the integer matrice SLAVE.
!   The rearrangement is performed by means of the index array.
!
!
! Arguments
! _________
!
!   INDEX           (INPUT) integer(i4b), dimension(:)
!                   The index vector.
!
!   SLAVE           (INPUT/OUTPUT) integer(i4b), dimension(:,:)
!                   Integer matrix to rearrange according to index.
!
!   ASCENDING       (INPUT, OPTIONAL) logical(lgl)
!                   Rearrange the columns of SLAVE according to ascending order 
!                   if ASCENDING = true, or to descending order if ASCENDING = false.
!                   The default is true.
!
!
! Further Details
! _______________
!
!   The size of INDEX must match the number of columns of SLAVE.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Utilities,         only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)    :: index(:)
    integer(i4b), intent(inout) :: slave(:,:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)                                                  :: k, l, n
    integer(i4b), dimension(size(slave,dim=1),size(slave,dim=2 )) :: tmp
!
    logical(lgl) :: test
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( slave, dim=2 )
    l = size( index )
!
    n = assert_eq( k, l, 'reorder' )
!
    test = false
    if ( present(ascending) ) test = .not. ascending
!
!   REARRANGE slave IN ASCENDING ORDER.
!
    tmp(:,:n) = slave(:,index)
!
    if ( test ) then
!
!       REARRANGE slave IN DESCENDING ORDER.
!
        slave(:,:n) = tmp(:,n:1_i4b:-1_i4b)
!
    else
!
!       REARRANGE slave IN ASCENDING ORDER.
!
        slave(:,:n) = tmp(:,:n)
!
    end if
!
!
! END OF SUBROUTINE reorder_im
! ____________________________
!
    end subroutine reorder_im
!
! ===========================================================================================
!
    subroutine reorder_rm( index, slave, ascending )
!
! Purpose
! _______
!
!   Given INDEX as output from the routine DO_INDEX, this routine makes
!   the corresponding rearrangement of the columns of the real matrice SLAVE.
!   The rearrangement is performed by means of the index array.
!
!
! Arguments
! _________
!
!   INDEX           (INPUT) integer(i4b), dimension(:)
!                   The index vector.
!
!   SLAVE           (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                   Real matrix to rearrange according to index.
!
!   ASCENDING       (INPUT, OPTIONAL) logical(lgl)
!                   Rearrange the columns of SLAVE according to ascending order 
!                   if ASCENDING = true, or to descending order if ASCENDING = false.
!                   The default is true.
!
!
! Further Details
! _______________
!
!   The size of INDEX must match the number of columns of SLAVE.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Utilities,         only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: index(:)
!
    real(stnd),   intent(inout) :: slave(:,:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(slave,dim=1),size(slave,dim=2 )) :: tmp
!
    integer(i4b) :: k, l, n
!
    logical(lgl) :: test
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( slave, dim=2 )
    l = size( index )
!
    n = assert_eq( k, l, 'reorder' )
!
    test = false
    if ( present(ascending) ) test = .not. ascending
!
!   REARRANGE slave IN ASCENDING ORDER.
!
    tmp(:,:n) = slave(:,index)
!
    if ( test ) then
!
!       REARRANGE slave IN DESCENDING ORDER.
!
        slave(:,:n) = tmp(:,n:1_i4b:-1_i4b)
!
    else
!
!       REARRANGE slave IN ASCENDING ORDER.
!
        slave(:,:n) = tmp(:,:n)
!
    end if
!
!
! END OF SUBROUTINE reorder_rm
! ____________________________
!
    end subroutine reorder_rm
!
! ===========================================================================================
!
    subroutine reorder_cm( index, slave, ascending )
!
! Purpose
! _______
!
!   Given INDEX as output from the routine DO_INDEX, this routine makes
!   the corresponding rearrangement of the columns of the complex matrice SLAVE.
!   The rearrangement is performed by means of the index array.
!
!
! Arguments
! _________
!
!   INDEX           (INPUT) integer(i4b), dimension(:)
!                   The index vector.
!
!   SLAVE           (INPUT/OUTPUT) complex(stnd), dimension(:,:)
!                   Complex matrix to rearrange according to index.
!
!   ASCENDING       (INPUT, OPTIONAL) logical(lgl)
!                   Rearrange the columns of SLAVE according to ascending order 
!                   if ASCENDING = true, or to descending order if ASCENDING = false.
!                   The default is true.
!
!
! Further Details
! _______________
!
!   The size of INDEX must match the number of columns of SLAVE.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Utilities,         only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: index(:)
!
    complex(stnd),   intent(inout) :: slave(:,:)
!
    logical(lgl), intent(in), optional   :: ascending
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(size(slave,dim=1),size(slave,dim=2 )) :: tmp
!
    integer(i4b) :: k, l, n
!
    logical(lgl) :: test
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    k = size( slave, dim=2 )
    l = size( index )
!
    n = assert_eq( k, l, 'reorder' )
!
    test = false
    if ( present(ascending) ) test = .not. ascending
!
!   REARRANGE slave IN ASCENDING ORDER.
!
    tmp(:,:n) = slave(:,index)
!
    if ( test ) then
!
!       REARRANGE slave IN DESCENDING ORDER.
!
        slave(:,:n) = tmp(:,n:1_i4b:-1_i4b)
!
    else
!
!       REARRANGE slave IN ASCENDING ORDER.
!
        slave(:,:n) = tmp(:,:n)
!
    end if
!
!
! END OF SUBROUTINE reorder_cm
! ____________________________
!
    end subroutine reorder_cm
!
! ===========================================================================================
!
!
! *******************************
! END OF MODULE Sort_Procedures *
! *******************************
!    
end module Sort_Procedures
