module Utilities_With_Pnter
! ####################################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING UTILITIES TO MANIPULATE FORTRAN90 POINTERS.                       *
!                                                                                    * 
! THESE ROUTINES ARE ADAPTED AND EXTENDED FROM PUBLIC DOMAIN ROUTINES FROM           *
! Numerical Recipes.                                                                 *
!                                                                                    * 
! LATEST REVISION : 21/03/2018                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!
!
! USED MODULES 
! ============
!
    use Select_Parameters, only : i4b, stnd
!
! OTHER MODULES USED IN SPECIFIC SUBROUTINES
! ==========================================
!
!   use Char_Constants,   only :
!   use Utilities,        only :
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public :: reallocate, realloc
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface reallocate
        module procedure     reallocate_iv, reallocate_im,   &
                             reallocate_rv, reallocate_rm,   &
                             reallocate_cv, reallocate_cm,   &
                             reallocate_hv
    end interface
!
    interface realloc
        module procedure     realloc_iv, realloc_im,   &
                             realloc_rv, realloc_rm,   &
                             realloc_cv, realloc_cm,   &
                             realloc_hv
    end interface
!    
!    
! MODULE VARIABLES 
! ================
!
!
! MODULE PARAMETERS 
! =================
!
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
!
! =========================================================================================
!
! ROUTINES THAT REALLOCATE A POINTER TO NEW SIZE, WHILE PRESERVING ITS PREVIOUS CONTENTS.
! ADAPTED AND EXTENDED FROM Numerical Recipes.
!
! =========================================================================================
!
    function reallocate_iv( p, n )
!
! Purpose
! _______
!
!   Reallocates a pointer P to an integer one dimensional array with a new size N,
!   while preserving its contents. The pointer P is deallocated on return.
!
!
! Arguments
! _________
!
!   P      integer(i4b), dimension(:), pointer
!          On entry, an allocated pointer to an integer vector.
!
!          On exit, the pointer is deallocated.
!
!   N      (INPUT) integer(i4b)
!          The size N of the new pointer.
!
!
! __________________________________________________________________________________________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : merror
!
    integer(i4b), dimension(:), pointer :: p, reallocate_iv
    integer(i4b), intent(in) :: n
!
    integer(i4b) :: nold, i
    integer      :: iok
!
    allocate( reallocate_iv(n), stat=iok )
!
    if ( iok/=0 ) call merror( 'reallocate'//allocate_error )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( .not.associated( p ) ) return
!
    nold = size( p )
    i = min( nold, n )
!
    reallocate_iv(1_i4b:i) = p(1_i4b:i)
!
    deallocate( p )
!
    end function reallocate_iv
!
! =========================================================================================
!
    function reallocate_rv( p, n )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a real one dimensional array with a new size N,
!   while preserving its contents. The pointer P is deallocated on return.
!
!
! Arguments
! _________
!
!   P      real(stnd), dimension(:), pointer
!          On entry, an allocated pointer to a real vector.
!
!          On exit, the pointer is deallocated.
!
!   N      (INPUT) integer(i4b)
!          The size N of the new pointer.
!
!
! __________________________________________________________________________________________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : merror
!
    real(stnd), dimension(:), pointer :: p, reallocate_rv
    integer(i4b), intent(in) :: n
!
    integer(i4b) :: nold, i
    integer      :: iok
!
    allocate( reallocate_rv(n), stat=iok )
!
    if ( iok/=0 ) call merror( 'reallocate'//allocate_error )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( .not.associated( p ) ) return
!
    nold = size( p )
    i = min( nold, n )
!
    reallocate_rv(1_i4b:i) = p(1_i4b:i)
!
    deallocate( p )
!
    end function reallocate_rv
!
! =========================================================================================
!
    function reallocate_cv( p, n )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a complex one dimensional array with a new size N,
!   while preserving its contents. The pointer P is deallocated on return.
!
!
! Arguments
! _________
!
!   P      complex(stnd), dimension(:), pointer
!          On entry, an allocated pointer to a complex vector.
!
!          On exit, the pointer is deallocated.
!
!   N      (INPUT) integer(i4b)
!          The size N of the new pointer.
!
!
! __________________________________________________________________________________________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : merror
!
    complex(stnd), dimension(:), pointer :: p, reallocate_cv
    integer(i4b), intent(in) :: n
!
    integer(i4b) :: nold, i
    integer      :: iok
!
    allocate( reallocate_cv(n), stat=iok )
!
    if ( iok/=0 ) call merror( 'reallocate'//allocate_error )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( .not.associated( p ) ) return
!
    nold = size( p )
    i = min( nold, n )
!
    reallocate_cv(1_i4b:i) = p(1_i4b:i)
!
    deallocate( p )
!
    end function reallocate_cv
!
! =========================================================================================
!
    function reallocate_hv( p, n )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a character one dimensional array with a new size N,
!   while preserving its contents. The pointer P is deallocated on return.
!
!
! Arguments
! _________
!
!   P      character(1), dimension(:), pointer
!          On entry, an allocated pointer to a character vector.
!
!          On exit, the pointer is deallocated.
!
!   N      (INPUT) integer(i4b)
!          The size N of the new pointer.
!
!
! __________________________________________________________________________________________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : merror
!
    character(1), dimension(:), pointer :: p, reallocate_hv
    integer(i4b), intent(in) :: n
!
    integer(i4b) :: nold, i
    integer      :: iok
!
    allocate( reallocate_hv(n), stat=iok )
!
    if ( iok/=0 ) call merror( 'reallocate'//allocate_error )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( .not.associated( p ) ) return
!
    nold = size( p )
    i = min( nold, n )
!
    reallocate_hv(1_i4b:i) = p(1_i4b:i)
!
    deallocate( p )
!
    end function reallocate_hv
!
! =========================================================================================
!
    function reallocate_im( p, n, m )
!
! Purpose
! _______
!
!   Reallocates a pointer P to an integer two dimensional array with a new shape (N,M)
!   while preserving its contents. The pointer P is deallocated on return.
!
!
! Arguments
! _________
!
!   P      integer(i4b), dimension(:,:), pointer
!          On entry, an allocated pointer to an integer matrix.
!
!          On exit, the pointer is deallocated.
!
!   N, M   (INPUT) integer(i4b)
!          The shape (N,M) of the new pointer.
!
!
! __________________________________________________________________________________________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : merror
!
    integer(i4b), dimension(:,:), pointer :: p, reallocate_im
    integer(i4b), intent(in) :: n, m
!
    integer(i4b) :: old, i, j
    integer      :: iok
!
    allocate( reallocate_im(n,m), stat=iok )
!
    if ( iok/=0 ) call merror( 'reallocate'//allocate_error )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( .not.associated( p ) ) return
!
    old = size( p, 1 )
    i = min( old, n )
    old = size( p, 2 )
    j = min( old, m )
!
    reallocate_im(1_i4b:i,1_i4b:j) = p(1_i4b:i,1_i4b:j)
!
    deallocate( p )
!
    end function reallocate_im
!
! =========================================================================================
!
    function reallocate_rm( p, n, m )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a real two dimensional array with a new shape (N,M)
!   while preserving its contents. The pointer P is deallocated on return.
!
!
! Arguments
! _________
!
!   P      real(stnd), dimension(:,:), pointer
!          On entry, an allocated pointer to a real matrix.
!
!          On exit, the pointer is deallocated.
!
!   N, M   (INPUT) integer(i4b)
!          The shape (N,M) of the new pointer.
!
!
! __________________________________________________________________________________________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : merror
!
    real(stnd), dimension(:,:), pointer :: p, reallocate_rm
    integer(i4b), intent(in) :: n, m
!
    integer(i4b) :: old, i, j
    integer      :: iok
!
    allocate( reallocate_rm(n,m), stat=iok )
!
    if ( iok/=0 ) call merror( 'reallocate'//allocate_error )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( .not.associated( p ) ) return
!
    old = size( p, 1 )
    i = min( old, n )
    old = size( p, 2 )
    j = min( old, m )
!
    reallocate_rm(1_i4b:i,1_i4b:j) = p(1_i4b:i,1_i4b:j)
!
    deallocate( p )
!
    end function reallocate_rm
!
! =========================================================================================
!
    function reallocate_cm( p, n, m )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a complex two dimensional array with a new shape (N,M)
!   while preserving its contents. The pointer P is deallocated on return.
!
!
! Arguments
! _________
!
!   P      complex(stnd), dimension(:,:), pointer
!          On entry, an allocated pointer to a complex matrix.
!
!          On exit, the pointer is deallocated.
!
!   N, M   (INPUT) integer(i4b)
!          The shape (N,M) of the new pointer.
!
!
! __________________________________________________________________________________________
!
    use Char_Constants,   only : allocate_error
    use Utilities,        only : merror
!
    complex(stnd), dimension(:,:), pointer :: p, reallocate_cm
    integer(i4b), intent(in) :: n, m
!
    integer(i4b) :: old, i, j
    integer      :: iok
!
    allocate( reallocate_cm(n,m), stat=iok )
!
    if ( iok/=0 ) call merror( 'reallocate'//allocate_error )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( .not.associated( p ) ) return
!
    old = size( p, 1 )
    i = min( old, n )
    old = size( p, 2 )
    j = min( old, m )
!
    reallocate_cm(1_i4b:i,1_i4b:j) = p(1_i4b:i,1_i4b:j)
!
    deallocate( p )
!
    end function reallocate_cm
!
! =========================================================================================
!
    subroutine realloc_iv( p, n, ialloc )
!
! Purpose
! _______
!
!   Reallocates a pointer P to an integer one dimensional array with a new size N,
!   while preserving its contents.
!
!
! Arguments
! _________
!
!   P      integer(i4b), dimension(:), pointer
!          On entry, an allocated pointer to an integer vector.
!
!          On exit, the allocated pointer with a new size of N.
!
!   N      (INPUT) integer(i4b)
!          The new size N of the pointer.
!
!   IALLOC (OUTPUT) integer
!          On exit, IALLOC = 0 indicates successful exit.
!          Any other values indicate an allocation problem.
!
!
! __________________________________________________________________________________________
!
    integer(i4b), dimension(:), pointer :: p
    integer(i4b), intent(in)  :: n
    integer,      intent(out) :: ialloc
!
    integer(i4b), dimension(n) :: tmp
    integer(i4b)               :: nold, i
!
    if ( .not.associated( p ) ) then
!
        allocate( p(n), stat=ialloc )
!
    else
!
        nold = size( p )
        i = min( nold, n )
        tmp(:i) = p(:i)
!
        deallocate( p )
        allocate( p(n), stat=ialloc )
!
        if ( ialloc==0 ) p(:i) = tmp(:i)
!
    end if
!
    end subroutine realloc_iv
!
! =========================================================================================
!
    subroutine realloc_rv( p, n, ialloc )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a real one dimensional array with a new size N,
!   while preserving its contents.
!
!
! Arguments
! _________
!
!   P      real(stnd), dimension(:), pointer
!          On entry, an allocated pointer to a real vector.
!
!          On exit, the allocated pointer with a new size of N.
!
!   N      (INPUT) integer(i4b)
!          The new size N of the pointer.
!
!   IALLOC (OUTPUT) integer
!          On exit, IALLOC = 0 indicates successful exit.
!          Any other values indicate an allocation problem.
!
!
! __________________________________________________________________________________________
!
    real(stnd), dimension(:), pointer :: p
    integer(i4b), intent(in)  :: n
    integer,      intent(out) :: ialloc
!
    real(stnd), dimension(n)   :: tmp
    integer(i4b)               :: nold, i
!
    if ( .not.associated( p ) ) then
!
        allocate( p(n), stat=ialloc )
!
    else
!
        nold = size( p )
        i = min( nold, n )
        tmp(:i) = p(:i)
!
        deallocate( p )
        allocate( p(n), stat=ialloc )
!
        if ( ialloc==0 ) p(:i) = tmp(:i)
!
    end if
!
    end subroutine realloc_rv
!
! =========================================================================================
!
    subroutine realloc_cv( p, n, ialloc )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a complex one dimensional array with a new size N,
!   while preserving its contents.
!
!
! Arguments
! _________
!
!   P      complex(stnd), dimension(:), pointer
!          On entry, an allocated pointer to a complex vector.
!
!          On exit, the allocated pointer with a new size of N.
!
!   N      (INPUT) integer(i4b)
!          The new size N of the pointer.
!
!   IALLOC (OUTPUT) integer
!          On exit, IALLOC = 0 indicates successful exit.
!          Any other values indicate an allocation problem.
!
!
! __________________________________________________________________________________________
!
    complex(stnd), dimension(:), pointer :: p
    integer(i4b), intent(in)  :: n
    integer,      intent(out) :: ialloc
!
    complex(stnd), dimension(n) :: tmp
    integer(i4b)                :: nold, i
!
    if ( .not.associated( p ) ) then
!
        allocate( p(n), stat=ialloc )
!
    else
!
        nold = size( p )
        i = min( nold, n )
        tmp(:i) = p(:i)
!
        deallocate( p )
        allocate( p(n), stat=ialloc )
!
        if ( ialloc==0 ) p(:i) = tmp(:i)
!
    end if
!
    end subroutine realloc_cv
!
! =========================================================================================
!
    subroutine realloc_hv( p, n, ialloc )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a character one dimensional array with a new size N,
!   while preserving its contents.
!
!
! Arguments
! _________
!
!   P      character(1), dimension(:), pointer
!          On entry, an allocated pointer to a character vector.
!
!          On exit, the allocated pointer with a new size of N.
!
!   N      (INPUT) integer(i4b)
!          The new size N of the pointer.
!
!   IALLOC (OUTPUT) integer
!          On exit, IALLOC = 0 indicates successful exit.
!          Any other values indicate an allocation problem.
!
!
! __________________________________________________________________________________________
!
    character(1), dimension(:), pointer :: p
    integer(i4b), intent(in)  :: n
    integer,      intent(out) :: ialloc
!
    character(1), dimension(n) :: tmp
    integer(i4b)               :: nold, i
!
    if ( .not.associated( p ) ) then
!
        allocate( p(n), stat=ialloc )
!
    else
!
        nold = size( p )
        i = min( nold, n )
        tmp(:i) = p(:i)
!
        deallocate( p )
        allocate( p(n), stat=ialloc )
!
        if ( ialloc==0 ) p(:i) = tmp(:i)
!
    end if
!
    end subroutine realloc_hv
!
! =========================================================================================
!
    subroutine realloc_im( p, n, m, ialloc )
!
! Purpose
! _______
!
!   Reallocates a pointer P to an integer two dimensional array with a new shape (N,M)
!   while preserving its contents.
!
!
! Arguments
! _________
!
!   P      integer(i4b), dimension(:,:), pointer
!          On entry, an allocated pointer to an integer matrix.
!
!          On exit, the allocated pointer with a new shape (N,M).
!
!   N, M   (INPUT) integer(i4b)
!          The new shape (N,M) of the pointer.
!
!   IALLOC (OUTPUT) integer
!          On exit, IALLOC = 0 indicates successful exit.
!          Any other values indicate an allocation problem.
!
!
! __________________________________________________________________________________________
!
    integer(i4b), dimension(:,:), pointer :: p
    integer(i4b), intent(in)  :: n, m
    integer,      intent(out) :: ialloc
!
    integer(i4b), dimension(n,m) :: tmp
    integer(i4b)                 :: old, i, j
!
    if ( .not.associated( p ) ) then
!
        allocate( p(n,m), stat=ialloc )
!
    else
!
        old = size( p, 1 )
        i = min( old, n )
        old = size( p, 2 )
        j = min( old, m )
        tmp(1_i4b:i,1_i4b:j) = p(1_i4b:i,1_i4b:j)
!
        deallocate( p )
        allocate( p(n,m), stat=ialloc )
!
        if ( ialloc==0 ) p(1_i4b:i,1_i4b:j) = tmp(1_i4b:i,1_i4b:j)
!
    end if
!
    end subroutine realloc_im
!
! =========================================================================================
!
    subroutine realloc_rm( p, n, m, ialloc )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a real two dimensional array with a new shape (N,M)
!   while preserving its contents.
!
!
! Arguments
! _________
!
!   P      real(stnd), dimension(:,:), pointer
!          On entry, an allocated pointer to a real matrix.
!
!          On exit, the allocated pointer with a new shape (N,M).
!
!   N, M   (INPUT) integer(i4b)
!          The new shape (N,M) of the pointer.
!
!   IALLOC (OUTPUT) integer
!          On exit, IALLOC = 0 indicates successful exit.
!          Any other values indicate an allocation problem.
!
!
! __________________________________________________________________________________________
!
    real(stnd),   dimension(:,:), pointer :: p
    integer(i4b), intent(in)  :: n, m
    integer,      intent(out) :: ialloc
!
    real(stnd), dimension(n,m) :: tmp
    integer(i4b)               :: old, i, j
!
    if ( .not.associated( p ) ) then
!
        allocate( p(n,m), stat=ialloc )
!
    else
!
        old = size( p, 1 )
        i = min( old, n )
        old = size( p, 2 )
        j = min( old, m )
        tmp(1_i4b:i,1_i4b:j) = p(1_i4b:i,1_i4b:j)
!
        deallocate( p )
        allocate( p(n,m), stat=ialloc )
!
        if ( ialloc==0 ) p(1_i4b:i,1_i4b:j) = tmp(1_i4b:i,1_i4b:j)
!
    end if
!
    end subroutine realloc_rm
!
! =========================================================================================
!
    subroutine realloc_cm( p, n, m, ialloc )
!
! Purpose
! _______
!
!   Reallocates a pointer P to a complex two dimensional array with a new shape (N,M)
!   while preserving its contents.
!
!
! Arguments
! _________
!
!   P      complex(stnd), dimension(:,:), pointer
!          On entry, an allocated pointer to a complex matrix.
!
!          On exit, the allocated pointer with a new shape (N,M).
!
!   N, M   (INPUT) integer(i4b)
!          The new shape (N,M) of the pointer.
!
!   IALLOC (OUTPUT) integer
!          On exit, IALLOC = 0 indicates successful exit.
!          Any other values indicate an allocation problem.
!
!
! __________________________________________________________________________________________
!
    complex(stnd), dimension(:,:), pointer :: p
    integer(i4b), intent(in)  :: n, m
    integer,      intent(out) :: ialloc
!
    complex(stnd), dimension(n,m) :: tmp
    integer(i4b)                  :: old, i, j
!
    if ( .not.associated( p ) ) then
!
        allocate( p(n,m), stat=ialloc )
!
    else
!
        old = size( p, 1 )
        i = min( old, n )
        old = size( p, 2 )
        j = min( old, m )
        tmp(1_i4b:i,1_i4b:j) = p(1_i4b:i,1_i4b:j)
!
        deallocate( p )
        allocate( p(n,m), stat=ialloc )
!
        if ( ialloc==0 ) p(1_i4b:i,1_i4b:j) = tmp(1_i4b:i,1_i4b:j)
!
    end if
!
    end subroutine realloc_cm
!
!
! =========================================================================================
!
!
! ************************************
! END OF MODULE Utilities_With_Pnter *
! ************************************
!    
end module Utilities_With_Pnter
