program mach_char
!! #############################################################################
!!
!! Copyright 2016 IRD
!!  
!! This file is part of statpack.
!!  
!! statpack is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! statpack is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the statpack/doc directory.
!!
!! #############################################################################
!
!
! Purpose
! =======
!
!   This program is intended to determine the parameters of the floating-point
!   arithmetic system for a given real kind type.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!   The program as given here must be modified before compiling. In order to obtain 
!   machine-specific parameters for a given real kind type, it is necessary to comment 
!   out all but one of the following use statements, before compiling MACH_CHAR.
!
!   LATEST REVISION : 20/01/2017
!                                                                              
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!  
! ----------------------------------------------------------------------
! BY SIMPLY ENSURING THAT A LEADING '!' APPEARS ON ALL BUT EXACTLY ONE !
! OF THE FOLLOWING use STATEMENTS, AND THEN RECOMPILING THE PROGRAM,   !
! MACHINE-SPECIFIC PARAMETERS FOR ANY real KIND TYPE CAN BE OBTAINED.  !
! ----------------------------------------------------------------------
!
!   use The_Kinds, only : stnd=>sp
!                          
   use The_Kinds, only : stnd=>dp
!
!   use The_Kinds, only : stnd=>qp
!
!   use The_Kinds, only : stnd=>low
!
!   use The_Kinds, only : stnd=>normal
!
!   use The_Kinds, only : stnd=>extended
!   
!   
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!   
!   
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    integer      :: base, basedigits, irnd, iuflow, igrd, expepspos, expepsneg,      &
                    minexp, maxexp, minexpbase, maxexpbase, iexp, ifloat
!
    real(stnd)   :: epspos, epsneg, epsilpos, epsilneg, machtiny, machhuge, rndunit
!   
!
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   mach RETURNS ALL THE MACHINE-SPECIFIC PARAMETERS.
!
    call mach( base, basedigits, irnd, iuflow, igrd, iexp, ifloat, minexp,         &
               maxexp, minexpbase, maxexpbase, expepspos, expepsneg,               &
               epspos, epsneg, epsilpos, epsilneg, machtiny, machhuge, rndunit )
!
!   PRINT RESULTS.
!
    write( prtunit, 10 ) stnd, base, basedigits
!
    select case (irnd)
        case(0) ; write( prtunit, '(a,/)' ) '(3)  Rounding upward or downward or toward zero' 
        case(1) ; write( prtunit, '(a,/)' ) '(3)  Non-IEEE-form of rounding to nearest'
        case(2) ; write( prtunit, '(a,/)' ) '(3)  IEEE-style of rounding'
    end select
!
    if ( iuflow==0 ) then
        write( prtunit, '(a,/)' ) '(4)  Full underflow   '
    else
        write( prtunit, '(a,/)' ) '(4)  Partial underflow'
    end if
!
    write( prtunit, 20 ) rndunit, igrd, minexp, maxexp, minexpbase, maxexpbase,        &
                         expepspos, expepsneg, epspos, epsneg, epsilpos, epsilneg,     &
                         machtiny, machhuge
!
    if ( ifloat>=0 ) write( prtunit, 30 ) ifloat, iexp
!
!
! FORMATS 
! =======
!
    10  format(                                                              &
!
    // '  Machine-specific parameters for REAL KIND type ', i3,  /           &
       '========================================================'///         &
       '(1)  Floating-point radix, "base", =',i3,//                          &
       '(2)  Number of base-"base" digits in the mantissa =',i6,/  )
!
    20  format(                                                                               &
!
    '(5)  Unit roundoff =',g17.8,//                                                           &
    '(6)  Number of base-"base" guard digits used in multiplication =',i2,//                  &
    '(7)  Minimum exponent before (gradual) underflow  =',i8,//                               &
    '(8)  Smallest power of base that causes overflow  =',i8,//                               &
    '(9)  Exponent for the smallest power of "base" that allows full'/                        &
    '     significance in the significand  =',i8,//                                           &
    '(10) Exponent for the largest power of "base" before overflow  =',i8,//                  &
    '(11) Exponent for the smallest power of "base" whose sum with 1. is > 1. =',i8,//        &
    '(12) Exponent for the smallest power of "base" whose difference with 1. is < 1. =',i8,// &
    '(13) Smallest power of "base" whose sum with 1. > 1. =',g17.8,//                         &
    '(14) Smallest power of "base" whose difference with 1. < 1. =',g17.8,//                  &
    '(15) Smallest floating point number whose sum with 1. > 1. =',g17.8,//                   &
    '(16) Smallest floating point number whose difference with 1. < 1. =',g17.8,//            &
    '(17) Smallest positive normalized floating point number =',g17.8,//                      &
    '(18) Largest positive normalized floating point number =',g17.8,/    ) 
!
    30  format(                                                                               &
!
    '(19) Size in bits of a floating point number (guess) =',i4,//                            &
    '(20) Number of bits reserved for the representation of the exponent (guess) =',i4,//    ) 
!
! ===========================================================================================
!
                             contains
!                            ========
!
!
! ===========================================================================================
!
        subroutine mach( base, basedigits, irnd, iuflow, igrd, iexp, ifloat, minexp,       &
                         maxexp, minexpbase, maxexpbase, expepspos, expepsneg,             &
                         epspos, epsneg, epsilpos, epsilneg, machtiny, machhuge, rndunit )
!
!
!   Purpose
!   _______
!
!       This subroutine is intended to determine the parameters of the floating-point
!       arithmetic system specified below.
!                                                                              
!
!   Arguments
!   _________
!
!       BASE        (OUTPUT) integer
!                   The radix for the floating point representation.                                     
!
!       BASEDIGITS  (OUTPUT) integer         
!                   The number of base BASE digits in the floating-point significand.
!
!       IRND        (OUTPUT) integer
!                   A parameter indicating whether proper rounding or chopping 
!                   (rounding upward, downward, toward zero) occurs in addition: 
!
!                       0    if floating-point addition rounds upward, downward
!                            or toward zero;
!                       1    if floating-point addition rounds to nearest,
!                            but not in the IEEE style;
!                       2    if floating-point addition rounds in the IEEE style.
!
!       IUFLOW      (OUTPUT) integer
!                   A parameter indicating whether underflow is full or partial:
!
!                       0    if there is full underflow (flush to zero, etc);
!                       1    if there is partial underflow.
!
!       IGRD        (OUTPUT) integer
!                   The number of guard digits for multiplication with chopping
!                   arithmetic (IRND = 0). It is
!
!                       0    if floating-point arithmetic rounds, or if it
!                            chops and only  BASEDIGITS  digits participate
!                            in the post-normalization shift of the
!                            floating-point significand in multiplication;
!                       1    if floating-point arithmetic chops and more
!                            than  BASEDIGITS  digits participate in the
!                            post-normalization shift of the floating-point
!                            significand in multiplication.
!
!       IEXP        (OUTPUT) integer
!                   The number of bits dedicated to the representation            
!                   of the exponent of a floating point number if BASE is
!                   a power of two and -1 otherwise.
!
!       IFLOAT      (OUTPUT) integer
!                   The number of bits dedicated to the representation            
!                   of a floating point number if BASE is a power of two 
!                   and -1 otherwise.                    
!                                                                              
!       MINEXP      (OUTPUT) integer
!                   Minimum exponent before (gradual) underflow.         
!
!       MAXEXP      (OUTPUT) integer
!                   Largest exponent before overflow. 
!
!       MINEXPBASE  (OUTPUT) integer
!                   The largest in magnitude negative integer such that         
!                   float(base)**minexpbase is positive and normalized.  
!
!       MAXEXPBASE  (OUTPUT) integer
!                   The largest in magnitude positive integer such that 
!                   float(base)**maxexpbase is positive and normalized.  
!
!       EXPEPSPOS   (OUTPUT) integer
!                   The largest in magnitude negative integer such that
!                   1.0+float(base)**expepspos /= 1.         
!
!       EXPEPSNEG   (OUTPUT) integer         
!                   The largest in magnitude negative integer such that            
!                   1.0-float(base)**expepsneg /= 1.         
!
!       EPSPOS      (OUTPUT) real(stnd)         
!                   The smallest power of base whose sum with 1. is         
!                   greater than 1. That is, float(base)**expepspos.                    
!
!       EPSNEG      (OUTPUT) real(stnd)          
!                   The smallest power of base whose difference with 1.           
!                   is less than 1. That is, float(base)**expepsneg.                 
!
!       EPSILPOS    (OUTPUT) real(stnd)
!                   The smallest positive floating point number whose           
!                   sum with 1. is greater than 1.             
!
!       EPSILNEG    (OUTPUT) real(stnd)
!                   The smallest positive floating point number whose           
!                   difference with 1. is less than 1.             
!
!       MACHTINY    (OUTPUT) real(stnd)
!                   The smallest positive normalized floating point number.       
!                   That is, base**minexpbase.                               
!
!       MACHHUGE    (OUTPUT) real(stnd)
!                   The largest positive normalized floating point number.        
!                   That is, (1-base**(-basedigits))*(base**maxexp).
!
!       RNDUNIT     (OUTPUT) real(stnd)
!                   Unit roundoff of the machine.
!                                                                              
!                                                                              
!   Further Details
!   _______________
!
!       This subroutine is based on the routines  MACHAR  by Cody and  LAMCH  in
!       LAPACK77 (version 3). See
!
!            Malcolm M. A. (1972) Algorithms to reveal properties of
!                floating-point arithmetic. Comms. of the ACM, 15, 949-951.
!
!            Gentleman W. M. and Marovich S. B. (1974) More on algorithms
!                that reveal properties of floating point arithmetic units.
!                Comms. of the ACM, 17, 276-277.
!
!            W.J.Cody W. J. (1988) MACHAR: A subroutine to dynamically 
!                determine machine parameters,TOMS 14, No. 4, 303-311.                                 
!                                                                              
!                                                                              
! _________________________________________________________________________________________________
!
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        integer, intent(out)    :: base, basedigits, irnd, iuflow, igrd, iexp, ifloat, minexp,  &
                                   maxexp, minexpbase, maxexpbase, expepspos, expepsneg
!
        real(stnd), intent(out) :: epspos, epsneg, epsilpos, epsilneg,    &
                                   machtiny, machhuge, rndunit
!   
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   __________________________________
!
        integer    :: ia, ip, iexpsum
!
        real(stnd) :: zero, one , two, machbase, baseh, a, b, c, tempa, eps1, eps2
!   
!
!   PARAMETERS
!   __________
!
        real(stnd) :: c2000=2000._stnd
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
!       THROUGHOUT THIS ROUTINE WE USE THE FUNCTION add TO ENSURE
!       THAT RELEVANT VALUES ARE STORED AND NOT HELD IN REGISTERS,
!       OR ARE NOT AFFECTED BY OPTIMIZERS.
!
!
!       DEFINE FLOATING POINT NUMBERS FOR INTEGER LITERALS SINCE THE 
!       VALUES COMPILERS PRODUCE FROM THESE LITERALS MAY NOT BE EXACT.
!
        zero = real( 0, stnd )
        one = real( 1, stnd )
        two = real( 2, stnd )
!   
!       DETERMINE THE BASE, base, FROM INTRINSIC FUNCTION radix(). 
!
        base = radix( one )
        machbase = real( base, stnd )
!   
!       NOW DETERMINE THE MANTISSA, nbasedigits. IT MAY BE DETERMINE BY
!       POWERING. nbasedigits IS THE SMALLEST POSITIVE INTEGER
!       FOR WHICH
!
!              real( base**nbasedigits + 1 ) - real( base**nbasedigits ) /= 1
!   
!       HOWEVER, IT IS SAFER TO USE THE INTRINSIC FUNCTION digits() .
!
!        basedigits = 0
!        a = one
!        c = one
!        do while( c==one )
!            basedigits = basedigits + 1
!            a = a * machbase
!            b = add( a, one )
!            c = add( b, -a )
!        end do
!
        basedigits = digits( one )
        a = machbase**basedigits        
!
!       DETERMINE WETHER ROUNDING OR CHOPPING OCCURS, BY ADDING
!       A BIT LESS THAN base/2 AND A BIT MORE THAN base/2 TO a.
!
        baseh = machbase/two
        b = nearest( baseh, -one )
        c = add( a, b )
        irnd = merge( 1, 0, c==a )
        b = nearest( baseh, one )
        c = add( a, b )
        if ( irnd==1 .and. c==a ) irnd = 0
!
!       COMPUTE UNIT ROUNDOFF OF THE MACHINE, DETERMINE
!       THE STYLE OF ROUNDING OR THE NUMBER OF GUARD DIGITS
!       FOR MULTIPLICATION WITH TRUNCATING ARITHMETIC.
!
        if ( irnd==1 ) then
!
            rndunit = epsilon( one )/two
            igrd = 0
!
!           TRY AND DECIDE WHETHER ROUNDIND ID DONE IN THE IEEE 'ROUND
!           TO NEAREST' STYLE. base/2 IS HALF A UNIT IN THE LAST PLACE
!           OF THE TWO NUMBERS a AND a+base. FURTHERMORE, a IS EVEN,
!           I.E. HAS LAST BIT zero, AND a+base IS ODD, I.E. HAS LAST
!           BIT one. THUS ADDING base/2 TO a SHOULD NOT CHANGE a, BUT
!           ADDING base/2 TO tempa SHOULD CHANGE tempa.
!
            tempa = add( a, machbase )
            b = add( a, baseh )
            c = add( tempa, baseh )
            if ( ( a==b ) .and. ( c>tempa ) ) irnd = 2
!
        else
!
            rndunit = epsilon( one )
!
!           DETERMINE THE NUMBER OF GUARD DIGITS FOR
!           MULTIPLICATION WITH TRUNCATING ARITHMETIC.
!
            b = nearest( one, one )
            c = add( b*one, zero )
            igrd = merge( 1, 0, c/=one )
!
        end if
!
!       DETERMINE epsneg AND expepsneg.
!
        baseh = one/machbase
!
        epsneg = spacing( one )*machbase
        do
            b = epsneg*baseh
            c = add( one, -b )
            if ( c==one ) exit
            epsneg = b
        end do
        expepsneg = exponent( epsneg ) - 1
!
!       DETERMINE epsilneg, THE SMALLEST POSITIVE FLOATING POINT NUMBER
!       WHOSE DIFFERENCE WITH ONE IS LESS THAN ONE.
!
        eps1 = epsneg
        eps2 = b
!
        do
            b = add( eps1, eps2 )/two
            c = add( one, -b )
            if ( c<one ) then
                if ( b==eps1 ) exit
                eps1 = b
            else
                if ( b==eps2 ) exit
                eps2 = b
            end if
!
        end do
!
        do
!
            b = nearest( eps1, -one )
            c = add( one, -b )
            if ( c>=one ) exit
            eps1 = b
!
            eps2 = nearest( eps2, one )
            c = add( one, -eps2 )
            if ( c<one ) exit
!
        end do
        epsilneg = merge( eps1, eps2, c>=one )
!
!       DETERMINE epspos AND expepspos.
!
        epspos = spacing( two )*machbase
        do
            b = epspos*baseh
            c = add( one, b )
            if ( c==one ) exit
            epspos = b
        end do
        expepspos = exponent( epspos ) - 1
!
!       DETERMINE epsilpos, THE SMALLEST POSITIVE FLOATING POINT NUMBER
!       WHOSE SUM WITH ONE IS GREATER THAN ONE.
!
        eps1 = epspos
        eps2 = b
!
        do
            b = add( eps1, eps2 )/two
            c = add( one, b )
            if ( c>one ) then
                if ( b==eps1 ) exit
                eps1 = b
            else
                if ( b==eps2 ) exit
                eps2 = b
            end if
!
        end do
!
        do
!
            b = nearest( eps1, -one )
            c = add( one, b )
            if ( c<=one ) exit
            eps1 = b
!
            eps2 = nearest( eps2, one )
            c = add( one, eps2 )
            if ( c>one ) exit
!
        end do
        epsilpos = merge( eps1, eps2, c<=one )
!
!       FIND minexp, maxexp, minexpbase, maxexpbase, 
!       machhuge, machtiny FROM INTRINSIC FUNCTIONS.
!
        minexp = minexponent( one )
        maxexp = maxexponent( one )
        minexpbase = minexp - 1
        maxexpbase = maxexp - 1
        machhuge = huge( one )
        machtiny = tiny( one )
!
!       CHECK FOR PARTIAL UNDERFLOW. WE ASSUME PARTIAL UNDERFLOW
!       IF WE FOUND NON VANISHING DENORMALIZED FLOATING POINT NUMBERS.
!
        a = add( machtiny/two, zero )
        b = a*one
        c = add( b, b )
        iuflow = merge( 1, 0 , (c/=zero) .and. (zero<a) .and. (a<machtiny) )
!
!       CHECK IF base IS A POWER OF 2.
!
        ia = base
        ip = 0
        do while ( ia/=1 )
            ia = ia/2
            ip = ip + 1
        end do
!
        if ( mod( base, 2**ip )==0 ) then
!
!           DETERMINE iexp THE NUMBER OF BITS NEEDED TO STORE THE EXPONENT.
!           WE ASSUME THAT THERE ARE iexp BINARY DIGITS SET ASIDE FOR 
!           THE REPRESENTATION OF THE EXPONENT. THAT IS  maxexp+abs(minexp)+1  
!           SUM APPROXIMATELY TO A POWER OF 2.
!
            iexpsum = maxexp - minexp + 1
!
            if ( iuflow==1 .and. irnd==2 )  then
!
!               CORRECT iexpsum FOR IEEE MACHINES.
!
!               WE ASSUME IEEE ARITHMETIC IF WE FOUND DENORMALIZED NUMBERS,
!               IF ARITHMETIC SEEMS TO ROUND IN THE IEEE STYLE AND IF QUIET NaNs EXIST.
!               IF THIS IS TRUE, WE NEED TO AUGMENT iexpsum BY 2 (ONE EXPONENT FOR 
!               infinity AND NaNs AND ONE EXPONENT FOR zero AND DENORMALIZED NUMBERS).
!
!               CHECK FOR THE EXISTENCE OF NaNs.
!
                a = nan()
                b = nan()
                if ( a/=b ) iexpsum = iexpsum + 2
            end if
!
            iexp = nint( log( real( iexpsum, stnd ) )/log( two ) )
!
            if ( base==2 .and. log10(machhuge)>c2000 ) then
!
!               CORRECT iexp FOR CRAY MACHINES WITH ONE RESERVED EXPONENT 
!               (BINARY) DIGIT. WE ASSUME THAT WE ARE ON A CRAY IF THE OVERFLOW
!               TRESHOLD IS SUFFICIENTLY LARGE AND IF base=2 .
!
                iexp = iexp + 1
            end if
!
!           DETERMINE ifloat THE TOTAL NUMBER OF BITS NEEDED TO STORE A
!           FLOATING-POINT NUMBER.
!
            ifloat = 1 + iexp + ( ip*basedigits )
!
!           EITHER THERE ARE AN ODD NUMBER OF BITS USED TO STORE A
!           FLOATING-POINT NUMBER, WHICH IS UNLIKELY, OR SOME BITS ARE
!           NOT USED IN THE REPRESENTATION OF NUMBERS, WHICH IS POSSIBLE
!           (E.G. CRAY MACHINES) OR THE MANTISSA HAS AN IMPLICIT BIT,
!           (E.G. IEEE MACHINES, DEC VAX MACHINES), WHICH IS PERHAPS THE
!           MOST LIKELY. WE ASSUME THAT WE ARE ON A CRAY IF THE OVERFLOW
!           TRESHOLD IS SUFFICIENTLY LARGE. IF THIS IS FALSE, WE ASSUME
!           THE LAST ALTERNATIVE AND WE NEED TO REDUCE ifloat BY ONE.
!
            if ( mod(ifloat,2)==1 .and. base==2 .and. log10(machhuge)<=c2000 ) then
                ifloat = ifloat - 1
            end if
!
!           CORRECT ifloat IF SOME BITS ARE NOT USED IN THE REPRESENTATION 
!           OF NUMBERS AS IN CRAY OR RS6000 MACHINES.
!
            select case( ifloat )
                case(33:64)
                    ifloat = 64
                case(65:128)
                    ifloat = 128
                case(129:)
                    ifloat = -1
            end select
!
        else
!
            iexp = -1
            ifloat = -1
!
        end if
!
!
!   END OF SUBROUTINE mach
!   ______________________
!
        end subroutine mach
!
! ===========================================================================================
!
        real(stnd) function add( y, z )
!
!   Purpose
!   _______
!
!       ADD  is intended to force  Y  and  Z  to be stored prior to doing
!       the addition of  Y  and  Z ,  for use in situations where optimizers
!       might hold one of these in a register.
!
!
!   Arguments
!   _________
!
!       Y, Z    (INPUT) real(stnd)
!               The values y and z.
!                                                                              
!                                                                              
! ___________________________________________________________________________________________
!
!   SPECIFICATIONS FOR ARGUMENTS
!   ____________________________
!
        real(stnd), intent(in) :: y, z
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        add = y + z
!
!
!   END OF FUNCTION add
!   ___________________
!
        end function add
!
! =============================================================================================
!
        real(stnd) function nan( )
!
!   Purpose
!   _______
!
!       NAN returns as a scalar, the bit pattern corresponding to a quiet NaN
!       in the IEEE standard 754 for binary floating-point arithmetic.
!
!
!   Arguments
!   _________
!
!       None
!
!
! _____________________________________________________________________________________________
!   
!
!   SPECIFICATIONS FOR LOCAL VARIABLES
!   __________________________________
!
        integer :: idum(8)
!
        real(stnd), save   :: nan_tmp
!   
        logical :: first= .true.
!
!
!   EXECUTABLE STATEMENTS
!   _____________________
!
        if ( first ) then
            first = .false.
!
!           SET THE BIT PATTERN FOR A QUIET NaN, THE FIRST TIME.
!
            idum(:) = not( 0 )
            nan_tmp = transfer( idum, nan_tmp )
        end if
!
!       RETURN A QUIET NaN.
!
        nan = nan_tmp
!
!
!   END OF FUNCTION nan
!   ___________________
!
        end function nan
!
! =============================================================================================
!
!
! END OF PROGRAM mach_char
! ========================
!
end program mach_char
