program test2_chol_solve
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines CHOL_CMP and CHOL_SOLVE
!   in module Lin_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Specific linear least squares problems with known exact solutions are generated for
!    checking the accuracy of subroutines CHOL_CMP and CHOL_SOLVE.
!
!       
! LATEST REVISION : 30/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, false, zero, tenth, one, two, six, ten,   &
                         chol_cmp, chol_solve, norm,  is_nan, merror, singvalues,  &
                         arth, geop, allocate_error
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=60_i4b, n2=6_i4b, n3=-1_i4b
!   
    real(stnd), parameter  :: fudge=ten, max_conda=50000._stnd
!
    character(len=*), parameter :: name_proc='chol_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: eps, eps2, ulp, err, errx,    &
                                               normx, conda, condax, tmp, d1
    real(stnd), allocatable, dimension(:,:) :: a, ata
    real(stnd), allocatable, dimension(:)   :: b, x, x2, singval, invdiag
!
    integer(i4b) :: i, j, k, n, m, krank0, success, notdone
    integer      :: iok
!
    logical(lgl) :: details, upper
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*) 
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n1,6_i4b), b(n1), singval(6_i4b),    &
              x(6_i4b), x2(6_i4b), ata(6_i4b,6_i4b), &
              invdiag(6_i4b), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
!   COMPUTE COEFFICIENT MATRIX OF LEAST SQUARES PROBLEMS.
!
    a(:n1,1_i4b) = one
    a(:n1,2_i4b) = arth( zero, one, n1 )
!
    do i = 3_i4b, 6_i4b
!
        a(:n1,i) = a(:n1,i-1_i4b)*a(:n1,2_i4b)
!
    end do
!
    upper = false
!
    do i = 1_i4b, 2_i4b
!
!       COMPUTE SOLUTIONS AND RIGHT HAND SIDES OF LEAST SQUARES PROBLEMS.
!
        if ( i==1_i4b ) then
!
            x(:6_i4b) = one
            normx     = six
!
            b(:n1) = sum( a(:n1,:6_i4b), dim=2 )
!
        else
!
            x(:6_i4b) = geop( one, tenth, 6_i4b)
            normx     = norm( x(:6_i4b) )
!
            b(:n1) = matmul( a(:n1,:6_i4b), x(:6_i4b) )
!
        end if
!
        do n = n1, n2, n3
!
            singval(:6_i4b) = singvalues( a(:n,:6_i4b) )
!
            if ( is_nan( singval(:6_i4b) ) ) cycle
!
            krank0 = count( singval(:6_i4b)/=zero )
!
!           SKIP DEFICIENT MATRICES.
!
            if ( krank0/=6_i4b ) cycle
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            conda = maxval( singval(:6_i4b) )/minval( singval(:6_i4b) )
!
            tmp =  min( conda, max_conda )
!
            condax = normx*tmp*tmp
!
!           FORM NORMAL EQUATIONS OF THE PROBLEM.
!
            ata(:6_i4b,:6_i4b) = matmul( transpose(a(:n,:6_i4b)), a(:n,:6_i4b) )
            x2(:6_i4b)         = matmul( b(:n), a(:n,:6_i4b) )
!
!           COMPUTE CHOLESKY FACTORIZATION OF ata .
!
            upper = .not.upper
!
            call chol_cmp( ata(:6_i4b,:6_i4b), invdiag(:6_i4b), d1, upper=upper )
!
            if ( d1==zero ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
!           COMPUTE THE SOLUTION FOR LINEAR LEAST SQUARES PROBLEM
!
!                       a*x = b .
!
            call chol_solve( ata(:6_i4b,:6_i4b), invdiag(:6_i4b), x2(:6_i4b), upper=upper )
!
!           COMPUTE ERRORS.
!
            x2(:6_i4b) = x2(:6_i4b) - x(:6_i4b)
!
            errx = norm( x2(:6_i4b) )
!
!           CHECK THE RESULTS FOR SMALL RESIDUALS.
!       
            m = m + 1_i4b
!
            err = errx/condax
!
            eps2 = real( n, stnd)*eps
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n,' , case= ', i
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of the coefficient matrix         = ',   &
                                      conda
!
!                   PRINT ERRORS OF SOLUTIONS.
!
                    write (prtunit,*) 'Accuracy of the computed solution (absolute error) = ',   &
                                      errx
!
                    write (prtunit,*) 'Accuracy of the computed solution (relative error) = ',   &
                                      errx/normx
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ', eps2, ' , n= ', n,' , case= ', i
!
!               PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                write (prtunit,*) 'Condition number of the coefficient matrix         = ',   &
                                  conda
!
!               PRINT ERRORS OF SOLUTIONS.
!
                write (prtunit,*) 'Accuracy of the computed solutions (absolute error) = ',   &
                                  errx
!
                write (prtunit,*) 'Accuracy of the computed solutions (relative error) = ',   &
                                  errx/normx
!
                write (prtunit,*)
!
            end if
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, x, x2, b, singval, ata, invdiag )
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_chol_solve
! ===============================
!
end program test2_chol_solve
