program test2_hous1
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines HOUS1 and APPLY_HOUS1 in module Hous_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Specific linear problems with known exact solutions are generated for checking the accuracy
!    of subroutines HOUS1 and APPLY_HOUS1.
!
!                                                                              
! LATEST REVISION : 03/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,  &
                         true, false, zero, one, c100, hous1,     &
                         apply_hous1, triang_solve, norm
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='hous1'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: eps, eps2, ulp, err, errx, errb,      &
                                               normx, normb, norma, beta, scale
    real(stnd), allocatable, dimension(:,:) :: a, a2
    real(stnd), allocatable, dimension(:)   :: b, x, x2, res, diagr
!
    integer(i4b) :: i, j, k, n, m, success
    integer      :: iok
!
    logical(lgl) :: details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*) 
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(n,n), a2(n,n), x2(n), res(n),  &
                  diagr(n), x(n), b(n), stat=iok   )
!
        if ( iok/=0 ) cycle
!
        do i = 1_i4b, 2_i4b
!
            select case( i )
!
                case( 1_i4b )
!
!                   TEST 1 : a(k,j) = abs(k-j) AND b(k) = 1 .
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
                            a(k,j) = real( abs(k-j), stnd )
                        end do
!
                    end do
!
!                   SET RIGHT HANDSIDE.
!
                    b(1_i4b:n) = one
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b)         = one/real( n-1_i4b, stnd )
                    x(2_i4b:n-1_i4b) = zero
                    x(n)             = x(1_i4b)
!
                case( 2_i4b )
!
!                   TEST 2 : a(k,j) = abs(k,j) AND b(k) = k .
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
                            a(k,j) = real( max(k,j), stnd )
                        end do
!
!                       SET RIGHT HANDSIDE.
!
                        b(j) = real( j, stnd )
!
                    end do
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b)   = one
                    x(2_i4b:n) = zero
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( x ) ) then
                if ( .not.all( ieee_is_normal( x(1_i4b:n) ) ) ) cycle
            end if
#endif
!
            a2(:n,:n) = a(:n,:n)
!
            norma = norm( a(:n,:n) )
            normx = norm( x(:n) )
            normb = norm( b(:n) )
!
!           COMPUTE THE SOLUTION MATRIX FOR LINEAR SYSTEM
!
!                         a*x = b .
!    
!           TRANSFORM THE MATRIX a TO UPPER TRIANGULAR FORM BY APPLYING 
!           A SERIE OF HOUSEHOLDER REFLECTIONS ON THE COLUMNS OF a AND
!           APPLY THE TRANSFORMATIONS TO b .
!
            x2(:n) = b(:n)
!
            do j = 1_i4b, n
!
                call hous1( a(j:n,j), beta, diagr(j) )
                call apply_hous1( a(j:n,j), beta, a(j:n,j+1_i4b:n), left=true )
                call apply_hous1( a(j:n,j), beta, x2(j:n) )
!
                a(j,j) = diagr(j)
!
            end do
!        
            call triang_solve( a(:n,:n), x2(:n), scale, upper=true, trans=false )
!
#ifdef _F2003
            if ( ieee_support_datatype( scale ) ) then
                if ( .not.ieee_is_normal( scale ) ) cycle
            end if
#endif
!
            m = m + 1_i4b
!
!           UNSCALE THE SOLUTION VECTOR.
!
            x2(:n) = x2(:n)/scale
!
!           COMPUTE RESIDUALS AND ERRORS.
!
            res(:n) = b(:n) - matmul( a2(:n,:n), x2(:n) )
            errb  = norm( res(:n) )
!
            res(:n) = x2(:n) - x(:n)
            errx  = norm( res(:n) )
!
            err = max( errb/(normb+norma), errx/normx )
!            err = errx/normx
!
            eps2 = real( n, stnd)*eps
!
!           CHECK THE RESULTS FOR SMALL RESIDUALS.
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n,' , case= ', i
!
                    write (prtunit,*) 'Accuracy of the computed solution (absolute error)        = ',   &
                                      errx
!
                    write (prtunit,*) 'Accuracy of the computed solution (relative error)        = ',   &
                                      errx/normx
!
                    write (prtunit,*) 'Accuracy of the computed residual (absolute error)        = ',   &
                                      errb
!
                    write (prtunit,*) 'Accuracy of the computed residual (relative error)        = ',   &
                                      errb/(normb+norma)
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ', eps2, ' , n= ', n,' , case= ', i
!
                write (prtunit,*) 'Accuracy of the computed solution (absolute error)        = ',   &
                                  errx
!
                write (prtunit,*) 'Accuracy of the computed solution (relative error)        = ',   &
                                  errx/normx
!
                write (prtunit,*) 'Accuracy of the computed residual (absolute error)        = ',   &
                                  errb
!
                write (prtunit,*) 'Accuracy of the computed residual (relative error)        = ',   &
                                  errb/(normb+norma)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, a2, res, x, x2, b, diagr )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_hous1
! ==========================
!
end program test2_hous1
