program test2_lq_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines LQ_CMP and ORTHO_GEN_LQ
!   in module QR_Procedures.
!                                                                              
! LATEST REVISION : 02/02/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, lgl, i4b, n2_def, stnd, false, zero, half, one, c50,   &
                         lq_cmp, ortho_gen_lq, norm, lamch
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, blksz IS THE DEFAULT BLOCK SIZE USED IN MODULE QR_Procedures.
!
    integer(i4b), parameter :: prtunit=6, blksz=20, m=n2_def
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='lq_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                               :: err1, err2, err, eps, ulp, safmin, tmp
    real(stnd), dimension(:),   allocatable  :: diagl, tau
    real(stnd), dimension(:,:), allocatable  :: a, q, l, res
!
    integer(i4b) :: i, j, k, n, p, o, success
    integer      :: iok
!
    logical(lgl) :: use_qr, details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
    use_qr = false
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
!   COUNTER FOR THE TEST NUMBER.
!
    success = 0_i4b
    p = 0_i4b
!
    do o = 1_i4b, 2_i4b
!
        if ( o==1_i4b ) then
            n = m + blksz
        else
            n = m - blksz
        end if
!
        do i = m, m+blksz
!
            p = p + 1_i4b
!
!           ALLOCATE WORK ARRAYS.
!
            k = min( i, n )
!
            allocate( a(i,n), res(i,n), q(k,n), l(i,k), diagl(k), tau(k), stat=iok )
!
            if ( iok==0 ) then
!
!               TEST p : LQ DECOMPOSITION OF a(:i,:n) .
!
!               GENERATE RANDOM DATA MATRIX.
!
                call random_number( a(:i,:n) )
!
                a(:i,:n) = a(:i,:n) - half
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    a(:i,:n) = a(:i,:n)/tmp
                end if
!
!               SAVE DATA MATRIX .
!
                res(:i,:n) = a(:i,:n)
!
!               COMPUTE LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
                call lq_cmp( res(:i,:n), diagl(:k), tau(:k), use_qr=use_qr )
!
!               GENERATE ORTHOGONAL MATRIX Q OF LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
                q(:k,:n) = res(:k,:n)
                call ortho_gen_lq( q(:k,:n), tau(:k), use_qr=use_qr )
!
!               RESTORE TRIANGULAR FACTOR L OF LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
                l(1_i4b,1_i4b) = diagl(1_i4b)
                l(2_i4b:i,1_i4b) = res(2_i4b:i,1_i4b)
!
                do j = 2_i4b, k
                    l(1_i4b:j-1_i4b,j) = zero
                    l(j,j) = diagl(j)
                    l(j+1_i4b:i,j) = res(j+1_i4b:i,j)
                end do
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:i,:n) - l(:i,:k)*q(:k,:n).
!
                res(:i,:n) = a(:i,:n) - matmul( l(:i,:k), q(:k,:n) )
!
                err1 = maxval(  norm( res(:i,:n), dim=1) / norm( a(:i,:n), dim=1) )/real(n,stnd)
!
!               CHECK ORTHOGONALITY OF q(:k,:n).
!
                res(:k,:k) = matmul( q(:k,:n), transpose(q(:k,:n)) )
!
                do j = 1_i4b, k
                    res(j,j) =  res(j,j) - one
                end do
!
                err2 = maxval( norm(res(:k,:k), dim=2_i4b) )/real(k,stnd)
!
                err = max( err1, err2 )
!
!               PRINT THE RESULTS OF THE TESTS.
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',p,' is correct'
!
                        write (prtunit,*) 'nr= ', i,' , nc= ', n,    &
                                          ' , use_qr= ', use_qr
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a - l*q
!                       WHERE q AND l ARE THE ORTHOGONAL MATRIX Q AND L FACTOR FROM THE
!                       LQ DECOMPOSITION OF a.
!
                        write (prtunit,*) 'Accuracy of the LQ decomposition       = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q*q**(t)
!                       WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE LQ DECOMPOSITION OF a .
!
                        write (prtunit,*) 'Orthogonality of the computed Q factor = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',p,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', i,' , nc= ', n,    &
                                      ' , use_qr= ', use_qr
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a - l*q
!                   WHERE q AND l ARE THE ORTHOGONAL MATRIX Q AND L FACTOR FROM THE
!                   LQ DECOMPOSITION OF a.
!
                    write (prtunit,*) 'Accuracy of the LQ decomposition       = ', &
                                     err1/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q*q**(t)
!                   WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE LQ DECOMPOSITION OF a .
!
                    write (prtunit,*) 'Orthogonality of the computed Q factor = ', &
                                     err2/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
!               DEALLOCATE WORK ARRAYS.
!
                deallocate( a, q, l, res, diagl, tau )
!
            else
!
                write (prtunit,*)    &
                'Problem in attempt to allocate memory for test ',p
!
            end if
!
        end do
!
    end do
!
    if ( p/=success ) then
        write (prtunit,*) p - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_lq_cmp
! ===========================
!
end program test2_lq_cmp
