program test2_probq
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBQ and  PINVQ2 in module Prob_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 21/05/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : i4b, stnd, lgl, n1_def, n2_def, n3_def,   &
                         true, false, zero, one, c99,              &
                         allocate_error, merror, probq, pinvq2
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, &
                               ndf=30, nprob=100
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    real(stnd),  parameter :: eps = 1.0e-4_stnd 
!
    character(len=*), parameter :: name_proc='probq'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err
    real(stnd), allocatable, dimension(:)   :: df, prob, prob2
    real(stnd), allocatable, dimension(:,:) :: xx, p, x
!
    integer(i4b), allocatable, dimension(:) :: ndf_vec
    integer(i4b)                            :: i, k, l, m, n, success
    integer                                 :: iok
!   
    logical(lgl) :: upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( xx(nprob,ndf), df(ndf), ndf_vec(ndf), prob(nprob),  prob2(nprob), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(nprob,n), x(nprob,n), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       GENERATE THE DEGREES OF FREEDOM OF THE CHI-SQUARED DISTRIBUTION .
!
        call random_number( df(:ndf) )
!
        df(:ndf) = c99*df(:ndf)
        where( df(:ndf)<one )  df(:ndf) = df(:ndf) + one
!
        df(1_i4b) = one
!
        ndf_vec(:ndf) = int( df(:ndf), i4b )
        df(:ndf)      = real( ndf_vec(:ndf), stnd )
!
!       GENERATE A RANDOM PROBABILITY VECTOR prob(:nprob) .
!
        call random_number( prob(:nprob) )
!
!       RECOMPUTE THE PROBABILITIES FROM THE CHI-SQUARED DEVIATES .
!
        do k = 1_i4b, 12_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( k )
!
            case( 1_i4b, 4_i4b, 7_i4b, 10_i4b )
!
                if ( k==1_i4b .or. i==7_i4b ) then
!
!                   TESTING LOWER TAIL FOR TESTS 1 TO 3 AND 7 TO 9.
!
                    upper = false
!
                    prob2(:nprob) = prob(:nprob)
!
                    if ( k==7_i4b ) then
!
!                       ADJUST THE NUMBERS OF DEGREES OF FREEDOM TO TEST NORMAL APPROXIMATION.
!
                        ndf_vec(:ndf) = 200_i4b + ndf_vec(:ndf)
                        df(:ndf)      = real( ndf_vec(:ndf), stnd )
!                
                    end if
!
!                   COMPUTE THE REQUIRED LOWER TAILS QUANTILES.
!
                    do i = 1_i4b, ndf
!
                        do l = 1_i4b, nprob
!
!                           COMPUTE THE CHI-SQUARED DEVIATES CORRESPONDING TO LOWER TAIL AREAS OF prob(:nprob)
!                           ACCURATELY WITH pinvq2 FUNCTION.
!
                            xx(l,i) = pinvq2( prob(l), df(i), prec=0.5e-10_stnd )
!
                        end do
!
                    end do
!
                else
!
!                   TESTING UPPER TAIL FOR TESTS 4 TO 6 AND 10 TO 12.
!
                    upper = true
!
!                   ADJUST THE PROBABILITIES.
!
                    prob2(:nprob) = one - prob(:nprob)
!
                end if
!
!               TESTS 1, 4, 7 AND 10 : SCALAR FORM OF probq2 GENERIC FUNCTION.
!
                do i = 1_i4b, ndf
!
                    do l = 1_i4b, nprob
!
                        p(1_i4b,1_i4b) = probq( xx(l,i), ndf_vec(i), upper=upper, ndf_max=100_i4b )
!
!                       CHECK THAT p(1,1) AND prob(l) AGREE.
!
                        err = max( abs( p(1_i4b,1_i4b) - prob2(l) ), err )
!
                    end do
!
                end do
!
            case( 2_i4b, 5_i4b, 8_i4b, 11_i4b )
!
!               TESTS 2, 5, 8 AND 11 : VECTOR FORM OF probq2 GENERIC FUNCTION.
!
                do i = 1_i4b, ndf
!
                    p(:nprob,1_i4b) = probq( xx(:nprob,i), ndf_vec(i), upper=upper, ndf_max=100_i4b )
!
!                   CHECK THAT p(:nprob,1) AND prob(:nprob) AGREE.
!
                    err = max( maxval( abs(  p(:nprob,1_i4b) - prob2(:nprob) ) ), err )
!
                end do
!
            case( 3_i4b, 6_i4b, 9_i4b, 12_i4b )
!
!               TESTS 3, 6, 9 AND 12 : MATRIX FORM OF probq2 GENERIC FUNCTION.
!
                do i = 1_i4b, ndf
!
                    x(:nprob,:n) = spread( xx(:nprob,i), dim=2, ncopies=n )
                    p(:nprob,:n) = probq( x(:nprob,:n), ndf_vec(i), upper=upper, ndf_max=100_i4b )
!
!                   CHECK THAT EACH COLUMN OF p(:nprob,:n) AGREES WITH prob(:nprob).
!
                    x(:nprob,:n) = spread( prob2(:nprob), dim=2, ncopies=n )
!
                    err = max( maxval( abs(  p(:nprob,:n) - x(:nprob,:n) ) ), err )
!
                end do
!
            end select
!
!           PRINT THE RESULTS OF THE TESTS.
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', k,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, x )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( xx, df, ndf_vec, prob, prob2 )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_probq
! ==========================
!
end program test2_probq
