program test2_qr_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines QR_CMP and ORTHO_GEN_QR
!   in module QR_Procedures.
!                                                                              
! LATEST REVISION : 02/02/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, n2_def, zero, half, one, c50, qr_cmp, &
                         ortho_gen_qr, norm, lamch
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, blksz IS THE DEFAULT BLOCK SIZE USED IN MODULE QR_Procedures.
!
    integer(i4b), parameter :: prtunit=6, blksz=20, n=n2_def
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='qr_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd), dimension(:,:), allocatable  :: a, q, r, res
    real(stnd), dimension(:),   allocatable  :: diagr, beta
    real(stnd)                               :: err1, err2, err, eps, ulp, safmin, tmp
!
    integer(i4b) :: i, j, k, l, m, p, success
    integer      :: iok
!   
    logical(lgl) :: details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
!   COUNTERS FOR THE TEST NUMBER AND THE NUMBER OF SUCCES.
!
    success = 0_i4b
    p = 0_i4b
!
    do l = 1_i4b, 2_i4b
!
        if ( l==1_i4b ) then
            m = n + blksz
        else
            m = n - blksz
        end if
!
        do i = n, n+blksz
!
            p = p + 1_i4b
!
!           ALLOCATE WORK ARRAYS.
!
            k = min( m, i )
!
            allocate( a(m,i), res(m,i), q(m,k), r(k,i), diagr(k), beta(k), stat=iok )
!
            if ( iok==0 ) then
!
!               TEST p : QR DECOMPOSITION OF a(:m,:i) .
!
!               GENERATE RANDOM DATA MATRIX.
!
                call random_number( a(:m,:i) )
!
                a(:m,:i) = a(:m,:i) - half
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    a(:m,:i) = a(:m,:i)/tmp
                end if
!
!               SAVE DATA MATRIX .
!
                res(:m,:i) = a(:m,:i)
!
!               COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
                call qr_cmp( res(:m,:i), diagr(:k), beta(:k) )
!
!               GENERATE ORTHOGONAL MATRIX Q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
                q(:m,:k) = res(:m,:k)
                call ortho_gen_qr( q(:m,:k), beta(:k) )
!
!               RESTORE TRIANGULAR FACTOR R OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
                r(1_i4b,1_i4b) = diagr(1_i4b)
                r(2_i4b:k,1_i4b) = zero
!
                do j = 2_i4b, k
                    r(1_i4b:j-1_i4b,j) = res(1_i4b:j-1_i4b,j)
                    r(j,j) = diagr(j)
                    r(j+1_i4b:k,j) = zero
                end do
!
                r(1_i4b:k,k+1_i4b:i) = res(1_i4b:k,k+1_i4b:i)
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:i) - q(:m,:k)*r(:k,:i).
!
                res(:m,:i) = a(:m,:i) - matmul( q(:m,:k), r(:k,:i) )
!
                err1 = maxval(  norm( res(:m,:i) , dim=2)  /  norm( a(:m,:i) , dim=2) )/real(m,stnd)
!
!               CHECK ORTHOGONALITY OF q(:m,:k).
!
                res(:k,:k) = matmul( transpose(q(:m,:k)), q(:m,:k) )
!
                do j = 1_i4b, k
                    res(j,j) =  res(j,j) - one
                end do
!
                err2 = maxval( norm(res(:k,:k), dim=2_i4b) )/real(k,stnd)
!
                err = max( err1, err2 )
!
!               PRINT THE RESULTS OF THE TESTS.
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',p,' is correct'
!
                        write (prtunit,*) 'nr= ', m,' , nc= ', i
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a - q*r
!                       WHERE q AND r ARE THE ORTHOGONAL MATRIX Q AND R FACTOR FROM THE
!                       QR DECOMPOSITION OF a.
!
                        write (prtunit,*) 'Accuracy of the QR decomposition       = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!                       WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE QR DECOMPOSITION OF a .
!
                        write (prtunit,*) 'Orthogonality of the computed Q factor = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',p,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', m,' , nc= ', i
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a - q*r
!                   WHERE q AND r ARE THE ORTHOGONAL MATRIX Q AND R FACTOR FROM THE
!                   QR DECOMPOSITION OF a.
!
                    write (prtunit,*) 'Accuracy of the QR decomposition       = ', &
                                     err1/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q**(t)*q
!                   WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE QR DECOMPOSITION OF a .
!
                    write (prtunit,*) 'Orthogonality of the computed Q factor = ', &
                                     err2/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
!               DEALLOCATED WORK ARRAYS.
!
                deallocate( a, q, r, res, diagr, beta )
!
            else
!
                write (prtunit,*)    &
                'Problem in attempt to allocate memory for test ',p
!
            end if
!
        end do
!
    end do
!
    if ( p/=success ) then
        write (prtunit,*) p - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_qr_cmp
! ===========================
!
end program test2_qr_cmp
