program test2_qr_solve2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines QR_CMP2 and QR_SOLVE2 in module QR_Procedures. 
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Specific linear problems with known exact solutions are generated for checking the accuracy
!    of subroutines QR_CMP2 and  QR_SOLVE2.
!
!                                                                              
! LATEST REVISION : 31/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,   &
                         zero, one, two, ten, qr_cmp2, qr_solve2,  &
                         norm, singvalues, merror, is_nan, lamch,  &
                         maxexp, allocate_error
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nrhs=10
!   
    real(stnd), parameter  :: fudge=ten, max_conda=1000._stnd
!
    character(len=*), parameter :: name_proc='qr_solve2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: eps, eps2, ulp, err, errx,  &
                                               normx, conda, condax, tmp,  &
                                               tol, safmin
    real(stnd), allocatable, dimension(:,:) :: a, x2, res
    real(stnd), allocatable, dimension(:)   :: b, x, diagr, beta, singval
!
    integer(i4b) :: i, j, k, n, n0, nrhs2, m, success, krank, krank0
    integer(i4b), allocatable, dimension(:) :: ip
    integer      :: iok
!
    logical(lgl) :: details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*) 
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   ALLOCATE WORK ARRAYS.
!
    n = max( n1, n2 )
!
    allocate( a(n,n), x2(n,nrhs), res(n,nrhs), diagr(n), &
              beta(n), x(n), b(n), ip(n), singval(n),    &
              stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        do i = 1_i4b, 4_i4b
!
            select case( i )
!
                case( 1_i4b )
!
                    n0 = n
!
!                   TESTS 1 AND 2 : a(k,j) = abs(k-j) AND b(k) = 1 .
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
                            a(k,j) = real( abs(k-j), stnd )
                        end do
!
                    end do
!
!                   SET RIGHT HANDSIDE.
!
                    b(1_i4b:n) = one
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b)         = one/real( n-1_i4b, stnd )
                    x(2_i4b:n-1_i4b) = zero
                    x(n)             = x(1_i4b)
!
                case( 3_i4b )
!
                    n0 = n
!
!                   TESTS 3 AND 4 : a(k,j) = abs(k,j) AND b(k) = k .
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
                            a(k,j) = real( max(k,j), stnd )
                        end do
!
!                       SET RIGHT HANDSIDE.
!
                        b(j) = real( j, stnd )
!
                    end do
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b)   = one
                    x(2_i4b:n) = zero
!
                case( 5_i4b )
!
                    n0 = min( n, maxexp - 1_i4b )
!
!                   TESTS 5 and 6 : WILKINSON'S TEST LOWER TRIANGLE.
!
                    do j = 1_i4b, n0
!
                        a(1_i4b:j-1_i4b,j) = zero
                        a(j,j)             = one
                        a(j+1_i4b:n0,j)    = -one
!
                    end do
!
!                   SET RIGHT HANDSIDE.
!
                    b(1_i4b:n0) = one
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b) = one                
!
                    do j = 2_i4b, n0
                        x(j) = x(j-1_i4b)*two             
                    end do
!
                case( 7_i4b )
!
                    n0 = min( n, maxexp - 1_i4b )
!
!                   TESTS 7 and 8 : WILKINSON'S TEST WITH UPPER TRIANGLE.
!
                    do j = 1_i4b, n0
!
                        a(j,1_i4b:j-1_i4b) = zero
                        a(j,j)             = one
                        a(j,j+1_i4b:n0)    = -one
!
                    end do
!
!                   SET RIGHT HANDSIDE.
!
                    b(1_i4b:n0) = one
!
!                   SET EXACT SOLUTION.
!
                    x(n0) = one                
!
                    do j = n0-1_i4b, 1_i4b, -1_i4b
                        x(j) = x(j+1_i4b)*two             
                    end do
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( x ) ) then
                if ( .not.all( ieee_is_normal( x(1_i4b:n0) ) ) ) cycle
            end if
#endif
!
            if ( (i/2_i4b)*2_i4b/=i ) then
!
                singval(:n0) = singvalues( a(:n0,:n0) )
!
                if ( is_nan( singval(:n0) ) ) cycle
!
                krank0 = count( singval(:n0)/=zero )
!
!               SKIP DEFICIENT MATRICES.
!
                if ( krank0/=n0 ) cycle
!
                normx = norm( x(:n0) )
!
!               COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
                conda = maxval( singval(:n0) )/minval( singval(:n0) )
!
                tmp =  min( conda, max_conda )
!
                condax = normx*tmp
!
!                condax = normx*conda
!
!               COMPUTE QR DECOMPOSITION OF MATRIX FOR LINEAR SYSTEM
!
                krank = 0_i4b
                tol   = safmin
!        
                call qr_cmp2( a(:n0,:n0), diagr(:n0), beta(:n0), ip(:n0), krank, tol=tol )
!
            end if
!
            m = m + 1_i4b
!        
            select case( i )
!
                case( 1_i4b, 3_i4b, 5_i4b, 7_i4b )
!
!                   TESTS 1, 3, 5, 7 : ONE RIGHT HAND-SIDE.
!
                    res(:n0,1_i4b) = b(:n0)         
!        
                    call qr_solve2( a(:n0,:n0), diagr(:n0), beta(:n0), ip(:n0), krank, res(:n0,1_i4b),    &
                                    x2(:n0,1_i4b) )
!
                case( 2_i4b, 4_i4b, 6_i4b, 8_i4b  )
!
!                   TESTS 2, 4, 6, 8 : SEVERAL RIGHT HAND-SIDES.
!
                    do j = 1_i4b, nrhs
                        res(:n0,j) = b(:n0)         
                    end do
!        
                    call qr_solve2( a(:n0,:n0), diagr(:n0), beta(:n0), ip(:n0), krank, res(:n0,:nrhs),     &
                                    x2(:n0,:nrhs) )
!
            end select
!
!           COMPUTE ERRORS FOR SOLUTION.
!       
            if ( (i/2_i4b)*2_i4b==i ) then
                nrhs2 = nrhs
            else
                nrhs2 = 1_i4b
            end if
!
            errx = zero
!
            do j = 1_i4b, nrhs2
!
                res(:n0,1_i4b) = x2(:n0,j) - x(:n0)
!
                err  = norm( res(:n0,1_i4b) )
                errx = max( errx, err )
!
            end do
!
            err = errx/condax
!
            eps2 = real( n0, stnd)*eps
!
!           CHECK THE RESULTS FOR SMALL ERRORS.
!
            if ( err<=eps2 .or. krank/=n0 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n0,' , case= ', i,' , krank= ', krank,' , nrhs= ', nrhs2
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of the coefficient matrix          = ', conda
!
!                   PRINT ERRORS OF SOLUTIONS.
!
                    write (prtunit,*) 'Accuracy of the computed solutions (absolute error) = ',   &
                                      errx
!
                    write (prtunit,*) 'Accuracy of the computed solutions (relative error) = ',   &
                                      errx/normx
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ', eps2, ' , n= ', n0,' , case= ', i,    &
                                  ' , krank= ', krank,' , nrhs= ', nrhs2
!
!               PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                write (prtunit,*) 'Condition number of the coefficient matrix          = ', conda
!
!               PRINT ERRORS OF SOLUTIONS.
!
                write (prtunit,*) 'Accuracy of the computed solutions (absolute error) = ',   &
                                  errx
!
                write (prtunit,*) 'Accuracy of the computed solutions (relative error) = ',   &
                                  errx/normx
!
                write (prtunit,*)
!
            end if
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, res, x, x2, b, diagr, beta, ip, singval )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_qr_solve2
! ==============================
!
end program test2_qr_solve2
