program test2_select_singval_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SELECT_SINGVAL_CMP
!   in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking the accuracy
!    of subroutine SELECT_SINGVAL_CMP.
!
! LATEST REVISION : 31/07/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,  &
                         zero, one, two, five, ten, c100,         &
                         eigval_sort, select_singval_cmp, arth,   &
                         gen_random_mat, lamch
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nsing_def=5
!   
    real(stnd), parameter :: fudge=ten
!
    character(len=*), parameter :: name_proc='select_singval_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, eps, eps1, eps2, ulp,   &
                                               abstol, safmin, tmp
    real(stnd), allocatable, dimension(:)   :: eigval, singval, singval2
    real(stnd), allocatable, dimension(:,:) :: a, a2
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, nsing, nsing2, success, notdone
    logical(lgl) :: failure, details
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err1 )
    eps    = fudge*ulp
    eps2   = fudge*sqrt( ulp )
    safmin = lamch( 'S' )
    abstol = sqrt( safmin )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), singval2(n), stat=iok )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )   deallocate( eigval )
            if ( allocated(singval) )  deallocate( singval )
            if ( allocated(singval2) ) deallocate( singval2 )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
            case( 1_i4b )
!
!               TEST 1: SVD OF a(:nr,:nc) WITH nr=nc .
!
                nr = n
                nc = n
!
            case( 2_i4b )
!
!               TEST 2: SVD OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                nr = n + 1_i4b
                nc = n
!
            case( 3_i4b )
!
!               TEST 3: SVD OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                nr = n + n
                nc = n
!
            case( 4_i4b )
!
!               TEST 4: SVD OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                nr = n
                nc = n + 1_i4b
!
            case( 5_i4b )
!
!               TEST 5: SVD OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                nr = n
                nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), a2(nr,nc), stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )  deallocate( a )
                if ( allocated(a2) ) deallocate( a2 )
!
                cycle
!
            end if
!
            do j = 1_i4b, 16_i4b
!
                select case( j )
!
                case( 1_i4b )
!
!                   TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b)   = one
                    eigval(2_i4b:n) = ulp
!
                case( 2_i4b )
!
!                   TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n-1_i4b) = one
                    eigval(n)             = ulp
!
                case( 3_i4b )
!
!                   TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                    eigval(n)             = one
!
                case( 4_i4b )
!
!                   TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                    eigval(1_i4b)         = ulp
                    eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                    eigval(n)             = two
!
                case( 5_i4b )
!
!                   TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                case( 6_i4b )
!
!                   TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                    tmp = real( n-1_i4b, stnd )
!
                    do l = 1_i4b, n
                        eigval(l) = ulp**(real(l-1_i4b,stnd)/tmp)
                    end do
!
                case( 7_i4b )
!
!                   TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                    eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                case( 8_i4b )
!
!                   TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                    eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                case( 9_i4b )
!
!                   TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                case( 10_i4b )
!
!                   TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                case( 11_i4b )
!
!                   TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                    call random_number( eigval(:n) )
!
                    eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                case( 12_i4b )
!
!                   TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                case( 13_i4b )
!
!                   TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                   SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                    tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                    eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                case( 14_i4b )
!
!                   TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                   SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                    tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                    eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                case( 15_i4b )
!
!                   TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                   SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                    tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                    eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                case default
!
!                   DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                    call random_number( eigval(:n) )
!
                    call random_number( tmp )
!
                    if ( tmp>safmin ) then
                        eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                    end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
!               SORT THE SINGULAR VALUES.
!
                call eigval_sort( sort, singval(:n) )
!
                where ( singval(:n)==zero )
!
                    eigval(:n) = one
!
                elsewhere
!
                    eigval(:n) = singval(:n)
!
                end where
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                call gen_random_mat( singval(:n), a(:nr,:nc) )
!
                do l = 1_i4b, 2_i4b
!
!                   COMPUTE nsing SINGULAR VALUES OF RANDOM DATA MATRIX.
!
                    if ( l==1_i4b ) then
                        nsing = min( n, nsing_def )
                    else
                        nsing = n
                    end if
!
!                   MAKE A COPY THE ORIGINAL MATRIX.
!
                    a2(:nr,:nc) = a(:nr,:nc)
!
!                   COMPUTE SVD OF RANDOM DATA MATRIX.
!
                    call select_singval_cmp( a2(:nr,:nc), nsing2, singval2(:n), failure, &
                                             ls=nsing, abstol=abstol, sort=sort )
!
                    if ( failure ) then
!
                        notdone = notdone + 1_i4b
!
                        cycle
!
                    end if
!
                    m = m + 1_i4b
!
!                   ABSOLUTE ERRORS OF EIGENVALUES.
!
                    err1 = maxval( abs( singval2(:nsing) - singval(:nsing) ) )
!
!                   RELATIVE ERRORS OF EIGENVALUES.
!
                    err2 = maxval( abs( (singval2(:nsing) - singval(:nsing))/eigval(:nsing) ) )
!
                    eps1 = real( n, stnd)*eps*maxval( singval(:nsing) )
!
                    if ( err1<=eps1 .and. err2<=eps2 ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , nsing= ', nsing,' , case= ', j
!
!                           COMPUTE ABSOLUE ERRORS OF SINGULAR VALUES.
!
                            write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                              err1
!
!                           COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                            write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                              err2
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        write (prtunit,*) '=> ', err1,' > ', eps1,' or ', err2,' > ',eps2,' , nr= ', nr, &
                                          ' , nc= ', nc, ' , nsing= ', nsing,' , case= ', j,             &
                                          ' , failure= ', failure
!
!                       COMPUTE ABSOLUE ERRORS OF SINGULAR VALUES.
!
                        write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                          err1
!
!                       COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                        write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                          err2
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, a2 )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, singval2 )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_select_singval_cmp
! =======================================
!
end program test2_select_singval_cmp
