program test2_solve_lin
!
!
! Purpose
! =======
!
!   This program is intended to test function SOLVE_LIN in module Lin_Procedures. 
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Specific linear problems with known exact solutions are generated for checking the accuracy
!    of function SOLVE_LIN.
!
!       
! LATEST REVISION : 31/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,  &
                         zero, one, two, solve_lin, norm,         &
                         is_nan, merror, singvalues, maxexp,      &
                         allocate_error
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nrhs=10
!   
    real(stnd), parameter  :: fudge=one, max_conda=1000._stnd
!
    character(len=*), parameter :: name_proc='solve_lin'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: eps, eps2, ulp, err, errx,  &
                                               normx, conda, condax, tmp
    real(stnd), allocatable, dimension(:,:) :: a, x2, b2
    real(stnd), allocatable, dimension(:)   :: res, b, x, singval
!
    integer(i4b) :: i, j, k, n, n0, nrhs2, m, krank0, success, notdone
    integer      :: iok
!
    logical(lgl) :: details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*) 
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   ALLOCATE WORK ARRAYS.
!
    n = max( n1, n2 )
!
    allocate( a(n,n), x2(n,nrhs), singval(n),  &
              x(n), b(n), b2(n,nrhs), res(n),  &
              stat=iok     )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        do i = 1_i4b, 8_i4b
!
            select case( i )
!
                case( 1_i4b )
!
                    n0 = n
!
!                   TESTS 1 AND 2 : a(k,j) = abs(k-j) AND b(k) = 1 .
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
                            a(k,j) = real( abs(k-j), stnd )
                        end do
!
                    end do
!
!                   SET RIGHT HANDSIDE.
!
                    b(1_i4b:n) = one
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b)         = one/real( n-1_i4b, stnd )
                    x(2_i4b:n-1_i4b) = zero
                    x(n)             = x(1_i4b)
!
                case( 3_i4b )
!
                    n0 = n
!
!                   TESTS 3 AND 4 : a(k,j) = max(k,j) AND b(k) = k .
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
                            a(k,j) = real( max(k,j), stnd )
                        end do
!
!                       SET RIGHT HANDSIDE.
!
                        b(j) = real( j, stnd )
!
                    end do
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b)   = one
                    x(2_i4b:n) = zero
!
                case( 5_i4b )
!
                    n0 = min( n, maxexp - 1_i4b )
!
!                   TESTS 5 and 6 : WILKINSON'S TEST WITH LOWER TRIANGLE.
!
                    do j = 1_i4b, n0
!
                        a(1_i4b:j-1_i4b,j) = zero
                        a(j,j)             = one
                        a(j+1_i4b:n0,j)    = -one
!
                    end do
!
!                   SET RIGHT HANDSIDE.
!
                    b(1_i4b:n0) = one
!
!                   SET EXACT SOLUTION.
!
                    x(1_i4b) = one                
!
                    do j = 2_i4b, n0
                        x(j) = x(j-1_i4b)*two             
                    end do
!
                case( 7_i4b )
!
                    n0 = min( n, maxexp - 1_i4b )
!
!                   TESTS 7 and 8 : WILKINSON'S TEST WITH UPPER TRIANGLE.
!
                    do j = 1_i4b, n0
!
                        a(j,1_i4b:j-1_i4b) = zero
                        a(j,j)             = one
                        a(j,j+1_i4b:n0)    = -one
!
                    end do
!
!                   SET RIGHT HANDSIDE.
!
                    b(1_i4b:n0) = one
!
!                   SET EXACT SOLUTION.
!
                    x(n0) = one                
!
                    do j = n0-1_i4b, 1_i4b, -1_i4b
                        x(j) = x(j+1_i4b)*two             
                    end do
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( x ) ) then
                if ( .not.all( ieee_is_normal( x(1_i4b:n0) ) ) ) cycle
            end if
#endif
!
            if ( (i/2_i4b)*2_i4b/=i ) then
!
!               TESTS 1, 3, 5, 7 : ONE RIGHT HAND-SIDE.
!
                singval(:n0) = singvalues( a(:n0,:n0) )
!
                if ( is_nan( singval(:n0) ) ) cycle
!
                krank0 = count( singval(:n0)/=zero )
!
!               SKIP DEFICIENT MATRICES.
!
                if ( krank0/=n0 ) cycle
!
                normx = norm( x(:n0) )
!
!               COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
                conda = maxval( singval(:n0) )/minval( singval(:n0) )
!
                if ( i==5_i4b .or. i==7_i4b ) then
!
                    condax = normx
!
                else
!
                    tmp =  min( conda, max_conda)
!
                    condax = normx*tmp
!
!                    condax = normx*conda
!
                end if
!
!               COMPUTE THE SOLUTION MATRIX FOR LINEAR SYSTEM
!
!                         a*x = b .
!
                x2(:n0,1_i4b) = solve_lin( a(:n0,:n0), b(:n0) )
!
            else
!
!               TESTS 2, 4, 6, 8 : SEVERAL RIGHT HAND-SIDES.
!
                do j = 1_i4b, nrhs
                    b2(:n0,j) = b(:n0)         
                end do
!
!               COMPUTE THE SOLUTION MATRIX FOR LINEAR SYSTEM
!
!                         a*x = b .
!        
                x2(:n0,:nrhs) = solve_lin( a(:n0,:n0), b2(:n0,:nrhs) )
!
            end if
!       
            if ( (i/2_i4b)*2_i4b==i ) then
                nrhs2 = nrhs
            else
                nrhs2 = 1_i4b
            end if
!
            if ( is_nan( x2(:n0,:nrhs2) ) ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
!           COMPUTE RESIDUALS AND ERRORS.
!
            m = m + 1_i4b
!
            errx = zero
!
            do j = 1_i4b, nrhs2
!
                res(:n0) = x2(:n0,j) - x(:n0)
!
                err  = norm( res(:n0) )
                errx = max( errx, err )
!
            end do
!
            err = errx/condax
!
            eps2 = real( n0, stnd)*eps
!
!           CHECK THE RESULTS FOR SMALL RESIDUALS.
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n0,' , case= ', i,' , nrhs= ', nrhs2
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of the coefficient matrix          = ', conda
!
!                   PRINT ERRORS OF SOLUTIONS.
!
                    write (prtunit,*) 'Accuracy of the computed solutions (absolute error) = ',   &
                                      errx
!
                    write (prtunit,*) 'Accuracy of the computed solutions (relative error) = ',   &
                                      errx/normx
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ', eps2, ' , n= ', n0,' , case= ', i,' , nrhs= ', nrhs2
!
!               PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                write (prtunit,*) 'Condition number of the coefficient matrix          = ', conda
!
!               PRINT ERRORS OF SOLUTIONS.
!
                write (prtunit,*) 'Accuracy of the computed solutions (absolute error) = ',   &
                                  errx
!
                write (prtunit,*) 'Accuracy of the computed solutions (relative error) = ',   &
                                  errx/normx
!
                write (prtunit,*)
!
            end if
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, res, x, x2, b, b2, singval )
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_solve_lin
! ==============================
!
end program test2_solve_lin
