program test2_sym_trid_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SYM_TRID_CMP and SYM_TRID_SOLVE in
!   module LIN_Procedures.
!                                                                              
!
! LATEST REVISION : 26/09/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,    &
                         zero, c100, gen_symtrid_mat,               &
                         sym_trid_cmp, sym_trid_solve, norm, lamch
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
!
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!
    character(len=*), parameter :: name_proc='sym_trid_cmp'
!   
    real(stnd), parameter  :: fudge=c100
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err, eps, eps2, ulp, safmin, normx,  &
                                                 tmp, scale, tol
    real(stnd), allocatable, dimension(:)     :: d, e, sub, diag, sup1, sup2, b, x, res
    real(stnd), allocatable, dimension(:,:)   :: a
!
    integer(i4b)     :: i, l, m, n, success, notdone
    integer          :: iok
!
    logical(lgl)                            :: details
    logical(lgl), allocatable, dimension(:) :: perm
!
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(n,n), d(n), e(n), sub(n), diag(n), sup1(n), sup2(n),   &
                  perm(n), b(n), x(n), res(n),  stat=iok   )
!
        if ( iok/=0 ) cycle
!
        a(:n,:n) = zero
!
        do i = 1_i4b, 50_i4b
!
!           SKIP THE TOO DIFFICULT CASES.
!
            if ( i==11_i4b .or. i==12_i4b .or. i==35_i4b .or. i==36_i4b ) cycle
!
!           GENERATE A TRIDIAGONAL MATRIX.
!           THE DIAGONAL ELEMENTS ARE STORED IN d.
!           THE OFF-DIAGONAL ELEMENTS ARE STORED IN e.
!
!            call gen_symtrid_mat( i, d(:n), e(:n) )
            call gen_symtrid_mat( i, d(:n), e(:n), eigval=diag(:n) )
!
#ifdef _F2003
            if ( ieee_support_datatype( d ) ) then
                if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                if ( .not.all( ieee_is_normal( e(:n) ) ) ) cycle
            end if
#endif
!
            tol = real( n, stnd)*sqrt( eps )
!
            if ( any( abs(diag(:n))<=tol ) ) cycle
!
!           FACTORIZE THE TRIDIAGONAL MATRIX.
!
            call sym_trid_cmp( d, e, sub, diag, sup1, sup2, perm, tol=tol )
!
            if ( perm(n) ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
!           FORM THE TRIDIAGONAL MATRIX.
!
            do l = 1_i4b, n-1_i4b
                a(l,l)       = d(l)
                a(l+1_i4b,l) = e(l)
                a(l,l+1_i4b) = e(l)
            end do
!
            a(n,n) = d(n)
!
!           GENERATE A RANDOM SOLUTION VECTOR x .
!
            call random_number( x(:n) )
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                x(:n) = x(:n)/tmp
            end if
!
            normx = norm( x(:n) )
!
!           COMPUTE THE RIGHT HAND SIDE VECTOR b.
!
            b(:n) = matmul( a, x )
!
!           SOLVE THE TRIDIAGONAL SYSTEM.
!
            call sym_trid_solve( sub, diag, sup1, sup2, perm, b, scale )
!
            if ( scale==zero ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
            m = m + 1_i4b
!
!           CHECK THE RESULTS FOR SMALL RESIDUALS.
!
            res(:n) = x(:n) - b(:n)/scale
!
            err = norm( res(:n) )/normx
!
            eps2 = real( n, stnd)*eps
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n,' , case= ', i
!
!                   PRINT RELATIVE ERROR OF SOLUTION.
!
                    write (prtunit,*) 'Relative error of the computed solution = ', err
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps2,' , n= ', n,' , case= ', i
!
!               PRINT RELATIVE ERROR OF SOLUTION.
!
                write (prtunit,*) 'Relative error of the computed solution = ', err
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, d, e, sub, diag, sup1, sup2,   &
                    perm, b, x, res  )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_sym_trid_cmp
! =================================
!
end program test2_sym_trid_cmp
