program test2_triang_solve
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine TRIANG_SOLVE in module Lin_Procedures. 
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Specific triangular linear problems with known exact solutions are generated for
!    checking the accuracy of subroutine TRIANG_SOLVE.
!
!                                                                              
! LATEST REVISION : 30/11/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, false, zero, one, two, c50, triang_solve, norm, &
                         merror, allocate_error, n1_def, n2_def, maxexp
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=max(n1_def,n2_def), n2=4_i4b, n3=-1_i4b
!
    real(stnd), parameter :: fudge=c50
!
    character(len=*), parameter :: name_proc='triang_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, normx, scal, eps
    real(stnd), allocatable, dimension(:,:) :: a
    real(stnd), allocatable, dimension(:)   :: b, x
!
    integer(i4b) :: i, j, k, n, n0, m, success
    integer      :: iok
!
    logical(lgl) :: details, upper, trans, scaling
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*) 
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( eps )
!
!   ALLOCATE WORK ARRAYS.
!
    n0 = min( n1, maxexp-1_i4b )
!
    allocate( a(n0,n0), x(n0), b(n0), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SET EXACT SOLUTIONS.
!
    x(1_i4b) = one                
!
    do j = 2_i4b, n0
        x(j) = x(j-1_i4b)*two             
    end do
!
    upper   = false
    trans   = false
    scaling = false
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n0, n2, n3
!
        normx = norm( x(:n) )
!
        do i = 1_i4b, 2_i4b
!
            upper   = .not.upper
!
            if ( upper ) then
!
!               TESTS 1, 2, 3, 4 : WILKINSON'S TEST WITH UPPER TRIANGULAR SYSTEM.
!
                do j = 1_i4b, n
!
                    a(j,1_i4b:j-1_i4b) = zero
                    a(j,j)             = one
                    a(j,j+1_i4b:n)     = -one
!
                end do
!
            else
!
!               TESTS 5, 6, 7, 8 : WILKINSON'S TEST WITH LOWER TRIANGULAR SYSTEM.
!
                do j = 1_i4b, n
!
                    a(1_i4b:j-1_i4b,j) = zero
                    a(j,j)             = one
                    a(j+1_i4b:n,j)     = -one
!
                end do
!
            end if
!
            do j = 1_i4b, 2_i4b
!
                trans = .not.trans
!
                do k = 1_i4b, 2_i4b
!
                    scaling = .not.scaling
!
!                   RESET RIGHT HANDSIDE.
!
                    b(1_i4b:n) = one
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR TRIANGULAR SYSTEM
!
!                       a*x(:)=b(:) OR a**t*x(:)=b(:) .
!
                    if ( scaling ) then
!
                        call triang_solve( a(:n,:n), b(:n), scal=scal, upper=upper, trans=trans )
!
                    else
!
                        call triang_solve( a(:n,:n), b(:n), upper=upper, trans=trans )
!
                        scal = one
!
                    end if
!
                    m = m + 1_i4b
!
!                   COMPUTE ERRORS FOR SOLUTION.
!       
                    if ( upper.eqv.trans ) then
                        b(:n) = b(:n)/scal - x(:n)
                    else
                        b(:n) = b(:n)/scal - x(n:1_i4b:-1_i4b)
                    end if
!
                    err = norm( b(:n) )
!
!                   CHECK THE RESULTS FOR EXACT SOLUTION.
!
                    if ( err/normx<=eps ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            write (prtunit,*) 'n= ', n, ' ,  upper= ', upper, ' , trans= ', trans, &
                                              ' , scaling= ', scaling
!
!                           PRINT ERRORS OF SOLUTIONS.
!
                            write (prtunit,*) 'Accuracy of the computed solutions (absolute error) = ',   &
                                              err
!
                            write (prtunit,*) 'Accuracy of the computed solutions (relative error) = ',   &
                                              err/normx
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        write (prtunit,*) '=> ',err/normx,' > ', eps, ' , n= ', n, ' , upper= ', upper, &
                                      ' , trans= ', trans, ' , scaling= ', scaling
!
!                       PRINT ERRORS OF SOLUTIONS.
!
                        write (prtunit,*) 'Accuracy of the computed solutions (absolute error) = ',   &
                                          err
!
                        write (prtunit,*) 'Accuracy of the computed solutions (relative error) = ',   &
                                          err/normx
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, x, b )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_triang_solve
! =================================
!
end program test2_triang_solve
