program test2_trid_deflate
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine TRID_DEFLATE in module Eig_Procedures.
!                                                                              
!
! LATEST REVISION : 31/07/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,                     &
                         true, false, zero, one, ten, c50, gen_symtrid_mat, lamch,   &
                         symtrid_bisect, trid_deflate, norm
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, neig_def=5
!   
    real(stnd), parameter :: fudge=c50
!
    character(len=*), parameter :: name_proc='trid_deflate'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err1, err2, err, eps, eps0, eps2, ulp, safmin,  &
                                                 abstol, norma, pertol, sep, prec
    real(stnd), allocatable, dimension(:)     :: d, e, eigval
    real(stnd), allocatable, dimension(:,:)   :: a, eigvec, res
!
    integer(i4b)     :: i, j, k, l, m, n, neig, success, notdone, max_qr_steps=4
    integer          :: iok
!
    logical(lgl)     :: failure1, failure2, details, ortho
!   
    character        :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp  = epsilon( err )
    eps  = fudge*ulp
    eps0 = sqrt( ulp )
!    eps0 = eps
!
    safmin = lamch('s')
    abstol = sqrt( safmin )
    prec = ten*lamch( 'p' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( d(n), e(n), eigval(n), eigvec(n,n),    &
                  a(n,n), res(n,n), stat=iok   )
!
        if ( iok/=0 ) then
!
            if ( allocated(d) )        deallocate( d )
            if ( allocated(e) )        deallocate( e )
            if ( allocated(eigval) )   deallocate( eigval )
            if ( allocated(eigvec) )   deallocate( eigvec )
            if ( allocated(a) )        deallocate( a )
            if ( allocated(res) )      deallocate( res )
!
            cycle
!
        end if
!
        a(:n,:n) = zero
!
        do i = 1_i4b, 50_i4b
!
!           SKIP GRADED MATRICES WITH A GEOMETRIC PROGRESSION TO AVOID
!           OVERFLOW PROBLEMS IN SINGLE PRECISION.
!
            if ( i==35_i4b .or. i==36_i4b ) cycle
!
!           GENERATE A TRIDIAGONAL MATRIX.
!           THE DIAGONAL ELEMENTS ARE STORED IN d.
!           THE OFF-DIAGONAL ELEMENTS ARE STORED IN e.
!
            call gen_symtrid_mat( i, d(:n), e(:n) )
!
#ifdef _F2003
            if ( ieee_support_datatype( d ) ) then
                if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                if ( .not.all( ieee_is_normal( e(:n) ) ) ) cycle
            end if
#endif
!
!           COMPUTE ALL EIGENVALUES OF THE TRIDIAGONAL MATRIX BY A BISECTION METHOD WITH HIGH ACCURACY.
!           THE EIGENVALUES ARE COMPUTED WITH MAXIMUM ACCURACY WHEN OPTIONAL PARAMETER abstol IS SET TO
!           sqrt( lamch('s') ), WHICH IS THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
            call symtrid_bisect( d(:n), e(:n), l, eigval(:n), failure1,   &
                                 sort=sort, le=n, abstol=abstol )
!
!           COMPUTE THE NORM OF THE TRIDIAGONAL MATRIX.
!
            norma = norm( eigval(:n) )
!
            if ( failure1 ) cycle
!
!           COMPUTE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
            do k = 1_i4b, 2_i4b
!
                if ( k==1_i4b ) then
!
!                    neig = min( n, neig_def )
!
                   l = min( n, neig_def )
!
                   do neig = l, n-1_i4b
!
                       pertol = abs( prec*eigval(neig) )
                       sep    = abs( eigval(neig) - eigval(neig+1_i4b) )
!
                       if ( sep>=pertol ) exit
!
                   end do
!
                else
!
                    neig = n
!
                end if
!
                do j = 1_i4b, 2_i4b
!
                    if ( j==1_i4b ) then
!
                        eps2 = eps
!
                        ortho = true
!
                    else
!
                        eps2 = eps0
!
                        ortho = false
!
                    end if
!
!                   COMPUTE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
                    call trid_deflate( d(:n), e(:n), eigval(:neig),           &
                                       eigvec(:n,:neig), failure2,            &
                                       ortho=ortho, max_qr_steps=max_qr_steps )
!
                    if ( failure2 ) then
!
                        notdone = notdone + 1_i4b
!
                        cycle
!
                    end if
!
                    m = m + 1_i4b
!
!                   FORM THE TRIDIAGONAL MATRIX .
!
                    do l = 1_i4b, n-1_i4b
                        a(l,l)       = d(l)
                        a(l+1_i4b,l) = e(l)
                        a(l,l+1_i4b) = e(l)
                    end do
!
                    a(n,n) = d(n)
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*diag(s)
!                   WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
                    res(:n,:neig) = matmul( a(:n,:n), eigvec(:n,:neig) )        &
                                    - eigvec(:n,:neig)*spread(eigval(:neig),1,n)
!
                    if ( norma/=zero ) then
                        err1 = maxval( norm(res(:n,:neig), dim=2_i4b) )/( norma*real(n,stnd) )
                    else
                        err1 = maxval( norm(res(:n,:neig), dim=2_i4b) )/real(n,stnd)
                    end if
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                   WHERE u ARE THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
                    res(:neig,:neig) = matmul( transpose( eigvec(:n,:neig) ), eigvec(:n,:neig) )
!
                    do l = 1_i4b, neig
                        res(l,l) = one - res(l,l)
                    end do
!
                    err2 = maxval( norm(res(:neig,:neig), dim=2_i4b) )/real(n,stnd)
!
                    if ( err1<=eps .and. err2<=eps2 ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            write (prtunit,*) 'n= ', n,' , neig= ', neig,' , case= ', i,   &
                                              ' , ortho= ', ortho
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*diag(s)
!                           WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE
!                           TRIDIAGONAL MATRIX.
!
                            write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', &
                                              err1/ulp, ' (must be less than :', fudge, ' )'
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                           WHERE u are EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
                            write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', &
                                              err2/ulp, ' (must be less than :', fudge, ' )'
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        if ( j==1_i4b ) then
!
                            err = max( err1, err2 )
!
                            write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,             &
                                              ' , neig= ', neig,' , case= ', i,            &
                                              ' , ortho= ', ortho,' , failure= ', failure2
!
                        else
!
                            write (prtunit,*) '=> ',err1,' > ',eps,' or ', err2,' > ',eps2, &
                                              ' , n= ', n,' , neig= ', neig,' , case= ', i, &
                                              ' , ortho= ', ortho,' , failure= ', failure2
!
                        end if
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*diag(s)
!                       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE
!                       TRIDIAGONAL MATRIX.
!
                        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                       WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, a, res, eigval, eigvec )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_trid_deflate
! =================================
!
end program test2_trid_deflate
