program test_apply_q_bd
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines APPLY_Q_BD and APPLY_P_BD
!   in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutines APPLY_Q_BD and APPLY_P_BD.
!
! LATEST REVISION : 21/05/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,    &
                         true, false, zero, half, ten,              &
                         bd_cmp, bd_svd, apply_q_bd, apply_p_bd,    &
                         unit_matrix, norm, lamch
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter :: fudge=ten
!
    character(len=*), parameter :: name_proc='apply_q_bd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, ulp, safmin, tmp, norma
    real(stnd), allocatable, dimension(:)   :: d, e, taup, tauq
    real(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, leftvec2, rightvec, work, proj
!
    integer      :: iok
    integer(i4b) :: i, k, m, n, nr, nc, success, notdone
!
    logical(lgl) :: failure, details, upper, left, trans
!   
    character :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( d(n), e(n), tauq(n), taup(n), stat=iok )
!
        if ( iok/=0 ) cycle
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
            case( 1_i4b )
!
!               TEST 1: a(:nr,:nc) WITH nr=nc .
!
                nr = n
                nc = n
!
            case( 2_i4b )
!
!               TEST 2: a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                nr = n + 1_i4b
                nc = n
!
            case( 3_i4b )
!
!               TEST 3: a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                nr = n + n
                nc = n
!
            case( 4_i4b )
!
!               TEST 4: a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                nr = n
                nc = n + 1_i4b
!
            case( 5_i4b )
!
!               TEST 5: a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                nr = n
                nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            i = max( nr, nc)
!
            allocate( a(nr,nc), a2(nr,nc), proj(nr,n), &
                 leftvec(nr,n), rightvec(nc,n),        &
                 leftvec2(nr,n), work(i,i), stat=iok   )
!
            if ( iok/=0 ) cycle
!
!           GENERATE A RANDOM COEFFICIENT MATRIX a(:nr,:nc) .
!
            call random_number( a(:nr,:nc) )
!
            a(:nr,:nc) = a(:nr,:nc) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                a(:nr,:nc) = a(:nr,:nc)/tmp
            end if
!
!           COMPUTE THE NORM OF THE MATRIX.
!
            norma = norm( a(:nr,:nc) )
!
!           MAKE A COPY THE ORIGINAL MATRIX.
!
            a2(:nr,:nc) = a(:nr,:nc)
!
!           CALL bd_cmp AND ortho_gen_bd TO REDUCE THE MATRIX TO BIDIAGONAL FORM
!
!                             a = Q*BD*P**(t)
!
!           WHERE Q AND P ARE ORTHOGONAL AND BD IS AN UPPER OR LOWER BIDIAGONAL MATRIX.
!
            call bd_cmp( a2(:nr,:nc), d(:n), e(:n), tauq(:n), taup(:n) )
!
!           COMPUTE SINGULAR VALUES AND SINGULAR VECTORS OF THE BIDIAGONAL MATRIX.
!
            call unit_matrix( leftvec(:nr,:n)  )
            call unit_matrix( rightvec(:nc,:n) )
!
            upper = nr>=nc
!
            call bd_svd( upper, d(:n), e(:n), failure, leftvec(:nr,:n),  &
                         rightvec(:nc,:n), sort=sort )
!
            if ( failure ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
!           MULTIPLY LEFT SINGULAR VECTORS BY SINGULAR VALUES.
!
            leftvec(:nr,:n) = leftvec(:nr,:n)*spread(d(:n),1,nr)
!
            do i = 1_i4b, 4_i4b
!
                m = m + 1_i4b
!
                select case( i )
!
                case( 1_i4b )
!
!                   TESTING apply_q_bd AND apply_p_bd SUBROUTINES WITH left=true AND trans=false.
!
                    left  = true
                    trans = false
!
                    work(:nc,:n) = rightvec(:nc,:n)
!
                    call apply_p_bd( a2(:nr,:nc), taup(:n), work(:nc,:n), left=true, trans=false )
!
                    proj(:nr,:n) = matmul( a(:nr,:nc), work(:nc,:n) )
!
                    leftvec2(:nr,:n) = leftvec(:nr,:n)
!
                    call apply_q_bd( a2(:nr,:nc), tauq(:n), leftvec2(:nr,:n), left=true, trans=false )
!
                case( 2_i4b )
!
!                   TESTING apply_q_bd AND apply_p_bd SUBROUTINES WITH left=false AND trans=false.
!
                    left  = false
                    trans = false
!
                    work(:nr,:nc) = a(:nr,:nc)
!
                    call apply_p_bd( a2(:nr,:nc), taup(:n), work(:nr,:nc), left=false, trans=false )
!
                    proj(:nr,:n) = matmul( work(:nr,:nc), rightvec(:nc,:n) )
!
                    call unit_matrix( work(:nr,:nr) )
!
                    call apply_q_bd( a2(:nr,:nc), tauq(:n), work(:nr,:nr), left=false, trans=false )
!
                    leftvec2(:nr,:n) = matmul( work(:nr,:nr), leftvec(:nr,:n) )
!
                case( 3_i4b )
!
!                   TESTING apply_q_bd AND apply_p_bd SUBROUTINES WITH left=true AND trans=true.
!
                    left  = true
                    trans = true
!
                    work(:nc,:nr) = transpose( a(:nr,:nc) )
!
                    call apply_p_bd( a2(:nr,:nc), taup(:n), work(:nc,:nr), left=true, trans=true )
!
                    proj(:nr,:n) = matmul( transpose(work(:nc,:nr)), rightvec(:nc,:n) )
!
                    call unit_matrix( work(:nr,:nr) )
!
                    call apply_q_bd( a2(:nr,:nc), tauq(:n), work(:nr,:nr), left=true, trans=true )
!
                    leftvec2(:nr,:n) = matmul( transpose(work(:nr,:nr)), leftvec(:nr,:n) )
!
                case( 4_i4b )
!
!                   TESTING apply_q_bd AND apply_p_bd SUBROUTINES WITH left=false AND trans=true.
!
                    left  = false
                    trans = true
!
                    work(:n,:nc) = transpose( rightvec(:nc,:n) )
!
                    call apply_p_bd( a2(:nr,:nc), taup(:n), work(:n,:nc), left=false, trans=true )
!
                    proj(:nr,:n) = matmul( a(:nr,:nc), transpose(work(:n,:nc)) )
!
                    work(:n,:nr) = transpose( leftvec(:nr,:n) )
!
                    call apply_q_bd( a2(:nr,:nc), tauq(:n), work(:n,:nr), left=false, trans=true )
!
                    leftvec2(:nr,:n) = transpose( work(:n,:nr) )
!
                end select
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!               WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!               v THE RIGHT SINGULAR VECTORS OF THE RANDOM MATRIX.
!
                work(:nr,:n) = proj(:nr,:n) - leftvec2(:nr,:n)
!
                if ( norma/=zero ) then
                    err = maxval( norm(work(:nr,:n), dim=2_i4b) )/( norma*real(n,stnd) )
                else
                    err = maxval( norm(work(:nr,:n), dim=2_i4b) )/real(n,stnd)
                end if
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr, ' , nc= ', nc,  &
                             ' , left= ', left,' , trans= ', trans
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                       WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                       v THE RIGHT SINGULAR VECTORS OF THE RANDOM MATRIX.
!
                        write (prtunit,*) 'Accuracy of the computed singular triplets = ', &
                                          err/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ', err,' > ', eps, ' , nr= ', nr, &
                         ' , nc= ', nc, ' , left= ', left, ' , trans= ', trans
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                   WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                   v THE RIGHT SINGULAR VECTORS OF THE RANDOM MATRIX.
!
                    write (prtunit,*) 'Accuracy of the computed singular triplets = ', &
                         err/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, a2, work, proj, leftvec, leftvec2, rightvec )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, tauq, taup )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_apply_q_bd
! ==============================
!
end program test_apply_q_bd
