program test_bd_singval2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine BD_SINGVAL2 in module SVD_Procedures.
!                                                                              
!
! LATEST REVISION : 01/08/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,    &
                         true, false, zero, one, c50, gen_bd_mat,   &
                         lamch, bd_singval2
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter :: fudge=c50
!
    character(len=*), parameter :: name_proc='bd_singval2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err1, err2, eps, eps1, eps2, ulp, safmin, abstol
    real(stnd), allocatable, dimension(:) :: d, e, singval0, singval, temp
!
    integer(i4b)     :: i, j, l, m, n, success, notdone
    integer          :: iok
!
    logical(lgl)     :: failure, details, high_accuracy, test_ok
!   
    character        :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err1 )
    eps = fudge*ulp
!
    safmin = lamch('s')
    abstol = sqrt( safmin )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( d(n), e(n), singval0(n), singval(n), temp(n), stat=iok   )
!
        if ( iok/=0 ) then
!
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(singval0) )  deallocate( singval0 )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(temp) )      deallocate( temp )
!
            cycle
!
        end if
!
        do i = 1_i4b, 17_i4b
!
!           GENERATE A BIDIAGONAL MATRIX.
!           THE DIAGONAL ELEMENTS ARE STORED IN d.
!           THE OFF-DIAGONAL ELEMENTS ARE STORED IN e.
!           THE SINGULAR VALUES ARE STORED IN singval0.
!
            call gen_bd_mat( i, d(:n), e(:n), singval=singval0(:n), sort=sort, failure=failure )
!
#ifdef _F2003
            if ( ieee_support_datatype( d ) ) then
!
                if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                if ( .not.all( ieee_is_normal( e(:n) ) ) ) cycle
!
            end if
#endif
!
            if ( failure ) cycle
!
            do j = 1_i4b, 2_i4b
!
                if ( j==1_i4b ) then
!
                    high_accuracy = true
!
!                   COMPUTE ALL SINGULAR VALUES OF THE BIDIAGONAL MATRIX BY A BISECTION METHOD WITH HIGH ACCURACY.
!                   THE SINGULAR VALUES ARE COMPUTED WITH MAXIMUM ACCURACY WHEN OPTIONAL PARAMETER abstol IS SET TO
!                   sqrt( lamch('s') ), WHICH IS THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
                    call bd_singval2( d(:n), e(:n), l, singval(:n), failure, sort=sort, ls=n, abstol=abstol )
!
                else
!
                    high_accuracy = false
!
!                   COMPUTE ALL SINGULAR VALUES OF THE BIDIAGONAL MATRIX BY A BISECTION METHOD WITH ABSOLUTE ACCURACY.
!
                    call bd_singval2( d(:n), e(:n), l, singval(:n), failure, sort=sort, ls=n )
!
                end if
!
                if ( failure ) then
!
                    notdone = notdone + 1_i4b
!
                    cycle
!
                end if
!
                m = m + 1_i4b
!
!               NOW, TEST ACCURACY OF THE RESULTS.
!
                where( singval0(:n)/=zero )
                    temp(:n) = singval0(:n)
                elsewhere
                    temp(:n) = one
                end where
!
!               ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                err1 = maxval( abs( singval(:n) - singval0(:n) ) )
!
!               RELATIVE ERRORS OF SINGULAR VALUES.
!
                err2 = maxval( abs( (singval(:n) - singval0(:n))/temp(:n) ) )
!
                eps1 = real( n, stnd)*eps*maxval( singval0(:n) )
!
                test_ok = err1<=eps1
!
                if ( high_accuracy ) then
!
                    eps2 = real( n, stnd)*eps
!
                    test_ok = test_ok .and. err2<=eps2
!
                end if
!
                if ( test_ok ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
                        write (prtunit,*) 'n= ', n,' , case= ', i
!
                        write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', err1
                        write (prtunit,*) 'Relative accuracy of the computed singular values    = ', err2
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    if ( high_accuracy ) then
!
                        write (prtunit,*) '=> ',err1,' > ',eps1,' or ', err2,' > ',eps2,' , n= ', n,              &
                             ' , case= ', i,' , high_accuracy= ', high_accuracy,' , failure= ', failure
!
                    else
!
                        write (prtunit,*) '=> ',err1,' > ',eps1,' , n= ', n,' , case= ', i,              &
                             ' , high_accuracy= ', high_accuracy,' , failure= ', failure
!
                    end if
!
                    write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', err1
                    write (prtunit,*) 'Relative accuracy of the computed singular values    = ', err2
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, singval0, singval, temp )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_bd_singval2
! ===============================
!
end program test_bd_singval2
