program test_comp_composite
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine comp_composite in module Stat_Procedures.
!                                                                              
! LATEST REVISION : 08/04/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, comp_composite, zero, merror, &
                         allocate_error, n1_def, n2_def
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, p IS THE NUMBER OF OBSERVATIONS, xmiss IS THE FLAG FOR MISSING VALUES.
!
    integer(i4b), parameter :: prtunit=6, ngrp=2, nobs_grp=20, n=n1_def, m=n2_def, p=ngrp*nobs_grp+5
!
    real(stnd), parameter :: xmiss=-999.99_stnd
!
    character(len=*), parameter :: name_proc='comp_composite'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err_mean, err_std, err_comp, err_mean_grp,    &
                                                 err_std_grp, err, eps
    real(stnd), allocatable, dimension(:,:)   :: xmean, xmean1, xmean2, xstd, xstd1, xstd2, xn
    real(stnd), allocatable, dimension(:,:,:) :: xmean_grp, xmean_grp1, xmean_grp2,             &
                                                 xstd_grp, xstd_grp1, xstd_grp2,                &
                                                 xn_grp, xcomp, xcomp1, xcomp2, x1, x2, tmp
!
    integer                   :: iok
    integer(i4b)              :: i, j, k, l, i1, i2, inc, success
    integer(i4b), allocatable :: ind(:), cnt(:,:,:)
!
    logical(lgl)                                :: first, last
    logical(lgl), allocatable, dimension(:,:,:) :: mask, tmp_mask
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( xmean(n,m), xmean1(n,m), xmean2(n,m),                             &
              xstd(n,m), xstd1(n,m), xstd2(n,m), xn(n,m),                       &
              xmean_grp(n,m,ngrp), xmean_grp1(n,m,ngrp), xmean_grp2(n,m,ngrp),  &
              xstd_grp(n,m,ngrp), xstd_grp1(n,m,ngrp), xstd_grp2(n,m,ngrp),     &
              xcomp(n,m,ngrp), xcomp1(n,m,ngrp), xcomp2(n,m,ngrp),              &
              x1(n,m,p), x2(n,m,p), tmp(n,m,nobs_grp),                          &
              xn_grp(n,m,ngrp), ind(p), cnt(n,m,ngrp), mask(n,m,p),             &
              tmp_mask(n,m,nobs_grp), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM OBSERVATION ARRAY x1 .
!
    call random_number( x1 )
!
!   COMPUTE THE MEANS AND STANDARD-DEVIATIONS OF x1.
!
    xmean1 = sum( x1, dim=3_i4b )/p
    xstd1  = sqrt( sum( (x1 - spread( xmean1, dim=3_i4b, ncopies=p ) )**2, dim=3_i4b )/p )
!
!   GENERATE A RANDOM OBSERVATION ARRAY x2 WITH MISSING VALUES.
!
    mask = x1<=0.15_stnd
    where( mask )
       x2 = xmiss
    elsewhere
       x2 = x1
    end where
!
!   COMPUTE THE MEANS AND STANDARD-DEVIATIONS OF x2.
!
    mask = .not.mask
    xn = count( mask, dim=3_i4b )
    xmean2 = sum( x2, dim=3_i4b, mask=mask )/xn
    xstd2  = sqrt( sum( (x2 - spread( xmean2, dim=3_i4b, ncopies=p ))**2, dim=3_i4b, mask=mask )/xn )
!
!   GENERATE GROUPS OF OBSERVATIONS.
!   COMPUTE THE MEANS OF x1 BY GROUPS.
!   COMPUTE THE STANDARD-DEVIATIONS OF x1 BY GROUPS.
!   COMPUTE THE COMPOSITES OF x1.
!   COMPUTE THE MEANS OF x2 BY GROUPS.
!   COMPUTE THE STANDARD-DEVIATIONS OF x2 BY GROUPS.
!   COMPUTE THE COMPOSITES OF x2.
!
    ind = 0_i4b
    i1 = 1_i4b
    i2 = 0_i4b
!
    do i = 1_i4b, ngrp
!
        i2 = i2 + nobs_grp
!
        ind(i1:i2) = i
        tmp_mask(:,:,:) = mask(:,:,i1:i2)
        cnt(:,:,i) = count( tmp_mask, dim=3_i4b )
!
        tmp(:,:,:) = x1(:,:,i1:i2)
        xmean_grp1(:,:,i) = sum( tmp, dim=3_i4b )/nobs_grp
        tmp(:,:,:) = (tmp(:,:,:)-spread( xmean_grp1(:,:,i),dim=3_i4b,ncopies=nobs_grp))**2
        xstd_grp1(:,:,i)  = sqrt( sum( tmp, dim=3_i4b )/nobs_grp )
        xcomp1(:,:,i) = ( xmean_grp1(:,:,i) - xmean1 )/xstd1
!
        where ( tmp_mask(:,:,:) )
            tmp(:,:,:) = x2(:,:,i1:i2)
        elsewhere
            tmp(:,:,:) = zero
        end where
!
        xmean_grp2(:,:,i) = sum( tmp, dim=3_i4b )/cnt(:,:,i)
!
        where ( tmp_mask(:,:,:) )
            tmp(:,:,:) = (tmp(:,:,:)-spread( xmean_grp2(:,:,i),dim=3_i4b,ncopies=nobs_grp))**2
        elsewhere
            tmp(:,:,:) = zero
        end where
!
        xstd_grp2(:,:,i)  =  sqrt( sum( tmp, dim=3_i4b ) /cnt(:,:,i) )
!
        xcomp2(:,:,i) = ( xmean_grp2(:,:,i) - xmean2 )/xstd2
!
        i1 = i1 + nobs_grp
!
    end do
!
!   RECOMPUTE THE STATISTICS WITH comp_composite SUBROUTINE.
!
    success = 0_i4b
!
    do i = 1_i4b, 12_i4b
!
        inc = merge( 1_i4b, p, i<=6_i4b )
!
        select case( i )
!
            case( 1_i4b, 7_i4b )
!
!               TESTS 1 AND 7 :
!
!               COMPUTE THE COMPOSITES OF x1 WITH comp_composite_rt .
!
                i2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    i2 = min( i2+inc, p )
!
                    first = k==1_i4b
                    last = i2==p
!
                    call comp_composite( x1(:,:,k:i2), first, last, ngrp, ind(k:i2),   &
                                         xmean, xstd, xn(1_i4b,1_i4b),                 &
                                         xmean_grp, xstd_grp, xn_grp(1_i4b,1_i4b,:),   &
                                         xcomp=xcomp  ) 
                end do
!
            case( 2_i4b, 8_i4b )
!
!               TESTS 2 AND 8 :
!
!               COMPUTE THE COMPOSITES OF x2 WITH comp_composite_miss_rt .
!
                i2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    i2 = min( i2+inc, p )
!
                    first = k==1_i4b
                    last = i2==p
!
                    call comp_composite( x2(:,:,k:i2), first, last, ngrp, ind(k:i2),   &
                                         xmean, xstd, xn,                              &
                                         xmean_grp, xstd_grp, xn_grp,                  &
                                         xmiss=xmiss, xcomp=xcomp  ) 
                end do
!
            case( 3_i4b, 9_i4b )
!
!               TESTS 3 AND 9 :
!
!               COMPUTE THE COMPOSITES OF x1 WITH comp_composite_rm .
!
                do j = 1_i4b, m
!
                    i2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        i2 = min( i2+inc, p )
!
                        first = k==1_i4b
                        last = i2==p
!
                        call comp_composite( x1(:,j,k:i2), first, last, ngrp, ind(k:i2),                &
                                             xmean(:,j), xstd(:,j), xn(1_i4b,1_i4b),                    &
                                             xmean_grp(:,j,:), xstd_grp(:,j,:), xn_grp(1_i4b,1_i4b,:),  &
                                             xcomp=xcomp(:,j,:)  ) 
                    end do
!
                end do
!
            case( 4_i4b, 10_i4b )
!
!               TESTS 4 AND 10 :
!
!               COMPUTE THE COMPOSITES OF x2 WITH comp_composite_miss_rm .
!
                do j = 1_i4b, m
!
                    i2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        i2 = min( i2+inc, p )
!
                        first = k==1_i4b
                        last = i2==p
!
                        call comp_composite( x2(:,j,k:i2), first, last, ngrp, ind(k:i2),         &
                                             xmean(:,j), xstd(:,j), xn(:,j),                     &
                                             xmean_grp(:,j,:), xstd_grp(:,j,:), xn_grp(:,j,:),   &
                                             xmiss=xmiss, xcomp=xcomp(:,j,:)  ) 
                    end do
!
                end do
!
            case( 5_i4b, 11_i4b )
!
!               TESTS 5 AND 11 :
!
!               COMPUTE THE COMPOSITES OF x1 WITH comp_composite_rv .
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        i2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            i2 = min( i2+inc, p )
!
                            first = k==1_i4b
                            last = i2==p
!
                            call comp_composite( x1(l,j,k:i2), first, last, ngrp, ind(k:i2),                &
                                                 xmean(l,j), xstd(l,j), xn(1_i4b,1_i4b),                    &
                                                 xmean_grp(l,j,:), xstd_grp(l,j,:), xn_grp(1_i4b,1_i4b,:),  &
                                                 xcomp=xcomp(l,j,:)  ) 
                        end do
!
                    end do
!
                end do
!
            case( 6_i4b, 12_i4b )
!
!               TESTS 6 AND 12 :
!
!               COMPUTE THE COMPOSITES OF x2 WITH comp_composite_miss_rv .
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        i2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            i2 = min( i2+inc, p )
!
                            first = k==1_i4b
                            last = i2==p
!
                            call comp_composite( x2(l,j,k:i2), first, last, ngrp, ind(k:i2),       &
                                                 xmean(l,j), xstd(l,j), xn(l,j),                   &
                                                 xmean_grp(l,j,:), xstd_grp(l,j,:), xn_grp(l,j,:), &
                                                 xmiss=xmiss, xcomp=xcomp(l,j,:)  ) 
                        end do
!
                    end do
!
                end do
!
        end select
!
!       CHECK THAT THE TWO SETS OF STATISTICS AGREE.
!
        if ( mod(i,2_i4b)==1_i4b ) then
 !
            err_mean = maxval( abs( (xmean-xmean1)/xmean ) )
            err_std  = maxval( abs( (xstd-xstd1)/xstd ) )
            err_comp = maxval( abs( xcomp - xcomp1 ) )
            err_mean_grp = maxval( abs( (xmean_grp-xmean_grp1)/xmean_grp ) )
            err_std_grp = maxval( abs( (xstd_grp-xstd_grp1)/xstd_grp ) )
!
        else
!
            err_mean = maxval( abs( (xmean-xmean2)/xmean ) )
            err_std  = maxval( abs( (xstd-xstd2)/xstd ) )
            err_comp = maxval( abs( xcomp - xcomp2 ) )
            err_mean_grp  = maxval( abs( (xmean_grp-xmean_grp2)/xmean_grp ) )
            err_std_grp = maxval( abs( (xstd_grp-xstd_grp2)/xstd_grp ) )
!
        end if
!
!       PRINT THE RESULTS OF THE TESTS.
!
        err = max( err_mean, err_std, err_comp, err_mean_grp, err_std_grp )
!
        if ( err<=eps ) then
!
            success = success + 1_i4b
!
!            write (prtunit,*) 'Test ',i,' is correct'
!
        else
!
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ', err,' > ', eps
!
            write (prtunit,*)
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( xmean, xmean1, xmean2, xstd, xstd1, xstd2,  &
                xn, xmean_grp, xmean_grp1, xmean_grp2,      &
                xstd_grp, xstd_grp1, xstd_grp2,             &
                xcomp, xcomp1, xcomp2, x1, x2, tmp,         &
                xn_grp, ind, cnt, mask, tmp_mask            )
!
    if ( success/=12_i4b ) then
        write (prtunit,*) 12_i4b - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_composite
! ==================================
!
end program test_comp_composite
