program test_comp_inv
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_INV in module Lin_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking the accuracy
!    of subroutine COMP_INV.
!
! LATEST REVISION : 30/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,                   &
                         true, false, zero, half, one, two, five, ten, c50, c100,  &
                         qr_cmp, ortho_gen_qr, comp_inv, arth, lamch, norm
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c50, max_conda=1000._stnd
!
    character(len=*), parameter :: name_proc='comp_inv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, ulp, safmin, tmp,   &
                                               shift, conda
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta, norma
    real(stnd), allocatable, dimension(:,:) :: a, ainv, ainv2, leftvec, rightvec, res
!
    integer(i4b) :: i, j, l, n, m, krank0, success, notdone
    integer      :: iok
!
    logical(lgl) :: failure, in_place, details
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    safmin = lamch( 'S' )
!
!   SPECIFY EVENTUALLY A SHIFT TO AVOID ZERO SINGULAR VALUE.
!
    shift = 0.04_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), diagr(n),          &
                  beta(n), norma(n),  a(n,n), ainv(n,n),    &
                  ainv2(n,n), leftvec(n,n),  rightvec(n,n), &
                  res(n,n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(norma) )     deallocate( norma )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(ainv) )      deallocate( ainv )
            if ( allocated(ainv2) )     deallocate( ainv2 )
            if ( allocated(leftvec) )   deallocate( leftvec )
            if ( allocated(rightvec) )  deallocate( rightvec )
            if ( allocated(res) )       deallocate( res )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:n,:n) .
!
        call random_number( leftvec(:n,:n) )
!
        leftvec(:n,:n) = leftvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            leftvec(:n,:n) = leftvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( leftvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( leftvec(:n,:n), beta(:n) )
!
!       GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:n,:n) .
!
        call random_number( rightvec(:n,:n) )
!
        rightvec(:n,:n) = rightvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            rightvec(:n,:n) = rightvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( rightvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( rightvec(:n,:n), beta(:n) )
!
!       COMPUTE FIRST THE EIGENVALUES OF THE MATRIX PRODUCT
!                     a(:n,:n)**(t)*a(:n,:n).
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
            case( 1_i4b )
!
!               TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b)   = one
                eigval(2_i4b:n) = ulp
!
            case( 2_i4b )
!
!               TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = one
                eigval(n)             = ulp
!
            case( 3_i4b )
!
!               TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                eigval(n)             = one
!
            case( 4_i4b )
!
!               TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                eigval(1_i4b)         = ulp
                eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                eigval(n)             = two
!
            case( 5_i4b )
!
!               TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
            case( 6_i4b )
!
!               TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                err = real( n-1_i4b, stnd )
                do l = 1_i4b, n
                    eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                end do
!
            case( 7_i4b )
!
!               TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
            case( 8_i4b )
!
!               TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
            case( 9_i4b )
!
!               TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
            case( 10_i4b )
!
!               TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
            case( 11_i4b )
!
!               TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                call random_number( eigval(:n) )
!
                eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
            case( 12_i4b )
!
!               TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
            case( 13_i4b )
!
!               TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 14_i4b )
!
!               TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 15_i4b )
!
!               TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case default
!
!               DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                call random_number( eigval(:n) )
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
!           COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
            singval(:n) = sqrt( eigval(:n) )
!            singval(:n) = eigval(:n)
!
            singval(:n) = singval(:n) + shift
!
            krank0 = count( singval(:n)/=zero )
!
!           SKIP DEFICIENT MATRICES.
!
            if ( krank0/=n ) cycle
!
!           FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
            a(:n,:n) = matmul( leftvec(:n,:n)*spread(singval,1,n), transpose(rightvec(:n,:n)) )
!
!           FORM INVERSE OF RANDOM MATRIX.
!
            eigval(:n) = one/singval(:n)
!
            ainv(:n,:n) = matmul( rightvec(:n,:n)*spread(eigval,1,n), transpose(leftvec(:n,:n)) )
!
!           COMPUTE THE NORMS OF THE COLUMNS OF INVERSE OF RANDOM MATRIX.
!
            norma(:n) = norm( ainv(:n,:n), dim=2_i4b )
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            conda = maxval( singval(:n) )/minval( singval(:n) )
!
!            conda = ( sum( singval(:n) )*sum( one/singval(:n) ) )/real( n, stnd)
!
            tmp =  min( conda, max_conda )
!
            beta(:n) = norma(:n)*tmp
!
            do i = 1_i4b, 2_i4b
!
                select case( i )
!
                case( 1_i4b )
!
!                   TEST 1 : IN-PLACE INVERSION.
!
                    ainv2(:n,:n) = a(:n,:n)
!
                    call comp_inv( ainv2(:n,:n), failure )
!
                    in_place = true
!
                case( 2_i4b )
!
!                   TEST 2 : OUT-OF-PLACE INVERSION.
!
                    call comp_inv( a(:n,:n), failure, ainv2(:n,:n) )
!
                    in_place = false
!
                end select
!
                if ( failure ) then
!
                    notdone = notdone + 1_i4b
!
                    cycle
!
                end if
!
!               CHECK THE RESULTS FOR SMALL RESIDUALS.
!
                res(:n,:n) = ainv2(:n,:n) - ainv(:n,:n)
!
                err = maxval( norm(res(:n,:n), dim=2_i4b ) /    &
                              beta(:n)         )
!
                m = m + 1_i4b
!
                eps2 = real( n, stnd)*eps
!
                if ( err<=eps2 ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'n= ', n, ' , case= ', j, ' , in_place= ', in_place
!
!                       PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                        write (prtunit,*) 'Condition number of the coefficient matrix          = ', conda
!
!                       PRINT SCALED ERRORS OF INVERSE.
!
                        write (prtunit,*) 'Maximum scaled error of columns of computed inverse = ', err
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ', err, ' > ', eps2, ' , n= ', n, ' , case= ', j,       &
                                      ' , in_place= ', in_place
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of the coefficient matrix          = ', conda
!
!                   PRINT SCALED ERRORS OF INVERSE.
!
                    write (prtunit,*) 'Maximum scaled error of columns of computed inverse = ', err
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta, norma,   &
                    a, ainv, ainv2, leftvec, rightvec, res )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_inv
! ============================
!
end program test_comp_inv
