program test_comp_mvs_grp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine comp_mvs_grp in module Stat_Procedures .
!                                                                              
! LATEST REVISION : 10/04/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, zero, comp_mvs_grp, merror, allocate_error, &
                         n1_def, n2_def
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, p IS THE NUMBER OF OBSERVATIONS, xmiss IS THE FLAG FOR MISSING VALUES.
!
    integer(i4b), parameter :: prtunit=6, ngrp=2, nobs_grp=20, n=n1_def, m=n2_def, p=ngrp*nobs_grp+5
!
    real(stnd), parameter :: xmiss=-999.99_stnd
!
    character(len=*), parameter :: name_proc='comp_mvs_grp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err_mean_grp, err_std_grp, err, eps
    real(stnd), allocatable, dimension(:,:,:) :: xmean_grp, xmean_grp1, xmean_grp2,   &
                                                 xstd_grp, xstd_grp1, xstd_grp2,      &
                                                 xn_grp, x1, x2, tmp
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, i1, i2, inc, success
    integer(i4b), allocatable :: ind(:), cnt(:,:,:)
!
    logical(lgl)                                :: first, last
    logical(lgl), allocatable, dimension(:,:,:) :: mask, tmp_mask
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( xmean_grp(n,m,ngrp), xmean_grp1(n,m,ngrp), xmean_grp2(n,m,ngrp), &
              xstd_grp(n,m,ngrp), xstd_grp1(n,m,ngrp), xstd_grp2(n,m,ngrp),    &
              xn_grp(n,m,ngrp), tmp(n,m,nobs_grp), x1(n,m,p), x2(n,m,p),       &
              ind(p), cnt(n,m,ngrp), mask(n,m,p), tmp_mask(n,m,nobs_grp),      &
              stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM OBSERVATION ARRAY x1 .
!
    call random_number( x1 )
!
!   GENERATE A RANDOM OBSERVATION ARRAY x2 WITH MISSING VALUES .
!
    mask = x1<=0.25_stnd
    where( mask )
       x2 = xmiss
    elsewhere
       x2 = x1
    end where
!
    mask = .not.mask
!
!   GENERATE GROUPS OF OBSERVATIONS.
!   COMPUTE THE MEANS OF x1 BY GROUPS.
!   COMPUTE THE STANDARD-DEVIATIONS OF x1 BY GROUPS.
!   COMPUTE THE MEANS OF x2 BY GROUPS.
!   COMPUTE THE STANDARD-DEVIATIONS OF x2 BY GROUPS.
!
    ind = 0_i4b
    i1 = 1_i4b
    i2 = 0_i4b
!
    do i = 1_i4b, ngrp
!
        i2 = i2 + nobs_grp
!
        ind(i1:i2) = i
        tmp_mask(:,:,:) = mask(:,:,i1:i2)
        cnt(:,:,i) = count( tmp_mask, dim=3_i4b )
!
        tmp(:,:,:) = x1(:,:,i1:i2)
        xmean_grp1(:,:,i) = sum( tmp, dim=3_i4b )/nobs_grp
        tmp(:,:,:) = (tmp(:,:,:)-spread( xmean_grp1(:,:,i),dim=3_i4b,ncopies=nobs_grp))**2
        xstd_grp1(:,:,i)  = sqrt( sum( tmp, dim=3_i4b )/nobs_grp )
!
        where ( tmp_mask(:,:,:) )
            tmp(:,:,:) = x2(:,:,i1:i2)
        elsewhere
            tmp(:,:,:) = zero
        end where
!
        xmean_grp2(:,:,i) = sum( tmp, dim=3_i4b )/cnt(:,:,i)
!
        where ( tmp_mask(:,:,:) )
            tmp(:,:,:) = (tmp(:,:,:)-spread( xmean_grp2(:,:,i),dim=3_i4b,ncopies=nobs_grp))**2
        elsewhere
            tmp(:,:,:) = zero
        end where
!
        xstd_grp2(:,:,i)  =  sqrt( sum( tmp, dim=3_i4b ) /cnt(:,:,i) )
!
        i1 = i1 + nobs_grp
!
    end do
!
!   RECOMPUTE THE STATISTICS WITH comp_mvs_grp SUBROUTINE.
!
    success = 0_i4b
!
    do i = 1_i4b, 12_i4b
!
        inc = merge( 1_i4b, p, i<=6_i4b )
!
        select case( i )
!
            case( 1_i4b, 7_i4b )
!
!               TESTS 1 AND 7 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_mvs_grp_rt.
!
                i2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    i2 = min( i2+inc, p )
!
                    first = k==1_i4b
                    last = i2==p
!
                    call comp_mvs_grp( x1(:,:,k:i2), first, last, ngrp, ind(k:i2),   &
                                       xmean_grp, xstd_grp, xn_grp(1_i4b,1_i4b,:)  )
!
                end do
!
            case( 2_i4b, 8_i4b )
!
!               TESTS 2 AND 8 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_mvs_grp_miss_rt.
!
                i2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    i2 = min( i2+inc, p )
!
                    first = k==1_i4b
                    last = i2==p
!
                    call comp_mvs_grp( x2(:,:,k:i2), first, last, ngrp, ind(k:i2),   &
                                       xmean_grp, xstd_grp, xn_grp,                  &
                                       xmiss=xmiss  ) 
!
                end do
!
            case( 3_i4b, 9_i4b )
!
!               TESTS 3 AND 9 :
!
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_mvs_grp_rm.
!
                do j = 1_i4b, m
!
                    i2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        i2 = min( i2+inc, p )
!
                        first = k==1_i4b
                        last = i2==p
!
                        call comp_mvs_grp( x1(:,j,k:i2), first, last, ngrp, ind(k:i2),                &
                                           xmean_grp(:,j,:), xstd_grp(:,j,:), xn_grp(1_i4b,1_i4b,:)  )
                    end do
!
                end do
!
            case( 4_i4b, 10_i4b )
!
!               TESTS 4 AND 10 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_mvs_grp_miss_rm.
!
                do j = 1_i4b, m
!
                    i2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        i2 = min( i2+inc, p )
!
                        first = k==1_i4b
                        last = i2==p
!
                        call comp_mvs_grp( x2(:,j,k:i2), first, last, ngrp, ind(k:i2),         &
                                           xmean_grp(:,j,:), xstd_grp(:,j,:), xn_grp(:,j,:),   &
                                           xmiss=xmiss  ) 
                    end do
!
                end do
!
            case( 5_i4b, 11_i4b )
!
!               TESTS 5 AND 11 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_mvs_grp_rv.
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        i2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            i2 = min( i2+inc, p )
!
                            first = k==1_i4b
                            last = i2==p
!
                            call comp_mvs_grp( x1(l,j,k:i2), first, last, ngrp, ind(k:i2),              &
                                               xmean_grp(l,j,:), xstd_grp(l,j,:), xn_grp(1_i4b,1_i4b,:) )
                        end do
!
                    end do
!
                end do
!
            case( 6_i4b, 12_i4b )
!
!               TESTS 6 AND 12 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_mvs_grp_miss_rv.
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        i2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            i2 = min( i2+inc, p )
!
                            first = k==1_i4b
                            last = i2==p
!
                            call comp_mvs_grp( x2(l,j,k:i2), first, last, ngrp, ind(k:i2),       &
                                               xmean_grp(l,j,:), xstd_grp(l,j,:), xn_grp(l,j,:), &
                                               xmiss=xmiss  ) 
!
                        end do
!
                    end do
!
                end do
!
        end select
!
!       CHECK THAT THE TWO SETS OF STATISTICS AGREE.
!
        if ( mod(i,2_i4b)==1_i4b ) then
!
            err_mean_grp = maxval( abs( (xmean_grp-xmean_grp1)/xmean_grp ) )
            err_std_grp = maxval( abs( (xstd_grp-xstd_grp1)/xstd_grp ) )
!
        else
!
            err_mean_grp  = maxval( abs( (xmean_grp-xmean_grp2)/xmean_grp ) )
            err_std_grp = maxval( abs( (xstd_grp-xstd_grp2)/xstd_grp ) )
!
        end if
!
!       PRINT THE RESULTS OF THE TESTS.
!
        err = max( err_mean_grp, err_std_grp )
!
        if ( err<=eps ) then
!
            success = success + 1_i4b
!
!            write (prtunit,*) 'Test ',i,' is correct'
!
        else
!
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ', err,' > ', eps
!
            write (prtunit,*)
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( xmean_grp, xmean_grp1, xmean_grp2, &
                xstd_grp, xstd_grp1, xstd_grp2,    &
                xn_grp, tmp, x1, x2, ind, cnt,     &
                mask, tmp_mask  )
!
    if ( success/=12_i4b ) then
        write (prtunit,*) 12_i4b - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_mvs_grp
! ================================
!
end program test_comp_mvs_grp
