program test_comp_pc_mca
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines comp_mca and comp_pc_mca 
!   in module Mul_Stat_Procedures .
!                                                                              
! LATEST REVISION : 23/05/2012
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, false, one, comp_mca, comp_pc_mca
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, mx=20, my=10, m=min(mx,my), p=50, mm=(m*(m+1))/2
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                         :: err_xstat, err_ystat, err_pc, err, eps, xn
    real(stnd), dimension(mx,my)       :: xsingvec
    real(stnd), dimension(mx,2)        :: xstat1, xstat2
    real(stnd), dimension(my,2)        :: ystat1, ystat2
    real(stnd), dimension(m)           :: xysingval
    real(stnd), dimension(mm)          :: pccorp
    real(stnd), dimension(mx,m)        :: xpccor
    real(stnd), dimension(p,m)         :: xpc
    real(stnd), dimension(mx,p)        :: x, xt, xt_std
    real(stnd), dimension(my,p)        :: y, yt
!
    integer(i4b) :: i, k, k2, inc
!
    logical(lgl) :: first, last, cov, failure
!
    character    :: sort = 'd'
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_pc_mca'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   GENERATE RANDOM OBSERVATION ARRAY x .
!
    call random_number( x )
!
!   COMPUTE THE MEANS AND STANDARD-DEVIATIONS OF x  .
!
    xstat1(:,1_i4b) = sum( x, dim=2_i4b )/p
    xt = x - spread( xstat1(:,1_i4b), dim=2_i4b, ncopies=p )
    xstat1(:,2_i4b) = sqrt( sum( xt**2, dim=2_i4b )/p )
!
!   GENERATE A RANDOM OBSERVATION ARRAY y .
!
    call random_number( y )
!
!   COMPUTE THE MEANS AND STANDARD-DEVIATIONS OF y  .
!
    ystat1(:,1_i4b) = sum( y, dim=2_i4b )/p
    yt = y - spread( ystat1(:,1_i4b), dim=2_i4b, ncopies=p )
    ystat1(:,2_i4b) = sqrt( sum( yt**2, dim=2_i4b )/p )
!
!   COMPUTE STANDARDIZED VARIABLES .
!
    yt = yt*spread( one/ystat1(:,2_i4b), dim=2_i4b, ncopies=p )
    xt_std = xt*spread( one/xstat1(:,2_i4b), dim=2_i4b, ncopies=p )
!
    cov = false
!
    do i = 1_i4b, 4_i4b
!
        cov = .not.cov
        inc = merge( 1_i4b, p, i<=2_i4b )
!
!       COMPUTE LEFT SINGULAR VECTORS OF THE COVARIANCE OR CORRELATION MATRIX .
!
        k2 = 0_i4b
        do k = 1_i4b, p, inc
!
            k2 = min( k2+inc, p )
!
            first = k==1_i4b
            last = k2==p
!
            call comp_mca( x(:,k:k2), y(:,k:k2), first, last, xstat2, ystat2, xysingval, xsingvec, failure,    &
                           cov=cov, sort=sort  )
        end do
!
        if ( cov ) then
!
            k2 = 0_i4b
            do k = 1_i4b, p, inc
!
                k2 = min( k2+inc, p )
!
                first = k==1_i4b
                last = k2==p
!
                call comp_pc_mca( x(:,k:k2), xsingvec, first, last, xpccor, pccorp, xpc(k:k2,:), xn,  &
                                  xmean=xstat2(:,1_i4b)                            )
            end do
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION xt**t*xsingvec - xpc
!           WHERE xsingvec ARE THE LEFT SINGULAR VECTORS OF THE COVARIANCE MATRIX, xt THE CENTERED DATA
!           AND xpc THE UNNORMALIZED SINGULAR VARIABLES.
!
            err_pc =         &
            sum( abs(matmul(transpose(xt),xsingvec)-xpc) )/sum( abs(xt) )
!
        else
!
            k2 = 0_i4b
            do k = 1_i4b, p, inc
!
                k2 = min( k2+inc, p )
!
                first = k==1_i4b
                last = k2==p
!
                call comp_pc_mca( x(:,k:k2), xsingvec, first, last, xpccor, pccorp, xpc(k:k2,:), xn,  &
                                  xmean=xstat2(:,1_i4b), xstd=xstat2(:,2_i4b)                         )
            end do
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION xt_std**t*xsingvec - xpc
!           WHERE xsingvec ARE THE LEFT SINGULAR VECTORS OF THE CORRELATION MATRIX, xt_std THE STANDARDIZED DATA
!           AND xpc THE UNNORMALIZED SINGULAR VARIABLES.
!
            err_pc =         &
            sum( abs(matmul(transpose(xt_std),xsingvec)-xpc) )/sum( abs(xt_std) )
!
        end if
!
!       CHECK THAT THE TWO SETS OF UNIVARIATE STATISTICS AGREE.
!
        err_xstat  = maxval( abs( (xstat1-xstat2)/xstat2 ) )
        err_ystat  = maxval( abs( (ystat1-ystat2)/ystat2 ) )
!
!       PRINT THE RESULTS OF THE TESTS.
!
        err = max( err_xstat, err_ystat, err_pc )
!
        if ( err<=eps .and. .not.failure ) then
            write (prtunit,*) 'Test ',i,' is correct'
        else
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ',err,' > ',eps
        end if
!
    end do
!
!
! END OF PROGRAM test_comp_pc_mca
! ===============================
!
end program test_comp_pc_mca
