program test_define_rot_givens
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine DEFINE_ROT_GIVENS in module Giv_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!   The program assumes that subroutine APPLY_ROT_GIVENS in module Giv_Procedures is correct.
!                                                                              
! LATEST REVISION : 01/12/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
! 
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,   &
                         zero, half, c50, define_rot_givens,       &
                         apply_rot_givens, norm, lamch, singvalues
!
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter :: fudge=c50, max_conda=1000._stnd
!
    character(len=*), parameter :: name_proc='define_rot_givens'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, safmin, cs, sn, tmp, &
                                               conda, normx
    real(stnd), dimension(:),   allocatable :: x, x2, res, singval
    real(stnd), dimension(:,:), allocatable :: a, syst
!
    integer(i4b) :: i, j, k, n, np1, m, success
    integer      :: iok
!
    logical(lgl) :: details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        np1 = n + 1_i4b
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( x(n), x2(n), res(n), singval(n),   &
                  a(n,n), syst(n,np1), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(x) )         deallocate( x )
            if ( allocated(x2) )        deallocate( x2 )
            if ( allocated(res) )       deallocate( res )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(syst) )      deallocate( syst )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX a .
!
        call random_number( a )
!
        a = a - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            a = a/tmp
        end if
!
        singval(:n) = singvalues( a(:n,:n) )
!
#ifdef _F2003
        if ( ieee_support_datatype( singval ) ) then
            if ( .not.all( ieee_is_normal( singval(1_i4b:n) ) ) ) cycle
        end if
#endif
!
        j = count( singval(:n)/=zero )
!
!       SKIP DEFICIENT MATRICES.
!
        if ( j/=n ) cycle
!
!       COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
        conda = maxval( singval(:n) )/minval( singval(:n) )
!
        tmp = min( conda, max_conda )
!
        eps2 = real( n, stnd)*eps*tmp
!
!       GENERATE A RANDOM SOLUTION VECTOR x .
!
        call random_number( x )
!
        x = x - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x = x/tmp
        end if
!
        normx = norm( x )
!
!       MAKE A COPY OF COEFFICIENT MATRIX a.
!    
        syst(:n,:n) = a(:n,:n)
!
!       COMPUTE RIGHT HAND-SIDE VECTOR b .
!
        syst(:n,np1) = matmul( a, x )
!
!       COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!            a(:n,:n)*x(:n) = b(:n) .
!
!       TRANSFORM THE MATRIX a TO UPPER TRAPEZOIDAL FORM BY APPLYING 
!       A SERIE OF GIVENS PLANE ROTATIONS ON THE ROWS OF a.
!       STORE THE ROTATIONS AND APPLY THEM TO b .
!
        do k = 1_i4b, n
!    
            do i = k+1_i4b, n
!    
                call define_rot_givens( syst(k,k), syst(i,k), cs, sn )
                call apply_rot_givens( syst(k,k:n), syst(i,k:n), cs, sn   )
                call apply_rot_givens( syst(k,np1), syst(i,np1), cs, sn   )
!    
            end do
!    
        end do
!    
!       SOLVE THE n BY n TRIANGULAR SYSTEM.
!    
        do i = n, 1_i4b, -1_i4b
!    
            x2(i) = syst(i,np1)/syst(i,i)
            syst(1_i4b:i-1_i4b,np1) = syst(1_i4b:i-1_i4b,np1) - x2(i)*syst(1_i4b:i-1_i4b,i)
!    
        end do
!
        m = m + 1_i4b
!
!       CHECK THE ERROR FOR SOLUTION.
!
        res = x2 - x
!
        err = norm( res ) / normx
!
        if ( err<=eps2 ) then
!
            success = success + 1_i4b
!
            if ( details ) then
!
                write (prtunit,*) 'Test ',m,' is correct'
!
                write (prtunit,*) 'n= ', n
!
!               PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
                write (prtunit,*) 'Relative error of the computed solution = ', err
!
                write (prtunit,*)
!
            end if
!
        else
!
            write (prtunit,*) 'Test ',m,' is incorrect'
!
            write (prtunit,*) '=> ',err,' > ', eps2, ' , n= ', n
!
!           PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
            write (prtunit,*) 'Relative error of the computed solution = ', err
!
            write (prtunit,*)
!
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, syst, res, singval, x, x2 )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_define_rot_givens
! =====================================
!
end program test_define_rot_givens
