program test_det
!
!
! Purpose
! =======
!
!   This program is intended to test function DET
!   in module Lin_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of function DET.
!
! LATEST REVISION : 03/10/2018
!                                                                         
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,         &
                         zero, half, one, two, five, ten, c50, c100,     &
                         qr_cmp, ortho_gen_qr, det, arth, lamch, is_nan, &
                         minexp, maxexp, machhuge, machbase
                         
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c50
!
    real(stnd), parameter :: invmachbase=one/machbase
!
    character(len=*), parameter :: name_proc='det'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, ulp, safmin, tmp,      &
                                               shift, conda, adet, adet2, dd1
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec
!
    integer(i4b) :: i, j, l, n, m, dd2, success, notdone
    integer      :: iok
!
    logical(lgl) :: details
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
!   SPECIFY A SHIFT TO AVOID ZERO SINGULAR VALUE.
!
    shift = 0.1_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), diagr(n),        &
                  beta(n), a(n,n),                        &
                  leftvec(n,n),  rightvec(n,n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:n,:n) .
!
        call random_number( leftvec(:n,:n) )
!
        leftvec(:n,:n) = leftvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            leftvec(:n,:n) = leftvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( leftvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( leftvec(:n,:n), beta(:n) )
!
!       GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:n,:n) .
!
        call random_number( rightvec(:n,:n) )
!
        rightvec(:n,:n) = rightvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            rightvec(:n,:n) = rightvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( rightvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( rightvec(:n,:n), beta(:n) )
!
!       COMPUTE FIRST THE EIGENVALUES OF THE MATRIX PRODUCT
!                     a(:n,:n)**(t)*a(:n,:n).
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
            case( 1_i4b )
!
!               TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b)   = one
                eigval(2_i4b:n) = ulp
!
            case( 2_i4b )
!
!               TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = one
                eigval(n)             = ulp
!
            case( 3_i4b )
!
!               TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                eigval(n)             = one
!
            case( 4_i4b )
!
!               TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                eigval(1_i4b)         = ulp
                eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                eigval(n)             = two
!
            case( 5_i4b )
!
!               TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
            case( 6_i4b )
!
!               TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                tmp = real( n-1_i4b, stnd )
                do l = 1_i4b, n
                    eigval(l) = ulp**(real(l-1_i4b,stnd)/tmp)
                end do
!
            case( 7_i4b )
!
!               TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
            case( 8_i4b )
!
!               TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
            case( 9_i4b )
!
!               TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
            case( 10_i4b )
!
!               TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
            case( 11_i4b )
!
!               TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                call random_number( eigval(:n) )
!
                eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
            case( 12_i4b )
!
!               TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
            case( 13_i4b )
!
!               TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 14_i4b )
!
!               TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 15_i4b )
!
!               TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case default
!
!               DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                call random_number( eigval(:n) )
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
!           COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
            singval(:n) = sqrt( eigval(:n) )
!
            singval(:n) = singval(:n) + shift
!
!           FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
            a(:n,:n) = matmul( leftvec(:n,:n)*spread(singval,1,n), transpose(rightvec(:n,:n)) )
!
!           COMPUTE THE DETERMINANT.
!
            adet = det( a )
!
            if ( is_nan( adet ) ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            conda = ( sum( singval(:n) )*sum( one/singval(:n) ) )/real( n, stnd)
!
!           RECOMPUTE THE DETERMINANT OF a FROM THE SINGULAR VALUES.
!
            dd1 = one
            dd2 = zero
!
            do l=1_i4b, n
!
                dd1 = dd1*singval(l)
!
                do
                    if ( abs(dd1)<= one ) exit
                    dd1 = dd1*invmachbase
                    dd2 = dd2 + 1_i4b
                end do
!
                do
                    if ( abs(dd1)>= invmachbase ) exit
                    dd1 = dd1*machbase
                    dd2 = dd2 - 1_i4b
                end do
!
            end do
!
            select case( dd2 )
!
                case ( :minexp-1_i4b )
!
                    adet2 = sign( zero, dd1 )
!
                case ( maxexp+1_i4b: )
!
                    adet2 = sign( machhuge, dd1 )
!
                case default
!
!                    adet2 = scale( dd1, dd2 )
                    adet2 = dd1*( machbase**dd2 )
!
            end select
!
!            adet2 = product( singval(:n) )
!
!           CHECK THE RESULTS.
!
#ifdef _F2003
            if ( ieee_support_datatype( adet2 ) ) then
!
                if ( ieee_is_normal( adet2 ) .and. ieee_is_normal( adet ) ) then
!
                    if ( min( abs(adet), abs(adet2) )==zero ) then
                        tmp = one
                    else
                        tmp = max( abs(adet), abs(adet2) )
                    end if
!
                    err = abs(adet - adet2) / tmp
!
                else
!
                    err = zero
!
                end if
!
            else
!
#else
                if ( abs( adet2 )<=machhuge .and. abs( adet )<=machhuge ) then
!
                    if ( min( abs(adet), abs(adet2) )==zero ) then
                        tmp = one
                    else
                        tmp = max( abs(adet), abs(adet2) )
                    end if
!
                    err = abs(adet - adet2) / tmp
!
                else
!
                    err = zero
!
                end if
#endif
!
#ifdef _F2003
            end if
#endif
!
            m = m + 1_i4b
!
            eps2 = real( n, stnd)*eps
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n,' , case= ', j,       &
                                      ' , adet= ', adet,' , adet2= ', adet2
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of the coefficient matrix = ', conda
!
!                   PRINT ABSOLUTE AND RELATIVE ERRORS OF COMPUTED DETERMINANT.
!
                    write (prtunit,*) 'Absolute error of the computed determinant = ', abs(adet - adet2)
                    write (prtunit,*) 'Relative error of the computed determinant = ', err
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps2,' , n= ', n,' , case= ', j,   &
                                  ' , adet= ', adet,' , adet2= ', adet2
!
!               PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                write (prtunit,*) 'Condition number of the coefficient matrix  = ', conda
!
!               PRINT ABSOLUTE AND RELATIVE ERRORS OF COMPUTED DETERMINANT.
!
                write (prtunit,*) 'Absolute error of the computed determinant  = ', abs(adet - adet2)
                write (prtunit,*) 'Relative error of the computed determinant  = ', err
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta,   &
                    a, leftvec, rightvec            )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_det
! =======================
!
end program test_det
