program test_eig_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine EIG_CMP2
!   in module EIG_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given eigenvalue value distributions are generated for checking
!    the accuracy of subroutine EIG_CMP2.
!
! LATEST REVISION : 31/07/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,            &
                         true, zero, half, one, two, five, ten, c50, c100,  &
                         qr_cmp, ortho_gen_qr, eigval_sort, eig_cmp2, norm, &
                         arth, lamch
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='eig_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, ulp, safmin, tmp, norma
    real(stnd), allocatable, dimension(:)   :: eigval, eigval2, diagr, beta, temp
    real(stnd), allocatable, dimension(:,:) :: a, eigvec, eigvec2, res
!
    integer      :: iok
    integer(i4b) :: i, j, l, n, m, q, success, notdone
!
    logical(lgl) :: failure, details, upper
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), eigval2(n), diagr(n), beta(n), temp(n),    &
                  a(n,n), res(n,n), eigvec(n,n), eigvec2(n,n), stat=iok )
!
        if ( iok/=0 ) then
!
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(eigvec) )    deallocate( eigvec )
            if ( allocated(eigval2) )   deallocate( eigval2 )
            if ( allocated(eigvec2) )   deallocate( eigvec2 )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(res) )       deallocate( res )
            if ( allocated(temp) )      deallocate( temp )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX eigvec(:n,:n) .
!
        call random_number( eigvec(:n,:n) )
!
        eigvec(:n,:n) = eigvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            eigvec(:n,:n) = eigvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( eigvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( eigvec(:n,:n), beta(:n) )
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
                case( 1_i4b )
!
!                   TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b)   = one
                    eigval(2_i4b:n) = ulp
!
                case( 2_i4b )
!
!                   TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n-1_i4b) = one
                    eigval(n)             = ulp
!
                case( 3_i4b )
!
!                   TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                    eigval(n)             = one
!
                case( 4_i4b )
!
!                   TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                    eigval(1_i4b)         = ulp
                    eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                    eigval(n)             = two
!
                case( 5_i4b )
!
!                   TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                case( 6_i4b )
!
!                   TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                    err = real( n-1_i4b, stnd )
!
                    do l = 1_i4b, n
                        eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                    end do
!
                case( 7_i4b )
!
!                   TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                    eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                case( 8_i4b )
!
!                   TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                    eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                case( 9_i4b )
!
!                   TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                case( 10_i4b )
!
!                   TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                case( 11_i4b )
!
!                   TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                    call random_number( eigval(:n) )
!
                    eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                case( 12_i4b )
!
!                   TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                case( 13_i4b )
!
!                   TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
                case( 14_i4b )
!
!                   TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
                case( 15_i4b )
!
!                   TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
                case default
!
!                   DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                    call random_number( eigval(:n) )
!
                    call random_number( tmp )
!
                    if ( tmp>safmin ) then
                        eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                    end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
            do q = 1_i4b, 2_i4b
!
                if ( q==2_i4b ) then
!
!                   EIGENVALUES ARE RANDOMLY NEGATED.
!
                    call random_number( beta(:n) )
!
                    where ( beta(:n)>half ) eigval(:n) = -eigval(:n)
!
                end if
!
!               SORT THE EIGENVALUES.
!
                call eigval_sort( sort, eigval(:n) )
!
                where ( eigval(:n)==zero )
!
                    temp(:n) = one
!
                elsewhere
!
                    temp(:n) = eigval(:n)
!
                end where
!
!               FORM A SYMMETRIC MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
                a(:n,:n) = matmul( eigvec(:n,:n)*spread(eigval,1,n), transpose(eigvec(:n,:n)) )
!
!               COMPUTE THE NORM OF THE SYMMETRIC MATRIX.
!
!                norma = norm( a(:n,:n) )
                norma = norm( eigval(:n) )
!
                upper = true
!
                do i = 1_i4b, 3_i4b
!
!                   MAKE A COPY OF THE ORIGINAL SYMMETRIC MATRIX.
!
                    eigvec2(:n,:n) = a(:n,:n)
!
!                   COMPUTE EIGENVECTORS OF THE SYMMETRIC MATRIX.
!
                    if ( i==1_i4b ) then
!
                        call eig_cmp2( eigvec2(:n,:n), eigval2(:n), failure, sort=sort )
!
                    else
!
                        upper = .not.upper
!
                        call eig_cmp2( eigvec2(:n,:n), eigval2(:n), failure, upper=upper, sort=sort )
!
                    end if
!
                    if ( failure ) then
!
                        notdone = notdone + 1_i4b
!
                        cycle
!
                    end if
!
                    m = m + 1_i4b
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                   WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                    res(:n,:n) = matmul( a(:n,:n), eigvec2(:n,:n) )        &
                                 - eigvec2(:n,:n)*spread(eigval2(:n),1,n)
!
                    if ( norma/=zero ) then
                        err1 = maxval( norm(res(:n,:n), dim=2_i4b) )/( norma*real(n,stnd) )
                    else
                        err1 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
                    end if
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                   WHERE u are THE EIGENVECTORS OF THE MATRIX.
!
                    res(:n,:n) = matmul( transpose( eigvec2(:n,:n) ), eigvec2(:n,:n) )
!
                    do l = 1_i4b, n
                        res(l,l) = one - res(l,l)
                    end do
!
                    err2 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
!
                    err = max( err1, err2 )
!
                    if ( err<=eps ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            if ( i==1_i4b ) then
!
                                write (prtunit,*) 'n= ', n,' , case= ', j
!
                            else
!
                                write (prtunit,*) 'n= ', n,' , case= ', j,' , upper= ', upper
!
                            end if
!
!                           COMPUTE ABSOLUE ERRORS OF EIGENVALUES.
!
                            write (prtunit,*) 'Maximum absolute error of the computed eigenvalues = ', &
                                              maxval( abs( eigval2(:n)-eigval(:n) ) )
!
!                           COMPUTE RELATIVE ERRORS OF EIGENVALUES.
!
                            write (prtunit,*) 'Maximum relative error of the computed eigenvalues = ', &
                                              maxval( abs( (eigval2(:n)-eigval(:n) )/temp(:n) ) )
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!                           WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Accuracy of the computed eigenpairs                = ', &
                                              err1/ulp, ' (must be less than :', fudge, ' )'
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                           WHERE u are THE EIGENVECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Orthogonality of the computed eigenvectors         = ', &
                                              err2/ulp, ' (must be less than :', fudge, ' )'
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        if ( i==1_i4b ) then
!
                            write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,        &
                                              ' , case= ', j,' , failure= ', failure
!
                        else
!
                            write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,                           &
                                              ' , case= ', j,' , upper= ', upper,' , failure= ', failure
!
                        end if
!
!                       COMPUTE ABSOLUE ERRORS OF EIGENVALUES.
!
                        write (prtunit,*) 'Maximum absolute error of the computed eigenvalues = ', &
                                          maxval( abs( eigval2(:n)-eigval(:n) ) )
!
!                       COMPUTE RELATIVE ERRORS OF EIGENVALUES.
!
                        write (prtunit,*) 'Maximum relative error of the computed eigenvalues = ', &
                                          maxval( abs( (eigval2(:n)-eigval(:n) )/temp(:n) ) )
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!                       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Accuracy of the computed eigenpairs                = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                       WHERE u are THE EIGENVECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed eigenvectors         = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, eigval2, temp, diagr, beta, a, eigvec, eigvec2, res )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_eig_cmp2
! ============================
!
end program test_eig_cmp2
