program test_givens_mat_left
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine GIVENS_MAT_LEFT in module Giv_Procedures. 
!
! LATEST REVISION : 03/09/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,   &
                         half, c1000,                               &
                         givens_mat_left, lamch, norm
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c1000
!
    character(len=*), parameter :: name_proc='givens_mat_left'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, ulp, safmin, tmp,   &
                                               anorm, bnorm, resnorm
    real(stnd), allocatable, dimension(:)   :: res, b, x
    real(stnd), allocatable, dimension(:,:) :: a, syst
!
    integer(i4b) :: i, k, n, nr, nc, ncp1, m, success
    integer      :: iok
!
    logical(lgl) :: details
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        do k = 1_i4b, 3_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: QR OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: QR OF a(:nr,:nc) WITH nr>=nc .
!
                    nr = n + n
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: QR OF a(:nr,:nc) WITH nc>=nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
            ncp1 = nc + 1_i4b
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), syst(nr,ncp1), res(nr),   &
                      x(nc), b(nr), stat=iok  )
!
            if ( iok/=0 ) cycle
!
!           GENERATE A RANDOM COEFFICIENT MATRIX a(:nr,:nc) .
!
            call random_number( a(:nr,:nc) )
!
            a(:nr,:nc) = a(:nr,:nc) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                a(:nr,:nc) = a(:nr,:nc)/tmp
            end if
!
!           COMPUTE THE NORM OF RANDOM COEFFICIENT MATRIX a .
!
            anorm = norm( a(:nr,:nc) )
!
!           GENERATE A RANDOM RIGHT HAND SIDE VECTOR b .
!
            call random_number( b(:nr) )
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                b(:nr) = b(:nr)/tmp
            end if
!
!           COMPUTE THE NORM OF THE RIGHT HAND SIDE VECTOR b .
!
            bnorm = norm( b(:nr) )
!
            syst(:nr,:nc) = a(:nr,:nc)
            syst(:nr,ncp1) = b(:nr)
!
!           COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                  a(:nr,:nc)*x(:nc) = b(:nr) .
!
!    
!           TRANSFORM THE MATRIX a TO UPPER TRAPEZOIDAL FORM BY APPLYING 
!           A SERIE OF GIVENS PLANE ROTATIONS ON THE ROWS OF a AND APPLY
!           THE ROTATIONS TO b .
!
            call givens_mat_left( syst(:nr,:ncp1) )
!    
!           SOLVE THE n BY n UPPER TRIANGULAR SYSTEM.
!    
            do i = n, 1_i4b, -1_i4b
!    
                x(i) = syst(i,ncp1)/syst(i,i)
                syst(1_i4b:i-1_i4b,ncp1) = syst(1_i4b:i-1_i4b,ncp1) - x(i)*syst(1_i4b:i-1_i4b,i)
!    
            end do
!
!           CHECK THE RESULTS FOR SMALL RESIDUALS.
!
            res(:nr) = b(:nr) - matmul( a(:nr,:n), x(:n) )
!    
!           COMPLETE COMPUTATION OF SOLUTION VECTOR.
!    
!            x(n+1_i4b:nc) = zero
!
!           CHECK THE RESULTS FOR SMALL RESIDUALS.
!
!            res(:nr) = b(:nr) - matmul( a(:nr,:nc), x(:nc) )
!
            err = norm( matmul(transpose(a),res) )
!
            if ( k==1_i4b .or. k==3_i4b  ) then
!
                resnorm = norm( res )
!
                err = max( err, resnorm )
!
            end if
!
            err = err/( anorm + bnorm )
!
            m = m + 1_i4b
!
            eps2 = real( n, stnd)*eps
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'nr= ', nr, ' , nc= ', nc
!
!                   PRINT RELATIVE ERRORS OF RESIDUALS.
!
                    write (prtunit,*) 'Maximum relative error of the computed residuals = ', err
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ', err,' > ', eps2, ' , nr= ', nr, ' , nc= ', nc
!
!               PRINT RELATIVE ERRORS OF RESIDUALS.
!
                write (prtunit,*) 'Maximum relative error of the computed residuals = ', err
!
                write (prtunit,*)
!
            end if
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, syst, res, x, b )
!
        end do
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_givens_mat_left
! ===================================
!
end program test_givens_mat_left
