program test_int_ran_num
!
!
! Purpose
! =======
!
!   This program is intended to test the intrinsic RANDOM_NUMBER and RANDOM_SEED subroutines. 
!                                                                              
! LATEST REVISION : 30/07/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, true, zero, one, three, c90, c1_m3,    &
                         base, arth, probn, probq, allocate_error, merror
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, nk IS THE NUMBER OF ELEMENTS
! BY CLASS FOR CHI-SQUARE GOODNESS-OF-FIT TEST, THE OTHER VARIABLES ARE
! USED BY THE CHI-SQUARE TEST FOR EQUIDISTRIBUTION PROPERTY OF THE RNG.
!
    integer(i4b), parameter :: prtunit = 6, nk = 100_i4b, nobs=100000_i4b,   &
                               nfreq=10_i4b, nfreqp1 = nfreq + 1_i4b
!
! eps IS THE SIGNIFICANCE LEVEL FOR THE UNIFORMITY, EQUIDISTRIBUTION AND RANDOMNESS TESTS.
!
    real(stnd), parameter :: eps = c1_m3
!
    character(len=*), parameter :: name_proc='intrinsic random_number'
!   
!   
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: y1, y2, y3, x, mean, std, prob, chi_sq, expctd
    real(stnd), allocatable, dimension(:)       :: harvest1, harvest2, bounds, sample
    real(stnd), allocatable, dimension(:,:,:,:) :: freq
!   
    integer(i4b) :: ndf, i, j, k, kp1, l, m, n, p, runs, success
!   
    integer                            :: nseed, iok
    integer, allocatable, dimension(:) :: histo, seed
!   
    logical(lgl) :: failure, details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
    success = 0_i4b
!
!   k IS THE NUMBER OF CLASSES FOR THE CHI-SQUARE GOODNESS-OF-FIT TEST.
!
    k = base
!
    do
!
        if ( k>100_i4b ) exit
!
        k = k*base
!
    end do
!
!   n IS THE TOTAL NUMBER OF OBSERVATIONS IN THE SAMPLE,
!   ndf IS THE NUMBER OF DEGREES OF FREEDOM FOR THE CHI-SQUARE
!   GOODNESS-OF-FIT TEST.
!
    n   = nk*k
    ndf = k - 1_i4b
    kp1 = k + 1_i4b
!
!   GET THE SIZE OF THE SEED ARRAY.
!
    call random_seed( size=nseed )
!
!   ALLOCATE WORK VARIABLES.
!
    allocate( bounds(kp1), sample(n), histo(k), seed(nseed),                  &
              harvest1(20_i4b), harvest2(20_i4b),                             &
              freq(0_i4b:nfreq,0_i4b:nfreq,0_i4b:nfreq,0_i4b:nfreq), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   COMPUTE THE BOUNDS OF THE k CLASSES.
!
    bounds(:k) = arth( zero, one, k )/real( k, stnd )
    bounds(kp1) = nearest( one, one )
!
!   SEED THE GENERATOR.
!
    call random_seed()
!
!   GENERATE A VECTOR OF RANDOM NUMBERS.
!
    call random_number( sample(:n) )
!
    do m = 1_i4b, 7_i4b
!
!       TEST THE RANDOM GENERATOR.
!
        select case( m )
!
            case( 1_i4b )
!
!               TEST 1 : TEST FOR UNIFORMITY WITH CHI-SQUARE GOODNESS-OF-FIT TEST
!
!               COMPUTE CHI-SQUARE TEST CRITERION.
!
                do i = 1_i4b, k
                    histo(i) = count( sample(:n)>=bounds(i) .and. sample(:n)<bounds(i+1_i4b) )
                end do
!
                chi_sq = real( sum( (histo(:k)-nk)**2 ), stnd )/real( nk, stnd )
!
!               COMPUTE PROBABILITY.
!
                prob = probq( chi_sq, ndf=ndf, upper=true )
!
                failure = prob<=eps
!
            case( 2_i4b )
!
!               TEST 2 : TEST FOR RANDOMNESS WITH THE TOTAL NUMBERS OF RUN UP OR DOWN.
!
!               COUNT THE TOTAL NUMBER OF RUNS.
!
                runs = 0_i4b
!
                do i = 2_i4b, n-1_i4b
!
                    y1 = sample(i-1_i4b)
                    y2 = sample(i)
                    y3 = sample(i+1_i4b)
!
                    if ( (y1<y2 .and. y3<y2) .or. (y1>y2 .and. y3>y2) ) runs = runs + 1_i4b
!
                end do
!
!               COMPUTE TEST CRITERION.
!
                mean = real( 2_i4b*n - 1_i4b, stnd )/three
                std = sqrt( real( 16_i4b*n - 29_i4b, stnd )/c90 )
                x = abs( ( real( runs, stnd ) - mean )/std )
!
!               COMPUTE PROBABILITY.
!
                prob = probn( x, upper=true )
!
                failure = prob<=eps
!
            case( 3_i4b )
!
!               TEST 3 : TEST FOR EQUIDISTRIBUTION PROPERTY OF THE RNG.
!
                ndf = (nfreqp1**4) - 1_i4b
                expctd = real( nobs, stnd)/( ndf + 1_i4b )
!
                call random_number( x )
                i = nfreqp1*x
!
                call random_number( x )
                j = nfreqp1*x
!
                call random_number( x )
                k = nfreqp1*x
!
                freq(:,:,:,:) = zero
!
                do p = 1_i4b, nobs
!
                    call random_number( x )
                    l = nfreqp1*x
                    freq(i,j,k,l) = freq(i,j,k,l) + one
                    i = j
                    j = k
                    k = l
!        
                end do
!
                chi_sq = zero
!
                do k = 0_i4b, nfreq
                    do j = 0_i4b, nfreq
                        do i = 0_i4b, nfreq
                            chi_sq = chi_sq + sum( (freq(:,i,j,k) - expctd)**2 )/expctd
                        end do
                    end do
                end do
!
!               COMPUTE PROBABILITY.
!
                prob = probq( chi_sq, ndf=ndf, upper=true )
!
                failure = prob<=eps
!
            case( 4_i4b )
!
!               TEST 4 : TEST THAT random_number RETURNS VALUES IN THE INTERVAL [0,1[.
!
                failure = any( sample(:)<zero .or. sample(:)>=one )
!
            case( 5_i4b )
!
!               TEST 5 : TEST FOR REPEATABLE CALLS OF SUBROUTINE random_number.
!
!               GET THE CURRENT STATE OF THE GENERATOR.
!
                call random_seed( get=seed(:) )
!
!               GENERATE A FIRST VECTOR OF RANDOM NUMBERS.
!
                call random_number( harvest1(:) )
!
!               RESET THE GENERATOR.
!
                call random_seed( put=seed(:) )
!
!               GENERATE A SECOND VECTOR OF RANDOM NUMBERS.
!
                call random_number( harvest2(:) )
!
!               THE TWO RANDOM NUMBERS VECTORS MUST MATCH.
!
                failure = any( harvest1(:)/=harvest2(:) )
!
            case( 6_i4b )
!
!               TEST 6: TEST FOR SUCCESSIVE CALLS OF SUBROUTINE random_number.
!
!               GENERATE FIRST VECTOR OF RANDOM NUMBERS.
!
                harvest1(:) = zero
                call random_number( harvest1(:) )
!
!               GENERATE SECOND RANDOM NUMBERS VECTOR.
!
                harvest2(:) = zero
                call random_number( harvest2(:) )
!
!               THE TWO RANDOM NUMBERS VECTORS MUST NOT MATCH.
!
                failure = any( harvest1(:)==harvest2(:) )
!
            case( 7_i4b )
!
!               TEST 7: TEST FOR SUCCESSIVE CALLS OF SUBROUTINE random_number WITH A RESET
!               BY A CALL TO random_seed WITHOUT ARGUMENTS.
!
!               GENERATE FIRST VECTOR OF RANDOM NUMBERS.
!
                harvest1(:) = zero
                call random_seed( )
                call random_number( harvest1(:) )
!
!               GENERATE SECOND VECTOR OF RANDOM NUMBERS.
!
                 harvest2(:) = zero
                 call random_seed( )
                 call random_number( harvest2(:) )
!
!               THE TWO RANDOM NUMBERS VECTORS MUST MATCH.
!
                failure = any( harvest1(:)/=harvest2(:) )
!
        end select
!
!       CHECK THE RESULTS OF THE TESTS.
!
        if ( .not.failure ) then
!
            success = success + 1_i4b
!
            if ( details ) then
!
                write (prtunit,*) 'Test ',m,' is correct'
!
                if ( m<=3_i4b ) then
!
!                   PRINT PROBABILITY OF THE TEST.
!
                    write (prtunit,*) 'Probability of the result = ', prob
!
                end if
!
                write (prtunit,*)
!
            end if
!
        else
!
            if ( m<=3_i4b ) then
!
                write (prtunit,*) 'Test ', m ,' is incorrect'
!
!               PRINT PROBABILITY OF THE TEST.
!
                write (prtunit,*) '=> ',prob,' <= ', eps
!
            else if ( m==7_i4b ) then
!
                success = success + 1_i4b
!
                write (prtunit,*) 'Warning information '
!
                write (prtunit,*) &
         '=> Non repeatable reset of intrinsic RNG with a call to random_seed() without argument!'
!
            end if
!
            write (prtunit,*)
!
        end if
!
    end do
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( harvest1, harvest2, histo, bounds, sample, seed, freq )
!
    p = m - 1_i4b
!
    if ( p/=success ) then
        write (prtunit,*) p - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_int_ran_num
! ===============================
!
end program test_int_ran_num
