program test_lq_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines LQ_CMP and ORTHO_GEN_LQ
!   in module QR_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutines LQ_CMP and ORTHO_GEN_LQ.
!
! LATEST REVISION : 23/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,                       &
                         true, zero, half, one, two, five, ten, c50, c100,             &
                         qr_cmp, ortho_gen_qr, lq_cmp, ortho_gen_lq, norm, arth, lamch
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='lq_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, ulp, safmin, tmp, shift, tol
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta, diagl, tau
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, res, q, ll
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, krank0, success
    logical(lgl) :: details, use_qr
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    safmin = lamch( 'S' )
!
!   SPECIFY EVENTUALLY A SHIFT TO AVOID ZERO SINGULAR VALUE
!   OR SHIFT THE SINGULAR VALUES DISTRIBUTION.
!
    shift = zero
!
    success = 0_i4b
    m       = 0_i4b
!
    use_qr = true
!
    do n = n1, n2, n3
!
        use_qr = .not.use_qr
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n),      &
                  diagr(n), beta(n),          &
                  diagl(n), tau(n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(diagl) )     deallocate( diagl )
            if ( allocated(tau) )       deallocate( tau )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: LQ DECOMPOSITION OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: LQ DECOMPOSITION OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: LQ DECOMPOSITION OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: LQ DECOMPOSITION OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: LQ DECOMPOSITION OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), res(nr,nc), ll(nr,n), q(n,nc),     &
                      leftvec(nr,n),   rightvec(nc,n),  stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(ll) )        deallocate( ll )
                if ( allocated(q) )         deallocate( q )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n).
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n).
!
            call random_number( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
            do j = 1_i4b, 16_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + sqrt( ulp )
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + sqrt( ulp )
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + sqrt( ulp )
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
                m = m + 1_i4b
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
                singval(:n) = singval(:n) + shift
!
                tol = real( n, stnd)*sqrt( ulp )
!
                where( singval(:n)<=tol ) singval(:n) = zero
!
                krank0 = count( singval(:n)/=zero )
!
!               FORM A RANDOM MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                a(:nr,:nc) = matmul( leftvec(:nr,:n)*spread(singval,1,nr), transpose(rightvec(:nc,:n)) )
!
!               SAVE DATA MATRIX.
!
                res(:nr,:nc) = a(:nr,:nc)
!
!               COMPUTE LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
                call lq_cmp( res(:nr,:nc), diagl(:n), tau(:n), use_qr=use_qr )
!
!               GENERATE ORTHOGONAL MATRIX Q OF LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
                q(:n,:nc) = res(:n,:nc)
                call ortho_gen_lq( q(:n,:nc), tau(:n), use_qr=use_qr )
!
!               RESTORE TRIANGULAR FACTOR L OF LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
                ll(1_i4b,1_i4b)    = diagl(1_i4b)
                ll(2_i4b:nr,1_i4b) = res(2_i4b:nr,1_i4b)
!
                do i = 2_i4b, n
!
                    ll(1_i4b:i-1_i4b,i) = zero
                    ll(i,i) = diagl(i)
                    ll(i+1_i4b:nr,i) = res(i+1_i4b:nr,i)
!
                end do
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:nr,:nc) - l(:nr,:n)*q(:n,:nc)
!               WHERE q AND l ARE THE ORTHOGONAL MATRIX Q AND L FACTOR FROM THE
!               LQ DECOMPOSITION OF a.
!
                res(:nr,:nc) = a(:nr,:nc) - matmul( ll(:nr,:n), q(:n,:nc)  )
!
                err1 = maxval(  norm( res(:nr,:nc) , dim=1)  /  norm( a(:nr,:nc) , dim=1) )/real(nc,stnd)
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q*q**(t)
!               WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE LQ DECOMPOSITION.
!
                res(:n,:n) = matmul( q(:n,:nc), transpose( q(:n,:nc) ) )
!
                do l = 1_i4b, n
                    res(l,l) = one - res(l,l)
                end do
!
                err2 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
!
                err = max( err1, err2 )
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j,  &
                                          ' , krank0= ', krank0, ' , use_qr= ', use_qr
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a - l*q
!                       WHERE q AND l ARE THE ORTHOGONAL MATRIX Q AND L FACTOR FROM THE
!                       LQ DECOMPOSITION OF a.
!
                        write (prtunit,*) 'Accuracy of the LQ decomposition       = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q*q**(t)
!                       WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE LQ DECOMPOSITION OF a .
!
                        write (prtunit,*) 'Orthogonality of the computed Q factor = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', nc,  &
                                      ' , case= ', j, ' , krank0= ', krank0,            &
                                      ' , use_qr= ', use_qr
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a - l*q
!                   WHERE q AND l ARE THE ORTHOGONAL MATRIX Q AND L FACTOR FROM THE
!                   LQ DECOMPOSITION OF a.
!
                    write (prtunit,*) 'Accuracy of the LQ decomposition       = ', &
                                     err1/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - q*q**(t)
!                   WHERE q IS THE ORTHOGONAL MATRIX Q FROM THE LQ DECOMPOSITION OF a .
!
                    write (prtunit,*) 'Orthogonality of the computed Q factor = ', &
                                     err2/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec, rightvec, res, ll, q )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta, diagl, tau )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_lq_cmp
! ==========================
!
end program test_lq_cmp
