program test_ortho_gen_p_bd
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines BD_CMP, ORTHO_GEN_Q_BD and ORTHO_GEN_P_BD
!   in module SVD_Procedures .
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking the accuracy
!    of subroutines BD_CMP, ORTHO_GEN_Q_BD and ORTHO_GEN_P_B.
!
! LATEST REVISION : 04/04/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,           &
                         zero, half, one, two, five, ten, c50, c100,       &
                         qr_cmp, ortho_gen_qr, bd_cmp, ortho_gen_q_bd,     &
                         ortho_gen_p_bd, norm, arth, lamch
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='ortho_gen_p_bd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, ulp, safmin, tmp, norma
    real(stnd), allocatable, dimension(:)   :: singval, diagr, beta, d, e, tauq, taup
    real(stnd), allocatable, dimension(:,:) :: a, bd, leftvec, rightvec, leftvec2, rightvec2, res
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, success
!
    logical(lgl) :: details
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( singval(n), diagr(n), beta(n),          &
                  d(n), e(n), tauq(n), taup(n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: BIDIAGONALIZATION OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: BIDIAGONALIZATION OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: BIDIAGONALIZATION OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: BIDIAGONALIZATION OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: BIDIAGONALIZATION OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), bd(n,n), res(nr,n),               &
                      leftvec(nr,n), rightvec(nc,n),              &
                      leftvec2(nr,nc), rightvec2(nc,n), stat=iok  )
!
            if ( iok/=0 ) cycle
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n) .
!
            call random_number( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
            do j = 1_i4b, 16_i4b
!
                m = m + 1_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b)   = one
                        d(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b:n-1_i4b) = one
                        d(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        d(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        d(1_i4b)         = ulp
                        d(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        d(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
                        do l = 1_i4b, n
                            d(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        d(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        d(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        d(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        d(1_i4b:i) = arth( one, one, i )
!
                        d(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        d(1_i4b:i) = arth( one, one, i )
!
                        d(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( d(:n) )
!
                        d(1_i4b:n) = 0.9_stnd + 0.1_stnd*d(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( d(1_i4b:n) ) )
!
                        d(1_i4b:n) = tmp + d(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( d(1_i4b:n) ) )
!
                        d(1_i4b:n) = tmp + d(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( d(1_i4b:n) ) )
!
                        d(1_i4b:n) = tmp + d(1_i4b:n)
!
                    case default
!
!                       DEFAULT TEST: RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( d(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                            d(1_i4b:n) = d(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( d ) ) then
                    if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( d(:n) )
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                a(:nr,:nc) = matmul( leftvec(:nr,:n)*spread(singval,1,nr), transpose(rightvec(:nc,:n)) )
!
!               COMPUTE THE NORM OF THE MATRIX.
!
!                norma = norm( a(:nr,:nc) )
                norma = norm( singval(:n) )
!
!               MAKE A COPY THE ORIGINAL MATRIX.
!
                leftvec2(:nr,:nc) = a(:nr,:nc)
!
!               CALL bd_cmp, ortho_gen_q_bd AND ortho_gen_p_bd TO REDUCE THE MATRIX TO BIDIAGONAL FORM
!
!                             a = Q*BD*P**(t)
!
!               WHERE Q AND P ARE ORTHOGONAL AND BD IS AN UPPER OR LOWER BIDIAGONAL MATRIX.
!
                call bd_cmp( leftvec2(:nr,:nc), d(:n), e(:n), tauq(:n), taup(:n) )
!
                call ortho_gen_p_bd( leftvec2(:nr,:nc), taup(:n), rightvec2(:nc,:n) )
!
                call ortho_gen_q_bd( leftvec2(:nr,:nc), tauq(:n) )
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*P - Q*BD.
!
                bd(:n,:n) = zero
!
                if ( nr>=nc ) then
!
!                   BD IS UPPER BIDIAGONAL.
!
                    do l = 1_i4b, n-1_i4b
                        bd(l,l)       = d(l)
                        bd(l,l+1_i4b) = e(l+1_i4b)
                    end do
!
                    bd(n,n) = d(n)
!
                else
!
!                   BD IS LOWER BIDIAGONAL.
!
                    bd(1_i4b,1_i4b) = d(1_i4b)
!
                    do l = 2_i4b, n
                        bd(l,l-1_i4b) = e(l)
                        bd(l,l)       = d(l)
                    end do
!
                endif
!
                res(:nr,:n) = matmul( a(:nr,:nc), rightvec2(:nc,:n) )   &
                              - matmul( leftvec2(:nr,:n), bd(:n,:n) )
!
                if ( norma/=zero ) then
                    err1 = maxval( norm(res(:nr,:n), dim=2_i4b) )/( norma*real(n,stnd) )
                else
                    err1 = maxval( norm(res(:nr,:n), dim=2_i4b) )/real(n,stnd)
                end if
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - P**(t)*P
!               WHERE P are THE LEFT ORTHOGONAL VECTORS OF THE BIDIAGONAL DECOMPOSITION.
!
                res(:n,:n) = matmul( transpose( leftvec2(:nr,:n) ), leftvec2(:nr,:n) )
!
                do l = 1_i4b, n
                    res(l,l) = one - res(l,l)
                end do
!
                err2 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q
!               WHERE Q are THE RIGHT ORTHOGONAL VECTORS OF THE BIDIAGONAL DECOMPOSITION.
!
                res(:n,:n) = matmul( transpose( rightvec2(:nc,:n) ), rightvec2(:nc,:n) )
!
                do l = 1_i4b, n
                    res(l,l) = one - res(l,l)
                end do
!
                err3 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
!
                err = max( err1, err2, err3 )
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*P - Q*BD.
!
                        write (prtunit,*) 'Accuracy of the computed bidiagonal decomposition        = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - P**(t)*P
!                       WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed left orthogonal vectors    = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q
!                       WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed right orthogonal vectors   = ', &
                                          err3/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', nc,' , case= ', j
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*P - Q*BD.
!
                    write (prtunit,*) 'Accuracy of the computed bidiagonal decomposition        = ', &
                                      err1/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - P**(t)*P
!                   WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                    write (prtunit,*) 'Orthogonality of the computed left orthogonal vectors    = ', &
                                      err2/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q
!                   WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                    write (prtunit,*) 'Orthogonality of the computed right orthogonal vectors   = ', &
                                      err3/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, bd, leftvec, rightvec, leftvec2, rightvec2, res )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( singval, diagr, beta, d, e, tauq, taup )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_ortho_gen_p_bd
! ==================================
!
end program test_ortho_gen_p_bd
