program test_phase_scramble_cor
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine PHASE_SCRAMBLE_COR in module Mul_Stat_Procedures .
!                                                                              
!                                                                              
! LATEST REVISION : 30/11/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, false, zero, half, one, comp_cor, phase_scramble_cor, pinvn
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
! p       IS THE NUMBER OF OBSERVATIONS OF THE TIME SERIES VECTORS
! nrep    IS THE NUMBER OF SHUFFLES FOR THE PHASE-SCRAMBLED BOOTSTRAP TEST
! nsample IS THE NUMBER OF SAMPLES TO EVALUATE THE REJECTION RATE OF THE TEST
!
    integer(i4b), parameter :: prtunit=6, p=50, nrep=99, nsample=3000
!
! sign_level IS THE SIGNIFICANCE LEVEL OF PHASE-SCRAMBLED BOOTSTRAP TEST
! eps        IS THE ALLOWED RELATIVE ERROR FOR THE REJECTION RATE
! b          IS THE LAG-1 AUTOCORRELATION FOR THE AR(1) MODEL USED
!            TO GENERATE THE TIME SERIES
!
    real(stnd),  parameter :: sign_level=0.05, eps=0.2, b=0.2
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                       :: xyn, err_prob, prob_rej_rate
    real(stnd), dimension(nsample)   :: xycor, prob
    real(stnd), dimension(nsample,2) :: xstat
    real(stnd), dimension(2)         :: ystat
    real(stnd), dimension(p,nsample) :: x, e
    real(stnd), dimension(nsample,p) :: x2
    real(stnd), dimension(p)         :: y, e2
!
    integer(i4b) :: i, j, rej_rate
!
    logical(lgl) :: first, last, norm
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='phase_scramble_cor'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   INITIALIZE THE RANDOM GENERATOR.
!
    call random_seed()
!
!   GENERATE A TIME SERIES USING AN AR(1) MODEL OF THE FORM
!
!               y(i+1) = b*y(i) + e2(i)
!
!     WHERE b IS THE SPECIFIED LAG-1 AUTOCORRELATION AND e2(i)
!     IS A NORMALLY DISTRIBUTED RANDOM VARIABLE WITH A 0 MEAN
!     AND A VARIANCE OF 1.
!
    call random_number( y(:p) )
!
    where( y(:p)<=zero .or. y(:p)>=one ) y(:p) = half
!
    e2(:p) = pinvn( y(:p) )
!
    y(1_i4b) = e2(1_i4b)
    do j=2_i4b, p
        y(j) = b*y(j-1_i4b) + e2(j)
    end do
!
    first = true
    last  = true
    norm  = true
!
    do i = 1_i4b, 6_i4b
!
!       GENERATE nsample INDEPENDENT TIME SERIES FROM THE SAME AR(1) MODEL.
!
        call random_number( x(:p,:nsample) )
!
        where( x(:p,:nsample)<=zero .or. x(:p,:nsample)>=one ) x(:p,:nsample) = half
!
        e(:p,:nsample) = pinvn( x(:p,:nsample) )
!
        x(1_i4b,:nsample) = e(1_i4b,:nsample)
        do j=2_i4b, p
            x(j,:nsample) = b*x(j-1_i4b,:nsample) + e(j,:nsample)
        end do
!
!       COMPUTE THE CORRELATIONS BETWEEN x AND y
!       FOR THE p OBSERVATIONS .
!
        call comp_cor( x(:p,:nsample), y(:p), first, last, xstat(:nsample,:2_i4b), ystat(:2_i4b),    &
                       xycor(:nsample), xyn, dimvar=2_i4b )
!
!       USE EMPIRICAL MARGINS FOR TESTS 4 TO 6.
!
        if ( i==4_i4b ) norm = false
!
!       NOW COMPUTE THE PHASE-SCRAMBLED BOOTSTRAP TEST OF THE CORRELATION
!       BETWEEN x AND y WITH SUBROUTINE phase_scramble_cor WITH nrep SHUFFLES .
!
        select case( i )
!
            case( 1_i4b, 4_i4b )
!
!               TESTS 1 AND 4  :
!
!               COMPUTE THE PHASE-SCRAMBLED BOOTSTRAP TEST WITH phase_scramble_cor_rv .
!
                do j = 1_i4b, nsample
                    call phase_scramble_cor( x(:p,j), y(:p), xstat(j,:2_i4b), ystat(:2_i4b),      &
                                             xycor(j), prob(j), nrep=nrep, norm=norm )
                end do
!
            case( 2_i4b, 5_i4b )
!
!               TESTS 2 AND 5  :
!
!               COMPUTE THE PHASE-SCRAMBLED BOOTSTRAP TEST WITH phase_scramble_cor_rm AND dimvar=2 .
!
                call phase_scramble_cor( x(:p,:nsample), y(:p), xstat(:nsample,:2_i4b), ystat(:2_i4b),          &
                                         xycor(:nsample), prob(:nsample), nrep=nrep, norm=norm, dimvar=2_i4b )
!
            case( 3_i4b, 6_i4b  )
!
!               TESTS 3 AND 6  :
!
!               COMPUTE THE PHASE-SCRAMBLED BOOTSTRAP TEST WITH phase_scramble_cor_rm AND dimvar=1 .
!
                x2(:nsample,:p) = transpose( x(:p,:nsample) )
                call phase_scramble_cor( x2(:nsample,:p), y(:p), xstat(:nsample,:2_i4b), ystat(:2_i4b),   &
                                         xycor(:nsample), prob(:nsample), nrep=nrep, norm=norm )
!
        end select
!
!       NOW COMPUTE THE REJECTION RATE OF THE TEST. IDEALLY, FOR THE sign_level
!       SIGNIFICANCE LEVEL, WE WOULD REJECT ONLY THE NULL HYPOTHESIS sign_level %
!       OF THE TIME. IN OTHER WORDS, THE REJECTION RATE MUST BE APPROXIMATELY EQUAL
!       TO THE SIGNIFICANCE LEVEL sign_level .
!
        rej_rate      = count( prob(:nsample)<=sign_level )
        prob_rej_rate = real( rej_rate, stnd )/real( nsample, stnd )
!
!       COMPUTE THE RELATIVE ERROR OF THE REJECTION RATE.
!
        err_prob = abs( (prob_rej_rate-sign_level)/sign_level )
!
!       CHECK THAT THE RELATIVE ERROR OF THE REJECTION RATE IS LESS THAN eps .
!
        if ( err_prob<=eps ) then
            write (prtunit,*) 'Test ',i,' is correct'
        else
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ',err_prob,' > ',eps
        end if
!
    end do
!
!
! END OF PROGRAM test_phase_scramble_cor
! ======================================
!
end program test_phase_scramble_cor
