program test_probgamma
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBGAMMA and  PINVGAMMA in module Prob_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 03/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, n1_def, n2_def, n3_def,    &
                         zero, c1_m11, one, ten, c2000,        &
                         allocate_error, merror, probgamma,    &
                         pinvgamma
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, min_eps IS THE MAXIMUM ACCURACY WHICH CAN BE
! OBTAINED.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, np=50, nprob=100
!   
    real(stnd), parameter :: fudge=ten
!   
    real(stnd), parameter :: min_eps=c1_m11
!
    character(len=*), parameter :: name_proc='probgamma'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: ulp, eps, err_max
    real(stnd), allocatable, dimension(:)   :: gamp, prob
    real(stnd), allocatable, dimension(:,:) :: xx, p, x
!
    integer(i4b) :: i, k, l, m, n, success
    integer      :: iok
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err_max )
    eps = max( fudge*sqrt(ulp), min_eps )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( gamp(np), prob(nprob), xx(nprob,np), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE THE p PARAMETER OF THE GAMMA DISTRIBUTION .
!
    gamp(1_i4b) = one
!
    call random_number( gamp(2_i4b:np) )
!
    gamp(2_i4b:np) = c2000*gamp(2_i4b:np)
!
    where( gamp(2_i4b:np)<one )  gamp(2_i4b:np) = gamp(2_i4b:np) + one
!
!   GENERATE A RANDOM PROBABILITY VECTOR prob(:nprob) .
!
    call random_number( prob(:nprob) )
!
!   COMPUTE THE REQUIRED LOWER TAILS QUANTILES.
!
    do i = 1_i4b, np
!
        do l = 1_i4b, nprob
!
!           COMPUTE THE GAMMA DEVIATES CORRESPONDING TO LOWER TAIL AREAS OF prob(:nprob) .
!
            xx(l,i) = pinvgamma( prob(l), gamp(i) )
!
        end do
!
    end do
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(nprob,n), x(nprob,n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       RECOMPUTE THE PROBABILITIES FROM THE GAMMA DEVIATES .
!
        do k = 1_i4b, 3_i4b
!
            m = m + 1_i4b
!
            err_max = zero
!
            select case( k )
!
            case( 1_i4b )
!
!               TEST 1 : SCALAR FORM OF FUNCTION probgamma.
!
                do i = 1_i4b, np
!
                    do l = 1_i4b, nprob
!
                        p(1_i4b,1_i4b) = probgamma( xx(l,i),  gamp(i) )
!
!                       CHECK THAT p(1,1) AND prob(l)  AGREE.
!
                        err_max = max( abs( p(1_i4b,1_i4b) - prob(l) ), err_max )
!
                    end do
!
                end do
!
            case( 2_i4b )
!
!               TEST 2 : VECTOR FORM OF FUNCTION probgamma.
!
                do i = 1_i4b, np
!
                    p(:nprob,1_i4b) = probgamma( xx(:nprob,i), gamp(i) )
!
!                   CHECK THAT p(:nprob,1) AND prob(:nprob)  AGREE.
!
                    err_max = max( maxval( abs(  p(:nprob,1_i4b) - prob(:nprob) ) ), err_max )
!
                end do
!
            case( 3_i4b )
!
!               TEST 3 : MATRIX FORM OF FUNCTION probgamma.
!
                do i = 1_i4b, np
!
                    x(:nprob,:n) = spread( xx(:nprob,i), dim=2, ncopies=n )
                    p(:nprob,:n) = probgamma( x(:nprob,:n), gamp(i) )
!
!                   CHECK THAT EACH COLUMN OF p(:nprob,:n) AGREES WITH prob(:nprob).
!
                    x(:nprob,:n) = spread( prob(:nprob), dim=2, ncopies=n )
!
                    err_max = max( maxval( abs(  p(:nprob,:n) - x(:nprob,:n) ) ), err_max )
!
                end do
!
            end select
!
!           PRINT THE RESULTS OF THE TESTS.
!
            if ( err_max<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err_max,' > ',eps,' , n= ', n,' , case= ', k
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, x )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( gamp, prob, xx )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probgamma
! =============================
!
end program test_probgamma
