program test_probn
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBN and PINVN in module Prob_Procedures.
!
!              
! LATEST REVISION : 03/10/2018
!                                                       
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, n1_def, n2_def, n3_def,           &
                         true, zero, c1_m7, half, one, c100, probn, pinvn
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, min_eps IS THE MAXIMUM ACCURACY WHICH CAN BE
! OBTAINED FROM THE pinvn FUNCTION.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c100
!   
    real(stnd), parameter  :: min_eps=c1_m7
!
    character(len=*), parameter :: name_proc='probn'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, ulp, eps
    real(stnd), allocatable, dimension(:,:) :: p, p2, x
!
    integer(i4b) :: i, j, k, m, n, success
    integer      :: iok
!
    logical(lgl) :: upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = max( fudge*sqrt(ulp), min_eps )
!
    success = 0_i4b
    m       = 0_i4b
!
    upper = true
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), x(n,n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM PROBABILITY MATRIX p .
!
        call random_number( p(:n,:n) )
!
        where( p(:n,:n)==zero .or. p(:n,:n)==one )  p(:n,:n) = half
!
        do i = 1_i4b, 6_i4b
!
            upper = .not.upper
!
            m = m + 1_i4b
!
            select case( i )
!
            case( 1_i4b )
!
!               TEST 1 :
!
!               COMPUTE THE NORMAL DEVIATES CORRESPONDING TO LOWER TAIL AREAS OF P .
!
                x(:n,:n) = pinvn( p(:n,:n) ) 
!
!               RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                p2(:n,:n) = probn( x(:n,:n), upper=upper )
!
            case( 2_i4b )
!
!               TEST 2 :
!
!               COMPUTE THE NORMAL DEVIATES CORRESPONDING TO UPPER TAIL AREAS OF P .
!
                x(:n,:n) = pinvn( one - p(:n,:n) ) 
!
!               RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                p2(:n,:n) = probn( x(:n,:n), upper=upper )
!
            case( 3_i4b )
!
!               TEST 3 :
!
                do j = 1_i4b, n
!
!                   COMPUTE THE NORMAL DEVIATES CORRESPONDING TO LOWER TAIL AREAS OF P(:,j) .
!
                    x(:n,j) = pinvn( p(:n,j) )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                    p2(:n,j) = probn( x(:n,j), upper=upper )
!
                end do
!
            case( 4_i4b )
!
!               TEST 4 :
!
                do j = 1_i4b, n
!
!                   COMPUTE THE NORMAL DEVIATES CORRESPONDING TO UPPER TAIL AREAS OF P(:,j) .
!
                    x(:n,j) = pinvn( one - p(:n,j) )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                    p2(:n,j) = probn( x(:n,j), upper=upper )
!
                end do
!
            case( 5_i4b )
!
!               TEST 5 :
!
                do j = 1_i4b, n
!
                    do k = 1_i4b, n
!
!                       COMPUTE THE NORMAL DEVIATE CORRESPONDING TO LOWER TAIL AREA OF P(k,j) .
!
                        x(k,j) = pinvn( p(k,j) )
!
!                       RECOMPUTE THE PROBABILITY FROM THE NORMAL DEVIATE .
!
                        p2(k,j) = probn( x(k,j), upper=upper )
!
                    end do
!
                end do
!
            case( 6_i4b )
!
!               TEST 6 :
!
                do j = 1_i4b, n
!
                    do k = 1_i4b, n
!
!                       COMPUTE THE NORMAL DEVIATE CORRESPONDING TO UPPER TAIL AREA OF P(k,j) .
!
                        x(k,j) = pinvn( one - p(k,j) )
!
!                       RECOMPUTE THE PROBABILITY FROM THE NORMAL DEVIATE .
!
                        p2(k,j) = probn( x(k,j), upper=upper )
!
                    end do
!
                end do
!
            end select
!
!           CHECK THAT p AND p2 AGREE.
!
            err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, x )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probn
! =========================
!
end program test_probn
