program test_probn2
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBN2 and PINVN2 in module Prob_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION 03/10/2018
!                                                 
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, extd, lgl, n1_def, n2_def, n3_def,   &
                         true, zero_extd, c1_m16_extd, half_extd,  &
                         one_extd, c1000_extd, probn2, pinvn2
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, min_eps IS THE MAXIMUM ACCURACY WHICH CAN BE
! OBTAINED FROM THE pinvn2 FUNCTION.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(extd), parameter  :: fudge=c1000_extd
!   
    real(extd), parameter  :: min_eps=c1_m16_extd
!
    character(len=*), parameter :: name_proc='probn2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(extd)                              :: err, ulp, eps
    real(extd), allocatable, dimension(:,:) :: p, p2, x
!
    integer(i4b) :: i, j, k, m, n, success
    integer      :: iok
!
    logical(lgl) :: upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = max( fudge*ulp, min_eps )
!
    success = 0_i4b
    m       = 0_i4b
!
    upper = true
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), x(n,n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM PROBABILITY MATRIX p .
!
        call random_number( p(:n,:n) )
!
        where( p(:n,:n)==zero_extd .or. p(:n,:n)==one_extd )  p(:n,:n) = half_extd
!
        do i = 1_i4b, 6_i4b
!
            upper = .not.upper
!
            m = m + 1_i4b
!
            select case( i )
!
            case( 1_i4b )
!
!               TEST 1 :
!
!               COMPUTE THE NORMAL DEVIATES CORRESPONDING TO LOWER TAIL AREAS OF P .
!
                x(:n,:n) = pinvn2( p(:n,:n) ) 
!
!               RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                p2(:n,:n) = probn2( x(:n,:n), upper=upper )
!
            case( 2_i4b )
!
!               TEST 2 :
!
!               COMPUTE THE NORMAL DEVIATES CORRESPONDING TO UPPER TAIL AREAS OF P .
!
                x(:n,:n) = pinvn2( one_extd - p(:n,:n) ) 
!
!               RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                p2(:n,:n) = probn2( x(:n,:n), upper=upper )
!
            case( 3_i4b )
!
!               TEST 3 :
!
                do j = 1_i4b, n
!
!                   COMPUTE THE NORMAL DEVIATES CORRESPONDING TO LOWER TAIL AREAS OF P(:,j) .
!
                    x(:n,j) = pinvn2( p(:n,j) )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                    p2(:n,j) = probn2( x(:n,j), upper=upper )
!
                end do
!
            case( 4_i4b )
!
!               TEST 4 :
!
                do j = 1_i4b, n
!
!                   COMPUTE THE NORMAL DEVIATES CORRESPONDING TO UPPER TAIL AREAS OF P(:,j) .
!
                    x(:n,j) = pinvn2( one_extd - p(:n,j) )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE NORMAL DEVIATES .
!
                    p2(:n,j) = probn2( x(:n,j), upper=upper )
!
                end do
!
            case( 5_i4b )
!
!               TEST 5 :
!
                do j = 1_i4b, n
!
                    do k = 1_i4b, n
!
!                       COMPUTE THE NORMAL DEVIATE CORRESPONDING TO LOWER TAIL AREA OF P(k,j) .
!
                        x(k,j) = pinvn2( p(k,j) )
!
!                       RECOMPUTE THE PROBABILITY FROM THE NORMAL DEVIATE .
!
                        p2(k,j) = probn2( x(k,j), upper=upper )
!
                    end do
!
                end do
!
            case( 6_i4b )
!
!               TEST 6 :
!
                do j = 1_i4b, n
!
                    do k = 1_i4b, n
!
!                       COMPUTE THE NORMAL DEVIATE CORRESPONDING TO UPPER TAIL AREA OF P(k,j) .
!
                        x(k,j) = pinvn2( one_extd - p(k,j) )
!
!                       RECOMPUTE THE PROBABILITY FROM THE NORMAL DEVIATE .
!
                        p2(k,j) = probn2( x(k,j), upper=upper )
!
                    end do
!
                end do
!
            end select
!
!           CHECK THAT p AND p2 AGREE.
!
            err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, x )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probn2
! ==========================
!
end program test_probn2
