program test_probq
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBQ and PINVQ in module Prob_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 01/03/2018
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, n1_def, n2_def, n3_def,   &
                         true, false, zero, half, one, c20, c99,   &
                         allocate_error, merror, probq, pinvq
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, ndf_max=10
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
!    real(stnd),  parameter :: eps = 1.0e-3_stnd 
!
    character(len=*), parameter :: name_proc='probq'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, ulp
    real(stnd), allocatable, dimension(:)   :: df_vec
    real(stnd), allocatable, dimension(:,:) :: prob, p, p2, x
!
    integer(i4b), allocatable, dimension(:) :: ndf_vec
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!
    logical(lgl) :: upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
    ulp = sqrt( err )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( df_vec(ndf_max), ndf_vec(ndf_max), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), prob(n,n), x(n,n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM PROBABILITY MATRIX p .
!
        call random_number( p(:n,:n) )
!
        where( p(:n,:n)<=ulp .or. p(:n,:n)>=one-ulp )  p(:n,:n) = half
!
!       GENERATE THE DEGREES OF FREEDOM OF CHI-SQUARED DISTRIBUTION .
!
        call random_number( df_vec(:ndf_max) )
!
        df_vec(:ndf_max) = c99*df_vec(:ndf_max)
!
!       WE TEST ONLY FOR NUMBER OF DEGREES OF FREEDOM GREATER THAN 20 .
!
        where( df_vec(:ndf_max)<c20 )  df_vec(:ndf_max) = df_vec(:ndf_max) + c20
!
        ndf_vec(:ndf_max) = int( df_vec(:ndf_max), i4b )
!
        do i = 1_i4b, 12_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( i )
!
            case( 1_i4b, 4_i4b, 7_i4b, 10_i4b )
!
                if ( i==1_i4b .or. i==7_i4b ) then
!
!                   TESTING LOWER TAIL FOR TESTS 1 TO 3 AND 7 TO 9.
!
                    upper = false
!
                    prob(:n,:n) = p(:n,:n)
!
                    if ( i==1_i4b ) then
!
!                       SET THE REQUIRED PRECISION OF THE RESULTS FOR TESTS 1 TO 6.
!
                        eps = 1.0e-5_stnd 
!
                    else
!
!                       ADJUST THE NUMBERS OF DEGREES OF FREEDOM TO TEST NORMAL APPROXIMATION.
!
                        ndf_vec(:ndf_max) = ndf_vec(:ndf_max) + 100_i4b
!
!                       SET THE REQUIRED PRECISION OF THE RESULTS FOR TESTS 7 TO 12.
!
                        eps = 1.0e-4_stnd 
!         
                    end if
!
                else
!
!                   TESTING UPPER TAIL FOR TESTS 4 TO 6 AND 10 TO 12.
!
                    upper = true
!
!                   ADJUST THE PROBABILITIES.
!
                    prob(:n,:n) = one - p(:n,:n)
!
                end if
!
!               TESTS 1, 4, 7 AND 10 : MATRIX FORM OF pinvq AND probq GENERIC FUNCTIONS.
!
                do l = 1_i4b, ndf_max
!
!                   COMPUTE THE QUANTILES x(:n,:n) OF CHI-SQUARED DISTRIBUTION WITH ndf_vec(l) DEGREES OF FREEDOM
!                   CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,:n) .
!
                    x(:n,:n) = pinvq( prob(:n,:n), ndf_vec(l) ) 
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p2(:n,:n) = probq( x(:n,:n), ndf_vec(l), upper=upper, ndf_max=100_i4b )
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 2_i4b, 5_i4b, 8_i4b, 11_i4b )
!
!               TESTS 2, 5, 8 AND 11 : VECTOR FORM OF pinvq AND probq GENERIC FUNCTIONS.
!
                do l = 1_i4b, ndf_max
!
                    do j = 1_i4b, n
!
!                       COMPUTE THE QUANTILES x(:n,j) OF CHI-SQUARED DISTRIBUTION WITH ndf_vec(l) DEGREES
!                       OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,j) .
!
                        x(:n,j) = pinvq( prob(:n,j), ndf_vec(l) )
!
!                       RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                        p2(:n,j) = probq( x(:n,j), ndf_vec(l), upper=upper, ndf_max=100_i4b )
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 3_i4b, 6_i4b, 9_i4b, 12_i4b )
!
!               TESTS 3, 6, 9 AND 12 : SCALAR FORM OF pinvq AND probq GENERIC FUNCTIONS.
!
                do l = 1_i4b, ndf_max
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
!
!                           COMPUTE THE QUANTILE x(k,j) OF CHI-SQUARED DISTRIBUTION WITH ndf_vec(l) DEGREES
!                           OF FREEDOM CORRESPONDING TO LOWER TAIL AREA OF prob(k,j) .
!
                            x(k,j) = pinvq( prob(k,j), ndf_vec(l) )
!
!                           RECOMPUTE THE PROBABILITY FROM THE QUANTILE .
!
                            p2(k,j) = probq( x(k,j), ndf_vec(l), upper=upper, ndf_max=100_i4b )
!
                        end do
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            end select
!
!           PRINT THE RESULTS OF THE TESTS.
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, prob, x )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( df_vec, ndf_vec )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probq
! =========================
!
end program test_probq
