program test_probstudent
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBSTUDENT in module Prob_Procedures.
!                                                                              
!
! Further Details
! ===============
!                                                                              
!    The program assumes that function PROBBETA in module Prob_Procedures is correct.
!                                                                              
! LATEST REVISION : 08/02/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, n1_def, n2_def, n3_def,                  &
                         zero, half, four, c99, c100, probstudent, probbeta, &
                         merror, allocate_error
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, ndf=10
!
    real(stnd),  parameter :: eps = 2.0e-3_stnd 
!
    character(len=*), parameter :: name_proc='probstudent'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err, a
    real(stnd), allocatable, dimension(:)     :: df_vec
    real(stnd), allocatable, dimension(:,:)   :: p, p2, x
    real(stnd), allocatable, dimension(:,:,:) :: t
!
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( df_vec(ndf), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), x(n,n), t(n,n,ndf), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM QUANTILE MATRIX t .
!
        call random_number( t(:n,:n,:ndf) )
!
        t(:n,:n,:ndf) = c100*(t(:n,:n,:ndf) - half)
!
!       GENERATE THE DEGREES OF FREEDOM OF STUDENT T-DISTRIBUTION .
!
        call random_number( df_vec(:ndf) )
!
        df_vec(:ndf) = c99*df_vec(:ndf)
!
!       WE TEST ONLY FOR DEGREES OF FREEDOM GREATER THAN 4.
!
        where( df_vec(:ndf)<four )  df_vec(:ndf) = df_vec(:ndf) + four
!
        do i = 1_i4b, 3_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( i )
!
            case( 1_i4b )
!
!               TEST 1 : MATRIX FORM OF probstudent GENERIC FUNCTION.
!
                do l = 1_i4b, ndf
!
!                   COMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p(:n,:n) = probstudent( t(:n,:n,l), df_vec(l) )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES
!                   WITH probbeta FUNCTION.
!
                    x(:n,:n) =  df_vec(l)/(df_vec(l)+ t(:n,:n,l)*t(:n,:n,l) )
!
                    a = half*df_vec(l)
!
                    p2(:n,:n) = probbeta( x(:n,:n), a, half )
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 2_i4b )
!
!               TEST 2 : VECTOR FORM OF probstudent GENERIC FUNCTION.
!
                do l = 1_i4b, ndf
!
                    a  = half*df_vec(l)
!
                    do j = 1_i4b, n
!
!                       COMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                        p(:n,j) = probstudent( t(:n,j,l), df_vec(l) )
!
!                       RECOMPUTE THE PROBABILITIES FROM THE QUANTILES
!                       WITH probbeta FUNCTION.
!
                        x(:n,j) =  df_vec(l)/(df_vec(l)+ t(:n,j,l)*t(:n,j,l) )
!
                        p2(:n,j) = probbeta( x(:n,j), a, half )
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 3_i4b )
!
!               TEST 3 : SCALAR FORM OF probstudent GENERIC FUNCTION.
!
                do l = 1_i4b, ndf
!
                    a = half*df_vec(l)
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
!
!                           COMPUTE THE PROBABILITY FROM THE QUANTILE .
!
                            p(k,j) = probstudent( t(k,j,l), df_vec(l) )
!
!                           RECOMPUTE THE PROBABILITIES FROM THE QUANTILES
!                           WITH probbeta FUNCTION.
!
                            x(k,j) =  df_vec(l)/(df_vec(l)+ t(k,j,l)*t(k,j,l) )
!
                            p2(k,j) = probbeta( x(k,j), a, half )
!
                        end do
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            end select
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, x, t )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( df_vec )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probstudent
! ===============================
!
end program test_probstudent
