program test_probt
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBT in module Prob_Procedures.
!                                                                              
!
! Further Details
! ===============
!                                                                              
!    The program assumes that function PROBBETA in module Prob_Procedures is correct.
!                                                                              
! LATEST REVISION : 21/05/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : i4b, stnd, lgl, n1_def, n2_def, n3_def,        &
                         true, false, zero, half, one, c50, c99, c100,  &
                         allocate_error, merror, probt, probbeta
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, ndf=10, ndf_max=c50
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    real(stnd),  parameter :: eps = 1.0e-3_stnd 
!
    character(len=*), parameter :: name_proc='probt'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err, a
    real(stnd), allocatable, dimension(:)     :: df_vec
    real(stnd), allocatable, dimension(:,:)   :: p, p2, x
    real(stnd), allocatable, dimension(:,:,:) :: t
!
    integer(i4b), allocatable, dimension(:) :: ndf_vec
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!
    logical(lgl) :: upper
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( df_vec(ndf), ndf_vec(ndf), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), x(n,n), t(n,n,ndf), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM QUANTILE MATRIX t .
!
        call random_number( t(:n,:n,:ndf) )
!
        t(:n,:n,:ndf) = c100*(t(:n,:n,:ndf) - half)
!
!       GENERATE THE DEGREES OF FREEDOM OF STUDENT T-DISTRIBUTION .
!
        call random_number( df_vec(:ndf) )
!
        df_vec(:ndf) = c99*df_vec(:ndf)
!
        where( df_vec(:ndf)<one )  df_vec(:ndf) = df_vec(:ndf) + one
!
        ndf_vec(:ndf) = int( df_vec(:ndf), i4b )
        df_vec(:ndf)  = real( ndf_vec(:ndf), stnd )
!
        do i = 1_i4b, 6_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( i )
!
            case( 1_i4b, 4_i4b )
!
                if ( i==1_i4b ) then
!
                    upper = false
!
                else
!
                    upper = true
!
                end if
!
!               TESTS 1 AND 4: MATRIX FORM OF probt GENERIC FUNCTION.
!
                do l = 1_i4b, ndf
!
!                   COMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p(:n,:n) = probt( t(:n,:n,l), ndf_vec(l), upper=upper, ndf_max=ndf_max )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES
!                   WITH probbeta FUNCTION.
!
                    a = half*df_vec(l)
!
                    x(:n,:n) = t(:n,:n,l)*t(:n,:n,l)
                    x(:n,:n) = x(:n,:n)/(df_vec(l)+ x(:n,:n) )
!
                    p2(:n,:n) = probbeta( x(:n,:n), half, a )
!
                    if ( upper ) then
!
                        where( t(:n,:n,l)<=zero )
                            p2(:n,:n) = half*( one + p2(:n,:n) )
                        elsewhere
                            p2(:n,:n) = half*( one - p2(:n,:n) )
                        end where
!
                    else
!
                        where( t(:n,:n,l)<=zero )
                            p2(:n,:n) = half*( one - p2(:n,:n) )
                        elsewhere
                            p2(:n,:n) = half*( one + p2(:n,:n) )
                        end where
!
                    end if
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 2_i4b, 5_i4b )
!
!               TESTS 2 AND 5 : VECTOR FORM OF probt GENERIC FUNCTION.
!
                do l = 1_i4b, ndf
!
                    a  = half*df_vec(l)
!
                    do j = 1_i4b, n
!
!                       COMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                        p(:n,j) = probt( t(:n,j,l), ndf_vec(l), upper=upper, ndf_max=ndf_max )
!
!                       RECOMPUTE THE PROBABILITIES FROM THE QUANTILES
!                       WITH probbeta FUNCTION.
!
                        x(:n,j) = t(:n,j,l)*t(:n,j,l)
                        x(:n,j) = x(:n,j)/(df_vec(l)+ x(:n,j) )
!
                        p2(:n,j) = probbeta( x(:n,j), half, a )
!
                        if ( upper ) then
!
                            where( t(:n,j,l)<=zero )
                                p2(:n,j) = half*( one + p2(:n,j) )
                            elsewhere
                                p2(:n,j) = half*( one - p2(:n,j) )
                            end where
!
                        else
!
                            where( t(:n,j,l)<=zero )
                                p2(:n,j) = half*( one - p2(:n,j) )
                            elsewhere
                                p2(:n,j) = half*( one + p2(:n,j) )
                            end where
!
                        end if
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 3_i4b, 6_i4b )
!
!               TESTS 3 AND 6 : SCALAR FORM OF probt GENERIC FUNCTION.
!
                do l = 1_i4b, ndf
!
                    a = half*df_vec(l)
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
!
!                           COMPUTE THE PROBABILITY FROM THE QUANTILE .
!
                            p(k,j) = probt( t(k,j,l), ndf_vec(l), upper=upper, ndf_max=ndf_max  )
!
!                           RECOMPUTE THE PROBABILITIES FROM THE QUANTILES
!                           WITH probbeta FUNCTION.
!
                            x(k,j) = t(k,j,l)*t(k,j,l)
!
                            if ( ( t(k,j,l)<=zero .and. upper ) .or. ( t(k,j,l)>zero .and. .not.upper ) ) then
                                x(k,j)  = x(k,j)/(df_vec(l)+ x(k,j) )
                                p2(k,j) = probbeta( x(k,j), half, a )
                                p2(k,j) = half*( one + p2(k,j) )
                            else
                                x(k,j)  = df_vec(l)/(df_vec(l)+ x(k,j) )
                                p2(k,j) = probbeta( x(k,j), a, half )
                                p2(k,j) = half*p2(k,j)
                            end if
!
                        end do
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            end select
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, x, t )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( df_vec, ndf_vec )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probt
! =========================
!
end program test_probt
