program test_ran_num
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines RANDOM_NUMBER_ and RANDOM_SEED_
!   in module Random. 
!                                                                              
! LATEST REVISION : 30/07/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, lgl, stnd, true, zero, one, three, c90, c1_m3,    &
                         base, arth, probn, probq, allocate_error, merror,      &
                         rand_integer32, random_number_, random_seed_
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, nk IS THE NUMBER OF ELEMENTS
! BY CLASS FOR CHI-SQUARE GOODNESS-OF-FIT TEST, THE OTHER VARIABLES ARE
! USED BY THE CHI-SQUARE TEST FOR EQUIDISTRIBUTION PROPERTY OF THE RNG.
!
    integer(i4b), parameter :: prtunit = 6, nk = 100_i4b, nobs=100000_i4b,   &
                               nfreq=10_i4b, nfreqp1 = nfreq + 1_i4b
!
! eps IS THE SIGNIFICANCE LEVEL FOR THE UNIFORMITY, EQUIDISTRIBUTION AND RANDOMNESS TESTS.
!
    real(stnd), parameter :: eps = c1_m3
!
    character(len=*), parameter :: name_proc='random_number_'
!   
!   
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                  :: y1, y2, y3, x, mean, std, prob, chi_sq, expctd
    real(stnd), allocatable, dimension(:)       :: harvest1, harvest2, bounds, sample
    real(stnd), allocatable, dimension(:,:,:,:) :: freq
!   
    integer(i4b) :: ndf, ndf2, nclass, nclassp1, i, j, k, l, m, n, o, p, runs, success, notdone
    integer(i4b), allocatable, dimension(:) :: int_sample, int_sample_orig
!   
    integer                            :: nseed, iok, alg
    integer, allocatable, dimension(:) :: histo, seed
!   
    logical(lgl) :: failure, details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
!   FIRST CHECK THAT THE RNGS GIVE THE SAME RESULTS AS THE ORIGINAL C
!   OR FORTRAN VERSIONS OF THE CODES FOR GENERATING INTEGERS.
!
!   ALLOCATE WORK VARIABLE FOR THE TESTS.
!
    allocate( int_sample(1000_i4b), int_sample_orig(5_i4b), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   alg IS THE NUMBER OF RNG.
!
    do alg = 1, 5
!
#ifdef _RANDOM_NOINT32
        if ( alg==1 .or. alg==2 ) then
!
!           IF 32 BIT INTEGERS ARE NOT AVAILABLE TESTING IS NOT POSSIBLE FOR RNG 1 AND 2.
!
            notdone = notdone + 1_i4b
!
            cycle
!
        else
#endif
!
            m = m + 1_i4b
!
!           SELECT THE RNG TO USE.
!
            call random_seed_( alg=alg )
!
!           GENERATE A SAMPLE OF 1000 INTEGERS.
!
            do i=1_i4b, 1000_i4b
                int_sample(i) = rand_integer32()
            end do
!
!           TEST THE LAST FIVE GENERATED INTEGERS WITH THOSE IN THE ORIGINAL RNG.
!
            select case( alg )
!
              case( 1 )
!
                int_sample_orig(1_i4b:5_i4b) = (/ -1390712647,  -187269233, -1644844144,  223674629,  1545121474 /)
!
              case( 2 )
!
                int_sample_orig(1_i4b:5_i4b) = (/ -50458538,  404295651,  297825295, -1238377686, -823192275  /)
!
              case( 3 )
!
                int_sample_orig(1_i4b:5_i4b) = (/ -461859560, 585543541, -548165902, 246995255, -785412809  /)
!
              case( 4 )
!
                int_sample_orig(1_i4b:5_i4b) = (/ -1651815433, -398763161, -1877971395, 1397735321, -834941650  /)
!
              case( 5 )
!
                int_sample_orig(1_i4b:5_i4b) = (/ -1285066579, -1186202517, -1541128585, -294185871, 1501963172 /)
!
            end select
!
            failure = any( int_sample(996_i4b:1000_i4b)/=int_sample_orig(1_i4b:5_i4b) )
!
!           CHECK THE RESULTS OF THE TESTS.
!
            if ( .not.failure ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',alg,' is correct'
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',alg,' is incorrect'
!
                write (prtunit,*) '=>  ', int_sample_orig(1_i4b), ' /= ', int_sample(996_i4b),    &
                                  ' or ', int_sample_orig(2_i4b), ' /= ', int_sample(997_i4b),    &
                                  ' or ', int_sample_orig(3_i4b), ' /= ', int_sample(998_i4b),    &
                                  ' or ', int_sample_orig(4_i4b), ' /= ', int_sample(999_i4b),    &
                                  ' or ', int_sample_orig(5_i4b), ' /= ', int_sample(1000_i4b),   &
                                  ' , RNG= ', alg
!
            end if
!
#ifdef _RANDOM_NOINT32
        end if
#endif
!
    end do
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( int_sample, int_sample_orig )
!
!   NOW TEST THE GENERATION OF REAL NUMBERS WITH THE RNGs.
!
!   nclass IS THE NUMBER OF CLASSES FOR THE CHI-SQUARE GOODNESS-OF-FIT TEST.
!
    nclass = base
!
    do
!
        if ( nclass>100_i4b ) exit
!
        nclass = nclass*base
!
    end do
!
!   n IS THE TOTAL NUMBER OF OBSERVATIONS IN THE SAMPLE,
!   ndf IS THE NUMBER OF DEGREES OF FREEDOM FOR THE CHI-SQUARE
!   GOODNESS-OF-FIT TEST.
!
    n   = nk*nclass
    ndf = nclass - 1_i4b
    nclassp1 = nclass + 1_i4b
!
!   ALLOCATE WORK VARIABLES FOR THE NEXT TESTS.
!
    allocate( bounds(nclassp1), sample(n), histo(nclass),                     &
              harvest1(20_i4b), harvest2(20_i4b),                             &
              freq(0_i4b:nfreq,0_i4b:nfreq,0_i4b:nfreq,0_i4b:nfreq), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   COMPUTE THE BOUNDS OF THE k CLASSES FOR THE CHI-SQUARE GOODNESS-OF-FIT TEST.
!
    bounds(:nclass) = arth( zero, one, nclass )/real( nclass, stnd )
    bounds(nclassp1) = nearest( one, one )
!
!   ndf2 AND expctd ARE, RESPECTIVELY, THE NUMBER OF DEGREES OF FREEDOM AND THE
!   EXPECTED VALUE FOR THE CHI-SQUARE EQUIDISTRIBUTION TEST.
!
    ndf2 = (nfreqp1**4) - 1_i4b
    expctd = real( nobs, stnd)/( ndf2 + 1_i4b )
!
!   alg IS THE RNG NUMBER.
!
    do alg = 1, 10
!
#ifdef _RANDOM_NOINT32
!
        if ( alg==1 .or. alg==2 .or.  alg==6 .or. alg==7 ) then
!
!           IF 32 BIT INTEGERS ARE NOT AVAILABLE TESTING IS NOT POSSIBLE FOR RNGs 1, 2, 6 AND 7.
!
            notdone = notdone + 1_i4b
!
            cycle
!
        end if
#endif
!
!       SELECT THE GENERATOR FOR THE NEXT 5 TESTS.
!
        call random_seed_( alg=alg )
!
!       SEED THE GENERATOR.
!
        call random_seed_()
!
!       GENERATE A VECTOR OF RANDOM NUMBERS.
!
        call random_number_( sample )
!
        do o = 1_i4b, 5_i4b
!
            m = m + 1_i4b
!
!           TEST THE RANDOM GENERATORS.
!
            select case( o )
!
                case( 1_i4b )
!
!                   TEST 1 : TEST FOR UNIFORMITY WITH CHI-SQUARE GOODNESS-OF-FIT TEST
!
!                   COMPUTE CHI-SQUARE TEST CRITERION.
!
                    do i = 1_i4b, nclass
                        histo(i) = count( sample(:n)>=bounds(i) .and. sample(:n)<bounds(i+1_i4b) )
                    end do
!
                    x = real( sum( (histo(:nclass)-nk)**2 ), stnd )/real( nk, stnd )
!
!                   COMPUTE PROBABILITY.
!
                    prob = probq( x, ndf=ndf, upper=true )
!
                    failure = prob<=eps
!
                case( 2_i4b )
!
!                   TEST 2 : TEST FOR RANDOMNESS WITH THE TOTAL NUMBERS OF RUN UP OR DOWN.
!
!                   COUNT THE TOTAL NUMBER OF RUNS.
!
                    runs = 0_i4b
!
                    do i = 2_i4b, n-1_i4b
!
                        y1 = sample(i-1_i4b)
                        y2 = sample(i)
                        y3 = sample(i+1_i4b)
!
                        if ( (y1<y2 .and. y3<y2) .or. (y1>y2 .and. y3>y2) ) runs = runs + 1_i4b
!
                    end do
!
!                   COMPUTE TEST CRITERION.
!
                    mean = real( 2_i4b*n - 1_i4b, stnd )/three
                    std = sqrt( real( 16_i4b*n - 29_i4b, stnd )/c90 )
                    x = abs( ( real( runs, stnd ) - mean )/std )
!
!                   COMPUTE PROBABILITY.
!
                    prob = probn( x, upper=true )
!
                    failure = prob<=eps
!
                case( 3_i4b )
!
!                   TEST 3 : TEST FOR EQUIDISTRIBUTION PROPERTY OF THE RNG.
!
                    call random_number_( x )
                    i = nfreqp1*x
!
                    call random_number_( x )
                    j = nfreqp1*x
!
                    call random_number_( x )
                    k = nfreqp1*x
!
                    freq(:,:,:,:) = zero
!
                    do p = 1_i4b, nobs
!
                        call random_number_( x )
!
                        l = nfreqp1*x
                        freq(i,j,k,l) = freq(i,j,k,l) + one
                        i = j
                        j = k
                        k = l
!        
                    end do
!
                    chi_sq = zero
!
                    do k = 0_i4b, nfreq
!
                        do j = 0_i4b, nfreq
!
                            do i = 0_i4b, nfreq
!
                                chi_sq = chi_sq + sum( ((freq(:,i,j,k) - expctd)**2)/expctd )
!
                            end do
!
                        end do
!
                    end do
!
!                   COMPUTE PROBABILITY.
!
                    prob = probq( chi_sq, ndf=ndf2, upper=true )
!
                    failure = prob<=eps
!
                case( 4_i4b )
!
!                   TEST 4 : TEST THAT random_number_ RETURNS VALUES IN THE INTERVAL ]0,1[ OR [0,1[.
!
#ifdef _RANDOM_WITH0
                    failure = any( sample(:)<zero .or. sample(:)>=one )
#else
                    failure = any( sample(:)<=zero .or. sample(:)>=one )
#endif
!
                case( 5_i4b )
!
!                   TEST 5 : TEST FOR REPEATABLE CALLS OF SUBROUTINE random_number_.
!
!                   GET THE SIZE OF THE SEED ARRAY OF THE CURRENT RNG.
!
                    call random_seed_( size=nseed )
!
!                   ALLOCATE SEED ARRAY FOR THE CURRENT RNG.
!
                    allocate( seed(nseed), stat=iok )

                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
!                   GET THE CURRENT STATE OF THE GENERATOR.
!
                    call random_seed_( get=seed(:) )
!
!                   GENERATE A FIRST VECTOR OF RANDOM NUMBERS.
!
                    call random_number_( harvest1(:) )
!
!                   RESET THE GENERATOR.
!
                    call random_seed_( put=seed(:) )
!
!                   GENERATE A SECOND VECTOR OF RANDOM NUMBERS.
!
                    call random_number_( harvest2(:) )
!
!                   THE TWO RANDOM NUMBERS VECTORS MUST MATCH.
!
                    failure = any( harvest1(:)/=harvest2(:) )
!
!                   DEALLOCATE SEED ARRAY.
!
                    deallocate( seed )
!
            end select
!
!           CHECK THE RESULTS OF THE TESTS.
!
            if ( .not.failure ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'RNG= ', alg,' , case= ', o
!
                    if ( o<=3_i4b ) then
!
!                       PRINT PROBABILITY OF THE TEST.
!
                        write (prtunit,*) 'Probability of the result = ', prob
!
                    end if
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ', m , ' is incorrect'
!
                if ( o<=3_i4b ) then
!
!                   PRINT PROBABILITY OF THE TEST.
!
                    write (prtunit,*) '=> ',prob,' <= ', eps, ' , RNG= ', alg,' , case= ', o
!
                else
!
                    write (prtunit,*) 'RNG= ', alg,' , case= ', o
!
                end if
!
                write (prtunit,*)
!
            end if
!
        end do
!
    end do
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( histo, bounds, sample, freq, harvest1, harvest2 )
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_ran_num
! ===========================
!
end program test_ran_num
