program test_real_fft
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines REAL_FFT and REAL_FFT_BACKWARD
!   in module FFT_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 21/05/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, n1_def, n2_def, n3_def,                      &
                         true, false, zero, c1000, half, one, init_fft, end_fft, &
                         fft, real_fft, real_fft_backward, lamch, norm
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c1000
!
    character(len=*), parameter :: name_proc='real_fft'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, ulp, safmin, tmp
    real(stnd), allocatable, dimension(:,:) :: x, x2, xr, xi
!   
    complex(stnd), allocatable, dimension(:,:) :: xc, xc2
!   
    integer(i4b), dimension(2) :: init_vec
    integer(i4b)               :: j, m, n, p, p1, p2, success
    integer                    :: iok
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*sqrt( ulp )
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        p2 = (n/2_i4b)
        p1 = p2 + 1_i4b
        p  = p2*2_i4b
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( x(p,p), x2(p,p), xr(p,p1), xi(p,p1), xc(p,p), xc2(p,p), stat=iok )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A REAL RANDOM NUMBER ARRAY.
!
        call random_number( x2(:p,:p) )
!
        x2(:p,:p) = x2(:p,:p) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x2(:p,:p) = x2(:p,:p)/tmp
        end if
!
        do j = 1_i4b, 4_i4b
!
            m = m + 1_i4b
!
            select case( j )
!
                case( 1_i4b )
!
!                   TEST 1: TEST 1D REAL_FFT AND REAL_FFT_BACKWARD .
!
                    x(:p,1_i4b) = x2(:p,1_i4b)
!
!                   INITIALIZE THE FFT SUBROUTINE.
!
                    call init_fft( p2 )
!
!                   TRANSFORM AND THEN INVERT THE SEQUENCE BACK.
!
                    call real_fft( x(:p,1_i4b), xc(:p1,1_i4b), true  )
!
                    xr(:p1,1_i4b) = real( xc(:p1,1_i4b) )
                    xi(:p1,1_i4b) = aimag( xc(:p1,1_i4b) )
!
                    call real_fft_backward( xr(:p1,1_i4b), xi(:p1,1_i4b), x(:p,1_i4b)  )
!
!                   CHECK THAT INVERSE(TRANSFORM(SEQUENCE)) = SEQUENCE.
!
                    x(:p,1_i4b) = x(:p,1_i4b) - x2(:p,1_i4b)
!
                    err = norm( x(:p,1_i4b) )/norm( x2(:p,1_i4b) )
!
                case( 2_i4b )
!
!                   TEST 2: TEST 2D REAL_FFT AND REAL_FFT_BACKWARD.
!
                    x(:p,:p) = x2(:p,:p)
!
!                   INITIALIZE THE FFT SUBROUTINE.
!
                    init_vec(:2_i4b) = (/ p, p2 /)
                    call init_fft( init_vec(:2_i4b), dim=2_i4b )
!
!                   TRANSFORM AND THEN INVERT THE SEQUENCE BACK.
!
                    call real_fft( x(:p,:p), xc(:p,:p1), true )
!
                    xr(:p,:p1) = real( xc(:p,:p1) )
                    xi(:p,:p1) = aimag( xc(:p,:p1) )
!
                    call real_fft_backward( xr(:p,:p1), xi(:p,:p1), x(:p,:p), 2_i4b  )
!
!                   CHECK THAT INVERSE(TRANSFORM(SEQUENCE)) = SEQUENCE.
!
                    x(:p,:p) = x(:p,:p) - x2(:p,:p)
!
                    err = maxval( norm( x(:p,:p), dim=1_i4b )/norm( x2(:p,:p), dim=1_i4b ) )
!
                case( 3_i4b )
!
!                   TEST 3: TEST 1D REAL_FFT AND FFT .
!
!                   GENERATE A COMPLEX RANDOM NUMBER ARRAY WITH A REAL FORWARD FFT TRANSFORM.
!
                    call random_number( xr(:p2,1_i4b) )
!
                    xr(:p2,1_i4b) = xr(:p2,1_i4b) - half
!
                    call random_number( tmp )
!
                    if ( tmp>safmin ) then
                        xr(:p2,1_i4b) = xr(:p2,1_i4b)/tmp
                    end if
!
                    call random_number( xi(:p2,1_i4b) )
!
                    xi(:p2,1_i4b) = xi(:p2,1_i4b) - half
!
                    call random_number( tmp )
!
                    if ( tmp>safmin ) then
                        xi(:p2,1_i4b) = xi(:p2,1_i4b)/tmp
                    end if
!
!                   ADJUST THE COMPLEX ARRAY SO THAT THE FFT TRANSFORM IS REAL.
!
                    xc(1_i4b,1_i4b)    = cmplx( xr(1_i4b,1_i4b), zero, stnd )
                    xc(2_i4b:p2,1_i4b) = cmplx( xr(2_i4b:p2,1_i4b), xi(2_i4b:p2,1_i4b), stnd )
                    xc(p1,1_i4b)       = cmplx( one, zero, stnd )
                    xc(p:p2+2_i4b:-1_i4b,1_i4b) = conjg( xc(2_i4b:p2,1_i4b) )
!
                    xc2(:p,1_i4b) = xc(:p,1_i4b)
!
!                   INITIALIZE THE FFT SUBROUTINE.
!
                    call init_fft( p )
!
!                   TRANSFORM THE SEQUENCE.
!
                    call fft( xc(:p,1_i4b), forward=true  )
!
                    call end_fft()
!
                    x(:p,1_i4b) = real( xc(:p,1_i4b) )
!
!                   INITIALIZE THE FFT SUBROUTINE.
!
                    call init_fft( p2 )
!
!                   INVERT THE SEQUENCE BACK.
!
                    call real_fft( x(:p,1_i4b), xc(:p1,1_i4b), false  )
!
                    xc(p:p2+2_i4b:-1_i4b,1_i4b) = conjg( xc(2_i4b:p2,1_i4b) )
!
!                   CHECK THAT INVERSE(TRANSFORM(SEQUENCE)) = SEQUENCE.
!
                    xc(:p,1_i4b) = xc(:p,1_i4b) - xc2(:p,1_i4b)
!
                    err = norm( xc(:p,1_i4b) )/norm( xc2(:p,1_i4b) )
!
                case( 4_i4b )
!
!                   TEST 4: TEST 2D REAL_FFT AND FFT .
!
!                   GENERATE A COMPLEX RANDOM NUMBER ARRAY WITH A REAL FORWARD FFT TRANSFORM BY ROW.
!
                    call random_number( xr(:p,:p2) )
!
                    xr(:p,:p2) = xr(:p,:p2) - half
!
                    call random_number( tmp )
!
                    if ( tmp>safmin ) then
                        xr(:p,:p2) = xr(:p,:p2)/tmp
                    end if
!
                    call random_number( xi(:p,:p2) )
!
                    xi(:p,:p2) = xi(:p,:p2) - half
!
                    call random_number( tmp )
!
                    if ( tmp>safmin ) then
                        xi(:p,:p2) = xi(:p,:p2)/tmp
                    end if
!
!                   ADJUST THE COMPLEX ARRAYS SO THAT THE FFT TRANSFORM OF THE ROWS IS REAL.
!
                    xc(:p,1_i4b)    = cmplx( xr(:p,1_i4b), zero, stnd )
                    xc(:p,2_i4b:p2) = cmplx( xr(:p,2_i4b:p2), xi(:p,2_i4b:p2), stnd )
                    xc(:p,p1)       = cmplx( one, zero, stnd )
                    xc(:p,p:p2+2_i4b:-1_i4b) = conjg( xc(:p,2_i4b:p2) )
!
                    xc2(:p,:p) = xc(:p,:p)
!
!                   INITIALIZE THE FFT SUBROUTINE.
!    
                    init_vec(:2_i4b) = (/ p, p /)
!
                    call init_fft( init_vec(:2_i4b), dim=2_i4b )
!
!                   TRANSFORM THE SEQUENCE.
!
                    call fft( xc(:p,:p), forward=true, dim=2_i4b  )
!
                    call end_fft()
!
                    x(:p,:p) = real( xc(:p,:p) )
!
!                   INITIALIZE THE FFT SUBROUTINE.
!
                    init_vec(:2_i4b) = (/ p, p2 /)
                    call init_fft( init_vec(:2_i4b), dim=2_i4b )
!
!                   TRANSFORM AND THEN INVERT THE SEQUENCE BACK.
!
                    call real_fft( x(:p,:p), xc(:p,:p1), false )
!
                    xc(:p,p:p2+2_i4b:-1_i4b) = conjg( xc(:p,2_i4b:p2) )
!
!                   CHECK THAT INVERSE(TRANSFORM(SEQUENCE)) = SEQUENCE.
!
                    xc(:p,:p) = xc(:p,:p) - xc2(:p,:p)
!
                    err = maxval( norm( xc(:p,:p), dim=1_i4b )/norm( xc2(:p,:p), dim=1_i4b ) )
!
            end select
!
!           PRINT THE RESULTS OF THE TESTS.
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , p= ', p,' , case= ', j
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        call end_fft()
!
        deallocate( x, x2, xr, xi, xc, xc2 )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_real_fft
! ============================
!
end program test_real_fft
