program test_svd_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SVD_CMP
!   in module SVD_Procedures .
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking the accuracy
!    of subroutine SVD_CMP.
!
! LATEST REVISION : 31/07/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,                       &
                         true, false, zero, half, one, two, five, ten, c50, c100,      &
                         qr_cmp, ortho_gen_qr, eigval_sort, svd_cmp, norm, arth, lamch
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='svd_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, ulp, safmin, tmp, norma
    real(stnd), allocatable, dimension(:)   :: eigval, singval, singval2, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, leftvec2, rightvec2, res
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, success, notdone
    logical(lgl) :: failure, details, perfect_shift, use_svd2
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), singval2(n),  &
                  diagr(n), beta(n), stat=iok          )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(singval2) )  deallocate( singval2 )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: SVD OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: SVD OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc),        res(nr,n),                 &
                      leftvec(nr,n),   rightvec(nc,n),            &
                      leftvec2(nr,nc), rightvec2(nc,n), stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
                if ( allocated(leftvec2) )  deallocate( leftvec2 )
                if ( allocated(rightvec2) ) deallocate( rightvec2 )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n) .
!
            call random_number( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
            do j = 1_i4b, 16_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
!               SORT THE SINGULAR VALUES.
!
                call eigval_sort( sort, singval(:n) )
!
                where ( singval(:n)==zero )
!
                    eigval(:n) = one
!
                elsewhere
!
                    eigval(:n) = singval(:n)
!
                end where
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                a(:nr,:nc) = matmul( leftvec(:nr,:n)*spread(singval,1,nr), transpose(rightvec(:nc,:n)) )
!
!               COMPUTE THE NORM OF THE MATRIX.
!
!                norma = norm( a(:nr,:nc) )
                norma = norm( singval(:n) )
!
!               COMPUTE FULL SVD OF RANDOM DATA MATRIX.
!
                perfect_shift = true
                use_svd2      = false
!
                do i = 1_i4b, 4_i4b
!
                    perfect_shift = .not.perfect_shift
!
                    if ( i==3_i4b ) then
!
                        if ( k==1_i4b .or. k==3_i4b ) exit
!
                        use_svd2 = true
!
                    end if
!
!                   MAKE A COPY THE ORIGINAL MATRIX.
!
                    leftvec2(:nr,:nc) = a(:nr,:nc)
!
!                   COMPUTE SVD OF RANDOM DATA MATRIX.
!
                    call svd_cmp( leftvec2(:nr,:nc), singval2(:n), failure, rightvec2(:nc,:n), &
                                   sort=sort, use_svd2=use_svd2, perfect_shift=perfect_shift )
!
                    if ( failure ) then
!
                        notdone = notdone + 1_i4b
!
                        cycle
!
                    end if
!
                    m = m + 1_i4b
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                   WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                   v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                    res(:nr,:n) = matmul( a(:nr,:nc), rightvec2(:nc,:n) )    &
                                  - leftvec2(:nr,:n)*spread(singval2(:n),1,nr)
!
                    if ( norma/=zero ) then
                        err1 = maxval( norm(res(:nr,:n), dim=2_i4b) )/( norma*real(nr,stnd) )
                    else
                        err1 = maxval( norm(res(:nr,:n), dim=2_i4b) )/real(nr,stnd)
                    end if
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                   WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                    res(:n,:n) = matmul( transpose( leftvec2(:nr,:n) ), leftvec2(:nr,:n) )
!
                    do l = 1_i4b, n
                        res(l,l) = one - res(l,l)
                    end do
!
                    err2 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(nr,stnd)
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                   WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                    res(:n,:n) = matmul( transpose( rightvec2(:nc,:n) ), rightvec2(:nc,:n) )
!
                    do l = 1_i4b, n
                        res(l,l) = one - res(l,l)
                    end do
!
                    err3 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(nc,stnd)
!
                    err = max( err1, err2, err3 )
!
                    if ( err<=eps ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j,  &
                                              ' , perfect_shift= ', perfect_shift,      &
                                              ' , use_svd2= ', use_svd2
!
!                           COMPUTE ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                            write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                              maxval( abs( singval2(:n)-singval(:n) ) )
!
!                           COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                            write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                              maxval( abs( singval2(:n)-singval(:n) )/abs(eigval(:n)) )
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                           WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                           v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                              err1/ulp, ' (must be less than :', fudge, ' )'
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                           WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                              err2/ulp, ' (must be less than :', fudge, ' )'
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                           WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                              err3/ulp, ' (must be less than :', fudge, ' )'
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', nc,' , case= ', j,   &
                                          ' , perfect_shift= ', perfect_shift,' , use_svd2= ', use_svd2,    &
                                          ' , failure= ', failure
!
!                       COMPUTE ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                        write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                          maxval( abs( singval2(:n)-singval(:n) ) )
!
!                       COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                        write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                          maxval( abs( singval2(:n)-singval(:n) )/abs(eigval(:n)) )
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                       WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                       v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                       WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                       WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                          err3/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec, rightvec, leftvec2, rightvec2, res )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, singval2, diagr, beta )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_svd_cmp
! ===========================
!
end program test_svd_cmp
