program test_symtrid_bisect
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SYMTRID_BISECT in module Eig_Procedures.
!                                                                              
!
! LATEST REVISION : 02/08/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,   &
                         true, false, zero, one, c50,              &
                         gen_symtrid_mat, lamch, symtrid_bisect
!
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!
    character(len=*), parameter :: name_proc='symtrid_bisect'
!   
    real(stnd), parameter  :: fudge=c50
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err1, err2, eps, eps1, eps2, ulp, safmin, abstol
    real(stnd), allocatable, dimension(:) :: d, e, eigval0, eigval, temp
!
    integer(i4b)     :: i, j, l, m, n, success, notdone
    integer          :: iok
!
    logical(lgl)     :: failure, details, high_accuracy, test_ok
!   
    character        :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err1 )
    eps    = fudge*ulp
!
    safmin = lamch('s')
    abstol = sqrt( safmin )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( d(n), e(n), eigval0(n), eigval(n), temp(n), stat=iok   )
!
        if ( iok/=0 ) then
!
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(eigval0) )   deallocate( eigval0 )
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(temp) )      deallocate( temp )
!
            cycle
!
        end if
!
        do i = 1_i4b, 17_i4b
!
!           GENERATE A TRIDIAGONAL MATRIX.
!           THE DIAGONAL ELEMENTS ARE STORED IN d.
!           THE OFF-DIAGONAL ELEMENTS ARE STORED IN e.
!           THE EIGENVALUES ARE STORED IN eigval0.
!
            call gen_symtrid_mat( i, d(:n), e(:n), eigval=eigval0(:n), sort=sort )
!
#ifdef _F2003
            if ( ieee_support_datatype( d ) ) then
                if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                if ( .not.all( ieee_is_normal( e(:n) ) ) ) cycle
            end if
#endif
!
            do j = 1_i4b, 2_i4b
!
                if ( j==1_i4b ) then
!
                    high_accuracy = true
!
!                   COMPUTE EIGENVALUES OF THE TRIDIAGONAL MATRIX TO HIGH ACCURACY.
!
                    call symtrid_bisect( d(:n), e(:n), l, eigval(:n), failure, abstol=abstol, sort=sort )
!
                else
!
                    high_accuracy = false
!
!                   COMPUTE EIGENVALUES OF THE TRIDIAGONAL MATRIX TO ABSOLUTE ACCURACY.
!
                    call symtrid_bisect( d(:n), e(:n), l, eigval(:n), failure, sort=sort )
!
                end if
!
                if ( failure ) then
!
                    notdone = notdone + 1_i4b
!
                    cycle
!
                end if
!
                m = m + 1_i4b
!
!               NOW, TEST ACCURACY OF THE RESULTS.
!
                where( eigval0(:n)/=zero )
                    temp(:n) = eigval0(:n)
                elsewhere
                    temp(:n) = one
                end where
!
!               ABSOLUTE ERRORS OF EIGENVALUES.
!
                err1 = maxval( abs( eigval(:n) - eigval0(:n) ) )
!
!               RELATIVE ERRORS OF EIGENVALUES.
!
                err2 = maxval( abs( (eigval(:n) - eigval0(:n))/temp(:n) ) )
!
                eps1 = real( n, stnd)*eps*maxval( abs( eigval0(:n) ) )
!
                test_ok = err1<=eps1
!
                if ( high_accuracy ) then
!
                    eps2 = real( n, stnd)*fudge*eps
!
                    test_ok = test_ok .and. err2<=eps2
!
                end if
!
                if ( test_ok ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
                        write (prtunit,*) 'n= ', n,' , case= ', i,' , high_accuracy= ', high_accuracy
!
                        write (prtunit,*) 'Absolute accuracy of the computed eigenvalues    = ', err1
                        write (prtunit,*) 'Relative accuracy of the computed eigenvalues    = ', err2
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    if ( high_accuracy ) then
!
                        write (prtunit,*) '=> ',err1,' > ',eps1,' or ', err2,' > ',eps2,' , n= ', n,    &
                             ' , case= ', i,' , high_accuracy= ', high_accuracy,' , failure= ', failure
!
                    else
!
                        write (prtunit,*) '=> ',err1,' > ',eps1,' , n= ', n,' , case= ', i,              &
                             ' , high_accuracy= ', high_accuracy,' , failure= ', failure
!
                    end if
!
                    write (prtunit,*) 'Absolute accuracy of the computed eigenvalues    = ', err1
                    write (prtunit,*) 'Relative accuracy of the computed eigenvalues    = ', err2
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, eigval0, eigval, temp )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_symtrid_bisect
! ==================================
!
end program test_symtrid_bisect
