module FFT_Procedures
! ##############################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ##############################################################################
!                                                                              *
! ******************************************************************************
! MODULE EXPORTING FAST FOURIER TRANSFORMS.                                    *
!                                                                              *
! LATEST REVISION : 27/09/2018                                                 *
!                                                                              *
! ##############################################################################
!                                                                              *
! ******************************************************************************
!                                                                              *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                              *
!                                                                              *
!  _OPENMP      FOR ACTIVATING OPENMP PARALLELIZATION                          *
!  _ALLOC       FOR ALLOCATING LOCAL VARIABLES INSTEAD OF                      *
!               PLACING THEM ON THE STACK IN SUBROUTINES AND                   *
!               FUNCTIONS                                                      *
!  _TRANSPOSE   FOR REPLACING THE transpose INTRINSIC FUNCTION                 *
!               WITH STATPACK transpose2 FUNCTION                              *
!                                                                              *
! ******************************************************************************
!
!
! USED MODULES 
! ============
!
    use Select_Parameters,  only : lgl, i4b, stnd, extd,   &
                                   npar_geop, npar2_geop
#ifdef _TRANSPOSE
    use Utilities,          only : transpose=>transpose2
#endif
#ifdef _OPENMP
    use Select_Parameters,  only : omp_limit
    use omp_lib,            only : omp_get_max_threads, omp_in_parallel, omp_get_nested
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!
!   use Logical_Constants
!   use Reals_Constants
!   use Char_Constants
!   use Utilities
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::  init_fft, fftxy, fft, fft_row, end_fft,         &
               real_fft, real_fft_forward, real_fft_backward 
!
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface init_fft
        module procedure    init_fft_dim, init_fftn, init_fft_cm, init_fft_ct
    end interface
!
    interface fftxy
        module procedure    fftxy_rv, fftxy_rm, fftxy_rt, fftxy_dim_rm, fftxy_dim_rt
    end interface
!
    interface fft
        module procedure    fftn, fft_cm, fft_ct, fft_dim_cm, fft_dim_ct
    end interface
!
    interface fft_row
        module procedure    fftn_row, fft_row_cm
    end interface
!
    interface real_fft
        module procedure    real_fft_rv, real_fft_rm
    end interface
!
    interface real_fft_forward
        module procedure    real_fft_forward_rv, real_fft_forward_rm
    end interface
!
    interface real_fft_backward
        module procedure    real_fft_backward_rv, real_fft_backward_rm
    end interface
!
!
! MODULE VARIABLES 
! ================
!
    complex(stnd), dimension(:), allocatable,   save :: y_1d, y2_1d, w_1d,   &
                                                        y_2d, y2_2d, w_2d,   &
                                                        y_3d, y2_3d, w_3d,   &
                                                        y2_n, w1_n, w2_n
    complex(stnd), dimension(:,:), allocatable, save :: y_n, v_n
!
    integer(i4b),    save :: n_1d=-1, nfft_1d=-1, lognfft_1d=-1,   &
                             n_2d=-1, nfft_2d=-1, lognfft_2d=-1,   &
                             n_3d=-1, nfft_3d=-1, lognfft_3d=-1,   &
                             n_n=-1,  nfft_n=-1,                   &
                             nfft1_n=-1, lognfft1_n=-1,            &
                             nfft2_n=-1, lognfft2_n=-1
!
!
! MODULE PARAMETERS 
! =================
#ifdef _OPENMP
!
!   BLOCK SIZE FOR THE PARALLEL TRANSPOSITION ALGORITHM.
!
    integer(i4b), parameter :: blksz_t=20_i4b
#endif
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                                INITIALIZATION SUBROUTINES
! =========================================================================================
!
!
    subroutine init_fft_dim( shap, dim )
!
! Purpose
! _______
!
!   Subroutine INIT_FFT sets up constants, the Chirp functions
!   and the Fourier transform of the Chirp functions for use by generic subroutines FFT,
!   FFTXY, FFT_ROW and REAL_FFT for a complex valued array of shape SHAP.
!
! Arguments
! _________
!
!   SHAP         (INPUT) integer(i4b), dimension(:)
!                Rank-one integer holding the shape of the complex valued array
!                to be transformed. Size(SHAP) must be less or equal to 3.
!
!   DIM          (INPUT, OPTIONAL) integer(i4b)
!                Eventually specifies the index for the Fourier transform.
!                Fourier transform on DIM-index-sections, only. DIM must be
!                less or equal to size(SHAP).
!
!
! Further Details
! _______________
!
!   INIT_FFT is first called to establish and transform the Chirp functions and
!   other constants. Then, subroutines FFT, FFTXY, FFT_ROW and REAL_FFT can be called
!   any number of times without the precalculated constants being destroyed; a further
!   call to INIT_FFT will only be necessary if Fourier transforms for a new length
!   (or shape) are required.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : fft_error1, fft_error2, fft_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in), dimension(:)  :: shap
    integer(i4b), intent(in), optional      :: dim
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='init_fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( shap )
!
    select case( m )
!
        case( :0_i4b )
!
             return
!
        case( 1_i4b:3_i4b )
!
            if ( any(shap<=0_i4b)  )  then
!
                call merror( name_proc//fft_error1 )
!
            end if
!
            if ( present(dim) ) then
!
                if ( dim<=0_i4b .or. dim>m ) then
!
                    call merror( name_proc//fft_error2 )
!
                end if    
!
                if ( m==1_i4b ) then
!
                    call init_fftn( shap(1) )
!
                else
!
                    call init_fft_cv( shap(dim) )
!
                end if
!
            else
!
                select case( m )
!
                    case( 1_i4b )
!
                        call init_fftn( shap(1) )
!
                    case( 2_i4b )
!
                        call init_fft_cm( shap(1), shap(2) )
!
                    case( 3_i4b )
!
                        call init_fft_ct( shap(1), shap(2), shap(3) )
!
                end select
!
            end if
!
        case( 4_i4b: )
!
             call merror( name_proc//fft_error3 )
!
    end select
!
!
! END OF SUBROUTINE init_fft_dim
! ______________________________
!
    end subroutine init_fft_dim
!
! =========================================================================================
!
    subroutine init_fftn( length1 )
!
! Purpose
! _______
!
!   Subroutine INIT_FFT sets up constants, the bit reverse tables, the Chirp function
!   and the Fourier transform of the Chirp function for use by generic subroutines FFT
!   and FFT_ROW for a series of length LENGTH1.
!
! Arguments
! _________
!
!   LENGTH1     (INPUT) integer(i4b)
!                The length of the complex valued sequence to be transformed.
!                LENGTH1 may be any positive integer.
!
!
! Further Details
! _______________
!
!   INIT_FFT is first called to establish and transform the Chirp function and
!   other constants. Then, subroutine FFT (or FFT_ROW) can be called any number of times
!   without the precalculated constants being destroyed; a further call to INIT_FFT will only
!   be necessary if a new length is required.
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83, see
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!          general length, Appl. Statist., 26 (3), 351-361.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error, fft_error4
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: length1
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)  :: lognfft
    integer       :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='init_fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENT.
!
    if ( length1<0_i4b ) then
!
        call merror( name_proc//fft_error4 )
!
    end if
!
    n_n = length1
!
!   COMPUTE POWERS OF TWO.
!
    nfft_n = 1_i4b
!
    do lognfft=0_i4b, bit_size( n_n ) - 2_i4b
!
        if ( n_n<=nfft_n ) exit
!
        nfft_n = 2_i4b*nfft_n
!
    end do
!
    if ( n_n/=nfft_n ) then
!
!       n_n IS NOT A POWER OF TWO. FIND INTEGER nfft_n
!       WHICH IS A POWER OF TWO AND AT LEAST 2*n_n .
!
        nfft_n = 2*nfft_n
        lognfft = lognfft + 1_i4b
!
    end if
!
    lognfft1_n = lognfft/2_i4b
!    lognfft2_n = merge( lognfft1_n, lognfft1_n+1_i4b, lognfft1_n*2_i4b==lognfft )
!
!BUG: For compiling STATPACK with NAG compiler
!     replace the preceding line with
!
    if ( lognfft1_n*2_i4b==lognfft ) then
!
        lognfft2_n = lognfft1_n
!
    else
!
        lognfft2_n = lognfft1_n+1_i4b
!
    end if
!
    nfft1_n    = 2_i4b**lognfft1_n
    nfft2_n    = nfft_n/nfft1_n
!
!   ALLOCATE WORK VARIABLES FOR THE BAILEY's FFT.
!
    if ( allocated( w1_n   ) )  deallocate( w1_n   )
    if ( allocated( w2_n   ) )  deallocate( w2_n   )
    if ( allocated( v_n    ) )  deallocate( v_n )
!
    allocate( w1_n(nfft1_n/2_i4b),            &
              w2_n(nfft2_n/2_i4b),            &
              v_n(nfft1_n,nfft2_n), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!   GENERATE POWERS OF ww = e**((-2*PI)/nfft) SUCH THAT
!
!       w(i) = ww**(i-1) .
!
    call gen_w_table( w1_n(:) )
    call gen_w_table( w2_n(:) )
!
!   GENERATE 2D TWIDDLE FACTORS.
!
    call gen_v_table( v_n(:,:) )
!
    if ( n_n/=nfft_n ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE CHIRP FUNCTION
!       AND TRANSFORM.
!
        if ( allocated( y_n  ) ) deallocate( y_n  )
        if ( allocated( y2_n ) ) deallocate( y2_n )
!
        allocate( y_n(nfft1_n,nfft2_n), y2_n(n_n), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE CHIRP FUNCTION AND TRANSFORM.
!
        call gen_chirp_tables_n( y_n(:,:), y2_n(:) )
!
    end if
!
!
! END OF SUBROUTINE init_fftn
! ___________________________
!
    end subroutine init_fftn
!
! =========================================================================================
!
    subroutine init_fft_cm( length1, length2 )
!
! Purpose
! _______
!
!   Subroutine INIT_FFT sets up constants, the Chirp functions
!   and the Fourier transforms of the Chirp functions for use by generic subroutines FFT
!   or FFTXY for a complex matrix of shape (LENGTH1,LENGTH2).
!
! Arguments
! _________
!
!   LENGTH1      (INPUT) integer(i4b)
!                The number of rows of the complex matrix to be transformed.
!                LENGTH1 may be any positive integer.
!
!   LENGTH2      (INPUT) integer(i4b)
!                The number of columns of the complex matrix to be transformed.
!                LENGTH2 may be any positive integer.
!
!
! Further Details
! _______________
!
!   INIT_FFT is first called to establish and transform the Chirp functions and
!   other constants. Then, subroutine FFT (or FFTXY) can be called any number of times
!   without the precalculated constants being destroyed; a further call to INIT_FFT will only
!   be necessary if a new shape is required.
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83, see
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error, fft_error5 
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: length1, length2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='init_fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENT.
!
    if ( length1<0_i4b .or. length2<0_i4b ) then
         call merror( name_proc//fft_error5 )
    end if
!
    n_1d = length1
    n_2d = length2
!
!   COMPUTE POWERS OF TWO.
!
    call powers_of_two( n_1d, nfft_1d, lognfft_1d )
    call powers_of_two( n_2d, nfft_2d, lognfft_2d )
!
!   ALLOCATE WORK VARIABLES FOR THE COOLEY-TUKEY FFT.
!
    if ( allocated( w_1d   ) )  deallocate( w_1d   )
    if ( allocated( w_2d   ) )  deallocate( w_2d   )
!
    allocate( w_1d(nfft_1d/2_i4b), w_2d(nfft_2d/2_i4b), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!   GENERATE POWERS OF ww = e**((-2*PI)/nfft) SUCH THAT
!
!       w(i) = ww**(i-1) .
!
    call gen_w_table( w_1d(:) )
    call gen_w_table( w_2d(:) )
!
    if ( n_1d/=nfft_1d ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE CHIRP FUNCTION
!       AND TRANSFORM.
!
        if ( allocated( y_1d  ) ) deallocate( y_1d  )
        if ( allocated( y2_1d ) ) deallocate( y2_1d )
!
        allocate( y_1d(nfft_1d), y2_1d(n_1d), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE CHIRP FUNCTION AND TRANSFORM.
!
        call gen_chirp_tables( lognfft_1d, w_1d(:), y_1d(:), y2_1d(:) )
!
    end if
!
    if ( n_2d/=nfft_2d ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE CHIRP FUNCTION
!       AND TRANSFORM.
!
        if ( allocated( y_2d  ) ) deallocate( y_2d  )
        if ( allocated( y2_2d ) ) deallocate( y2_2d )
!
        allocate( y_2d(nfft_2d), y2_2d(n_2d), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE CHIRP FUNCTION AND TRANSFORM.
!
        call gen_chirp_tables( lognfft_2d, w_2d(:), y_2d(:), y2_2d(:) )
!
    end if
!
!
! END OF SUBROUTINE init_fft_cm
! _____________________________
!
    end subroutine init_fft_cm
!
! =========================================================================================
!
    subroutine init_fft_ct( length1, length2, length3 )
!
! Purpose
! _______
!
!   Subroutine INIT_FFT sets up constants, the Chirp functions
!   and the Fourier transforms of the Chirp functions for use by generic subroutines FFT
!   or FFTXY for a complex 3d array of shape (LENGTH1,LENGTH2,LENGTH3).
!
!
! Arguments
! _________
!
!   LENGTH1      (INPUT) integer(i4b)
!                The extent in the first dimension of the complex array to be transformed.
!                LENGTH1 may be any positive integer.
!
!   LENGTH2      (INPUT) integer(i4b)
!                The extent in the second dimension of the complex array to be transformed.
!                LENGTH2 may be any positive integer.
!
!   LENGTH3      (INPUT) integer(i4b)
!                The extent in the third dimension of the complex array to be transformed.
!                LENGTH3 may be any positive integer.
!
!
! Further Details
! _______________
!
!   INIT_FFT is first called to establish and transform the Chirp functions and
!   other constants. Then, subroutine FFT (or FFTXY) can be called any number of times without
!   the precalculated constants being destroyed; a further call to INIT_FFT will only
!   be necessary if a new shape is required.
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83, see
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error, fft_error6
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: length1, length2, length3
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='init_fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENT.
!
    if ( length1<0_i4b .or. length2<0_i4b .or. length3<0_i4b ) then
        call merror( name_proc//fft_error6 )
    end if
!
    n_1d = length1
    n_2d = length2
    n_3d = length3
!
!   COMPUTE POWERS OF TWO.
!
    call powers_of_two( n_1d, nfft_1d, lognfft_1d )
    call powers_of_two( n_2d, nfft_2d, lognfft_2d )
    call powers_of_two( n_3d, nfft_3d, lognfft_3d )
!
!   ALLOCATE WORK VARIABLES FOR THE COOLEY-TUKEY FFT.
!
    if ( allocated( w_1d   ) )  deallocate( w_1d   )
    if ( allocated( w_2d   ) )  deallocate( w_2d   )
    if ( allocated( w_3d   ) )  deallocate( w_3d   )
!
    allocate( w_1d(nfft_1d/2_i4b),                 &
              w_2d(nfft_2d/2_i4b),                 &
              w_3d(nfft_3d/2_i4b), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE POWERS OF ww = e**((-2*PI)/nfft) SUCH THAT
!
!       w(i) = ww**(i-1) .
!
    call gen_w_table( w_1d(:) )
    call gen_w_table( w_2d(:) )
    call gen_w_table( w_3d(:) )
!
    if ( n_1d/=nfft_1d ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE CHIRP FUNCTION
!       AND TRANSFORM.
!
        if ( allocated( y_1d  ) ) deallocate( y_1d  )
        if ( allocated( y2_1d ) ) deallocate( y2_1d )
!
        allocate( y_1d(nfft_1d), y2_1d(n_1d), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE CHIRP FUNCTION AND TRANSFORM.
!
        call gen_chirp_tables( lognfft_1d, w_1d(:), y_1d(:), y2_1d(:) )
!
    end if
!
    if ( n_2d/=nfft_2d ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE CHIRP FUNCTION
!       AND TRANSFORM.
!
        if ( allocated( y_2d  ) ) deallocate( y_2d  )
        if ( allocated( y2_2d ) ) deallocate( y2_2d )
!
        allocate( y_2d(nfft_2d), y2_2d(n_2d), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE CHIRP FUNCTION AND TRANSFORM.
!
        call gen_chirp_tables( lognfft_2d, w_2d(:), y_2d(:), y2_2d(:) )
!
    end if
!
    if ( n_3d/=nfft_3d ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE CHIRP FUNCTION
!       AND TRANSFORM.
!
        if ( allocated( y_3d  ) ) deallocate( y_3d  )
        if ( allocated( y2_3d ) ) deallocate( y2_3d )
!
        allocate( y_3d(nfft_3d), y2_3d(n_3d), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE CHIRP FUNCTION AND TRANSFORM.
!
        call gen_chirp_tables( lognfft_3d, w_3d(:), y_3d(:), y2_3d(:) )
!
    end if
!
!
! END OF SUBROUTINE init_fft_ct
! _____________________________
!
    end subroutine init_fft_ct
!
!
! =========================================================================================
!                             REAL (FAST) FOURIER TRANSFORMS
! =========================================================================================
!
!
    subroutine real_fft_rv( vec, vect, forward)
!
! Purpose
! _______
!
!   Subroutine REAL_FFT computes the Fast Fourier Transform (FFT) for 
!   a real valued sequence VEC of even length.
!
!
! Arguments
! _________
!
!   VEC      (INPUT) real(stnd), dimension(:)
!            On entry, the real valued sequence to be transformed.
!
!            Size(VEC) must be an even (positive) integer.
!
!   VECT     (OUTPUT) complex(stnd), dimension(:)
!            On exit, a complex vector of length size(VEC)/2+1 containing
!            the first size(VEC)/2+1 coefficients of the Fourier transform  
!            of the real sequence VEC. These coefficients are the positive
!            frequency half of the full complex Fourier transform of the 
!            the real value sequence VEC.
!
!            VECT must verify  size(VECT) = size(VEC)/2 + 1.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!
! Further Details
! _______________
!
!   REAL_FFT computes the Forward Fourier Transform, VECT, according to the
!   following formula
!
!       VECT(j+1)  =  [ sum k=0 to nn-1 ]  VEC(k+1) exp( - i 2 pi j k / nn )
!
!   for j=0, 1, ..., nn/2 and where i=sqrt( -1 ), nn=size(VEC) and pi=3.1415923565...
!
!
!   REAL_FFT computes the Backward Fourier Transform, VECT, according to the
!   following formula
!
!       VECT(j+1)  =  (1/nn) [ sum k=0 to nn-1 ]  VEC(k+1) exp( 2 pi j k / nn )
!
!   for j=0, 1, ..., nn/2 and where i=sqrt( -1 ), nn=size(VEC) and pi=3.1415923565...
!
!   The remaining values of the Fourier Transform may be computed by the following
!   lines of code
!
!                nn   = size(VEC)
!
!                nnd2 = nn/2
!
!                vect(nn:nnd2+2:-1) = conjg( vect(2:nnd2) )
!
!   Before using REAL_FFT, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( size(vec)/2 )     
!
!   For more details on the Discrete Fourier Transform, see
!
!   (1) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!   (2) Bloomfield, P., 1976: Fourier analysis of time series- An introduction, 
!              John Wiley and Sons, New York.
!      
!      
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,         only : assert, merror
    use Char_Constants,    only : fft_error7, fft_error8
    use Reals_Constants,   only : zero, half, one, pi
    use Logical_Constants, only : false
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),    intent(in),  dimension(:) :: vec
!
    complex(stnd), intent(out), dimension(:) :: vect
!
    logical(lgl),  intent(in) :: forward
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)  :: m, md2, md4, md2p1, mp2, i, k
!
    real(stnd)    :: theta, tp
!
    complex(stnd) :: alph, ximag, s1, beta
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='real_fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m   = size( vec )
    md2 = m/2_i4b
!   
    if ( m/=2*md2 ) then
!   
         call merror( name_proc//fft_error7 )
!   
    end if
!
    md2p1 = md2 + 1_i4b
!   
    call assert( logical(int(size(vect),i4b)==md2p1,lgl),    &
                 name_proc )
!   
    if ( m==2_i4b ) then
!   
!       size(vec) IS EQUAL TO TWO.   
!   
        theta =  vec(1_i4b)
        tp    =  vec(2_i4b)
        vect(2_i4b) = cmplx( theta-tp, zero, kind=stnd )
        vect(1_i4b) = cmplx( theta+tp, zero, kind=stnd )
!   
    else
!   
!       size(vec) IS GREATER THAN TWO.   
!   
        if ( md2/=n_n ) then
            call merror( name_proc//fft_error8 )
        end if
!   
!       MOVE vec TO vect.   
!   
        vect(:md2) = cmplx( vec(1_i4b:m:2_i4b), vec(2_i4b:m:2_i4b) , kind=stnd )
!   
!       COMPUTE THE CENTER COEFFICIENT.   
!   
        tp          = sum( vec(1_i4b:m:2_i4b) ) - sum( vec(2_i4b:m:2_i4b) )
        vect(md2p1) = cmplx( tp, zero, kind=stnd )
!
!       CALL SUBROUTINE FFTN_ROW TO COMPUTE THE FOURIER TRANSFORM.
!
        call fftn_row( vect(:md2), false )
!
        vect(:md2) = vect(:md2)*real( md2, stnd )
!
        vect(1_i4b) = cmplx( real(vect(1_i4b))+aimag(vect(1_i4b)) , kind=stnd )
!
        md4 = md2p1/2_i4b
!   
        if ( md4>=2_i4b ) then
!
!          DECOMPOSE THE COMPLEX VECTOR vect INTO THE COMPONENTS 
!          OF THE TRANSFORM OF THE INPUT VECTOR vec.
!
           mp2   = md2 + 2_i4b
           theta = pi/real( md2, kind=stnd )
           tp    = theta
           ximag = cmplx( zero, one, kind=stnd )
!
           do k=2_i4b, md4
!   
               i       = mp2 - k
               s1      = conjg( vect(i) )
               alph    = vect(k) + s1
               beta    = ximag*( s1 - vect(k) )
               s1      = cmplx( cos(theta), sin(theta), kind=stnd )
               vect(k) = half*( alph + beta*s1 )
               vect(i) = half*conjg( alph - beta*s1 )
               theta   = theta + tp
!   
           end do           
!
       end if
!
    end if
!
    if ( forward ) then
!   
        vect(:md2p1) = conjg( vect(:md2p1) )
!   
    else
!   
        vect(:md2p1) = vect(:md2p1)*(one/real(m,stnd))
!   
    end if
!
!
! END OF SUBROUTINE real_fft_rv
! _____________________________
!
    end subroutine real_fft_rv
!
! =========================================================================================
!
    subroutine real_fft_rm( mat, matt, forward)
!
! Purpose
! _______
!
!   Subroutine REAL_FFT computes the Fast Fourier Transform (FFT) for 
!   each row of a real valued matrix MAT. size(mat,2) must be of even length.
!
!
! Arguments
! _________
!
!   MAT      (INPUT) real(stnd), dimension(:,:)
!            On entry, the real valued matrix to be transformed.
!            Size(MAT,2) must be an even (positive) integer.
!
!   MATT     (OUTPUT) complex(stnd), dimension(:,:)
!            On exit, a complex matrix of shape size(MAT,1) by size(MAT,2)/2+1.
!            each row of MATT contains the first size(MAT,2)/2+1 coefficients of
!            the Fourier transform of the corresponding row of the real matrix MAT.
!            These coefficients are the positive frequency half of the full complex
!            Fourier transform of the corresponding row of MAT.
!
!            The shape of MATT must verify:
!
!            - size(MATT,1) = size(MAT,1)
!            - size(MATT,2) = size(MAT,2)/2 + 1.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!
! Further Details
! _______________
!
!   REAL_FFT computes the Forward Fourier Transform, MATT, according to the
!   following formula
!
!       MATT(l,j+1)  =  [ sum k=0 to nn-1 ]  MAT(l,k+1) exp( - i 2 pi j k / nn )
!
!   for j=0, 1, ..., nn/2 , l=1, ..., size(MAT,1) and where i=sqrt( -1 ), nn=size(VEC)
!   and pi=3.1415923565...
!
!   REAL_FFT computes the Backward Fourier Transform, MATT, according to the
!   following formula
!
!       MATT(l,j+1)  =  (1/nn) [ sum k=0 to nn-1 ]  MAT(l,k+1) exp( 2 pi j k / nn )
!
!   for j=0, 1, ..., nn/2 , l=1, ..., size(MAT,1) and where i=sqrt( -1 ), nn=size(VEC)
!   and pi=3.1415923565...
!
!   The remaining values of the Fourier Transform may be computed by the following
!   lines of code
!
!                nn   = size(mat,2)
!
!                nnd2 = nn/2
!
!                matt(l,nn:nnd2+2:-1) = conjg( matt(l,2:nnd2) )
!
!   Before using REAL_FFT, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( (/ size(MAT,1), size(MAT,2)/2 /), dim=2)     
!
!   For more details on the Discrete Fourier Transform, see
!
!   (1) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!   (2) Bloomfield, P., 1976: Fourier analysis of time series- An introduction, 
!              John Wiley and Sons, New York.
!      
!      
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,         only : assert, merror
    use Char_Constants,    only : fft_error8, fft_error9
    use Reals_Constants,   only : zero, half, one, pi
    use Logical_Constants, only : false
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),    intent(in),  dimension(:,:) :: mat
!
    complex(stnd), intent(out), dimension(:,:) :: matt
!
    logical(lgl),  intent(in) :: forward
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, n, md2, md4, md2p1, mp2, i, k
!
    real(stnd)   :: theta, tp
!
    complex(stnd)                         ::  ximag, s2
    complex(stnd), dimension(size(mat,1)) :: alph, s1, beta
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='real_fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n   = size( mat, 1 )
    m   = size( mat, 2 )
!
!   TEST THE ARGUMENTS.
!
    md2 = m/2_i4b
!   
    if ( m/=2_i4b*md2 )  then
!
        call merror( name_proc//fft_error9 )
!
    end if
!
    md2p1 = md2 + 1_i4b
!   
    call assert( logical(int(size(matt,1),i4b)==n,lgl),        &
                 logical(int(size(matt,2),i4b)==md2p1,lgl),    &
                 name_proc )
!   
    if ( m==2_i4b ) then
!   
!       size(mat,2) IS EQUAL TO TWO.   
!   
        do i=1_i4b, n
!   
            theta         =  mat(i,1_i4b)
            tp            =  mat(i,2_i4b)
            matt(i,2_i4b) = cmplx( theta-tp, zero, kind=stnd )
            matt(i,1_i4b) = cmplx( theta+tp, zero, kind=stnd )
!   
        end do
!   
    else
!   
!       size(mat,2) IS GREATER THAN TWO.   
!   
        if ( md2/=n_1d ) then
!   
            call merror( name_proc//fft_error8 )
!   
        end if
!   
!       MOVE mat TO matt.   
!   
        matt(:n,:md2) = cmplx( mat(:n,1_i4b:m:2_i4b), mat(:n,2_i4b:m:2_i4b) , kind=stnd )
!   
!       COMPUTE THE CENTER COEFFICIENT.   
!   
        do i=1_i4b, n
!
            tp            = sum( mat(i,1_i4b:m:2_i4b) ) - sum( mat(i,2_i4b:m:2_i4b) )
            matt(i,md2p1) = cmplx( tp, zero, kind=stnd )
!
        end do
!
!       CALL SUBROUTINE FFT_ROW_cm TO COMPUTE THE FOURIER TRANSFORM.
!
        call fft_row_cm( matt(:n,:md2), false )
!
        matt(:n,:md2) = matt(:n,:md2)*real( md2, stnd )
!
        matt(:n,1_i4b) = cmplx( real(matt(:n,1_i4b))+aimag(matt(:n,1_i4b)), zero , kind=stnd )
!
        md4 = md2p1/2_i4b
!
        if ( md4>=2_i4b ) then
!
!          DECOMPOSE THE COMPLEX VECTOR vect INTO THE COMPONENTS 
!          OF THE TRANSFORM OF THE INPUT VECTOR vec.
!
           mp2   = md2 + 2_i4b
           theta = pi/real( md2, kind=stnd )
           tp    = theta
           ximag = cmplx( zero, one, kind=stnd )
!
           do k=2_i4b, md4
!
               i           = mp2 - k
               s1(:n)      = conjg( matt(:n,i) )
               alph(:n)    = matt(:n,k) + s1(:n)
               beta(:n)    = ximag*( s1(:n) - matt(:n,k) )
               s2          = cmplx( cos(theta), sin(theta), kind=stnd )
               matt(:n,k)  = half*( alph(:n) + s2*beta(:n) )
               matt(:n,i)  = half*conjg( alph(:n) - s2*beta(:n) )
               theta       = theta + tp
!
           end do           
!
       end if
!
    end if
!
    if ( forward ) then
!
        matt(:n,:md2p1) = conjg( matt(:n,:md2p1) )
!
    else
!
        matt(:n,:md2p1) = matt(:n,:md2p1)*(one/real(m,stnd))
!
    end if
!
!
! END OF SUBROUTINE real_fft_rm
! _____________________________
!
    end subroutine real_fft_rm
!
! =========================================================================================
!
    subroutine real_fft_forward_rv( vec, vecr, veci)
!
! Purpose
! _______
!
!   Subroutine REAL_FFT_FORWARD implements the forward discrete Fourier Transform for 
!   a real valued sequence VEC of general length.
!
!
! Arguments
! _________
!
!   VEC      (INPUT) real(stnd), dimension(:)
!            On entry, the real valued sequence to be transformed.
!
!   VECR     (OUTPUT) real(stnd), dimension(:)
!            On exit, the real part of the forward discrete Fourier Transform
!            of the sequence VEC.
!
!            VECR must verify  size(VECR) = size(VEC)/2 + 1.
!
!   VECI     (OUTPUT) real(stnd), dimension(:)
!            On exit, the imaginary part of the forward discrete Fourier Transform
!            of the sequence VEC.
!
!            VECI must verify  size(VECI) = size(VEC)/2 + 1.
!
!
! Further Details
! _______________
!
!   Only, the part of the discrete Fourier Transform corresponding to the positive
!   frequencies are computed and output in VECR and VECI.
!
!   The forward Discrete Fourier Transform is computed using Goertzel method.
!
!   For more details on the Goertzel method for computing the Discrete Fourier Transform.
!   For further details, see
!
!   (1) Goertzel, G., 1958: An Algorithm for the Evaluation of Finite Trigonometric Series,
!           The American Mathematical Monthly, Vol. 65, No. 1, pp. 34-35
!
!   (2) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,         only : assert
    use Reals_Constants,   only : zero, half, two, four, pi
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:) :: vec
    real(stnd),   intent(out), dimension(:) :: vecr, veci
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, i, k, km
!
    real(stnd)   :: a, b, c, d, e
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='real_fft_forward'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m  = size( vec )
    km = m/2_i4b
!
    i = km + 1_i4b
!
    call assert( logical(int(size(vecr),i4b)==i,lgl),    &
                 logical(int(size(veci),i4b)==i,lgl),    &
                 name_proc )
!   
#ifdef _OPENMP
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                 &
!$OMP         ,PRIVATE(i,k,a,b,c,d,e)       &
!$OMP         ,SHARED(m,km,vec,vecr,veci)
!
!   COMPUTE FILTERED FOURIER TRANSFORM OF THE SERIES USING GOERTZEL METHOD.
!
    c = pi/real( m, stnd )
!
!$OMP DO SCHEDULE(STATIC) 
!
    do k = 0_i4b, km
!
        e = c*real( k, stnd )
        d = -four*sin( e )**2
        a = zero
        b = zero
!
        do i = m, 1_i4b, -1_i4b
!
            a = a + b
            b = b + d*a+ vec(i)
!
        end do
!
        vecr(k+1_i4b) = b - half*d*a
        veci(k+1_i4b) = -a*sin( two*e )
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE real_fft_forward_rv
! _____________________________________
!
    end subroutine real_fft_forward_rv
!
! =========================================================================================
!
    subroutine real_fft_backward_rv( vecr, veci, vec)
!
! Purpose
! _______
!
!   Subroutine REAL_FFT_BACKWARD implements the (real) backward discrete Fourier
!   Transform for a complex valued sequence stored in the vector VECR (real part of
!   the complex sequence) and VECI (imaginary part of the sequence).
!   The resulting real discrete Fourier Transform is stored in the real vector VEC.
!
!   Size(VEC) which gives the size of the transform may be of general length.
!
!
! Arguments
! _________
!
!   VECR     (INPUT) real(stnd), dimension(:)
!            On entry, the real part of the complex sequence to be transformed.
!
!            VECR must verify  size(VECR) = size(VEC)/2 + 1.
!
!   VECI     (INPUT) real(stnd), dimension(:)
!            On entry, the imaginary part of the complex sequence to be transformed.
!
!            VECI must verify  size(VECI) = size(VEC)/2 + 1.
!
!   VEC      (OUTPUT) real(stnd), dimension(:)
!            On exit, the discrete Fourier transform real valued sequence.
!
!
! Further Details
! _______________
!
!   The backward Discrete Fourier Transform is computed using Goertzel method.
!
!   For more details on the Goertzel method for computing the Discrete Fourier Transform.
!   For further details, see
!
!   (1) Goertzel, G., 1958: An Algorithm for the Evaluation of Finite Trigonometric Series,
!           The American Mathematical Monthly, Vol. 65, No. 1, pp. 34-35
!
!   (2) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,         only : assert
    use Reals_Constants,   only : zero, one, half, two, four, pi
    use Logical_Constants, only : true, false
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:) :: vecr, veci
    real(stnd),   intent(out), dimension(:) :: vec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, i, k, km
!
    real(stnd)   :: a, b, c, d, e, ar, ai, br, bi
!
    logical(lgl) :: odd
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='real_fft_backward'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m  = size( vec )
!
    km = m/2_i4b
!
    if ( m==2_i4b*km ) then
!
        odd = false
!
    else
!
        odd = true
!
    end if
!
    i = km + 1_i4b
!
    call assert( logical(int(size(vecr),i4b)==i,lgl),    &
                 logical(int(size(veci),i4b)==i,lgl),    &
                 name_proc )
!   
#ifdef _OPENMP
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                            &
!$OMP         ,PRIVATE(i,k,a,b,c,d,e,ar,br,ai,bi)      &
!$OMP         ,SHARED(m,km,vec,vecr,veci,odd)
!
    c = pi/real( m, stnd )
!
!   SET ODD/EVEN NUMBER OF POINTS LAST SPECTRUM VALUE.
!
    if ( odd ) then
!
        b = vecr(km+1_i4b)
        a = veci(km+1_i4b)
!
    else
!
        b = half*vecr(km+1_i4b)
        a = zero
!
    end if
!
!   COMPUTE INVERSE FOURIER TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC) 
!
    do i = 0_i4b, m-1_i4b
!
        e  = c*real( i, stnd )
        d  = -four*sin( e )**2
        ar = b
        br = b
        ai = a
        bi = a
!
        do k = km, 2_i4b, -1_i4b
!
            br = br + d*ar + vecr(k)
            ar = ar + br
            bi = bi + d*ai + veci(k)
            ai = ai + bi
!
        end do
!
        vec(i+1_i4b) = two*(br-sin(two*e)*ai) + d*ar + vecr(1_i4b)
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
    vec(:m) = (one/real(m,stnd))*vec(:m)
!
!
! END OF SUBROUTINE real_fft_backward_rv
! ______________________________________
!
    end subroutine real_fft_backward_rv
!
! =========================================================================================
!
    subroutine real_fft_forward_rm( mat, matr, mati, dim )
!
! Purpose
! _______
!
!   Subroutine REAL_FFT_FORWARD computes the forward discrete Fourier Transform of
!   each row (DIM=2) or each column (DIM=1) of the real matrix MAT.
!
!   Size(MAT,DIM) which gives the size of the transform may be of general length.
!
!   The real parts of the forward discrete Fourier Transforms are stored (rowwise) in MATR
!   and the corresponding imaginary parts of the transforms are stored in MATI.
!
!
! Arguments
! _________
!
!   MAT      (INPUT) real(stnd), dimension(:,:)
!            On entry, the real valued sequences to be transformed.
!
!   MATR     (OUTPUT) real(stnd), dimension(:,:)
!            On exit, the real part of the forward discrete Fourier Transform
!            of the sequences stored in MAT.
!
!            The shape of MATR must verify:
!
!            - size(MATR,1) = size(MAT,3-DIM)
!            - size(MATR,2) = size(MAT,DIM)/2 + 1.
!
!   MATI     (OUTPUT) real(stnd), dimension(:,:)
!            On exit, the imaginary part of the forward discrete Fourier Transform
!            of the sequences stored in MAT.
!
!            The shape of MATI must verify:
!
!            - size(MATI,1) = size(MAT,3-DIM)
!            - size(MATI,2) = size(MAT,DIM)/2 + 1.
!
!   DIM      (INPUT) integer(i4b)
!            Specifies the index for the Fourier transform. If:
!
!            - DIM = 1 :  Fourier transform on first index.
!            - DIM = 2 :  Fourier transform on second index.
!
!
! Further Details
! _______________
!
!   Only, the parts of the discrete Fourier Transforms corresponding to the positive
!   frequencies are computed and output in MATR and MATI.
!
!   The forward Discrete Fourier Transform is computed using Goertzel method.
!
!   For more details on the Goertzel method for computing the Discrete Fourier Transform.
!   For further details, see
!
!   (1) Goertzel, G., 1958: An Algorithm for the Evaluation of Finite Trigonometric Series,
!           The American Mathematical Monthly, Vol. 65, No. 1, pp. 34-35
!
!   (2) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,         only : assert, merror
    use Char_Constants,    only : fft_error10
    use Reals_Constants,   only : zero, half, two, four, pi
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:,:) :: mat
    real(stnd),   intent(out), dimension(:,:) :: matr, mati
!
    integer(i4b), intent(in) :: dim
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, m, i, k, km
!
    real(stnd)                                  :: c, d, e
    real(stnd), dimension(size(mat,3-int(dim))) :: a, b
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='real_fft_forward'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( dim/=1_i4b .and. dim/=2_i4b ) then
        call merror( name_proc//fft_error10 )
    end if
!
    n  = size( mat, 3-int(dim) )
!
    call assert( logical(int(size(matr,1),i4b)==n,lgl),    &
                 logical(int(size(mati,1),i4b)==n,lgl),    &
                 name_proc )
!
    m  = size( mat, int(dim) )
    km = m/2_i4b
!
    i = km + 1_i4b
!
    call assert(  logical(int(size(matr,2),i4b)==i,lgl),  &
                  logical(int(size(mati,2),i4b)==i,lgl),  &
                  name_proc )
!   
#ifdef _OPENMP
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m*n>=omp_limit                      .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                     &
!$OMP         ,PRIVATE(i,k,a,b,c,d,e)           &
!$OMP         ,SHARED(n,m,km,dim,mat,matr,mati)
!
!   COMPUTE FILTERED FOURIER TRANSFORM OF THE SERIES USING GOERTZEL METHOD.
!
    c = pi/real( m, stnd )
!
    if ( dim==1_i4b ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
        do k = 0_i4b, km
!
            e     = c*real( k, stnd )
            d     = -four*sin( e )**2
            a(:n) = zero
            b(:n) = zero
!
            do i = m, 1_i4b, -1_i4b
!
                a(:n) = a(:n) + b(:n)
                b(:n) = b(:n) + d*a(:n)+ mat(i,:n)
!
            end do
!
            matr(:n,k+1_i4b) = b(:n) - half*d*a(:n)
            mati(:n,k+1_i4b) = -sin( two*e )*a(:n)
!
        end do
!
!$OMP END DO
!
    else
!
!$OMP DO SCHEDULE(STATIC) 
!
        do k = 0_i4b, km
!
            e     = c*real( k, stnd )
            d     = -four*sin( e )**2
            a(:n) = zero
            b(:n) = zero
!
            do i = m, 1_i4b, -1_i4b
!
                a(:n) = a(:n) + b(:n)
                b(:n) = b(:n) + d*a(:n)+ mat(:n,i)
!
            end do
!
            matr(:n,k+1_i4b) = b(:n) - half*d*a(:n)
            mati(:n,k+1_i4b) = -sin( two*e )*a(:n)
!
        end do
!
!$OMP END DO
!
    end if
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE real_fft_forward_rm
! _____________________________________
!
    end subroutine real_fft_forward_rm
!
! =========================================================================================
!
    subroutine real_fft_backward_rm( matr, mati, mat, dim )
!
! Purpose
! _______
!
!   Subroutine REAL_FFT_BACKWARD computes the (real) backward discrete Fourier Transform for 
!   complex valued sequences stored in the matrices MATR (real part of the sequences stored 
!   rowwise) and MATI (imaginary part of the sequences stored rowwise).
!   The resulting real discrete Fourier Transforms are stored in the rows (DIM=2) or the
!   columns (DIM=1) of the real matrix MAT.
!
!   Size(MAT,DIM) which gives the size of the transform may be of general length.
!
!
! Arguments
! _________
!
!   MATR     (INPUT) real(stnd), dimension(:,:)
!            On entry, the real part of the complex sequences to be transformed.
!
!            The shape of MATR must verify:
!
!            - size(MATR,1) = size(MAT,3-DIM)
!            - size(MATR,2) = size(MAT,DIM)/2 + 1.
!
!
!   MATI     (INPUT) real(stnd), dimension(:,:)
!            On entry, the imaginary part of the complex sequences to be transformed.
!
!            The shape of MATI must verify:
!
!            - size(MATI,1) = size(MAT,3-DIM)
!            - size(MATI,2) = size(MAT,DIM)/2 + 1.
!
!   MAT      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, the real backward discrete Fourier transforms
!            of the complex sequences stored rowwise in MATR and MATI.
!
!   DIM      (INPUT) integer(i4b)
!            Specifies the index for the Fourier transform. If:
!
!            - DIM = 1 :  Fourier transform on first index.
!            - DIM = 2 :  Fourier transform on second index.
!
!
! Further Details
! _______________
!
!   The backward Discrete Fourier Transform is computed using Goertzel method.
!
!   For more details on the Goertzel method for computing the Discrete Fourier Transform.
!   For further details, see
!
!   (1) Goertzel, G., 1958: An Algorithm for the Evaluation of Finite Trigonometric Series,
!           The American Mathematical Monthly, Vol. 65, No. 1, pp. 34-35
!
!   (2) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,         only : assert, merror
    use Char_Constants,    only : fft_error10
    use Reals_Constants,   only : zero, one, half, two, four, pi
    use Logical_Constants, only : true, false
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:,:) :: matr, mati
    real(stnd),   intent(out), dimension(:,:) :: mat
!
    integer(i4b), intent(in) :: dim
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, m, i, k, km
!
    real(stnd)                          :: c, d, e
    real(stnd), dimension(size(matr,1)) :: a, b, ar, ai, br, bi
!
    logical(lgl) :: odd
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='real_fft_backward'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( dim/=1_i4b .and. dim/=2_i4b ) then
!
        call merror( name_proc//fft_error10 )
!
    end if
!
    n  = size( mat, 3-int(dim) )
!
    call assert( logical(int(size(matr,1),i4b)==n,lgl),    &
                 logical(int(size(mati,1),i4b)==n,lgl),    &
                 name_proc )
!
    m  = size( mat, int(dim) )
!
    km = m/2_i4b
!
    if ( m==2_i4b*km ) then
!
        odd = false
!
    else
!
        odd = true
!
    end if
!
    i = km + 1_i4b
!
    call assert(  logical(int(size(matr,2),i4b)==i,lgl),  &
                  logical(int(size(mati,2),i4b)==i,lgl),  &
                  name_proc )
!   
#ifdef _OPENMP
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m*n>=omp_limit                      .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                            &
!$OMP         ,PRIVATE(i,k,a,b,c,d,e,ar,br,ai,bi)      &
!$OMP         ,SHARED(n,m,km,dim,mat,matr,mati,odd)
!
    c = pi/real( m, stnd )
!
!   SET ODD/EVEN NUMBER OF POINTS LAST SPECTRUM VALUE.
!
    if ( odd ) then
!
        b(:n) = matr(:n,km+1_i4b)
        a(:n) = mati(:n,km+1_i4b)
!
    else
!
        b(:n) = half*matr(:n,km+1_i4b)
        a(:n) = zero
!
    end if
!
!   COMPUTE INVERSE FOURIER TRANSFORM.
!
    if ( dim==1_i4b ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
        do i = 0_i4b, m-1_i4b
!
            e  = c*real( i, stnd )
            d  = -four*sin( e )**2
!
            ar(:n) = b(:n)
            br(:n) = b(:n)
            ai(:n) = a(:n)
            bi(:n) = a(:n)
!
            do k = km, 2_i4b, -1_i4b
!
                br(:n) = br(:n) + d*ar(:n) + matr(:n,k)
                ar(:n) = ar(:n) + br(:n)
                bi(:n) = bi(:n) + d*ai(:n) + mati(:n,k)
                ai(:n) = ai(:n) + bi(:n)
!
            end do
!
            mat(i+1_i4b,:n) = two*(br(:n)-sin(two*e)*ai(:n)) + d*ar(:n) + matr(:n,1_i4b)
!
        end do
!
!$OMP END DO
!
   else
!
!$OMP DO SCHEDULE(STATIC) 
!
        do i = 0_i4b, m-1_i4b
!
            e  = c*real( i, stnd )
            d  = -four*sin( e )**2
!
            ar(:n) = b(:n)
            br(:n) = b(:n)
            ai(:n) = a(:n)
            bi(:n) = a(:n)
!
            do k = km, 2_i4b, -1_i4b
!
                br(:n) = br(:n) + d*ar(:n) + matr(:n,k)
                ar(:n) = ar(:n) + br(:n)
                bi(:n) = bi(:n) + d*ai(:n) + mati(:n,k)
                ai(:n) = ai(:n) + bi(:n)
!
            end do
!
            mat(:n,i+1_i4b) = two*(br(:n)-sin(two*e)*ai(:n)) + d*ar(:n) + matr(:n,1_i4b)
!
        end do
!
!$OMP END DO
!
   end if
!
!$OMP END PARALLEL
!
    mat(:,:) = ( one/real(m,stnd) )*mat(:,:)
!
!
! END OF SUBROUTINE real_fft_backward_rm
! ______________________________________
!
    end subroutine real_fft_backward_rm
!
! =========================================================================================
!
    subroutine fftxy_rv( x, y, fftx, ffty)
!
! Purpose
! _______
!
!   Given two real valued sequences of the same length, X and Y, FFTXY returns
!   the Fast Fourier Transforms of these sequences in the two complex valued sequences
!   FFTX and FFTY.
!
!
! Arguments
! _________
!
!   X, Y         (INPUT) real(stnd), dimension(:)
!                On entry, the real valued sequences to be transformed.
!
!                X and Y must verify  size(X) = size(Y).
!
!   FFTX, FFTY   (OUTPUT) complex(stnd), dimension(:)
!                On exit, FFTX, FFTY are replaced by the Fourier transforms
!                of X and Y, respectively.
!
!                FFTX and FFTY must verify size(FFTX) = size(FFTY) = size(X) = size(Y).
!
!
! Further Details
! _______________
!
!   Size(FFTX) = size(FFTY) = size(X) = size(Y) may be of general length.
!
!   If size(X) is an exact power of two, Bailey's Four-Step FFT algorithm is used,
!   otherwise the CHIRP-Z transform is employed.
!
!   Before using FFTXY, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( size(X) )     
!   
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83 and
!   Bailey (1990). For more details, see:
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!   (2) Bailey, D., 1990: FFTs in External or Hierarchical Memory, 
!           The Journal of Supercomputing, 4, 23-35.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Char_Constants,    only : fft_error8
    use Reals_Constants,   only : zero, half
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),   dimension(:) :: x, y
!
    complex(stnd), intent(out), dimension(:) :: fftx, ffty
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(n_n)    :: zsum
!
    integer(i4b) :: length1
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fftxy'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    length1 =  assert_eq( int(size(x),i4b) ,          &
                          int(size(y),i4b) ,          &
                          int(size(fftx),i4b)     ,   &
                          int(size(ffty),i4b)     ,   &
                          name_proc )
!
    if ( length1/=n_n ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
    fftx(:n_n) = cmplx( x(:n_n), y(:n_n), kind=stnd )
!
    call fftn_row( fftx(:n_n), true )
!
!   SEPARATE THE TWO TRANSFORMS.
!
    ffty(1_i4b) = cmplx( aimag( fftx(1_i4b) ), zero, kind=stnd )
    fftx(1_i4b) = cmplx( real( fftx(1_i4b) ),  zero, kind=stnd )
!
    ffty(2_i4b:n_n) = fftx(n_n:2_i4b:-1_i4b)
    zsum(2_i4b:n_n) = half*( fftx(2_i4b:n_n) + ffty(2_i4b:n_n) )
    ffty(2_i4b:n_n) = half*( fftx(2_i4b:n_n) - ffty(2_i4b:n_n) )
!
    fftx(2_i4b:n_n) = cmplx( real( zsum(2_i4b:n_n) ) , aimag( ffty(2_i4b:n_n) ), kind=stnd )
    ffty(2_i4b:n_n) = cmplx( aimag( zsum(2_i4b:n_n) ), -real( ffty(2_i4b:n_n) ), kind=stnd )
!
!
! END OF SUBROUTINE fftxy_rv
! __________________________
!
    end subroutine fftxy_rv
!
! =========================================================================================
!
    subroutine fftxy_rm( x, y, fftx, ffty )
!
! Purpose
! _______
!
!   Given two real valued matrices of the same shape, X and Y, FFTXY returns
!   the Fast Fourier Transforms of X and Y in the two complex valued matrices 
!   FFTX and FFTY, respectively.
!   
!
! Arguments
! _________
!
!   X, Y         (INPUT) real(stnd), dimension(:,:)
!                On entry, the real valued matrices to be transformed.
!
!                X and Y must verify the equality: shape(X) = shape(Y).
!
!   FFTX, FFTY   (OUTPUT) complex(stnd), dimension(:,:)
!                On exit, FFTX and FFTY are replaced
!                by the Fourier transforms of X and Y, respectively.
!
!                FFTX and FFTY must verify the equalities shape(FFTX) = shape(FFTY) = shape(X) = shape(Y).
!
!
! Further Details
! _______________
!
!   Depending if size(X,1) and size(X,2) are exact powers of two or not, a
!   radix-2 decimation-in-time Cooley-Tukey algorithm or a CHIRP-Z transform is
!   employed.
!
!   Before using FFTXY, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( size(X,1), size(X,2) )     
!
!   For more details, see:
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!   (2) Cooley, J.W., Lewis, P., and Welch, P., 1969: The Fast Fourier Transform
!           and its Applications", IEEE Trans on Education, 12, 1, 28-34.
!
!   (3) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half
    use Char_Constants,    only : allocate_error, fft_error8
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:) :: x, y
!
    complex(stnd), intent(out), dimension(:,:) :: fftx, ffty
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(:),   allocatable  :: tmpvec, tmpvec2
    complex(stnd), dimension(:,:), allocatable  :: tmp
!
    integer(i4b) :: m, n, i, k
!
    integer  :: iok
!
#ifdef _OPENMP
    integer(i4b) :: i2, k2
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fftxy'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(x,1),i4b) ,          &
                    int(size(y,1),i4b) ,          &
                    int(size(fftx,1),i4b)     ,   &
                    int(size(ffty,1),i4b)     ,   &
                    name_proc )
!
    n =  assert_eq( int(size(x,2),i4b) ,          &
                    int(size(y,2),i4b) ,          &
                    int(size(fftx,2),i4b)     ,   &
                    int(size(ffty,2),i4b)     ,   &
                    name_proc )
!
    if ( m/=n_1d .or.  n/=n_2d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
!   ALLOCATE WORK VARIABLE.
!
    allocate( tmp(n,m), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
#ifdef _OPENMP
    i           = nfft_2d*nfft_1d
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  i>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,i2,k,k2,iok,tmpvec,tmpvec2),   &
!$OMP          SHARED(lognfft_1d,lognfft_2d)
!
!   ALLOCATE WORK VARIABLES.
!
    k = max( nfft_1d, nfft_2d )
!
    allocate( tmpvec(k), tmpvec2(m), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        fftx(:m,i) = cmplx( x(:m,i), y(:m,i), kind=stnd  )
!
    end do
!
!$OMP END DO            
!
!   TRANSPOSE THE ARRAY.
!
#ifdef _OPENMP
    if ( test_par ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
        do k = 1_i4b, n, blksz_t
!
            do i = 1_i4b, m, blksz_t
!
                k2 = min( n, k+blksz_t-1_i4b )
                i2 = min( m, i+blksz_t-1_i4b )
!
                tmp(k:k2,i:i2) = transpose( fftx(i:i2,k:k2) )
!
            end do
!
        end do
!
!$OMP END DO
!
    else
#endif
!
        tmp(:n,:m) = transpose( fftx(:m,:n) )
!
#ifdef _OPENMP
    end if
#endif
!
!   TRANSFORM SECOND SUBSCRIPT.
!
    if ( n_2d==nfft_2d ) then
!
!       n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, m
!
            call fft2( tmp(:n,i), w_2d(:), lognfft_2d )
!
        end do
!
!$OMP END DO
!
    else
!
!       n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, m
!
            tmpvec(:n) = tmp(:n,i)
!
            call chirp( tmpvec(:nfft_2d), w_2d(:), y_2d(:), y2_2d(:), lognfft_2d )
!
            tmp(:n,i) = tmpvec(:n)
!
        end do
!
!$OMP END DO
!
    end if
!
!   TRANSPOSE BACK THE ARRAY.
!
#ifdef _OPENMP
    if ( test_par ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
        do k = 1_i4b, m, blksz_t
!
            do i = 1_i4b, n, blksz_t
!
                k2 = min( m, k+blksz_t-1_i4b )
                i2 = min( n, i+blksz_t-1_i4b )
!
                fftx(k:k2,i:i2) = transpose( tmp(i:i2,k:k2) )
!
            end do
!
        end do
!
!$OMP END DO
!
    else
#endif
!
        fftx(:m,:n) = transpose( tmp(:n,:m) )
!
#ifdef _OPENMP
    end if
#endif
!
!   SEPARATE THE TWO TRANSFORMS.
!
!$OMP DO SCHEDULE(STATIC)
!
    do k = 2_i4b, n
!
        i = n - k + 2_i4b
!
        ffty(:m,k) = fftx(:m,i)
!
    end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC)
!
    do k = 2_i4b, n
!
        tmpvec(:m)  = ffty(:m,k)
        tmpvec2(:m) = half*( fftx(:m,k) + tmpvec(:m) )
        tmpvec(:m)  = half*( fftx(:m,k) - tmpvec(:m) )
        fftx(:m,k)  = cmplx( real(  tmpvec2(:m) ), aimag( tmpvec(:m) ), kind=stnd )
        ffty(:m,k)  = cmplx( aimag( tmpvec2(:m) ), -real( tmpvec(:m) ), kind=stnd )
!
    end do
!
!$OMP END DO NOWAIT
!
!$OMP SINGLE
!
    tmpvec(:m)     = fftx(:m,1_i4b)
    fftx(:m,1_i4b) = cmplx( real( tmpvec(:m)  ), zero, kind=stnd )
    ffty(:m,1_i4b) = cmplx( aimag( tmpvec(:m) ), zero, kind=stnd )
!
!$OMP END SINGLE
!
!   TRANSFORM FIRST SUBSCRIPT.
!
    if ( n_1d==nfft_1d ) then
!
!       m IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            call fft2( fftx(:m,k), w_1d(:), lognfft_1d )
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            call fft2( ffty(:m,k), w_1d(:), lognfft_1d )
!
        end do
!
!$OMP END DO
!
    else
!
!       m IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            tmpvec(:m) = fftx(:m,k)
!
            call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
            fftx(:m,k) = tmpvec(:m)
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            tmpvec(:m) = ffty(:m,k)
!
            call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
            ffty(:m,k) = tmpvec(:m)
!
        end do
!
!$OMP END DO
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( tmpvec, tmpvec2 )
!
!$OMP END PARALLEL
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( tmp )
!
!
! END OF SUBROUTINE fftxy_rm
! __________________________
!
    end subroutine fftxy_rm
!
! =========================================================================================
!
    subroutine fftxy_rt( x, y, fftx, ffty )
!
! Purpose
! _______
!
!   Given two real valued 3D arrays of the same shape, X and Y, FFTXY returns
!   the Fast Fourier Transforms of X and Y in the two complex valued 3D arrays 
!   FFTX and FFTY, respectively.
!   
!
! Arguments
! _________
!
!   X, Y         (INPUT) real(stnd), dimension(:,:,:)
!                On entry, the real valued 3D arrays to be transformed.
!
!                X and Y must verify shape(X) = shape(Y).
!
!   FFTX, FFTY   (OUTPUT) complex(stnd), dimension(:,:,:)
!                On exit, FFTX and FFTY are replaced
!                by the Fourier transforms of X and Y, respectively.
!
!                FFTX and FFTY must verify shape(FFTX) = shape(FFTY) = shape(X) = shape(Y).
!
!
! Further Details
! _______________
!
!   Depending if size(X,1), size(X,2) and size(X,3) are exact powers of two or not, a
!   radix-2 decimation-in-time Cooley-Tukey algorithm or a CHIRP-Z transform is
!   employed.
!
!   Before using FFTXY, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( size(X,1), size(X,2), size(X,3) )     
!
!   For more details, see:
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!   (2) Cooley, J.W., Lewis, P., and Welch, P., 1969: The Fast Fourier Transform
!           and its Applications", IEEE Trans on Education, 12, 1, 28-34.
!
!   (3) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half
    use Char_Constants,    only : allocate_error, fft_error8
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:,:) :: x, y
!
    complex(stnd), intent(out), dimension(:,:,:) :: fftx, ffty
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(:),   allocatable  :: tmpvec, tmpvec2
    complex(stnd), dimension(:,:), allocatable  :: tmp
!
    integer(i4b) :: m, n, p, i, k
!
    integer  :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fftxy'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(x,1),i4b) ,          &
                    int(size(y,1),i4b) ,          &
                    int(size(fftx,1),i4b)     ,   &
                    int(size(ffty,1),i4b)     ,   &
                    name_proc )
!
    p =  assert_eq( int(size(x,2),i4b) ,          &
                    int(size(y,2),i4b) ,          &
                    int(size(fftx,2),i4b)     ,   &
                    int(size(ffty,2),i4b)     ,   &
                    name_proc )
!
    n =  assert_eq( int(size(x,3),i4b) ,          &
                    int(size(y,3),i4b) ,          &
                    int(size(fftx,3),i4b)     ,   &
                    int(size(ffty,3),i4b)     ,   &
                    name_proc )
!
    if ( m/=n_1d .or.  p/=n_2d .or.  n/=n_3d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
#ifdef _OPENMP
    i           = nfft_3d*nfft_2d*nfft_1d
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  i>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,k,iok,tmp,tmpvec,tmpvec2),  &
!$OMP          SHARED(lognfft_1d,lognfft_2d,lognfft_3d)
!
!   ALLOCATE WORK VARIABLES.
!
    i = max( n, p)
    k = max( nfft_1d, nfft_2d, nfft_3d )
!
    allocate( tmp(i,m), tmpvec(k), tmpvec2(m), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
    do k = 1_i4b, n
!
        do i = 1_i4b, p
!
            fftx(:m,i,k) = cmplx( x(:m,i,k), y(:m,i,k), kind=stnd  )
!
        end do
!
    end do
!
!$OMP END DO            
!
!   TRANSFORM THIRD SUBSCRIPT.
!
    if ( n_3d==nfft_3d ) then
!
!       n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, p
!
            tmp(:n,:m) = transpose( fftx(:m,k,:n) )
!
            do i = 1_i4b, m
!
                call fft2( tmp(:n,i), w_3d(:), lognfft_3d )
!
            end do
!
            fftx(:m,k,:n) = transpose( tmp(:n,:m) )
!
        end do
!
!$OMP END DO
!
    else
!
!       n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, p
!
            tmp(:n,:m) = transpose( fftx(:m,k,:n) )
!
            do i = 1_i4b, m
!
                tmpvec(:n) = tmp(:n,i)
!
                call chirp( tmpvec(:nfft_3d), w_3d(:), y_3d(:), y2_3d(:), lognfft_3d )
!
                tmp(:n,i) = tmpvec(:n)
!
            end do
!
            fftx(:m,k,:n) = transpose( tmp(:n,:m) )
!
        end do
!
!$OMP END DO
!
    end if
!
!   SEPARATE THE TWO TRANSFORMS.
!
!$OMP DO SCHEDULE(STATIC)
!
    do k = 2_i4b, n
!
        i = n - k + 2_i4b
!
        ffty(:m,:p,k) = fftx(:m,:p,i)
!
    end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC)
!
    do k = 2_i4b, n
!
        do i = 1_i4b, p
!
            tmpvec(:m)   = ffty(:m,i,k)
            tmpvec2(:m)  = half*( fftx(:m,i,k) + tmpvec(:m) )
            tmpvec(:m)   = half*( fftx(:m,i,k) - tmpvec(:m) )
            fftx(:m,i,k) = cmplx( real(  tmpvec2(:m) ), aimag( tmpvec(:m) ), kind=stnd )
            ffty(:m,i,k) = cmplx( aimag( tmpvec2(:m) ), -real( tmpvec(:m) ), kind=stnd )
!
        end do
!
    end do
!
!$OMP END DO NOWAIT
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, p
!
        tmpvec(:m) = fftx(:m,i,1_i4b)
        fftx(:m,i,1_i4b) = cmplx( real( tmpvec(:m)  ), zero, kind=stnd )
        ffty(:m,i,1_i4b) = cmplx( aimag( tmpvec(:m) ), zero, kind=stnd )
!
    end do
!
!$OMP END DO
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( tmpvec2 )
!
!   TRANSFORM SECOND SUBSCRIPT OF X.
!
    if ( n_2d==nfft_2d ) then
!
!       p IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            tmp(:p,:m) = transpose( fftx(:m,:p,k) )
!
            do i = 1_i4b, m
!
                call fft2( tmp(:p,i), w_2d(:), lognfft_2d )
!
            end do
!
            fftx(:m,:p,k) = transpose( tmp(:p,:m) )
!
        end do
!
!$OMP END DO NOWAIT
!
    else
!
!       p IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            tmp(:p,:m) = transpose( fftx(:m,:p,k) )
!
            do i = 1_i4b, m
!
                tmpvec(:p) = tmp(:p,i)
!
                call chirp( tmpvec(:nfft_2d), w_2d(:), y_2d(:), y2_2d(:), lognfft_2d )
!
                tmp(:p,i) = tmpvec(:p)
!
            end do
!
            fftx(:m,:p,k) = transpose( tmp(:p,:m) )
!
        end do
!
!$OMP END DO NOWAIT
!
    end if
!
!   TRANSFORM FIRST SUBSCRIPT OF X.
!
    if ( n_1d==nfft_1d ) then
!
!       m IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            do i = 1_i4b, p
!
                call fft2( fftx(:m,i,k), w_1d(:), lognfft_1d )
!
            end do
!
        end do
!
!$OMP END DO NOWAIT
!
    else
!
!       m IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            do i = 1_i4b, p
!
                tmpvec(:m) = fftx(:m,i,k)
!
                call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                fftx(:m,i,k) = tmpvec(:m)
!
            end do
!
        end do
!
!$OMP END DO NOWAIT
!
    end if
!
!   TRANSFORM SECOND SUBSCRIPT OF Y.
!
    if ( n_2d==nfft_2d ) then
!
!       p IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            tmp(:p,:m) = transpose( ffty(:m,:p,k) )
!
            do i = 1_i4b, m
!
                call fft2( tmp(:p,i), w_2d(:), lognfft_2d )
!
            end do
!
            ffty(:m,:p,k) = transpose( tmp(:p,:m) )
!
        end do
!
!$OMP END DO NOWAIT
!
    else
!
!       p IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            tmp(:p,:m) = transpose( ffty(:m,:p,k) )
!
            do i = 1_i4b, m
!
                tmpvec(:p) = tmp(:p,i)
!
                call chirp( tmpvec(:nfft_2d), w_2d(:), y_2d(:), y2_2d(:), lognfft_2d )
!
                tmp(:p,i) = tmpvec(:p)
!
            end do
!
            ffty(:m,:p,k) = transpose( tmp(:p,:m) )
!
        end do
!
!$OMP END DO NOWAIT
!
    end if
!
!   TRANSFORM FIRST SUBSCRIPT OF Y.
!
    if ( n_1d==nfft_1d ) then
!
!       m IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            do i = 1_i4b, p
!
                call fft2( ffty(:m,i,k), w_1d(:), lognfft_1d )
!
            end do
!
        end do
!
!$OMP END DO
!
    else
!
!       m IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n
!
            do i = 1_i4b, p
!
                tmpvec(:m) = ffty(:m,i,k)
!
                call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                ffty(:m,i,k) = tmpvec(:m)
!
            end do
!
        end do
!
!$OMP END DO
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( tmp, tmpvec )
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE fftxy_rt
! __________________________
!
    end subroutine fftxy_rt
!
! =========================================================================================
!
    subroutine fftxy_dim_rm( x, y, fftx, ffty, dim )
!
! Purpose
! _______
!
!   Given two real valued matrices of the same shape, X and Y, FFTXY returns
!   the Fast Fourier Transforms of the rows (DIM=2) or the columns (DIM=1) of X and Y
!   in the two complex valued matrices FFTX and FFTY, respectively.
!   
!
! Arguments
! _________
!
!   X, Y         (INPUT) real(stnd), dimension(:,:)
!                On entry, the real valued matrices to be transformed.
!
!                X and Y must verify shape(X) = shape(Y).
!
!   FFTX, FFTY   (OUTPUT) complex(stnd), dimension(:,:)
!                On exit:
!
!                - each row of FFTX and FFTY are replaced
!                  by the Fourier transforms of the rows of X
!                  and Y, respectively, if DIM=2 .
!                - each column of FFTX and FFTY are replaced
!                  by the Fourier transforms of the colums of X
!                  and Y, respectively, if DIM=1 .
!
!                FFTX and FFTY must verify shape(FFTX) = shape(FFTY) = shape(X) = shape(Y).
!
!   DIM          (INPUT) integer(i4b)
!                Specifies the index for the Fourier transform. If:
!
!                - DIM = 1 :  Fourier transform on first index,
!                - DIM = 2 :  Fourier transform on second index.
!
!
! Further Details
! _______________
!
!   Size(FFTX,DIM) = size(FFTY,DIM) = size(X,DIM) = size(Y,DIM) may be of general length.
!
!   Before using FFTXY, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( (/ size(X,1), size(X,2) /), dim=DIM )     
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, half
    use Char_Constants,    only : allocate_error, fft_error8, fft_error11
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:) :: x, y
!
    complex(stnd), intent(out), dimension(:,:) :: fftx, ffty
!
    integer(i4b), intent(in) :: dim
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(:), allocatable :: tmpvec, tmpvec2
!
    integer(i4b) :: m, n, i, k
!
    integer  :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fftxy'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( dim/=1_i4b .and. dim/=2_i4b ) then
!
        call merror( name_proc//fft_error11 )
!
    end if
!
    m =  assert_eq( int(size(x,3-int(dim)),i4b) ,          &
                    int(size(y,3-int(dim)),i4b) ,          &
                    int(size(fftx,3-int(dim)),i4b)     ,   &
                    int(size(ffty,3-int(dim)),i4b)     ,   &
                    name_proc )
!
    n =  assert_eq( int(size(x,int(dim)),i4b) ,          &
                    int(size(y,int(dim)),i4b) ,          &
                    int(size(fftx,int(dim)),i4b)     ,   &
                    int(size(ffty,int(dim)),i4b)     ,   &
                    name_proc )
!
    if ( n/=n_1d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
    fftx(:,:) = cmplx( x(:,:), y(:,:), kind=stnd )
!
    call fft_dim_cm( fftx(:,:), true, dim)
!
#ifdef _OPENMP
    i           = m*n
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  i>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,k,iok,tmpvec,tmpvec2)
!
!   ALLOCATE WORK ARRAYS.
!
    i = size( x, 1)
!
    allocate( tmpvec(i), tmpvec2(i), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!   SEPARATE THE TWO TRANSFORMS.
!
    if ( dim==2_i4b ) then
!
!       TRANSFORM ON SECOND SUBSCRIPT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 2_i4b, n
!
            i = n - k + 2_i4b
!
            ffty(:m,k) = fftx(:m,i)
!
        end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 2_i4b, n
!
            tmpvec(:m)  = ffty(:m,k)
            tmpvec2(:m) = half*( fftx(:m,k) + tmpvec(:m) )
            tmpvec(:m)  = half*( fftx(:m,k) - tmpvec(:m) )
            fftx(:m,k)  = cmplx( real(  tmpvec2(:m) ), aimag( tmpvec(:m) ), kind=stnd )
            ffty(:m,k)  = cmplx( aimag( tmpvec2(:m) ), -real( tmpvec(:m) ), kind=stnd )
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP SINGLE
!
        tmpvec(:m)     = fftx(:m,1_i4b)
        fftx(:m,1_i4b) = cmplx( real( tmpvec(:m)  ), zero, kind=stnd )
        ffty(:m,1_i4b) = cmplx( aimag( tmpvec(:m) ), zero, kind=stnd )
!
!$OMP END SINGLE
!
    else
!
!       TRANSFORM ON FIRST SUBSCRIPT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, m
!
            tmpvec(1_i4b)    = fftx(1_i4b,k)
            tmpvec(2_i4b:n)  = fftx(n:2_i4b:-1_i4b,k)
            tmpvec2(2_i4b:n) = half*( fftx(2_i4b:n,k) + tmpvec(2_i4b:n) )
            tmpvec(2_i4b:n)  = half*( fftx(2_i4b:n,k) - tmpvec(2_i4b:n) )
!
            fftx(1_i4b,k)    = cmplx( real(  tmpvec(1_i4b) ), zero, kind=stnd )
            fftx(2_i4b:n,k)  = cmplx( real(  tmpvec2(2_i4b:n) ), aimag( tmpvec(2_i4b:n) ), kind=stnd )
!
            ffty(1_i4b,k)    = cmplx( aimag( tmpvec(1_i4b) ), zero, kind=stnd )
            ffty(2_i4b:n,k)  = cmplx( aimag( tmpvec2(2_i4b:n) ), -real( tmpvec(2_i4b:n) ), kind=stnd )
!
        end do
!
!$OMP END DO
!
    end if
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( tmpvec, tmpvec2 )
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE fftxy_dim_rm
! ______________________________
!
    end subroutine fftxy_dim_rm
!
! =========================================================================================
!
    subroutine fftxy_dim_rt( x, y, fftx, ffty, dim )
!
! Purpose
! _______
!
!   Given two real valued 3D arrays of the same shape, X and Y, FFTXY returns
!   the Fast Fourier Transforms of each DIM-index section of X and Y
!   in the two complex valued 3D arrays FFTX and FFTY, respectively.
!   
!
! Arguments
! _________
!
!   X, Y         (INPUT) real(stnd), dimension(:,:,:)
!                On entry, the real valued 3D arrays to be transformed.
!
!                X and Y must verify shape(X) = shape(Y).
!
!   FFTX, FFTY   (OUTPUT) complex(stnd), dimension(:,:,:)
!                On exit, the DIM-index sections of FFTX and FFTY are replaced 
!                by the Fourier transforms of the DIM-index sections of X and Y,
!                respectively.
!
!                FFTX and FFTY must verify shape(FFTX) = shape(FFTY) = shape(X) = shape(Y).
!
!   DIM          (INPUT) integer(i4b)
!                Specifies the index for the Fourier transform. If:
!
!                - DIM  = 1 :  Fourier transform on first-index-sections.
!                - DIM  = 2 :  Fourier transform on second-index-sections.
!                - DIM  = 3 :  Fourier transform on third-index-sections.
!
!
! Further Details
! _______________
!
!   Size(FFTX,DIM) = size(FFTY,DIM) = size(X,DIM) = size(Y,DIM) may be of general length.
!
!   Before using FFTXY, the user must call subroutine INIT_FFT as follows:
!
!                call init_fft( (/ size(X,1), size(X,2), size(X,3) /), dim=DIM )     
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, half
    use Char_Constants,    only : allocate_error, fft_error8, fft_error12
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:,:) :: x, y
!
    complex(stnd), intent(out), dimension(:,:,:) :: fftx, ffty
!
    integer(i4b), intent(in) :: dim
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(:), allocatable :: tmpvec, tmpvec2
!
    integer(i4b)  :: m, n, p, q, i, j, k
!
    integer  :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fftxy'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( dim<1_i4b .or. dim>3_i4b ) then
!
        call merror( name_proc//fft_error12 )
!
    end if
!
    n = size( x, int(dim) )
!
    if ( n/=n_1d ) then
!
         call merror( name_proc//fft_error8 )
!
    end if
!
    m =  assert_eq( int(size(x,1),i4b) ,          &
                    int(size(y,1),i4b) ,          &
                    int(size(fftx,1),i4b)     ,   &
                    int(size(ffty,1),i4b)     ,   &
                    name_proc )
!
    p =  assert_eq( int(size(x,2),i4b) ,          &
                    int(size(y,2),i4b) ,          &
                    int(size(fftx,2),i4b)     ,   &
                    int(size(ffty,2),i4b)     ,   &
                    name_proc )
!
    q =  assert_eq( int(size(x,3),i4b) ,          &
                    int(size(y,3),i4b) ,          &
                    int(size(fftx,3),i4b)     ,   &
                    int(size(ffty,3),i4b)     ,   &
                    name_proc )
!
    fftx(:m,:p,:q) = cmplx( x(:m,:p,:q), y(:m,:p,:q), kind=stnd  )
!
    call fft_dim_ct( fftx(:m,:p,:q), true, dim)
!
#ifdef _OPENMP
    i           = m*p*q
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  i>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,j,k,iok,tmpvec,tmpvec2)
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( tmpvec(m), tmpvec2(m), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!   SEPARATE THE TWO TRANSFORMS.
!
    select case( dim )
!
        case( 1_i4b )
!
!           TRANSFORM ON FIRST SUBSCRIPT.
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
            do k = 1_i4b, q
!
                do j = 1_i4b, p
!
                    tmpvec(1_i4b)    = fftx(1_i4b,j,k)
                    tmpvec(2_i4b:n)  = fftx(n:2_i4b:-1_i4b,j,k)
                    tmpvec2(2_i4b:n) = half*( fftx(2_i4b:n,j,k) + tmpvec(2_i4b:n) )
                    tmpvec(2_i4b:n)  = half*( fftx(2_i4b:n,j,k) - tmpvec(2_i4b:n) )
!
                    fftx(1_i4b,j,k)    = cmplx( real(  tmpvec(1_i4b) ), zero, kind=stnd )
                    fftx(2_i4b:n,j,k)  = cmplx( real(  tmpvec2(2_i4b:n) ), aimag( tmpvec(2_i4b:n) ), kind=stnd )
!
                    ffty(1_i4b,j,k)    = cmplx( aimag( tmpvec(1_i4b) ), zero, kind=stnd )
                    ffty(2_i4b:n,j,k)  = cmplx( aimag( tmpvec2(2_i4b:n) ), -real( tmpvec(2_i4b:n) ), kind=stnd )
!
                end do
!
            end do
!
!$OMP END DO
!
        case( 2_i4b )
!
!           TRANSFORM ON SECOND SUBSCRIPT.
!
!$OMP DO SCHEDULE(STATIC)
!
            do k = 1_i4b, q
!
                ffty(:m,2_i4b:n,k)  = fftx(:m,n:2_i4b:-1_i4b,k)
!
            end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC)
!
            do k = 1_i4b, q
!
                do j = 2_i4b, n
!
                    tmpvec(:m)  = ffty(:m,j,k)
                    tmpvec2(:m) = half*( fftx(:m,j,k) + tmpvec(:m) )
                    tmpvec(:m)  = half*( fftx(:m,j,k) - tmpvec(:m) )
!
                    fftx(:m,j,k)  = cmplx( real(  tmpvec2(:m) ), aimag( tmpvec(:m) ), kind=stnd )
                    ffty(:m,j,k)  = cmplx( aimag( tmpvec2(:m) ), -real( tmpvec(:m) ), kind=stnd )
!
                end do
!
                tmpvec(:m)       = fftx(:m,1_i4b,k)
                fftx(:m,1_i4b,k) = cmplx( real( tmpvec(:m)  ), zero, kind=stnd )
                ffty(:m,1_i4b,k) = cmplx( aimag( tmpvec(:m) ), zero, kind=stnd )
!
            end do
!
!$OMP END DO
!
        case( 3_i4b )
!
!           TRANSFORM ON THIRD SUBSCRIPT.
!
!$OMP DO SCHEDULE(STATIC)
!
            do k = 2_i4b, n
!
                i = n - k + 2_i4b
!
                ffty(:m,:p,k) = fftx(:m,:p,i)
!
            end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
            do k = 2_i4b, n
!
                do i = 1_i4b, p
!
                    tmpvec(:m)    = ffty(:m,i,k)
                    tmpvec2(:m)   = half*( fftx(:m,i,k) + tmpvec(:m) )
                    tmpvec(:m)    = half*( fftx(:m,i,k) - tmpvec(:m) )
                    fftx(:m,i,k)  = cmplx( real(  tmpvec2(:m) ), aimag( tmpvec(:m) ), kind=stnd )
                    ffty(:m,i,k)  = cmplx( aimag( tmpvec2(:m) ), -real( tmpvec(:m) ), kind=stnd )
!
                end do
!
            end do
!
!$OMP END DO NOWAIT
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 1_i4b, p
!
                tmpvec(:m) = fftx(:m,i,1_i4b)
                fftx(:m,i,1_i4b) = cmplx( real( tmpvec(:m)  ) , zero, kind=stnd )
                ffty(:m,i,1_i4b) = cmplx( aimag( tmpvec(:m) ), zero, kind=stnd )
!
            end do
!
!$OMP END DO
!
    end select
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( tmpvec, tmpvec2 )
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE fftxy_dim_rt
! ______________________________
!
    end subroutine fftxy_dim_rt
!
!
! =========================================================================================
!                                 COMPLEX FOURIER TRANSFORMS
! =========================================================================================
!
!
    subroutine fftn( dat, forward)
!
! Purpose
! _______
!
!   Subroutine FFT implements the Fast Fourier Transform for a complex valued sequence
!   DAT of general length.
!
!   Forward discrete Fourier transform of a vector DAT(:) is given by              
!
!   t( DAT )(j)  =  [ sum k=0 to nn-1 ]  DAT(k) exp( - i 2 pi j k / nn )
!
!   Backward discrete Fourier transform of a vector DAT(:) is given by              
!
!   t( DAT )(j)  =  (1/nn) [ sum k=0 to nn-1 ]  DAT(k) exp( i 2 pi j k / nn )
!
!   where  i  =  sqrt( -1 ), nn =  size(DAT) and  pi =  3.1415923565...
!
!   Note that the indexing of DAT is shifted by one :                     
!   DAT(0) stored in DAT(1), ... , DAT( nn-1 ) stored in DAT( nn ).
!
!
! Arguments
! _________
!
!   DAT      (INPUT/OUTPUT) complex(stnd), dimension(:)
!            On entry, the complex valued sequence to be transformed.
!            On exit, DAT is replaced by the Fourier transform.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!
! Further Details
! _______________
!
!   If size(DAT) is an exact power of two, Bailey's Four-Step FFT algorithm is used,
!   otherwise the CHIRP-Z transform is employed.
!
!   Before using FFT, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( size(DAT) )     
!   
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83 and
!   Bailey (1990). For more details, see:
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!   (2) Bailey, D., 1990: FFTs in External or Hierarchical Memory, 
!           The Journal of Supercomputing, 4, 23-35.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : fft_error8
    use Reals_Constants,   only : zero, one
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:) :: dat
!
    logical(lgl), intent(in) :: forward
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(nfft1_n,nfft2_n) :: tmp2
!
    integer(i4b) :: length, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    length = size( dat )
!
    if ( length/=n_n ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
    if ( .not.forward ) then
!
        dat(:n_n) = conjg( dat(:n_n) )
!
    end if
!
    if ( n_n==nfft_n ) then
!
!       length IS A POWER OF TWO. USE Bailey's FFT.
!
!       RESHAPE THE INPUT ARRAY.
!     
        tmp2(:nfft1_n,:nfft2_n) = reshape( dat, shape(tmp2) )
!     
!       STEP 1.
!     
        call fft2_row( tmp2(:,:), w2_n(:), lognfft2_n )
!     
!       STEPS 2, 3, 4.
!     
        do i=1_i4b, nfft2_n
!     
!           STEP 2.
!     
            tmp2(:nfft1_n,i) = tmp2(:nfft1_n,i)*v_n(:nfft1_n,i)
!     
!           STEPS 3 AND 4.
!     
            call fft2( tmp2(:nfft1_n,i), w1_n(:), lognfft1_n )
!     
!           RESHAPE THE RESULT BACK TO ONE DIMENSION.  
!     
            dat(i:n_n:nfft2_n) = tmp2(:nfft1_n,i)
!     
        end do
!
    else
!
!       length IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!       FIRST MULTIPLY THE COMPLEX SEQUENCE BY THE CHIRP FUNCTION
!       IN THE TIME DOMAIN AND RESHAPE THE INPUT ARRAY.
!
        tmp2(:nfft1_n,:nfft2_n) = reshape( dat(:n_n)*y2_n(:n_n), shape(tmp2),     &
                                           (/ cmplx(zero,zero,kind=stnd) /) )
        
!       FOURIER TRANSFORM THE CHIRPED SEQUENCE.
!
!       STEP 1.
!     
        call fft2_row( tmp2(:,:), w2_n(:), lognfft2_n )
!     
!       STEPS 2, 3, 4.
!     
        do i=1_i4b, nfft2_n
!     
!           STEP 2.
!     
            tmp2(:nfft1_n,i) = tmp2(:nfft1_n,i)*v_n(:nfft1_n,i)
!     
!           STEPS 3 AND 4.
!     
            call fft2( tmp2(:nfft1_n,i), w1_n(:), lognfft1_n )
!     
!           CONVOLVE BY FREQUENCY DOMAIN MULTIPLICATION WITH THE CHIRP TRANSFORM.  
!     
            tmp2(:nfft1_n,i) = conjg( tmp2(:nfft1_n,i)*y_n(:nfft1_n,i) )
!
!           BACKWARD FOURIER TRANSFORM THE SEQUENCE.
!
!           STEPS 1 AND 2.
!     
            call fft2( tmp2(:nfft1_n,i), w1_n(:), lognfft1_n )
!     
!           STEP 3.
!     
            tmp2(:nfft1_n,i) = tmp2(:nfft1_n,i)*v_n(:nfft1_n,i)
!     
        end do
!     
!       STEP 4.
!     
        call fft2_row( tmp2(:,:), w2_n(:), lognfft2_n )
!     
!       RESHAPE THE RESULT BACK TO ONE DIMENSION.  
!     
        dat(:n_n) = reshape( tmp2, shape(dat) )
!
!       MULTIPLY BY CHIRP FUNCTION.
!
        dat(:n_n) = ( conjg( dat(:n_n) )*y2_n(:n_n) )*(one/real(nfft_n,stnd))
!
    end if
!
    if ( .not.forward ) then
!
        dat(:n_n) = conjg( dat(:n_n) )*(one/real(n_n,stnd))
!
    end if
!
!
! END OF SUBROUTINE fftn
! ______________________
!
    end subroutine fftn
!
! =========================================================================================
!
    subroutine fft_cm( dat, forward)
!
! Purpose
! _______
!
!   Subroutine FFT implements the Fast Fourier Transform for a complex matrix
!   DAT of general shape.
!
!
! Arguments
! _________
!
!   DAT      (INPUT/OUTPUT) complex(stnd), dimension(:,:)
!            On entry, the complex matrix to be transformed.
!            On exit, DAT is replaced by its Fourier transform.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!
! Further Details
! _______________
!
!   Depending if size(DAT,1) and size(DAT,2) are exact powers of two or not, a
!   radix-2 decimation-in-time Cooley-Tukey algorithm or a CHIRP-Z transform is
!   employed.
!
!   Before using FFT, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( (/ size(DAT,1), size(DAT,2) /) )     
!
!   For more details, see:
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!   (2) Cooley, J.W., Lewis, P., and Welch, P., 1969: The Fast Fourier Transform
!           and its Applications", IEEE Trans on Education, 12, 1, 28-34.
!
!   (3) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Char_Constants,    only : allocate_error, fft_error8
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:) :: dat
!
    logical(lgl), intent(in) :: forward
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: scal
!
    complex(stnd), dimension(:),   allocatable :: tmpvec
    complex(stnd), dimension(:,:), allocatable :: x
!
    integer(i4b) :: length1, length2, i
!
    integer      :: iok
!
#ifdef _OPENMP
    integer(i4b) :: i2, k, k2
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    length1 = size( dat, 1 )
    length2 = size( dat, 2 )
!
!   TEST THE ARGUMENTS.
!
    if ( length1/=n_1d .or. length2/=n_2d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
!   ALLOCATE WORK VARIABLE FOR THE FFT OR DISCRETE CHIRP-Z TRANSFORM.
!
    allocate( x(n_2d,n_1d), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nfft_1d*nfft_2d>=omp_limit          .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,i2,k,k2,iok,scal,tmpvec),   &
!$OMP          SHARED(lognfft_1d,lognfft_2d)
!
    if ( n_1d/=nfft_1d .or. n_2d/=nfft_2d ) then
!
!       ALLOCATE WORK VARIABLE FOR THE DISCRETE CHIRP-Z TRANSFORM.
!
        i = max( nfft_1d, nfft_2d)
!
        allocate( tmpvec(i), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
    end if
!
!   TRANSPOSE THE ARRAY.
!
#ifdef _OPENMP
    if ( test_par ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
        do k = 1_i4b, n_2d, blksz_t
!
            do i = 1_i4b, n_1d, blksz_t
!
                k2 = min( n_2d, k+blksz_t-1_i4b )
                i2 = min( n_1d, i+blksz_t-1_i4b )
!
                x(k:k2,i:i2) = transpose( dat(i:i2,k:k2) )
!
                if ( .not.forward ) then
!
                    x(k:k2,i:i2) = conjg( x(k:k2,i:i2) )
!
                end if
!
            end do
!
        end do
!
!$OMP END DO
!
    else
#endif
!
        x(:n_2d,:n_1d) = transpose( dat(:n_1d,:n_2d) )
!
        if ( .not.forward ) then
!
            x(:n_2d,:n_1d) = conjg( x(:n_2d,:n_1d) )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!   TRANSFORM SECOND SUBSCRIPT.
!
    if ( n_2d==nfft_2d ) then
!
!       length2 IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, n_1d
!
            call fft2( x(:n_2d,i), w_2d(:), lognfft_2d )
!
        end do
!
!$OMP END DO
!
    else
!
!       length2 IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, n_1d
!
            tmpvec(:n_2d) = x(:n_2d,i)
!
            call chirp( tmpvec(:nfft_2d), w_2d(:), y_2d(:), y2_2d(:), lognfft_2d )
!
            x(:n_2d,i) = tmpvec(:n_2d)
!
        end do
!
!$OMP END DO
!
    end if
!
!   TRANSPOSE BACK THE ARRAY.
!
#ifdef _OPENMP
    if ( test_par ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
        do k = 1_i4b, n_1d, blksz_t
!
            do i = 1_i4b, n_2d, blksz_t
!
                k2 = min( n_1d, k+blksz_t-1_i4b )
                i2 = min( n_2d, i+blksz_t-1_i4b )
!
                dat(k:k2,i:i2) = transpose( x(i:i2,k:k2) )
!
            end do
!
        end do
!
!$OMP END DO
!
    else
#endif
!
        dat(:n_1d,:n_2d) = transpose( x(:n_2d,:n_1d) )
!
#ifdef _OPENMP
    end if
#endif
!
    scal = one/real( n_1d*n_2d, stnd)
!
!   TRANSFORM FIRST SUBSCRIPT.
!
    if ( n_1d==nfft_1d ) then
!
!       length1 IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, n_2d
!
            call fft2( dat(:n_1d,i), w_1d(:), lognfft_1d  )
!
            if ( .not.forward ) then
!
                dat(:n_1d,i) = conjg( dat(:n_1d,i) )*scal
!
            end if
!
        end do
!
!$OMP END DO
!
    else
!
!       length1 IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, n_2d
!
            tmpvec(:n_1d) = dat(:n_1d,i)
!
            call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
            if ( forward ) then
!
                dat(:n_1d,i) = tmpvec(:n_1d)
!
            else
!
                dat(:n_1d,i) = conjg( tmpvec(:n_1d) )*scal
!
            end if
!
        end do
!
!$OMP END DO
!
    end if
!
    if ( n_1d/=nfft_1d .or. n_2d/=nfft_2d ) then
!
!       DEALLOCATE WORK VARIABLE.
!
        deallocate( tmpvec )
!
    end if
!
!$OMP END PARALLEL
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( x )
!
!
! END OF SUBROUTINE fft_cm
! ________________________
!
    end subroutine fft_cm
!
! =========================================================================================
!
    subroutine fft_ct( dat, forward)
!
! Purpose
! _______
!
!   Subroutine FFT implements the Fast Fourier Transform for a complex 3D array
!   DAT of general shape.
!
!
! Arguments
! _________
!
!   DAT      (INPUT/OUTPUT) complex(stnd), dimension(:,:,:)
!            On entry, the complex array to be transformed.
!            On exit, DAT is replaced by its Fourier transform.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!
! Further Details
! _______________
!
!   Depending if size(DAT,1), size(DAT,2) and size(DAT,3) are exact powers of two or not,
!   a radix-2 decimation-in-time Cooley-Tukey algorithm or a CHIRP-Z transform is
!   employed.
!
!   Before using FFT, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( (/ size(DAT,1), size(DAT,2), size(DAT,3) /) )     
!
!   For more details, see:
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!   (2) Cooley, J.W., Lewis, P., and Welch, P., 1969: The Fast Fourier Transform
!           and its Applications", IEEE Trans on Education, 12, 1, 28-34.
!
!   (3) Oppenheim, A.V., and Schafer, R.W., 1999: Discrete-Time Signal Processing,
!           Second Edition. Prentice-Hall, New Jersey.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Char_Constants,    only : allocate_error, fft_error8
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:,:) :: dat
!
    logical(lgl), intent(in) :: forward
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: scal
!
    complex(stnd), dimension(:),   allocatable :: tmpvec
    complex(stnd), dimension(:,:), allocatable :: x
!
    integer(i4b) :: length1, length2, length3, j, k
!
    integer      :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    length1 = size( dat, 1 )
    length2 = size( dat, 2 )
    length3 = size( dat, 3 )
!
!   TEST THE ARGUMENTS.
!
    if ( length1/=n_1d .or. length2/=n_2d .or. length3/=n_3d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
#ifdef _OPENMP
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nfft_1d*nfft_2d*nfft_3d>=omp_limit  .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(j,k,x,tmpvec,scal,iok),   &
!$OMP          SHARED(lognfft_1d,lognfft_2d,lognfft_3d)
!
!   ALLOCATE WORK VARIABLES FOR THE FFT OR DISCRETE CHIRP-Z TRANSFORMS.
!
    k = max( n_2d, n_3d)
!
    if ( n_1d/=nfft_1d .or. n_2d/=nfft_2d .or. n_3d/=nfft_3d ) then
!
        j = max( nfft_1d, nfft_2d, nfft_3d)
!
        allocate( x(k,n_1d), tmpvec(j), stat = iok )
!
    else
!
        allocate( x(k,n_1d), stat = iok )
!
    end if
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
    if ( .not.forward ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
        do k = 1_i4b, n_3d
!
            do j = 1_i4b, n_2d
!
                dat(:n_1d,j,k) = conjg( dat(:n_1d,j,k) )
!
            end do
!
        end do
!
!$OMP END DO
!
    end if
!
!   TRANSFORM THIRD SUBSCRIPT.
!
    if ( n_3d==nfft_3d ) then
!
!       length3 IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n_2d
!
            x(:n_3d,:n_1d) = transpose( dat(:n_1d,k,:n_3d) )
!
            do j = 1_i4b, n_1d
!
                call fft2( x(:n_3d,j), w_3d(:), lognfft_3d )
!
            end do
!
            dat(:n_1d,k,:n_3d) = transpose( x(:n_3d,:n_1d) )
!
        end do
!
!$OMP END DO
!
    else
!
!       length3 IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n_2d
!
            x(:n_3d,:n_1d) = transpose( dat(:n_1d,k,:n_3d) )
!
            do j = 1_i4b, n_1d
!
                tmpvec(:n_3d) = x(:n_3d,j)
!
                call chirp( tmpvec(:nfft_3d), w_3d(:), y_3d(:), y2_3d(:), lognfft_3d )
!
                x(:n_3d,j) = tmpvec(:n_3d)
!
            end do
!
            dat(:n_1d,k,:n_3d) = transpose( x(:n_3d,:n_1d) )
!
        end do
!
!$OMP END DO
!
    end if
!
!   TRANSFORM SECOND SUBSCRIPT.
!
    if ( n_2d==nfft_2d ) then
!
!       length2 IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n_3d
!
            x(:n_2d,:n_1d) = transpose( dat(:n_1d,:n_2d,k) )
!
            do j = 1_i4b, n_1d
!
                call fft2( x(:n_2d,j), w_2d(:), lognfft_2d )
!
            end do
!
            dat(:n_1d,:n_2d,k) = transpose( x(:n_2d,:n_1d) )
!
        end do
!
!$OMP END DO NOWAIT
!
    else
!
!       length2 IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n_3d
!
            x(:n_2d,:n_1d) = transpose( dat(:n_1d,:n_2d,k) )
!
            do j = 1_i4b, n_1d
!
                tmpvec(:n_2d) = x(:n_2d,j)
!
                call chirp( tmpvec(:nfft_2d), w_2d(:), y_2d(:), y2_2d(:), lognfft_2d )
!
                x(:n_3d,j) = tmpvec(:n_2d)
!
            end do
!
            dat(:n_1d,:n_2d,k) = transpose( x(:n_2d,:n_1d) )
!
        end do
!
!$OMP END DO NOWAIT
!
    end if
!
!   TRANSFORM FIRST SUBSCRIPT.
!
    if ( n_1d==nfft_1d ) then
!
!       length1 IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n_3d
!
            do j = 1_i4b, n_2d
!
                call fft2( dat(:n_1d,j,k), w_1d(:), lognfft_1d )
!
            end do
!
        end do
!
!$OMP END DO NOWAIT
!
    else
!
!       length1 IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n_3d
!
            do j = 1_i4b, n_2d
!
                tmpvec(:n_1d) = dat(:n_1d,j,k)
!
                call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                dat(:n_1d,j,k) = tmpvec(:n_1d)
!
            end do
!
        end do
!
!$OMP END DO NOWAIT
!
    end if
!
    if ( .not.forward ) then
!
        scal = one/real( n_1d*n_2d*n_3d, stnd)
!
!$OMP DO SCHEDULE(STATIC)
!
        do k = 1_i4b, n_3d
!
            do j = 1_i4b, n_2d
!
                dat(:n_1d,j,k) = conjg( dat(:n_1d,j,k) )*scal
!
            end do
!
        end do
!
!$OMP END DO
!
    end if
!
!   DEALLOCATE WORK VARIABLES.
!
    if ( n_1d/=nfft_1d .or. n_2d/=nfft_2d .or. n_3d/=nfft_3d ) then
!
        deallocate( x, tmpvec )
!
    else
!
        deallocate( x )
!
    end if
!
!$OMP END PARALLEL
!
!
!
! END OF SUBROUTINE fft_ct
! ________________________
!
    end subroutine fft_ct
!
! =========================================================================================
!
    subroutine fft_dim_cm( dat, forward, dim)
!
! Purpose
! _______
!
!   Subroutine FFT replaces each row of DAT by its Fourier transform.
!   (DIM=2) or each column of DAT by its Fourier transform (DIM=1).
!   Size(DAT,DIM) may be of general length.
!
!
! Arguments
! _________
!
!   DAT      (INPUT/OUTPUT) complex(stnd), dimension(:,:)
!            On entry, the complex valued sequences to be transformed.
!            On exit, each row of DAT is replaced by its Fourier transform
!            if DIM=2 or each column of DAT is replaced by its Fourier transform
!            if DIM=1.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!   DIM      (INPUT) integer(i4b)
!            Specifies the index for the Fourier transform. If:
!
!            - DIM = 1 :  Fourier transform on first index,
!            - DIM = 2 :  Fourier transform on second index.
!
!
! Further Details
! _______________
!
!   If size(DAT,DIM) is an exact power of two, a 1D in-place complex-complex radix-2
!   decimation-in-time Cooley-Tukey FFT algorithm is used, otherwise the
!   CHIRP-Z transform is employed.
!
!   Before using FFT, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( (/ size(DAT,1), size(DAT,2) /), dim=DIM )     
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83, see
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!        general length, Appl. Statist., 26 (3), 351-361.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Utilities,         only : merror
    use Char_Constants,    only : allocate_error, fft_error8, fft_error11
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:) :: dat
!
    logical(lgl), intent(in) :: forward
!
    integer(i4b), intent(in) :: dim
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: scal
!
    complex(stnd), dimension(:),   allocatable :: tmpvec
    complex(stnd), dimension(:,:), allocatable :: x
!
    integer(i4b)  :: n, n1, n2, j
!
    integer :: iok
!
#ifdef _OPENMP
!
    integer(i4b)  :: j2, k, k2
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( dim/=1_i4b .and. dim/=2_i4b ) then
!
        call merror( name_proc//fft_error11 )
!
    end if
!
    n = size( dat, int(dim) )
!
    if ( n/=n_1d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
    n1 = size( dat, 1 )
    n2 = size( dat, 2 )
!
#ifdef _OPENMP
!
    j = ((n1*n2)/n)*nfft_1d
!
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  j>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
    if ( dim==1_i4b ) then
!
!       TRANSFORM FIRST SUBSCRIPT.
!
!$OMP PARALLEL IF(test_par), PRIVATE(iok,j,scal,tmpvec),    &
!$OMP          SHARED(nfft_1d,lognfft_1d)
!
        if ( .not.forward ) then
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n2
!
                dat(:n,j) = conjg( dat(:n,j) )
!
            end do
!
!$OMP END DO NOWAIT
!
        end if
!
        if ( n_1d==nfft_1d ) then
!
!           n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n2
!
                call fft2( dat(:n,j), w_1d(:), lognfft_1d )
!
            end do
!
!$OMP END DO NOWAIT
!
        else
!
!           ALLOCATE WORK VARIABLE.
!
            allocate( tmpvec(nfft_1d), stat = iok )
!
            if ( iok/=0 ) then
!
                call merror( name_proc//allocate_error )
!
            end if
!
!           n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n2
!
                tmpvec(:n) = dat(:n,j)
!
                call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                dat(:n,j) = tmpvec(:n)
!
            end do
!
!$OMP END DO NOWAIT
!
!           DEALLOCATE WORK VARIABLE.
!
            deallocate( tmpvec )
!
        end if
!
        if ( .not.forward ) then
!
            scal = one/real( n, stnd)
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n2
!
                dat(:n,j) = conjg( dat(:n,j) )*scal
!
            end do
!
!$OMP END DO
!
        end if
!
!$OMP END PARALLEL
!
    else
!
!       TRANSFORM SECOND SUBSCRIPT.
!
!       ALLOCATE WORK VARIABLE FOR THE FFT OR DISCRETE CHIRP-Z TRANSFORMS.
!
        allocate( x(n,n1), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!$OMP PARALLEL IF(test_par), PRIVATE(iok,j,j2,k,k2,scal,tmpvec), SHARED(nfft_1d,lognfft_1d)
!
!   TRANSPOSE THE ARRAY.
!
#ifdef _OPENMP
        if ( test_par ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
            do k = 1_i4b, n, blksz_t
!
                do j = 1_i4b, n1, blksz_t
!
                    k2 = min( n, k+blksz_t-1_i4b )
                    j2 = min( n1, j+blksz_t-1_i4b )
!
                    x(k:k2,j:j2) = transpose( dat(j:j2,k:k2) )
!
                end do
!
            end do
!
!$OMP END DO
!
        else
#endif
!
            x(:n,:n1) = transpose( dat(:n1,:n) )
!
#ifdef _OPENMP
        end if
#endif
!
        if ( .not.forward ) then
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n1
!
                x(:n,j) = conjg( x(:n,j) )
!
            end do
!
!$OMP END DO NOWAIT
!
        end if
!
        if ( n_1d==nfft_1d ) then
!
!           n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n1
!
                call fft2( x(:n,j), w_1d(:), lognfft_1d )
!
            end do
!
!$OMP END DO
!
        else
!
!           n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
            allocate( tmpvec(nfft_1d), stat = iok )
!
            if ( iok/=0 ) then
!
                call merror( name_proc//allocate_error )
!
            end if
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n1
!
                tmpvec(:n) = x(:n,j)
!
                call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                x(:n,j) = tmpvec(:n)
!
            end do
!
!$OMP END DO
!
!           DEALLOCATE WORK VARIABLE.
!
            deallocate( tmpvec )
!
        end if
!
        if ( .not.forward ) then
!
            scal = one/real( n, stnd)
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n1
!
                x(:n,j) = conjg( x(:n,j) )*scal
!
            end do
!
!$OMP END DO
!
        end if
!
!       TRANSPOSE BACK THE ARRAY.
!
#ifdef _OPENMP
        if ( test_par ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
            do k = 1_i4b, n1, blksz_t
!
                do j = 1_i4b, n, blksz_t
!
                    k2 = min( n1, k+blksz_t-1_i4b )
                    j2 = min( n, j+blksz_t-1_i4b )
!
                    dat(k:k2,j:j2) = transpose( x(j:j2,k:k2) )
!
                end do
!
            end do
!
!$OMP END DO
!
    else
#endif
!
        dat(:n1,:n) = transpose( x(:n,:n1) )
!
#ifdef _OPENMP
    end if
#endif
!
!$OMP END PARALLEL
!
!       DEALLOCATE WORK VARIABLE.
!
        deallocate( x )
!
    end if
!
!
! END OF SUBROUTINE fft_dim_cm
! ____________________________
!
    end subroutine fft_dim_cm
!
! =========================================================================================
!
    subroutine fft_dim_ct( dat, forward, dim)
!
! Purpose
! _______
!
!   Subroutine FFT replaces each DIM-index section of DAT by 
!   its Fourier transform.
!   Size(DAT,DIM) may be of general length.
!
!
! Arguments
! _________
!
!   DAT      (INPUT/OUTPUT) complex(stnd), dimension(:,:,:)
!            On entry, the complex valued sequences to be transformed.
!            On exit, the DIM-index sections of DAT are replaced by 
!            their Fourier transforms.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!   DIM      (INPUT) integer(i4b)
!            Specifies the index for the Fourier transform. If:
!
!            - DIM = 1 :  Fourier transform on first-index-sections,
!            - DIM = 2 :  Fourier transform on second-index-sections,
!            - DIM = 3 :  Fourier transform on third-index-sections.
!
!
! Further Details
! _______________
!
!   If size(DAT,DIM) is an exact power of two, a 1D in-place complex-complex radix-2
!   decimation-in-time Cooley-Tukey FFT algorithm is used, otherwise the
!   CHIRP-Z transform is employed.
!
!   Before using FFT_DIM_CT, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( (/ size(DAT,1), size(DAT,2), size(DAT,3) /), dim=DIM )     
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83, see
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!        general length, Appl. Statist., 26 (3), 351-361.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Utilities,         only : merror
    use Char_Constants,    only : allocate_error, fft_error8, fft_error12
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:,:) :: dat
!
    logical(lgl), intent(in) :: forward
!
    integer(i4b), intent(in) :: dim
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: scal
!
    complex(stnd), dimension(:),   allocatable :: tmpvec
    complex(stnd), dimension(:,:), allocatable :: x
!
    integer(i4b) :: n, n1, n2, n3, j, k
!
    integer :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( dim<1_i4b .or. dim>3_i4b ) then
!
        call merror( name_proc//fft_error12 )
!
    end if
!
    n = size( dat, int(dim) )
!
    if ( n/=n_1d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
    n1 = size( dat, 1 )
    n2 = size( dat, 2 )
    n3 = size( dat, 3 )
!
#ifdef _OPENMP
!
    j = ((n1*n2*n2)/n)*nfft_1d
!
    k           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  j>=omp_limit                        .and.      &
                  k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(iok,j,k,x,tmpvec,scal), SHARED(nfft_1d,lognfft_1d)
!
!   ALLOCATE WORK VARIABLES FOR THE FFT OR DISCRETE CHIRP-Z TRANSFORMS.
!
    if ( n_1d/=nfft_1d ) then
!
        allocate( tmpvec(nfft_1d), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
    end if
!
    if ( dim/=1_i4b ) then
!
        allocate( x(n,n1), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
    end if
!
    if ( .not.forward ) then
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
        do k = 1_i4b, n3
!
            do j = 1_i4b, n2
!
                dat(:n1,j,k) = conjg( dat(:n1,j,k) )
!
            end do
!
        end do
!
!$OMP END DO
!
    end if
!
    select case( dim )
!
        case( 1_i4b )
!
!           TRANSFORM FIRST SUBSCRIPT.
!
            if ( n_1d==nfft_1d ) then
!
!               n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
                do k = 1_i4b, n3
!
                    do j = 1_i4b, n2
!
                        call fft2( dat(:n,j,k), w_1d(:), lognfft_1d )
!
                    end do
!
                end do
!
!$OMP END DO
!
            else
!
!               n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
                do k = 1_i4b, n3
!
                    do j = 1_i4b, n2
!
                        tmpvec(:n) = dat(:n,j,k)
!
                        call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                        dat(:n,j,k) = tmpvec(:n)
!
                    end do
!
                end do
!
!$OMP END DO
!
            end if
!
        case( 2_i4b )
!
!           TRANSFORM SECOND SUBSCRIPT.
!
            if ( n_1d==nfft_1d ) then
!
!               n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
                do k = 1_i4b, n3
!
                    x(:n,:n1) = transpose( dat(:n1,:n,k) )
!
                    do j = 1_i4b, n1
!
                        call fft2( x(:n,j), w_1d(:), lognfft_1d )
!
                    end do
!
                    dat(:n1,:n,k) = transpose( x(:n,:n1) )
!
                end do
!
!$OMP END DO
!
            else
!
!               n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
                do k = 1_i4b, n3
!
                    x(:n,:n1) = transpose( dat(:n1,:n,k) )
!
                    do j = 1_i4b, n1
!
                        tmpvec(:n) = x(:n,j)
!
                        call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                        x(:n,j) = tmpvec(:n)
!
                    end do
!
                    dat(:n1,:n,k) = transpose( x(:n,:n1) )
!
                end do
!
!$OMP END DO
!
            end if
!
        case( 3_i4b )
!
!           TRANSFORM THIRD SUBSCRIPT.
!
            if ( n_1d==nfft_1d ) then
!
!               n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
!$OMP DO SCHEDULE(STATIC)
!
                do k = 1_i4b, n2
!
                    x(:n,:n1) = transpose( dat(:n1,k,:n) )
!
                    do j = 1_i4b, n1
!
                        call fft2( x(:n,j), w_1d(:), lognfft_1d )
!
                    end do
!
                    dat(:n1,k,:n) = transpose( x(:n,:n1) )
!
                end do
!
!$OMP END DO
!
            else
!
!               n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!$OMP DO SCHEDULE(STATIC)
!
                do k = 1_i4b, n2
!
                    x(:n,:n1) = transpose( dat(:n1,k,:n) )
!
                    do j = 1_i4b, n1
!
                        tmpvec(:n) = x(:n,j)
!
                        call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                        x(:n,j) = tmpvec(:n)
!
                    end do
!
                    dat(:n1,k,:n) = transpose( x(:n,:n1) )
!
                end do
!
!$OMP END DO
!
            end if
!
    end select
!
    if ( .not.forward ) then
!
        scal = one/real( n, stnd)
!
!$OMP DO SCHEDULE(STATIC), COLLAPSE(2)
!
        do k = 1_i4b, n3
!
            do j = 1_i4b, n2
!
                dat(:n1,j,k) = conjg( dat(:n1,j,k) )*scal
!
            end do
!
        end do
!
!$OMP END DO
!
    end if
!
!   DEALLOCATE WORK VARIABLES.
!
    if ( n_1d/=nfft_1d ) then
!
        deallocate( tmpvec )
!
    end if
!
    if ( dim/=1_i4b ) then
!
        deallocate( x )
!
    end if
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE fft_dim_ct
! ____________________________
!
    end subroutine fft_dim_ct
!
! =========================================================================================
!
    subroutine fftn_row( dat, forward)
!
! Purpose
! _______
!
!   Subroutine FFT_ROW implements the Fast Fourier Transform for a complex valued sequence
!   DAT of general length.
!
!   Forward discrete Fourier transform of a vector DAT(:) is given by              
!
!   t( DAT )(j)  =  [ sum k=0 to nn-1 ]  DAT(k) exp( - i 2 pi j k / nn )
!
!   Backward discrete Fourier transform of a vector DAT(:) is given by              
!
!   t( DAT )(j)  =  (1/nn) [ sum k=0 to nn-1 ]  DAT(k) exp( i 2 pi j k / nn )
!
!   where  i  =  sqrt( -1 ), nn =  size(DAT) and pi =  3.1415923565...
!
!   Note that the indexing of DAT is shifted by one :                     
!   DAT(0) stored in DAT(1), ... , DAT( nn-1 ) stored in DAT( nn ).
!
!
! Arguments
! _________
!
!   DAT      (INPUT/OUTPUT) complex(stnd), dimension(:)
!            On entry, the complex valued sequence to be transformed.
!            On exit, DAT is replaced by the Fourier transform.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!
! Further Details
! _______________
!
!   If size(DAT) is an exact power of two, Bailey's Four-Step FFT algorithm is used,
!   otherwise the CHIRP-Z transform is employed.
!
!   This is the parallelized version of FFT subroutine (Parallelization is done with
!   OPENMP directives).
!
!   Before using FFT_ROW, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( size(DAT) )     
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83.
!   For more details, see:
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!           general length, Appl. Statist., 26 (3), 351-361.
!
!   (2) Bailey, D., 1990: FFTs in External or Hierarchical Memory, 
!           The Journal of Supercomputing, 4, 23-35.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : fft_error8
    use Reals_Constants,   only : zero, one
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:) :: dat
!
    logical(lgl), intent(in) :: forward
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(nfft1_n,nfft2_n) :: tmp2
!
    integer(i4b) :: length, i, i0, i1, i2, ncol
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fft_row'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    length = size( dat )
!
    if ( length/=n_n ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nfft_n>=omp_limit                   .and.      &
                  i>1_i4b
#endif
!
    if ( n_n==nfft_n ) then
!
!       length IS A POWER OF TWO. USE Bailey's FFT.
!
!       RESHAPE THE INPUT ARRAY.
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,i0,i1,i2)
!
!$OMP DO SCHEDULE(STATIC)
!
        do i=1_i4b, nfft2_n
!     
            i0 = (i-1_i4b)*nfft1_n
            i1 = i0 + 1_i4b
            i2 = i0 + nfft1_n
!
            if ( forward ) then
!
                tmp2(:nfft1_n,i) = dat(i1:i2)
!
            else
!
                tmp2(:nfft1_n,i) = conjg( dat(i1:i2) )
!
            end if
!     
        end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!     
!       STEP 1.
!     
        call fft2_row_par( tmp2(:,:), w2_n(:), lognfft2_n )
!     
!       STEPS 2, 3, 4.
!
!$OMP PARALLEL IF(test_par), PRIVATE(i)
!
!$OMP DO SCHEDULE(STATIC)
!
        do i=1_i4b, nfft2_n
!     
!           STEP 2.
!     
            tmp2(:nfft1_n,i) = tmp2(:nfft1_n,i)*v_n(:nfft1_n,i)
!     
!           STEPS 3 AND 4.
!     
            call fft2( tmp2(:nfft1_n,i), w1_n(:), lognfft1_n )
!
            if ( .not.forward ) then
!
                tmp2(:nfft1_n,i) = conjg( tmp2(:nfft1_n,i) )*(one/real(n_n,stnd))
!
            end if
!     
!           RESHAPE THE RESULT BACK TO ONE DIMENSION.  
!     
            dat(i:n_n:nfft2_n) = tmp2(:nfft1_n,i)
!     
        end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
    else
!
!       length IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!       ncol+1 IS THE LAST COLUMN OF tmp2, WHICH CONTAINS THE END OF
!       THE ARAAY DAT.
!
        ncol = n_n/nfft1_n
!
!       FIRST MULTIPLY THE COMPLEX SEQUENCE BY THE CHIRP FUNCTION
!       IN THE TIME DOMAIN AND RESHAPE THE INPUT ARRAY.
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,i0,i1,i2)
!
!$OMP DO SCHEDULE(STATIC)
!
        do i=1_i4b, ncol
!     
            i0 = (i-1_i4b)*nfft1_n
            i1 = i0 + 1_i4b
            i2 = i0 + nfft1_n
!
!           MULTIPLY BY CHIRP FUNCTION.
!
            if ( forward ) then
!     
                tmp2(1_i4b:nfft1_n,i) = dat(i1:i2)*y2_n(i1:i2)
!
            else
!
                tmp2(1_i4b:nfft1_n,i) = conjg( dat(i1:i2) )*y2_n(i1:i2)
!
            end if
!     
        end do
!
!$OMP END DO NOWAIT
!
!$OMP SINGLE
!
        i0 = ncol*nfft1_n
        i1 = i0 + 1_i4b
        i2 = n_n - i0
!
!       MULTIPLY BY CHIRP FUNCTION.
!
        if ( forward ) then
!
            tmp2(1_i4b:i2,ncol+1_i4b) = dat(i1:n_n)*y2_n(i1:n_n)
!
        else
!
            tmp2(1_i4b:i2,ncol+1_i4b) = conjg( dat(i1:n_n) )*y2_n(i1:n_n)
!
        end if
!
!       ZERO PADDING.
!
        tmp2(i2+1_i4b:nfft1_n,ncol+1_i4b)      = cmplx( zero, zero, kind=stnd)
        tmp2(1_i4b:nfft1_n,ncol+2_i4b:nfft2_n) = cmplx( zero, zero, kind=stnd)
!
!$OMP END SINGLE
!
!$OMP END PARALLEL
!        
!       FOURIER TRANSFORM THE CHIRPED SEQUENCE.
!
!       STEP 1.
!     
        call fft2_row_par( tmp2(:,:), w2_n(:), lognfft2_n )
!     
!       STEPS 2, 3, 4.
!
!$OMP PARALLEL IF(test_par), PRIVATE(i),SHARED(lognfft1_n)
!
!$OMP DO SCHEDULE(STATIC)
!
        do i=1_i4b, nfft2_n
!     
!           STEP 2.
!     
            tmp2(:nfft1_n,i) = tmp2(:nfft1_n,i)*v_n(:nfft1_n,i)
!     
!           STEPS 3 AND 4.
!     
            call fft2( tmp2(:nfft1_n,i), w1_n(:), lognfft1_n )
!     
!           CONVOLVE BY FREQUENCY DOMAIN MULTIPLICATION WITH THE CHIRP TRANSFORM.  
!     
            tmp2(:nfft1_n,i) = conjg( tmp2(:nfft1_n,i)*y_n(:nfft1_n,i) )
!
!           BACKWARD FOURIER TRANSFORM THE SEQUENCE.
!
!           STEPS 1 AND 2.
!     
            call fft2( tmp2(:nfft1_n,i), w1_n(:), lognfft1_n )
!     
!           STEP 3.
!     
            tmp2(:nfft1_n,i) = tmp2(:nfft1_n,i)*v_n(:nfft1_n,i)
!     
        end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!       STEP 4.
!     
        call fft2_row_par( tmp2(:,:), w2_n(:), lognfft2_n )
!     
!       RESHAPE THE RESULT BACK TO ONE DIMENSION.  
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,i0,i1,i2)
!
!$OMP DO SCHEDULE(STATIC)
!
        do i=1_i4b, ncol
!     
            i0 = (i-1_i4b)*nfft1_n
            i1 = i0 + 1_i4b
            i2 = i0 + nfft1_n
!
!           MULTIPLY BY CHIRP FUNCTION.
!     
!
            if ( forward ) then
!
                dat(i1:i2) = ( conjg( tmp2(:nfft1_n,i) )*y2_n(i1:i2) )*(one/real(nfft_n,stnd))
!
            else
!
                dat(i1:i2) = ( ( tmp2(:nfft1_n,i)*conjg( y2_n(i1:i2) ) )*(one/real(nfft_n,stnd)) ) &
                             *(one/real(n_n,stnd))
!
            end if
!     
        end do
!
!$OMP END DO NOWAIT
!
!$OMP SINGLE
!
        i0 = ncol*nfft1_n
        i1 = i0 + 1_i4b
        i2 = n_n - i0
!
!       MULTIPLY BY CHIRP FUNCTION.
!     
        if ( forward ) then
!
            dat(i1:n_n) = ( conjg( tmp2(1_i4b:i2,ncol+1_i4b) )*y2_n(i1:n_n) )*(one/real(nfft_n,stnd))
!
        else
!
            dat(i1:n_n) = ( ( tmp2(1_i4b:i2,ncol+1_i4b)*conjg( y2_n(i1:n_n) ) )*(one/real(nfft_n,stnd)) ) &
                          *(one/real(n_n,stnd))
!
        end if
!
!$OMP END SINGLE
!
!$OMP END PARALLEL
!
    end if
!
!
! END OF SUBROUTINE fftn_row
! __________________________
!
    end subroutine fftn_row
!
! =========================================================================================
!
    subroutine fft_row_cm( dat, forward)
!
! Purpose
! _______
!
!   Subroutine FFT_ROW replaces each row of DAT by its Fourier transform.
!   Size(DAT,2) may be of general length.
!
!
! Arguments
! _________
!
!   DAT      (INPUT/OUTPUT) complex(stnd), dimension(:,:)
!            On entry, the complex valued sequences to be transformed.
!            On exit, each row of DAT is replaced by its Fourier transform.
!
!   FORWARD  (INPUT) logical(lgl)
!            Specifies whether a forward or backward Fourier transform
!            is desired. If:
!
!            - FORWARD = true:  a forward Fourier transform is computed
!            - FORWARD = false:  a backward Fourier transform is computed.
!
!
! Further Details
! _______________
!
!   If size(DAT,2) is an exact power of two, a 1D complex-complex radix-2
!   decimation-in-time Cooley-Tukey FFT algorithm is used, otherwise the
!   CHIRP-Z transform is employed.
!
!   This is a parallelized FFT subroutine if OPENMP is used (Parallelization is
!   done with OPENMP directives).
!
!   Before using FFT_ROW, the user must call subroutine INIT_FFT as follows :
!
!                call init_fft( (/ size(DAT,1), size(DAT,2) /), dim=2 )     
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83.
!   For further details, see :
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!          general length, Appl. Statist., 26 (3), 351-361.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Char_Constants,    only : allocate_error, fft_error8
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:) :: dat
!
    logical(lgl), intent(in) :: forward
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd), dimension(:,:), allocatable :: x
!
    integer(i4b) :: m, n, p, i, i1, i2, i3, nblksz
    integer      :: iok
!
#ifdef _OPENMP
    complex(stnd), dimension(:), allocatable :: tmpvec
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fft_row'
!
!   max_alloc IS A FACTOR WHICH ALLOWS TO REDUCE THE WORKSPACE USED
!   TO COMPUTE THE FOURIER TRANSFORM OF THE DATA IF NECESSARY FOR THE SERIAL
!   VERSION OF THE ALGORITHM.
!   blksz IS THE BLOCK SIZE FOR THE PARALLELIZED VERSION OF THE ALGORITHM,
!   PARALLELIZATION IS DONE OVER THE ROWS OF dat.
!   nblksz_min IS THE MINIMUM OF BLOCKS OF ROWS OF SIZE blksz FOR USING
!   THE PARALLELIZED VERSION OVER THE ROWS OF dat.
!
#ifdef _OPENMP
    integer(i4b),  parameter :: max_alloc=4_i4b, blksz=1024_i4b, nblksz_min=4_i4b
#else
    integer(i4b),  parameter :: max_alloc=4_i4b
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( dat, 1 )
    n = size( dat, 2 )
!
!   TEST THE ARGUMENTS.
!
    if ( n/=n_1d ) then
!
        call merror( name_proc//fft_error8 )
!
    end if
!
#ifdef _OPENMP
    nblksz = m/blksz
!
    if ( m/=nblksz*blksz ) then
!
        nblksz = nblksz + 1_i4b
!
    end if
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m*n>=omp_limit                      .and.      &
                  nblksz>=nblksz_min                  .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!$OMP PARALLEL IF(test_par),                           &
!$OMP          PRIVATE(iok, i, i1, i2, i3, p, x, tmpvec)
!
!       ALLOCATE WORK VARIABLE FOR THE Cooley-Tukey FFT OR
!       DISCRETE CHIRP-Z TRANSFORM.
!
        if ( n==nfft_1d ) then
!
            allocate( x(n,blksz), stat=iok )
!
        else
!
            allocate( x(n,blksz), tmpvec(nfft_1d), stat=iok )
!
        end if
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!$OMP DO SCHEDULE(STATIC)
!
        do  i=0_i4b, m-1_i4b, blksz
!
            i1 = i + 1_i4b
            i2 = min( m, i+blksz )
            i3 = i2 - i1 + 1_i4b        
!
            x(:n,:i3) = transpose( dat(i1:i2,:n) )
!
            do p=1_i4b, i3
!
                if ( .not.forward ) then
!
                    x(:n,p) = conjg( x(:n,p) )
!
                end if
!
                if ( n==nfft_1d ) then
!
!                   n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
                    call fft2( x(:n,p), w_1d(:), lognfft_1d )
!
                else
!
!                   n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
                    tmpvec(:n) = x(:n,p)
!
                    call chirp( tmpvec(:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                    x(:n,p) = tmpvec(:n)
!
                end if
!
                if ( .not.forward ) then
!
                    x(:n,p) = conjg( x(:n,p) )*(one/real(n,stnd))
!
                end if
!
            end do
!
            dat(i1:i2,:n) = transpose( x(:n,:i3) )
!
        end do
!
!$OMP END DO
!
!       DEALLOCATE WORK VARIABLES.
!
        if ( n==nfft_1d ) then
!
            deallocate( x )
!
        else
!
            deallocate( x, tmpvec )
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
        if ( n==nfft_1d ) then
!
            if ( .not.forward ) then
!
                dat(:m,:n) = conjg( dat(:m,:n) )
!
            end if
!
!           n IS A POWER OF TWO. USE Cooley-Tukey FFT.
!
            call fft2_row( dat(:m,:n), w_1d(:), lognfft_1d )
!
            if ( .not.forward ) then
!
                dat(:m,:n) = conjg( dat(:m,:n) )*(one/real(n,stnd))
!
            end if
!
        else
!
!           ALLOCATE WORK VARIABLE FOR THE DISCRETE CHIRP-Z TRANSFORM.
!
            p = m
!
            do i=1_i4b, max_alloc
!
                allocate( x(p,nfft_1d), stat=iok )
!
                if ( iok==0 ) exit
!
                p = p/2_i4b
!
            end do
!
            if ( iok==0 ) then
!
                nblksz = m/p
!
                if ( nblksz*p/=m ) nblksz = nblksz + 1_i4b
!
            else
!
                call merror( name_proc//allocate_error )
!
            end if
!
            i2 = 0_i4b
!
            do i=1_i4b, nblksz
!
                i1 = i2 + 1_i4b
                i2 = min( m, i2+p )
                i3 = i2 - i1 + 1_i4b        
!
!               n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
                if ( forward ) then
!
                    x(:i3,:n) = dat(i1:i2,:n)
!
                else
!
                    x(:i3,:n) = conjg( dat(i1:i2,:n) )
!
                end if
!
                call chirp_row( x(:i3,:nfft_1d), w_1d(:), y_1d(:), y2_1d(:), lognfft_1d )
!
                if ( forward ) then
!
                    dat(i1:i2,:n) = x(:i3,:n)
!
                else
!
                    dat(i1:i2,:n) = conjg( x(:i3,:n) )*(one/real(n,stnd))
!
                end if
!
            end do
!
!           DEALLOCATE WORK VARIABLE.
!
            deallocate( x )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE fft_row_cm
! ____________________________
!
    end subroutine fft_row_cm
!
! =========================================================================================
!
    subroutine end_fft()
!
! Purpose
! _______
!
!   END_FFT deallocates the workspace previously allocated by a call to INIT_FFT.
!
! Arguments
! _________
!
!   None
!
!
! _________________________________________________________________________________________
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n_1d       = -1_i4b
    n_2d       = -1_i4b
    n_3d       = -1_i4b
    n_n        = -1_i4b
!
    nfft_1d    = -1_i4b
    nfft_2d    = -1_i4b
    nfft_3d    = -1_i4b
    nfft_n     = -1_i4b
    nfft1_n    = -1_i4b
    nfft2_n    = -1_i4b
!
    lognfft_1d    = -1_i4b
    lognfft_2d    = -1_i4b
    lognfft_3d    = -1_i4b
    lognfft1_n    = -1_i4b
    lognfft2_n    = -1_i4b
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( allocated( w_1d  )   ) deallocate( w_1d   )
    if ( allocated( w_2d  )   ) deallocate( w_2d   )
    if ( allocated( w_3d  )   ) deallocate( w_3d   )
    if ( allocated( w1_n  )   ) deallocate( w1_n   )
    if ( allocated( w2_n  )   ) deallocate( w2_n   )
!
    if ( allocated( y_1d  )   ) deallocate( y_1d   )
    if ( allocated( y_2d  )   ) deallocate( y_2d   )
    if ( allocated( y_3d  )   ) deallocate( y_3d   )
    if ( allocated( y_n  )    ) deallocate( y_n    )
!
    if ( allocated( y2_1d )   ) deallocate( y2_1d  )
    if ( allocated( y2_2d )   ) deallocate( y2_2d  )
    if ( allocated( y2_3d )   ) deallocate( y2_3d  )
    if ( allocated( y2_n )    ) deallocate( y2_n   )
!
    if ( allocated( v_n )     ) deallocate( v_n    )
!
!
! END OF SUBROUTINE end_fft
! _________________________
!
    end subroutine end_fft
!
!
! =========================================================================================
!                                 PRIVATE SUBROUTINES
! =========================================================================================
!
!
    subroutine init_fft_cv( length1 )
!
! Purpose
! _______
!
!   Subroutine INIT_FFT_CV sets up constants, the Chirp function and the
!   Fourier transform of the Chirp function for use by generic subroutines FFT
!   and FFTXY (when the DIM argument is used) for a series of length LENGTH1.
!
! Arguments
! _________
!
!   LENGTH1      (INPUT) integer(i4b)
!                The length of the complex valued sequence to be transformed.
!                LENGTH1 may be any positive integer.
!
!
! Further Details
! _______________
!
!   INIT_FFT_CV is a private subroutine.
!
!   INIT_FFT_CV is first called to establish and transform the Chirp function and
!   other constants. Then, subroutine FFT (or FFTXY) can be called any number of times without
!   the precalculated constants being destroyed; a further call to INIT_FFT_CV will only
!   be necessary if a new length is required.
!
!   This subroutine is adapted from Applied Statistics algorithms AS 117 and AS 83, see
!
!   (1) Monro, D.M., and Branch, J.L., 1977: The Chirp discrete Fourier transform of
!        general length, Appl. Statist., 26 (3), 351-361.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error, fft_error4
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: length1
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='init_fft'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENT.
!
    if ( length1<0_i4b ) then
        call merror( name_proc//fft_error4 )
    end if
!
    n_1d = length1
!
!   COMPUTE POWERS OF TWO.
!
    call powers_of_two( n_1d, nfft_1d, lognfft_1d )
!
!   ALLOCATE WORK VARIABLES FOR THE COOLEY-TUKEY FFT.
!
    if ( allocated( w_1d   ) )  deallocate( w_1d   )
!
    allocate( w_1d(nfft_1d/2_i4b), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!   GENERATE POWERS OF ww = e**((-2*PI)/nfft) SUCH THAT
!
!       w(i) = ww**(i-1) .
!
    call gen_w_table( w_1d(:) )
!
    if ( n_1d/=nfft_1d ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE CHIRP FUNCTION
!       AND TRANSFORM.
!
        if ( allocated( y_1d  ) ) deallocate( y_1d  )
        if ( allocated( y2_1d ) ) deallocate( y2_1d )
!
        allocate( y_1d(nfft_1d), y2_1d(n_1d), stat = iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!
!       GENERATE CHIRP FUNCTION AND TRANSFORM.
!
        call gen_chirp_tables( lognfft_1d, w_1d(:), y_1d(:), y2_1d(:) )
!
    end if
!
!
! END OF SUBROUTINE init_fft_cv
! _____________________________
!
    end subroutine init_fft_cv
!
! =========================================================================================
!
    subroutine powers_of_two( n, nfft, lognfft)
!
! Purpose
! _______
!
!   Check if N is an exact power of two. If it is the case,
!   set NFFT to N.
!   if N is not a power of two. find integer NFFT
!   which is a power of two and at least 2 * N .
!
!   Then, define LOGNFFT as NFFT=2**(LOGNFFT)
!
!
! Arguments
! _________
!
!   N       (INPUT) integer(i4b)
!           On entry, the value of N.
!
!   NFFT    (OUTPUT) integer(i4b)
!           On output,  the value of NFFT.
!
!   LOGNFFT (OUTPUT) integer(i4b)
!           On output, the value of LOGNFFT.
!
!
! Further Details
! _______________
!            
!   POWERS_OF_TWO is a private subroutine.
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: n
    integer(i4b), intent(out) :: nfft, lognfft
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   COMPUTE POWERS OF TWO.
!
    nfft = 1_i4b
!
    do lognfft=0_i4b, bit_size( n ) - 2_i4b
!
        if ( n<=nfft ) exit
        nfft = 2_i4b*nfft
!
    end do
!
    if ( n/=nfft ) then
!
!       n IS NOT A POWER OF TWO. FIND INTEGER nfft
!       WHICH IS A POWER OF TWO AND AT LEAST 2*n .
!
        nfft = 2*nfft
        lognfft = lognfft + 1_i4b
!
    end if
!
!
! END OF SUBROUTINE powers_of_two
! _______________________________
!
    end subroutine powers_of_two
!
! =========================================================================================
!
    subroutine gen_w_table( w )
!
! Purpose
! _______
!
!   Generate powers of ww = e**((- pi)/size( W )) such that
!
!       W(i) = ww**(i-1)
!
!   for i=1 to size( W ).
!
!
! Arguments
! _________
!
!   W     (OUTPUT) complex(stnd), dimension(:)
!         On output, the powers of ww.
!
!
! Further Details
! _______________
!            
!   GEN_W_TABLE is a private subroutine.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : one_extd, zero_extd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(out), dimension(:)  :: w
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n
!
    real(extd) :: y
!
    complex(extd) :: ww, x
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( w )
!
!   COMPUTE ww .
!
    y = acos( -one_extd )/real( n, kind=extd )
    ww = cmplx( cos( y ), -sin( y ), kind=extd )
!
!   COMPUTE W .
!
    x = cmplx( one_extd, zero_extd, kind=extd )
    w = geop_extd_c( x, ww, n )
!
!
! END OF SUBROUTINE gen_w_table
! _____________________________
!
    end subroutine gen_w_table
!
! =========================================================================================
!
    function geop_extd_c( first, factor, n )
!
! Purpose
! _______
!
!   Returns an one-dimensional complex array of length N containing a geometric 
!   progression whose first value is FIRST and whose multiplier is FACTOR.
!
!
! Further Details
! _______________
!            
!   GEOP_EXTD_C is a private subroutine.
!   
!
! __________________________________________________________________________________________
!
    complex(extd), intent(in) :: first, factor
    integer(i4b),  intent(in) :: n
!
    complex(extd), dimension(n) :: geop_extd_c
    complex(extd)               :: temp
    integer(i4b)                :: k, k2
!
    if ( n>0_i4b ) geop_extd_c(1) = first
!
    if ( n<=npar_geop ) then
!
        do k = 2_i4b, n
!
            geop_extd_c(k) = geop_extd_c(k-1_i4b)*factor
!
        end do
!
    else
!
        do k = 2_i4b, npar2_geop
!
            geop_extd_c(k) = geop_extd_c(k-1_i4b)*factor
!
        end do
!
        temp = factor**npar2_geop
        k = npar2_geop
!
        do
!
            if ( k>=n ) exit
!
            k2 = k + k
            geop_extd_c(k+1_i4b:min(k2,n)) = temp*geop_extd_c(1_i4b:min(k,n-k))
            temp = temp*temp
            k = k2
!
        end do
!
    end if
!
    end function geop_extd_c
!
! =========================================================================================
!
    subroutine gen_v_table( v )
!
! Purpose
! _______
!
!   Generate 2d twiddle factors.
!
!
! Arguments
! _________
!
!   V     (OUTPUT) complex(stnd), dimension(:,:)
!         On output, the 2d twiddle factors.
!
!
! Further Details
! _______________
!            
!   GEN_V_TABLE is a private subroutine.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : one_extd, two_extd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(out), dimension(:,:)  :: v
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nfft1, nfft2, nfft, j, k
!
    real(extd)   :: tc
!
    complex(extd) :: wn
!
#ifdef _OPENMP
    logical  :: test_par
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!   
    nfft1 = size( v , 1 )
    nfft2 = size( v , 2 )
    nfft = nfft1*nfft2
!   
    tc = two_extd*acos(-one_extd)/nfft
    wn = cmplx( cos( tc ), -sin( tc ), kind=extd )
!   
#ifdef _OPENMP
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )   .and.      &
               nfft1*nfft2>=omp_limit       .and.      &
               k>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(k,j)
!
!$OMP DO SCHEDULE(STATIC)
!
    do k=1_i4b, nfft2
!
        do j=1_i4b, nfft1
            v(j,k) = wn**((j-1_i4b)*(k-1_i4b))
        end do
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE gen_v_table
! _____________________________
!
    end subroutine gen_v_table
!
! =========================================================================================
!
    subroutine gen_chirp_tables_n( y, y2 )
!
! Purpose
! _______
!
!   Set up Chirp function and Fourier transformed Chirp function for FFTN subroutine.
!
!
! Arguments
! _________
!
!   Y        (OUTPUT) complex(stnd), dimension(:,:)
!            On output, the Fourier transformed Chirp function reshaped as a 2d array.
!
!   Y2       (OUTPUT) complex(stnd), dimension(:)
!            On output, the Chirp function.
!
!
! Further Details
! _______________
!            
!   GEN_CHIRP_TABLES_N is a private subroutine.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : one_extd, zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(out), dimension(nfft1_n,nfft2_n) :: y
    complex(stnd), intent(out), dimension(n_n)             :: y2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: j
!
    real(extd) :: tc, z
!
    complex(stnd), dimension(nfft_n) :: tmp
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   SET UP BOTTOM SEGMENT OF CHIRP FUNCTION.
!
    tc = acos( -one_extd )/real( n_n, extd )
!
    do j = 1_i4b, n_n
!
        z = real( j-1_i4b, extd )
        z = z*z*tc
        tmp(j) = cmplx( cos(z), sin(z), kind=extd )
!
    end do
!
!   CLEAR MIDDLE SEGMENT OF CHIRP FUNCTION.
!
    tmp(n_n+1_i4b:nfft_n-n_n+1_i4b) = cmplx( zero,zero, kind=stnd )
!
!   SET UP UPPER SEGMENT OF CHIRP FUNCTION.
!
    tmp(nfft_n-n_n+2_i4b:nfft_n) = tmp(n_n:2_i4b:-1_i4b)
!
!   SAVE THE CHIRP FUNCTION.
!
    y2(:n_n) = conjg( tmp(:n_n) )
!
!   FORWARD FOURIER TRANSFORM THE CHIRP FUNCTION.
!
!   RESHAPE THE CHIRP FUNCTION.
!     
    y(:nfft1_n,:nfft2_n) = reshape( tmp, shape(y) )
!     
!   STEP 1.
!     
    call fft2_row_par( y(:,:), w2_n(:), lognfft2_n )
!
#ifdef _OPENMP
    j = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )   .and.      &
               nfft_n>=omp_limit            .and.      &
               j>1_i4b
#endif
!
!   STEPS 2, 3, 4.
!
!$OMP PARALLEL IF(test_par), PRIVATE(j), SHARED(lognfft1_n)
!
!$OMP DO SCHEDULE(STATIC)
!
    do j=1_i4b, nfft2_n
!     
!       STEP 2.
!     
        y(:,j) = y(:,j)*v_n(:,j)
!     
!       STEPS 3 AND 4.
!     
        call fft2( y(:,j), w1_n(:), lognfft1_n )
!     
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE gen_chirp_tables_n
! ____________________________________
!
    end subroutine gen_chirp_tables_n
!
! =========================================================================================
!
    subroutine gen_chirp_tables( lognfft, w, y, y2 )
!
! Purpose
! _______
!
!   Set up Chirp function and Fourier transformed Chirp function.
!
!
! Arguments
! _________
!
!   LOGNFFT  (OUTPUT) integer(i4b)
!            On entry, the value LOGNFFT such that size(Y)=2**(LOGNFFT).
!
!   W        (INPUT) complex(stnd), dimension(:)
!            On entry, the powers of ww = e**((-PI)/size(W)).
!                         W(i) = ww**(i-1) .
!                   
!   Y        (OUTPUT) complex(stnd), dimension(:)
!            On output, the Fourier transformed Chirp function.
!
!   Y2       (OUTPUT) complex(stnd), dimension(:)
!            On output, the Chirp function.
!
!
! Further Details
! _______________
!            
!   GEN_CHIRP_TABLES is a private subroutine.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : one_extd, zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: lognfft
!
    complex(stnd), intent(in),  dimension(:) :: w
    complex(stnd), intent(out), dimension(:) :: y, y2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, nfft, nfftdv2, j
!
    real(extd) :: tc, z
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nfft    = size( y  )
    n       = size( y2 )
    nfftdv2 = size( w  )
!
!   SET UP BOTTOM SEGMENT OF CHIRP FUNCTION.
!
    tc = acos( -one_extd )/real( n, extd )
!
    do j = 1_i4b, n
!
        z = real( j-1_i4b, extd )
        z = z*z*tc
        y(j) = cmplx( cos(z), sin(z), kind=extd )
!
    end do
!
!   CLEAR MIDDLE SEGMENT OF CHIRP FUNCTION.
!
    y(n+1_i4b:nfft-n+1_i4b) = cmplx( zero, zero, kind=stnd )
!
!   SET UP UPPER SEGMENT OF CHIRP FUNCTION.
!
    y(nfft-n+2_i4b:nfft) = y(n:2_i4b:-1_i4b)
!
!   SAVE THE CHIRP FUNCTION.
!
    y2(:n) = conjg( y(:n) )
!
!   FORWARD FOURIER TRANSFORM THE CHIRP FUNCTION.
!
    call fft2( y(:nfft), w(:nfftdv2), lognfft )
!
!
! END OF SUBROUTINE gen_chirp_tables
! __________________________________
!
    end subroutine gen_chirp_tables
!
! =========================================================================================
!
    subroutine chirp( x, w, y, y2, lognfft  )
!
! Purpose
! _______
!
!   Subroutine CHIRP replaces the vector X by its Chirp transform.
!
!
! Arguments
! _________
!
!   X        (INPUT/OUTPUT) complex(stnd), dimension(:)
!            On entry, the complex valued sequence to be transformed.
!            On exit, X is replaced by its Chirp transform.
!            Size(X) must be an exact power of 2.
!
!   W        (INPUT) complex(stnd), dimension(:)
!            On entry, the powers of ww = e**((-PI)/size(W)).
!
!                         W(i) = ww**(i-1) .
!
!            size(W) must verify size(W) = size(X)/2.
!
!   Y        (INPUT) complex(stnd), dimension(:)
!            On entry, the appropriate Fourier transformed Chirp function.
!
!            size(Y) must verify size(Y) = size(X) = 2**(LOGNFFT).
!
!   Y2       (INPUT) complex(stnd), dimension(:)
!            On entry, the appropriate Chirp function. size(Y2) is the exact
!            size of the complex valued sequence to be transformed and stored
!            in argument X before extension of this sequence to an exact power of 2.
!
!            size(Y2) must verify size(Y2) <= size(X).
!
!   LOGNFFT  (INPUT) integer(i4b)
!            On entry, LOGNFFT must verify size(Y) = size(X,2) = 2**(LOGNFFT).
!
!
! Further Details
! _______________
!
!   Since subroutine CHIRP is a private entity, no checks on the sizes of the
!   arguments are done inside the subroutine.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:) :: x
!
    complex(stnd), intent(in), dimension(size(x))   :: y
    complex(stnd), intent(in), dimension(size(x)/2) :: w
    complex(stnd), intent(in), dimension(:)         :: y2
!
    integer(i4b), intent(in) :: lognfft
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, nfft
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n       = size( y2 )
    nfft    = size( x )
!
!   n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!   FIRST MULTIPLY THE COMPLEX SEQUENCE BY THE CHIRP FUNCTION
!   IN THE TIME DOMAIN.
!
    x(:n) = x(:n)*y2(:n)
!
    x(n+1_i4b:nfft) = cmplx( zero, zero, kind=stnd )
!
!   FOURIER TRANSFORM THE CHIRPED SEQUENCE.
!
    call fft2( x(:nfft), w(:), lognfft )
!
!   CONVOLVE BY FREQUENCY DOMAIN MULTIPLICATION WITH THE CHIRP TRANSFORM.
!
    x(:nfft) = conjg( x(:nfft)*y(:nfft) )
!
!   BACKWARD FOURIER TRANSFORM THE SEQUENCE.
!
    call fft2( x(:nfft), w(:), lognfft  )
!
!   MULTIPLY BY CHIRP FUNCTION.
!
    x(:n) = ( conjg( x(:n) )*y2(:n) )*(one/real(nfft,stnd))
!
!
! END OF SUBROUTINE chirp
! _______________________
!
    end subroutine chirp
!
! =========================================================================================
!
    subroutine chirp_row( x, w, y, y2, lognfft  )
!
! Purpose
! _______
!
!   Subroutine CHIRP_ROW replaces each row of X by its Chirp transform.
!
!
! Arguments
! _________
!
!   X        (INPUT/OUTPUT) complex(stnd), dimension(:,:)
!            On entry, the complex valued sequences to be transformed.
!            On exit, each row of X is replaced by its Chirp transform.
!            Size(X,2) must be an exact power of 2.
!
!   W        (INPUT) complex(stnd), dimension(:)
!            On entry, the powers of ww = e**((-PI)/size(W)).
!
!                         W(i) = ww**(i-1) .
!
!            size(W) must verify size(W) = size(X,2)/2.
!
!   Y        (INPUT) complex(stnd), dimension(:)
!            On entry, the appropriate Fourier transformed Chirp function.
!
!            size(Y) must verify size(Y) = size(X,2) = 2**(LOGNFFT).
!
!   Y2       (INPUT) complex(stnd), dimension(:)
!            On entry, the appropriate Chirp function. size(Y2) is the exact
!            size of the complex valued sequences to be transformed and stored
!            in argument X before extension of these sequences to an exact power of 2.
!
!            size(Y2) must verify size(Y2) <= size(X,2).
!
!   LOGNFFT  (INPUT) integer(i4b)
!            On entry, LOGNFFT must verify size(Y) = size(X,2) = 2**(LOGNFFT).
!
!
! Further Details
! _______________
!
!   Since subroutine CHIRP_ROW is a private entity, no checks on the sizes of the
!   arguments are done inside the subroutine.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:) :: x
!
    complex(stnd), intent(in), dimension(size(x,2))   :: y
    complex(stnd), intent(in), dimension(size(x,2)/2) :: w
    complex(stnd), intent(in), dimension(:)           :: y2
!
    integer(i4b), intent(in) :: lognfft
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, n, nfft, j
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n       = size( y2 )
    m       = size( x, 1 )
    nfft    = size( x, 2 )
!
!   n IS NOT A POWER OF TWO. USE DISCRETE CHIRP-Z TRANSFORM.
!
!   FIRST MULTIPLY THE COMPLEX SEQUENCE BY THE CHIRP FUNCTION
!   IN THE TIME DOMAIN.
!
    do j = 1_i4b, n
!
        x(:m,j) = x(:m,j)*y2(j)
!
    end do
!
    x(:m,n+1_i4b:nfft) = cmplx( zero, zero, kind=stnd )
!
!   FOURIER TRANSFORM THE CHIRPED SEQUENCE.
!
    call fft2_row( x(:m,:nfft), w(:), lognfft )
!
!   CONVOLVE BY FREQUENCY DOMAIN MULTIPLICATION WITH THE CHIRP TRANSFORM.
!
    do j = 1_i4b, nfft
!
        x(:m,j) = conjg( x(:m,j)*y(j) )
!
    end do
!
!   BACKWARD FOURIER TRANSFORM THE SEQUENCE.
!
    call fft2_row( x(:m,:nfft), w(:), lognfft  )
!
!   MULTIPLY BY CHIRP FUNCTION.
!
    do j = 1_i4b, n
!
        x(:m,j) = ( conjg( x(:m,j) )*y2(j) )*(one/real(nfft,stnd))
!
    end do
!
!
! END OF SUBROUTINE chirp_row
! ___________________________
!
    end subroutine chirp_row
!
! =========================================================================================
!
    subroutine fft2( vec, w, lognfft )
!
! Purpose
! _______
!
!   Subroutine FFT2 implements the fast Fourier transform for a complex valued sequence
!   whose length is constrained to be an exact power of two.
!
!
! Arguments
! _________
!
!   VEC      (INPUT/OUTPUT) complex(stnd), dimension(:)
!            On entry, the complex valued sequence to be transformed.
!            On exit, VEC is replaced by the Fourier transform.
!
!   W        (INPUT) complex(stnd), dimension(:)
!            On entry, the powers of ww = e**((-PI)/size(W)).
!
!                         W(i) = ww**(i-1) .
!
!   LOGNFFT  (INPUT) integer(i4b)
!            On entry, LOGNFFT must verify size(VEC) = 2**(LOGNFFT).
!
! Further Details
! _______________
!
!   This private subroutine is used by subroutines INIT_FFT and FFT. This is a 1D
!   complex-complex radix-2 decimation-in-time Cooley-Tukey algorithm.
!
!   VEC and W must verify size(VEC) = 2 * size(W).
!
!   For more details, see:
!
!   (1) Cooley, J.W., Lewis, P., and Welch, P., 1969: The Fast Fourier Transform
!           and its Applications", IEEE Trans on Education, 12, 1, 28-34.
!
!      
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
#ifdef _ALLOC
    use Char_Constants,    only : allocate_error
    use Utilities,         only : merror
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:)           :: vec
    complex(stnd), intent(in),    dimension(size(vec)/2) :: w
!
    integer(i4b), intent(in) :: lognfft
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd) :: ww
#ifdef _ALLOC
    complex(stnd), dimension(:), allocatable :: tmp, vecw
#else
    complex(stnd), dimension(size(vec)/4)  :: tmp
    complex(stnd), dimension(size(vec))    :: vecw
#endif
!
    integer(i4b) :: nfft, nfftdv2, nfftdv2m1, i, j, l, m, stage, spowerofw
#ifdef _ALLOC
    integer      :: iok
#endif
!
    logical(lgl)  :: data_in_vecw
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='fft2'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nfft      = size( vec )
    nfftdv2   = nfft/2_i4b
    nfftdv2m1 = nfftdv2 - 1_i4b
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLES FOR THE Cooley-Tukey FFT.
!
    i = nfft/4_i4b
!
    allocate( vecw(nfft), tmp(i), stat=iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
!   FIRST STAGE.
!
    vecw(:nfftdv2)           = vec(:nfftdv2) + vec(nfftdv2+1_i4b:nfft)
    vecw(nfftdv2+1_i4b:nfft) = vec(:nfftdv2) - vec(nfftdv2+1_i4b:nfft)
!
    data_in_vecw = true
!     
    spowerofw = nfftdv2
!     
    do stage = 2_i4b, lognfft
!     
        spowerofw = spowerofw/2_i4b
!
        if ( data_in_vecw ) then
!
            do i=0_i4b, nfftdv2m1, spowerofw
!     
                j = 2_i4b*i
                l = j + spowerofw
                m = nfftdv2 + i
!     
!               INVARIANT : ww = W**(i)
!     
                ww = w(i+1_i4b)
!     
                tmp(:spowerofw)          = ww*vecw(l+1_i4b:l+spowerofw)
                vec(i+1_i4b:i+spowerofw) = vecw(j+1_i4b:j+spowerofw) + tmp(:spowerofw)
                vec(m+1_i4b:m+spowerofw) = vecw(j+1_i4b:j+spowerofw) - tmp(:spowerofw)
!     
            end do      
!
        else
!
            do i=0_i4b, nfftdv2m1, spowerofw
!     
                j = 2_i4b*i
                l = j + spowerofw
                m = nfftdv2 + i
!     
!               INVARIANT : ww = W**(i)
!     
                ww = w(i+1_i4b)
!     
                tmp(:spowerofw)           = ww*vec(l+1_i4b:l+spowerofw)
                vecw(i+1_i4b:i+spowerofw) = vec(j+1_i4b:j+spowerofw) + tmp(:spowerofw)
                vecw(m+1_i4b:m+spowerofw) = vec(j+1_i4b:j+spowerofw) - tmp(:spowerofw)
!     
            end do      
!
        end if
!
        data_in_vecw = .not.data_in_vecw
!
    end do
!
    if ( data_in_vecw ) then
!
        vec(:nfft) = vecw(:nfft)
!
    end if
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( vecw, tmp )
#endif
!
!
! END OF SUBROUTINE fft2
! ______________________
!
    end subroutine fft2
!
! =========================================================================================
!
    subroutine fft2_row( mat, w, lognfft )
!
! Purpose
! _______
!
!   Subroutine FFT2_ROW replaces each row of MAT by its Fourier transform.
!   size(MAT,2) is constrained to be an exact power of two.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) complex(stnd), dimension(:,:)
!            On entry, the complex valued sequences to be transformed.
!            On exit, each row of MAT is replaced by its Fourier transform.
!
!   W        (INPUT) complex(stnd), dimension(:)
!            On entry, the powers of ww = e**((-PI)/size(W)).
!
!                         W(i) = ww**(i-1) .
!
!   LOGNFFT  (INPUT) integer(i4b)
!            On entry, LOGNFFT must verify size(MAT,2) = 2**(LOGNFFT).
!
! Further Details
! _______________
!
!   This private subroutine is used by subroutines INIT_FFT and FFT. This is a 1D
!   complex-complex radix-2 decimation-in-time Cooley-Tukey algorithm.
!
!   MAT and W must verify size(MAT,2) = 2 * size(W).
!
!   For more details, see:
!
!   (1) Cooley, J.W., Lewis, P., and Welch, P., 1969: The Fast Fourier Transform
!           and its Applications", IEEE Trans on Education, 12, 1, 28-34.
!      
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
#ifdef _ALLOC
    use Char_Constants,    only : allocate_error
    use Utilities,         only : merror
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:)           :: mat
    complex(stnd), intent(in),    dimension(size(mat,2)/2) :: w
!
    integer(i4b), intent(in) :: lognfft
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    complex(stnd) :: ww
#ifdef _ALLOC
    complex(stnd), dimension(:), allocatable   :: tmp
    complex(stnd), dimension(:,:), allocatable :: matw
#else
    complex(stnd), dimension(size(mat,1))             :: tmp
    complex(stnd), dimension(size(mat,1),size(mat,2)) :: matw
#endif
!
    integer(i4b) :: n1, nfft, nfftdv2, nfftdv2m1, i, j, k, l,  &
                    m, stage, spowerofw
#ifdef _ALLOC
    integer      :: iok
#endif
!
    logical(lgl)  :: data_in_matw
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='fft2_row'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n1        = size( mat, 1 )
    nfft      = size( mat, 2 )
    nfftdv2   = nfft/2_i4b
    nfftdv2m1 = nfftdv2 - 1_i4b
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLES FOR THE Cooley-Tukey FFT.
!
    allocate( matw(n1,nfft), tmp(n1), stat=iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
!   FIRST STAGE.
!
    do i = 1_i4b, nfftdv2
!
        j = i + nfftdv2
!
        matw(:n1,i) = mat(:n1,i) + mat(:n1,j)
        matw(:n1,j) = mat(:n1,i) - mat(:n1,j)
!
    end do      
!
    data_in_matw = true
!     
    spowerofw = nfftdv2
!     
    do stage = 2_i4b, lognfft
!     
        spowerofw = spowerofw/2_i4b
!
        if ( data_in_matw ) then
!
            do i=0_i4b, nfftdv2m1, spowerofw
!     
                j = 2_i4b*i
                l = j + spowerofw
                m = nfftdv2 + i
!     
!               INVARIANT : ww = W**(i)
!     
                ww = w(i+1_i4b)
!     
                do k=1_i4b, spowerofw
!     
                    tmp(:n1)     = ww*matw(:n1,l+k)
                    mat(:n1,i+k) = matw(:n1,j+k) + tmp(:n1)
                    mat(:n1,m+k) = matw(:n1,j+k) - tmp(:n1)
!     
                end do      
!     
            end do      
!
        else
!
            do i=0_i4b, nfftdv2m1, spowerofw
!     
                j = 2_i4b*i
                l = j + spowerofw
                m = nfftdv2 + i
!     
!               INVARIANT : ww = W**(i)
!     
                ww = w(i+1_i4b)
!     
                do k=1_i4b, spowerofw
!     
                    tmp(:n1)      = ww*mat(:n1,l+k)
                    matw(:n1,i+k) = mat(:n1,j+k) + tmp(:n1)
                    matw(:n1,m+k) = mat(:n1,j+k) - tmp(:n1)
!     
                end do      
!     
            end do      
!
        end if
!
        data_in_matw = .not.data_in_matw
!
    end do
!
    if ( data_in_matw ) then
!
        mat(:n1,:nfft) = matw(:n1,:nfft)
!
    end if
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( matw, tmp )
#endif
!
!
! END OF SUBROUTINE fft2_row
! __________________________
!
    end subroutine fft2_row
!
! =========================================================================================
!
    subroutine fft2_row_par( mat, w, lognfft )
!
! Purpose
! _______
!
!   Subroutine FFT2_ROW_PAR replaces each row of MAT by its Fourier transform.
!   size(MAT,2) is constrained to be an exact power of two.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) complex(stnd), dimension(:,:)
!            On entry, the complex valued sequences to be transformed.
!            On exit, each row of MAT is replaced by its Fourier transform.
!
!   W        (INPUT) complex(stnd), dimension(:)
!            On entry, the powers of ww = e**((-PI)/size(W)).
!
!                         W(i) = ww**(i-1) .
!
!   LOGNFFT  (INPUT) integer(i4b)
!            On entry, LOGNFFT must verify size(MAT,2) = 2**(LOGNFFT).
!
!
! Further Details
! _______________
!
!   This private subroutine is used by subroutines INIT_FFT and FFT. This is a 1D
!   complex-complex radix-2 decimation-in-time Cooley-Tukey algorithm.
!
!   This is the parallelized version of FFT2_ROW subroutine (Parallelization is done with
!   OPENMP directives).
!
!   MAT and W must verify ize(MAT,2) = 2 * size(W).
!
!   For more details, see:
!
!   (1) Cooley, J.W., Lewis, P., and Welch, P., 1969: The Fast Fourier Transform
!           and its Applications", IEEE Trans on Education, 12, 1, 28-34.
!      
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _ALLOC
    use Char_Constants,    only : allocate_error
    use Utilities,         only : merror
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    complex(stnd), intent(inout), dimension(:,:)           :: mat
    complex(stnd), intent(in),    dimension(size(mat,2)/2) :: w
!
    integer(i4b), intent(in) :: lognfft
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _ALLOC
    complex(stnd), dimension(:,:), allocatable :: mat2
#else
    complex(stnd), dimension(size(mat,2),size(mat,1)) :: mat2
#endif
!
    integer(i4b) :: n1, nfft, i
#ifdef _ALLOC
    integer :: iok
#endif
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
#ifdef _ALLOC
    character(len=*),  parameter :: name_proc='fft2_row_par'
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n1        = size( mat, 1 )
    nfft      = size( mat, 2 )
#ifdef _ALLOC
!
!   ALLOCATE WORK VARIABLE FOR THE Cooley-Tukey FFT.
!
    allocate( mat2(nfft,n1), stat=iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
#endif
!
    mat2(:nfft,:n1) = transpose( mat(:n1,:nfft) )
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nfft*n1>=omp_limit                  .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i), SHARED(nfft,lognfft)
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, n1
!
        call fft2( mat2(:nfft,i), w(:), lognfft )
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
    mat(:n1,:nfft) = transpose( mat2(:nfft,:n1) )
!
#ifdef _ALLOC
!
!   DEALLOCATE WORK VARIABLE.
!
    deallocate( mat2 )
#endif
!
!
! END OF SUBROUTINE fft2_row_par
! ______________________________
!
    end subroutine fft2_row_par
!
!
! =========================================================================================
!
!
! ******************************
! END OF MODULE FFT_Procedures *
! ******************************
!    
end module FFT_Procedures
