module Prob_Procedures
! ####################################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR PROBABILITY DISTRIBUTION            *
! FUNCTIONS, INVERSES, AND OTHER PARAMETERS.                                         *
!                                                                                    *
! LATEST REVISION : 01/10/2018                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!                                                                                    *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                                    *
!                                                                                    *
!  _OPENMP  FOR ACTIVATING OPENMP PARALLELIZATION                                    *
!                                                                                    *
! ************************************************************************************
!
!
! USED MODULES
! ============
!
    use Select_Parameters,  only : lgl, i4b, stnd, extd
#ifdef _OPENMP
    use Select_Parameters,  only : omp_limit, omp_limit2
    use omp_lib,            only : omp_get_num_procs, omp_get_max_threads,    &
                                   omp_in_parallel, omp_get_nested
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!
!   use Logical_Constants
!   use Reals_Constants
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   lngamma,                                                      &
                probgamma, probgamma2, probgamma3, pinvgamma, pinvgamma2,     &
                probbeta, pinvbeta,                                           &
                probn, probn2, pinvn, pinvn2,                                 &
                probt, probstudent, pinvt, pinvstudent,                       &
                probq, probq2, probq3, pinvq, pinvq2, pinvq3,                 &
                probf, probf2, pinvf2,                                        &
                probbinom, rangen
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface lngamma
        module procedure    lngamma_r, lngamma_rv, lngamma_rm
    end interface
!
    interface probgamma
        module procedure    probgamma_r, probgamma_rv, probgamma_rm, probgamma_rv2, probgamma_rm2
    end interface
!
    interface probgamma2
        module procedure    probgamma2_r, probgamma2_rv, probgamma2_rm, probgamma2_rv2, probgamma2_rm2
    end interface
!
    interface probgamma3
        module procedure    probgamma3_r, probgamma3_rv, probgamma3_rm, probgamma3_rv2, probgamma3_rm2
    end interface
!
    interface pinvgamma
        module procedure    pinvgamma_r
    end interface
!
    interface pinvgamma2
        module procedure    pinvgamma_r
    end interface
!
    interface probbeta
        module procedure    probbeta_r, probbeta_rv, probbeta_rm
    end interface
!
    interface pinvbeta
        module procedure    pinvbeta_r
    end interface
!
    interface probn
        module procedure    probn_r, probn_rv, probn_rm
    end interface
!
    interface probn2
        module procedure    probn2_r, probn2_rv, probn2_rm
    end interface
!
    interface pinvn
        module procedure    pinvn_r, pinvn_rv, pinvn_rm
    end interface
!
    interface pinvn2
        module procedure    pinvn2_r, pinvn2_rv, pinvn2_rm
    end interface
!
    interface probt
        module procedure    probt_r, probt_rv, probt_rm, probt_rv2, probt_rm2
    end interface
!
    interface probstudent
        module procedure    probstudent_r, probstudent_rv, probstudent_rm, probstudent_rv2, probstudent_rm2
    end interface
!
    interface pinvt
        module procedure    pinvt_r, pinvt_rv, pinvt_rm, pinvt_rv2, pinvt_rm2
    end interface
!
    interface pinvstudent
        module procedure    pinvstudent_r, pinvstudent_rv, pinvstudent_rm, pinvstudent_rv2, pinvstudent_rm2
    end interface
!
    interface probq
        module procedure    probq_r, probq_rv, probq_rm, probq_rv2, probq_rm2
    end interface
!
    interface probq2
        module procedure    probq2_r, probq2_rv, probq2_rm, probq2_rv2, probq2_rm2
    end interface
!
    interface probq3
        module procedure    probq3_r, probq3_rv, probq3_rm, probq3_rv2, probq3_rm2
    end interface
!
    interface pinvq
        module procedure    pinvq_r, pinvq_rv, pinvq_rm, pinvq_rv2, pinvq_rm2
    end interface
!
    interface pinvq2
        module procedure    pinvq2_r, pinvq2_rv, pinvq2_rm, pinvq2_rv2, pinvq2_rm2
    end interface
!
    interface pinvq3
        module procedure    pinvq2_r, pinvq2_rv, pinvq2_rm, pinvq2_rv2, pinvq2_rm2
    end interface
!
    interface probf
        module procedure    probf_r, probf_rv, probf_rm, probf_rv2, probf_rm2
    end interface
!
    interface probf2
        module procedure    probf2_r, probf2_rv, probf2_rm, probf2_rv2, probf2_rm2
    end interface
!
    interface pinvf2
        module procedure    pinvf2_r
    end interface
!
    interface probbinom
        module procedure    probbinom_r
    end interface
!
    interface rangen
        module procedure    rangen_r, rangen_rv
    end interface
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                            LOGARITHM OF THE GAMMA FUNCTION
! =========================================================================================
!
!
    function lngamma_r( x )
!
! Purpose
! _______
!
!   Evaluates the logarithm of the gamma function: ln( gamma(X) )
!   for a strictly positive real argument X.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd)
!          On entry, a strictly positive real argument X.
!
!
! Further Details
! _______________
!
!   This function uses a Lanczos-type approximation to ln(gamma(X)) for X > 0. Its accuracy
!   is about 14 significant digits except for small regions in the vicinity of 1 and 2.
!
!   This function is adapted from
!
!   (1)    Lanczos, C., 1964: A precision approximation of the gamma function,
!                              J. SIAM Numer. Anal., B, 1, 86-96.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, zero_extd, one_extd, half_extd,       &
                                  seven_extd, c6_5_extd, lnsqrt2pi_extd
    use Char_Constants,    only : prob_error6
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: x
    real(stnd)             :: lngamma_r
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)    :: lanczos, tmp
!
    integer(i4b)  :: j
!   
!   
! PARAMETERS 
! __________
!
    real(extd), dimension(9), parameter ::     a=                                    &
                           (/ 0.9999999999995183_extd, 676.5203681218835_extd,       &
                              -1259.139216722289_extd, 771.3234287757674_extd,       &
                              -176.6150291498386_extd, 12.50734324009056_extd,       &
                              -0.1385710331296526_extd, 0.9934937113930748E-05_extd, &
                               0.1659470187408462E-06_extd /)
!
    character(len=*),  parameter :: name_proc='lngamma'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENT.
!
    if ( x<=zero )  then
        call merror( name_proc//prob_error6 )
    end if
!
    lanczos = zero_extd
    tmp     = x + seven_extd
!
    do j = 9_i4b, 2_i4b, -1_i4b
!
        lanczos = lanczos + a(j)/tmp
        tmp     = tmp - one_extd
!
    end do
!
    lanczos   = lanczos + a(1_i4b)
    lngamma_r = log( lanczos ) + lnsqrt2pi_extd - (x + c6_5_extd) &
                + (x - half_extd)*log( x + c6_5_extd )
!
!
! END OF FUNCTION lngamma_r
! _________________________
!
    end function lngamma_r
!
! =========================================================================================
!
    function lngamma_rv( x )
!
! Purpose
! _______
!
!   Evaluates the logarithm of the gamma function: ln( gamma(X(:)) )
!   for a strictly positive real vector argument X(:).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, a strictly positive real vector argument X.
!
!
! Further Details
! _______________
!
!   This function uses a Lanczos-type approximation to ln(gamma(X)) for X > 0. Its accuracy
!   is about 14 significant digits except for small regions in the vicinity of 1 and 2.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)    Lanczos, C., 1964: A precision approximation of the gamma function,
!                             J. SIAM Numer. Anal., B, 1, 86-96.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, zero_extd, one_extd, half_extd,       &
                                  seven_extd, c6_5_extd, lnsqrt2pi_extd
    use Char_Constants,    only : prob_error6
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:)        :: x
    real(stnd),             dimension(size(x))  :: lngamma_rv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)                     :: lanczosi, tmp
    real(extd), dimension(size(x)) :: lanczos
!
    integer(i4b)  :: i, j, n
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(extd), dimension(9), parameter ::            a=                             &
                           (/ 0.9999999999995183_extd, 676.5203681218835_extd,       &
                              -1259.139216722289_extd, 771.3234287757674_extd,       &
                              -176.6150291498386_extd, 12.50734324009056_extd,       &
                              -0.1385710331296526_extd, 0.9934937113930748E-05_extd, &
                               0.1659470187408462E-06_extd /)
!
    character(len=*),  parameter :: name_proc='lngamma'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENT.
!
    if ( any( x(:n)<=zero ) )  then
        call merror( name_proc//prob_error6 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,tmp,lanczosi), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        lanczosi = zero_extd
        tmp      = x(i) + seven_extd
!
        do j = 9_i4b, 2_i4b, -1_i4b
!
            lanczosi = lanczosi + a(j)/tmp
            tmp      = tmp - one_extd
!
        end do
!
        lanczosi   = lanczosi + a(1_i4b)
        lanczos(i) = log( lanczosi ) + lnsqrt2pi_extd - ( x(i) + c6_5_extd ) &
                     + ( x(i) - half_extd )*log( x(i) + c6_5_extd )
!
    end do
!
!$OMP END PARALLEL DO
!        
    lngamma_rv(:n) = lanczos(:n)
!
! END OF FUNCTION lngamma_rv
! __________________________
!
    end function lngamma_rv
!
! =========================================================================================
!
    function lngamma_rm( x )
!
! Purpose
! _______
!
!   Evaluates the logarithm of the gamma function: ln( gamma(X(:,:)) )
!   for a strictly positive real matrix argument X(:,:).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:)
!          On entry, a strictly positive real matrix argument X.
!
!
! Further Details
! _______________
!
!   This function uses a Lanczos-type approximation to ln(gamma(X)) for X > 0. Its accuracy
!   is about 14 significant digits except for small regions in the vicinity of 1 and 2.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)    Lanczos, C., 1964: A precision approximation of the gamma function,
!                             J. SIAM Numer. Anal., B, 1, 86-96.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, zero_extd, one_extd, half_extd,       &
                                  seven_extd, c6_5_extd, lnsqrt2pi_extd
    use Char_Constants,    only : prob_error6
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:)                 :: x
    real(stnd),             dimension(size(x,1),size(x,2)) :: lngamma_rm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)                                 :: lanczosik, tmp
    real(extd), dimension(size(x,1),size(x,2)) :: lanczos
!
    integer(i4b)  :: i, j, k, n, p
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(extd), dimension(9), parameter ::               a=                          &
                           (/ 0.9999999999995183_extd, 676.5203681218835_extd,       &
                              -1259.139216722289_extd, 771.3234287757674_extd,       &
                              -176.6150291498386_extd, 12.50734324009056_extd,       &
                              -0.1385710331296526_extd, 0.9934937113930748E-05_extd, &
                               0.1659470187408462E-06_extd /)
!
    character(len=*),  parameter :: name_proc='lngamma'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 1 )
    p = size( x, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. p<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENT.
!
    if ( any( x(:n,:p)<=zero ) )  then
        call merror( name_proc//prob_error6 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*p>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,k,tmp,lanczosik), SCHEDULE(STATIC)
!
    do k=1_i4b, p
!
        do i=1_i4b, n
!
            lanczosik = zero_extd
            tmp       = x(i,k) + seven_extd
!
            do j = 9_i4b, 2_i4b, -1_i4b
!
                lanczosik = lanczosik + a(j)/tmp
                tmp      = tmp - one_extd
!
            end do
!
            lanczosik    = lanczosik + a(1_i4b)
            lanczos(i,k) = log( lanczosik ) + lnsqrt2pi_extd - ( x(i,k) + c6_5_extd ) &
                           + ( x(i,k) - half_extd )*log( x(i,k) + c6_5_extd )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!        
    lngamma_rm(:n,:p) = lanczos(:n,:p)
!
! END OF FUNCTION lngamma_rm
! __________________________
!
    end function lngamma_rm
!
!
! =========================================================================================
!                            GAMMA PROBABILITY DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function probgamma_r( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real argument X and a strictly positive value GAMP of the parameter p
!   of the Gamma distribution.
!   
!   PROBGAMMA computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X.
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd)
!           On entry, a positive real argument X which is the value of the upper integral limit.
!           X must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   Otherwise, a Pearson's series expansion is used for evaluating the integral, see reference (3),
!   Formula 6.5.29, p.262. The "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The default value
!   for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU, and also
!   varies slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but it may be slower.
!
!   This function is adapted from
!
!   (1)      Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                            Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)      Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                             Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)      Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, three, nine
    use Logical_Constants, only : false
    use Num_Constants,     only : machtiny, macheps
    use Char_Constants,    only : prob_error12, prob_error13
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: x, gamp
    real(stnd), intent(in), optional :: acu
    real(stnd)                       :: probgamma_r
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)    :: c, a, gam, acu2, elimit
!
    integer(i4b)  :: i, maxit
!
    logical(lgl) :: fail
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: xbig = 1.e+8_stnd,  plimit = 1000._stnd 
                             
!
    character(len=*),  parameter :: name_proc='probgamma'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( x<zero ) then
        call merror( name_proc//prob_error12 )
    end if
!
    if ( gamp<=zero ) then
        call merror( name_proc//prob_error13 )
    end if
!
    acu2 = macheps
    if ( present(acu) ) then
        acu2 = max( acu, macheps )
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( x==zero ) then
!
        probgamma_r = zero
!
        return
!
    else if ( gamp>plimit ) then
!
!       USE A NORMAL APPROXIMATION IF gamp > plimit.
!
        c = three*sqrt(gamp)*( (x/gamp)**( one/three) + one/(nine*gamp) - one )
        probgamma_r =  probn_r( c, false )
!
        return
!
    else if ( x>xbig ) then
!
!      IF x IS EXTREMELY LARGE COMPARED TO gamp THEN SET probgamma_r = 1
!
        probgamma_r =  one
!
        return
!
    end if
!
!   DETERMINE MACHINE DEPENDANT CONSTANT.
!
    elimit = log( machtiny )
!
!   USE PEARSON'S SERIES EXPANSION,
!   CALCULATE USING EQUATION NO. 6.5.29.
!  
    c   = one
    gam = one
    a   = gamp
!  
    do i = 1_i4b, maxit
!
        a   = a + one
        c   = c*x/a
        gam = gam + c
!
        if ( c<=min( acu2, acu2*gam) ) exit
!
    end do
!
    if ( fail .and. i>maxit ) then
!  
        gam = -one
!  
    else
!  
        c = gamp*log(x) - x - lngamma_r( gamp+one )
        c = c + log( gam )
!
!       TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!  
        c = max( c, elimit )
        gam = exp( c )
!      
    end if
!      
    probgamma_r = gam
!
!
! END OF FUNCTION probgamma_r
! ___________________________
!
    end function probgamma_r
!
! =========================================================================================
!
    function probgamma_rv( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real vector argument X and a strictly positive value GAMP of the parameter P
!   of the Gamma distribution.
!   
!   PROBGAMMA computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X(i) for i=1 to size(X).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:)
!           On entry, a positive real vector argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   Otherwise, a Pearson's series expansion is used for evaluating the integral, see reference (3),
!   Formula 6.5.29, p.262. The "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The default value
!   for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU, and also varies
!   slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but it may be slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)     Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                           Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)     Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                            Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)     Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                        :: gamp
    real(stnd), intent(in), dimension(:)          :: x
    real(stnd), intent(in),              optional :: acu
    real(stnd), dimension(size(x))                :: probgamma_rv
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: xi
    real(stnd), dimension(size(x)) :: prob
!
    integer(i4b)  :: n, i
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,xi), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        xi = x(i)
!
        prob(i) = probgamma_r( xi, gamp, acu=acu, maxiter=maxiter, failure=failure )
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma_rv(:n) = prob(:n)
!
!
! END OF FUNCTION probgamma_rv
! ____________________________
!
    end function probgamma_rv
!
! =========================================================================================
!
    function probgamma_rv2( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real vector argument X and a strictly positive vector argument GAMP of the
!   parameter P of the Gamma distribution.
!   
!   PROBGAMMA computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP(i) will be less than or equal to X(i) for i=1 to size(X).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:)
!           On entry, a positive real vector argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd), dimension(:)
!           On entry, a strictly positive real vector argument which are the values of the parameter p
!           of the Gamma distribution.
!           Elements in GAMP(:) must be greater than zero.
!
!           The size of GAMP must verify  size(GAMP) = size(X)  .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP(i) (e.g. GAMP(i)>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   Otherwise, a Pearson's series expansion is used for evaluating the integral, see reference (3),
!   Formula 6.5.29, p.262. The "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The default value
!   for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU, and also varies
!   slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but it may be slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)     Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                           Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)     Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                            Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)     Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:)          :: x, gamp
    real(stnd), intent(in),              optional :: acu
    real(stnd), dimension(size(x))                :: probgamma_rv2
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: xi, gampi
    real(stnd), dimension(size(x)) :: prob
!
    integer(i4b)  :: n, i
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probgamma'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(x),i4b)     ,   &
                    int(size(gamp),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,xi,gampi), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        xi = x(i)
!
        gampi = gamp(i)
!
        prob(i) = probgamma_r( xi, gampi, acu=acu, maxiter=maxiter, failure=failure )
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma_rv2(:n) = prob(:n)
!
!
! END OF FUNCTION probgamma_rv2
! _____________________________
!
    end function probgamma_rv2
!
! =========================================================================================
!
    function probgamma_rm( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real matrix argument X and a strictly positive value GAMP of the parameter P
!   of the Gamma distribution.
!   
!   PROBGAMMA computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X(i,j) for i=1 to size(X,1) and j=1 
!   to size(X,2).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:,:)
!           On entry, a positive real matrix argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:,:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   Otherwise, a Pearson's series expansion is used for evaluating the integral, see reference (3),
!   Formula 6.5.29, p.262. The "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The default value
!   for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU, and also varies
!   slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but it may be slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)     Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                           Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)     Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                            Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)     Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                (formulae 6.5.29 and 26.4.14),
!                                New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                          :: gamp
    real(stnd), intent(in), dimension(:,:)          :: x
    real(stnd), intent(in),                optional :: acu
    real(stnd), dimension(size(x,1),size(x,2))      :: probgamma_rm
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: xij
    real(stnd), dimension(size(x,1),size(x,2)) :: prob
!
    integer(i4b)  :: n, m, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 1 )
    m = size( x, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,xij), SCHEDULE(STATIC)
!
    do j=1_i4b, m
!
        do i=1_i4b, n
!
            xij = x(i,j)
!
            prob(i,j) = probgamma_r( xij, gamp, acu=acu, maxiter=maxiter, failure=failure )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma_rm(:n,:m) = prob(:n,:m)
!
!
! END OF FUNCTION probgamma_rm
! ____________________________
!
    end function probgamma_rm
!
! =========================================================================================
!
    function probgamma_rm2( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real matrix argument X and a strictly positive matrix argument GAMP of the
!   parameter P of the Gamma distribution.
!   
!   PROBGAMMA computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP(i,j) will be less than or equal to X(i,j) for i=1 to size(X,1) and j=1 
!   to size(X,2).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:,:)
!           On entry, a positive real matrix argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd), dimension(:,:)
!           On entry, a strictly positive real matrix argument which are the values of the parameter p
!           of the Gamma distribution.
!           Elements in GAMP(:,:) must be greater than zero.
!
!           The shape of GAMP must verify:
!
!           - size(GAMP,1) = size(X,1)
!           - size(GAMP,2) = size(X,2).
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP(i,j) (e.g. GAMP(i,j)>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   Otherwise, a Pearson's series expansion is used for evaluating the integral, see reference (3),
!   Formula 6.5.29, p.262. The "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The default value
!   for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU, and also varies
!   slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but it may be slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)     Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                           Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)     Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                            Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)     Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                (formulae 6.5.29 and 26.4.14),
!                                New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:)          :: x, gamp
    real(stnd), intent(in),                optional :: acu
    real(stnd), dimension(size(x,1),size(x,2))      :: probgamma_rm2
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: xij, gampij
    real(stnd), dimension(size(x,1),size(x,2)) :: prob
!
    integer(i4b)  :: n, m, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probgamma'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(x,1),i4b)     ,   &
                    int(size(gamp,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(x,2),i4b)     ,   &
                    int(size(gamp,2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,xij,gampij), SCHEDULE(STATIC)
!
    do j=1_i4b, m
!
        do i=1_i4b, n
!
            xij = x(i,j)
!
            gampij = gamp(i,j)
!
            prob(i,j) = probgamma_r( xij, gampij, acu=acu, maxiter=maxiter, failure=failure )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma_rm2(:n,:m) = prob(:n,:m)
!
!
! END OF FUNCTION probgamma_rm2
! _____________________________
!
    end function probgamma_rm2
!
! =========================================================================================
!
    function probgamma2_r( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real argument X and a strictly positive value GAMP of the parameter p of
!   the Gamma distribution.
!   
!   PROBGAMMA2 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X.
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd)
!           On entry, a positive real argument X which is the value of the upper integral limit.
!           X must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (1), Formula 26.4.14, p.941.
!
!   For X<=1 ou X<GAMP, a Pearson's series expansion is used, see reference (1),
!   Formula 6.5.29, p.262. For other values of X, a continued fraction expansion is
!   used since this expansion tends to converge more quickly than Pearson's series
!   expansion, see reference (1), Formula 6.5.31, p.263. In both cases, the "integrating"
!   process is terminated when both the absolute and relative contributions to the
!   integral is not greater than the value of ACU. The default value for ACU gives
!   the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but is slower.
!
!   This function is adapted from
!
!   (1)  Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!   (2)  Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, nine
    use Logical_Constants, only : false
    use Num_Constants,     only : machtiny, machhuge, macheps
    use Char_Constants,    only : prob_error12, prob_error13
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: x, gamp
    real(stnd), intent(in), optional :: acu
    real(stnd)                       :: probgamma2_r
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)    :: pn1, pn2, pn3, pn4, pn5, pn6, arg, c, rn, a, b,  &
                     an, gam, acu2, oflo, elimit
!
    integer(i4b)  :: i, maxit
!
    logical(lgl) :: fail
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: xbig = 1.e+8_stnd,  plimit = 1000._stnd 
                             
!
    character(len=*),  parameter :: name_proc='probgamma2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( x<zero )  then
        call merror( name_proc//prob_error12 )
    end if
!
    if ( gamp<=zero )  then
        call merror( name_proc//prob_error13 )
    end if
!
!   DETERMINE THE REQUIRED ACCURACY OF THE RESULT.
!
    acu2 = macheps
    if ( present(acu) ) then
        acu2 = max( acu, macheps )
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( x==zero ) then
!
        probgamma2_r = zero
!
        return
!
    else if ( gamp>plimit ) then
!
!       USE A NORMAL APPROXIMATION IF gamp > plimit.
!
        pn1 = three*sqrt(gamp)*( (x/gamp)**( one/three) + one/(nine*gamp) - one )
        probgamma2_r =  probn_r( pn1, false )
!
        return
!
    else if ( x>xbig ) then
!
!      IF x IS EXTREMELY LARGE COMPARED TO gamp THEN SET probgamma2_r = 1
!
        probgamma2_r =  one
!
        return
!
    end if
!
!   DETERMINE MACHINE DEPENDANT CONSTANTS.
!
    elimit = log( machtiny )
    oflo   = sqrt( machhuge )
!
    if ( x<=one .or. x<gamp ) then
!
!       USE PEARSON'S SERIES EXPANSION,
!       CALCULATE USING EQUATION NO. 6.5.29.
!  
        c   = one
        gam = one
        a   = gamp
!
        do i = 1_i4b, maxit
!
            a   = a + one
            c   = c*x/a
            gam = gam + c
!
            if ( c<=min( acu2, acu2*gam) ) exit
!
        end do
!
        if ( fail .and. i>maxit ) then
!  
            gam = -one
!  
        else
!  
            arg = gamp*log(x) - x - lngamma_r( gamp+one )
            arg = arg + log( gam )
!
!           TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
            arg = max( arg, elimit )
            gam = exp( arg )
!    
        end if
!    
   else
!  
!      USE A CONTINUED FRACTION EXPANSION,
!      CALCULATE USING EQUATION NO. 6.5.31.
!  
       a   = one - gamp
       b   = a + x + one
       c   = zero
       pn1 = one
       pn2 = x
       pn3 = x + one
       pn4 = x*b
       gam = pn3/pn4
!  
       do i = 1_i4b, maxit
!  
           a = a + one
           b = b + two
           c = c + one
           an = a*c
           pn5 = b*pn3 - an*pn1
           pn6 = b*pn4 - an*pn2
 !  
           if ( abs(pn6)>zero ) then
!
               rn = pn5/pn6
!
               if ( abs(gam - rn)<=min( acu2, acu2*rn) ) exit
!
               gam = rn
!
           end if
!  
           pn1 = pn3
           pn2 = pn4
           pn3 = pn5
           pn4 = pn6
!      
           if ( abs(pn5)>=oflo ) then
!      
!              RE-SCALE TERMS IN CONTINUED FRACTION IF TERMS ARE LARGE.
!      
               pn1 = pn1/oflo
               pn2 = pn2/oflo
               pn3 = pn3/oflo
               pn4 = pn4/oflo
!      
           end if
!      
       end do
!
       if ( fail .and. i>maxit ) then
!  
           gam = -one
!  
       else
!      
           arg = gamp*log(x) - x - lngamma_r( gamp )
           arg = arg + log( gam )
!
!          TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
            arg = max( arg, elimit )
            gam = one - exp( arg )
!
       end if
!
   end if
!      
   probgamma2_r = gam
!
!
! END OF FUNCTION probgamma2_r
! ____________________________
!
    end function probgamma2_r
!
! =========================================================================================
!
    function probgamma2_rv( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real vector argument X and a strictly positive value GAMP of the parameter p
!   of the Gamma distribution.
!   
!   PROBGAMMA2 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X(i) for i=1 to size(X).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:)
!           On entry, a positive real vector argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (1), Formula 26.4.14, p.941.
!
!   For X<=1 ou X<GAMP, a Pearson's series expansion is used, see reference (1),
!   Formula 6.5.29, p.262. For other values of X, a continued fraction expansion is
!   used since this expansion tends to converge more quickly than Pearson's series
!   expansion, see reference (1), Formula 6.5.31, p.263. In both cases, the "integrating"
!   process is terminated when both the absolute and relative contributions to the
!   integral is not greater than the value of ACU. The default value for ACU gives
!   the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but is slower.
!
!   This function is adapted from
!
!   (1)  Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!   (2)  Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                        :: gamp
    real(stnd), intent(in), dimension(:)          :: x
    real(stnd), intent(in),              optional :: acu
    real(stnd), dimension(size(x))                :: probgamma2_rv
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: xi
    real(stnd), dimension(size(x)) :: prob
!
    integer(i4b)  :: n, i
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,xi), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        xi = x(i)
!
        prob(i) = probgamma2_r( xi, gamp, acu=acu, maxiter=maxiter, failure=failure )
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma2_rv(:n) = prob(:n)
!
!
! END OF FUNCTION probgamma2_rv
! _____________________________
!
    end function probgamma2_rv
!
! =========================================================================================
!
    function probgamma2_rv2( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real vector argument X and a strictly positive vector argument GAMP of the
!   parameter P of the Gamma distribution.
!   
!   PROBGAMMA2 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP(i) will be less than or equal to X(i) for i=1 to size(X).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:)
!           On entry, a positive real vector argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd), dimension(:)
!           On entry, a strictly positive real vector argument which are the values of the parameter p
!           of the Gamma distribution.
!           Elements in GAMP(:) must be greater than zero.
!
!           The size of GAMP must verify  size(GAMP) = size(X)  .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP(i) (e.g. GAMP(i)>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (1), Formula 26.4.14, p.941.
!
!   For X(i)<=1 ou X(i)<GAMP(i), a Pearson's series expansion is used, see reference (1),
!   Formula 6.5.29, p.262. For other values of X(i), a continued fraction expansion is
!   used since this expansion tends to converge more quickly than Pearson's series
!   expansion, see reference (1), Formula 6.5.31, p.263. In both cases, the "integrating"
!   process is terminated when both the absolute and relative contributions to the
!   integral is not greater than the value of ACU. The default value for ACU gives
!   the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but is slower.
!
!   This function is adapted from
!
!   (1)  Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!   (2)  Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:)          :: x, gamp
    real(stnd), intent(in),              optional :: acu
    real(stnd), dimension(size(x))                :: probgamma2_rv2
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: xi, gampi
    real(stnd), dimension(size(x)) :: prob
!
    integer(i4b)  :: n, i
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probgamma2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(x),i4b)     ,   &
                    int(size(gamp),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,xi,gampi), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        xi = x(i)
!
        gampi = gamp(i)
!
        prob(i) = probgamma2_r( xi, gampi, acu=acu, maxiter=maxiter, failure=failure )
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma2_rv2(:n) = prob(:n)
!
!
! END OF FUNCTION probgamma2_rv2
! ______________________________
!
    end function probgamma2_rv2
!
! =========================================================================================
!
    function probgamma2_rm( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real matrix argument X and a strictly positive value GAMP of the parameter p
!   of the Gamma distribution.
!   
!   PROBGAMMA2 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X(i,j) for i=1 to size(X,1) and j=1 
!   to size(X,2).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:,:)
!           On entry, a positive real matrix argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:,:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or  continued fraction expansion of the
!           incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (1), Formula 26.4.14, p.941.
!
!   For X<=1 ou X<GAMP, a Pearson's series expansion is used, see reference (1),
!   Formula 6.5.29, p.262. For other values of X, a continued fraction expansion is
!   used since this expansion tends to converge more quickly than Pearson's series
!   expansion, see reference (1), Formula 6.5.31, p.263. In both cases, the "integrating"
!   process is terminated when both the absolute and relative contributions to the
!   integral is not greater than the value of ACU. The default value for ACU gives
!   the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but is slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)  Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!   (2)  Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: gamp
    real(stnd), intent(in), dimension(:,:)              :: x
    real(stnd), intent(in),                    optional :: acu
    real(stnd), dimension(size(x,1),size(x,2))          :: probgamma2_rm
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: xij
    real(stnd), dimension(size(x,1),size(x,2)) :: prob
!
    integer(i4b)  :: n, m, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 1 )
    m = size( x, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,xij), SCHEDULE(STATIC)
!
    do j=1_i4b, m
!
        do i=1_i4b, n
!
            xij = x(i,j)
!
            prob(i,j) = probgamma2_r( xij, gamp, acu=acu, maxiter=maxiter, failure=failure )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma2_rm(:n,:m) = prob(:n,:m)
!
!
! END OF FUNCTION probgamma2_rm
! _____________________________
!
    end function probgamma2_rm
!
! =========================================================================================
!
    function probgamma2_rm2( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real matrix argument X and a strictly positive matrix argument GAMP of the
!   parameter P of the Gamma distribution.
!   
!   PROBGAMMA2 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP(i,j) will be less than or equal to X(i,j) for i=1 to size(X,1) and j=1 
!   to size(X,2).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:,:)
!           On entry, a positive real matrix argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:,:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd), dimension(:,:)
!           On entry, a strictly positive real matrix argument which are the values of the parameter p
!           of the Gamma distribution.
!           Elements in GAMP(:,:) must be greater than zero.
!
!           The shape of GAMP must verify:
!
!           - size(GAMP,1) = size(X,1)
!           - size(GAMP,2) = size(X,2).
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or  continued fraction expansion of the
!           incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP(i,j) (e.g. GAMP(i,j)>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (1), Formula 26.4.14, p.941.
!
!   For X(i,j)<=1 ou X(i,j)<GAMP(i,j), a Pearson's series expansion is used, see reference (1),
!   Formula 6.5.29, p.262. For other values of X(i,j), a continued fraction expansion is
!   used since this expansion tends to converge more quickly than Pearson's series
!   expansion, see reference (1), Formula 6.5.31, p.263. In both cases, the "integrating"
!   process is terminated when both the absolute and relative contributions to the
!   integral is not greater than the value of ACU. The default value for ACU gives
!   the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   This function is more accurate than PROBGAMMA3, but is slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)  Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!   (2)  Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:)              :: x, gamp
    real(stnd), intent(in),                    optional :: acu
    real(stnd), dimension(size(x,1),size(x,2))          :: probgamma2_rm2
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: xij, gampij
    real(stnd), dimension(size(x,1),size(x,2)) :: prob
!
    integer(i4b)  :: n, m, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probgamma2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n =  assert_eq( int(size(x,1),i4b)     ,   &
                    int(size(gamp,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(x,2),i4b)     ,   &
                    int(size(gamp,2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,xij,gampij), SCHEDULE(STATIC)
!
    do j=1_i4b, m
!
        do i=1_i4b, n
!
            xij = x(i,j)
!
            gampij = gamp(i,j)
!
            prob(i,j) = probgamma2_r( xij, gampij, acu=acu, maxiter=maxiter, failure=failure )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma2_rm2(:n,:m) = prob(:n,:m)
!
!
! END OF FUNCTION probgamma2_rm2
! ______________________________
!
    end function probgamma2_rm2
!
! =========================================================================================
!
    function probgamma3_r( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real argument X and a strictly positive value GAMP of the parameter p
!   of the Gamma distribution.
!   
!   PROBGAMMA3 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X.
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd)
!           On entry, a positive real argument X which is the value of the upper integral limit.
!           X must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   For X<=max(GAMP/2,13), a Pearson's series expansion is used, see reference (3),
!   Formula 6.5.29, p.262. For larger values of X, an alternate Pearson's asymptotic
!   series expansion is used since this expansion tends to converge more quickly,
!   see reference (2), Formula 6.5.32, p.263.
!
!   In both cases, the "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The
!   default value for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   PROBGAMMA3 is faster, but less accurate than PROBGAMMA or PROBGAMMA2, since
!   for large values of X, the alternate Pearson's series expansion is only asymptotic.
!
!   This function is adapted from
!
!   (1)   Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)   Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)   Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                 Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one, three, nine, c13
    use Logical_Constants, only : false
    use Num_Constants,     only : machtiny, macheps
    use Char_Constants,    only : prob_error12, prob_error13
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: x, gamp
    real(stnd), intent(in), optional :: acu
    real(stnd)                       :: probgamma3_r
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)    :: c, b, gam, acu2, elimit, cst, x1, lngamp
!
    integer(i4b) :: i, maxit
!
    logical(lgl) :: fail
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: xbig = 1.e+8_stnd,  plimit = 1000._stnd 
                             
!
    character(len=*),  parameter :: name_proc='probgamma3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( x<zero )  then
        call merror( name_proc//prob_error12 )
    end if
!
    if ( gamp<=zero ) then
        call merror( name_proc//prob_error13 )
    end if
!
    acu2 = macheps
    if ( present(acu) ) then
        acu2 = max( acu, macheps )
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( x==zero ) then
!
        probgamma3_r = zero
!
        return
!
    else if ( gamp>plimit ) then
!
!       USE A NORMAL APPROXIMATION IF gamp > plimit.
!
        c = three*sqrt(gamp)*( (x/gamp)**( one/three) + one/(nine*gamp) - one )
        probgamma3_r =  probn_r( c, false )
!
        return
!
    else if ( x>xbig ) then
!
!      IF x IS EXTREMELY LARGE COMPARED TO gamp THEN SET probgamma3_r = 1
!
        probgamma3_r =  one
!
        return
!
    end if
!
    lngamp = lngamma_r( gamp )
    cst    = max( half*gamp, c13 )
!
!   DETERMINE MACHINE DEPENDANT CONSTANT.
!
    elimit = log( machtiny )
!
    if ( x<cst ) then
!
!       USE PEARSON'S SERIES EXPANSION.
!       CALCULATE USING EQUATION NO. 6.5.29
!
        gam  = one/gamp
        c    = gam
        b    = gamp
!
        do i = 1_i4b, maxit
!
            b = b + one
            c = c*x/b
!
            if ( c<=min( acu2, acu2*gam) ) exit
!
            gam = gam + c
!
        end do
!
        if ( fail .and. i>maxit ) then
!  
            gam = -one
!  
        else
!
            c = gamp*log( x ) - x - lngamp          
            c = c + log( gam )
!
!           TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!      
            c = max( c, elimit )
            gam = exp( c )
!      
        end if
!
    else
!
!       USE ALTERNATE PEARSON'S SERIES EXPANSION.
!       CALCULATE USING EQUATION NO. 6.5.32
!
        x1   = one/x
        b    = gamp - one
        c    = b*x1
        gam  = c + one
!
        do i = 2_i4b, maxit
!
            b = b - one
            c = c*b*x1
!
            if ( c<=min( acu2, acu2*gam) ) exit
!            if ( c<=acu2*gam ) exit
!
            gam = gam + c
!
        end do
!
        if ( fail .and. i>maxit ) then
!  
            gam = -one
!  
        else
!
            c = (gamp-one)*log(x) - x - lngamp         
            c = c + log( gam )
!
!           TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!      
            c = max( c, elimit )
            gam = one - exp( c )
!
        end if
!
    end if
!      
    probgamma3_r = gam
!
!
! END OF FUNCTION probgamma3_r
! ____________________________
!
    end function probgamma3_r
!
! =========================================================================================
!
    function probgamma3_rv( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real vector argument X and a strictly positive value GAMP of the parameter P
!   of the Gamma distribution.
!   
!   PROBGAMMA3 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X(i) for i=1 to size(X).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:)
!           On entry, a positive real vector argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   For X<=max(GAMP/2,13), a Pearson's series expansion is used, see reference (3),
!   Formula 6.5.29, p.262. For larger values of X, an alternate Pearson's asymptotic
!   series expansion is used since this expansion tends to converge more quickly,
!   see reference (2), Formula 6.5.32, p.263.
!
!   In both cases, the "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The
!   default value for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   PROBGAMMA3 is faster, but less accurate than PROBGAMMA or PROBGAMMA2, since
!   for large values of X, the alternate Pearson's series expansion is only asymptotic.
!
!   This function is adapted from
!
!   (1)   Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)   Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)   Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                 Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                        :: gamp
    real(stnd), intent(in), dimension(:)          :: x
    real(stnd), intent(in),              optional :: acu
    real(stnd), dimension(size(x))                :: probgamma3_rv
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: xi
    real(stnd), dimension(size(x)) :: prob
!
    integer(i4b)  :: n, i
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,xi), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        xi = x(i)
!
        prob(i) = probgamma3_r( xi, gamp, acu=acu, maxiter=maxiter, failure=failure )
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma3_rv(:n) = prob(:n)
!
!
! END OF FUNCTION probgamma3_rv
! ____________________________
!
    end function probgamma3_rv
!
! =========================================================================================
!
    function probgamma3_rv2( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real vector argument X and a strictly positive vector argument GAMP of the
!   parameter P of the Gamma distribution.
!   
!   PROBGAMMA3 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP(i) will be less than or equal to X(i) for i=1 to size(X).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:)
!           On entry, a positive real vector argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd), dimension(:)
!           On entry, a strictly positive real vector argument which are the values of the parameter p
!           of the Gamma distribution.
!           Elements in GAMP(:) must be greater than zero.
!
!           The size of GAMP must verify  size(GAMP) = size(X)  .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP(i) (e.g. GAMP(i)>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   For X(i)<=max(GAMP(i)/2,13), a Pearson's series expansion is used, see reference (3),
!   Formula 6.5.29, p.262. For larger values of X(i), an alternate Pearson's asymptotic
!   series expansion is used since this expansion tends to converge more quickly,
!   see reference (2), Formula 6.5.32, p.263.
!
!   In both cases, the "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The
!   default value for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   PROBGAMMA3 is faster, but less accurate than PROBGAMMA or PROBGAMMA2, since
!   for large values of X, the alternate Pearson's series expansion is only asymptotic.
!
!   This function is adapted from
!
!   (1)   Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)   Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)   Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                 Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:)          :: x, gamp
    real(stnd), intent(in),              optional :: acu
    real(stnd), dimension(size(x))                :: probgamma3_rv2
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: xi, gampi
    real(stnd), dimension(size(x)) :: prob
!
    integer(i4b)  :: n, i
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probgamma3'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(x),i4b)     ,   &
                    int(size(gamp),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,xi,gampi), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        xi = x(i)
!
        gampi = gamp(i)
!
        prob(i) = probgamma3_r( xi, gampi, acu=acu, maxiter=maxiter, failure=failure )
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma3_rv2(:n) = prob(:n)
!
!
! END OF FUNCTION probgamma3_rv2
! ______________________________
!
    end function probgamma3_rv2
!
! =========================================================================================
!
    function probgamma3_rm( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real matrix argument X and a strictly positive value GAMP of the parameter P
!   of the Gamma distribution.
!   
!   PROBGAMMA3 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP will be less than or equal to X(i,j) for i=1 to size(X,1) and j=1 
!   to size(X,2).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:,:)
!           On entry, a positive real matrix argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:,:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd)
!           On entry, a strictly positive real argument which is the value of the parameter p
!           of the Gamma distribution.
!           GAMP must be greater than zero.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP (e.g. GAMP>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   For X<=max(GAMP/2,13), a Pearson's series expansion is used, see reference (3),
!   Formula 6.5.29, p.262. For larger values of X, an alternate Pearson's asymptotic
!   series expansion is used since this expansion tends to converge more quickly,
!   see reference (2), Formula 6.5.32, p.263.
!
!   In both cases, the "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The
!   default value for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   PROBGAMMA3 is faster, but less accurate than PROBGAMMA or PROBGAMMA2, since
!   for large values of X, the alternate Pearson's series expansion is only asymptotic.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)   Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)   Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                 Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                               :: gamp
    real(stnd), intent(in), dimension(:,:)               :: x
    real(stnd), intent(in),                     optional :: acu
    real(stnd), dimension(size(x,1),size(x,2))           :: probgamma3_rm
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: xij
    real(stnd), dimension(size(x,1),size(x,2)) :: prob
!
    integer(i4b)  :: n, m, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 1 )
    m = size( x, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,xij), SCHEDULE(STATIC)
!
    do j=1_i4b, m
!
        do i=1_i4b, n
!
            xij = x(i,j)
!
            prob(i,j) = probgamma3_r( xij, gamp, acu=acu, maxiter=maxiter, failure=failure )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma3_rm(:n,:m) = prob(:n,:m)
!
!
! END OF FUNCTION probgamma3_rm
! _____________________________
!
    end function probgamma3_rm
!
! =========================================================================================
!
    function probgamma3_rm2( x, gamp, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the gamma probability distribution function (e.g. the Incomplete Gamma Integral)
!   for a positive real matrix argument X and a strictly positive matrix argument GAMP of the
!   parameter P of the Gamma distribution.
!   
!   PROBGAMMA3 computes the probability that a random variable having a Gamma distribution
!   with parameter GAMP(i,j) will be less than or equal to X(i,j) for i=1 to size(X,1) and j=1 
!   to size(X,2).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:,:)
!           On entry, a positive real matrix argument X which gives the values of the 
!           upper integral limit.
!           Elements in X(:,:) must be greater or equal to zero.
!
!   GAMP    (INPUT) real(stnd), dimension(:,:)
!           On entry, a strictly positive real matrix argument which are the values of the parameter p
!           of the Gamma distribution.
!           Elements in GAMP(:,:) must be greater than zero.
!
!           The shape of GAMP must verify:
!
!           - size(GAMP,1) = size(X,1)
!           - size(GAMP,2) = size(X,2).
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. If l decimal places of accuracy are
!           required then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series expansion of the incomplete Gamma integral.
!           The default value is 1000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   For large GAMP(i,j) (e.g. GAMP(i,j)>1000), this function used a normal approximation, based
!   on the Wilson-Hilferty transformation, see reference (3), Formula 26.4.14, p.941.
!
!   For X(i,j)<=max(GAMP(i,j)/2,13), a Pearson's series expansion is used, see reference (3),
!   Formula 6.5.29, p.262. For larger values of X(i,j), an alternate Pearson's asymptotic
!   series expansion is used since this expansion tends to converge more quickly,
!   see reference (2), Formula 6.5.32, p.263.
!
!   In both cases, the "integrating" process is terminated when both the absolute and
!   relative contributions to the integral is not greater than the value of ACU. The
!   default value for ACU gives the maximum precision of this function.
!
!   The time taken by this function thus depends in the precision requested through ACU,
!   and also varies slightly with the input arguments X and GAMP.
!
!   PROBGAMMA3 is faster, but less accurate than PROBGAMMA or PROBGAMMA2, since
!   for large values of X, the alternate Pearson's series expansion is only asymptotic.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Lau, C.L., 1980: Algorithm AS 147: A simple series for the Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 29, No. 1, pp. 113-114
!
!   (2)   Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                          Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)   Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                 Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:)               :: x, gamp
    real(stnd), intent(in),                     optional :: acu
    real(stnd), dimension(size(x,1),size(x,2))           :: probgamma3_rm2
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: xij, gampij
    real(stnd), dimension(size(x,1),size(x,2)) :: prob
!
    integer(i4b)  :: n, m, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probgamma3'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
    n =  assert_eq( int(size(x,1),i4b)     ,   &
                    int(size(gamp,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(x,2),i4b)     ,   &
                    int(size(gamp,2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,xij,gampij), SCHEDULE(STATIC)
!
    do j=1_i4b, m
!
        do i=1_i4b, n
!
            xij = x(i,j)
!
            gampij = gamp(i,j)
!
            prob(i,j) = probgamma3_r( xij, gampij, acu=acu, maxiter=maxiter, failure=failure )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!      
    probgamma3_rm2(:n,:m) = prob(:n,:m)
!
!
! END OF FUNCTION probgamma3_rm2
! ______________________________
!
    end function probgamma3_rm2
!
!
! =========================================================================================
!                            INVERSE GAMMA PROBABILITY DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function pinvgamma_r( p, gamp, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the inverse gamma probability distribution function.
!
!   For given arguments P (0<=P<=1) and GAMP (GAMP>0), PINVGAMMA returns the value X such that P is 
!   the probability that a random variable distributed as a gamma distribution with parameter GAMP
!   is less than or equal to X.
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd)
!           On entry, input probability.
!           P must be in the range (0,1) inclusive.
!
!   GAMP    (INPUT) real(stnd)
!           on entry, the parameter p of the gamma distribution.
!           GAMP must be greater or equal to 0.25 .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result when computing the incomplete 
!           Gamma integral in the evaluation of the seven term Taylor series in function
!           PINVQ2 .
!
!           If l decimal places of accuracy are required then ACU should
!           be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!
!           The default value is 1000.
!
!           See the description of the PROBGAMMA2 function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function actually uses PINVQ2 function and is adapted from
!
!   (1)   Best, D.J., and Roberts, D.E., 1975: Algorithm AS 91: The Percentage Points of 
!                     the chi2 Distribution. Appl. Statist., Vol.24, No. 3, pp.385-388
!
!   (2)   Shea, B.L., 1988: Algorithm AS 239: Chi-squared and Incomplete Gamma Integral,
!                       Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)   Shea, B.L., 1991: Algorithm AS R85: A remark on AS 91: The Percentage Points of
!                     the chi2 Distribution. Appl. Statist. Vol.40, No. 1, pp.233-235.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, quarter, half, one, two
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error1, prob_error18
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: p, gamp
    real(stnd), intent(in), optional :: acu
!
    real(stnd)                       :: pinvgamma_r
!
    integer(i4b), intent(in), optional :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: df
!   
!   
! PARAMETERS 
! __________
!
!
    character(len=*),  parameter :: name_proc='pinvgamma'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( gamp<quarter ) then
        call merror( name_proc//prob_error18 )
    end if
!
    if ( p<zero .or. p>one )  then
        call merror( name_proc//prob_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( p==zero .or. p==one ) then
!
        if ( p==zero ) then
            pinvgamma_r = zero
        else
            pinvgamma_r = machhuge
        end if
!  
!        pinvgamma_r = merge( zero, machhuge, p==zero )
!
        return
!
    end if
!
!   INITIALISE FOR INVERSE CHI2 DISTRIBUTION USAGE.
!
    df = two*gamp
!
    pinvgamma_r = half*pinvq2_r( p, df, acu=acu, maxiter=maxiter )
!
!
! END OF FUNCTION pinvgamma_r
! ___________________________
!
    end function pinvgamma_r
!
!
! =========================================================================================
!                            BETA PROBABILITY DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function probbeta_r( x, a, b, beta, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the beta probability distribution function (e.g the incomplete beta function).
!
!   For given arguments X (0<=X<=1), A (A>0), B (B>0), PROBBETA returns the probability that
!   a random variable from a beta distribution having parameters A and B will be less than or
!   equal to X.
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd)
!           On entry, the value to which function is to be integrated.
!           X must be in range (0,1) inclusive.
!
!   A       (INPUT) real(stnd)
!           on entry, the (1st) parameter of the beta distribution.
!           A must be greater than 0.
!
!   B       (INPUT) real(stnd)
!           On entry, the (2nd) parameter of the beta distribution.
!           B must be greater than 0.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(A,B).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. The "integrating"
!           process is terminated when both the absolute and relative contributions
!           to the integral is not greater than the value of ACU.
!
!           ACU  is a small strictly positive integer. If the number of decimal digits'
!           accuracy required is r, ACU should be set to 10**(-(r+1)).
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   This function is adapted from
!
!   (1)     Majumder, K.L., and Bhattacharjee, G.P., 1973: Algorithm AS 63: the Incomplete
!                   Beta Integral. Appl. Statist., Vol.22, No.3, pp 409-411.
!
!   (2)     Cran, G.W., Martin, K.J., and  Thomas, G.E., 1977: Remark AS R19 and Algorithm AS 109:
!                A remark on Algorithms: AS 63 the Incomplete Beta integral, AS 64 Inverse of
!                the Incomplete Beta Function Ratio. Appl. Statist., Vol.26, No.1, pp 111-114.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : macheps
    use Logical_Constants, only : true, false
    use Char_Constants,    only : prob_error7, prob_error8, prob_error9
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: x, a, b
    real(stnd), intent(in), optional :: beta, acu
    real(stnd)                       :: probbeta_r
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: acu2, beta2, betain, asb, cx, xx, aa, bb, term, ai, rx, temp
!
    integer(i4b) :: ns, i, maxit
!
    logical(lgl) :: indx, fail
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probbeta'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( a<=zero .or. b<=zero ) then
        call merror( name_proc//prob_error7 )
    end if
!
    if ( x<zero .or. x>one ) then
        call merror( name_proc//prob_error8 )
    end if
!
    acu2 = macheps
    if ( present(acu) ) then
        if ( acu<=zero ) call merror( name_proc//prob_error9 )
        acu2 = acu
    end if
!
    maxit = 2000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( x==zero .or. x==one ) then
!
        probbeta_r = x
!
        return
!
    end if
!
!   CALCULATE log( BETA(a,b)) IF REQUIRED.
!
    asb  = a + b
!
    if ( present(beta) ) then
        beta2 = beta
    else
        beta2 = lngamma_r( a ) + lngamma_r( b ) - lngamma_r( asb )
    end if
!
!   CHANGE TAIL FOR MORE EFFICIENT USE OF THE POWER SERIES.
!
    if ( a<asb*x ) then
!
        xx   = one - x
        cx   = x
        aa   = b
        bb   = a
        indx = true
!
    else
!
        xx   = x
        cx   = one - x
        aa   = a
        bb   = b
        indx = false
!
    end if
!
    term   = one
    ai     = one
    betain = one
    ns     = int( bb + cx*asb, i4b )
!
!   USE SOPER'S REDUCTION FORMULAE.
!
!    rx = merge( xx, xx/cx, ns==0_i4b )
!  
    if ( ns==0_i4b ) then
        rx = xx
    else
        rx = xx/cx
    end if
!
    temp = bb - ai
!
    do i = 1_i4b, maxit
!
        term   = term*temp*rx/(aa+ai)
        betain = betain + term
        temp   = abs( term )
!
        if ( temp<=acu2 .and. temp<=acu2*betain ) exit
!
        ai = ai + one
        ns = ns - 1_i4b
!
        if ( ns>=0_i4b ) then
            temp = bb - ai
            if ( ns==0_i4b ) rx = xx
        else
            temp = asb
            asb = asb + one
        end if
!
    end do
!
    if ( fail .and. i>maxit ) then
!
        probbeta_r = -one
!
    else
!
        betain = betain*exp( aa*log(xx) + (bb-one)*log(cx) - beta2 )/aa
!
        if ( indx ) then
            probbeta_r = one - betain
        else
            probbeta_r = betain
        end if
!
    end if
!
!
! END OF FUNCTION probbeta_r
! __________________________
!
    end function probbeta_r
!
! =========================================================================================
!
    function probbeta_rv( x, a, b, beta, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the beta probability distribution function (e.g the incomplete beta function).
!
!   For given arguments X(:) (0<=X(:)<=1), A (A>0), B (B>0), PROBBETA returns the probabilities
!   that a random variable from a beta distribution having parameters A and B will be less than or
!   equal to X(:).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:)
!           On entry, the values to which function is to be integrated.
!           Elements in X(:) must be in the range (0,1) inclusive.
!
!   A       (INPUT) real(stnd)
!           on entry, the (1st) parameter of the beta distribution.
!           A must be greater than 0.
!
!   B       (INPUT) real(stnd)
!           On entry, the (2nd) parameter of the beta distribution.
!           B must be greater than 0.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(A,B).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. The "integrating"
!           process is terminated when both the absolute and relative contributions
!           to the integral is not greater than the value of ACU.
!
!           ACU  is a small strictly positive integer. If the number of decimal digits'
!           accuracy required is r, ACU should be set to 10**(-(r+1)).
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)     Majumder, K.L., and Bhattacharjee, G.P., 1973: Algorithm AS 63: the Incomplete
!                   Beta Integral. Appl. Statist., Vol.22, No.3, pp 409-411.
!
!   (2)     Cran, G.W., Martin, K.J., and  Thomas, G.E., 1977: Remark AS R19 and Algorithm AS 109:
!                A remark on Algorithms: AS 63 the Incomplete Beta integral, AS 64 Inverse of
!                the Incomplete Beta Function Ratio. Appl. Statist., Vol.26, No.1, pp 111-114.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : macheps
    use Logical_Constants, only : true, false
    use Char_Constants,    only : prob_error7, prob_error8, prob_error9
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                :: a, b
    real(stnd), intent(in), dimension(:)  :: x
    real(stnd), intent(in), optional      :: beta, acu
    real(stnd), dimension(size(x))        :: probbeta_rv
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: acu2, beta2, betain, asb, cx, xx, aa, bb,   &
                                      term, ai, rx, temp, xj
    real(stnd), dimension(size(x)) :: prob
!
    integer(i4b) :: n, i, j, maxit, ns
!
    logical(lgl) :: indx, fail
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probbeta'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( a<=zero .or. b<=zero ) then
        call merror( name_proc//prob_error7 )
    end if
!
    if ( any(x(:n)<zero) .or. any(x(:n)>one) )  then
        call merror( name_proc//prob_error8 )
    end if
!
    acu2 = macheps
    if ( present(acu) ) then
        if ( acu<=zero ) call merror( name_proc//prob_error9 )
        acu2 = acu
    end if
!
    maxit = 2000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
!   CALCULATE log( BETA(a,b)) IF REQUIRED.
!
    if ( present(beta) ) then
        beta2 = beta
    else
        beta2 = lngamma_r( a ) + lngamma_r( b ) - lngamma_r( a + b )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(j,i,ns,indx,xj,xx,cx,aa,bb,term,ai,betain,rx,temp,asb)
!
    do j = 1_i4b, n
!
       xj = x(j)
!
       if ( xj==zero .or. xj==one ) then
!
           prob(j) = xj
!
       else
!
           asb  = a + b
!
!          CHANGE TAIL FOR MORE EFFICIENT USE OF THE POWER SERIES.
!
            if ( a<asb*xj ) then
!
                xx   = one - xj
                cx   = xj
                aa   = b
                bb   = a
                indx = true
!
            else
!
                xx   = xj
                cx   = one - xj
                aa   = a
                bb   = b
                indx = false
!
            end if
!
            term   = one
            ai     = one
            betain = one
            ns     = int( bb + cx*asb, i4b )
!
!           USE SOPER'S REDUCTION FORMULAE.
!
!            rx = merge( xx, xx/cx, ns==0_i4b )
            if ( ns==0_i4b ) then
                rx = xx
            else
                rx = xx/cx
            end if
!
            temp = bb - ai
!
            do i = 1_i4b, maxit
!
                term   = term*temp*rx/(aa+ai)
                betain = betain + term
                temp   = abs( term )
!
                if ( temp<=acu2 .and. temp<=acu2*betain ) exit
!
                ai = ai + one
                ns = ns - 1_i4b
!
                if ( ns>=0_i4b ) then
                    temp = bb - ai
                    if ( ns==0_i4b ) rx = xx
                else
                    temp = asb
                    asb = asb + one
                end if
!
            end do
!
            if ( fail .and. i>maxit ) then
!
                prob(j) = -one
!
            else
!
                betain = betain*exp( aa*log(xx) + (bb-one)*log(cx) - beta2 )/aa
!
                if ( indx ) then
                    prob(j) = one - betain
                else
                    prob(j) = betain
                end if
!
            end if
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
    probbeta_rv(:n) = prob(:n)
!
! END OF FUNCTION probbeta_rv
! ___________________________
!
    end function probbeta_rv
!
! =========================================================================================
!
    function probbeta_rm( x, a, b, beta, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the beta probability distribution function (e.g the incomplete beta function).
!
!   For given arguments X(:,:) (0<=X(:,:)<=1), A (A>0), B (B>0), PROBBETA returns the probabilities
!   that a random variable from a beta distribution having parameters A and B will be less than or
!   equal to X(:,:).
!
!
! Arguments
! _________
!
!   X       (INPUT) real(stnd), dimension(:,:)
!           On entry, the values to which function is to be integrated.
!           Elements in X(:,:) must be in the range (0,1) inclusive.
!
!   A       (INPUT) real(stnd)
!           on entry, the (1st) parameter of the beta distribution.
!           A must be greater than 0.
!
!   B       (INPUT) real(stnd)
!           On entry, the (2nd) parameter of the beta distribution.
!           B must be greater than 0.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(A,B).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result. The "integrating"
!           process is terminated when both the absolute and relative contributions
!           to the integral is not greater than the value of ACU.
!
!           ACU  is a small strictly positive integer. If the number of decimal digits'
!           accuracy required is r, ACU should be set to 10**(-(r+1)).
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)     Majumder, K.L., and Bhattacharjee, G.P., 1973: Algorithm AS 63: the Incomplete
!                   Beta Integral. Appl. Statist., Vol.22, No.3, pp 409-411.
!
!   (2)     Cran, G.W., Martin, K.J., and  Thomas, G.E., 1977: Remark AS R19 and Algorithm AS 109:
!                A remark on Algorithms: AS 63 the Incomplete Beta integral, AS 64 Inverse of
!                the Incomplete Beta Function Ratio. Appl. Statist., Vol.26, No.1, pp 111-114.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : macheps
    use Logical_Constants, only : true, false
    use Char_Constants,    only : prob_error7, prob_error8, prob_error9
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: a, b
    real(stnd), intent(in), dimension(:,:)              :: x
    real(stnd), intent(in),                    optional :: beta, acu
    real(stnd), dimension(size(x,1),size(x,2))          :: probbeta_rm
!
    integer(i4b), intent(in), optional    :: maxiter
!
    logical(lgl), intent(in), optional    :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: acu2, beta2, betain, asb, cx, xx, aa, bb,   &
                                                  term, ai, rx, temp, xjl
    real(stnd), dimension(size(x,1),size(x,2)) :: prob
!
    integer(i4b) :: n, p, i, j, l, maxit, ns
!
    logical(lgl) :: indx, fail
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probbeta'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 1 )
    p = size( x, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. p<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( a<=zero .or. b<=zero ) then
        call merror( name_proc//prob_error7 )
    end if
!
    if ( any(x(:n,:p)<zero) .or. any(x(:n,:p)>one) ) then
        call merror( name_proc//prob_error8 )
    end if
!
    acu2 = macheps
    if ( present(acu) ) then
        if ( acu<=zero ) call merror( name_proc//prob_error9 )
        acu2 = acu
    end if
!
    maxit = 2000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
!   CALCULATE log( BETA(a,b)) IF REQUIRED.
!
    if ( present(beta) ) then
        beta2 = beta
    else
        beta2 = lngamma_r( a ) + lngamma_r( b ) - lngamma_r( a + b )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*p>=omp_limit                      .and.      &
                  p/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(l,j,i,ns,indx,xjl,xx,cx,aa,bb,term,ai,betain,rx,temp,asb)
!
    do l = 1_i4b, p
!
        do j = 1_i4b, n
!
            xjl = x(j,l)
!
            if ( xjl==zero .or. xjl==one ) then
!
                prob(j,l) = xjl
!
            else
!
                asb  = a + b
!
!               CHANGE TAIL FOR MORE EFFICIENT USE OF THE POWER SERIES.
!
                if ( a<asb*xjl ) then
!
                    xx   = one - xjl
                    cx   = xjl
                    aa   = b
                    bb   = a
                    indx = true
!
                else
!
                    xx   = xjl
                    cx   = one - xjl
                    aa   = a
                    bb   = b
                    indx = false
!
                end if
!
                term   = one
                ai     = one
                betain = one
                ns     = int( bb + cx*asb, i4b )
!
!               USE SOPER'S REDUCTION FORMULAE.
!
!                rx = merge( xx, xx/cx, ns==0_i4b )
!
                if ( ns==0_i4b ) then
                    rx = xx
                else
                    rx = xx/cx
                end if
!
                temp = bb - ai
!
                do i = 1_i4b, maxit
!
                    term   = term*temp*rx/(aa+ai)
                    betain = betain + term
                    temp   = abs( term )
!
                    if ( temp<=acu2 .and. temp<=acu2*betain ) exit
!
                    ai = ai + one
                    ns = ns - 1_i4b
!
                    if ( ns>=0_i4b ) then
                        temp = bb - ai
                        if ( ns==0_i4b ) rx = xx
                    else
                        temp = asb
                        asb = asb + one
                    end if
!
                end do
!
                if ( fail .and. i>maxit ) then
!
                    prob(j,l) = -one
!
                else
!
                    betain = betain*exp( aa*log(xx) + (bb-one)*log(cx) - beta2 )/aa
!
                    if ( indx ) then
                        prob(j,l) = one - betain
                    else
                        prob(j,l) = betain
                    end if
!
                end if
!
            end if
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probbeta_rm(:n,:p) = prob(:n,:p)
!
! END OF FUNCTION probbeta_rm
! ___________________________
!
    end function probbeta_rm
!
!
! =========================================================================================
!                            INVERSE BETA PROBABILITY DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function pinvbeta_r( p, a, b, beta, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the inverse beta probability distribution function (e.g. the incomplete beta function).
!
!   For given arguments P (0<=P<=1), A (A>0.1), B (B>0.1), PINVBETA returns the value X such that P is 
!   the probability that a random variable distributed as beta(A,B) is less than or equal to X.
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd)
!           On entry, input probability.
!           P must be in the range (0,1) inclusive.
!
!   A       (INPUT) real(stnd)
!           on entry, the (1st) parameter of the beta distribution.
!           A must be greater than 0.1 .
!
!   B       (INPUT) real(stnd)
!           On entry, the (2nd) parameter of the beta distribution.
!           B must be greater than 0.1 .
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function beta(A,B).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result, when computing the
!           incomplete beta function. The "integrating" process for evaluating
!           the incomplete beta function is terminated when the
!           relative contribution to the integral is not greater than
!           the value of ACU. ACU  is a small strictly positive integer.
!           See the description of the PROBBETA function for more details
!           on this argument.
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!           See the description of the PROBBETA function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function is not very accurate for small values of A and/or B (e.g. less than 0.5).
!
!   This function is adapted from
!
!   (1)     Majumder, K.L., and Bhattacharjee, G.P., 1973: Algorithm AS 64: Inverse of
!                the Incomplete Beta Function Ratio. Appl. Statist., Vol.22, No.3, pp 411-414.
!
!   (2)     Cran, G.W., Martin, K.J., and  Thomas, G.E., 1977: Remark AS R19 and Algorithm AS 109:
!                A remark on Algorithms: AS 63 the Incomplete Beta integral, AS 64 Inverse of
!                the Incomplete Beta Function Ratio. Appl. Statist., Vol.26, No.1, pp 111-114.
!
!   (3)     Berry, K.J.,  Mielke, P.W., and Cran, G.W., 1990: Algorithm AS R83: A remark on Algorithm
!                AS 109: Inverse of the Incomplete Beta Function Ratio. Appl. Statist., Vol.39,
!                No 2, pp. 309-310.
!
!   (4)     Berry, K.J.,  Mielke, P.W., and Cran, G.W., 1991: Correction to Algorithm AS R83: A remark
!                on Algorithm AS 109: Inverse of the Incomplete Beta Function Ratio.
!                Appl. Statist. Vol. 40, No. 1, p.236
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, tenth, fifth, half, one, two, three, four, five, &
                                  six, nine, ten, c13
    use Num_Constants,     only : macheps, machdecexpr
    use Logical_Constants, only : true, false
    use Char_Constants,    only : prob_error1, prob_error9, prob_error22
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: p, a, b
    real(stnd), intent(in), optional :: beta, acu
    real(stnd)                       :: pinvbeta_r
!
    integer(i4b), intent(in), optional    :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: acu2, beta2, asb, sae, fpu, iex, eps2, x, pp, aa, bb,  &
                  g, h, r, s, t, w, y, yprev, sq, prev, adj, tx, xl, xr, &
                  fxl, fcs, xrmxl
!
    integer(i4b) :: i
!
    logical(lgl) :: indx, use_bisect
!   
!   
! PARAMETERS 
! __________
!
!   DEFINE ACCURACY FOR BISECTION METHOD.
!
    real(stnd),  parameter :: eps=0.00001_stnd, sig=0.000001_stnd
!
!   MAXIMUM NUMBER OF ITERATIONS FOR BISECTION AND NEWTON METHODS.
!
    integer(i4b),  parameter :: itmax=50_i4b
!
    character(len=*),  parameter :: name_proc='pinvbeta'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( a<=tenth .or. b<=tenth ) then
        call merror( name_proc//prob_error22 )
    end if
!
    if ( p<zero .or. p>one ) then
        call merror( name_proc//prob_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( p==zero .or. p==one ) then
!
       pinvbeta_r = p
!
       return
!
    end if
!
!   DEFINE ACCURACY FOR INCOMPLETE BETA FUNCTION IF REQUIRED.
!
    acu2 = macheps
    if ( present(acu) ) then
        if ( acu<=zero ) call merror( name_proc//prob_error9 )
        acu2 = acu
    end if
!
!   CALCULATE log( BETA(A,B)) IF REQUIRED.
!
    asb  = a + b
!
    if ( present(beta) ) then
        beta2 = beta
    else
        beta2 = lngamma_r( a ) + lngamma_r( b ) - lngamma_r( asb )
    end if
!
    xl = min( a, b )
    xr = max( a, b )
!
    i = 0_i4b
    use_bisect = xl>one .and. ten*xl>xr
!
    if ( use_bisect ) then
!
!      USE A BISECTION METHOD.
!
       xl = zero
       xr = one
!
       fxl = -p
!
       do i = 1_i4b, itmax
!
           x = ( xl + xr )*half
           pp = probbeta_r( x, a, b, beta=beta2, acu=acu2, maxiter=maxiter )
           fcs = pp - p
!
           if ( fcs*fxl>zero ) then
               xl  = x
               fxl = fcs
           else
               xr  = x
           end if
!
           xrmxl = xr - xl
!
           if ( xrmxl<=sig .and. abs(fcs)<=eps ) exit
!
       end do
!
    end if
!
    if ( i>itmax ) use_bisect = false
!
    if ( .not.use_bisect ) then
!
!       USE NEWTON METHOD FOR SKEWED CASES OR
!       NO CONVERGENCE WITH BISECTION METHOD.
!
!       CHANGE TAIL IF NECESSARY.
!
        if ( p>half ) then
!
            pp   = one - p
            aa   = b
            bb   = a
            indx = true
!
        else
!
            pp   = p
            aa   = a
            bb   = b
            indx = false
!
        end if
!
!       CALCULATE THE INITIAL APPROXIMATION.
!
        r = sqrt( -log(pp*pp) )
        y = r - ( 2.30753_stnd+0.27061_stnd*r )/( one + (0.99229_stnd+0.04481_stnd*r)*r )
!
        if ( aa>one .and. bb>one ) then
!
            r = (y*y-three)/six
            s = one/( aa + aa -one )
            t = one/( bb + bb - one )
            h = two/( s + t )
            w = y*sqrt( h+r )/h - (t-s)*( r + five/six -two/(three*h) )
            x = aa/( aa + bb*exp( w + w ) )
!
        else
!
            r = bb + bb
            t = one/( nine*bb )
            t = r*( one - t +y*sqrt(t) )**3
!
            if ( t<=zero ) then
!
                x = one - exp( ( log( (one-pp)*bb) + beta2 )/bb )
!
            else
!
                t = (four*aa + r - two )/t
                if ( t<=one ) then
                    x = exp( ( log(pp*aa)+beta2 )/aa )
                else
                    x = one - two/( t + one )
                end if
!
            end if
!
        end if
!
!       SOLVE FOR x BY A MODIFIED NEWTON-RAPHSON METHOD,
!       USING THE FUNCTION probbeta_r .
!
        r     = one - aa
        t     = one - bb
        yprev = zero
        sq    = one
        prev  = one
!
        x =  min( max( x, 0.0001_stnd ), 0.9999_stnd ) 
!
!
!       DEFINE ACCURACY.
!       sae BELOW IS THE MOST NEGATIVE DECIMAL EXPONENT WHICH DOES NOT 
!       CAUSE AN UNDERFLOW.
!
        sae = -machdecexpr
        fpu = ten**sae
        iex = max(-five/aa**2 - one/pp**fifth - c13, sae)
        eps2 = ten**iex
!
        b_ext : do i = 1_i4b, itmax
!
            y   = probbeta_r( x, aa, bb, beta=beta2, acu=acu2, maxiter=maxiter )
            y   = ( y - pp )*exp( beta2 + r*log(x) + t*log(one-x) )
!
            if ( y*yprev<=zero ) prev = max( sq, fpu )
!
            g = one
!
            do
!
                adj = g*y
                sq  = adj*adj
!
                if ( sq<prev ) then
!
                    tx = x - adj
!
                    if ( tx>=zero .and. tx<=one ) then
!
                        if ( prev<=eps2 .or. y*y<=eps2 ) exit b_ext
!
                        if ( tx/=zero .and. tx/=one ) then
!
                            if ( tx==x ) exit b_ext
!
                            x     = tx
                            yprev  = y
!
                            cycle b_ext
!
                        end if
!
                    end if
!
                end if
!
                g = g/three
!
            end do
!
        end do b_ext
!
        if ( indx ) then
            x = one - x
        end if
!
    end if
!
    pinvbeta_r = x
!
!
! END OF FUNCTION pinvbeta_r
! __________________________
!
    end function pinvbeta_r
!
!
! =========================================================================================
!                            GAUSSIAN DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function probn_r( x, upper )
!
! Purpose
! _______
!
!   Evaluates the standard normal (Gaussian) distribution function from X to infinity 
!   if UPPER is true or from minus infinity to X if UPPER is false. In other words, if:
!
!   - UPPER = true  :    PROBN = prob( U > X )  ,
!   - UPPER = false :    PROBN = prob( U < X )  ,
!   
!   , for U = Laplace_Gauss(0;1).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd)
!          On entry, upper or lower limit of integration.
!
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  : probability to the right of X is calculated.
!          - UPPER = false : probability to the left of X is calculated.
!
!
! Further Details
! _______________
!
!   This function is adapted from
!
!   (1)   Hill, I.D., 1973: Algorithm AS66: The Normal Integral, Applied Statistics,
!                           vol.22, no.3, pp.424-427
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, half, nine, three
    use Num_Constants,     only : machdecprec
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: x
    real(stnd)             :: probn_r
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: z, y
!
    logical(lgl) :: up
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: z1 = 1.28_stnd, z2 = (machdecprec+nine)/three, z3 = 18.66_stnd
!
    real(stnd), parameter :: p  = 0.398942280444_stnd,  q  = 0.39990348504_stnd,  &
                             r  = 0.398942280385_stnd,  a1 = 5.75885480458_stnd,  &
                             a2 =  2.62433121679_stnd,  a3 = 5.92885724438_stnd,  &
                             b1 = -29.8213557807_stnd,  b2 = 48.6959930692_stnd,  &
                             c1 = -3.8052E-8_stnd,      c2 = 3.98064794E-4_stnd,  &
                             c3 = -0.151679116635_stnd, c4 = 4.8385912808_stnd,   &
                             c5 = 0.742380924027_stnd,  c6 = 3.99019417011_stnd,  &
                             d1 = 1.00000615302_stnd,   d2 = 1.98615381364_stnd,  &
                             d3 = 5.29330324926_stnd,   d4 = -15.1508972451_stnd, &
                             d5 = 30.789933034_stnd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( x>=zero ) then
!
        up = upper
        z = x
!
    else
!
        up = .not. upper
        z = -x
!
    end if
!
    if ( z<=z2 .or. (up .and. z<=z3) ) then
!
        y = half*z*z
!
        if ( z>z1 ) then
            probn_r = r*exp(-y)/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))))
        else
            probn_r = half - z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))))
        end if
!
    else
!
        probn_r = zero
!
    end if
!
    if ( .not. up )      probn_r = one - probn_r
    if ( probn_r<zero )  probn_r = zero
    if ( probn_r>one )   probn_r = one
!
!
! END OF FUNCTION probn_r
! _______________________
!
    end function probn_r
!
! =========================================================================================
!
    function probn_rv( x, upper )
!
! Purpose
! _______
!
!   Evaluates the standard normal (Gaussian) distribution function from X(i) to infinity 
!   if UPPER is true or from minus infinity to X(i) if UPPER is false, for i=1 to
!   size(X). In other words, if:
!
!   - UPPER = true  :    PROBN( i ) = prob( U > X(i) )  ,
!   - UPPER = false :    PROBN( i ) = prob( U < X(i) )  ,
!   
!   , for U = Laplace_Gauss(0;1) and i=1 to size(X).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, upper or lower limits of integration.
!
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  : probability to the right of X is calculated.
!          - UPPER = false : probability to the left of X is calculated.
!
!
! Further Details
! _______________
!
!   The subroutine is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, I.D., 1973: Algorithm AS66: The Normal Integral, Applied Statistics,
!                           vol.22, no.3, pp.424-427
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, half, nine, three
    use Num_Constants,     only : machdecprec
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),      intent(in) :: x
    real(stnd), dimension(size(x))            :: probn_rv
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: z, y, xi
    real(stnd), dimension(size(x)) :: probvec
!
    logical(lgl) :: up
!   
    integer(i4b) :: n, i
!
#ifdef _OPENMP
!   
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: z1 = 1.28_stnd, z2 = (machdecprec+nine)/three, z3 = 18.66_stnd
!
    real(stnd), parameter :: p  = 0.398942280444_stnd,  q  = 0.39990348504_stnd,  &
                             r  = 0.398942280385_stnd,  a1 = 5.75885480458_stnd,  &
                             a2 =  2.62433121679_stnd,  a3 = 5.92885724438_stnd,  &
                             b1 = -29.8213557807_stnd,  b2 = 48.6959930692_stnd,  &
                             c1 = -3.8052E-8_stnd,      c2 = 3.98064794E-4_stnd,  &
                             c3 = -0.151679116635_stnd, c4 = 4.8385912808_stnd,   &
                             c5 = 0.742380924027_stnd,  c6 = 3.99019417011_stnd,  &
                             d1 = 1.00000615302_stnd,   d2 = 1.98615381364_stnd,  &
                             d3 = 5.29330324926_stnd,   d4 = -15.1508972451_stnd, &
                             d5 = 30.789933034_stnd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,up,z,y,xi)
!
    do i = 1_i4b, n
!
        xi = x(i)
!
        if ( xi>=zero ) then
            up = upper
            z  = xi
        else
            up = .not. upper
            z  = -xi
        end if
!
        if ( z<=z2 .or. (up .and. z<=z3) ) then
!
            y = half*z*z
!
            if ( z>z1 ) then
                xi = r*exp(-y)/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))))
            else
                xi = half - z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))))
            end if
!
        else
!
            xi = zero
!
        end if
!
        if ( .not. up ) xi = one - xi
        if ( xi<zero )  xi = zero
        if ( xi>one )   xi = one
!
        probvec(i) = xi
!
    end do
!
!$OMP END PARALLEL DO
!
    probn_rv(1_i4b:n) = probvec(1_i4b:n)
!
!
! END OF FUNCTION probn_rv
! ________________________
!
    end function probn_rv
!
! =========================================================================================
!
    function probn_rm( x, upper )
!
! Purpose
! _______
!
!   Evaluates the standard normal (Gaussian) distribution function from X(i,j) to infinity 
!   if UPPER is true or from minus infinity to X(i,j) if UPPER is false, for i=1 to
!   size(X,1) and j=1 to size(X,2). In other words, if:
!
!   - UPPER = true  :    PROBN( i, j ) = prob( U > X(i,j) )  ,
!   - UPPER = false :    PROBN( i, j ) = prob( U < X(i,j) )  ,
!   
!   , for U = Laplace_Gauss(0;1) and i=1 to size(X,1) and j=1 to size(X,2).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:)
!          On entry, upper or lower limits of integration.
!
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  : probability to the right of X is calculated.
!          - UPPER = false : probability to the left of X is calculated.
!
!
! Further Details
! _______________
!
!   The subroutine is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, I.D., 1973: Algorithm AS66: The Normal Integral, Applied Statistics,
!                           vol.22, no.3, pp.424-427
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, half, nine, three
    use Num_Constants,     only : machdecprec
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                 intent(in) :: x
    real(stnd), dimension(size(x,1),size(x,2))             :: probn_rm
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: z, y, xji
    real(stnd), dimension(size(x,1),size(x,2)) :: probmat
!
    logical(lgl) :: up
!   
    integer(i4b) :: n, m, i, j
!
#ifdef _OPENMP
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: z1 = 1.28_stnd, z2 = (machdecprec+nine)/three, z3 = 18.66_stnd
!
    real(stnd), parameter :: p  = 0.398942280444_stnd,  q  = 0.39990348504_stnd,  &
                             r  = 0.398942280385_stnd,  a1 = 5.75885480458_stnd,  &
                             a2 =  2.62433121679_stnd,  a3 = 5.92885724438_stnd,  &
                             b1 = -29.8213557807_stnd,  b2 = 48.6959930692_stnd,  &
                             c1 = -3.8052E-8_stnd,      c2 = 3.98064794E-4_stnd,  &
                             c3 = -0.151679116635_stnd, c4 = 4.8385912808_stnd,   &
                             c5 = 0.742380924027_stnd,  c6 = 3.99019417011_stnd,  &
                             d1 = 1.00000615302_stnd,   d2 = 1.98615381364_stnd,  &
                             d3 = 5.29330324926_stnd,   d4 = -15.1508972451_stnd, &
                             d5 = 30.789933034_stnd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( x, 1 )
    n = size( x, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)             &
!$OMP            ,PRIVATE(i,j,up,z,y,xji)
!
    do i = 1_i4b, n
!
        do j = 1_i4b, m
!
            xji = x(j,i)
!
            if ( xji>=zero ) then
                up = upper
                z  = xji
            else
                up = .not. upper
                z  = -xji
            end if
!
            if ( z<=z2 .or. (up .and. z<=z3) ) then
!
                y = half*z*z
!
                if ( z>z1 ) then
                    xji = r*exp(-y)/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))))
                else
                    xji = half - z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))))
                end if
!
            else
!
                xji = zero
!
            end if
!
            if ( .not. up ) xji = one - xji
            if ( xji<zero ) xji = zero
            if ( xji>one )  xji = one
!
            probmat(j,i) = xji
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probn_rm(1_i4b:m,1_i4b:n) = probmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION probn_rm
! ________________________
!
    end function probn_rm
!
! =========================================================================================
!
    function probn2_r( x, upper )
!
! Purpose
! _______
!
!   Evaluates the standard normal (Gaussian) distribution function from X to infinity 
!   if UPPER is true or from minus infinity to X if UPPER is false. In other words, if:
!
!   - UPPER = true  :    PROBN2 = prob( U > X )  ,
!   - UPPER = false :    PROBN2 = prob( U < X )  ,
!   
!   , for U = Laplace_Gauss(0;1).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(extd)
!          On entry, upper or lower limit of integration.
!
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if: 
!
!          - UPPER = true  : probability to the right of X is calculated.
!          - UPPER = false : probability to the left of X is calculated.
!
!
! Further Details
! _______________
!
!   This function gives higher accuracy than PROBN.
!
!   This function is based upon algorithm 5666 for the error function, from:
!
!   (1)   Hart, J.F. et al, 1968: Computer Approximations, 354 pp, New York, Wiley.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero_extd, one_extd, two_extd, three_extd,    &
                                  four_extd, half_extd, c37_extd, c0_65_extd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), intent(in) :: x
    real(extd)             :: probn2_r
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)   :: xabs, expntl
!
    logical(lgl) :: up
!   
!   
! PARAMETERS 
! __________
!
    real(extd), parameter :: cutoff=7.071_extd,               &
                             root2pi=2.506628274631001_extd
!
    real(extd), parameter ::                                  &
                             p0=220.2068679123761_extd,       &
                             p1=221.2135961699311_extd,       &
                             p2=112.0792914978709_extd,       &
                             p3=33.91286607838300_extd,       &
                             p4=6.373962203531650_extd,       &
                             p5=.7003830644436881_extd,       &
                             p6=.3526249659989109e-01_extd
!
    real(extd), parameter ::                                  &
                             q0=440.4137358247522_extd,       &
                             q1=793.8265125199484_extd,       &
                             q2=637.3336333788311_extd,       &
                             q3=296.5642487796737_extd,       &
                             q4=86.78073220294608_extd,       &
                             q5=16.06417757920695_extd,       &
                             q6=1.755667163182642_extd,       &
                             q7=.8838834764831844e-1_extd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( x>=zero_extd ) then
!
        up = upper
        xabs = x
!
    else
!
        up = .not. upper
        xabs = -x
!
    end if
!
    if ( xabs>c37_extd ) then
!
!        abs( X ) > 37.
!
        probn2_r = merge( one_extd, zero_extd, x>zero_extd )
!
    else
!
!       abs( X ) <= 37.
!
        expntl = exp( -half_extd*xabs*xabs )
        if ( xabs<cutoff ) then
!
!            abs( X ) < CUTOFF = 10/sqrt(2).
!
            probn2_r = expntl*((((((p6*xabs + p5)*              &
                                       xabs + p4)*              &
                                       xabs + p3)*              &
                                       xabs + p2)*              &
                                       xabs + p1)*              &
                                       xabs + p0)  /            &
                              (((((((q7*xabs + q6)*             &
                                       xabs + q5)*              &
                                       xabs + q4)*              &
                                       xabs + q3)*              &
                                       xabs + q2)*              &
                                       xabs + q1)*              &
                                       xabs + q0)
!
        else
!
!           abs( X ) >= CUTOFF.
!
            probn2_r = (expntl/root2pi)/(xabs + one_extd             &
                                       /(xabs + two_extd             &
                                       /(xabs + three_extd           &
                                       /(xabs + four_extd            &
                                       /(xabs + c0_65_extd)))))
!
        end if
!
    end if
!
    if ( .not. up ) probn2_r = one_extd - probn2_r
!
!
! END OF FUNCTION probn2_r
! ________________________
!
    end function probn2_r
!
! =========================================================================================
!
    function probn2_rv( x, upper )
!
! Purpose
! _______
!
!   Evaluates the standard normal (Gaussian) distribution function from X(i) to infinity 
!   if UPPER is true or from minus infinity to X(i) if UPPER is false, for i=1 to
!   size(X). In other words, if:
!
!   - UPPER = true  :    PROBN2( i ) = prob( U > X(i) )  ,
!   - UPPER = false :    PROBN2( i ) = prob( U < X(i) )  ,
!   
!   , for U = Laplace_Gauss(0;1) and i=1 to size(X).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(extd), dimension(:)
!          On entry, upper or lower limits of integration.
!
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  : probability to the right of X is calculated.
!          - UPPER = false : probability to the left of X is calculated.
!
!
! Further Details
! _______________
!
!   The subroutine is parallelized if OPENMP is used.
!
!   This function gives higher accuracy than PROBN.
!
!   This function is based upon algorithm 5666 for the error function, from:
!
!   (1)   Hart, J.F. et al, 1968: Computer Approximations, 354 pp, New York, Wiley.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero_extd, one_extd, two_extd, three_extd,    &
                                  four_extd, half_extd, c37_extd, c0_65_extd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), dimension(:),       intent(in) :: x
    real(extd), dimension(size(x))             :: probn2_rv
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)                     :: xabs, expntl, xi
    real(extd), dimension(size(x)) :: probvec
!
    logical(lgl) :: up
!   
    integer(i4b) :: n, i
!
#ifdef _OPENMP
!   
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(extd), parameter :: cutoff=7.071_extd,               &
                             root2pi=2.506628274631001_extd
!
    real(extd), parameter ::                                  &
                             p0=220.2068679123761_extd,       &
                             p1=221.2135961699311_extd,       &
                             p2=112.0792914978709_extd,       &
                             p3=33.91286607838300_extd,       &
                             p4=6.373962203531650_extd,       &
                             p5=.7003830644436881_extd,       &
                             p6=.3526249659989109e-01_extd
!
    real(extd), parameter ::                                  &
                             q0=440.4137358247522_extd,       &
                             q1=793.8265125199484_extd,       &
                             q2=637.3336333788311_extd,       &
                             q3=296.5642487796737_extd,       &
                             q4=86.78073220294608_extd,       &
                             q5=16.06417757920695_extd,       &
                             q6=1.755667163182642_extd,       &
                             q7=.8838834764831844e-1_extd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)             &
!$OMP            ,PRIVATE(i,up,xabs,expntl,xi)
!
    do i = 1_i4b, n
!
        xi = x(i)
!
        if ( xi>=zero_extd ) then
!
            up   = upper
            xabs = xi
!
        else
!
            up   = .not. upper
            xabs = -xi
!
        end if
!
        if ( xabs>c37_extd ) then
!
!           abs( X ) > 37.
!
            if ( xi>zero_extd ) then
                xi = one_extd
            else
                xi = zero_extd
            end if
!
        else
!
!           abs( X ) <= 37.
!
            expntl = exp( -half_extd*xabs*xabs )
!
            if ( xabs<cutoff ) then
!
                xi = expntl*((((((p6*xabs + p5)*                &
                                     xabs + p4)*                &
                                     xabs + p3)*                &
                                     xabs + p2)*                &
                                     xabs + p1)*                &
                                     xabs + p0)  /              &
                           (((((((q7*xabs + q6)*                &
                                     xabs + q5)*                &
                                     xabs + q4)*                &
                                     xabs + q3)*                &
                                     xabs + q2)*                &
                                     xabs + q1)*                &
                                     xabs + q0)
!
            else
!
                xi = (expntl/root2pi)/(xabs + one_extd          &
                                     /(xabs + two_extd          &
                                     /(xabs + three_extd        &
                                     /(xabs + four_extd         &
                                     /(xabs + c0_65_extd)))))
!
            end if
!
        end if
!
        if ( .not. up ) xi = one_extd - xi
!
        probvec(i) = xi
!
    end do
!
!$OMP END PARALLEL DO
!
    probn2_rv(1_i4b:n) = probvec(1_i4b:n)
!
!
! END OF FUNCTION probn2_rv
! _________________________
!
    end function probn2_rv
!
! =========================================================================================
!
    function probn2_rm( x, upper )
!
! Purpose
! _______
!
!   Evaluates the standard normal (Gaussian) distribution function from X(i,j) to infinity 
!   if UPPER is true or from minus infinity to X(i,j) if UPPER is false, for i=1 to
!   size(X,1) and j=1 to size(X,2). In other words, if:
!
!   - UPPER = true  :    PROBN2( i, j ) = prob( U > X(i,j) )  ,
!   - UPPER = false :    PROBN2( i, j ) = prob( U < X(i,j) )  ,
!   
!   , for U = Laplace_Gauss(0;1) and i=1 to size(X,1) and j=1 to size(X,2).
!
!
! Arguments
! _________
!
!   X      (INPUT) real(extd), dimension(:,:)
!          On entry, upper or lower limits of integration.
!
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  : probability to the right of X is calculated.
!          - UPPER = false : probability to the left of X is calculated.
!
!
! Further Details
! _______________
!
!   The subroutine is parallelized if OPENMP is used.
!
!   This function gives higher accuracy than PROBN.
!
!   This function is based upon algorithm 5666 for the error function, from:
!
!   (1)   Hart, J.F. et al, 1968: Computer Approximations, 354 pp, New York, Wiley.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero_extd, one_extd, two_extd, three_extd,    &
                                  four_extd, half_extd, c37_extd, c0_65_extd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), dimension(:,:),                 intent(in) :: x
    real(extd), dimension(size(x,1),size(x,2))             :: probn2_rm
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)                                 :: xabs, expntl, xji
    real(extd), dimension(size(x,1),size(x,2)) :: probmat
!
    logical(lgl) :: up
!   
    integer(i4b) :: n, m, i, j
!
#ifdef _OPENMP
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(extd), parameter :: cutoff=7.071_extd,               &
                             root2pi=2.506628274631001_extd
!
    real(extd), parameter ::                                  &
                             p0=220.2068679123761_extd,       &
                             p1=221.2135961699311_extd,       &
                             p2=112.0792914978709_extd,       &
                             p3=33.91286607838300_extd,       &
                             p4=6.373962203531650_extd,       &
                             p5=.7003830644436881_extd,       &
                             p6=.3526249659989109e-01_extd
!
    real(extd), parameter ::                                  &
                             q0=440.4137358247522_extd,       &
                             q1=793.8265125199484_extd,       &
                             q2=637.3336333788311_extd,       &
                             q3=296.5642487796737_extd,       &
                             q4=86.78073220294608_extd,       &
                             q5=16.06417757920695_extd,       &
                             q6=1.755667163182642_extd,       &
                             q7=.8838834764831844e-1_extd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    m = size( x, 1 )
    n = size( x, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)             &
!$OMP            ,PRIVATE(i,j,up,xabs,expntl,xji)
!
    do i = 1_i4b, n
!
        do j = 1_i4b, m
!
            xji = x(j,i)
!
            if ( xji>=zero_extd ) then
                up   = upper
                xabs = xji
            else
                up   = .not. upper
                xabs = -xji
            end if
!
            if ( xabs>c37_extd ) then
!
!               abs( X ) > 37.
!
                if ( xji>zero_extd ) then
                    xji = one_extd
                else
                    xji = zero_extd
                end if
!
            else
!
!               abs( X ) <= 37.
!
                expntl = exp( -half_extd*xabs*xabs )
!
                if ( xabs<cutoff ) then
!
                    xji = expntl*((((((p6*xabs + p5)*                &
                                          xabs + p4)*                &
                                          xabs + p3)*                &
                                          xabs + p2)*                &
                                          xabs + p1)*                &
                                          xabs + p0)  /              &
                                (((((((q7*xabs + q6)*                &
                                          xabs + q5)*                &
                                          xabs + q4)*                &
                                          xabs + q3)*                &
                                          xabs + q2)*                &
                                          xabs + q1)*                &
                                          xabs + q0)
!
                else
!
                    xji = (expntl/root2pi)/(xabs + one_extd          &
                                          /(xabs + two_extd          &
                                          /(xabs + three_extd        &
                                          /(xabs + four_extd         &
                                          /(xabs + c0_65_extd)))))
!
                end if
!
            end if
!
            if ( .not. up ) xji = one_extd - xji
!
            probmat(j,i) = xji
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probn2_rm(1_i4b:m,1_i4b:n) = probmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION probn2_rm
! _________________________
!
    end function probn2_rm
!
!
! =========================================================================================
!                    INVERSE OF THE GAUSSIAN DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function pinvn_r( p )
!
! Purpose
! _______
!
!   Evaluates the inverse of the standard normal (Gaussian) distribution function:
!
!                           X0 = PINVN( P ) 
!   
!   , if P = probability( U < X0 )  for U = Laplace_Gauss(0;1).
!
!   PINVN returns the normal deviate X0 corresponding to a given lower tail area of P.
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd)
!          On entry, the probability. P must verify 0. < P < 1.
!
!
! Further Details
! _______________
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 2 and 3) by the subroutine PPND7 described in the reference (1).
!
!   This function is accurate to about seven decimal figures for min(P,1-P) > 10**(-316).
!
!   If P is very close to unity, a serious loss of significance may be incurred in forming
!   1 - P = c in the code of the function. In this circumstance the user should, if possible, evaluate c directly
!   or in extended precision and evaluate X0(P) = PINVN( P ) as -X0(c) = PINVN( c ).
!
!   The hash sums below are the sums of the mantissas of the coefficients.
!   They are included for use in checking transcription.
!
!   This function is adapted from the routine PPND7 described in
!
!   (1)    Wichura, M.J., 1988: Algorithm AS 241: The percentage points of the normal
!                                 distribution.  Appl. Statist., Vol. 37, No. 3, pp. 477-484.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, half
    use Char_Constants,    only : prob_error1
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: p
    real(stnd)             :: pinvn_r
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: q, r
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::   split1 = 0.425_stnd, split2 = 5._stnd,          &
                               const1 = 0.180625_stnd, const2 = 1.6_stnd
!   
!   Coefficients for P close to 0.5.
!   
    real(stnd), parameter :: a0 = 3.3871327179e+00_stnd, &
                             a1 = 5.0434271938e+01_stnd, &
                             a2 = 1.5929113202e+02_stnd, &
                             a3 = 5.9109374720e+01_stnd, &
                             b1 = 1.7895169469e+01_stnd, &
                             b2 = 7.8757757664e+01_stnd, &
                             b3 = 6.7187563600e+01_stnd
! HASH SUM AB                    32.3184577772
!   
!   Coefficients for P not close to 0, 0.5 or 1.
!   

    real(stnd), parameter :: c0 = 1.4234372777e+00_stnd, &
                             c1 = 2.7568153900e+00_stnd, &
                             c2 = 1.3067284816e+00_stnd, &
                             c3 = 1.7023821103e-01_stnd, &
                             d1 = 7.3700164250e-01_stnd, &
                             d2 = 1.2021132975e-01_stnd
! HASH SUM CD                    15.7614929821
!   
!   Coefficients for P near 0 or 1.
!   
    real(stnd), parameter :: e0 = 6.6579051150e+00_stnd, &
                             e1 = 3.0812263860e+00_stnd, &
                             e2 = 4.2868294337e-01_stnd, &
                             e3 = 1.7337203997e-02_stnd, &
                             f1 = 2.4197894225e-01_stnd, &
                             f2 = 1.2258202635e-02_stnd
! HASH SUM EF                    19.4052910204
!
    character(len=*),  parameter :: name_proc='pinvn'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( zero>=p .or. p>=one )  call merror( name_proc//prob_error1 )
!
    q = p - half
!
    if ( abs(q)<=split1 ) then
!
        r = const1 - q*q
        pinvn_r = q*(((a3*r + a2)*r + a1)*r + a0) /(((b3*r + b2)*r + b1)*r + one)
!
    else
!
!        r = merge( p, one-p, q<zero )
        if ( q<zero ) then
!
            r = p
!
        else
!
            r = one - p
!
        end if
!
        r = sqrt( -log(r) )
!
        if ( r<=split2 ) then
!
            r = r - const2
            pinvn_r = (((c3*r + c2)*r + c1)*r + c0)/((d2*r + d1)*r + one)
!
        else
!
            r = r - split2
            pinvn_r = (((e3*r + e2)*r + e1)*r + e0)/((f2*r + f1)*r + one)
!
        end if
!
        if ( q<zero ) pinvn_r = -pinvn_r
!
    end if
!
!
! END OF FUNCTION pinvn_r
! _______________________
!
    end function pinvn_r
!
! =========================================================================================
!
    function pinvn_rv( p )
!
! Purpose
! _______
!
!   Evaluates the inverse of the standard normal (Gaussian) distribution function:
!
!                         X0(i) = PINVN( P(i) ) 
!   
!   , if P(i) = probability( U < X0(i) )  for U = Laplace_Gauss(0;1) and  i=1 to size(P).
!
!   PINVN returns the normal deviate X0(i) corresponding to a given lower tail area 
!   of P(i), for i=1 to size(P).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:)
!          On entry, the probabilities. P(i) must verify 0. < P(i) < 1,
!          for i=1 to size(P) .
!
!
! Further Details
! _______________
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 2 and 3) by the subroutine PPND7 described in the reference (1).
!
!   This function is accurate to about seven decimal figures for min(P,1-P) > 10**(-316).
!
!   If P is very close to unity, a serious loss of significance may be incurred in forming
!   1 - P = c in the code of the function. In this circumstance the user should, if possible, evaluate c directly
!   or in extended precision and evaluate X0(P) = PINVN( P ) as -X0(c) = PINVN( c ).
!
!   The subroutine is parallelized if OPENMP is used.
!
!   The hash sums below are the sums of the mantissas of the coefficients.
!   They are included for use in checking transcription.
!
!   This function is adapted from the routine PPND7 described in
!
!   (1)    Wichura, M.J., 1988: Algorithm AS 241: The percentage points of the normal
!                                 distribution.  Appl. Statist., Vol. 37, No. 3, pp. 477-484.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, half
    use Char_Constants,    only : prob_error1
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),      intent(in) :: p
    real(stnd), dimension(size(p))            :: pinvn_rv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: pinv, q, r, pi
    real(stnd), dimension(size(p)) :: pinvvec
!   
    integer(i4b) :: n, i
!
#ifdef _OPENMP
!   
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::   split1 = 0.425_stnd, split2 = 5._stnd,          &
                               const1 = 0.180625_stnd, const2 = 1.6_stnd
!   
!   Coefficients for P close to 0.5.
!   
    real(stnd), parameter :: a0 = 3.3871327179e+00_stnd, &
                             a1 = 5.0434271938e+01_stnd, &
                             a2 = 1.5929113202e+02_stnd, &
                             a3 = 5.9109374720e+01_stnd, &
                             b1 = 1.7895169469e+01_stnd, &
                             b2 = 7.8757757664e+01_stnd, &
                             b3 = 6.7187563600e+01_stnd
! HASH SUM AB                    32.3184577772
!   
!   Coefficients for P not close to 0, 0.5 or 1.
!   

    real(stnd), parameter :: c0 = 1.4234372777e+00_stnd, &
                             c1 = 2.7568153900e+00_stnd, &
                             c2 = 1.3067284816e+00_stnd, &
                             c3 = 1.7023821103e-01_stnd, &
                             d1 = 7.3700164250e-01_stnd, &
                             d2 = 1.2021132975e-01_stnd
! HASH SUM CD                    15.7614929821
!   
!   Coefficients for P near 0 or 1.
!   
    real(stnd), parameter :: e0 = 6.6579051150e+00_stnd, &
                             e1 = 3.0812263860e+00_stnd, &
                             e2 = 4.2868294337e-01_stnd, &
                             e3 = 1.7337203997e-02_stnd, &
                             f1 = 2.4197894225e-01_stnd, &
                             f2 = 1.2258202635e-02_stnd
! HASH SUM EF                    19.4052910204
!
    character(len=*),  parameter :: name_proc='pinvn'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( p )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   TEST THE ARGUMENTS.
!
    if ( any(zero>=p(:n)) .or. any(p(:n)>=one) )    &
    call merror( name_proc//prob_error1 )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i,pinv,q,r,pi)
!
    do i = 1_i4b, n
!
        pi = p(i)
        q  = pi - half
!
        if ( abs(q)<=split1 ) then
!
            r    = const1 - q*q
            pinv = q*(((a3*r + a2)*r + a1)*r + a0) /(((b3*r + b2)*r + b1)*r + one)
!
        else
!
!            r = merge( pi, one-pi, q<zero )
            if ( q<zero ) then
                r = pi
            else
                r = one - pi
            end if
!
            r = sqrt( -log(r) )
!
            if ( r<=split2 ) then
!
                r    = r - const2
                pinv = (((c3*r + c2)*r + c1)*r + c0)/((d2*r + d1)*r + one)
!
            else
!
                r    = r - split2
                pinv = (((e3*r + e2)*r + e1)*r + e0)/((f2*r + f1)*r + one)
!
            end if
!
            if ( q<zero ) then
!
                pinv = -pinv
!
            end if
!
        end if
!
        pinvvec(i) = pinv
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvn_rv(1_i4b:n) = pinvvec(1_i4b:n)
!
!
! END OF FUNCTION pinvn_rv
! ________________________
!
    end function pinvn_rv
!
! =========================================================================================
!
    function pinvn_rm( p )
!
!
! Purpose
! _______
!
!   Evaluates the inverse of the standard normal (Gaussian) distribution function:
!
!              X0(i,j) = PINVN( P(i,j) )
!   
!   , if P(i,j) = probability( U < X0(i,j) )  for U = Laplace_Gauss(0;1), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVN returns the normal deviate X0(i,j) corresponding to a given lower tail area 
!   of P(i,j) for i=1 to size(P,1) and j=1 to size(P,2).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:,:)
!          On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1,
!          for i=1 to size(P,1) and j=1 to size(P,2) .
!
!
! Further Details
! _______________
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 2 and 3) by the subroutine PPND7 described in the reference (1).
!
!   This function is accurate to about seven decimal figures for min(P,1-P) > 10**(-316).
!
!   If P is very close to unity, a serious loss of significance may be incurred in forming
!   1 - P = c in the code of the function. In this circumstance the user should, if possible, evaluate c directly
!   or in extended precision and evaluate X0(P) = PINVN( P ) as -X0(c) = PINVN( c ).
!
!   The subroutine is parallelized if OPENMP is used.
!
!   The hash sums below are the sums of the mantissas of the coefficients.
!   They are included for use in checking transcription.
!
!   This function is adapted from the routine PPND7 described in
!
!   (1)    Wichura, M.J., 1988: Algorithm AS 241: The percentage points of the normal
!                                 distribution.  Appl. Statist., Vol. 37, No. 3, pp. 477-484.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, half
    use Char_Constants,    only : prob_error1
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                intent(in) :: p
    real(stnd), dimension(size(p,1),size(p,2))            :: pinvn_rm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: pinv, q, r, pji
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvmat
!   
    integer(i4b) :: n, m, i, j
!
#ifdef _OPENMP
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::   split1 = 0.425_stnd, split2 = 5._stnd,          &
                               const1 = 0.180625_stnd, const2 = 1.6_stnd
!   
!    Coefficients for P close to 0.5.
!   
    real(stnd), parameter :: a0 = 3.3871327179e+00_stnd, &
                             a1 = 5.0434271938e+01_stnd, &
                             a2 = 1.5929113202e+02_stnd, &
                             a3 = 5.9109374720e+01_stnd, &
                             b1 = 1.7895169469e+01_stnd, &
                             b2 = 7.8757757664e+01_stnd, &
                             b3 = 6.7187563600e+01_stnd
! HASH SUM AB                    32.3184577772
!   
!    Coefficients for P not close to 0, 0.5 or 1.
!   

    real(stnd), parameter :: c0 = 1.4234372777e+00_stnd, &
                             c1 = 2.7568153900e+00_stnd, &
                             c2 = 1.3067284816e+00_stnd, &
                             c3 = 1.7023821103e-01_stnd, &
                             d1 = 7.3700164250e-01_stnd, &
                             d2 = 1.2021132975e-01_stnd
! HASH SUM CD                    15.7614929821
!   
!    Coefficients for P near 0 or 1.
!   
    real(stnd), parameter :: e0 = 6.6579051150e+00_stnd, &
                             e1 = 3.0812263860e+00_stnd, &
                             e2 = 4.2868294337e-01_stnd, &
                             e3 = 1.7337203997e-02_stnd, &
                             f1 = 2.4197894225e-01_stnd, &
                             f2 = 1.2258202635e-02_stnd
! HASH SUM EF                    19.4052910204
!
    character(len=*),  parameter :: name_proc='pinvn'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( p, 1 )
    n = size( p, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) return
!
!   TEST THE ARGUMENTS.
!
    if ( any(zero>=p(:m,:n)) .or. any(p(:m,:n)>=one) )  &
    call merror( name_proc//prob_error1 )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i,j,pinv,q,r,pji)
!
    do i = 1_i4b, n
!
        do j = 1_i4b, m
!
            pji = p(j,i)
            q   = pji - half
!
            if ( abs(q)<=split1 ) then
!
                r    = const1 - q*q
                pinv = q*(((a3*r + a2)*r + a1)*r + a0) /(((b3*r + b2)*r + b1)*r + one)
!
            else
!
!                r = merge( pji, one-pji, q<zero )
                if ( q<zero ) then
                    r = pji
                else
                    r = one - pji
                end if
!
                r = sqrt( -log(r) )
!
                if ( r<=split2 ) then
!
                    r    = r - const2
                    pinv = (((c3*r + c2)*r + c1)*r + c0)/((d2*r + d1)*r + one)
!
                else
!
                    r    = r - split2
                    pinv = (((e3*r + e2)*r + e1)*r + e0)/((f2*r + f1)*r + one)
!
                end if
!
                if ( q<zero ) then
!
                    pinv = -pinv
!
                end if
!
            end if
!
            pinvmat(j,i) = pinv
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvn_rm(1_i4b:m,1_i4b:n) = pinvmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION pinvn_rm
! ________________________
!
    end function pinvn_rm
!
! =========================================================================================
!
    function pinvn2_r( p )
!
! Purpose
! _______
!
!   Evaluates the inverse of the standard normal (Gaussian) distribution function:
!
!                           X0 = PINVN2( P ) 
!   
!   , if P = probability( U < X0 )  for U = Laplace_Gauss(0;1).
!
!   PINVN2 returns the normal deviate X0 corresponding to a given lower tail area of P.
!
!
! Arguments
! _________
!
!   P      (INPUT) real(extd)
!          On entry, the probability. P must verify 0. < P < 1.
!
!
! Further Details
! _______________
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 7) by the subroutine PPND16 described in the reference (1).
!
!   This function is accurate to about 16 decimal figures for min(P,1-P) > 10**(-316) and gives higher
!   accuracy than PINVN function, but it is slower.
!
!   On a machine, that uses only 32 bits to represent real variables, PINVN2 should be implemented in double
!   precision (e.g. with a correct choice of the kind EXTD in the module Select_Parameters ).
!
!   If P is very close to unity, a serious loss of significance may be incurred in forming
!   1 - P = c in the code of the function. In this circumstance the user should, if possible, evaluate c directly
!   or in extended precision and evaluate X0(P) = PINVN2( P ) as -X0(c) = PINVN2( c ) .
!
!   The hash sums below are the sums of the mantissas of the coefficients.
!   They are included for use in checking transcription.
!
!   This function is adapted from the routine PPND16 described in
!
!   (1)    Wichura, M.J., 1988: Algorithm AS 241: The percentage points of the normal
!                             distribution.  Appl. Statist., Vol. 37, No. 3, pp. 477-484.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero_extd, one_extd, half_extd
    use Char_Constants,    only : prob_error1
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), intent(in) :: p
    real(extd)             :: pinvn2_r
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd) :: q, r
!   
!   
! PARAMETERS 
! __________
!
    real(extd), parameter ::   split1 = 0.425_extd, split2 = 5._extd,          &
                               const1 = 0.180625_extd, const2 = 1.6_extd
!   
!   Coefficients for P close to 0.5.
!   
    real(extd), parameter :: a0 = 3.3871328727963666080_extd,    &
                             a1 = 1.3314166789178437745e+2_extd, &
                             a2 = 1.9715909503065514427e+3_extd, &
                             a3 = 1.3731693765509461125e+4_extd, &
                             a4 = 4.5921953931549871457e+4_extd, &
                             a5 = 6.7265770927008700853e+4_extd, &
                             a6 = 3.3430575583588128105e+4_extd, &
                             a7 = 2.5090809287301226727e+3_extd, &
                             b1 = 4.2313330701600911252e+1_extd, &
                             b2 = 6.8718700749205790830e+2_extd, &
                             b3 = 5.3941960214247511077e+3_extd, &
                             b4 = 2.1213794301586595867e+4_extd, &
                             b5 = 3.9307895800092710610e+4_extd, &
                             b6 = 2.8729085735721942674e+4_extd, &
                             b7 = 5.2264952788528545610e+3_extd
! HASH SUM AB                    55.8831928806149014439
!   
!   Coefficients for P not close to 0, 0.5 or 1.
!   
    real(extd), parameter :: c0 = 1.42343711074968357734_extd,    &
                             c1 = 4.63033784615654529590_extd,    &
                             c2 = 5.76949722146069140550_extd,    &
                             c3 = 3.64784832476320460504_extd,    &
                             c4 = 1.27045825245236838258_extd,    &
                             c5 = 2.41780725177450611770e-1_extd, &
                             c6 = 2.27238449892691845833e-2_extd, &
                             c7 = 7.74545014278341407640e-4_extd, &
                             d1 = 2.05319162663775882187_extd,    &
                             d2 = 1.67638483018380384940_extd,    &
                             d3 = 6.89767334985100004550e-1_extd, &
                             d4 = 1.48103976427480074590e-1_extd, &
                             d5 = 1.51986665636164571966e-2_extd, &
                             d6 = 5.47593808499534494600e-4_extd, &
                             d7 = 1.05075007164441684324e-9_extd
! HASH SUM CD                    49.33206503301610289036
!   
!   Coefficients for P near 0 or 1.
!   
    real(extd), parameter :: e0 = 6.65790464350110377720_extd,    &
                             e1 = 5.46378491116411436990_extd,    &
                             e2 = 1.78482653991729133580_extd,    &
                             e3 = 2.96560571828504891230e-1_extd, &
                             e4 = 2.65321895265761230930e-2_extd, &
                             e5 = 1.24266094738807843860e-3_extd, &
                             e6 = 2.71155556874348757815e-5_extd, &
                             e7 = 2.01033439929228813265e-7_extd, &
                             f1 = 5.99832206555887937690e-1_extd, &
                             f2 = 1.36929880922735805310e-1_extd, &
                             f3 = 1.48753612908506148525e-2_extd, &
                             f4 = 7.86869131145613259100e-4_extd, &
                             f5 = 1.84631831751005468180e-5_extd, &
                             f6 = 1.42151175831644588870e-7_extd, &
                             f7 = 2.04426310338993978564e-15_extd
! HASH SUM EF                    47.52583317549289671629
!
    character(len=*),  parameter :: name_proc='pinvn2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( zero_extd>=p .or. p>=one_extd ) call merror( name_proc//prob_error1 )
!
    q = p - half_extd
!
    if ( abs(q)<=split1 ) then
!
        r = const1 - q*q
        pinvn2_r = q * (((((((a7*r + a6)*r + a5)*r + a4)*r + a3)*r + a2)*r + a1)*r + a0) /      &
                       (((((((b7*r + b6)*r + b5)*r + b4)*r + b3)*r + b2)*r + b1)*r + one_extd)
!
    else
!
!        r = merge( p, one_extd-p, q<zero_extd )
        if ( q<zero_extd ) then
            r = p
        else
            r = one_extd - p
        end if
!
        r = sqrt( -log(r) )
!
        if ( r<=split2 ) then
!
            r = r - const2
            pinvn2_r = (((((((c7*r + c6)*r + c5)*r + c4)*r + c3)*r + c2)*r + c1)*r + c0) /       &
                       (((((((d7*r + d6)*r + d5)*r + d4)*r + d3)*r + d2)*r + d1)*r + one_extd)
!
        else
!
            r = r - split2
            pinvn2_r = (((((((e7*r + e6)*r + e5)*r + e4)*r + e3)*r + e2)*r + e1)*r + e0) /       &
                       (((((((f7*r + f6)*r + f5)*r + f4)*r + f3)*r + f2)*r + f1)*r + one_extd)
!
        end if
!
        if ( q<zero_extd ) pinvn2_r = -pinvn2_r
!
    end if
!
!
! END OF FUNCTION pinvn2_r
! ________________________
!
    end function pinvn2_r
!
! =========================================================================================
!
    function pinvn2_rv( p )
!
! Purpose
! _______
!
!   Evaluates the inverse of the standard normal (Gaussian) distribution function:
!
!                         X0(i) = PINVN2( P(i) ) 
!   
!   , if P(i) = probability( U < X0(i) )  for U = Laplace_Gauss(0;1) and i=1 to size(P).
!
!   PINVN returns the normal deviate X0(i) corresponding to a given lower tail area 
!   of P(i), for i=1 to size(P).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(extd), dimension(:)
!          On entry, the probabilities. P(i) must verify 0. < P(i) < 1,
!          for i=1 to size(P) .
!
!
! Further Details
! _______________
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 7) by the subroutine PPND16 described in the reference (1).
!
!   This function is accurate to about 16 decimal figures for min(P,1-P) > 10**(-316) and gives higher
!   accuracy than PINVN function, but it is slower.
!
!   On a machine, that uses only 32 bits to represent real variables, PINVN2 should be implemented in double
!   precision (e.g. with a correct choice of the kind EXTD in the module Select_Parameters ).
!
!   If P is very close to unity, a serious loss of significance may be incurred in forming
!   1 - P = c in the code of the function. In this circumstance the user should, if possible, evaluate c directly
!   or in extended precision and evaluate X0(P) = PINVN2( P ) as -X0(c) = PINVN2( c ).
!
!   The subroutine is parallelized if OPENMP is used.
!
!   The hash sums below are the sums of the mantissas of the coefficients.
!   They are included for use in checking transcription.
!
!   This function is adapted from the routine PPND16 described in
!
!   (1)    Wichura, M.J.,  1988: Algorithm AS 241: The percentage points of the normal
!                             distribution.  Appl. Statist., Vol. 37, No. 3, pp. 477-484.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero_extd, one_extd, half_extd
    use Char_Constants,    only : prob_error1
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), dimension(:),      intent(in) :: p
    real(extd), dimension(size(p))            :: pinvn2_rv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)                     :: pinv, q, r, pi
    real(extd), dimension(size(p)) :: pinvvec
!   
    integer(i4b) :: n, i
!
#ifdef _OPENMP
!   
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(extd), parameter ::   split1 = 0.425_extd, split2 = 5._extd,          &
                               const1 = 0.180625_extd, const2 = 1.6_extd
!   
!   Coefficients for P close to 0.5
!   
    real(extd), parameter :: a0 = 3.3871328727963666080_extd,    &
                             a1 = 1.3314166789178437745e+2_extd, &
                             a2 = 1.9715909503065514427e+3_extd, &
                             a3 = 1.3731693765509461125e+4_extd, &
                             a4 = 4.5921953931549871457e+4_extd, &
                             a5 = 6.7265770927008700853e+4_extd, &
                             a6 = 3.3430575583588128105e+4_extd, &
                             a7 = 2.5090809287301226727e+3_extd, &
                             b1 = 4.2313330701600911252e+1_extd, &
                             b2 = 6.8718700749205790830e+2_extd, &
                             b3 = 5.3941960214247511077e+3_extd, &
                             b4 = 2.1213794301586595867e+4_extd, &
                             b5 = 3.9307895800092710610e+4_extd, &
                             b6 = 2.8729085735721942674e+4_extd, &
                             b7 = 5.2264952788528545610e+3_extd
! HASH SUM AB                    55.8831928806149014439
!   
!   Coefficients for P not close to 0, 0.5 or 1.
!   
    real(extd), parameter :: c0 = 1.42343711074968357734_extd,    &
                             c1 = 4.63033784615654529590_extd,    &
                             c2 = 5.76949722146069140550_extd,    &
                             c3 = 3.64784832476320460504_extd,    &
                             c4 = 1.27045825245236838258_extd,    &
                             c5 = 2.41780725177450611770e-1_extd, &
                             c6 = 2.27238449892691845833e-2_extd, &
                             c7 = 7.74545014278341407640e-4_extd, &
                             d1 = 2.05319162663775882187_extd,    &
                             d2 = 1.67638483018380384940_extd,    &
                             d3 = 6.89767334985100004550e-1_extd, &
                             d4 = 1.48103976427480074590e-1_extd, &
                             d5 = 1.51986665636164571966e-2_extd, &
                             d6 = 5.47593808499534494600e-4_extd, &
                             d7 = 1.05075007164441684324e-9_extd
! HASH SUM CD                    49.33206503301610289036
!   
!   Coefficients for P near 0 or 1.
!   
    real(extd), parameter :: e0 = 6.65790464350110377720_extd,    &
                             e1 = 5.46378491116411436990_extd,    &
                             e2 = 1.78482653991729133580_extd,    &
                             e3 = 2.96560571828504891230e-1_extd, &
                             e4 = 2.65321895265761230930e-2_extd, &
                             e5 = 1.24266094738807843860e-3_extd, &
                             e6 = 2.71155556874348757815e-5_extd, &
                             e7 = 2.01033439929228813265e-7_extd, &
                             f1 = 5.99832206555887937690e-1_extd, &
                             f2 = 1.36929880922735805310e-1_extd, &
                             f3 = 1.48753612908506148525e-2_extd, &
                             f4 = 7.86869131145613259100e-4_extd, &
                             f5 = 1.84631831751005468180e-5_extd, &
                             f6 = 1.42151175831644588870e-7_extd, &
                             f7 = 2.04426310338993978564e-15_extd
! HASH SUM EF                    47.52583317549289671629
!
    character(len=*),  parameter :: name_proc='pinvn2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( p )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   TEST THE ARGUMENTS.
!
    if ( any(zero_extd>=p(:n)) .or. any(p(:n)>=one_extd) )  &
    call merror( name_proc//prob_error1 )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i,pinv,q,r,pi)
!
    do i = 1_i4b, n
!
        pi = p(i)
        q  = pi - half_extd
!
        if ( abs(q)<=split1 ) then
!
            r    = const1 - q*q
            pinv = q * (((((((a7*r + a6)*r + a5)*r + a4)*r + a3)*r + a2)*r + a1)*r + a0) /      &
                       (((((((b7*r + b6)*r + b5)*r + b4)*r + b3)*r + b2)*r + b1)*r + one_extd)
!
        else
!
!            r = merge( pi, one_extd-pi, q<zero_extd )
            if ( q<zero_extd ) then
                r = pi
            else
                r = one_extd - pi
            end if
!
            r = sqrt( -log(r) )
!
           if ( r<=split2 ) then
!
                r    = r - const2
                pinv = (((((((c7*r + c6)*r + c5)*r + c4)*r + c3)*r + c2)*r + c1)*r + c0) /        &
                       (((((((d7*r + d6)*r + d5)*r + d4)*r + d3)*r + d2)*r + d1)*r + one_extd)
!
            else
!
                r    = r - split2
                pinv = (((((((e7*r + e6)*r + e5)*r + e4)*r + e3)*r + e2)*r + e1)*r + e0) /        &
                       (((((((f7*r + f6)*r + f5)*r + f4)*r + f3)*r + f2)*r + f1)*r + one_extd)
!
            end if
!
            if ( q<zero_extd ) then
!
                pinv = -pinv
!
            end if
!
        end if
!
        pinvvec(i) = pinv
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvn2_rv(1_i4b:n) = pinvvec(1_i4b:n)
!
!
! END OF FUNCTION pinvn2_rv
! _________________________
!
    end function pinvn2_rv
!
! =========================================================================================
!
    function pinvn2_rm( p )
!
! Purpose
! _______
!
!   Evaluates the inverse of the standard normal (Gaussian) distribution function:
!
!              X0(i,j) = PINVN2( P(i,j) )
!   
!   , if P(i,j) = probability( U < X0(i,j) )  for U = Laplace_Gauss(0;1), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVN2 returns the normal deviate X0(i,j) corresponding to a given lower tail area 
!   of P(i,j) for i=1 to size(P,1) and j=1 to size(P,2).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(extd), dimension(:,:)
!          On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1,
!          for i=1 to size(P,1) and j=1 to size(P,2) .
!
!
! Further Details
! _______________
!
!   The inverse Gaussian CDF is approximated to high precision using rational approximations
!   (polynomials with degree 7) by the subroutine PPND16 described in the reference (1).
!
!   This function is accurate to about 16 decimal figures for min(P,1-P) > 10**(-316) and gives higher
!   accuracy than PINVN function, but it is slower.
!
!   On a machine, that uses only 32 bits to represent real variables, PINVN2 should be implemented in double
!   precision (e.g. with a correct choice of the kind EXTD in the module Select_Parameters ).
!
!   If P is very close to unity, a serious loss of significance may be incurred in forming
!   1 - P = c in the code of the function. In this circumstance the user should, if possible, evaluate c directly
!   or in extended precision and evaluate X0(P) = PINVN2( P ) as -X0(c) = PINVN2( c ).
!
!   The subroutine is parallelized if OPENMP is used.
!
!   The hash sums below are the sums of the mantissas of the coefficients.
!   They are included for use in checking transcription.
!
!   This function is adapted from the routine PPND16 described in
!
!   (1)    Wichura, M.J.,  1988: Algorithm AS 241: The percentage points of the normal
!                           distribution.  Appl. Statist., Vol. 37, No. 3, pp. 477-484.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero_extd, one_extd, half_extd
    use Char_Constants,    only : prob_error1
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), dimension(:,:),                intent(in) :: p
    real(extd), dimension(size(p,1),size(p,2))            :: pinvn2_rm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(extd)                                 :: pinv, q, r, pji
    real(extd), dimension(size(p,1),size(p,2)) :: pinvmat
!   
    integer(i4b) :: n, m, i, j
!
#ifdef _OPENMP
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(extd), parameter ::   split1 = 0.425_extd, split2 = 5._extd,          &
                               const1 = 0.180625_extd, const2 = 1.6_extd
!   
!   Coefficients for P close to 0.5
!   
    real(extd), parameter :: a0 = 3.3871328727963666080_extd,    &
                             a1 = 1.3314166789178437745e+2_extd, &
                             a2 = 1.9715909503065514427e+3_extd, &
                             a3 = 1.3731693765509461125e+4_extd, &
                             a4 = 4.5921953931549871457e+4_extd, &
                             a5 = 6.7265770927008700853e+4_extd, &
                             a6 = 3.3430575583588128105e+4_extd, &
                             a7 = 2.5090809287301226727e+3_extd, &
                             b1 = 4.2313330701600911252e+1_extd, &
                             b2 = 6.8718700749205790830e+2_extd, &
                             b3 = 5.3941960214247511077e+3_extd, &
                             b4 = 2.1213794301586595867e+4_extd, &
                             b5 = 3.9307895800092710610e+4_extd, &
                             b6 = 2.8729085735721942674e+4_extd, &
                             b7 = 5.2264952788528545610e+3_extd
! HASH SUM AB                    55.8831928806149014439
!   
!   Coefficients for P not close to 0, 0.5 or 1.
!   
    real(extd), parameter :: c0 = 1.42343711074968357734_extd,    &
                             c1 = 4.63033784615654529590_extd,    &
                             c2 = 5.76949722146069140550_extd,    &
                             c3 = 3.64784832476320460504_extd,    &
                             c4 = 1.27045825245236838258_extd,    &
                             c5 = 2.41780725177450611770e-1_extd, &
                             c6 = 2.27238449892691845833e-2_extd, &
                             c7 = 7.74545014278341407640e-4_extd, &
                             d1 = 2.05319162663775882187_extd,    &
                             d2 = 1.67638483018380384940_extd,    &
                             d3 = 6.89767334985100004550e-1_extd, &
                             d4 = 1.48103976427480074590e-1_extd, &
                             d5 = 1.51986665636164571966e-2_extd, &
                             d6 = 5.47593808499534494600e-4_extd, &
                             d7 = 1.05075007164441684324e-9_extd
! HASH SUM CD                    49.33206503301610289036
!   
!   Coefficients for P near 0 or 1.
!   
    real(extd), parameter :: e0 = 6.65790464350110377720_extd,    &
                             e1 = 5.46378491116411436990_extd,    &
                             e2 = 1.78482653991729133580_extd,    &
                             e3 = 2.96560571828504891230e-1_extd, &
                             e4 = 2.65321895265761230930e-2_extd, &
                             e5 = 1.24266094738807843860e-3_extd, &
                             e6 = 2.71155556874348757815e-5_extd, &
                             e7 = 2.01033439929228813265e-7_extd, &
                             f1 = 5.99832206555887937690e-1_extd, &
                             f2 = 1.36929880922735805310e-1_extd, &
                             f3 = 1.48753612908506148525e-2_extd, &
                             f4 = 7.86869131145613259100e-4_extd, &
                             f5 = 1.84631831751005468180e-5_extd, &
                             f6 = 1.42151175831644588870e-7_extd, &
                             f7 = 2.04426310338993978564e-15_extd
! HASH SUM EF                    47.52583317549289671629
!
    character(len=*),  parameter :: name_proc='pinvn2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( p, 1 )
    n = size( p, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) return
!
!   TEST THE ARGUMENTS.
!
    if ( any(zero_extd>=p(:m,:n)) .or. any(p(:m,:n)>=one_extd) )    &
    call merror( name_proc//prob_error1 )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)     &
!$OMP            ,PRIVATE(i,j,pinv,q,r,pji)
!
    do i = 1_i4b, n
!
        do j = 1_i4b, m
!
            pji = p(j,i)
            q   = pji - half_extd
!
            if ( abs(q)<=split1 ) then
!
                r    = const1 - q*q
                pinv = q * (((((((a7*r + a6)*r + a5)*r + a4)*r + a3)*r + a2)*r + a1)*r + a0) /      &
                           (((((((b7*r + b6)*r + b5)*r + b4)*r + b3)*r + b2)*r + b1)*r + one_extd)
!
            else
!
!                r = merge( pji, one_extd-pji, q<zero_extd )
                if ( q<zero_extd ) then
                    r = pji
                else
                    r = one_extd - pji
                end if
!
                r = sqrt( -log(r) )
!
                if ( r<=split2 ) then
!
                    r    = r - const2
                    pinv = (((((((c7*r + c6)*r + c5)*r + c4)*r + c3)*r + c2)*r + c1)*r + c0) /        &
                           (((((((d7*r + d6)*r + d5)*r + d4)*r + d3)*r + d2)*r + d1)*r + one_extd)
!
                else
!
                    r    = r - split2
                    pinv = (((((((e7*r + e6)*r + e5)*r + e4)*r + e3)*r + e2)*r + e1)*r + e0) /        &
                           (((((((f7*r + f6)*r + f5)*r + f4)*r + f3)*r + f2)*r + f1)*r + one_extd)
!
                end if
!
                if ( q<zero_extd ) then
!
                    pinv = -pinv
!
                end if
!
            end if
!
            pinvmat(j,i) = pinv
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvn2_rm(1_i4b:m,1_i4b:n) = pinvmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION pinvn2_rm
! _________________________
!
    end function pinvn2_rm
!
!
! =========================================================================================
!                            STUDENT'S T-DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function probt_r( t, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the Student's t-distribution function from T to infinity 
!   if UPPER is true or from minus infinity to T if UPPER is false. In otherwords, if:
!
!   - UPPER = true  :    PROBT = prob( U > T )  ,
!   - UPPER = false :    PROBT = prob( U < T )  ,
!   
!   , for U = Student(NDF).
!
!
! Arguments
! _________
!
!   T          (INPUT) real(stnd)
!              On entry, upper or lower limit of integration of the t-density.
!
!   NDF        (INPUT) integer(i4b)
!              On entry, degrees of freedom of the t-distribution. NDF must be greater
!              or equal to 1.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of T is calculated.
!              - UPPER = false : probability to the left of T is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the t_density is integrated.
!              - NDF is greater than NDF_MAX, an asymptotic series is used.
!
!              The default is 20.
!
!
! Further Details
! _______________
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-distribution,
!                        (Algorithm 395), Comm. A.C.M., vol.13, 617-619
!
!   (2)   Cooper, B.E., 1968: The integral of student's t distribution,
!                        (Algorithm AS3), Applied Statistics, vol.17, no.2, 189
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, half, pi,                &
                                  c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror, arth, cumprod
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: t
    real(stnd)             :: probt_r
!
    integer(i4b), intent(in)           :: ndf
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, f, s
!
    integer(i4b) :: i, ndf_max2
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: g1=one/pi
!
    character(len=*),  parameter :: name_proc='probt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( ndf<1_i4b ) then
        call merror( name_proc//prob_error2 )
    end if
!
    if ( present(ndf_max) ) then
!
        ndf_max2 = ndf_max
!
    else
!
        ndf_max2 = 20_i4b
!
    end if
!
    f = real( ndf, stnd )
    a = t*t
!
    if ( ndf<=ndf_max2 ) then
!
!       EXACT COMPUTATION FOR SMALL ndf.
!
        b = f/( f + a )
        a = t/sqrt( f )
!
        if ( ndf>=4_i4b ) then
!
            c = mod( f, two ) + two
            i = ndf/2_i4b - 1_i4b
            s = one + sum( cumprod( b - b/arth( c, two, i ) ) )
!
        else
!
            s = merge( zero, one, ndf==1_i4b )
!
        end if
!
        if ( mod(ndf,2_i4b)/=1_i4b ) then
            s = half*a*sqrt( b )*s
        else
            s = ( a*b*s + atan( a ) )*g1
        end if
!
        if ( upper ) then
            probt_r = half - s
        else
            probt_r = half + s
        end if
!
    else
!
!       ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE ndf.
!
        a = a/f
        b = one + a
        s = b - one
        if ( s/=zero ) a = a*( log( b )/s )
        c = f - half
        f = c48*c*c
        a = c*a
        s = a*a
        b = (((((-c0_4*a - c3_3)*a - c24)*a - c85_5) /                 &
              (c0_8*s + c100 + f) + a +three) / f + one) * sqrt( a )
!
        if ( t<zero ) b = -b
!
        probt_r = probn_r( b, upper )
!
    end if
!
!
! END OF FUNCTION probt_r
! _______________________
!
    end function probt_r
!
! =========================================================================================
!
    function probt_rv( t, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the Student's t-distribution function from T(i) to infinity 
!   if UPPER is true or from minus infinity to T(i) if UPPER is false,
!   for i=1 to size(T). In other words, if:
!
!   - UPPER = true  :    PROBT( i ) = prob( U > T(i) )  ,
!   - UPPER = false :    PROBT( i ) = prob( U < T(i) )  ,
!   
!   , for U = STUDENT(NDF) and i=1 to size(T).
!
!
! Arguments
! _________
!
!   T          (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration of the t-density.
!
!   NDF        (INPUT) integer(i4b)
!              On entry, degrees of freedom of the t-distribution. NDF must be greater
!              or equal to  1.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  :  probabilities to the right of T is calculated.
!              - UPPER = false :  probabilities to the left of T is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the t_density is integrated.
!              - NDF is greater than NDF_MAX, an asymptotic series is used.
!
!              The default is 20.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-distribution,
!                        (Algorithm 395), Comm. A.C.M., vol.13, 617-619
!
!   (2)   Cooper, B.E., 1968: The integral of student's t distribution,
!                        (Algorithm AS3), Applied Statistics, vol.17, no.2, 189
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, half, pi,                &
                                  c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),       intent(in) :: t
    real(stnd), dimension(size(t))             :: probt_rv
!
    integer(i4b), intent(in)           :: ndf
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(max(ndf/2,1)) :: vec
    real(stnd), dimension(size(t))      :: prob
    real(stnd)                          :: a, b, c, f, s, df, ti
!
    integer(i4b) :: i, j, l, n, ndf_max2
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: g1=one/pi
!
    character(len=*),  parameter :: name_proc='probt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( t )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( ndf<1_i4b ) then
        call merror( name_proc//prob_error2 )
    end if
!
    if ( present(ndf_max) ) then
!
        ndf_max2 = ndf_max
!
    else
!
        ndf_max2 = 20_i4b
!
    end if
!
    df = real( ndf, stnd )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,j,l,ti,a,b,c,f,s,vec)
!
    if ( ndf<=ndf_max2 ) then
!
!       EXACT COMPUTATION FOR SMALL ndf.
!
        if ( ndf>=4_i4b ) then
!
            c = mod( df, two ) + two
            j = ndf/2_i4b - 1_i4b
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                ti = t(i)
!
                b = df/( df + ti*ti )
                a = ti/sqrt( df )
!
                vec(1_i4b) = b - b/c
!
                s = c
!
                do l=2_i4b, j
                    s = s + two
                    vec(l) = vec(l-1_i4b)*( b - b/s )              
                end do
!
                s = one + sum( vec(:j) )
!
                if ( mod(ndf,2_i4b)/=1_i4b ) then
                    s = half*a*sqrt( b )*s
                else
                    s = ( a*b*s + atan( a ) )*g1
                end if
!
                if ( upper ) then
                    prob(i) = half - s
                else
                    prob(i) = half + s
                end if
!
            end do
!
!$OMP END DO
!
        else
!
!            c = merge( zero, one, ndf==1_i4b )
            if ( ndf==1_i4b ) then
                c = zero
            else
                c = one
            end if
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                ti = t(i)
!
                a = ti*ti
!
                b = df/( df + a )
                a = ti/sqrt( df )
!
                if ( mod(ndf,2_i4b)/=1_i4b ) then
                    s = half*a*sqrt( b )
                else
                    s = ( a*b*c + atan( a ) )*g1
                end if
!
                if ( upper ) then
                    prob(i) = half - s
                else
                    prob(i) = half + s
                end if
!
            end do
!
!$OMP END DO
!
        end if
!
    else
!
!       ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE ndf.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i=1_i4b, n
!
            ti = t(i)
!
            a = (ti*ti)/df
            b = one + a
            s = b - one
            if ( s/=zero ) a = a*( log( b )/s )
            c = df - half
            f = c48*c*c
            a = c*a
            s = a*a
            b = (((((-c0_4*a - c3_3)*a - c24)*a - c85_5) /                 &
                (c0_8*s + c100 + f) + a +three) / f + one) * sqrt( a )
            if ( ti<zero ) b = -b
!
            prob(i) = probn_r( b, upper )
!
        end do
!
!$OMP END DO
!
    end if
!
!$OMP END PARALLEL
!
    probt_rv(:n) = prob(:n)
!
!
! END OF FUNCTION probt_rv
! ________________________
!
    end function probt_rv
!
! =========================================================================================
!
    function probt_rv2( t, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the Student's t-distribution function from T(i) to infinity 
!   if UPPER is true or from minus infinity to T(i) if UPPER is false,
!   for i=1 to size(T). In other words, if:
!
!   - UPPER = true  :    PROBT( i ) = prob( U > T(i) )  ,
!   - UPPER = false :    PROBT( i ) = prob( U < T(i) )  ,
!   
!   , for U = STUDENT(NDF(i)) and i=1 to size(T).
!
!
! Arguments
! _________
!
!   T          (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration of the t-density.
!
!   NDF        (INPUT) integer(i4b), dimension(:)
!              On entry, degrees of freedom of the t-distribution. Any value in the
!              array NDF must be greater or equal to  1.
!
!              The size of NDF must be  size(NDF) = size(T) .
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  :  probabilities to the right of T is calculated.
!              - UPPER = false  : probabilities to the left of T is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the t_density is integrated.
!              - NDF is greater than NDF_MAX, an asymptotic series is used.
!
!              The default is 20.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-distribution,
!                        (Algorithm 395), Comm. A.C.M., vol.13, 617-619
!
!   (2)   Cooper, B.E., 1968: The integral of student's t distribution,
!                        (Algorithm AS3), Applied Statistics, vol.17, no.2, 189
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, half, pi,                &
                                  c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Char_Constants,    only : prob_error2, allocate_error
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),       intent(in) :: t
    real(stnd), dimension(size(t))             :: probt_rv2
!
    integer(i4b), dimension(:), intent(in)           :: ndf
    integer(i4b),               intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:), allocatable :: vec
    real(stnd), dimension(size(t))        :: prob
    real(stnd)                            :: a, b, c, f, s, ti
!
    integer      :: iok
    integer(i4b) :: i, j, l, n, ndfi, ndf_max2, ndfm
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: g1=one/pi
!
    character(len=*),  parameter :: name_proc='probt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(t),i4b)     ,   &
                    int(size(ndf),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any( ndf(:n)<1_i4b ) ) then
        call merror( name_proc//prob_error2 )
    end if
!
    if ( present(ndf_max) ) then
!
        ndf_max2 = ndf_max
!
    else
!
        ndf_max2 = 20_i4b
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
    ndfm = max( ndf_max2/2_i4b, 1_i4b )
!
!$OMP PARALLEL IF(test_par), PRIVATE(iok,i,j,l,ndfi,ti,a,b,c,f,s,vec)
!
!   ALLOCATE WORK ARRAY.
!
    allocate( vec(ndfm), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!$OMP DO SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        ti   = t(i)
        ndfi = ndf(i)
!
        f = real( ndfi, stnd )
        a = ti*ti
!
        if ( ndfi<=ndf_max2 ) then
!
!           EXACT COMPUTATION FOR SMALL ndf.
!
            b = f/( f + a )
            a = ti/sqrt( f )
!
            if ( ndfi>=4_i4b ) then
!
                c = mod( f, two ) + two
                j = ndfi/2_i4b - 1_i4b
!
                vec(1_i4b) = b - b/c
!
                do l=2_i4b, j
                    c = c + two
                    vec(l) = vec(l-1_i4b)*( b - b/c )              
                end do
!
                s = one + sum( vec(:j) )
!
            else
!
!                s = merge( zero, one, ndfi==1_i4b )
                if ( ndfi==1_i4b ) then
                    s = zero
                else
                    s = one
                end if
!
            end if
!
            if ( mod(ndfi,2_i4b)/=1_i4b ) then
                s = half*a*sqrt( b )*s
            else
                s = ( a*b*s + atan( a ) )*g1
            end if
!
            if ( upper ) then
                prob(i) = half - s
            else
                prob(i) = half + s
            end if
!
        else
!
!           ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE ndf.
!
            a = a/f
            b = one + a
            s = b - one
            if ( s/=zero ) a = a*( log( b )/s )
            c = f - half
            f = c48*c*c
            a = c*a
            s = a*a
            b = (((((-c0_4*a - c3_3)*a - c24)*a - c85_5) /                 &
                  (c0_8*s + c100 + f) + a +three) / f + one) * sqrt( a )
!
            if ( ti<zero ) b = -b
!
            prob(i) = probn_r( b, upper )
!
        end if
!
    end do
!
!$OMP END DO
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( vec )
!
!$OMP END PARALLEL
!
    probt_rv2(:n) = prob(:n)
!
!
! END OF FUNCTION probt_rv2
! _________________________
!
    end function probt_rv2
!
! =========================================================================================
!
    function probt_rm( t, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the Student's t-distribution function from T(i,j) to infinity 
!   if UPPER is true or from minus infinity to T(i,j) if UPPER is false,
!   for i=1 to size(T,1) and j=1 to size(T,2). In other words, if:
!
!   - UPPER = true  :    PROBT( i, j ) = prob( U > T(i,j) )  ,
!   - UPPER = false :    PROBT( i, j ) = prob( U < T(i,j) )  ,
!   
!   , for U = Student(NDF), i=1 to size(T,1) and j=1 to size(T,2).
!
!
! Arguments
! _________
!
!   T          (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration of the t-density.
!
!   NDF        (INPUT) integer(i4b)
!              On entry, degrees of freedom of the t-distribution. NDF must be greater
!              or equal to 1.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  :  probabilities to the right of T is calculated.
!              - UPPER = false  : probabilities to the left of T is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the t_density is integrated.
!              - NDF is greater than NDF_MAX, an asymptotic series is used.
!
!              The default is 20.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-distribution,
!                        (Algorithm 395), Comm. A.C.M., vol.13, 617-619
!
!   (2)   Cooper, B.E., 1968: The integral of student's t distribution,
!                        (Algorithm AS3), Applied Statistics, vol.17, no.2, 189
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, half, pi,                &
                                  c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                intent(in) :: t
    real(stnd), dimension(size(t,1),size(t,2))            :: probt_rm
!
    integer(i4b), intent(in)           :: ndf
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(max(ndf/2,1))        :: vec
    real(stnd), dimension(size(t,1),size(t,2)) :: prob
    real(stnd)                                 :: a, b, c, f, s, df, tji
!
    integer(i4b) :: i, j, l, k, n, m, ndf_max2
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: g1=one/pi
!
    character(len=*),  parameter :: name_proc='probt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( t, 1 )
    n = size( t, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m*n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( ndf<1_i4b )  then
        call merror( name_proc//prob_error2 )
    end if
!
    if ( present(ndf_max) ) then
!
        ndf_max2 = ndf_max
!
    else
!
        ndf_max2 = 20_i4b
!
    end if
!
    df = real( ndf, stnd )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,j,l,k,tji,a,b,c,f,s,vec)
!
    if ( ndf<=ndf_max2 ) then
!
!       EXACT COMPUTATION FOR SMALL ndf.
!
        if ( ndf>=4_i4b ) then
!
            c = mod( df, two ) + two
            l = ndf/2_i4b - 1_i4b
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                do j=1_i4b, m
!
                    tji = t(j,i)
!
                    b = df/( df + tji*tji )
                    a = tji/sqrt( df )
!
                    vec(1_i4b) = b - b/c
!
                    s = c
!
                    do k=2_i4b, l
                        s = s + two
                        vec(k) = vec(k-1_i4b)*( b - b/s )      
                    end do
!
                    s = one + sum( vec(:l) )
!
                    if ( mod(ndf,2_i4b)/=1_i4b ) then
                        s = half*a*sqrt( b )*s
                    else
                        s = ( a*b*s + atan( a ) )*g1
                    end if
!
                    if ( upper ) then
                        prob(j,i) = half - s
                    else
                        prob(j,i) = half + s
                    end if
!
                end do
!
            end do
!
!$OMP END DO
!
        else
!
!            c = merge( zero, one, ndf==1_i4b )
            if ( ndf==1_i4b ) then
                c = zero
            else
                c = one
            end if
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                do j=1_i4b, m
!
                    tji = t(j,i)
!
                    a = tji*tji
!
                    b = df/( df + a )
                    a = tji/sqrt( df )
!
                    if ( mod(ndf,2_i4b)/=1_i4b ) then
                        s = half*a*sqrt( b )
                    else
                        s = ( a*b*c + atan( a ) )*g1
                    end if
!
                    if ( upper ) then
                        prob(j,i) = half - s
                    else
                        prob(j,i) = half + s
                    end if
!
                end do
!
            end do
!
!$OMP END DO
!
        end if
!
    else
!
!       ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE ndf.
!
!$OMP DO SCHEDULE(STATIC)
!
        do i=1_i4b, n
!
            do j=1_i4b, m
!
                tji = t(j,i)
!
                a = (tji*tji)/df
                b = one + a
                s = b - one
                if ( s/=zero ) a = a*( log( b )/s )
                c = df - half
                f = c48*c*c
                a = c*a
                s = a*a
                b = (((((-c0_4*a - c3_3)*a - c24)*a - c85_5) /                 &
                    (c0_8*s + c100 + f) + a +three) / f + one) * sqrt( a )
!
                if ( tji<zero ) b = -b
!
                prob(j,i) = probn_r( b, upper )
!
            end do
!
        end do
!
!$OMP END DO
!
    end if
!
!$OMP END PARALLEL
!
    probt_rm(:m,:n) = prob(:m,:n)
!
!
! END OF FUNCTION probt_rm
! ________________________
!
    end function probt_rm
!
! =========================================================================================
!
    function probt_rm2( t, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the Student's t-distribution function from T(i,j) to infinity 
!   if UPPER is true or from minus infinity to T(i,j) if UPPER is false,
!   for i=1 to size(T,1) and j=1 to size(T,2). In other words, if:
!
!   - UPPER = true  :    PROBT( i, j ) = prob( U > T(i,j) )  ,
!   - UPPER = false :    PROBT( i, j ) = prob( U < T(i,j) )  ,
!   
!   , for U = Student(NDF(i,j)), i=1 to size(T,1) and j=1 to size(T,2).
!
!
! Arguments
! _________
!
!   T          (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration of the t-density.
!
!   NDF        (INPUT) integer(i4b), dimension(:,:)
!              On entry, degrees of freedom of the t-distribution. Any value in the array
!              NDF must be greater or equal to 1.
!
!              The shape of NDF must verify:
!
!              - size(NDF,1) = size(T,1)
!              - size(NDF,2) = size(T,2).
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  :  probabilities to the right of T is calculated.
!              - UPPER = false  : probabilities to the left of T is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the t_density is integrated.
!              - NDF is greater than NDF_MAX, an asymptotic series is used.
!
!              The default is 20.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-distribution,
!                        (Algorithm 395), Comm. A.C.M., vol.13, 617-619
!
!   (2)   Cooper, B.E., 1968: The integral of student's t distribution,
!                        (Algorithm AS3), Applied Statistics, vol.17, no.2, 189
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, half, pi,                &
                                  c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Char_Constants,    only : prob_error2, allocate_error
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                intent(in) :: t
    real(stnd), dimension(size(t,1),size(t,2))            :: probt_rm2
!
    integer(i4b), dimension(:,:), intent(in)           :: ndf
    integer(i4b),                 intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(:), allocatable      :: vec
    real(stnd), dimension(size(t,1),size(t,2)) :: prob
    real(stnd)                                 :: a, b, c, f, s, tji
!
    integer      :: iok
    integer(i4b) :: i, j, l, k, n, m, ndfji, ndf_max2, ndfm
!
#ifdef _OPENMP
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: g1=one/pi
!
    character(len=*),  parameter :: name_proc='probt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(t,1),i4b)     ,   &
                    int(size(ndf,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(t,2),i4b)     ,   &
                    int(size(ndf,2),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m*n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any( ndf(:m,:n)<1_i4b ) ) then
        call merror( name_proc//prob_error2 )
    end if
!
    if ( present(ndf_max) ) then
!
        ndf_max2 = ndf_max
!
    else
!
        ndf_max2 = 20_i4b
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
    ndfm = max( ndf_max2/2_i4b, 1_i4b )
!
!$OMP PARALLEL IF(test_par), PRIVATE(iok,i,j,l,k,ndfji,tji,a,b,c,f,s,vec)
!
!   ALLOCATE WORK ARRAY.
!
    allocate( vec(ndfm), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!$OMP DO SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        do j=1_i4b, m
!
            tji   = t(j,i)
            ndfji = ndf(j,i)
!
            f = real( ndfji, stnd )
            a = tji*tji
!
            if ( ndfji<=ndf_max2 ) then
!
!               EXACT COMPUTATION FOR SMALL ndf.
!
                b = f/( f + a )
                a = tji/sqrt( f )
!
                if ( ndfji>=4_i4b ) then
!
                    c = mod( f, two ) + two
                    l = ndfji/2_i4b - 1_i4b
!
                    vec(1_i4b) = b - b/c
!
                    do k=2_i4b, l
                        c = c + two
                        vec(k) = vec(k-1_i4b)*( b - b/c )              
                    end do
!
                    s = one + sum( vec(:l) )
!
                else
!
!                    s = merge( zero, one, ndfji==1_i4b )
                    if ( ndfji==1_i4b ) then
                        s = zero
                    else
                        s = one
                    end if
!
                end if
!
                if ( mod(ndfji,2_i4b)/=1_i4b ) then
                    s = half*a*sqrt( b )*s
                else
                    s = ( a*b*s + atan( a ) )*g1
                end if
!
                if ( upper ) then
                    prob(j,i) = half - s
                else
                    prob(j,i) = half + s
                end if
!
            else
!
!               ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE ndf.
!
                a = a/f
                b = one + a
                s = b - one
                if ( s/=zero ) a = a*( log( b )/s )
                c = f - half
                f = c48*c*c
                a = c*a
                s = a*a
                b = (((((-c0_4*a - c3_3)*a - c24)*a - c85_5) /                 &
                      (c0_8*s + c100 + f) + a +three) / f + one) * sqrt( a )
!
                if ( tji<zero ) b = -b
!
                prob(j,i) = probn_r( b, upper )
!
            end if
!
        end do
!
    end do
!
!$OMP END DO
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( vec )
!
!$OMP END PARALLEL
!
    probt_rm2(:m,:n) = prob(:m,:n)
!
!
! END OF FUNCTION probt_rm2
! _________________________
!
    end function probt_rm2
!
! =========================================================================================
!
    function probstudent_r( t, df )
!
! Purpose
! _______
!
!   Evaluates the two-tailed probability of Student's t.  PROBSTUDENT
!   computes the probability that a random variable following Student's t
!   distribution will exceed abs(T) in absolute value.
!
!
! Arguments
! _________
!
!   T      (INPUT) real(stnd)
!          On entry, input constant. PROBSTUDENT computes 
!          the probability that abs(T) will be exceeded in absolute value.
!
!   DF     (INPUT) real(stnd)
!          On entry, degrees of freedom of the t-distribution. DF must be greater
!          or equal to 1. DF is not necessarily an integer.
!
!
! Further Details
! _______________
!
!   This function is not very accurate for very small degrees of freedom (e.g. number of
!   degrees of freedom less than 5).
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Algorithm 395 Student's t-distribution,
!                            Comm. A.C.M., vol.13, 617-619
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, four, half, pi          &
                                  ,c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Num_Constants,     only : machtiny
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: t, df
    real(stnd)             :: probstudent_r
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: f, t2, a, b, xj, y, z
!
    integer(i4b) :: ndf 
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: twoonpi=two/pi
!
    character(len=*),  parameter :: name_proc='probstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( df<one ) then
        call merror( name_proc//prob_error2 )
    end if
!
    f = df
    ndf = int( f, i4b )
    t2 = t*t
    y = t2/f
    b = one + y
!
    if ( f/=real(ndf,stnd) .or. ( ndf>=20_i4b    .and. t2<f ) .or. ndf>200_i4b ) then
!
!       ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE OR NON-INTEGER df .
!
        if ( b>one ) y = y*(log( b )/(b - one))
!
        a = f - half
        b = c48*a*a
        y = a*y
        y = (((((-c0_4*y - c3_3)*y - c24)*y - c85_5) /                 &
              (c0_8*(y*y) + c100 + b) + y +three) / b + one)*sqrt( y )
!
        probstudent_r = two*probn_r( y, upper=true )
!
    else if ( ndf<20_i4b .and. t2<four ) then
!
!       NESTED SUMMATION OF COSINE SERIES.
!
        y = sqrt( y )
        a = merge( zero, y, ndf==1_i4b )
!
        do
!
            f = f - two
!
            if ( f<=one ) exit
!
            a = ( f - one )/( b*f )*a + y
!
        end do
!
        if ( f==zero ) then
            probstudent_r = one - a/sqrt( b )
        else
            probstudent_r = one - ( atan(y) + a/b )*twoonpi
        end if
!
    else
!
!       TAIL SERIES EXPANSION FOR LARGE t VALUES.
!
        a = one
        y = f
        xj = zero
        z = zero
!
        do
!
            if ( a==z ) exit
!
            xj = xj + two
            z = a
            y = y*( xj - one )/( b*xj )
            a = a + y/(f+xj )
!
        end do
!
        do
!
            if ( f<=one .or. a<machtiny ) exit
!
            a = ( f - one )/( b*f )*a
            f = f - two
!
        end do
!
        if ( f==zero ) then
            probstudent_r = a
        else
            probstudent_r = sqrt(b)*twoonpi*a/b
        end if
!
    end if
!
!
! END OF FUNCTION probstudent_r
! _____________________________
!
    end function probstudent_r
!
! =========================================================================================
!
    function probstudent_rv( t, df )
!
! Purpose
! _______
!
!   Evaluates the two-tailed probabilities of Student's t.  PROBSTUDENT
!   computes  probabilities that a random variable following Student's t
!   distribution will exceed abs(T(i)) in absolute value, for i=1 to size(T).
!
!
! Arguments
! _________
!
!   T      (INPUT) real(stnd), dimension(:)
!          On entry, input constants. PROBSTUDENT computes 
!          probabilities that abs(T(i)) will be exceeded in absolute value,
!          for i=1 to size(T).
!
!   DF     (INPUT) real(stnd)
!          On entry, degrees of freedom of the t-distribution. DF must be greater
!          or equal to 1. DF is not necessarily an integer.
!
!
! Further Details
! _______________
!
!   This function is not very accurate for very small degrees of freedom (e.g. number of
!   degrees of freedom less than 5).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Algorithm 395 Student's t-distribution,
!                            Comm. A.C.M., vol.13, 617-619
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, four, half, pi           &
                                  ,c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Num_Constants,     only : machtiny
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),      intent(in) :: t
    real(stnd),                    intent(in) :: df
    real(stnd), dimension(size(t))            :: probstudent_rv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: f, t2, a, b, xj, y, z, probi
    real(stnd), dimension(size(t)) :: probvec
!
    integer(i4b) :: ndf, i, n
!
#ifdef _OPENMP
    integer(i4b) :: j 
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: twoonpi=two/pi
!
    character(len=*),  parameter :: name_proc='probstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( t )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( df<one ) then
        call merror( name_proc//prob_error2 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,ndf,f,t2,a,b,xj,y,z,probi)
!
    do i=1_i4b, n
!
        f = df
        ndf = int( f, i4b )
!
        t2 = t(i)*t(i)
        y = t2/f
        b = one + y
!
        if ( f/=real(ndf,stnd) .or. ( ndf>=20_i4b    .and. t2<f ) .or. ndf>200_i4b ) then
!
!           ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE OR NON-INTEGER df .
!
            if ( b>one ) y = y*(log( b )/(b - one))
            a = f - half
            b = c48*a*a
            y = a*y
            y = (((((-c0_4*y - c3_3)*y - c24)*y - c85_5) /                 &
                (c0_8*(y*y) + c100 + b) + y +three) / b + one)*sqrt( y )
!
            probi = two*probn_r( y, upper=true )
!
        else if ( ndf<20_i4b .and. t2<four ) then
!
!           NESTED SUMMATION OF COSINE SERIES.
!
            y = sqrt( y )
!
!            a = merge( zero, y, ndf==1_i4b )
            if ( ndf==1_i4b ) then
                a = zero
            else
                a = y
            end if
!
            do
!
                f = f - two
!
                if ( f<=one ) exit
!
                a = ( f - one )/( b*f )*a + y
!
            end do
!
            if ( f==zero ) then
                probi = one - a/sqrt( b )
            else
                probi = one - ( atan(y) + a/b )*twoonpi
            end if
!
        else
!
!           TAIL SERIES EXPANSION FOR LARGE t VALUES.
!
            a = one
            y = f
            xj = zero
            z = zero
!
            do
!
                if ( a==z ) exit
!
                xj = xj + two
                z = a
                y = y*( xj - one )/( b*xj )
                a = a + y/(f+xj )
!
            end do
!
            do
!
                if ( f<=one .or. a<machtiny ) exit
!
                a = ( f - one )/( b*f )*a
                f = f - two
!
            end do
!
            if ( f==zero ) then
                probi = a
            else
                probi = sqrt(b)*twoonpi*a/b
            end if
!
        end if
!
        probvec(i) = probi
!
    end do
!
!$OMP END PARALLEL DO
!
    probstudent_rv(1_i4b:n) = probvec(1_i4b:n)
!
!
! END OF FUNCTION probstudent_rv
! ______________________________
!
    end function probstudent_rv
!
! =========================================================================================
!
    function probstudent_rv2( t, df )
!
! Purpose
! _______
!
!   Evaluates the two-tailed probabilities of Student's t.  PROBSTUDENT
!   computes  probabilities that a random variable following Student's t
!   distribution will exceed abs(T(i)) in absolute value, for i=1 to size(T).
!
!
! Arguments
! _________
!
!   T      (INPUT) real(stnd), dimension(:)
!          On entry, input constants. PROBSTUDENT computes 
!          probabilities that abs(T(i)) will be exceeded in absolute value,
!          for i=1 to size(T).
!
!   DF     (INPUT) real(stnd), dimension(:)
!          On entry, degrees of freedom of the t-distribution. Any value in the array DF
!          must be greater or equal to 1, but is not necessarily an integer.
!
!          The size of DF must be  size(DF) = size(T) .
!
!
! Further Details
! _______________
!
!   This function is not very accurate for very small degrees of freedom (e.g. number of
!   degrees of freedom less than 5).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Algorithm 395 Student's t-distribution,
!                            Comm. A.C.M., vol.13, 617-619
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, four, half, pi           &
                                  ,c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Num_Constants,     only : machtiny
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),      intent(in) :: t, df
    real(stnd), dimension(size(t))            :: probstudent_rv2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: f, t2, a, b, xj, y, z, probi
    real(stnd), dimension(size(t)) :: probvec
!
    integer(i4b) :: ndf, i, n
!
#ifdef _OPENMP
    integer(i4b) :: j 
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: twoonpi=two/pi
!
    character(len=*),  parameter :: name_proc='probstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(t),i4b)     ,   &
                    int(size(df),i4b)    ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any( df(:n)<one ) ) then
        call merror( name_proc//prob_error2 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,ndf,f,t2,a,b,xj,y,z,probi)
!
    do i=1_i4b, n
!
        f = df(i)
        ndf = int( f, i4b )
!
        t2 = t(i)*t(i)
        y = t2/f
        b = one + y
!
        if ( f/=real(ndf,stnd) .or. ( ndf>=20_i4b    .and. t2<f ) .or. ndf>200_i4b ) then
!
!           ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE OR NON-INTEGER df .
!
            if ( b>one ) y = y*(log( b )/(b - one))
            a = f - half
            b = c48*a*a
            y = a*y
            y = (((((-c0_4*y - c3_3)*y - c24)*y - c85_5) /                 &
                (c0_8*(y*y) + c100 + b) + y +three) / b + one)*sqrt( y )
!
            probi = two*probn_r( y, upper=true )
!
        else if ( ndf<20_i4b .and. t2<four ) then
!
!           NESTED SUMMATION OF COSINE SERIES.
!
            y = sqrt( y )
!
!            a = merge( zero, y, ndf==1_i4b )
            if ( ndf==1_i4b ) then
                a = zero
            else
                a = y
            end if
!
            do
!
                f = f - two
!
                if ( f<=one ) exit
!
                a = ( f - one )/( b*f )*a + y
!
            end do
!
            if ( f==zero ) then
                probi = one - a/sqrt( b )
            else
                probi = one - ( atan(y) + a/b )*twoonpi
            end if
!
        else
!
!           TAIL SERIES EXPANSION FOR LARGE t VALUES.
!
            a = one
            y = f
            xj = zero
            z = zero
!
            do
!
                if ( a==z ) exit
!
                xj = xj + two
                z = a
                y = y*( xj - one )/( b*xj )
                a = a + y/(f+xj )
!
            end do
!
            do
!
                if ( f<=one .or. a<machtiny ) exit
!
                a = ( f - one )/( b*f )*a
                f = f - two
!
            end do
!
            if ( f==zero ) then
                probi = a
            else
                probi = sqrt(b)*twoonpi*a/b
            end if
!
        end if
!
        probvec(i) = probi
!
    end do
!
!$OMP END PARALLEL DO
!
    probstudent_rv2(1_i4b:n) = probvec(1_i4b:n)
!
!
! END OF FUNCTION probstudent_rv2
! _______________________________
!
    end function probstudent_rv2
!
! =========================================================================================
!
    function probstudent_rm( t, df )
!
! Purpose
! _______
!
!   Evaluates  two-tailed probabilities of Student's t.  PROBSTUDENT
!   computes  probabilities that a random variable following Student's t
!   distribution will exceed abs(T(i,j)) in absolute value, for i=1 to size(T,1)
!   and j=1 to size(T,2).
!
!
! Arguments
! _________
!
!   T      (INPUT) real(stnd), dimension(:,:)
!          On entry, input constants. PROBSTUDENT computes 
!          probabilities that abs(T(i,j)) will be exceeded in absolute value,
!          for i=1 to size(T,1) and j=1 to size(T,2).
!
!   DF     (INPUT) real(stnd)
!          On entry, degrees of freedom of the t-distribution. DF must be greater
!          or equal to 1. DF is not necessarily an integer.
!
!
! Further Details
! _______________
!
!   This function is not very accurate for very small degrees of freedom (e.g. number of
!   degrees of freedom less than 5).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Algorithm 395 Student's t-distribution,
!                            Comm. A.C.M., vol.13, 617-619
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, four, half, pi          &
                                  ,c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Num_Constants,     only : machtiny
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                intent(in) :: t
    real(stnd),                                intent(in) :: df
    real(stnd), dimension(size(t,1),size(t,2))            :: probstudent_rm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: f, t2, a, b, xj, y, z, probji
    real(stnd), dimension(size(t,1),size(t,2)) :: probmat
!
    integer(i4b) :: ndf, i, j, n, m
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: twoonpi=two/pi
!
    character(len=*),  parameter :: name_proc='probstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( t, 1 )
    n = size( t, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( df<one ) then
        call merror( name_proc//prob_error2 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,ndf,f,t2,a,b,xj,y,z,probji)
!
    do i=1_i4b, n
!
        do j=1_i4b, m
!
            f = df
            ndf = int( f, i4b )
!
            t2 = t(j,i)*t(j,i)
            y = t2/f
            b = one + y
!
            if ( f/=real(ndf,stnd) .or. ( ndf>=20_i4b    .and. t2<f ) .or. ndf>200_i4b ) then
!
!               ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE OR NON-INTEGER df .
!
                if ( b>one ) y = y*(log( b )/(b - one))
                a = f - half
                b = c48*a*a
                y = a*y
                y = (((((-c0_4*y - c3_3)*y - c24)*y - c85_5) /                 &
                    (c0_8*(y*y) + c100 + b) + y +three) / b + one)*sqrt( y )
!
                probji = two*probn_r( y, upper=true )
!
            else if ( ndf<20_i4b .and. t2<four ) then
!
!               NESTED SUMMATION OF COSINE SERIES.
!
                y = sqrt( y )
!
!                a = merge( zero, y, ndf==1_i4b )
                if ( ndf==1_i4b ) then
                    a = zero
                else
                    a = y
                end if
!
                do
!
                    f = f - two
!
                    if ( f<=one ) exit
!
                    a = ( f - one )/( b*f )*a + y
!
                end do
!
                if ( f==zero ) then
                    probji = one - a/sqrt( b )
                else
                    probji = one - ( atan(y) + a/b )*twoonpi
                end if
!
            else
!
!               TAIL SERIES EXPANSION FOR LARGE t VALUES.
!
                a = one
                y = f
                xj = zero
                z = zero
!
                do
!
                    if ( a==z ) exit
!
                    xj = xj + two
                    z = a
                    y = y*( xj - one )/( b*xj )
                    a = a + y/(f+xj )
!
                end do
!
                do
!
                    if ( f<=one .or. a<machtiny ) exit
!
                    a = ( f - one )/( b*f )*a
                    f = f - two
!
                end do
!
                if ( f==zero ) then
                    probji = a
                else
                    probji = sqrt(b)*twoonpi*a/b
                end if
!
            end if
!
            probmat(j,i) = probji
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probstudent_rm(1_i4b:m,1_i4b:n) = probmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION probstudent_rm
! ______________________________
!
    end function probstudent_rm
!
! =========================================================================================
!
    function probstudent_rm2( t, df )
!
! Purpose
! _______
!
!   Evaluates  two-tailed probabilities of Student's t.  PROBSTUDENT
!   computes  probabilities that a random variable following Student's t
!   distribution will exceed abs(T(i,j)) in absolute value, for i=1 to size(T,1)
!   and j=1 to size(T,2).
!
!
! Arguments
! _________
!
!   T      (INPUT) real(stnd), dimension(:,:)
!          On entry, input constants. PROBSTUDENT computes 
!          probabilities that abs(T(i,j)) will be exceeded in absolute value,
!          for i=1 to size(T,1) and j=1 to size(T,2).
!
!   DF     (INPUT) real(stnd), dimension(:,:)
!          On entry, degrees of freedom of the t-distribution. Any value in the array DF
!          must be greater or equal to 1, but is not necessarily an integer.
!
!          The shape of DF must verify:
!
!          - size(DF,1) = size(T,1)
!          - size(DF,2) = size(T,2).
!
!
! Further Details
! _______________
!
!   This function is not very accurate for very small degrees of freedom (e.g. number of
!   degrees of freedom less than 5).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Algorithm 395 Student's t-distribution,
!                            Comm. A.C.M., vol.13, 617-619
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, four, half, pi          &
                                  ,c0_4, c0_8, c3_3, c24, c48, c85_5, c100
    use Num_Constants,     only : machtiny
    use Char_Constants,    only : prob_error2
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                intent(in) :: t, df
    real(stnd), dimension(size(t,1),size(t,2))            :: probstudent_rm2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: f, t2, a, b, xj, y, z, probji
    real(stnd), dimension(size(t,1),size(t,2)) :: probmat
!
    integer(i4b) :: ndf, i, j, n, m
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: twoonpi=two/pi
!
    character(len=*),  parameter :: name_proc='probstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(t,1),i4b)     ,   &
                    int(size(df,1),i4b)    ,   &
                    name_proc )
!
    n =  assert_eq( int(size(t,2),i4b)     ,   &
                    int(size(df,2),i4b)    ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any( df(:m,:n)<one ) ) then
        call merror( name_proc//prob_error2 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,ndf,f,t2,a,b,xj,y,z,probji)
!
    do i=1_i4b, n
!
        do j=1_i4b, m
!
            f = df(j,i)
            ndf = int( f, i4b )
!
            t2 = t(j,i)*t(j,i)
            y = t2/f
            b = one + y
!
            if ( f/=real(ndf,stnd) .or. ( ndf>=20_i4b    .and. t2<f ) .or. ndf>200_i4b ) then
!
!               ASYMPTOTIC SERIES (GAUSSIAN APPROXIMATION) FOR LARGE OR NON-INTEGER df .
!
                if ( b>one ) y = y*(log( b )/(b - one))
                a = f - half
                b = c48*a*a
                y = a*y
                y = (((((-c0_4*y - c3_3)*y - c24)*y - c85_5) /                 &
                    (c0_8*(y*y) + c100 + b) + y +three) / b + one)*sqrt( y )
!
                probji = two*probn_r( y, upper=true )
!
            else if ( ndf<20_i4b .and. t2<four ) then
!
!               NESTED SUMMATION OF COSINE SERIES.
!
                y = sqrt( y )
!
!                a = merge( zero, y, ndf==1_i4b )
                if ( ndf==1_i4b ) then
                    a = zero
                else
                    a = y
                end if
!
                do
!
                    f = f - two
!
                    if ( f<=one ) exit
!
                    a = ( f - one )/( b*f )*a + y
!
                end do
!
                if ( f==zero ) then
                    probji = one - a/sqrt( b )
                else
                    probji = one - ( atan(y) + a/b )*twoonpi
                end if
!
            else
!
!               TAIL SERIES EXPANSION FOR LARGE t VALUES.
!
                a = one
                y = f
                xj = zero
                z = zero
!
                do
!
                    if ( a==z ) exit
!
                    xj = xj + two
                    z = a
                    y = y*( xj - one )/( b*xj )
                    a = a + y/(f+xj )
!
                end do
!
                do
!
                    if ( f<=one .or. a<machtiny ) exit
!
                    a = ( f - one )/( b*f )*a
                    f = f - two
!
                end do
!
                if ( f==zero ) then
                    probji = a
                else
                    probji = sqrt(b)*twoonpi*a/b
                end if
!
            end if
!
            probmat(j,i) = probji
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probstudent_rm2(1_i4b:m,1_i4b:n) = probmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION probstudent_rm2
! _______________________________
!
    end function probstudent_rm2
!
!
! =========================================================================================
!                    INVERSE OF THE STUDENT'S T DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function pinvt_r( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the Student's t distribution function:
!
!                           T0 = PINVT( P, NDF ) 
!   
!   , if P = probability( U < T0 )  for U = Student(NDF).
!
!   PINVT returns the quantile T0 of Student's t-distribution  with NDF degrees of 
!   freedom corresponding to a given lower tail area of P.
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd)
!          On entry, the probability. P must verify 0. < P < 1.
!
!   NDF    (INPUT) integer(i4b)
!          On entry, degrees of freedom of the t-distribution. NDF must be greater
!          or equal to 1.
!
!
! Further Details
! _______________
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: p
!
    integer(i4b), intent(in) :: ndf
!
    real(stnd)             :: pinvt_r
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, d, df, q, x, y
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    q = merge( two*(one - p), two*p, p>half )
!
!   TEST THE ARGUMENTS.
!
    if ( q<=zero .or. ndf<1_i4b )  then
        call merror( name_proc//prob_error3 )
    end if
!
    select case( ndf )
!
        case( 1_i4b )
!
!           EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
            q = q * pio2
            pinvt_r = cos( q )/sin( q )
!
        case( 2_i4b )
!
!           EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
            pinvt_r = sqrt( abs( two/( q*(two - q) ) - two ) )
!
        case( 3_i4b: )
!
!           EXPANSION.
!
            df = real( ndf, stnd )
            a = one/( df - half )
            b = c48/( a*a )
            c = ( ( c20700*a/b - c98 )*a - c16 )*a + c96_36
            d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
            x = d*q
            y = x**(two/df)
!
            if ( y>c5_m2+a ) then
!
!               ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                x = pinvn_r( half*q )
                y = x*x
                if ( ndf<5_i4b ) c = c + c0_3*( df - c4_5 )*( x + c0_6 )
                c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
                y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
                y = a*y*y
!
                if ( y>c2_m3 ) then
                    y = exp(y) - one
                else
                    y = half*y*y + y
                end if
!
            else
                y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                        + half/(df + four) )*y - one ) * (df + one)/(df + two) + one/y
            end if
!
            pinvt_r = sqrt( df*y )
!
     end select
!
    if ( p<half ) pinvt_r = -pinvt_r
!
!
! END OF FUNCTION pinvt_r
! _______________________
!
    end function pinvt_r
!
! =========================================================================================
!
    function pinvt_rv( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the Student's t distribution function:
!
!                           T0(i) = PINVT( P(i), NDF ) 
!   
!   , if P(i) = probability( U < T0(i) )  for U = Student(NDF) and i=1 to size(P).
!
!   PINVT returns the quantiles T0(i) of Student's t-distribution with NDF degrees of
!   freedom corresponding to a given lower tail area of P(i), for i=1 to size(P).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:)
!          On entry, the probabilities. P(i) must verify 0. < P(i) < 1,
!          for i=1 to size(P).
!
!   NDF    (INPUT) integer(i4b)
!          On entry, degrees of freedom of the t-distribution. NDF must be greater
!          or equal to 1.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),  intent(in) :: p
!
    integer(i4b), intent(in) :: ndf
!
    real(stnd), dimension(size(p)) :: pinvt_rv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: a, b, c, d, df, x, y
    real(stnd), dimension(size(p)) :: q
!
    integer(i4b) :: i, n
!
#ifdef _OPENMP
    integer(i4b) :: j 
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( p )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
    where ( p(:n)>half )
        q(:n) = two*(one - p(:n))
    elsewhere
        q(:n) = two*p(:n)
    end where
!
!   TEST THE ARGUMENTS.
!
    if ( any( q(:n)<=zero ) .or. ndf<1_i4b ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,x,y)
!
    select case( ndf )
!
        case( 1_i4b )
!
!           EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                y = q(i) * pio2
                x = cos( y )/sin( y )
                if ( p(i)<half ) x = -x
                q(i) = x
!
            end do
!
!$OMP END DO
!
        case( 2_i4b )
!
!           EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                y = q(i)
                x = sqrt( abs( two/( y*(two - y) ) - two ) )
                if ( p(i)<half ) x = -x
                q(i) = x
!
            end do
!
!$OMP END DO
!
        case( 3_i4b: )
!
!           EXPANSION.
!
!$OMP SINGLE
!
            df = real( ndf, stnd )
            a = one/( df - half )
            b = c48/( a*a )
            c = ( ( c20700*a/b - c98 )*a - c16 )*a + c96_36
            d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                x = d*q(i)
                y = x**(two/df)
!
                if ( y>c5_m2+a ) then
!
!                   ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                    x = pinvn_r( half*q(i) )
                    y = x*x
                    if ( ndf<5_i4b ) c = c + c0_3*( df - c4_5 )*( x + c0_6 )
                    c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
                    y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
                    y = a*y*y
!
                    if ( y>c2_m3 ) then
                        y = exp(y) - one
                    else
                        y = half*y*y + y
                    end if
!
                else
                    y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                        + half/(df + four) )*y - one ) * (df + one)/(df + two) + one/y
                end if
!
                x = sqrt( df*y )
!
                if ( p(i)<half ) x = -x
!
                q(i) = x
!
            end do
!
!$OMP END DO
!
    end select
!
!$OMP END PARALLEL
!
    pinvt_rv(:n) = q(:n)
!
!
! END OF FUNCTION pinvt_rv
! ________________________
!
    end function pinvt_rv
!
! =========================================================================================
!
    function pinvt_rv2( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the Student's t distribution function:
!
!                           T0(i) = PINVT( P(i), NDF(i) )
!   
!   , if P(i) = probability( U < T0(i) )  for U = Student(NDF(i)) and i=1 to size(P).
!
!   PINVT returns the quantiles T0(i) of Student's t-distribution with NDF(i) degrees of
!   freedom corresponding to a given lower tail area of P(i), for i=1 to size(P).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:)
!          On entry, the probabilities. P(i) must verify 0. < P(i) < 1,
!          for i=1 to size(P).
!
!   NDF    (INPUT) integer(i4b), dimension(:)
!          On entry, degrees of freedom of the t-distribution. Any value in the array
!          NDF must be greater or equal to 1.
!
!          The size of NDF must be  size(NDF) = size(P).
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),  intent(in) :: p
!
    integer(i4b), dimension(:), intent(in) :: ndf
!
    real(stnd), dimension(size(p)) :: pinvt_rv2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: qi, a, b, c, d, df, x, y
    real(stnd), dimension(size(p)) :: q
!
    integer(i4b) :: i, n, ndfi
!
#ifdef _OPENMP
    integer(i4b) :: j 
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(p),i4b)     ,   &
                    int(size(ndf),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
    where ( p(:n)>half )
        q(:n) = two*(one - p(:n))
    elsewhere
        q(:n) = two*p(:n)
    end where
!
!   TEST THE ARGUMENTS.
!
    if ( any( q(:n)<=zero ) .or. any( ndf(:n)<1_i4b ) ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,ndfi,qi,a,b,c,d,df,x,y), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        qi   = q(i)
        ndfi = ndf(i)
!
        select case( ndfi )
!
            case( 1_i4b )
!
!               EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
                y = qi * pio2
                x = cos( y )/sin( y )
!
            case( 2_i4b )
!
!               EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
                x = sqrt( abs( two/( qi*(two - qi) ) - two ) )
!
            case( 3_i4b: )
!
!               EXPANSION.
!
                df = real( ndfi, stnd )
                a = one/( df - half )
                b = c48/( a*a )
                c = ( ( c20700*a/b - c98 )*a - c16 )*a + c96_36
                d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
                x = d*qi
                y = x**(two/df)
!
                if ( y>c5_m2+a ) then
!
!                   ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                    x = pinvn_r( half*qi )
                    y = x*x
                    if ( ndfi<5_i4b ) c = c + c0_3*( df - c4_5 )*( x + c0_6 )
                    c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
                    y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
                    y = a*y*y
!
                    if ( y>c2_m3 ) then
                        y = exp(y) - one
                    else
                        y = half*y*y + y
                    end if
!
                else
                    y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                        + half/(df + four) )*y - one ) * (df + one)/(df + two) + one/y
                end if
!
                x = sqrt( df*y )
!
        end select
!
        if ( p(i)<half ) x = -x
!
        q(i) = x
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvt_rv2(:n) = q(:n)
!
!
! END OF FUNCTION pinvt_rv2
! _________________________
!
    end function pinvt_rv2
!
! =========================================================================================
!
    function pinvt_rm( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the Student's t distribution function:
!
!               T0(i,j) = PINVT( P(i,j), NDF ) 
!   
!   , if P(i,j) = probability( U < T0(i,j) )  for U = Student(NDF), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVT returns the quantiles T0(i,j) of Student's t-distribution with NDF degrees of 
!   freedom corresponding to a given lower tail area of P(i,j) for i=1 to size(P,1) and 
!   j=1 to size(P,2).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:,:)
!          On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1,
!          for i=1 to size(P,1) and j=1 to size(P,2).
!
!   NDF    (INPUT) integer(i4b)
!          On entry, degrees of freedom of the t-distribution. NDF must be greater
!          or equal to 1.
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),  intent(in) :: p
!
    integer(i4b), intent(in) :: ndf
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvt_rm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: a, b, c, d, df, x, y
    real(stnd), dimension(size(p,1),size(p,2)) :: q
!
    integer(i4b) :: i, j, n, m
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( p, 1 )
    n = size( p, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m*n<=0_i4b ) then
        return
    end if
!
    where ( p(:m,:n)>half )
        q(:m,:n) = two*(one - p(:m,:n))
    elsewhere
        q(:m,:n) = two*p(:m,:n)
    end where
!
!   TEST THE ARGUMENTS.
!
    if ( any( q(:m,:n)<=zero ) .or. ndf<1_i4b ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(i,j,x,y)
!
    select case( ndf )
!
        case( 1_i4b )
!
!           EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                do j=1_i4b, m
!
                    y = q(j,i) * pio2
                    x = cos( y )/sin( y )
                    if ( p(j,i)<half ) x = -x
                    q(j,i) = x
!
                end do
!
            end do
!
!$OMP END DO
!
        case( 2_i4b )
!
!           EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                do j=1_i4b, m
!
                    y = q(j,i)
                    x = sqrt( abs( two/( y*(two - y) ) - two ) )
                    if ( p(j,i)<half ) x = -x
                    q(j,i) = x
!
                end do
!
            end do
!
!$OMP END DO
!
        case( 3_i4b: )
!
!           EXPANSION.
!
!$OMP SINGLE
!
            df = real( ndf, stnd )
            a = one/( df - half )
            b = c48/( a*a )
            c = ( ( c20700*a/b - c98 )*a - c16 )*a + c96_36
            d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC)
!
            do i=1_i4b, n
!
                do j=1_i4b, m
!
                x = d*q(j,i)
                y = x**(two/df)
!
                if ( y>c5_m2+a ) then
!
!                   ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                    x = pinvn_r( half*q(j,i) )
                    y = x*x
                    if ( ndf<5_i4b ) c = c + c0_3*( df - c4_5 )*( x + c0_6 )
                    c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
                    y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
                    y = a*y*y
!
                    if ( y>c2_m3 ) then
                        y = exp(y) - one
                    else
                        y = half*y*y + y
                    end if
!
                else
                    y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                        + half/(df + four) )*y - one ) * (df + one)/(df + two) + one/y
                end if
!
                x = sqrt( df*y )
!
                if ( p(j,i)<half ) x = -x
!
                q(j,i) = x
!
                end do
!
            end do
!
!$OMP END DO
!
    end select
!
!$OMP END PARALLEL
!
    pinvt_rm(:m,:n) = q(:m,:n)
!
!
! END OF FUNCTION pinvt_rm
! ________________________
!
    end function pinvt_rm
!
! =========================================================================================
!
    function pinvt_rm2( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the Student's t distribution function:
!
!               T0(i,j) = PINVT( P(i,j), NDF(i,j) )
!   
!   , if P(i,j) = probability( U < T0(i,j) )  for U = Student(NDF(i,j)), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVT returns the quantiles T0(i,j) of Student's t-distribution with NDF(i,j) degrees of 
!   freedom corresponding to a given lower tail area of P(i,j) for i=1 to size(P,1) and 
!   j=1 to size(P,2) .
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:,:)
!          On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1,
!          for i=1 to size(P,1) and j=1 to size(P,2).
!
!   NDF    (INPUT) integer(i4b), dimension(:,:)
!          On entry, degrees of freedom of the t-distribution. Any value in the array NDF must be greater
!          or equal to 1.
!
!          The shape of NDF must verify:
!
!          - size(NDF,1) = size(P,1)
!          - size(NDF,2) = size(P,2).
!
!
! Further Details
! _______________
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),  intent(in) :: p
!
    integer(i4b), dimension(:,:), intent(in) :: ndf
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvt_rm2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: qji, a, b, c, d, df, x, y
    real(stnd), dimension(size(p,1),size(p,2)) :: q
!
    integer(i4b) :: i, j, n, m, ndfji
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvt'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(p,1),i4b)     ,   &
                    int(size(ndf,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(p,2),i4b)     ,   &
                    int(size(ndf,2),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m*n<=0_i4b ) then
        return
    end if
!
    where ( p(:m,:n)>half )
        q(:m,:n) = two*(one - p(:m,:n))
    elsewhere
        q(:m,:n) = two*p(:m,:n)
    end where
!
!   TEST THE ARGUMENTS.
!
    if ( any( q(:m,:n)<=zero ) .or. any( ndf(:m,:n)<1_i4b ) ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,ndfji,qji,a,b,c,d,df,x,y), SCHEDULE(STATIC)
!
    do i=1_i4b, n
!
        do j=1_i4b, m
!
            qji   = q(j,i)
            ndfji = ndf(j,i)
!
            select case( ndfji )
!
                case( 1_i4b )
!
!                   EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
                    y = qji * pio2
                    x = cos( y )/sin( y )
!
                case( 2_i4b )
!
!                   EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
                    x = sqrt( abs( two/( qji*(two - qji) ) - two ) )
!
                case( 3_i4b: )
!
!                   EXPANSION.
!
                    df = real( ndfji, stnd )
                    a = one/( df - half )
                    b = c48/( a*a )
                    c = ( ( c20700*a/b - c98 )*a - c16 )*a + c96_36
                    d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
                    x = d*qji
                    y = x**(two/df)
!
                    if ( y>c5_m2+a ) then
!
!                       ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                        x = pinvn_r( half*qji )
                        y = x*x
                        if ( ndfji<5_i4b ) c = c + c0_3*( df - c4_5 )*( x + c0_6 )
                        c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
                        y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
                        y = a*y*y
!
                        if ( y>c2_m3 ) then
                            y = exp(y) - one
                        else
                            y = half*y*y + y
                        end if
!
                    else
                        y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                            + half/(df + four) )*y - one ) * (df + one)/(df + two) + one/y
                    end if
!
                    x = sqrt( df*y )
!
            end select
!
            if ( p(j,i)<half ) x = -x
!
            q(j,i) = x
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvt_rm2(:m,:n) = q(:m,:n)
!
!
! END OF FUNCTION pinvt_rm2
! _________________________
!
    end function pinvt_rm2
!
! =========================================================================================
!
    function pinvstudent_r( p, df )
!
! Purpose
! _______
!
!   Evaluates the inverse of a modification of Student's t probability
!   distribution function.
!
!   PINVSTUDENT calculates the two-tail quantiles of Student's t-distribution,
!   that is a value x such that the probability of the absolute value of t being
!   greater than X is P. 
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd)
!          On entry, the probability. P is the sum of the areas (equal) 
!          in both tails of the t-distribution.
!          P must verify   0. < P < 1.
!
!   DF     (INPUT) real(stnd)
!          On entry, degrees of freedom of the t-distribution. DF must be greater
!          or equal to 1. . DF is not necessarily an integer.
!
!
! Further Details
! _______________
!
!   Note that PINVSTUDENT  does not provide the actual Student's t inverse.
!   For q equal to the probability that a Student's t random variable is less
!   than x, that inverse can be obtained by the following rules:
!
!   - for q in the range (0.0,0.5), call PINVSTUDENT with
!     P = 2 * q and negate the result x.
!   - for q in the range (0.5,1.0), call PINVSTUDENT with 
!     P = 2 * (1-q).
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Num_Constants,     only : macheps
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: p, df
!
    real(stnd)             :: pinvstudent_r
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, d, x, y
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( p<=zero .or. p>=one .or. df<one )  then
        call merror( name_proc//prob_error3 )
    end if
!
    if ( abs(df-one)<=macheps ) then
!
!       EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
        x = p * pio2
        pinvstudent_r = cos( x )/sin( x )
!
    else if ( abs(df-two)<=macheps ) then
!
!       EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
        pinvstudent_r = sqrt( abs( two/( p*(two - p) ) - two ) )
!
    else
!
!       EXPANSION.
!
        a = one/( df - half )
        b = c48/( a*a )
        c = ( ( c20700*a/b - c98 )*a - c16 )*a + c96_36
        d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
        x = d*p
        y = x**(two/df)
!
        if ( y>c5_m2+a ) then
!
!           ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
            x = pinvn_r( half*p )
            y = x*x
!
            if ( df<five ) then
                c = c + c0_3*( df - c4_5 )*( x + c0_6 )
            end if
!
            c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
            y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
            y = a*y*y
!
            if ( y>c2_m3 ) then
                y = exp( y ) - one
            else
                y = half*y*y + y
            end if
!
        else
!
            y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                + half/(df + four) )*y - one )*(df + one)/(df + two) + one/y
!
        end if
!
        pinvstudent_r = sqrt( df*y )
!
     end if
!
!
! END OF FUNCTION pinvstudent_r
! _____________________________
!
    end function pinvstudent_r
!
! =========================================================================================
!
    function pinvstudent_rv( p, df )
!
! Purpose
! _______
!
!   Evaluates the inverse of a modification of Student's t probability
!   distribution function.
!
!   PINVSTUDENT calculates the two-tail quantiles of Student's t-distribution,
!   that is a value x(i) such that the probability of the absolute value of t being
!   greater than x(i) is P(i), for i=1 to size(P). 
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:)
!          On entry, the probabilities.  P(i) is the sum of the areas (equal)
!          in both tails of the t-distribution, for i=1 to size(P).
!          P(i) must verify 0. < P(i) < 1, for i=1 to size(P).
!
!   DF     (INPUT) real(stnd)
!          On entry, degrees of freedom of the t-distribution. DF must be greater
!          or equal to 1. . DF is not necessarily an integer.
!
!
! Further Details
! _______________
!
!   Note that PINVSTUDENT  does not provide the actual Student's t inverse.
!   For q(:) equal to the probabilities that a Student's t random variable is less
!   than x(:), that inverse can be obtained by the following rules:
!
!   - for q(:) in the range (0.0,0.5), call PINVSTUDENT with
!     P(:) = 2 * q(:) and negate the result x(:).
!   - for q(:) in the range (0.5,1.0), call PINVSTUDENT with 
!     P(:) = 2 * (1-q(:)).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Num_Constants,     only : macheps
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),  intent(in) :: p
    real(stnd)              ,  intent(in) :: df
!
    real(stnd), dimension(size(p)) :: pinvstudent_rv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: a, b, c, cc, d, x, y
    real(stnd), dimension(size(p)) :: pinvvec
!
    integer(i4b) :: i, n
!
#ifdef _OPENMP
!
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( p )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any( p(:n)<=zero) .or. any(p(:n)>=one) .or. df<one ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)
!
    if ( abs(df-one)<=macheps ) then
!
!       EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC), PRIVATE(i,x)
!
        do i=1_i4b, n
!
            x = p(i) * pio2
            pinvvec(i) = cos( x )/sin( x )
!
        end do
!
!$OMP END DO
!
    else if ( abs(df-two)<=macheps ) then
!
!       EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC), PRIVATE(i,x)
!
        do i=1_i4b, n
!
            x = p(i)
            pinvvec(i) = sqrt( abs( two/( x*(two - x) ) - two ) )
!
        end do
!
!$OMP END DO
!
    else
!
!       EXPANSION.
!
!$OMP SINGLE
!
        a = one/( df - half )
        b = c48/( a*a )
        c = ( ( c20700*a/b - c98 )*a - c16)*a + c96_36
        d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC), PRIVATE(i,x,y,cc)
!
        do i=1_i4b, n
!
            x = d*p(i)
            y = x**(two/df)
!
            if ( y>c5_m2+a ) then
!
!               ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                x = pinvn_r( half*p(i) )
                y = x*x
                cc = c
!
                if ( df<five ) then
                    cc = cc + c0_3*( df - c4_5 )*( x + c0_6 )
                end if
!
                cc = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + cc
                y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/cc - y - three )/b + one )*x
                y = a*y*y
!
                if ( y>c2_m3 ) then
                    y = exp( y ) - one
                else
                    y = half*y*y + y
                end if
!
            else
!
                y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                        + half/(df + four) )*y - one )*(df + one)/(df + two) + one/y
!
            end if
!
            pinvvec(i) = sqrt( df*y )
!
        end do
!
!$OMP END DO
!
     end if
!
!$OMP END PARALLEL
!
     pinvstudent_rv(1_i4b:n) = pinvvec(1_i4b:n)
!
!
! END OF FUNCTION pinvstudent_rv
! ______________________________
!
    end function pinvstudent_rv
!
! =========================================================================================
!
    function pinvstudent_rv2( p, df )
!
! Purpose
! _______
!
!   Evaluates the inverse of a modification of Student's t probability
!   distribution function.
!
!   PINVSTUDENT calculates the two-tail quantiles of Student's t-distribution,
!   that is a value x(i) such that the probability of the absolute value of t being
!   greater than x(i) is P(i), for i=1 to size(P). 
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:)
!          On entry, the probabilities.  P(i) is the sum of the areas (equal)
!          in both tails of the t-distribution, for i=1 to size(P).
!          P(i) must verify 0. < P(i) < 1, for i=1 to size(P).
!
!   DF     (INPUT) real(stnd), dimension(:)
!          On entry, degrees of freedom of the t-distribution. Any value in the array DF
!          must be greater or equal to 1, but is not necessarily an integer.
!
!          The size of DF must be  size(DF) = size(P) .
!
!
! Further Details
! _______________
!
!   Note that PINVSTUDENT  does not provide the actual Student's t inverse.
!   For q(:) equal to the probabilities that a Student's t random variable is less
!   than x(:), that inverse can be obtained by the following rules:
!
!   - for q(:) in the range (0.0,0.5), call PINVSTUDENT with
!     P(:) = 2 * q(:) and negate the result x(:).
!   - for q(:) in the range (0.5,1.0), call PINVSTUDENT with 
!     P(:) = 2 * (1-q(:)).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Num_Constants,     only : macheps
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),  intent(in) :: p, df
!
    real(stnd), dimension(size(p)) :: pinvstudent_rv2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: a, b, c, d, x, y, pi, dfi, pinv
    real(stnd), dimension(size(p)) :: pinvvec
!
    integer(i4b) :: i, n
!
#ifdef _OPENMP
!
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(p),i4b)     ,   &
                    int(size(df),i4b)    ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:n)<=zero) .or. any(p(:n)>=one) .or. any(df(:n)<one) ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(i,x,y,a,b,c,d,pi,dfi,pinv)
!
    do i=1_i4b, n
!
        pi  = p(i)
        dfi = df(i)
!
        if ( abs(dfi-one)<=macheps ) then
!
!           EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
            x    = pi * pio2
            pinv = cos( x )/sin( x )
!
        else if ( abs(dfi-two)<=macheps ) then
!
!           EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
            pinv = sqrt( abs( two/( pi*(two - pi) ) - two ) )
!
        else
!
!           EXPANSION.
!
            a = one/( dfi - half )
            b = c48/( a*a )
            c = ( ( c20700*a/b - c98 )*a - c16)*a + c96_36
            d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*dfi
!
            x = d*pi
            y = x**(two/dfi)
!
            if ( y>c5_m2+a ) then
!
!               ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                x = pinvn_r( half*pi )
                y = x*x
!
                if ( dfi<five ) then
                    c = c + c0_3*( dfi - c4_5 )*( x + c0_6 )
                end if
!
                c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
                y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
                y = a*y*y
!
                if ( y>c2_m3 ) then
                    y = exp( y ) - one
                else
                    y = half*y*y + y
                end if
!
            else
!
                y = ( ( one/( ( (dfi + six)/(dfi*y) - c0_089*d - c0_822 )*(dfi + two)*three )     &
                        + half/(dfi + four) )*y - one )*(dfi + one)/(dfi + two) + one/y
!
            end if
!
            pinv = sqrt( dfi*y )
!
        end if
!
        pinvvec(i) = pinv
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvstudent_rv2(1_i4b:n) = pinvvec(1_i4b:n)
!
!
! END OF FUNCTION pinvstudent_rv2
! _______________________________
!
    end function pinvstudent_rv2
!
! =========================================================================================
!
    function pinvstudent_rm( p, df )
!
! Purpose
! _______
!
!   Evaluates the inverse of a modification of Student's t probability
!   distribution function.
!
!   PINVSTUDENT calculates the two-tail quantiles of Student's t-distribution,
!   that is a value x(i,j) such that the probability of the absolute value of t being
!   greater than x(i,j) is P(i,j), for i=1 to size(P,1) and j=1 to size(P,2). 
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:,:)
!          On entry, the probabilities. P(i,j) is the sum of the areas (equal) in both
!          tails of the t-distribution, for i=1 to size(P) and j=1 to size(P,2) .
!          P(i,j) must verify 0. < P(i,j) < 1, for i=1 to size(P,1) and j=1 to size(P,2) .
!
!   DF     (INPUT) real(stnd)
!          On entry, degrees of freedom of the t-distribution. DF must be greater
!          or equal to 1. . DF is not necessarily an integer.
!
!
! Further Details
! _______________
!
!   Note that PINVSTUDENT  does not provide the actual Student's t inverse.
!   For q(:,:) equal to the probabilities that a Student's t random variable is less
!   than x(:,:), that inverse can be obtained by the following rules:
!
!   - for q(:,:) in the range (0.0,0.5), call PINVSTUDENT with
!     P(:,:) = 2 * q(:,:) and negate the result x(:,:).
!   - for q(:,:) in the range (0.5,1.0), call PINVSTUDENT with 
!     P(:,:) = 2 * (1-q(:,:)).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Num_Constants,     only : macheps
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),  intent(in) :: p
    real(stnd)                ,  intent(in) :: df
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvstudent_rm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: a, b, c, cc, d, x, y
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvmat
!
    integer(i4b) :: i, j, n, m
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( p, 1 )
    n = size( p, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:m,:n)<=zero) .or. any(p(:m,:n)>=one) .or. df<one ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)
!
    if ( abs(df-one)<=macheps ) then
!
!       EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC), PRIVATE(i,j,x)
!
        do i=1_i4b, n
!
            do j=1_i4b, m
!
                x = p(j,i) * pio2
                pinvmat(j,i) = cos( x )/sin( x )
!
            end do
!
        end do
!
!$OMP END DO
!
    else if ( abs(df-two)<=macheps ) then
!
!       EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
!$OMP DO SCHEDULE(STATIC), PRIVATE(i,j,x)
!
        do i=1_i4b, n
!
            do j=1_i4b, m
!
                x = p(j,i)
                pinvmat(j,i) = sqrt( abs( two/( x*(two - x) ) - two ) )
!
            end do
!
        end do
!
!$OMP END DO
!
    else
!
!       EXPANSION.
!
!$OMP SINGLE
!
        a = one/( df - half )
        b = c48/( a*a )
        c = ( ( c20700*a/b - c98 )*a - c16)*a + c96_36
        d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*df
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC), PRIVATE(i,j,x,y,cc)
!
        do i=1_i4b, n
!
            do j=1_i4b, m
!
                x = d*p(j,i)
                y = x**(two/df)
!
                if ( y>c5_m2+a ) then
!
!                   ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                    x = pinvn_r( half*p(j,i) )
                    y = x*x
                    cc = c
!
                    if ( df<five ) then
                        cc = cc + c0_3*( df - c4_5 )*( x + c0_6 )
                    end if
!
                    cc = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + cc
                    y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/cc - y - three )/b + one )*x
                    y = a*y*y
!
                    if ( y>c2_m3 ) then
                        y = exp( y ) - one
                    else
                        y = half*y*y + y
                    end if
!
                else
!
                    y = ( ( one/( ( (df + six)/(df*y) - c0_089*d - c0_822 )*(df + two)*three )     &
                            + half/(df + four) )*y - one )*(df + one)/(df + two) + one/y
!
                end if
!
                pinvmat(j,i) = sqrt( df*y )
!
            end do
!
        end do
!
!$OMP END DO
!
     end if
!
!$OMP END PARALLEL
!
     pinvstudent_rm(1_i4b:m,1_i4b:n) = pinvmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION pinvstudent_rm
! ______________________________
!
    end function pinvstudent_rm
!
! =========================================================================================
!
    function pinvstudent_rm2( p, df )
!
! Purpose
! _______
!
!   Evaluates the inverse of a modification of Student's t probability
!   distribution function.
!
!   PINVSTUDENT calculates the two-tail quantiles of Student's t-distribution,
!   that is a value x(i,j) such that the probability of the absolute value of t being
!   greater than x(i,j) is P(i,j), for i=1 to size(P,1) and j=1 to size(P,2). 
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:,:)
!          On entry, the probabilities. P(i,j) is the sum of the areas (equal) in both
!          tails of the t-distribution, for i=1 to size(P) and j=1 to size(P,2) .
!          P(i,j) must verify 0. < P(i,j) < 1, for i=1 to size(P,1) and j=1 to size(P,2) .
!
!   DF     (INPUT) real(stnd), dimension(:,:)
!          On entry, degrees of freedom of the t-distribution. Any value in the array DF
!          must be greater or equal to 1, but is not necessarily an integer.
!
!          The shape of DF must verify:
!
!          - size(DF,1) = size(P,1)
!          - size(DF,2) = size(P,2) .
!
!
! Further Details
! _______________
!
!   Note that PINVSTUDENT  does not provide the actual Student's t inverse.
!   For q(:,:) equal to the probabilities that a Student's t random variable is less
!   than x(:,:), that inverse can be obtained by the following rules:
!
!   - for q(:,:) in the range (0.0,0.5), call PINVSTUDENT with
!     P(:,:) = 2 * q(:,:) and negate the result x(:,:).
!   - for q(:,:) in the range (0.5,1.0), call PINVSTUDENT with 
!     P(:,:) = 2  * (1-q(:,:)).
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Hill, G.W., 1970: Student's t-quantiles (Algorithm 396),
!                      Comm. A.C.M., vol.13, no10, 620-621
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, three, four, five, six, seven, pio2, half,    &
                                  c4_5, c6_3, c16, c36, c48, c94_5, c96_36, c98, c20700,        &
                                  c2_m3, c0_089, c5_m2, c0_3, c0_4, c0_6, c0_822
    use Num_Constants,     only : macheps
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),  intent(in) :: p, df
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvstudent_rm2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: a, b, c, d, x, y, pji, dfji, pinv
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvmat
!
    integer(i4b) :: i, j, n, m
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvstudent'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(p,1),i4b)     ,   &
                    int(size(df,1),i4b)    ,   &
                    name_proc )
!
    n =  assert_eq( int(size(p,2),i4b)     ,   &
                    int(size(df,2),i4b)    ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n*m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:m,:n)<=zero) .or. any(p(:m,:n)>=one) .or. any(df(:m,:n)<one) ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC), PRIVATE(i,j,x,y,a,b,c,d,pji,dfji,pinv)
!
    do i=1_i4b, n
!
        do j=1_i4b, m
!
            pji  = p(j,i)
            dfji = df(j,i)
!
            if ( abs(dfji-one)<=macheps ) then
!
!               EXACT INTEGRAL FOR 1 DEGREE OF FREEDOM.
!
                x    = pji * pio2
                pinv = cos( x )/sin( x )
!
            else if ( abs(dfji-two)<=macheps ) then
!
!               EXACT INTEGRAL FOR 2 DEGREES OF FREEDOM.
!
                pinv = sqrt( abs( two/( pji*(two - pji) ) - two ) )
!
            else
!
!               EXPANSION.
!
                a = one/( dfji - half )
                b = c48/( a*a )
                c = ( ( c20700*a/b - c98 )*a - c16)*a + c96_36
                d = ( ( c94_5/(b+c) - three )/b + one )*sqrt( a*pio2 )*dfji
!
                x = d*pji
                y = x**(two/dfji)
!
                if ( y>c5_m2+a ) then
!
!                   ASYMPTOTIC INVERSE EXPANSION ABOUT NORMAL.
!
                    x = pinvn_r( half*pji )
                    y = x*x
!
                    if ( dfji<five ) then
                        c = c + c0_3*( dfji - c4_5 )*( x + c0_6 )
                    end if
!
                    c = ( ( (c5_m2*d*x - five)*x - seven )*x - two )*x + b + c
                    y = ( ( ( ( (c0_4*y + c6_3)*y + c36 )*y + c94_5 )/c - y - three )/b + one )*x
                    y = a*y*y
!
                    if ( y>c2_m3 ) then
                        y = exp( y ) - one
                    else
                        y = half*y*y + y
                    end if
!
                else
!
                    y = ( ( one/( ( (dfji + six)/(dfji*y) - c0_089*d - c0_822 )*(dfji + two)*three )     &
                            + half/(dfji + four) )*y - one )*(dfji + one)/(dfji + two) + one/y
!
                end if
!
                pinv = sqrt( dfji*y )
!
            end if
!
            pinvmat(j,i) = pinv
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvstudent_rm2(1_i4b:m,1_i4b:n) = pinvmat(1_i4b:m,1_i4b:n)
!
!
! END OF FUNCTION pinvstudent_rm2
! _______________________________
!
    end function pinvstudent_rm2
!
!
! =========================================================================================
!                             CHI-SQUARED DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function probq_r( x2, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2 to infinity 
!   if UPPER is true or from zero to X2 if UPPER is false. In other words, if:
!
!   - UPPER = true  :    PROBQ = prob( U > X2 )  ,
!   - UPPER = false :    PROBQ = prob( U < X2 )  ,
!   
!   , for U = Chi-squared(NDF).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd)
!              On entry, upper or lower limit of integration. X2 must be greater
!              or equal to zero.
!
!   NDF        (INPUT) integer(i4b)
!              On entry, degrees of freedom of the chi-squared distribution. NDF must be greater
!              or equal to 1.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  :  probability to the right of X2 is calculated.
!              - UPPER = false  : probability to the left of X2 is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the chi-squared density is integrated.
!              - NDF is greater than NDF_MAX, a gaussian approximation is used.
!
!              The default is 40.
!
!
! Further Details
! _______________
!
!   If NDF<=NDF_MAX, the chi-squared distribution function is integrating by using formulae 
!   26.4.4 and 26.4.5 in reference (1), otherwise a normal approximation based on the 
!   Wilson-Hilferty transformation is used (see the reference (1), Formula 26.4.14).
!
!   This function works only for integer degrees of freedom. It may be faster than PROBQ2 or
!   PROBQ3 functions for the default value of NDF_MAX, but it is less accurate.
!
!   This function is adapted from
!
!   (1) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 26.4.4, 26.4.5 and 26.4.14),
!                                     New York, Dover Publications
!
!   (2) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, half, third, c4_5, c200
    use Char_Constants,    only : prob_error4
    use Utilities,         only : merror, arth, cumprod
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: x2
    real(stnd)             :: probq_r
!
    integer(i4b), intent(in)           :: ndf
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: df, a, s, u
!
    integer(i4b) :: n2, ndf_max2, k
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: sqrt2opi=two/2.506628274631001_stnd
!
    character(len=*),  parameter :: name_proc='probq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( ndf<1_i4b .or. x2<zero )  then
        call merror( name_proc//prob_error4 )
    end if
!
    if ( present(ndf_max) ) then
        ndf_max2 = ndf_max
    else
        ndf_max2 = 40_i4b
    end if
!
    if ( ndf<=ndf_max2 ) then
!
!       EXACT PROBABILITY FOR SMALL ndf.
!
        n2 = ndf/2_i4b
        k = n2 - 1_i4b
!
        if ( n2*2_i4b==ndf ) then
!
!           ndf EVEN.
!
            if ( x2<=c200 ) then
!
                u = half*x2
                a = exp(-u)
                s = one + sum( cumprod( u/arth( one, one, k ) ) )
                probq_r = a*s
!
            else
!
                probq_r = zero
!
            end if
!
        else
!
!           ndf ODD.
!
            u = sqrt( x2 )
            probq_r = two*probn_r( u, true )
!
            if ( ndf>1_i4b.and.x2<=c200 ) then
!
                a = exp(-half*x2)
                s = one + sum( cumprod( x2/arth( three, two, k ) ) )
                probq_r = probq_r + sqrt2opi*u*a*s
!
            end if
!
        end if
!
        if ( .not. upper )  probq_r = one - probq_r
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE ndf.
!
        df = real( ndf, stnd )
!
!       WILSON-HILFERTY FORMULA.
!
        a = one/( c4_5*df )
        u = (x2/df)**third - ( one - a )
        u = u/sqrt( a )
!
!       GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
        probq_r = probn_r( u, upper )
!
    end if
!
!
! END OF FUNCTION probq_r
! _______________________
!
    end function probq_r
!
! =========================================================================================
!
    function probq_rv( x2, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the  chi-squared distribution function from X2(i) to infinity 
!   if UPPER is true or from zero to X2(i) if UPPER is false, for i=1 to size(X2). In other words, if:
!
!   - UPPER = true  :    PROBQ( i ) = prob( U > X2(i) )  ,
!   - UPPER = false :    PROBQ( i ) = prob( U < X2(i) )  ,
!   
!   , for U = Chi-squared(NDF) and i=1 to size(X2).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration. X2(i) must be greater
!              or equal to zero for i=1 to size(X2).
!
!   NDF        (INPUT) integer(i4b)
!              On entry, degrees of freedom of the chi-squared distribution. NDF must be greater
!              or equal to 1.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the chi-squared density is integrated.
!              - NDF is greater than NDF_MAX, a gaussian approximation is used.
!
!              The default is 40.
!
!
! Further Details
! _______________
!
!   If NDF<=NDF_MAX, the chi-squared distribution function is integrating by using formulae 
!   26.4.4 and 26.4.5 in reference (1), otherwise a normal approximation based on the 
!   Wilson-Hilferty transformation is used (see the reference (1), Formula 26.4.14).
!
!   This function works only for integer degrees of freedom. It may be faster than PROBQ2 or
!   PROBQ3 functions for the default value of NDF_MAX, but it is less accurate.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 26.4.4, 26.4.5 and 26.4.14),
!                                     New York, Dover Publications
!
!   (2) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, half, third, c4_5, c200
    use Char_Constants,    only : prob_error4
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),       intent(in) :: x2
    real(stnd), dimension(size(x2))            :: probq_rv
!
    integer(i4b), intent(in)           :: ndf
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                          :: df, a, sqrta, s, u, x2i, probi
    real(stnd), dimension(size(x2))     :: prob
    real(stnd), dimension(max(ndf/2,1)) :: vec
!
    integer(i4b) :: n, n2, ndf_max2, k, i, j
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: sqrt2opi=two/2.506628274631001_stnd
!
    character(len=*),  parameter :: name_proc='probq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( ndf<1_i4b .or. any(x2(:n)<zero) ) then
        call merror( name_proc//prob_error4 )
    end if
!
    if ( present(ndf_max) ) then
        ndf_max2 = ndf_max
    else
        ndf_max2 = 40_i4b
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
    if ( ndf<=ndf_max2 ) then
!
!       EXACT PROBABILITY FOR SMALL ndf.
!
        n2 = ndf/2_i4b
        k = n2 - 1_i4b
!
        if ( n2*2_i4b==ndf ) then
!
!           ndf EVEN.
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,x2i,a,u,s,probi,vec), SCHEDULE(STATIC)
!
            do i = 1_i4b, n
!
                x2i = x2(i)
!
                if ( x2i<=c200 ) then
!
                    a = exp(-half*x2i)
                    u = half*x2i
!
                    vec(1_i4b) = u
                    s = one
!
                    do j=2_i4b, k
                        s = s + one
                        vec(j) = vec(j-1_i4b)*( u/s )              
                    end do
!
                    s = one + sum( vec(:k) )
                    probi = a*s
!
                else
!
                    probi = zero
!
                end if
!
                prob(i) = probi
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!           ndf ODD.
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,x2i,a,u,s,probi,vec), SCHEDULE(STATIC)
!
            do i = 1_i4b, n
!
                x2i   = x2(i)
                u     = sqrt( x2i )
                probi = two*probn_r( u, true )
!
                if ( x2i<=c200 .and. ndf>1_i4b ) then
!
                    a = exp(-half*x2i)
!
                    vec(1_i4b) = x2i/three
                    s = three
!
                    do j=2_i4b, k
                        s = s + two
                        vec(j) = vec(j-1_i4b)*( x2i/s )              
                    end do
!
                    s = one + sum( vec(:k) )
                    probi = probi + sqrt2opi*u*a*s
!
                end if
!
                prob(i) = probi
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
        if ( .not. upper )  prob(:n) = one - prob(:n)
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE ndf.
!
        df = real( ndf, stnd )
!
!       WILSON-HILFERTY FORMULA.
!
        a = one/( c4_5*df )
        sqrta = sqrt( a )
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,x2i,u), SCHEDULE(STATIC)
!
        do i = 1_i4b, n
!
            x2i = x2(i)
!
            u = (x2i/df)**third - ( one - a )
            u = u/sqrta
!
!           GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
            prob(i) = probn_r( u, upper )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
    probq_rv = prob(:n)
!
!
! END OF FUNCTION probq_rv
! ________________________
!
    end function probq_rv
!
! =========================================================================================
!
    function probq_rv2( x2, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the  chi-squared distribution function from X2(i) to infinity 
!   if UPPER is true or from zero to X2(i) if UPPER is false, for i=1 to size(X2). In other words, if:
!
!   - UPPER = true  :    PROBQ( i ) = prob( U > X2(i) )  ,
!   - UPPER = false :    PROBQ( i ) = prob( U < X2(i) )  ,
!   
!   , for U = Chi-squared(NDF( i )) and i=1 to size(X2).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration. X2(i) must be greater
!              or equal to zero for i=1 to size(X2).
!
!   NDF        (INPUT) integer(i4b), dimension(:)
!              On entry, degrees of freedom of the chi-squared distribution. Any value in
!              the array NDF must be greater or equal to 1.
!
!              The size of NDF must verify  size(NDF) = size(X2)  .
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the chi-squared density is integrated.
!              - NDF is greater than NDF_MAX, a gaussian approximation is used.
!
!              The default is 40.
!
!
! Further Details
! _______________
!
!   If NDF(i)<=NDF_MAX, the chi-squared distribution function is integrating by using formulae 
!   26.4.4 and 26.4.5 in reference (1), otherwise a normal approximation based on the 
!   Wilson-Hilferty transformation is used (see the reference (1), Formula 26.4.14).
!
!   This function works only for integer degrees of freedom. It may be faster than PROBQ2 or
!   PROBQ3 functions for the default value of NDF_MAX, but it is less accurate.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 26.4.4, 26.4.5 and 26.4.14),
!                                     New York, Dover Publications
!
!   (2) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, half, third, c4_5, c200
    use Char_Constants,    only : prob_error4, allocate_error
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),       intent(in) :: x2
    real(stnd), dimension(size(x2))            :: probq_rv2
!
    integer(i4b), dimension(:), intent(in)           :: ndf
    integer(i4b),               intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: df, a, sqrta, s, u, x2i, probi
    real(stnd), dimension(size(x2))       :: prob
    real(stnd), dimension(:), allocatable :: vec
!
    integer      :: iok
    integer(i4b) :: n, n2, ndf_max2, k, i, j, ndfi, ndfm
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: sqrt2opi=two/2.506628274631001_stnd
!
    character(len=*),  parameter :: name_proc='probq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(x2),i4b)   ,   &
                    int(size(ndf),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(ndf(:n)<1_i4b) .or. any(x2(:n)<zero) ) then
        call merror( name_proc//prob_error4 )
    end if
!
    if ( present(ndf_max) ) then
        ndf_max2 = ndf_max
    else
        ndf_max2 = 40_i4b
    end if
!
    ndfm = max( ndf_max2/2_i4b, 1_i4b )
!    ndfm = maxval( ndf(:n) )/2_i4b
!
    k  = ndfm - 1_i4b
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(iok,i,j,n2,k,ndfi,x2i,a,sqrta,u,s,probi,df,vec)
!
!   ALLOCATE WORK ARRAY.
!
    allocate( vec(ndfm), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        x2i  = x2(i)
        ndfi = ndf(i)
!
        if ( ndfi<=ndf_max2 ) then
!
!           EXACT PROBABILITY FOR SMALL ndfi.
!
            n2 = ndfi/2_i4b
            k  = n2 - 1_i4b
!
            if ( n2*2_i4b==ndfi ) then
!
!               ndfi EVEN.
!
                if ( x2i<=c200 ) then
!
                    a = exp(-half*x2i)
                    u = half*x2i
!
                    vec(1_i4b) = u
                    s = one
!
                    do j=2_i4b, k
                        s = s + one
                        vec(j) = vec(j-1_i4b)*( u/s )              
                    end do
!
                    s = one + sum( vec(:k) )
                    probi = a*s
!
                else
!
                    probi = zero
!
                end if
!
            else
!
!               ndfi ODD.
!
                u     = sqrt( x2i )
                probi = two*probn_r( u, true )
!
                if ( x2i<=c200 .and. ndfi>1_i4b ) then
!
                    a = exp(-half*x2i)
!
                    vec(1_i4b) = x2i/three
                    s = three
!
                    do j=2_i4b, k
                        s = s + two
                        vec(j) = vec(j-1_i4b)*( x2i/s )              
                    end do
!
                    s = one + sum( vec(:k) )
                    probi = probi + sqrt2opi*u*a*s
!
                end if
!
            end if
!
            if ( .not. upper )  probi = one - probi
!
        else
!
!           APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE ndfi.
!
            df = real( ndfi, stnd )
!
!           WILSON-HILFERTY FORMULA.
!
            a = one/( c4_5*df )
            sqrta = sqrt( a )
!
            u = (x2i/df)**third - ( one - a )
            u = u/sqrta
!
!           GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
            probi = probn_r( u, upper )
!
        end if
!
        prob(i) = probi
!
    end do
!
!$OMP END DO
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( vec )
!
!$OMP END PARALLEL
!
    probq_rv2 = prob(:n)
!
!
! END OF FUNCTION probq_rv2
! _________________________
!
    end function probq_rv2
!
! =========================================================================================
!
    function probq_rm( x2, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the  chi-squared distribution function from X2(i,j) to infinity 
!   if UPPER is true or from zero to X2(i,j) if UPPER is false,
!   for i=1 to size(X2,1) and j=1 to size(X2,2). In other words, if:
!
!   - UPPER = true  :    PROBQ( i, j ) = prob( U > X2(i,j) )  ,
!   - UPPER = false :    PROBQ( i, j ) = prob( U < X2(i,j) )  ,
!   
!   , for U = Chi-squared(NDF), i=1 to size(X2,1) and j=1 to size(X2,2).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration. X2(i,j) must be greater
!              or equal to zero for i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   NDF        (INPUT) integer(i4b)
!              On entry, degrees of freedom of the chi-squared distribution. NDF must be greater
!              or equal to 1.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the chi-squared density is integrated.
!              - NDF is greater than NDF_MAX, a gaussian approximation is used.
!
!              The default is 40.
!
!
! Further Details
! _______________
!
!   If NDF<=NDF_MAX, the chi-squared distribution function is integrating by using formulae 
!   26.4.4 and 26.4.5 in reference (1), otherwise a normal approximation based on the 
!   Wilson-Hilferty transformation is used (see the reference (1), Formula 26.4.14).
!
!   This function works only for integer degrees of freedom. It may be faster than PROBQ2 or
!   PROBQ3 functions for the default value of NDF_MAX, but it is less accurate.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 26.4.4, 26.4.5 and 26.4.14),
!                                     New York, Dover Publications
!
!   (2) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, half, third, c4_5, c200
    use Char_Constants,    only : prob_error4
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                  intent(in) :: x2
    real(stnd), dimension(size(x2,1),size(x2,2))            :: probq_rm
!
    integer(i4b), intent(in)           :: ndf
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                   :: df, a, sqrta, s, u, x2ji, probji
    real(stnd), dimension(size(x2,1),size(x2,2)) :: prob
    real(stnd), dimension(max(ndf/2,1))          :: vec
!
    integer(i4b) :: n, m, n2, ndf_max2, k, i, j, l
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: sqrt2opi=two/2.506628274631001_stnd
!
    character(len=*),  parameter :: name_proc='probq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( x2, 1 )
    n = size( x2, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( ndf<1_i4b .or. any(x2(:m,:n)<zero) ) then
        call merror( name_proc//prob_error4 )
    end if
!
    if ( present(ndf_max) ) then
        ndf_max2 = ndf_max
    else
        ndf_max2 = 40_i4b
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
    if ( ndf<=ndf_max2 ) then
!
!       EXACT PROBABILITY FOR SMALL ndf.
!
        n2 = ndf/2_i4b
        k = n2 - 1_i4b
!
        if ( n2*2_i4b==ndf ) then
!
!           ndf EVEN.
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,l,x2ji,a,u,s,probji,vec), SCHEDULE(STATIC)
!
            do i = 1_i4b, n
!
                do j = 1_i4b, m
!
                    x2ji = x2(j,i)
!
                    if ( x2ji<=c200 ) then
!
                        a = exp(-half*x2ji)
                        u = half*x2ji
!
                        vec(1_i4b) = u
                        s = one
!
                        do l=2_i4b, k
                            s = s + one
                            vec(l) = vec(l-1_i4b)*( u/s )              
                        end do
!          
                        s = one + sum( vec(:k) )
                        probji = a*s
!
                    else
!
                        probji = zero
!
                    end if
!
                    prob(j,i) = probji
!
                end do
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!           ndf ODD.
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,l,x2ji,a,u,s,probji,vec), SCHEDULE(STATIC)
!
            do i = 1_i4b, n
!
                do j = 1_i4b, m
!
                    x2ji   = x2(j,i)
                    u     = sqrt( x2ji )
                    probji = two*probn_r( u, true )
!
                    if ( x2ji<=c200 .and. ndf>1_i4b ) then
!
                        a = exp(-half*x2ji)
!
                        vec(1_i4b) = x2ji/three
                        s = three
!
                        do l=2_i4b, k
                            s = s + two
                            vec(l) = vec(l-1_i4b)*( x2ji/s )              
                        end do
!             
                        s = one + sum( vec(:k) )
                        probji = probji + sqrt2opi*u*a*s
!
                    end if
!
                    prob(j,i) = probji
!
                end do
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
        if ( .not. upper )  prob(:m,:n) = one - prob(:m,:n)
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE ndf.
!
        df = real( ndf, stnd )
!
!       WILSON-HILFERTY FORMULA.
!
        a = one/( c4_5*df )
        sqrta = sqrt( a )
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,x2ji,u), SCHEDULE(STATIC)
!
        do i = 1_i4b, n
!
            do j = 1_i4b, m
!
                x2ji = x2(j,i)
!
                u = (x2ji/df)**third - ( one - a )
                u = u/sqrta
!
!               GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
                prob(j,i) = probn_r( u, upper )
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
    probq_rm = prob(:m,:n)
!
!
! END OF FUNCTION probq_rm
! ________________________
!
    end function probq_rm
!
! =========================================================================================
!
    function probq_rm2( x2, ndf, upper, ndf_max )
!
! Purpose
! _______
!
!   Evaluates the  chi-squared distribution function from X2(i,j) to infinity 
!   if UPPER is true or from zero to X2(i,j) if UPPER is false,
!   for i=1 to size(X2,1) and j=1 to size(X2,2). In other words, if:
!
!   - UPPER = true  :    PROBQ( i, j ) = prob( U > X2(i,j) )  ,
!   - UPPER = false :    PROBQ( i, j ) = prob( U < X2(i,j) )  ,
!   
!   , for U = Chi-squared(NDF( i, j )), i=1 to size(X2,1) and j=1 to size(X2,2).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration. X2(i,j) must be greater
!              or equal to zero for i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   NDF        (INPUT) integer(i4b), dimension(:,:)
!              On entry, degrees of freedom of the chi-squared distribution. Any value in
!              the array NDF must be greater or equal to 1.
!
!              The shape of NDF must verify:
!
!              - size(NDF,1) = size(X2,1)
!              - size(NDF,2) = size(X2,2) .
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   NDF_MAX    (INPUT, OPTIONAL) integer(i4b)
!              On entry, if:
!
!              - NDF is lower or equal to NDF_MAX, the chi-squared density is integrated.
!              - NDF is greater than NDF_MAX, a gaussian approximation is used.
!
!              The default is 40.
!
!
! Further Details
! _______________
!
!   If NDF(i,j)<=NDF_MAX, the chi-squared distribution function is integrating by using formulae 
!   26.4.4 and 26.4.5 in reference (1), otherwise a normal approximation based on the 
!   Wilson-Hilferty transformation is used (see the reference (1), Formula 26.4.14).
!
!   This function works only for integer degrees of freedom. It may be faster than PROBQ2 or
!   PROBQ3 functions for the default value of NDF_MAX, but it is less accurate.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 26.4.4, 26.4.5 and 26.4.14),
!                                     New York, Dover Publications
!
!   (2) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, half, third, c4_5, c200
    use Char_Constants,    only : prob_error4, allocate_error
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                  intent(in) :: x2
    real(stnd), dimension(size(x2,1),size(x2,2))            :: probq_rm2
!
    integer(i4b), dimension(:,:), intent(in)           :: ndf
    integer(i4b),                 intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                   :: df, a, sqrta, s, u, x2ji, probji
    real(stnd), dimension(size(x2,1),size(x2,2)) :: prob
    real(stnd), dimension(:), allocatable        :: vec
!
    integer      :: iok
    integer(i4b) :: n, m, n2, ndf_max2, k, i, j, l, ndfji, ndfm
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: sqrt2opi=two/2.506628274631001_stnd
!
    character(len=*),  parameter :: name_proc='probq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(x2,1),i4b)   ,   &
                    int(size(ndf,1),i4b)  ,   &
                    name_proc )
!
    n =  assert_eq( int(size(x2,2),i4b)   ,   &
                    int(size(ndf,2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(ndf(:m,:n)<1_i4b) .or. any(x2(:m,:n)<zero) ) then
        call merror( name_proc//prob_error4 )
    end if
!
    if ( present(ndf_max) ) then
        ndf_max2 = ndf_max
    else
        ndf_max2 = 40_i4b
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    ndfm = max( ndf_max2/2_i4b, 1_i4b )
!
    k  = ndfm - 1_i4b
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par), PRIVATE(iok,i,j,n2,k,l,ndfji,x2ji,a,sqrta,u,s,probji,df,vec)
!
!   ALLOCATE WORK ARRAY.
!
    allocate( vec(ndfm), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!$OMP DO SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        do j = 1_i4b, m
!
            x2ji  = x2(j,i)
            ndfji = ndf(j,i)
!
            if ( ndfji<=ndf_max2 ) then
!
!               EXACT PROBABILITY FOR SMALL ndfji.
!
                n2 = ndfji/2_i4b
                k = n2 - 1_i4b
!
                if ( n2*2_i4b==ndfji ) then
!
!                   ndfji EVEN.
!
                    if ( x2ji<=c200 ) then
!
                        a = exp(-half*x2ji)
                        u = half*x2ji
!
                        vec(1_i4b) = u
                        s = one
!
                        do l=2_i4b, k
                            s = s + one
                            vec(l) = vec(l-1_i4b)*( u/s )              
                        end do
!
                        s = one + sum( vec(:k) )
                        probji = a*s
!
                    else
!
                        probji = zero
!
                    end if
!
                else
!
!                   ndfji ODD.
!
                    u     = sqrt( x2ji )
                    probji = two*probn_r( u, true )
!
                    if ( x2ji<=c200 .and. ndfji>1_i4b ) then
!
                        a = exp(-half*x2ji)
!
                        vec(1_i4b) = x2ji/three
                        s = three
!
                        do l=2_i4b, k
                            s = s + two
                            vec(l) = vec(l-1_i4b)*( x2ji/s )              
                        end do
!
                        s = one + sum( vec(:k) )
                        probji = probji + sqrt2opi*u*a*s
!
                    end if
!
                end if
!
                if ( .not. upper )  probji = one - probji
!
            else
!
!               APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE ndfji.
!
                df = real( ndfji, stnd )
!
!               WILSON-HILFERTY FORMULA.
!
                a = one/( c4_5*df )
                sqrta = sqrt( a )
!
                u = (x2ji/df)**third - ( one - a )
                u = u/sqrta
!
!               GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
                probji = probn_r( u, upper )
!
            end if
!
            prob(j,i) = probji
!
        end do
!
    end do
!
!$OMP END DO
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( vec )
!
!$OMP END PARALLEL
!
    probq_rm2 = prob(:m,:n)
!
!
! END OF FUNCTION probq_rm2
! _________________________
!
    end function probq_rm2
!
! =========================================================================================
!
    function probq2_r( x2, df, upper, df_max, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2 to infinity 
!   if UPPER is true or from zero to X2 if UPPER is false. In other words, if:
!
!   - UPPER = true  :    PROBQ2 = prob( U > X2 )  ,
!   - UPPER = false :    PROBQ2 = prob( U <= X2 ) ,
!   
!   , for U = Chi-squared(DF).
!
!   PROBQ2 computes the probability that a random variable which follows the chi-squared
!   distribution with DF degrees of freedom is less than or equal to X2 (if UPPER is set to false)
!   or greater than to X2 (if UPPER is set to true).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd)
!              On entry, upper or lower limit of integration. X2 must be greater
!              or equal to zero.
!
!   DF         (INPUT) real(stnd)
!              On entry, degrees of freedom of the chi-squared distribution. DF must be greater
!              or equal to 0.5 and less than or equal to 200 000.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series expansion of the incomplete Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is faster than PROBQ3 function, but is less accurate.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                     Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
    use Logical_Constants, only : false
    use Reals_Constants,   only : zero, third, half, one, two, c4_5,    &
                                  c13, c25, c100
    use Num_Constants,     only : macheps, machtiny
    use Char_Constants,    only : prob_error14, prob_error16
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: x2, df
    real(stnd), intent(in), optional :: df_max
    real(stnd)                       :: probq2_r
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: df_max2, x, x1, p, lngamp, w, w1, b, elimit
!
    integer(i4b) :: i, maxit
!
    logical(lgl) :: fail
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: df_limit= 2.e5_stnd, x2big = 1.e+8_stnd
!
    character(len=*),  parameter :: name_proc='probq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( x2<zero ) then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( df<half .or. df>df_limit ) then
        call merror( name_proc//prob_error16 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
    if ( df<=df_max2 ) then
!
!       EXACT PROBABILITY FOR SMALL df .
!
        if ( x2>x2big ) then
!
            if ( upper )  then
                w = zero
            else
                w = one
            end if
!
        else
!
!           INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
            x = half*x2    
            p = half*df
!
            lngamp = lngamma_r( p )
            w = max( half*p, c13 )
!
!           DETERMINE MACHINE DEPENDANT CONSTANT.
!
            elimit = log( machtiny )
!
            if ( x<w ) then
!
                if ( df>c25 .and. x2<two ) then
!
                    if ( upper )  then
                        w = one
                    else
                        w = zero
                    end if
!
                else
!
!                   USE PEARSON'S SERIES EXPANSION.
!                   CALCULATE USING EQUATION NO. 6.5.29
!
                    w  = one/p
                    w1 = w
                    b  = p
!
                    do i = 1_i4b, maxit
!
                        b = b + one
                        w1 = w1*x/b
!
                        if ( w1<=macheps*w ) exit
!
                        w = w + w1
!
                    end do
!
                    if ( fail .and. i>maxit ) then
!  
                        w = -one
!  
                    else
!
                        w1 = p*log( x ) - x - lngamp          
                        w1 = w1 + log( w )
!
!                       TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                        if ( w1>=elimit ) then
                            w = exp( w1 )
                        else
                            w = zero
                        end if
!
                        if ( upper )  then
                            w = one - w
                        end if
!
                    end if
!
                end if
!
            else
!
!               USE ALTERNATE PEARSON'S SERIES EXPANSION.
!               CALCULATE USING EQUATION NO. 6.5.32
!
                x1 = one/x
                b  = p - one
                w1 = b*x1
                w  = w1 + one
!
                do i = 2_i4b, maxit
!
                    b = b - one
                    w1 = w1*b*x1
!
                    if ( w1<=macheps*w ) exit
!
                    w = w + w1
!
                end do
!
                if ( fail .and. i>maxit ) then
!  
                    w = -one
!  
                else
!
                    w1 = (p-one)*log(x) - x - lngamp         
                    w1 = w1 + log( w )
!
!                   TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                    if ( w1>=elimit ) then
                        w = exp( w1 )
                    else
                        w = zero
                    end if
!
                    if ( .not.upper )  then
                        w = one - w
                    end if
!
                end if
!
            end if
!
        end if
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!       WILSON-HILFERTY FORMULA .
!
        p = one/( c4_5*df )
        x = (x2/df)**third - ( one - p )
        x = x/sqrt( p )
!
!       GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
        w = probn_r( x, upper )
!
    end if
!
    probq2_r = w
!
!
! END OF FUNCTION probq2_r
! ________________________
!
    end function probq2_r
!
! =========================================================================================
!
    function probq2_rv( x2, df, upper, df_max, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2(i) to infinity 
!   if UPPER is true or from zero to X2(i) if UPPER is false, for i=1 to size(X2). In other words, if:
!
!   - UPPER = true  :    PROBQ2( i ) = prob( U > X2(i) )  ,
!   - UPPER = false :    PROBQ2( i ) = prob( U <= X2(i) ) ,
!   
!   , for U = Chi-squared(DF) and i=1 to size(X2).
!
!   PROBQ2 computes the probabilities that a random variable (vector) which 
!   follows the chi-squared distribution with DF degrees of freedom is less than or
!   equal to X2(:) (if UPPER is set to false) or greater than to X2(:) (if UPPER is
!   set to true).
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration. X2(i) must be greater
!              or equal to zero for i=1 to size(X2).
!
!   DF         (INPUT) real(stnd)
!              On entry, degrees of freedom of the chi-squared distribution. DF must be greater
!              or equal to 0.5 and less than or equal to 200 000.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series expansion of the incomplete Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is faster than PROBQ3 function, but is less accurate.
!
!   This functon is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                     Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Reals_Constants,   only : zero, third, half, one, two, c4_5,    &
                                  c13, c25, c100
    use Num_Constants,     only : macheps, machtiny
    use Char_Constants,    only : prob_error14, prob_error16
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(in)           :: x2
    real(stnd),               intent(in)           :: df
    real(stnd),               intent(in), optional :: df_max
    real(stnd), dimension(size(x2))                :: probq2_rv
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: df_max2, x, x1, x2j, xlim, p, sqrtp, &
                                       lngamp, w, w1, b, elimit
    real(stnd), dimension(size(x2)) :: prob
!
    integer(i4b) :: n, i, j, maxit
!
    logical(lgl) :: fail
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: df_limit= 2.e5_stnd, x2big = 1.e+8_stnd
!
    character(len=*),  parameter :: name_proc='probq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:n)<zero) )  then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( df<half .or. df>df_limit )  then
        call merror( name_proc//prob_error16 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
    if ( df<=df_max2 ) then
!
!       DETERMINE MACHINE DEPENDANT CONSTANT.
!
        elimit = log( machtiny )
!
        p = half*df
!
        lngamp = lngamma_r( p )
        xlim = max( half*p, c13 )
!
!       EXACT PROBABILITY FOR SMALL df .
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(j,i,w,w1,x,x1,x2j,b), SCHEDULE(STATIC)
!
        do j = 1_i4b, n
!
            x2j = x2(j)
!
            if ( x2j>x2big ) then
!
                if ( upper )  then
                    w = zero
                else
                    w = one
                end if
!
            else
!
!               INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
                x = half*x2j 
!
                if ( x<xlim ) then
!
                    if ( df>c25 .and. x2j<two ) then
!
                        if ( upper )  then
                            w = one
                        else
                            w = zero
                        end if
!
                    else
!
!                       USE PEARSON'S SERIES EXPANSION.
!                       CALCULATE USING EQUATION NO. 6.5.29
!
                        w  = one/p
                        w1 = w
                        b  = p
!
                        do i = 1_i4b, maxit
!
                            b = b + one
                            w1 = w1*x/b
!
                            if ( w1<=macheps*w ) exit
!
                            w = w + w1
!
                        end do
!
                        if ( fail .and. i>maxit ) then
!  
                            w = -one
!  
                        else
!
                            w1 = p*log( x ) - x - lngamp          
                            w1 = w1 + log( w )
!
!                           TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                            if ( w1>=elimit ) then
                                w = exp( w1 )
                            else
                                w = zero
                            end if
!
                            if ( upper )  then
                                w = one - w
                            end if
!
                        end if
!
                    end if
!
                else
!
!                   USE ALTERNATE PEARSON'S SERIES EXPANSION.
!                   CALCULATE USING EQUATION NO. 6.5.32
!
                    x1 = one/x
                    b  = p - one
                    w1 = b*x1
                    w  = w1 + one
!
                    do i = 2_i4b, maxit
!
                        b = b - one
                        w1 = w1*b*x1
!
                        if ( w1<=macheps*w ) exit
!
                        w = w + w1
!
                    end do
!
                    if ( fail .and. i>maxit ) then
!  
                        w = -one
!  
                    else
!
                        w1 = (p-one)*log(x) - x - lngamp         
                        w1 = w1 + log( w )
!
!                       TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                        if ( w1>=elimit ) then
                            w = exp( w1 )
                        else
                            w = zero
                        end if
!
                        if ( .not.upper )  then
                            w = one - w
                        end if
!
                    end if
!
                end if
!
            end if
!
            prob(j) = w
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!       WILSON-HILFERTY FORMULA .
!
        p = one/( c4_5*df )
        sqrtp = sqrt( p )
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(j,x2j,w), SCHEDULE(STATIC)
!
        do j = 1_i4b, n
!
            x2j = x2(j)
!
            w = (x2j/df)**third - ( one - p )
            w = w/sqrtp
!
!           GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
            prob(j) = probn_r( w, upper )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
   probq2_rv = prob(:n)
!
!
! END OF FUNCTION probq2_rv
! _________________________
!
    end function probq2_rv
!
! =========================================================================================
!
    function probq2_rv2( x2, df, upper, df_max, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2(i) to infinity 
!   if UPPER is true or from zero to X2(i) if UPPER is false, for i=1 to size(X2). In other words, if:
!
!   - UPPER = true  :    PROBQ2( i ) = prob( U > X2(i) )   ,
!   - UPPER = false :    PROBQ2( i ) = prob( U <= X2(i) )  ,
!   
!   , for U = Chi-squared(DF( i )) and i=1 to size(X2).
!
!   PROBQ2 computes the probabilities that a random variable (vector) which 
!   follows the chi-squared distribution with DF(:) degrees of freedom is less than or
!   equal to X2(:) (if UPPER is set to false) or greater than to X2(:) (if UPPER is
!   set to true).
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration. X2(i) must be greater
!              or equal to zero for i=1 to size(X2).
!
!   DF         (INPUT) real(stnd), dimension(:)
!              On entry, degrees of freedom of the chi-squared distribution. Any value
!              in the array DF must be greater or equal to 0.5 and less than or equal
!              to 200 000.
!
!              The size of DF must verify  size(DF) = size(X2).
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series expansion of the incomplete Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only for the values of DF(:) less than
!              or equal to DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF(i)<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is faster than PROBQ3 function, but is less accurate.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                     Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Reals_Constants,   only : zero, third, half, one, two, c4_5,    &
                                  c13, c25, c100
    use Num_Constants,     only : macheps, machtiny
    use Char_Constants,    only : prob_error14, prob_error16
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(in)           :: x2, df
    real(stnd),               intent(in), optional :: df_max
    real(stnd), dimension(size(x2))                :: probq2_rv2
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: df_max2, x, x1, x2j, dfj, xlim, p,   &
                                       lngamp, w, w1, b, elimit
    real(stnd), dimension(size(x2)) :: prob
!
    integer(i4b) :: n, i, j, maxit
!
    logical(lgl) :: fail
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: df_limit= 2.e5_stnd, x2big = 1.e+8_stnd
!
    character(len=*),  parameter :: name_proc='probq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(x2),i4b)   ,   &
                    int(size(df),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:n)<zero) ) then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( any(df(:n)<half) .or. any(df(:n)>df_limit) ) then
        call merror( name_proc//prob_error16 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!   DETERMINE MACHINE DEPENDANT CONSTANT.
!
    elimit = log( machtiny )
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(j,i,w,w1,x,x1,xlim,x2j,dfj,b,p,lngamp), SCHEDULE(STATIC)
!
    do j = 1_i4b, n
!
        x2j = x2(j)
        dfj = df(j)
!
        if ( dfj<=df_max2 ) then
!
!           EXACT PROBABILITY FOR SMALL df .
!
            if ( x2j>x2big ) then
!
                if ( upper )  then
                    w = zero
                else
                    w = one
                end if
!
            else
!
!               INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
                x = half*x2j 
                p = half*dfj
!
                lngamp = lngamma_r( p )
                xlim = max( half*p, c13 )
!
                if ( x<xlim ) then
!
                    if ( dfj>c25 .and. x2j<two ) then
!
                        if ( upper )  then
                            w = one
                        else
                            w = zero
                        end if
!
                    else
!
!                       USE PEARSON'S SERIES EXPANSION.
!                       CALCULATE USING EQUATION NO. 6.5.29
!
                        w  = one/p
                        w1 = w
                        b  = p
!
                        do i = 1_i4b, maxit
!
                            b = b + one
                            w1 = w1*x/b
!
                            if ( w1<=macheps*w ) exit
!
                            w = w + w1
!
                        end do
!
                        if ( fail .and. i>maxit ) then
!  
                            w = -one
!  
                        else
!
                            w1 = p*log( x ) - x - lngamp          
                            w1 = w1 + log( w )
!
!                           TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                            if ( w1>=elimit ) then
                                w = exp( w1 )
                            else
                                w = zero
                            end if
!
                            if ( upper )  then
                                w = one - w
                            end if
!
                        end if
!
                    end if
!
                else
!
!                   USE ALTERNATE PEARSON'S SERIES EXPANSION.
!                   CALCULATE USING EQUATION NO. 6.5.32
!
                    x1 = one/x
                    b  = p - one
                    w1 = b*x1
                    w  = w1 + one
!
                    do i = 2_i4b, maxit
!
                        b = b - one
                        w1 = w1*b*x1
!
                        if ( w1<=macheps*w ) exit
!
                        w = w + w1
!
                    end do
!
                    if ( fail .and. i>maxit ) then
!  
                        w = -one
!  
                    else
!
                        w1 = (p-one)*log(x) - x - lngamp         
                        w1 = w1 + log( w )
!
!                       TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                        if ( w1>=elimit ) then
                            w = exp( w1 )
                        else
                            w = zero
                        end if
!
                        if ( .not.upper )  then
                            w = one - w
                        end if
!
                    end if
!
                end if
!
            end if
!
        else
!
!           APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!           WILSON-HILFERTY FORMULA .
!
            p = one/( c4_5*dfj )
!
            b = (x2j/dfj)**third - ( one - p )
            b = b/sqrt( p )
!
!           GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
            w = probn_r( b, upper )
!
        end if
!
        prob(j) = w
!
    end do
!
!$OMP END PARALLEL DO
!
   probq2_rv2 = prob(:n)
!
!
! END OF FUNCTION probq2_rv2
! __________________________
!
    end function probq2_rv2
!
! =========================================================================================
!
    function probq2_rm( x2, df, upper, df_max, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2(i,j) to infinity 
!   if UPPER is true or from zero to X2(i,j) if UPPER is false, for i=1 to size(X2,1)
!   and j=1 to size(X2,2). In other words, if:
!
!   - UPPER = true  :    PROBQ2( i,j ) = prob( U >  X2(i,j) )  ,
!   - UPPER = false :    PROBQ2( i,j ) = prob( U <= X2(i,j) )  ,
!   
!   , for U = Chi-squared(DF), i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   PROBQ2 computes the probabilities that a random variable (matrix) which follows the chi-squared
!   distribution with DF degrees of freedom is less than or equal to X2(:,:) (if UPPER is set to false)
!   or greater than to X2(:,:) (if UPPER is set to true).
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration. X2(i,j) must be greater
!              or equal to zero for i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   DF         (INPUT) real(stnd)
!              On entry, degrees of freedom of the chi-squared distribution. DF must be greater
!              or equal to 0.5 and less than or equal to 200 000.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series expansion of the incomplete Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is faster than PROBQ3 function, but is less accurate.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                     Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Reals_Constants,   only : zero, third, half, one, two, c4_5,    &
                                  c13, c25, c100
    use Num_Constants,     only : macheps, machtiny
    use Char_Constants,    only : prob_error14, prob_error16
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: x2
    real(stnd),                 intent(in)           :: df
    real(stnd),                 intent(in), optional :: df_max
    real(stnd), dimension(size(x2,1),size(x2,2))     :: probq2_rm
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                   :: df_max2, x, x1, x2kj, xlim, p, sqrtp, &
                                                    lngamp, w, w1, b, elimit
    real(stnd), dimension(size(x2,1),size(x2,2)) :: prob
!
    integer(i4b) :: m, n, i, j, k, maxit
!
    logical(lgl) :: fail
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: df_limit= 2.e5_stnd, x2big = 1.e+8_stnd
!
    character(len=*),  parameter :: name_proc='probq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( x2, 1 )
    n = size( x2, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:m,:n)<zero) )  then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( df<half .or. df>df_limit )  then
        call merror( name_proc//prob_error16 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!
    if ( df<=df_max2 ) then
!
!       DETERMINE MACHINE DEPENDANT CONSTANT.
!
        elimit = log( machtiny )
!
        p = half*df
!
        lngamp = lngamma_r( p )
        xlim = max( half*p, c13 )
!
!       EXACT PROBABILITY FOR SMALL df .
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(j,k,i,w,w1,x,x1,x2kj,b), SCHEDULE(STATIC)
!
        do j = 1_i4b, n
!
            do k = 1_i4b, m
!
                x2kj = x2(k,j)
!
                if ( x2kj>x2big ) then
!
                    if ( upper )  then
                        w = zero
                    else
                        w = one
                    end if
!
                else
!
!                   INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
                    x = half*x2kj 
!
                    if ( x<xlim ) then
!
                        if ( df>c25 .and. x2kj<two ) then
!
                            if ( upper )  then
                                w = one
                            else
                                w = zero
                            end if
!
                        else
!
!                           USE PEARSON'S SERIES EXPANSION.
!                           CALCULATE USING EQUATION NO. 6.5.29
!
                            w  = one/p
                            w1 = w
                            b  = p
!
                            do i = 1_i4b, maxit
!
                                b = b + one
                                w1 = w1*x/b
!
                                if ( w1<=macheps*w ) exit
!
                                w = w + w1
!
                            end do
!
                            if ( fail .and. i>maxit ) then
!  
                                w = -one
!  
                            else
!
                                w1 = p*log( x ) - x - lngamp          
                                w1 = w1 + log( w )
!
!                               TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                                if ( w1>=elimit ) then
                                    w = exp( w1 )
                                else
                                    w = zero
                                end if
!
                                if ( upper )  then
                                    w = one - w
                                end if
!
                            end if
!
                        end if
!
                    else
!
!                       USE ALTERNATE PEARSON'S SERIES EXPANSION.
!                       CALCULATE USING EQUATION NO. 6.5.32
!
                        x1 = one/x
                        b  = p - one
                        w1 = b*x1
                        w  = w1 + one
!
                        do i = 2_i4b, maxit
!
                            b = b - one
                            w1 = w1*b*x1
!
                            if ( w1<=macheps*w ) exit
!
                            w = w + w1
!
                        end do
!
                        if ( fail .and. i>maxit ) then
!  
                            w = -one
!  
                        else
!
                            w1 = (p-one)*log(x) - x - lngamp         
                            w1 = w1 + log( w )
!
!                           TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                            if ( w1>=elimit ) then
                                w = exp( w1 )
                            else
                                w = zero
                            end if
!
                            if ( .not.upper )  then
                                w = one - w
                            end if
!
                        end if
!
                    end if
!
                end if
!
                prob(k,j) = w
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!       WILSON-HILFERTY FORMULA .
!
        p = one/( c4_5*df )
        sqrtp = sqrt( p )
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(j,k,x2kj,w), SCHEDULE(STATIC)
!
        do j = 1_i4b, n
!
            do k = 1_i4b, m
!
                x2kj = x2(k,j)
!
                w = (x2kj/df)**third - ( one - p )
                w = w/sqrtp
!
!               GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
                prob(k,j) = probn_r( w, upper )
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
    probq2_rm = prob(:m,:n)
!
!
! END OF FUNCTION probq2_rm
! _________________________
!
    end function probq2_rm
!
! =========================================================================================
!
    function probq2_rm2( x2, df, upper, df_max, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2(i,j) to infinity 
!   if UPPER is true or from zero to X2(i,j) if UPPER is false, for i=1 to size(X2,1)
!   and j=1 to size(X2,2). In other words, if:
!
!   - UPPER = true  :    PROBQ2( i,j ) = prob( U >  X2(i,j) )  ,
!   - UPPER = false :    PROBQ2( i,j ) = prob( U <= X2(i,j) )  ,
!   
!   , for U = Chi-squared(DF( i,j )), i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   PROBQ2 computes the probabilities that a random variable (matrix) which follows the chi-squared
!   distribution with DF(:,:) degrees of freedom is less than or equal to X2(:,:) (if UPPER is set to false)
!   or greater than to X2(:,:) (if UPPER is set to true).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration. X2(i,j) must be greater
!              or equal to zero for i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   DF         (INPUT) real(stnd), dimension(:,:)
!              On entry, degrees of freedom of the chi-squared distribution. Any value
!              in the array DF must be greater or equal to 0.5 and less than or equal
!              to 200 000.
!
!              The shape of DF must verify:
!
!              - size(DF,1) = size(X2,1)
!              - size(DF,2) = size(X2,2).
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series expansion of the incomplete Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only for the values of DF(:,:) less than
!              or equal to DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF(i,j)<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is faster than PROBQ3 function, but is less accurate.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                                   Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                     (formulae 6.5.29, 6.5.32 and 26.4.14), New York,
!                                     Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
    use Reals_Constants,   only : zero, third, half, one, two, c4_5,    &
                                  c13, c25, c100
    use Num_Constants,     only : macheps, machtiny
    use Char_Constants,    only : prob_error14, prob_error16
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: x2, df
    real(stnd),                 intent(in), optional :: df_max
    real(stnd), dimension(size(x2,1),size(x2,2))     :: probq2_rm2
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                   :: df_max2, x, x1, x2kj, dfkj, xlim, p, &
                                                    lngamp, w, w1, b, elimit
    real(stnd), dimension(size(x2,1),size(x2,2)) :: prob
!
    integer(i4b) :: m, n, i, j, k, maxit
!
    logical(lgl) :: fail
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter :: df_limit= 2.e5_stnd, x2big = 1.e+8_stnd
!
    character(len=*),  parameter :: name_proc='probq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(x2,1),i4b)   ,   &
                    int(size(df,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(x2,2),i4b)   ,   &
                    int(size(df,2),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:m,:n)<zero) )  then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( any(df(:m,:n)<half) .or. any(df(:m,:n)>df_limit) )  then
        call merror( name_proc//prob_error16 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    maxit = 1000_i4b
    if ( present(maxiter) ) then
        maxit = max( maxiter, 10_i4b )
    end if
!
    fail = false
    if ( present(failure) ) then
        fail = failure
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!   DETERMINE MACHINE DEPENDANT CONSTANT.
!
    elimit = log( machtiny )
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(j,k,i,w,w1,x,x1,xlim,x2kj,dfkj,b,p,lngamp), SCHEDULE(STATIC)
!
    do j = 1_i4b, n
!
        do k = 1_i4b, m
!
            x2kj = x2(k,j)
            dfkj = df(k,j)
!
            if ( dfkj<=df_max2 ) then
!
!               EXACT PROBABILITY FOR SMALL df .
!
                if ( x2kj>x2big ) then
!
                    if ( upper )  then
                        w = zero
                    else
                        w = one
                    end if
!
                else
!
!                   INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
                    x = half*x2kj 
                    p = half*dfkj
!
                    lngamp = lngamma_r( p )
                    xlim = max( half*p, c13 )
!
                    if ( x<xlim ) then
!
                        if ( dfkj>c25 .and. x2kj<two ) then
!
                            if ( upper )  then
                                w = one
                            else
                                w = zero
                            end if
!
                        else
!
!                           USE PEARSON'S SERIES EXPANSION.
!                           CALCULATE USING EQUATION NO. 6.5.29
!
                            w  = one/p
                            w1 = w
                            b  = p
!
                            do i = 1_i4b, maxit
!
                                b = b + one
                                w1 = w1*x/b
!
                                if ( w1<=macheps*w ) exit
!
                                w = w + w1
!
                            end do
!
                            if ( fail .and. i>maxit ) then
!  
                                w = -one
!  
                            else
!
                                w1 = p*log( x ) - x - lngamp          
                                w1 = w1 + log( w )
!
!                               TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                                if ( w1>=elimit ) then
                                    w = exp( w1 )
                                else
                                    w = zero
                                end if
!
                                if ( upper )  then
                                    w = one - w
                                end if
!
                            end if
!
                        end if
!
                    else
!
!                       USE ALTERNATE PEARSON'S SERIES EXPANSION.
!                       CALCULATE USING EQUATION NO. 6.5.32
!
                        x1 = one/x
                        b  = p - one
                        w1 = b*x1
                        w  = w1 + one
!
                        do i = 2_i4b, maxit
!
                            b = b - one
                            w1 = w1*b*x1
!
                            if ( w1<=macheps*w ) exit
!
                            w = w + w1
!
                        end do
!
                        if ( fail .and. i>maxit ) then
!  
                            w = -one
!  
                        else
!
                            w1 = (p-one)*log(x) - x - lngamp         
                            w1 = w1 + log( w )
!
!                           TAKE CARE OF UNDERFLOW IN INTRINSIC EXPONENTIAL FUNCTION.
!
                            if ( w1>=elimit ) then
                                w = exp( w1 )
                            else
                                w = zero
                            end if
!
                            if ( .not.upper )  then
                                w = one - w
                            end if
!
                        end if
!
                    end if
!
                end if
!
            else
!
!               APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!               WILSON-HILFERTY FORMULA .
!
                p = one/( c4_5*dfkj )
!
                b = (x2kj/dfkj)**third - ( one - p )
                b = b/sqrt( p )
!
!               GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
                w = probn_r( b, upper )
!
            end if
!
            prob(k,j) = w
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probq2_rm2 = prob(:m,:n)
!
!
! END OF FUNCTION probq2_rm2
! __________________________
!
    end function probq2_rm2
!
! =========================================================================================
!
    function probq3_r( x2, df, upper, df_max, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2 to infinity 
!   if UPPER is true or from zero to X2 if UPPER is false. In other words, if:
!
!   - UPPER = true  :    PROBQ3 = prob( U > X2 )   ,
!   - UPPER = false :    PROBQ3 = prob( U <= X2 )  ,
!   
!   , for U = Chi-squared(DF).
!
!   PROBQ3 computes the probability that a random variable which follows the chi-squared
!   distribution with DF degrees of freedom is less than or equal to X2 (if UPPER is set to false)
!   or greater than to X2 (if UPPER is set to true).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd)
!              On entry, upper or lower limit of integration. X2 must be greater
!              or equal to zero.
!
!   DF         (INPUT) real(stnd)
!              On entry, degrees of freedom of the chi-squared distribution.
!              DF must be greater or equal to 0.5.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   ACU        (INPUT, OPTIONAL) real(stnd)
!              On entry, the desired accuracy of the result if DF<=DF_MAX (e.g. if the incomplete 
!              Gamma integral is used). If l decimal places of accuracy are required
!              then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!              ACU should not be set smaller than the machine precision since the stated accuracy
!              cannot be attained. In that case the machine precision is used instead.
!              The default value for ACU is epsilon( ACU ).
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series or continued fraction expansion of the incomplete 
!              Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is more accurate than PROBQ and PROBQ2 functions, but it is slower.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                        Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3)  Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, third, half, one, c4_5, c100
    use Char_Constants,    only : prob_error14, prob_error15
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: x2, df
    real(stnd), intent(in), optional :: df_max, acu
    real(stnd)                       :: probq3_r
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: df_max2, x, p, prob
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probq3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( x2<zero ) then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( df<half ) then
        call merror( name_proc//prob_error15 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    if ( df<=df_max2 ) then
!
!       EXACT PROBABILITY FOR SMALL df .
!
!       INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
        x = half*x2    
        p = half*df
!
        prob = probgamma2_r( x, p, acu=acu, maxiter=maxiter, failure=failure )
!        prob = probgamma_r( x, p, acu=acu, maxiter=maxiter, failure=failure )
!
        if ( upper .and. prob/=-one )  prob = one - prob
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!       WILSON-HILFERTY FORMULA .
!
        p = one/( c4_5*df )
        x = (x2/df)**third - ( one - p )
        x = x/sqrt( p )
!
!       GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
        prob = probn_r( x, upper )
!
    end if
!
    probq3_r = prob
!
!
! END OF FUNCTION probq3_r
! ________________________
!
    end function probq3_r
!
! =========================================================================================
!
    function probq3_rv( x2, df, upper, df_max, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2(i) to infinity 
!   if UPPER is true or from zero to X2(i) if UPPER is false, for i=1 to size(X2). In other words, if:
!
!   - UPPER = true  :    PROBQ3( i ) = prob( U > X2(i) )  ,
!   - UPPER = false :    PROBQ3( i ) = prob( U <= X2(i) )  ,
!   
!   , for U = Chi-squared(DF) and i=1 to size(X2).
!
!   PROBQ3 computes the probabilities that a random variable (vector) which 
!   follows the chi-squared distribution with DF degrees of freedom is less than or 
!   equal to X2(:) (if UPPER is set to false) or greater than to X2(:) (if UPPER is
!   set to true).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration. X2(i) must be greater
!              or equal to zero for i=1 to size(X2).
!
!   DF         (INPUT) real(stnd)
!              On entry, degrees of freedom of the chi-squared distribution.
!              DF must be greater or equal to 0.5.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   ACU        (INPUT, OPTIONAL) real(stnd)
!              On entry, the desired accuracy of the result if DF<=DF_MAX (e.g. if the incomplete 
!              Gamma integral is used). If l decimal places of accuracy are required
!              then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!              ACU should not be set smaller than the machine precision since the stated accuracy
!              cannot be attained. In that case the machine precision is used instead.
!              The default value for ACU is epsilon( ACU ).
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series or  continued fraction expansion of the incomplete 
!              Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is more accurate than PROBQ and PROBQ2 functions, but it is slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                        Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, third, half, one, c4_5, c100
    use Char_Constants,    only : allocate_error, prob_error14, prob_error15
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(in)           :: x2
    real(stnd),               intent(in)           :: df
    real(stnd),               intent(in), optional :: df_max, acu
    real(stnd), dimension(size(x2))                :: probq3_rv
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: df_max2, p, xi, probi
    real(stnd),       dimension(size(x2)) :: prob
    real(stnd), allocatable, dimension(:) :: x
!
    integer(i4b) :: n, i
    integer      :: iok
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probq3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:n)<zero) ) then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( df<half ) then
        call merror( name_proc//prob_error15 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    if ( df<=df_max2 ) then
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n>=omp_limit                        .and.      &
                      i>1_i4b                             .and.      &
                      j>1_i4b
#endif
!
!       EXACT PROBABILITY FOR SMALL df .
!
        p = half*df
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,xi,probi), SCHEDULE(STATIC)
!
        do i = 1_i4b, n
!
!           INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
            xi = half*x2(i)
!
            probi = probgamma2_r( xi, p, acu=acu, maxiter=maxiter, failure=failure )
!            probi = probgamma_r( xi, p, acu=acu, maxiter=maxiter, failure=failure )
!
            prob(i) = probi
!
        end do
!
!$OMP END PARALLEL DO
!
        if ( upper )  then
            where( prob(:n)/=-one ) prob(:n) = one - prob(:n)
        end if
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!       ALLOCATE WORK ARRAY.
!
        allocate( x(n), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       WILSON-HILFERTY FORMULA .
!
        p = one/( c4_5*df )
        x(:n) = (x2(:n)/df)**third - ( one - p )
        x(:n) = x(:n)/sqrt( p )
!
!       GAUSSIAN APPROXIMATION FOR THE PROBABILITIES.
!
        prob(:n) = probn_rv( x(:n), upper )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x )
!
    end if
!
    probq3_rv(:n) = prob(:n)
!
!
! END OF FUNCTION probq3_rv
! _________________________
!
    end function probq3_rv
!
! =========================================================================================
!
    function probq3_rv2( x2, df, upper, df_max, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2(i) to infinity 
!   if UPPER is true or from zero to X2(i) if UPPER is false, for i=1 to size(X2). In other words, if:
!
!   - UPPER = true  :    PROBQ3( i ) = prob( U > X2(i) )   ,
!   - UPPER = false :    PROBQ3( i ) = prob( U <= X2(i) )  ,
!   
!   , for U = Chi-squared(NDF( i )) and i=1 to size(X2).
!
!   PROBQ3 computes the probabilities that a random variable (vector) which 
!   follows the chi-squared distribution with DF(:) degrees of freedom is less than or 
!   equal to X2(:) (if UPPER is set to false) or greater than to X2(:) (if UPPER is
!   set to true).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:)
!              On entry, upper or lower limits of integration. X2(i) must be greater
!              or equal to zero for i=1 to size(X2).
!
!   DF         (INPUT) real(stnd), dimension(:)
!              On entry, degrees of freedom of the chi-squared distribution. Any value in
!              the array DF must be greater or equal to 0.5.
!
!              The size of NDF must verify  size(DF) = size(X2)  .
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   ACU        (INPUT, OPTIONAL) real(stnd)
!              On entry, the desired accuracy of the result if DF<=DF_MAX (e.g. if the incomplete 
!              Gamma integral is used). If l decimal places of accuracy are required
!              then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!              ACU should not be set smaller than the machine precision since the stated accuracy
!              cannot be attained. In that case the machine precision is used instead.
!              The default value for ACU is epsilon( ACU ).
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series or  continued fraction expansion of the incomplete 
!              Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF(i)<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is more accurate than PROBQ and PROBQ2 functions, but it is slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                        Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, third, half, one, c4_5, c100
    use Char_Constants,    only : prob_error14, prob_error15
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(in)           :: x2, df
    real(stnd),               intent(in), optional :: df_max, acu
    real(stnd), dimension(size(x2))                :: probq3_rv2
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: x2i, dfi, x, df_max2, p, probi
    real(stnd), dimension(size(x2)) :: prob
!
    integer(i4b) :: n, i
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probq3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(x2),i4b)  ,   &
                    int(size(df),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:n)<zero) ) then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( any(df(:n)<half) ) then
        call merror( name_proc//prob_error15 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,dfi,x2i,x,p,probi), SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        x2i = x2(i)
        dfi = df(i)
!
        if ( dfi<=df_max2 ) then
!
!           EXACT PROBABILITY FOR SMALL dfi .
!
!           INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
            x = half*x2i    
            p = half*dfi
!
            probi = probgamma2_r( x, p, acu=acu, maxiter=maxiter, failure=failure )
!            probi = probgamma_r( x, p, acu=acu, maxiter=maxiter, failure=failure )
!
            if ( upper )  then
                if ( probi/=-one ) probi = one - probi
            end if
!
        else
!
!           APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE dfi .
!
!           WILSON-HILFERTY FORMULA .
!
            p = one/( c4_5*dfi )
            x = (x2i/dfi)**third - ( one - p )
            x = x/sqrt( p )
!
!           GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
            probi = probn_r( x, upper )
!
        end if
!
        prob(i) = probi
!
    end do
!
!$OMP END PARALLEL DO
!
    probq3_rv2(:n) = prob(:n)
!
!
! END OF FUNCTION probq3_rv2
! __________________________
!
    end function probq3_rv2
!
! =========================================================================================
!
    function probq3_rm( x2, df, upper, df_max, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the chi-squared distribution function from X2(i,j) to infinity 
!   if UPPER is true or from zero to X2(i,j) if UPPER is false, for i=1 to size(X2,1)
!   and j=1 to size(X2,2). In other words, if:
!
!   - UPPER = true  :    PROBQ3( i,j ) = prob( U >  X2(i,j) )  ,
!   - UPPER = false :    PROBQ3( i,j ) = prob( U <= X2(i,j) )  ,
!   
!   , for U = Chi-squared(DF), i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   PROBQ3 computes the probabilities that a random variable (matrix) which follows the chi-squared
!   distribution with DF degrees of freedom is less than or equal to X2(:,:) (if UPPER is set to false)
!   or greater than to X2(:,:) (if UPPER is set to true).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration. X2(i,j) must be greater
!              or equal to zero for i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   DF         (INPUT) real(stnd)
!              On entry, degrees of freedom of the chi-squared distribution.
!              DF must be greater or equal to 0.5.
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   ACU        (INPUT, OPTIONAL) real(stnd)
!              On entry, the desired accuracy of the result if DF<=DF_MAX (e.g. if the incomplete 
!              Gamma integral is used). If l decimal places of accuracy are required
!              then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!              ACU should not be set smaller than the machine precision since the stated accuracy
!              cannot be attained. In that case the machine precision is used instead.
!              The default value for ACU is epsilon( ACU ).
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series or  continued fraction expansion of the incomplete 
!              Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is more accurate than PROBQ and PROBQ2 functions, but it is slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                        Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, third, half, one, c4_5, c100
    use Char_Constants,    only : allocate_error, prob_error14, prob_error15
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: x2
    real(stnd),                 intent(in)           :: df
    real(stnd),                 intent(in), optional :: df_max, acu
    real(stnd), dimension(size(x2,1),size(x2,2))     :: probq3_rm
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                   :: df_max2, p, xji, probji
    real(stnd), dimension(size(x2,1),size(x2,2)) :: prob
    real(stnd), allocatable, dimension(:,:) :: x
!
    integer(i4b) :: n, m, i, j
    integer      :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probq3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( x2, 1 )
    n = size( x2, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:m,:n)<zero) )  then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( df<half )  then
        call merror( name_proc//prob_error15 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
    if ( df<=df_max2 ) then
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*m>=omp_limit                      .and.      &
                      i>1_i4b                             .and.      &
                      j>1_i4b
#endif
!
!       EXACT PROBABILITY FOR SMALL df .
!
        p = half*df
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,xji,probji), SCHEDULE(STATIC)
!
        do i = 1_i4b, n
!
            do j = 1_i4b, m
!
!               INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
                xji = half*x2(j,i)
!
                probji = probgamma2_r( xji, p, acu=acu, maxiter=maxiter, failure=failure )
!
                prob(j,i) = probji
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
        if ( upper )  then
            where( prob(:m,:n)/=-one ) prob(:m,:n) = one - prob(:m,:n)
        end if
!
    else
!
!       APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE df .
!
!       ALLOCATE WORK ARRAY.
!
        allocate( x(m,n), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       WILSON-HILFERTY FORMULA .
!
        p = one/( c4_5*df )
        x(:m,:n) = (x2(:m,:n)/df)**third - ( one - p )
        x(:m,:n) = x(:m,:n)/sqrt( p )
!
!       GAUSSIAN APPROXIMATION FOR THE PROBABILITIES.
!
        prob(:m,:n) = probn_rm( x(:m,:n), upper )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x )
!
    end if
!
    probq3_rm(:m,:n) = prob(:m,:n)
!
!
! END OF FUNCTION probq3_rm
! _________________________
!
    end function probq3_rm
!
! =========================================================================================
!
    function probq3_rm2( x2, df, upper, df_max, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the  chi-squared distribution function from X2(i,j) to infinity 
!   if UPPER is true or from zero to X2(i,j) if UPPER is false,
!   for i=1 to size(X2,1) and j=1 to size(X2,2). In other words, if:
!
!   - UPPER = true  :    PROBQ3( i, j ) = prob( U > X2(i,j) )  ,
!   - UPPER = false :    PROBQ3( i, j ) = prob( U < X2(i,j) )  ,
!   
!   , for U = Chi-squared(NDF( i, j )), i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   PROBQ3 computes the probabilities that a random variable (matrix) which 
!   follows the chi-squared distribution with DF(:,:) degrees of freedom is less than or 
!   equal to X2(:,:) (if UPPER is set to false) or greater than to X2(:,:) (if UPPER is
!   set to true).
!
!
! Arguments
! _________
!
!   X2         (INPUT) real(stnd), dimension(:,:)
!              On entry, upper or lower limits of integration. X2(i,j) must be greater
!              or equal to zero for i=1 to size(X2,1) and j=1 to size(X2,2).
!
!   DF         (INPUT) real(stnd), dimension(:,:)
!              On entry, degrees of freedom of the chi-squared distribution. Any value in
!              the array DF must be greater or equal to 0.5.
!
!              The shape of DF must verify:
!
!              - size(DF,1) = size(X2,1)
!              - size(DF,2) = size(X2,2) .
!
!   UPPER      (INPUT) logical(lgl)
!              On entry, if:
!
!              - UPPER = true  : probability to the right of X2 is calculated.
!              - UPPER = false : probability to the left of X2 is calculated.
!
!   DF_MAX     (INPUT, OPTIONAL) real(stnd)
!              On entry, if:
!
!              - DF is lower or equal to DF_MAX, the chi-squared density is integrated
!                using the incomplete Gamma integral.
!              - DF is greater than DF_MAX, a gaussian approximation is used.
!
!              The default is 100.
!
!   ACU        (INPUT, OPTIONAL) real(stnd)
!              On entry, the desired accuracy of the result if DF<=DF_MAX (e.g. if the incomplete 
!              Gamma integral is used). If l decimal places of accuracy are required
!              then ACU should be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!              ACU should not be set smaller than the machine precision since the stated accuracy
!              cannot be attained. In that case the machine precision is used instead.
!              The default value for ACU is epsilon( ACU ).
!
!   MAXITER    (INPUT, OPTIONAL) integer(i4b)
!              On entry, MAXITER controls the maximum number of iterations when evaluating
!              the Pearson's series or  continued fraction expansion of the incomplete 
!              Gamma integral if DF<=DF_MAX.
!              The default value is 1000.
!
!   FAILURE    (INPUT, OPTIONAL) logical(lgl)
!              On entry, if FAILURE is set to true, the values for which the "integrating"
!              process of the incomplete Gamma integral did not converge to the desired 
!              accuracy are set to -1. The algorithm fails to converge if the number
!              of iterations exceeds MAXITER.
!
!              This argument is actually used only if DF<=DF_MAX.
!
!              The default value is false.
!
!
! Further Details
! _______________
!
!   If DF(i,j)<=DF_MAX, the chi-squared distribution function is evaluated by integrating the
!   incomplete Gamma integral (see the references (2) and (3) for more details), otherwise a normal
!   approximation based on the Wilson-Hilferty transformation is used (see the reference (3),
!   Formula 26.4.14, p.941).
!
!   This function is more accurate than PROBQ and PROBQ2 functions, but it is slower.
!
!   The function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Wilson, E.B., and Hilferty, M.M., 1931: The distribution of Chi-square,
!                        Proceed. Nation. Academ. Scien., Vol. 17, 684-688
!
!   (2) Shea, B.L., 1988: Algorithm AS 239: Chi-Squared and incomplete Gamma integral,  
!                  Appl. Statist., Vol. 37, No. 3, pp. 466-473
!
!   (3) Abramowitz, M., and Stegun, I.A., 1970: Handbook of Mathematical Functions,
!                                 (formulae 6.5.29, 6.5.31 and 26.4.14),
!                                 New York, Dover Publications
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, third, half, one, c4_5, c100
    use Char_Constants,    only : prob_error14, prob_error15
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: x2, df
    real(stnd),                 intent(in), optional :: df_max, acu
    real(stnd), dimension(size(x2,1),size(x2,2))     :: probq3_rm2
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                   :: x2ji, dfji, x, df_max2, p, probji
    real(stnd), dimension(size(x2,1),size(x2,2)) :: prob
!
    integer(i4b) :: m, n, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probq3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(x2,1),i4b)  ,   &
                    int(size(df,1),i4b)  ,   &
                    name_proc )
!
    n =  assert_eq( int(size(x2,2),i4b)  ,   &
                    int(size(df,2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x2(:m,:n)<zero) ) then
        call merror( name_proc//prob_error14 )
    end if
!
    if ( any(df(:m,:n)<half) ) then
        call merror( name_proc//prob_error15 )
    end if
!
    if ( present(df_max) ) then
        df_max2 = df_max
    else
        df_max2 = c100
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,dfji,x2ji,x,p,probji), SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        do j = 1_i4b, m
!
            x2ji = x2(j,i)
            dfji = df(j,i)
!
            if ( dfji<=df_max2 ) then
!
!               EXACT PROBABILITY FOR SMALL dfji .
!
!               INITIALIZATION FOR CALCULATION USING INCOMPLETE GAMMA FUNCTION.
!
                x = half*x2ji    
                p = half*dfji
!
                probji = probgamma2_r( x, p, acu=acu, maxiter=maxiter, failure=failure )
!
                if ( upper )  then
                    if ( probji/=-one ) probji = one - probji
                end if
!
            else
!
!               APPROXIMATION WITH GAUSSIAN DISTRIBUTION FOR LARGE dfji .
!
!               WILSON-HILFERTY FORMULA .
!
                p = one/( c4_5*dfji )
                x = (x2ji/dfji)**third - ( one - p )
                x = x/sqrt( p )
!
!               GAUSSIAN APPROXIMATION FOR THE PROBABILITY.
!
                probji = probn_r( x, upper )
!
            end if
!
            prob(j,i) = probji
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probq3_rm2(:m,:n) = prob(:m,:n)
!
!
! END OF FUNCTION probq3_rm2
! __________________________
!
    end function probq3_rm2
!
!
! =========================================================================================
!                    INVERSE OF THE CHI-SQUARED DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function pinvq_r( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                           X0 = PINVQ( P, NDF ) 
!   
!   , if P = probability( U < X0 )  for U = Chi-squared(NDF).
!
!   PINVQ returns the quantile X0 of the chi-squared distribution with NDF degrees of 
!   freedom corresponding to a given lower tail area of P.
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd)
!          On entry, the probability. P must verify 0. < P < 1. .
!
!   NDF    (INPUT) integer(i4b)
!          On entry, degrees of freedom of the chi-squared distribution.
!          NDF must be greater or equal to 1.
!
!
! Further Details
! _______________
!
!   This function is fast, but not very accurate especially for small degrees of freedom,
!   e.g. for NDF<10 or 20. If high accuracy is desired, function PINVQ2 must be used
!   instead.
!
!   This function is adapted from
!
!   (1)   Goldstein, R.B., 1973: Chi-square quantiles,
!                           Comm. A.C.M., vol.16, no.8, 483-485
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, four, half
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: p
!
    integer(i4b), intent(in) :: ndf
!
    real(stnd) :: pinvq_r
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: df, f1, f2, pp, t
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::                             &
                             a1   =  1.264616e-2_stnd,   &
                             a2   = -1.425296e-2_stnd,   &
                             a3   =  1.400483e-2_stnd,   &
                             a4   = -5.886090e-3_stnd,   &
                             a5   = -1.091214e-2_stnd,   &
                             a6   = -2.304527e-2_stnd,   &
                             a7   =  3.135411e-3_stnd,   &
                             a8   = -2.728484e-4_stnd,   &
                             a9   = -9.699681e-3_stnd,   &
                             a10  =  1.316872e-2_stnd,   &
                             a11  =  2.618914e-2_stnd,   &
                             a12  = -2.222222e-1_stnd,   &
                             a13  =  5.406674e-5_stnd,   &
                             a14  =  3.483789e-5_stnd,   &
                             a15  = -7.274761e-4_stnd,   &
                             a16  =  3.292181e-3_stnd,   &
                             a17  = -8.729713e-3_stnd,   &
                             a18  =  4.714045e-1_stnd,   &
                             a19  =  one
!
    real(stnd), parameter ::                             &
                             c1   =  1.565326e-3_stnd,   &
                             c2   =  1.060438e-3_stnd,   &
                             c3   = -6.950356e-3_stnd,   &
                             c4   = -1.323293e-2_stnd,   &
                             c5   =  2.277679e-2_stnd,   &
                             c6   = -8.986007e-3_stnd,   &
                             c7   = -1.513904e-2_stnd,   &
                             c8   =  2.530010e-3_stnd,   &
                             c9   = -1.450117e-3_stnd,   &
                             c10  =  5.169654e-3_stnd,   &
                             c11  = -1.153761e-2_stnd,   &
                             c12  =  1.128186e-2_stnd,   &
                             c13  =  2.607083e-2_stnd,   &
                             c14  = -2.237368e-1_stnd,   &
                             c15  =  9.780499e-5_stnd,   &
                             c16  = -8.426812e-4_stnd,   &
                             c17  =  3.125580e-3_stnd,   &
                             c18  = -8.553069e-3_stnd,   &
                             c19  =  1.348028e-4_stnd,   &
                             c20  =  4.713941e-1_stnd,   &
                             c21  =  1.000089_stnd
!
    character(len=*),  parameter :: name_proc='pinvq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( p<=zero .or. p>=one .or. ndf<1_i4b )  then
        call merror( name_proc//prob_error3 )
    end if
!
    select case( ndf )
!
        case( 1_i4b )
!
            pp = half*( one - p )
            t = pinvn_r( pp )
            pinvq_r = t*t
!
        case( 2_i4b )
!
            pp = one - p
!
            if ( pp>zero ) then
                pinvq_r = -two*log( pp )
            else
                pinvq_r = machhuge
            end if
!
        case( 3_i4b: )
!
            df = real( ndf, stnd )
            f1 = one/df
            t = pinvn_r( p )
            f2 = sqrt( f1 )*t
!
            if ( df-two>=four*abs(t) ) then
!
                pp = (((   a1 + a2*f2 )*f1                                                          &
                   + (((   a3 + a4*f2 )*f2 + a5 )*f2 + a6 ))*f1                                     &
                   + ((((( a7 + a8*f2 )*f2 + a9 )*f2 + a10 )*f2 + a11 )*f2 + a12 ))*f1              &
                   + ((((( a13*f2 + a14 )*f2 + a15 )*f2 + a16 )*f2 + a17 )*f2*f2 + a18 )*f2 + a19
!
            else
!
                pp = ((((((( c1*f2 + c2 )*f2 + c3 )*f2 + c4 )*f2 + c5 )*f2 + c6 )*f2 + c7 )*f1      &
                   + (((((( c8 + c9*f2 )*f2 + c10 )*f2 + c11 )*f2 + c12 )*f2 + c13 )*f2 + c14 ))*f1 &
                   + (((((  c15*f2 + c16 )*f2 + c17 )*f2 + c18 )*f2 + c19 )*f2 + c20 )*f2 + c21 
!
            end if
!
            pinvq_r = pp*pp*pp*df
!
     end select
!
!
! END OF FUNCTION pinvq_r
! _______________________
!
    end function pinvq_r
!
! =========================================================================================
!
    function pinvq_rv( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                           X0(i) = PINVQ( P(i), NDF ) 
!   
!   , if P(i) = probability( U < X0(i) )  for U = Chi-squared(NDF) and i=1 to size(P).
!
!   PINVQ returns the quantiles X0(i) of the chi-squared distribution with NDF degrees of 
!   freedom corresponding to a given lower tail area of P(i), for i=1 to size(P).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:)
!          On entry, the probabilities. P(i) must verify 0. < P(i) < 1. ,
!          for i=1 to size(P).
!
!   NDF    (INPUT) integer(i4b)
!          On entry, degrees of freedom of the chi-squared distribution.
!          NDF must be greater or equal to 1.
!
!
! Further Details
! _______________
!
!   This function is fast, but not very accurate especially for small degrees of freedom,
!   e.g. for NDF<10 or 20. If high accuracy is desired, function PINVQ2 must be used
!   instead.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Goldstein, R.B., R.B., 1973: Chi-square quantiles,
!                           Comm. A.C.M., vol.16, no.8, 483-485
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, four, half
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : allocate_error, prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),  intent(in) :: p
!
    integer(i4b), intent(in) :: ndf
!
    real(stnd), dimension(size(p)) :: pinvq_rv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: df, f1, f2, pp, ti
    real(stnd), dimension(size(p))        :: t
    real(stnd), dimension(:), allocatable :: y 
!
    integer(i4b) :: n, i
    integer      :: iok
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::                             &
                             a1   =  1.264616e-2_stnd,   &
                             a2   = -1.425296e-2_stnd,   &
                             a3   =  1.400483e-2_stnd,   &
                             a4   = -5.886090e-3_stnd,   &
                             a5   = -1.091214e-2_stnd,   &
                             a6   = -2.304527e-2_stnd,   &
                             a7   =  3.135411e-3_stnd,   &
                             a8   = -2.728484e-4_stnd,   &
                             a9   = -9.699681e-3_stnd,   &
                             a10  =  1.316872e-2_stnd,   &
                             a11  =  2.618914e-2_stnd,   &
                             a12  = -2.222222e-1_stnd,   &
                             a13  =  5.406674e-5_stnd,   &
                             a14  =  3.483789e-5_stnd,   &
                             a15  = -7.274761e-4_stnd,   &
                             a16  =  3.292181e-3_stnd,   &
                             a17  = -8.729713e-3_stnd,   &
                             a18  =  4.714045e-1_stnd,   &
                             a19  =  one
!
    real(stnd), parameter ::                             &
                             c1   =  1.565326e-3_stnd,   &
                             c2   =  1.060438e-3_stnd,   &
                             c3   = -6.950356e-3_stnd,   &
                             c4   = -1.323293e-2_stnd,   &
                             c5   =  2.277679e-2_stnd,   &
                             c6   = -8.986007e-3_stnd,   &
                             c7   = -1.513904e-2_stnd,   &
                             c8   =  2.530010e-3_stnd,   &
                             c9   = -1.450117e-3_stnd,   &
                             c10  =  5.169654e-3_stnd,   &
                             c11  = -1.153761e-2_stnd,   &
                             c12  =  1.128186e-2_stnd,   &
                             c13  =  2.607083e-2_stnd,   &
                             c14  = -2.237368e-1_stnd,   &
                             c15  =  9.780499e-5_stnd,   &
                             c16  = -8.426812e-4_stnd,   &
                             c17  =  3.125580e-3_stnd,   &
                             c18  = -8.553069e-3_stnd,   &
                             c19  =  1.348028e-4_stnd,   &
                             c20  =  4.713941e-1_stnd,   &
                             c21  =  1.000089_stnd
!
    character(len=*),  parameter :: name_proc='pinvq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( p )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:n)<=zero) .or. any(p(:n)>=one) .or. ndf<1_i4b ) then
        call merror( name_proc//prob_error3 )
    end if
!
    select case( ndf )
!
        case( 1_i4b )
!
!           ALLOCATE WORK ARRAY.
!
            allocate( y(n), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            t(:n) = half*( one - p(:n) )
            y(:n) = pinvn_rv( t(:n) )
            t(:n) = y(:n)*y(:n)
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( y )
!
        case( 2_i4b )
!
!           ALLOCATE WORK ARRAY.
!
            allocate( y(n), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            y(:n) = one - p(:n)
!
            where ( y(:n)>zero )
                t(:n) = -two*log( y(:n) )
            elsewhere
                t(:n) = machhuge
            end where
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( y )
!
        case( 3_i4b: )
!
            df = real( ndf, stnd )
            f1 = one/df
!
            t(:n) = pinvn_rv( p(:n) )
!
#ifdef _OPENMP
            i = omp_get_num_procs()
            j = omp_get_max_threads()
            ompparallel = omp_in_parallel()
            ompnested   = omp_get_nested() .and. ompparallel
            test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                          n>=omp_limit                        .and.      &
                          i>1_i4b                             .and.      &
                          j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,ti,f2,pp)
!
            do i = 1_i4b, n
!
                ti = t(i)
!
                f2 = sqrt( f1 )*ti
!
                if ( df-two>=four*abs(ti) ) then
!
                    pp = (((   a1 + a2*f2 )*f1                                                          &
                       + (((   a3 + a4*f2 )*f2 + a5 )*f2 + a6 ))*f1                                     &
                       + ((((( a7 + a8*f2 )*f2 + a9 )*f2 + a10 )*f2 + a11 )*f2 + a12 ))*f1              &
                       + ((((( a13*f2 + a14 )*f2 + a15 )*f2 + a16 )*f2 + a17 )*f2*f2 + a18 )*f2 + a19
!
                else
!
                    pp = ((((((( c1*f2 + c2 )*f2 + c3 )*f2 + c4 )*f2 + c5 )*f2 + c6 )*f2 + c7 )*f1      &
                       + (((((( c8 + c9*f2 )*f2 + c10 )*f2 + c11 )*f2 + c12 )*f2 + c13 )*f2 + c14 ))*f1 &
                       + (((((  c15*f2 + c16 )*f2 + c17 )*f2 + c18 )*f2 + c19 )*f2 + c20 )*f2 + c21 
!
                end if
!
                t(i) = pp*pp*pp*df
!
            end do
!
!$OMP END PARALLEL DO
!
    end select
!
    pinvq_rv = t(:n)
!
!
! END OF FUNCTION pinvq_rv
! ________________________
!
    end function pinvq_rv
!
! =========================================================================================
!
    function pinvq_rv2( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                           X0(i) = PINVQ( P(i), NDF ) 
!   
!   , if P(i) = probability( U < X0(i) )  for U = Chi-squared(NDF(i)) and i=1 to size(P).
!
!   PINVQ returns the quantiles X0(i) of the chi-squared distribution with NDF(i) degrees of 
!   freedom corresponding to a given lower tail area of P(i), for i=1 to size(P).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:)
!          On entry, the probabilities. P(i) must verify 0. < P(i) < 1. ,
!          for i=1 to size(P).
!
!   NDF    (INPUT) integer(i4b), dimension(:)
!          On entry, degrees of freedom of the chi-squared distribution.
!          Any value in the array NDF must be greater or equal to 1.
!
!          The size of NDF must be  size(NDF) = size(P).
!
!
! Further Details
! _______________
!
!   This function is fast, but not very accurate especially for small degrees of freedom,
!   e.g. for NDF<10 or 20. If high accuracy is desired, function PINVQ2 must be used
!   instead.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Goldstein, R.B., R.B., 1973: Chi-square quantiles,
!                           Comm. A.C.M., vol.16, no.8, 483-485
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, four, half
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),  intent(in) :: p
!
    integer(i4b), dimension(:), intent(in) :: ndf
!
    real(stnd), dimension(size(p)) :: pinvq_rv2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: df, f1, f2, pp, t, pi
    real(stnd), dimension(size(p)) :: q
!
    integer(i4b) :: n, i, ndfi
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::                             &
                             a1   =  1.264616e-2_stnd,   &
                             a2   = -1.425296e-2_stnd,   &
                             a3   =  1.400483e-2_stnd,   &
                             a4   = -5.886090e-3_stnd,   &
                             a5   = -1.091214e-2_stnd,   &
                             a6   = -2.304527e-2_stnd,   &
                             a7   =  3.135411e-3_stnd,   &
                             a8   = -2.728484e-4_stnd,   &
                             a9   = -9.699681e-3_stnd,   &
                             a10  =  1.316872e-2_stnd,   &
                             a11  =  2.618914e-2_stnd,   &
                             a12  = -2.222222e-1_stnd,   &
                             a13  =  5.406674e-5_stnd,   &
                             a14  =  3.483789e-5_stnd,   &
                             a15  = -7.274761e-4_stnd,   &
                             a16  =  3.292181e-3_stnd,   &
                             a17  = -8.729713e-3_stnd,   &
                             a18  =  4.714045e-1_stnd,   &
                             a19  =  one
!
    real(stnd), parameter ::                             &
                             c1   =  1.565326e-3_stnd,   &
                             c2   =  1.060438e-3_stnd,   &
                             c3   = -6.950356e-3_stnd,   &
                             c4   = -1.323293e-2_stnd,   &
                             c5   =  2.277679e-2_stnd,   &
                             c6   = -8.986007e-3_stnd,   &
                             c7   = -1.513904e-2_stnd,   &
                             c8   =  2.530010e-3_stnd,   &
                             c9   = -1.450117e-3_stnd,   &
                             c10  =  5.169654e-3_stnd,   &
                             c11  = -1.153761e-2_stnd,   &
                             c12  =  1.128186e-2_stnd,   &
                             c13  =  2.607083e-2_stnd,   &
                             c14  = -2.237368e-1_stnd,   &
                             c15  =  9.780499e-5_stnd,   &
                             c16  = -8.426812e-4_stnd,   &
                             c17  =  3.125580e-3_stnd,   &
                             c18  = -8.553069e-3_stnd,   &
                             c19  =  1.348028e-4_stnd,   &
                             c20  =  4.713941e-1_stnd,   &
                             c21  =  1.000089_stnd
!
    character(len=*),  parameter :: name_proc='pinvq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(p),i4b)     ,   &
                    int(size(ndf),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:n)<=zero) .or. any(p(:n)>=one) .or. any(ndf(:n)<1_i4b) ) then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,df,f1,f2,pp,t,pi,ndfi)
!
    do i = 1_i4b, n
!
        pi   = p(i)
        ndfi = ndf(i)
!
        select case( ndfi )
!
            case( 1_i4b )
!
                f1 = half*( one - pi )
                f2 = pinvn_r( f1 )
                t  = f2*f2
!
            case( 2_i4b )
!
                f1 = one - pi
!
                if ( f1>zero ) then
                    t = -two*log( f1 )
                else
                    t = machhuge
                end if
!
            case( 3_i4b: )
!
                df = real( ndfi, stnd )
                f1 = one/df
                pp = pinvn_r( pi )
                f2 = sqrt( f1 )*pp
!
                if ( df-two>=four*abs(pp) ) then
!
                    pp = (((   a1 + a2*f2 )*f1                                                          &
                       + (((   a3 + a4*f2 )*f2 + a5 )*f2 + a6 ))*f1                                     &
                       + ((((( a7 + a8*f2 )*f2 + a9 )*f2 + a10 )*f2 + a11 )*f2 + a12 ))*f1              &
                       + ((((( a13*f2 + a14 )*f2 + a15 )*f2 + a16 )*f2 + a17 )*f2*f2 + a18 )*f2 + a19
!
                else
!
                    pp = ((((((( c1*f2 + c2 )*f2 + c3 )*f2 + c4 )*f2 + c5 )*f2 + c6 )*f2 + c7 )*f1      &
                       + (((((( c8 + c9*f2 )*f2 + c10 )*f2 + c11 )*f2 + c12 )*f2 + c13 )*f2 + c14 ))*f1 &
                       + (((((  c15*f2 + c16 )*f2 + c17 )*f2 + c18 )*f2 + c19 )*f2 + c20 )*f2 + c21 
!
                end if
!
                t = pp*pp*pp*df
!
        end select
!
        q(i) = t
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvq_rv2 = q(:n)
!
!
! END OF FUNCTION pinvq_rv2
! _________________________
!
    end function pinvq_rv2
!
! =========================================================================================
!
    function pinvq_rm( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!               X0(i,j) = PINVQ( P(i,j), NDF )
!   
!   , if P(i,j) = probability( U < T0(i,j) )  for U = Chi-squared(NDF), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVQ returns the quantiles X0(i,j) of the chi-squared distribution with NDF degrees of 
!   freedom corresponding to a given lower tail area of P(i,j) for i=1 to size(P,1) and
!   j=1 to size(P,2).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:,:)
!          On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1. ,
!          for i=1 to size(P,1) and j=1 to size(P,2).
!
!   NDF    (INPUT) integer(i4b)
!          On entry, degrees of freedom of the chi-squared distribution.
!          NDF must be greater or equal to 1.
!
!
! Further Details
! _______________
!
!   This function is fast, but not very accurate especially for small degrees of freedom,
!   e.g. for NDF<10 or 20. If high accuracy is desired, function PINVQ2 must be used
!   instead.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Goldstein, R.B., R.B., 1973: Chi-square quantiles,
!                           Comm. A.C.M., vol.16, no.8, 483-485
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, four, half
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : allocate_error, prob_error3
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),  intent(in) :: p
!
    integer(i4b), intent(in) :: ndf
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvq_rm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: df, f1, f2, pp, tji
    real(stnd), dimension(size(p,1),size(p,2)) :: t
    real(stnd), dimension(:,:),    allocatable :: y 
!
    integer(i4b) :: n, m, i, j
    integer      :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::                             &
                             a1   =  1.264616e-2_stnd,   &
                             a2   = -1.425296e-2_stnd,   &
                             a3   =  1.400483e-2_stnd,   &
                             a4   = -5.886090e-3_stnd,   &
                             a5   = -1.091214e-2_stnd,   &
                             a6   = -2.304527e-2_stnd,   &
                             a7   =  3.135411e-3_stnd,   &
                             a8   = -2.728484e-4_stnd,   &
                             a9   = -9.699681e-3_stnd,   &
                             a10  =  1.316872e-2_stnd,   &
                             a11  =  2.618914e-2_stnd,   &
                             a12  = -2.222222e-1_stnd,   &
                             a13  =  5.406674e-5_stnd,   &
                             a14  =  3.483789e-5_stnd,   &
                             a15  = -7.274761e-4_stnd,   &
                             a16  =  3.292181e-3_stnd,   &
                             a17  = -8.729713e-3_stnd,   &
                             a18  =  4.714045e-1_stnd,   &
                             a19  =  one
!
    real(stnd), parameter ::                             &
                             c1   =  1.565326e-3_stnd,   &
                             c2   =  1.060438e-3_stnd,   &
                             c3   = -6.950356e-3_stnd,   &
                             c4   = -1.323293e-2_stnd,   &
                             c5   =  2.277679e-2_stnd,   &
                             c6   = -8.986007e-3_stnd,   &
                             c7   = -1.513904e-2_stnd,   &
                             c8   =  2.530010e-3_stnd,   &
                             c9   = -1.450117e-3_stnd,   &
                             c10  =  5.169654e-3_stnd,   &
                             c11  = -1.153761e-2_stnd,   &
                             c12  =  1.128186e-2_stnd,   &
                             c13  =  2.607083e-2_stnd,   &
                             c14  = -2.237368e-1_stnd,   &
                             c15  =  9.780499e-5_stnd,   &
                             c16  = -8.426812e-4_stnd,   &
                             c17  =  3.125580e-3_stnd,   &
                             c18  = -8.553069e-3_stnd,   &
                             c19  =  1.348028e-4_stnd,   &
                             c20  =  4.713941e-1_stnd,   &
                             c21  =  1.000089_stnd
!
    character(len=*),  parameter :: name_proc='pinvq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( p, 1 )
    n = size( p, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:m,:n)<=zero) .or. any(p(:m,:n)>=one) .or. ndf<1_i4b ) then
        call merror( name_proc//prob_error3 )
    end if
!
    select case( ndf )
!
        case( 1_i4b )
!
!           ALLOCATE WORK ARRAY.
!
            allocate( y(m,n), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            t(:m,:n) = half*( one - p(:m,:n) )
            y(:m,:n) = pinvn_rm( t(:m,:n) )
            t(:m,:n) = y(:m,:n)*y(:m,:n)
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( y )
!
        case( 2_i4b )
!
!           ALLOCATE WORK ARRAY.
!
            allocate( y(m,n), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            y(:m,:n) = one - p(:m,:n)
!
            where ( y(:m,:n)>zero )
                t(:m,:n) = -two*log( y(:m,:n) )
            elsewhere
                t(:m,:n) = machhuge
            end where
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( y )
!
        case( 3_i4b: )
!
            df = real( ndf, stnd )
            f1 = one/df
!
            t(:m,:n) = pinvn_rm( p(:m,:n) )
!
#ifdef _OPENMP
            i = omp_get_num_procs()
            j = omp_get_max_threads()
            ompparallel = omp_in_parallel()
            ompnested   = omp_get_nested() .and. ompparallel
            test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                          n*m>=omp_limit                      .and.      &
                          n/j>=omp_limit2                     .and.      &
                          i>1_i4b                             .and.      &
                          j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,tji,f2,pp)
!
            do i = 1_i4b, n
!
                do j = 1_i4b, m
!
                    tji = t(j,i)
!
                    f2 = sqrt( f1 )*tji
!
                    if ( df-two>=four*abs(tji) ) then
!
                        pp = (((   a1 + a2*f2 )*f1                                                          &
                           + (((   a3 + a4*f2 )*f2 + a5 )*f2 + a6 ))*f1                                     &
                           + ((((( a7 + a8*f2 )*f2 + a9 )*f2 + a10 )*f2 + a11 )*f2 + a12 ))*f1              &
                           + ((((( a13*f2 + a14 )*f2 + a15 )*f2 + a16 )*f2 + a17 )*f2*f2 + a18 )*f2 + a19
!
                    else
!
                        pp = ((((((( c1*f2 + c2 )*f2 + c3 )*f2 + c4 )*f2 + c5 )*f2 + c6 )*f2 + c7 )*f1      &
                           + (((((( c8 + c9*f2 )*f2 + c10 )*f2 + c11 )*f2 + c12 )*f2 + c13 )*f2 + c14 ))*f1 &
                           + (((((  c15*f2 + c16 )*f2 + c17 )*f2 + c18 )*f2 + c19 )*f2 + c20 )*f2 + c21 
!
                    end if
!
                    t(j,i) = pp*pp*pp*df
!
                end do
!
            end do
!
!$OMP END PARALLEL DO
!
    end select
!
    pinvq_rm = t(:m,:n)
!
!
! END OF FUNCTION pinvq_rm
! ________________________
!
    end function pinvq_rm
!
! =========================================================================================
!
    function pinvq_rm2( p, ndf )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!               X0(i,j) = PINVQ( P(i,j), NDF )
!   
!   , if P(i,j) = probability( U < T0(i,j) )  for U = Chi-squared(NDF(i,j)), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVQ returns the quantiles X0(i,j) of the chi-squared distribution with NDF(i,j) degrees of 
!   freedom corresponding to a given lower tail area of P(i,j) for i=1 to size(P,1) and
!   j=1 to size(P,2).
!
!
! Arguments
! _________
!
!   P      (INPUT) real(stnd), dimension(:,:)
!          On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1. ,
!          for i=1 to size(P,1) and j=1 to size(P,2).
!
!   NDF    (INPUT) integer(i4b), dimension(:,:)
!          On entry, degrees of freedom of the chi-squared distribution.
!          Any value in the array NDF must be greater or equal to 1.
!
!          The shape of NDF must verify:
!
!          - size(NDF,1) = size(P,1)
!          - size(NDF,2) = size(P,2).
!
!
! Further Details
! _______________
!
!   This function is fast, but not very accurate especially for small degrees of freedom,
!   e.g. for NDF<10 or 20. If high accuracy is desired, function PINVQ2 must be used
!   instead.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Goldstein, R.B., R.B., 1973: Chi-square quantiles,
!                           Comm. A.C.M., vol.16, no.8, 483-485
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one, two, four, half
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error3
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),  intent(in) :: p
!
    integer(i4b), dimension(:,:), intent(in) :: ndf
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvq_rm2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: df, f1, f2, pp, t, pji
    real(stnd), dimension(size(p,1),size(p,2)) :: q
!
    integer(i4b) :: n, m, i, j, ndfji
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), parameter ::                             &
                             a1   =  1.264616e-2_stnd,   &
                             a2   = -1.425296e-2_stnd,   &
                             a3   =  1.400483e-2_stnd,   &
                             a4   = -5.886090e-3_stnd,   &
                             a5   = -1.091214e-2_stnd,   &
                             a6   = -2.304527e-2_stnd,   &
                             a7   =  3.135411e-3_stnd,   &
                             a8   = -2.728484e-4_stnd,   &
                             a9   = -9.699681e-3_stnd,   &
                             a10  =  1.316872e-2_stnd,   &
                             a11  =  2.618914e-2_stnd,   &
                             a12  = -2.222222e-1_stnd,   &
                             a13  =  5.406674e-5_stnd,   &
                             a14  =  3.483789e-5_stnd,   &
                             a15  = -7.274761e-4_stnd,   &
                             a16  =  3.292181e-3_stnd,   &
                             a17  = -8.729713e-3_stnd,   &
                             a18  =  4.714045e-1_stnd,   &
                             a19  =  one
!
    real(stnd), parameter ::                             &
                             c1   =  1.565326e-3_stnd,   &
                             c2   =  1.060438e-3_stnd,   &
                             c3   = -6.950356e-3_stnd,   &
                             c4   = -1.323293e-2_stnd,   &
                             c5   =  2.277679e-2_stnd,   &
                             c6   = -8.986007e-3_stnd,   &
                             c7   = -1.513904e-2_stnd,   &
                             c8   =  2.530010e-3_stnd,   &
                             c9   = -1.450117e-3_stnd,   &
                             c10  =  5.169654e-3_stnd,   &
                             c11  = -1.153761e-2_stnd,   &
                             c12  =  1.128186e-2_stnd,   &
                             c13  =  2.607083e-2_stnd,   &
                             c14  = -2.237368e-1_stnd,   &
                             c15  =  9.780499e-5_stnd,   &
                             c16  = -8.426812e-4_stnd,   &
                             c17  =  3.125580e-3_stnd,   &
                             c18  = -8.553069e-3_stnd,   &
                             c19  =  1.348028e-4_stnd,   &
                             c20  =  4.713941e-1_stnd,   &
                             c21  =  1.000089_stnd
!
    character(len=*),  parameter :: name_proc='pinvq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(p,1),i4b)     ,   &
                    int(size(ndf,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(p,2),i4b)     ,   &
                    int(size(ndf,2),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b )  then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:m,:n)<=zero) .or. any(p(:m,:n)>=one) .or. any(ndf(:m,:n)<1_i4b) )  then
        call merror( name_proc//prob_error3 )
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,df,f1,f2,pp,t,pji,ndfji)
!
    do i = 1_i4b, n
!
        do j = 1_i4b, m
!
            pji   = p(j,i)
            ndfji = ndf(j,i)
!
            select case( ndfji )
!
                case( 1_i4b )
!
                    f1 = half*( one - pji )
                    f2 = pinvn_r( f1 )
                    t = f2*f2
!
                case( 2_i4b )
!
                    f1 = one - pji
!
                    if ( f1>zero ) then
                        t = -two*log( f1 )
                    else
                        t = machhuge
                    end if
!
                case( 3_i4b: )
!
                    df = real( ndfji, stnd )
                    f1 = one/df
                    pp  = pinvn_r( pji )
                    f2 = sqrt( f1 )*pp
!
                    if ( df-two>=four*abs(pp) ) then
!
                        pp = (((   a1 + a2*f2 )*f1                                                          &
                           + (((   a3 + a4*f2 )*f2 + a5 )*f2 + a6 ))*f1                                     &
                           + ((((( a7 + a8*f2 )*f2 + a9 )*f2 + a10 )*f2 + a11 )*f2 + a12 ))*f1              &
                           + ((((( a13*f2 + a14 )*f2 + a15 )*f2 + a16 )*f2 + a17 )*f2*f2 + a18 )*f2 + a19
!
                    else
!
                        pp = ((((((( c1*f2 + c2 )*f2 + c3 )*f2 + c4 )*f2 + c5 )*f2 + c6 )*f2 + c7 )*f1      &
                           + (((((( c8 + c9*f2 )*f2 + c10 )*f2 + c11 )*f2 + c12 )*f2 + c13 )*f2 + c14 ))*f1 &
                           + (((((  c15*f2 + c16 )*f2 + c17 )*f2 + c18 )*f2 + c19 )*f2 + c20 )*f2 + c21 
!
                    end if
!
                    t = pp*pp*pp*df
!
            end select
!
            q(j,i) = t
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvq_rm2 = q(:m,:n)
!
!
! END OF FUNCTION pinvq_rm2
! _________________________
!
    end function pinvq_rm2
!
! =========================================================================================
!
    function pinvq2_r( p, df, prec, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                           X0 = PINVQ2( P, DF ) 
!   
!   , if P = probability( U < X0 )  for U = Chi-squared(DF).
!
!   PINVQ2 returns the quantile X0 of the chi-squared distribution with DF degrees of 
!   freedom corresponding to a given lower tail area of P. In other words, PINVQ2
!   outputs a chi-squared value, X0, such that a random variable, distributed as chi-squared
!   with DF degrees of freedom, will be less than or equal to X0 with probability P.
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd)
!           On entry, the probability. P must be in the inclusive range (0,1).
!
!   DF      (INPUT) real(stnd)
!           On entry, degrees of freedom of the chi-squared distribution.
!           DF must be greater or equal to 0.5.
!
!   PREC    (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accurary of the result. If more than six significant
!           digits are required, the default value of PREC (e.g. 0.5e-06_stnd) should be
!           altered appropriately (e.g. decreased). PREC is a small strictly positive 
!           integer less than 0.5e-06_stnd.
!
!           The default value for PREC is 0.5e-06_stnd .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result when computing the incomplete 
!           Gamma integral in the evaluation of the seven term Taylor series.
!           If l decimal places of accuracy are required then ACU should
!           be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or  continued fraction expansion of the
!           incomplete Gamma integral.
!
!           The default value is 1000.
!
!           See the description of the PROBGAMMA2 function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function is both more general (here the number of degrees of freedom, DF,
!   is not necessarily an integer) and more accurate (here the quantile X0 may be calculated
!   as exactly as the computer allows with the parameter PREC) than PINVQ function.
!
!   This function is adapted from
!
!   (1)   Best, D.J., and Roberts, D.E., 1975: Algorithm AS 91: The Percentage Points of 
!                     the chi2 Distribution. Appl. Statist., Vol.24, No. 3, pp.385-388
!
!   (2)   Shea, B.L., 1991: Algorithm AS R85: A remark on AS 91: The Percentage Points of
!                     the chi2 Distribution. Appl. Statist. Vol.40, No. 1, pp.233-235.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one, two, three, six
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error1, prob_error15
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: p, df
    real(stnd), intent(in), optional :: prec, acu
!
    real(stnd) :: pinvq2_r
!
    integer(i4b), intent(in), optional :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, p1, p2, q, s1, s2, s3, s4, s5, s6, &
                    t, x, xx, g, chi2, prec2
!
    integer(i4b) :: i
!   
!   
! PARAMETERS 
! __________
!
!   maxit IS THE MAXIMUM NUMBER OF ITERATIONS OF THE TAYLOR SERIES EXPANSION.
!
    integer(i4b), parameter ::   maxit = 20_i4b
!
    real(stnd), parameter :: aa = 0.6931471806_stnd
!
    real(stnd), parameter ::                      &
                             c1 = 0.01_stnd,      &
                             c2 = 0.222222_stnd,  &
                             c3 = 0.32_stnd,      &
                             c4 = 0.4_stnd,       &
                             c5 = 1.24_stnd,      &
                             c6 = 2.2_stnd,       &
                             c7 = 4.67_stnd,      &
                             c8 = 6.66_stnd,      &
                             c9 = 6.73_stnd,      &
                             c10 = 13.32_stnd,    &
                             c11 = 60.0_stnd,     &
                             c12 = 70.0_stnd,     &
                             c13 = 84.0_stnd,     &
                             c14 = 105.0_stnd,    &
                             c15 = 120.0_stnd,    &
                             c16 = 127.0_stnd,    &
                             c17 = 140.0_stnd,    &
                             c18 = 175.0_stnd,    &
                             c19 = 210.0_stnd,    &
                             c20 = 252.0_stnd,    &
                             c21 = 264.0_stnd,    &
                             c22 = 294.0_stnd,    &
                             c23 = 346.0_stnd,    &
                             c24 = 420.0_stnd,    &
                             c25 = 462.0_stnd,    &
                             c26 = 606.0_stnd,    &
                             c27 = 672.0_stnd,    &
                             c28 = 707.0_stnd,    &
                             c29 = 735.0_stnd,    &
                             c30 = 889.0_stnd,    &
                             c31 = 932.0_stnd,    &
                             c32 = 966.0_stnd,    &
                             c33 = 1141.0_stnd,   &
                             c34 = 1182.0_stnd,   &
                             c35 = 1278.0_stnd,   &
                             c36 = 1740.0_stnd,   &
                             c37 = 2520.0_stnd,   &
                             c38 = 5040.0_stnd
!
    character(len=*),  parameter :: name_proc='pinvq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( p<zero .or. p>one )  then
        call merror( name_proc//prob_error1 )
    end if
!
    if ( df<half ) then
        call merror( name_proc//prob_error15 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( p==zero .or. p==one ) then
!
        pinvq2_r = merge( zero, machhuge, p==zero )
!
        return
!
    end if
!
!   DETERMINE THE REQUIRED PRECISION.
!   prec2 DETERMINE THE PRECISION OF THE RESULT. IF MORE THAN
!   SIX SIGNIFICANT FIGURES ARE REQUIRED prec2 SHOULD BE DECREASED.
!
    prec2 = 0.5e-06_stnd
    if ( present(prec) ) then
        if ( prec>zero .and. prec<prec2 ) prec2 = prec
    end if
!
!   INITIALISE.
!
    xx = half*df
    g = lngamma_r( xx )
!
    c  = xx - one
!
    if ( df<-c5*log(p) ) then
!
!       STARTING APPROXIMATION FOR SMALL CHI-SQUARED.
!
        chi2 = ( p*xx*exp(g + xx*aa))**(one/xx)
!
        if ( chi2<prec2 ) then
!
           return
!
        end if
!
    else if ( df<=c3 ) then
!
!       STARTING APPROXIMATION FOR df LESS THAN OR EQUAL TO 0.32 .
!
        chi2 = c4
        a = log( one-p )
!
        do
!
            q  = chi2
            p1 = one + chi2*( c7 + chi2 )
            p2 = chi2*( c9 + chi2*( c8 + chi2 ) )
            t  = -half + ( c7 + two*chi2 )/p1 - ( c9 + chi2*( c10 + three*chi2 ) )/p2
            chi2 = chi2 - ( one - exp( a + g + half*chi2 + c*aa )*p2/p1 )/t
!
            if (abs(q / chi2 - one)<= c1) exit
!
        end do
!
    else
!
!       STARTING APPROXIMATION USING WILSON AND HILFERTY ESTIMATE.
!
        x = pinvn_r( p )
!
        p1 = c2/df
        chi2 = df*( x*sqrt(p1) + one - p1 )**3_i4b
!
!       STARTING APPROXIMATION FOR p TENDING TO 1.
!
        if (chi2 > c6*df + six)       &
        chi2 = -two*( log(one-p) - c*log(half*chi2) + g )
!
    end if
!
!   CALCULATION OF SEVEN TERM TAYLOR SERIES.
!
    do i = 1_i4b, maxit
!
        q  = chi2
        p1 = half*chi2
        p2 = p - probgamma2_r( p1, xx, acu=acu, maxiter=maxiter )
!
        t = p2*exp( xx*aa + g + p1 - c*log(chi2) )
        b = t/chi2
        a = half*t - b*c
        s1 = ( c19 + a*( c17 + a*(c14 + a*(c13 + a*(c12 + c11*a))))) / c24
        s2 = ( c24 + a*( c29 + a*(c32 + a*(c33 + c35*a)))) / c37
        s3 = ( c19 + a*( c25 + a*(c28 + c31*a))) / c37
        s4 = ( c20 + a*( c27 + c34*a) + c*(c22 + a*(c30 + c36*a))) / c38
        s5 = ( c13 + c21*a + c*(c18 + c26*a)) / c37
        s6 = ( c15 + c*( c23 + c16*c)) / c38
        chi2 = chi2 + t*(one + half*t*s1 - b*c*(s1 - b*  &
                   (s2 - b*(s3 - b*(s4 - b*(s5 - b*s6))))))
!
        if ( abs( q/chi2 - one )<=prec2 ) exit
!
    end do
!
!   HERE IF i>maxit THEN THE RESULT IS PROBABLY AS ACCURATE AS THE MACHINE WILL ALLOW.
!   BUT IT IS SAFE TO GENERATE AN ERROR MESSAGE.
!
!    if ( i>maxit )  call merror( name_proc//prob_error17 )
!
    pinvq2_r = chi2
!
!
! END OF FUNCTION pinvq2_r
! ________________________
!
    end function pinvq2_r
!
! =========================================================================================
!
    function pinvq2_rv( p, df, prec, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                           X0 = PINVQ2( P(i), DF ) 
!   
!   , if P(i) = probability( U < X0 )  for U = Chi-squared(DF) and i=1 to size(P).
!
!   PINVQ2 returns the quantiles X0(i) of the chi-squared distribution with DF degrees of 
!   freedom corresponding to a given lower tail area of  P(i), for i=1 to size(P).
!   In other words, PINVQ2 outputs chi-squared values, X0(:), such that random variables,
!   distributed as chi-squared with DF degrees of freedom, will be less than or equal to X0(:)
!   with associated probabilities P(:).
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd), dimension(:)
!           On entry, the probabilities. P(i) must verify 0. < P(i) < 1,
!           for i=1 to size(P). 
!
!   DF      (INPUT) real(stnd)
!           On entry, degrees of freedom of the chi-squared distribution.
!           DF must be greater or equal to 0.5.
!
!   PREC    (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accurary of the result. If more than six significant
!           digits are required, the default value of PREC (e.g. 0.5e-06_stnd) should be
!           altered appropriately(e.g. decreased). PREC is a small strictly positive 
!           integer less than 0.5e-06_stnd.
!
!           The default value for PREC is 0.5e-06_stnd .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result when computing the incomplete 
!           Gamma integral in the evaluation of the seven term Taylor series.
!           If l decimal places of accuracy are required then ACU should
!           be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!
!           The default value is 1000.
!
!           See the description of the PROBGAMMA2 function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function is both more general (here the number of degrees of freedom, DF,
!   is not necessarily an integer) and more accurate (here the quantiles X0(:) may be calculated
!   as exactly as the computer allows with the parameter PREC) than PINVQ function.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Best, D.J., and Roberts, D.E., 1975: Algorithm AS 91: The Percentage Points of 
!                     the chi2 Distribution. Appl. Statist., Vol.24, No. 3, pp.385-388
!
!   (2)   Shea, B.L., 1991: Algorithm AS R85: A remark on AS 91: The Percentage Points of
!                     the chi2 Distribution. Appl. Statist. Vol.40, No. 1, pp.233-235.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one, two, three, six
    use Logical_Constants, only : true, false
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error1, prob_error15
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(in)           :: p
    real(stnd),               intent(in)           :: df
    real(stnd),               intent(in), optional :: prec, acu
!
    real(stnd), dimension(size(p)) :: pinvq2_rv
!
    integer(i4b), intent(in), optional :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, p1, p2, q, s1, s2, s3, s4, s5, s6, &
                    t, x, xx, g, prec2, chi2i, pi
    real(stnd), dimension(size(p)) :: chi2
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: done
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
!   maxit IS THE MAXIMUM NUMBER OF ITERATIONS OF THE TAYLOR SERIES EXPANSION.
!
    integer(i4b), parameter ::   maxit = 20_i4b
!
    real(stnd), parameter :: aa = 0.6931471806_stnd
!
    real(stnd), parameter ::                      &
                             c1 = 0.01_stnd,      &
                             c2 = 0.222222_stnd,  &
                             c3 = 0.32_stnd,      &
                             c4 = 0.4_stnd,       &
                             c5 = 1.24_stnd,      &
                             c6 = 2.2_stnd,       &
                             c7 = 4.67_stnd,      &
                             c8 = 6.66_stnd,      &
                             c9 = 6.73_stnd,      &
                             c10 = 13.32_stnd,    &
                             c11 = 60.0_stnd,     &
                             c12 = 70.0_stnd,     &
                             c13 = 84.0_stnd,     &
                             c14 = 105.0_stnd,    &
                             c15 = 120.0_stnd,    &
                             c16 = 127.0_stnd,    &
                             c17 = 140.0_stnd,    &
                             c18 = 175.0_stnd,    &
                             c19 = 210.0_stnd,    &
                             c20 = 252.0_stnd,    &
                             c21 = 264.0_stnd,    &
                             c22 = 294.0_stnd,    &
                             c23 = 346.0_stnd,    &
                             c24 = 420.0_stnd,    &
                             c25 = 462.0_stnd,    &
                             c26 = 606.0_stnd,    &
                             c27 = 672.0_stnd,    &
                             c28 = 707.0_stnd,    &
                             c29 = 735.0_stnd,    &
                             c30 = 889.0_stnd,    &
                             c31 = 932.0_stnd,    &
                             c32 = 966.0_stnd,    &
                             c33 = 1141.0_stnd,   &
                             c34 = 1182.0_stnd,   &
                             c35 = 1278.0_stnd,   &
                             c36 = 1740.0_stnd,   &
                             c37 = 2520.0_stnd,   &
                             c38 = 5040.0_stnd
!
    character(len=*),  parameter :: name_proc='pinvq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( p )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:n)<zero) .or. any(p(:n)>one) ) then
        call merror( name_proc//prob_error1 )
    end if
!
    if ( df<half ) then
        call merror( name_proc//prob_error15 )
    end if
!
!   DETERMINE THE REQUIRED PRECISION.
!   prec2 DETERMINE THE PRECISION OF THE RESULT. IF MORE THAN
!   SIX SIGNIFICANT FIGURES ARE REQUIRED prec2 SHOULD BE DECREASED.
!
    prec2 = 0.5e-06_stnd
    if ( present(prec) ) then
        if ( prec>zero .and. prec<prec2 ) prec2 = prec
    end if
!
!   INITIALISE.
!
    xx = half*df
    g = lngamma_r( xx )
!
    c  = xx - one
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,done,pi,chi2i,a,b,p1,p2,q,s1,s2,s3,s4,s5,s6,t,x)
!
    do i = 1_i4b, n
!
        pi = p(i)
!
        if ( pi==zero ) then
!
            chi2i = zero
!
        else if ( pi==one ) then
!
            chi2i = machhuge
!
        else
!
            done = false
!
            if ( df<-c5*log(pi) ) then
!
!               STARTING APPROXIMATION FOR SMALL CHI-SQUARED.
!
                chi2i = ( pi*xx*exp(g + xx*aa))**(one/xx)
!
                if ( chi2i<prec2 ) done = true
!
            else if ( df<=c3 ) then
!
!               STARTING APPROXIMATION FOR df LESS THAN OR EQUAL TO 0.32 .
!
                chi2i = c4
                a = log( one-pi )
!
                do
!
                    q  = chi2i
                    p1 = one + chi2i*( c7 + chi2i )
                    p2 = chi2i*( c9 + chi2i*( c8 + chi2i ) )
                    t  = -half + ( c7 + two*chi2i )/p1 - ( c9 + chi2i*( c10 + three*chi2i ) )/p2
                    chi2i = chi2i - ( one - exp( a + g + half*chi2i + c*aa )*p2/p1 )/t
!
                    if (abs(q / chi2i - one)<= c1) exit
!
                end do
!
            else
!
!               STARTING APPROXIMATION USING WILSON AND HILFERTY ESTIMATE.
!
                x = pinvn_r( pi )
!
                p1 = c2/df
                chi2i = df*( x*sqrt(p1) + one - p1 )**3_i4b
!
!               STARTING APPROXIMATION FOR p TENDING TO 1.
!
                if (chi2i > c6*df + six) then
                    chi2i = -two*( log(one-pi) - c*log(half*chi2i) + g )
                end if
!
            end if
!
!           CALCULATION OF SEVEN TERM TAYLOR SERIES.
!
            do j = 1_i4b, maxit
!
                if ( done ) exit
!
                q  = chi2i
                p1 = half*chi2i
                p2 = pi - probgamma2_r( p1, xx, acu=acu, maxiter=maxiter )
!
                t = p2*exp( xx*aa + g + p1 - c*log(chi2i) )
                b = t/chi2i
                a = half*t - b*c
                s1 = ( c19 + a*( c17 + a*(c14 + a*(c13 + a*(c12 + c11*a))))) / c24
                s2 = ( c24 + a*( c29 + a*(c32 + a*(c33 + c35*a)))) / c37
                s3 = ( c19 + a*( c25 + a*(c28 + c31*a))) / c37
                s4 = ( c20 + a*( c27 + c34*a) + c*(c22 + a*(c30 + c36*a))) / c38
                s5 = ( c13 + c21*a + c*(c18 + c26*a)) / c37
                s6 = ( c15 + c*( c23 + c16*c)) / c38
                chi2i = chi2i + t*(one + half*t*s1 - b*c*(s1 - b*  &
                       (s2 - b*(s3 - b*(s4 - b*(s5 - b*s6))))))
!
                done = abs( q/chi2i - one )<=prec2
!
            end do
!
!           HERE IF j>maxit THEN THE RESULT IS PROBABLY AS ACCURATE AS THE MACHINE WILL ALLOW.
!           BUT IT IS SAFE TO GENERATE AN ERROR MESSAGE.
!
!            if ( j>maxit )  call merror( name_proc//prob_error17 )
!
        end if
!
        chi2(i) = chi2i
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvq2_rv = chi2(:n)
!
!
! END OF FUNCTION pinvq2_rv
! _________________________
!
    end function pinvq2_rv
!
! =========================================================================================
!
    function pinvq2_rv2( p, df, prec, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                           X0 = PINVQ2( P(i), DF(i) ) 
!   
!   , if P(i) = probability( U < X0 )  for U = Chi-squared(DF(i)) and i=1 to size(P).
!
!   PINVQ2 returns the quantiles X0(i) of the chi-squared distribution with DF(i) degrees of 
!   freedom corresponding to a given lower tail area of  P(i), for i=1 to size(P).
!   In other words, PINVQ2 outputs chi-squared values, X0(:), such that random variables,
!   distributed as chi-squared with DF(:) degrees of freedom, will be less than or equal to X0(:)
!   with associated probabilities P(:).
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd), dimension(:)
!           On entry, the probabilities. P(i) must verify 0. < P(i) < 1,
!           for i=1 to size(P) . 
!
!   DF      (INPUT) real(stnd), dimension(:)
!           On entry, degrees of freedom of the chi-squared distribution.
!           Any value in the array DF must be greater or equal to 0.5.
!
!           The size of DF must verify  size(DF) = size(P).
!
!   PREC    (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accurary of the result. If more than six significant
!           digits are required, the default value of PREC (e.g. 0.5e-06_stnd) should be
!           altered appropriately(e.g. decreased). PREC is a small strictly positive 
!           integer less than 0.5e-06_stnd.
!
!           The default value for PREC is 0.5e-06_stnd .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result when computing the incomplete 
!           Gamma integral in the evaluation of the seven term Taylor series.
!           If l decimal places of accuracy are required then ACU should
!           be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!
!           The default value is 1000.
!
!           See the description of the PROBGAMMA2 function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function is both more general (here the numbers of degrees of freedom, DF(:),
!   are not necessarily integers) and more accurate (here the quantiles X0(:) may be calculated
!   as exactly as the computer allows with the parameter PREC) than PINVQ function.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Best, D.J., and Roberts, D.E., 1975: Algorithm AS 91: The Percentage Points of 
!                     the chi2 Distribution. Appl. Statist., Vol.24, No. 3, pp.385-388
!
!   (2)   Shea, B.L., 1991: Algorithm AS R85: A remark on AS 91: The Percentage Points of
!                     the chi2 Distribution. Appl. Statist. Vol.40, No. 1, pp.233-235.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one, two, three, six
    use Logical_Constants, only : true, false
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error1, prob_error15
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(in)           :: p, df
    real(stnd),               intent(in), optional :: prec, acu
!
    real(stnd), dimension(size(p)) :: pinvq2_rv2
!
    integer(i4b), intent(in), optional :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, p1, p2, q, s1, s2, s3, s4, s5, s6, &
                    t, x, xx, g, prec2, chi2i, pi, dfi
    real(stnd), dimension(size(p)) :: chi2
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: done
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
!   maxit IS THE MAXIMUM NUMBER OF ITERATIONS OF THE TAYLOR SERIES EXPANSION.
!
    integer(i4b), parameter ::   maxit = 20_i4b
!
    real(stnd), parameter :: aa = 0.6931471806_stnd
!
    real(stnd), parameter ::                      &
                             c1 = 0.01_stnd,      &
                             c2 = 0.222222_stnd,  &
                             c3 = 0.32_stnd,      &
                             c4 = 0.4_stnd,       &
                             c5 = 1.24_stnd,      &
                             c6 = 2.2_stnd,       &
                             c7 = 4.67_stnd,      &
                             c8 = 6.66_stnd,      &
                             c9 = 6.73_stnd,      &
                             c10 = 13.32_stnd,    &
                             c11 = 60.0_stnd,     &
                             c12 = 70.0_stnd,     &
                             c13 = 84.0_stnd,     &
                             c14 = 105.0_stnd,    &
                             c15 = 120.0_stnd,    &
                             c16 = 127.0_stnd,    &
                             c17 = 140.0_stnd,    &
                             c18 = 175.0_stnd,    &
                             c19 = 210.0_stnd,    &
                             c20 = 252.0_stnd,    &
                             c21 = 264.0_stnd,    &
                             c22 = 294.0_stnd,    &
                             c23 = 346.0_stnd,    &
                             c24 = 420.0_stnd,    &
                             c25 = 462.0_stnd,    &
                             c26 = 606.0_stnd,    &
                             c27 = 672.0_stnd,    &
                             c28 = 707.0_stnd,    &
                             c29 = 735.0_stnd,    &
                             c30 = 889.0_stnd,    &
                             c31 = 932.0_stnd,    &
                             c32 = 966.0_stnd,    &
                             c33 = 1141.0_stnd,   &
                             c34 = 1182.0_stnd,   &
                             c35 = 1278.0_stnd,   &
                             c36 = 1740.0_stnd,   &
                             c37 = 2520.0_stnd,   &
                             c38 = 5040.0_stnd
!
    character(len=*),  parameter :: name_proc='pinvq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(p),i4b)   ,   &
                    int(size(df),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:n)<zero) .or. any(p(:n)>one) ) then
        call merror( name_proc//prob_error1 )
    end if
!
    if ( any(df(:n)<half) ) then
        call merror( name_proc//prob_error15 )
    end if
!
!   DETERMINE THE REQUIRED PRECISION.
!   prec2 DETERMINE THE PRECISION OF THE RESULT. IF MORE THAN
!   SIX SIGNIFICANT FIGURES ARE REQUIRED prec2 SHOULD BE DECREASED.
!
    prec2 = 0.5e-06_stnd
    if ( present(prec) ) then
        if ( prec>zero .and. prec<prec2 ) prec2 = prec
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,done,pi,dfi,chi2i,a,b,c,g,p1,p2,q,s1,s2,s3,s4,s5,s6,t,x,xx)
!
    do i = 1_i4b, n
!
        pi  = p(i)
        dfi = df(i)
!
        if ( pi==zero ) then
!
            chi2i = zero
!
        else if ( pi==one ) then
!
            chi2i = machhuge
!
        else
!
!           INITIALISE.
!
            xx = half*dfi
            g = lngamma_r( xx )
!
            c  = xx - one
!
            done = false
!
            if ( dfi<-c5*log(pi) ) then
!
!               STARTING APPROXIMATION FOR SMALL CHI-SQUARED.
!
                chi2i = ( pi*xx*exp(g + xx*aa))**(one/xx)
                if ( chi2i<prec2 ) done = true
!
            else if ( dfi<=c3 ) then
!
!               STARTING APPROXIMATION FOR df LESS THAN OR EQUAL TO 0.32 .
!
                chi2i = c4
                a = log( one-pi )
!
                do
!
                    q  = chi2i
                    p1 = one + chi2i*( c7 + chi2i )
                    p2 = chi2i*( c9 + chi2i*( c8 + chi2i ) )
                    t  = -half + ( c7 + two*chi2i )/p1 - ( c9 + chi2i*( c10 + three*chi2i ) )/p2
                    chi2i = chi2i - ( one - exp( a + g + half*chi2i + c*aa )*p2/p1 )/t
!
                    if (abs(q / chi2i - one)<= c1) exit
!
                end do
!
            else
!
!               STARTING APPROXIMATION USING WILSON AND HILFERTY ESTIMATE.
!
                x = pinvn_r( pi )
!
                p1 = c2/dfi
                chi2i = dfi*( x*sqrt(p1) + one - p1 )**3_i4b
!
!               STARTING APPROXIMATION FOR p TENDING TO 1.
!
                if (chi2i > c6*dfi + six) then
                    chi2i = -two*( log(one-pi) - c*log(half*chi2i) + g )
                end if
!
            end if
!
!           CALCULATION OF SEVEN TERM TAYLOR SERIES.
!
            do j = 1_i4b, maxit
!
                if ( done ) exit
!
                q  = chi2i
                p1 = half*chi2i
                p2 = pi - probgamma2_r( p1, xx, acu=acu, maxiter=maxiter )
!
                t = p2*exp( xx*aa + g + p1 - c*log(chi2i) )
                b = t/chi2i
                a = half*t - b*c
                s1 = ( c19 + a*( c17 + a*(c14 + a*(c13 + a*(c12 + c11*a))))) / c24
                s2 = ( c24 + a*( c29 + a*(c32 + a*(c33 + c35*a)))) / c37
                s3 = ( c19 + a*( c25 + a*(c28 + c31*a))) / c37
                s4 = ( c20 + a*( c27 + c34*a) + c*(c22 + a*(c30 + c36*a))) / c38
                s5 = ( c13 + c21*a + c*(c18 + c26*a)) / c37
                s6 = ( c15 + c*( c23 + c16*c)) / c38
                chi2i = chi2i + t*(one + half*t*s1 - b*c*(s1 - b*  &
                       (s2 - b*(s3 - b*(s4 - b*(s5 - b*s6))))))
!
                done = abs( q/chi2i - one )<=prec2
!
            end do
!
!           HERE IF j>maxit THEN THE RESULT IS PROBABLY AS ACCURATE AS THE MACHINE WILL ALLOW.
!           BUT IT IS SAFE TO GENERATE AN ERROR MESSAGE.
!
!            if ( j>maxit )  call merror( name_proc//prob_error17 )
!
        end if
!
        chi2(i) = chi2i
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvq2_rv2 = chi2(:n)
!
!
! END OF FUNCTION pinvq2_rv2
! __________________________
!
    end function pinvq2_rv2
!
! =========================================================================================
!
    function pinvq2_rm( p, df, prec, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                           X0 = PINVQ2( P(i,j), DF )
!   
!   , if P(i,j) = probability( U < X0 )  for U = Chi-squared(DF), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVQ2 returns the quantiles X0(i,j) of the chi-squared distribution with DF degrees of 
!   freedom corresponding to a given lower tail area of P(i,j), for i=1 to size(P) and j=1 to size(P,2).
!   In other words, PINVQ2 outputs chi-squared values, X0(:,:), such that random variables,
!   distributed as chi-squared with DF degrees of freedom, will be less than or equal to X0(:,:)
!   with associated probabilities P(:,:).
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd), dimension(:,:)
!           On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1,
!           for i=1 to size(P) and j=1 to size(P,2).
!
!   DF      (INPUT) real(stnd)
!           On entry, degrees of freedom of the chi-squared distribution.
!           DF must be greater or equal to 0.5.
!
!   PREC    (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accurary of the result. If more than six significant
!           digits are required, the default value of PREC (e.g. 0.5e-06_stnd) should be
!           altered appropriately(e.g. decreased). PREC is a small strictly positive 
!           integer less than 0.5e-06_stnd.
!
!           The default value for PREC is 0.5e-06_stnd .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result when computing the incomplete 
!           Gamma integral in the evaluation of the seven term Taylor series.
!           If l decimal places of accuracy are required then ACU should
!           be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!
!           The default value is 1000.
!
!           See the description of the PROBGAMMA2 function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function is both more general (here the number of degrees of freedom, DF,
!   is not necessarily an integer) and more accurate (here the quantiles X0(:,:) may be calculated
!   as exactly as the computer allows with the parameter PREC) than PINVQ function.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Best, D.J., and Roberts, D.E., 1975: Algorithm AS 91: The Percentage Points of 
!                     the chi2 Distribution. Appl. Statist., Vol.24, No. 3, pp.385-388
!
!   (2)   Shea, B.L., 1991: Algorithm AS R85: A remark on AS 91: The Percentage Points of
!                     the chi2 Distribution. Appl. Statist. Vol.40, No. 1, pp.233-235.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one, two, three, six
    use Logical_Constants, only : true, false
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error1, prob_error15
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: p
    real(stnd),                 intent(in)           :: df
    real(stnd),                 intent(in), optional :: prec, acu
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvq2_rm
!
    integer(i4b), intent(in), optional :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, p1, p2, q, s1, s2, s3, s4, s5, s6, &
                    t, x, xx, g, prec2, chi2ki, pki
    real(stnd), dimension(size(p,1),size(p,2)) :: chi2
!
    integer(i4b) :: m, n, i, j, k
!
    logical(lgl) :: done
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
!   maxit IS THE MAXIMUM NUMBER OF ITERATIONS OF THE TAYLOR SERIES EXPANSION.
!
    integer(i4b), parameter ::   maxit = 20_i4b
!
    real(stnd), parameter :: aa = 0.6931471806_stnd
!
    real(stnd), parameter ::                      &
                             c1 = 0.01_stnd,      &
                             c2 = 0.222222_stnd,  &
                             c3 = 0.32_stnd,      &
                             c4 = 0.4_stnd,       &
                             c5 = 1.24_stnd,      &
                             c6 = 2.2_stnd,       &
                             c7 = 4.67_stnd,      &
                             c8 = 6.66_stnd,      &
                             c9 = 6.73_stnd,      &
                             c10 = 13.32_stnd,    &
                             c11 = 60.0_stnd,     &
                             c12 = 70.0_stnd,     &
                             c13 = 84.0_stnd,     &
                             c14 = 105.0_stnd,    &
                             c15 = 120.0_stnd,    &
                             c16 = 127.0_stnd,    &
                             c17 = 140.0_stnd,    &
                             c18 = 175.0_stnd,    &
                             c19 = 210.0_stnd,    &
                             c20 = 252.0_stnd,    &
                             c21 = 264.0_stnd,    &
                             c22 = 294.0_stnd,    &
                             c23 = 346.0_stnd,    &
                             c24 = 420.0_stnd,    &
                             c25 = 462.0_stnd,    &
                             c26 = 606.0_stnd,    &
                             c27 = 672.0_stnd,    &
                             c28 = 707.0_stnd,    &
                             c29 = 735.0_stnd,    &
                             c30 = 889.0_stnd,    &
                             c31 = 932.0_stnd,    &
                             c32 = 966.0_stnd,    &
                             c33 = 1141.0_stnd,   &
                             c34 = 1182.0_stnd,   &
                             c35 = 1278.0_stnd,   &
                             c36 = 1740.0_stnd,   &
                             c37 = 2520.0_stnd,   &
                             c38 = 5040.0_stnd
!
    character(len=*),  parameter :: name_proc='pinvq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( p, 1 )
    n = size( p, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:m,:n)<zero) .or. any(p(:m,:n)>one) ) then
        call merror( name_proc//prob_error1 )
    end if
!
    if ( df<half ) then
        call merror( name_proc//prob_error15 )
    end if
!
!   DETERMINE THE REQUIRED PRECISION.
!   prec2 DETERMINE THE PRECISION OF THE RESULT. IF MORE THAN
!   SIX SIGNIFICANT FIGURES ARE REQUIRED prec2 SHOULD BE DECREASED.
!
    prec2 = 0.5e-06_stnd
    if ( present(prec) ) then
        if ( prec>zero .and. prec<prec2 ) prec2 = prec
    end if
!
!   INITIALISE.
!
    xx = half*df
    g = lngamma_r( xx )
!
    c  = xx - one
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,k,done,pki,chi2ki,a,b,p1,p2,q,s1,s2,s3,s4,s5,s6,t,x)
!
    do i = 1_i4b, n
!
        do k = 1_i4b, m
!
            pki = p(k,i)
!
            if ( pki==zero ) then
!
                chi2ki = zero
!
            else if ( pki==one ) then
!
                chi2ki = machhuge
!
            else
!
                done = false
!
                if ( df<-c5*log(pki) ) then
!
!                   STARTING APPROXIMATION FOR SMALL CHI-SQUARED.
!
                    chi2ki = ( pki*xx*exp(g + xx*aa))**(one/xx)
                    if ( chi2ki<prec2 ) done = true
!
                else if ( df<=c3 ) then
!
!                   STARTING APPROXIMATION FOR df LESS THAN OR EQUAL TO 0.32 .
!
                    chi2ki = c4
                    a = log( one-pki )
                    do
                        q  = chi2ki
                        p1 = one + chi2ki*( c7 + chi2ki )
                        p2 = chi2ki*( c9 + chi2ki*( c8 + chi2ki ) )
                        t  = -half + ( c7 + two*chi2ki )/p1 - ( c9 + chi2ki*( c10 + three*chi2ki ) )/p2
                        chi2ki = chi2ki - ( one - exp( a + g + half*chi2ki + c*aa )*p2/p1 )/t
                        if (abs(q / chi2ki - one)<= c1) exit
                    end do
!
                else
!
!                   STARTING APPROXIMATION USING WILSON AND HILFERTY ESTIMATE.
!
                    x = pinvn_r( pki )
!
                    p1 = c2/df
                    chi2ki = df*( x*sqrt(p1) + one - p1 )**3_i4b
!
!                   STARTING APPROXIMATION FOR p TENDING TO 1.
!
                    if (chi2ki > c6*df + six) then
                        chi2ki = -two*( log(one-pki) - c*log(half*chi2ki) + g )
                    end if
!
                end if
!
!               CALCULATION OF SEVEN TERM TAYLOR SERIES.
!
                do j = 1_i4b, maxit
!
                    if ( done ) exit
!
                    q  = chi2ki
                    p1 = half*chi2ki
                    p2 = pki - probgamma2_r( p1, xx, acu=acu, maxiter=maxiter )
!
                    t = p2*exp( xx*aa + g + p1 - c*log(chi2ki) )
                    b = t/chi2ki
                    a = half*t - b*c
                    s1 = ( c19 + a*( c17 + a*(c14 + a*(c13 + a*(c12 + c11*a))))) / c24
                    s2 = ( c24 + a*( c29 + a*(c32 + a*(c33 + c35*a)))) / c37
                    s3 = ( c19 + a*( c25 + a*(c28 + c31*a))) / c37
                    s4 = ( c20 + a*( c27 + c34*a) + c*(c22 + a*(c30 + c36*a))) / c38
                    s5 = ( c13 + c21*a + c*(c18 + c26*a)) / c37
                    s6 = ( c15 + c*( c23 + c16*c)) / c38
                    chi2ki = chi2ki + t*(one + half*t*s1 - b*c*(s1 - b*  &
                             (s2 - b*(s3 - b*(s4 - b*(s5 - b*s6))))))
!
                    done = abs( q/chi2ki - one )<=prec2
!
                end do
!
!               HERE IF j>maxit THEN THE RESULT IS PROBABLY AS ACCURATE AS THE MACHINE WILL ALLOW.
!               BUT IT IS SAFE TO GENERATE AN ERROR MESSAGE.
!
!                if ( j>maxit )  call merror( name_proc//prob_error17 )
!
            end if
!
            chi2(k,i) = chi2ki
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvq2_rm = chi2(:m,:n)
!
!
! END OF FUNCTION pinvq2_rm
! _________________________
!
    end function pinvq2_rm
!
! =========================================================================================
!
    function pinvq2_rm2( p, df, prec, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the inverse of the chi-squared distribution function:
!
!                X0 = PINVQ2( P(i,j), DF(i,j) )
!   
!   , if P(i,j) = probability( U < X0 )  for U = Chi-squared(DF(i,j)), i=1 to size(P,1) and j=1 to size(P,2).
!
!   PINVQ2 returns the quantiles X0(i,j) of the chi-squared distribution with DF(i,j) degrees of 
!   freedom corresponding to a given lower tail area of P(i,j), for i=1 to size(P) and j=1 to size(P,2).
!   In other words, PINVQ2 outputs chi-squared values, X0(:,:), such that random variables,
!   distributed as chi-squared with corresponding DF(:,:) degrees of freedom, will be less than or
!   equal to X0(:,:) with associated probabilities P(:,:).
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd), dimension(:,:)
!           On entry, the probabilities. P(i,j) must verify 0. < P(i,j) < 1,
!           for i=1 to size(P) and j=1 to size(P,2).
!
!   DF      (INPUT) real(stnd), dimension(:,:)
!           On entry, degrees of freedom of the chi-squared distribution.
!           Any value in the array DF must be greater or equal to 0.5.
!
!           The shape of DF must verify:
!
!           - size(DF,1) = size(P,1)
!           - size(DF,2) = size(P,2).
!
!   PREC    (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accurary of the result. If more than six significant
!           digits are required, the default value of PREC (e.g. 0.5e-06_stnd) should be
!           altered appropriately(e.g. decreased). PREC is a small strictly positive 
!           integer less than 0.5e-06_stnd.
!
!           The default value for PREC is 0.5e-06_stnd .
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result when computing the incomplete 
!           Gamma integral in the evaluation of the seven term Taylor series.
!           If l decimal places of accuracy are required then ACU should
!           be set to 10**(-(l+1)). ACU is a small strictly positive integer.
!           ACU should not be set smaller than the machine precision since the stated accuracy
!           cannot be attained. In that case the machine precision is used instead.
!
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the Pearson's series or continued fraction expansion of the
!           incomplete Gamma integral.
!
!           The default value is 1000.
!
!           See the description of the PROBGAMMA2 function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function is both more general (here the numbers of degrees of freedom, DF(:,:),
!   are not necessarily integers) and more accurate (here the quantiles X0(:,:) may be calculated
!   as exactly as the computer allows with the parameter PREC) than PINVQ function.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1)   Best, D.J., and Roberts, D.E., 1975: Algorithm AS 91: The Percentage Points of 
!                     the chi2 Distribution. Appl. Statist., Vol.24, No. 3, pp.385-388
!
!   (2)   Shea, B.L., 1991: Algorithm AS R85: A remark on AS 91: The Percentage Points of
!                     the chi2 Distribution. Appl. Statist. Vol.40, No. 1, pp.233-235.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one, two, three, six
    use Logical_Constants, only : true, false
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : prob_error1, prob_error15
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: p, df
    real(stnd),                 intent(in), optional :: prec, acu
!
    real(stnd), dimension(size(p,1),size(p,2)) :: pinvq2_rm2
!
    integer(i4b), intent(in), optional :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, p1, p2, q, s1, s2, s3, s4, s5, s6, &
                    t, x, xx, g, prec2, chi2ki, pki, dfki
    real(stnd), dimension(size(p,1),size(p,2)) :: chi2
!
    integer(i4b) :: m, n, i, j, k
!
    logical(lgl) :: done
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
!   maxit IS THE MAXIMUM NUMBER OF ITERATIONS OF THE TAYLOR SERIES EXPANSION.
!
    integer(i4b), parameter ::   maxit = 20_i4b
!
    real(stnd), parameter :: aa = 0.6931471806_stnd
!
    real(stnd), parameter ::                      &
                             c1 = 0.01_stnd,      &
                             c2 = 0.222222_stnd,  &
                             c3 = 0.32_stnd,      &
                             c4 = 0.4_stnd,       &
                             c5 = 1.24_stnd,      &
                             c6 = 2.2_stnd,       &
                             c7 = 4.67_stnd,      &
                             c8 = 6.66_stnd,      &
                             c9 = 6.73_stnd,      &
                             c10 = 13.32_stnd,    &
                             c11 = 60.0_stnd,     &
                             c12 = 70.0_stnd,     &
                             c13 = 84.0_stnd,     &
                             c14 = 105.0_stnd,    &
                             c15 = 120.0_stnd,    &
                             c16 = 127.0_stnd,    &
                             c17 = 140.0_stnd,    &
                             c18 = 175.0_stnd,    &
                             c19 = 210.0_stnd,    &
                             c20 = 252.0_stnd,    &
                             c21 = 264.0_stnd,    &
                             c22 = 294.0_stnd,    &
                             c23 = 346.0_stnd,    &
                             c24 = 420.0_stnd,    &
                             c25 = 462.0_stnd,    &
                             c26 = 606.0_stnd,    &
                             c27 = 672.0_stnd,    &
                             c28 = 707.0_stnd,    &
                             c29 = 735.0_stnd,    &
                             c30 = 889.0_stnd,    &
                             c31 = 932.0_stnd,    &
                             c32 = 966.0_stnd,    &
                             c33 = 1141.0_stnd,   &
                             c34 = 1182.0_stnd,   &
                             c35 = 1278.0_stnd,   &
                             c36 = 1740.0_stnd,   &
                             c37 = 2520.0_stnd,   &
                             c38 = 5040.0_stnd
!
    character(len=*),  parameter :: name_proc='pinvq2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(p,1),i4b)    ,   &
                    int(size(df,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(p,2),i4b)    ,   &
                    int(size(df,2),i4b)   ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(p(:m,:n)<zero) .or. any(p(:m,:n)>one) ) then
        call merror( name_proc//prob_error1 )
    end if
!
    if ( any(df(:m,:n)<half) ) then
        call merror( name_proc//prob_error15 )
    end if
!
!   DETERMINE THE REQUIRED PRECISION.
!   prec2 DETERMINE THE PRECISION OF THE RESULT. IF MORE THAN
!   SIX SIGNIFICANT FIGURES ARE REQUIRED prec2 SHOULD BE DECREASED.
!
    prec2 = 0.5e-06_stnd
    if ( present(prec) ) then
        if ( prec>zero .and. prec<prec2 ) prec2 = prec
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,k,done,pki,dfki,chi2ki,a,b,c,g,p1,p2,q,s1,s2,s3,s4,s5,s6,t,x,xx)
!
    do i = 1_i4b, n
!
        do k = 1_i4b, m
!
            pki  = p(k,i)
            dfki = df(k,i)
!
            if ( pki==zero ) then
!
                chi2ki = zero
!
            else if ( pki==one ) then
!
                chi2ki = machhuge
!
            else
!
!               INITIALISE.
!
                xx = half*dfki
                g = lngamma_r( xx )
!
                c  = xx - one
!
                done = false
!
                if ( dfki<-c5*log(pki) ) then
!
!                   STARTING APPROXIMATION FOR SMALL CHI-SQUARED.
!
                    chi2ki = ( pki*xx*exp(g + xx*aa))**(one/xx)
                    if ( chi2ki<prec2 ) done = true
!
                else if ( dfki<=c3 ) then
!
!                   STARTING APPROXIMATION FOR df LESS THAN OR EQUAL TO 0.32 .
!
                    chi2ki = c4
                    a = log( one-pki )
                    do
                        q  = chi2ki
                        p1 = one + chi2ki*( c7 + chi2ki )
                        p2 = chi2ki*( c9 + chi2ki*( c8 + chi2ki ) )
                        t  = -half + ( c7 + two*chi2ki )/p1 - ( c9 + chi2ki*( c10 + three*chi2ki ) )/p2
                        chi2ki = chi2ki - ( one - exp( a + g + half*chi2ki + c*aa )*p2/p1 )/t
                        if (abs(q / chi2ki - one)<= c1) exit
                    end do
!
                else
!
!                   STARTING APPROXIMATION USING WILSON AND HILFERTY ESTIMATE.
!
                    x = pinvn_r( pki )
!
                    p1 = c2/dfki
                    chi2ki = dfki*( x*sqrt(p1) + one - p1 )**3_i4b
!
!                   STARTING APPROXIMATION FOR p TENDING TO 1.
!
                    if (chi2ki > c6*dfki + six) then
                        chi2ki = -two*( log(one-pki) - c*log(half*chi2ki) + g )
                    end if
!
                end if
!
!               CALCULATION OF SEVEN TERM TAYLOR SERIES.
!
                do j = 1_i4b, maxit
!
                    if ( done ) exit
!
                    q  = chi2ki
                    p1 = half*chi2ki
                    p2 = pki - probgamma2_r( p1, xx, acu=acu, maxiter=maxiter )
!
                    t = p2*exp( xx*aa + g + p1 - c*log(chi2ki) )
                    b = t/chi2ki
                    a = half*t - b*c
                    s1 = ( c19 + a*( c17 + a*(c14 + a*(c13 + a*(c12 + c11*a))))) / c24
                    s2 = ( c24 + a*( c29 + a*(c32 + a*(c33 + c35*a)))) / c37
                    s3 = ( c19 + a*( c25 + a*(c28 + c31*a))) / c37
                    s4 = ( c20 + a*( c27 + c34*a) + c*(c22 + a*(c30 + c36*a))) / c38
                    s5 = ( c13 + c21*a + c*(c18 + c26*a)) / c37
                    s6 = ( c15 + c*( c23 + c16*c)) / c38
                    chi2ki = chi2ki + t*(one + half*t*s1 - b*c*(s1 - b*  &
                             (s2 - b*(s3 - b*(s4 - b*(s5 - b*s6))))))
!
                    done = abs( q/chi2ki - one )<=prec2
!
                end do
!
!               HERE IF j>maxit THEN THE RESULT IS PROBABLY AS ACCURATE AS THE MACHINE WILL ALLOW.
!               BUT IT IS SAFE TO GENERATE AN ERROR MESSAGE.
!
!                if ( j>maxit )  call merror( name_proc//prob_error17 )
!
            end if
!
            chi2(k,i) = chi2ki
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    pinvq2_rm2 = chi2(:m,:n)
!
!
! END OF FUNCTION pinvq2_rm2
! __________________________
!
    end function pinvq2_rm2
!
!
! =========================================================================================
!                             F  DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function probf_r( f, ndf1, ndf2, upper )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom NDF1 and NDF2 
!   from F to infinity if UPPER is true or from zero to F if UPPER is false. In other words, if:
!
!   - UPPER = true  :    PROBF = prob( U > F )  ,
!   - UPPER = false :    PROBF = prob( U < F )  ,
!   
!   , for U = Fisher(NDF1,NDF2).
!
!
! Arguments
! _________
!
!   F      (INPUT) real(stnd)
!          On entry, upper or lower limit of integration. F must be greater
!          or equal to zero.
!
!   NDF1   (INPUT) integer(i4b)
!          On entry, first degree of freedom of the F distribution (numerator).
!          NDF1 must be greater or equal to 1.
!
!   NDF2   (INPUT) integer(i4b)
!          On entry, second degree of freedom of the F distribution (denominator).
!          NDF2 must be greater or equal to 1.
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  :  probability to the right of F is calculated.
!          - UPPER = false  : probability to the left of F is calculated.
!
!
! Further Details
! _______________
!
!   This function accepts only integer values of degree of freedom and uses a normal
!   approximation. This normal approximation is not accurate for small values of degrees
!   of freedom.
!
!   This function is adapted from
!
!   (1) Peizer, D.B., and Pratt, J.W., 1968: A normal approximation for Binomial, F, Beta, and ...,
!                                (formula 2.24a), J.A.S.A., Vol. 63, 1457-1483
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, six, ten, half, third, sixth, tenth,   &
                                  c15, c21, c4_m2
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error5
    use Utilities,         only : merror, geop
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in) :: f
    real(stnd)               :: probf_r
!
    integer(i4b), intent(in) :: ndf1, ndf2
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: x, p, q, g, pg, qg, df1, df2, f1, f2, fn, safmin
!
    integer(i4b) :: k 
!   
!   
! PARAMETERS 
! __________
!
    real(stnd),  parameter, dimension(5) :: coef=(/ three, six, ten, c15, c21 /)
!
    character(len=*),  parameter :: name_proc='probf'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( ndf1<1_i4b .or. ndf2<1_i4b .or. f<zero )  then
        call merror( name_proc//prob_error5 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    safmin = lamch( 's' )
!
    if ( f<safmin ) then
!
        if ( upper ) then
            probf_r = one
        else
            probf_r = zero
        end if
!
        return
!
    end if
!
    x = f
!
    if ( ndf1==1_i4b ) then
!
!       USE STUDENT T-DISTRIBUTION FOR NDF1=1 .
!
        g = sqrt( x )
!
        probf_r = two*probt_r( g, ndf2, true )
!
        if ( .not. upper ) probf_r = one - probf_r
!
    else if ( ndf2==1_i4b ) then
!
!       USE STUDENT T-DISTRIBUTION FOR NDF2=1 .
!
        g = sqrt( one/x )
!
        probf_r = two*probt_r( g, ndf1, true )
!
        if ( upper ) probf_r = one - probf_r
!
    else
!
!       OTHERWISE USE PEIZER-PRATT FORMULA (GAUSSIAN APPROXIMATION).
!
        df1 = real( ndf1, stnd )
        df2 = real( ndf2, stnd )
        f1 = ( df1 - one )/two
        f2 = ( df2 - one )/two
        fn = ( df1 + df2 - two )/two
!
        p = df2/( df2 + x*df1 )
        q = one - p
!
        x = f2/fn/p
        pg = one - x
!
        do k = 1_i4b, 2_i4b
!
            if ( abs(pg)>tenth ) then
                g = ( one - x*x + two*x*log(merge(x,one,x/=zero)) )/( pg*pg ) 
            else
                g = sum( geop(pg, pg, 5_i4b)/coef(1_i4b:5_i4b) )
            end if
!
            if ( k==2_i4b ) exit
!
            qg = g
            x = f1/fn/q
            pg = one - x
!
        end do
!
        pg = g
        g = q/df2 - p/df1 + ( q - half )/( df1 + df2 )
        x = c4_m2*g + ( f2 + sixth ) - p*( fn + third )
        g = ( one + q*qg + p*pg )/( fn + sixth )/p/q
        q = x*sqrt( g )
!
        probf_r = probn_r( q, upper )
!
    end if
!
!
! END OF FUNCTION probf_r
! _______________________
!
    end function probf_r
!
! =========================================================================================
!
    function probf_rv( f, ndf1, ndf2, upper )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom NDF1 and NDF2 
!   from F(i) to infinity if UPPER is true or from zero to F(i) if UPPER is false,
!   for i=1 to size(F). In other words, if:
!
!   - UPPER = true  :    PROBF( i ) = prob( U > F(i) )  ,
!   - UPPER = false :    PROBF( i ) = prob( U < F(i) )  ,
!   
!   , for U = Fisher(NDF1,NDF2) and i=1 to size(F).
!
!
! Arguments
! _________
!
!   F      (INPUT) real(stnd), dimension(:)
!          On entry, upper or lower limits of integration. F(i) must be greater
!          or equal to zero for i=1 to size(F).
!
!   NDF1   (INPUT) integer(i4b)
!          On entry, first degree of freedom of the F distribution (numerator).
!          NDF1 must be greater or equal to 1.
!
!   NDF2   (INPUT) integer(i4b)
!          On entry, second degree of freedom of the F distribution (denominator).
!          NDF2 must be greater or equal to 1.
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  :  probability to the right of F is calculated.
!          - UPPER = false  : probability to the left of F is calculated.
!
!
! Further Details
! _______________
!
!   This function accepts only integer values of degree of freedom and uses a normal
!   approximation. This normal approximation is not accurate for small values of degrees
!   of freedom.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Peizer, D.B., and Pratt, J.W., 1968: A normal approximation for Binomial, F, Beta, and ...,
!                                (formula 2.24a), J.A.S.A., Vol. 63, 1457-1483
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, six, ten, half, third, sixth, tenth,   &
                                  c15, c21, c4_m2
    use Num_Constants,     only : lamch
    use Char_Constants,    only : allocate_error, prob_error5
    use Utilities,         only : merror, geop
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),       intent(in) :: f
    real(stnd), dimension(size(f))             :: probf_rv
!
    integer(i4b), intent(in) :: ndf1, ndf2
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: ffi, x, p, q, g, pg, qg, df1, df2, f1, f2, fn, safmin
    real(stnd), dimension(size(f))        :: ff, prob 
    real(stnd), dimension(:), allocatable :: y 
!
    integer(i4b) :: n, i, j
    integer      :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS
! __________
!
    real(stnd),  parameter, dimension(5) :: coef=(/ three, six, ten, c15, c21 /)
!
    character(len=*),  parameter :: name_proc='probf'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( f )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( ndf1<1_i4b .or. ndf2<1_i4b .or. any(f(:n)<zero) ) then
        call merror( name_proc//prob_error5 )
    end if
!
    safmin = lamch( 's' )
!
    where ( f(:n)<safmin )
        ff(:n) = one
    elsewhere
        ff(:n) = f(:n)
    end where
!
    if ( ndf1==1_i4b ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( y(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       USE STUDENT T-DISTRIBUTION FOR NDF1=1 .
!
        y(:n) = sqrt( ff(:n) )
!
        prob(:n) = two*probt_rv( y(:n), ndf2, true )
!
        if ( .not. upper ) prob(:n) = one - prob(:n)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( y )
!
    else if ( ndf2==1_i4b ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( y(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       USE STUDENT T-DISTRIBUTION FOR NDF2=1 .
!
        y(:n) = sqrt( one/ff(:n) )
!
        prob(:n) = two*probt_rv( y(:n), ndf1, true )
!
        if ( upper ) prob(:n) = one - prob(:n)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( y )
!
    else
!
!       OTHERWISE USE PEIZER-PRATT FORMULA (GAUSSIAN APPROXIMATION).
!
        df1 = real( ndf1, stnd )
        df2 = real( ndf2, stnd )
        f1 = ( df1 - one )/two
        f2 = ( df2 - one )/two
        fn = ( df1 + df2 - two )/two
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n>=omp_limit                        .and.      &
                      i>1_i4b                             .and.      &
                      j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,ffi,x,p,q,g,pg,qg)
!
        do i = 1_i4b, n
!
            ffi = ff(i)
!
            p = df2/( df2 + ffi*df1 )
            q = one - p
!
            x = f2/fn/p
            pg = one - x
!
            do j = 1_i4b, 2_i4b
!
                if ( abs(pg)>tenth ) then
                    g = ( one - x*x + two*x*log(merge(x,one,x/=zero)) )/( pg*pg ) 
                else
                    g = sum( geop(pg, pg, 5_i4b)/coef(1_i4b:5_i4b) )
                end if
!
                if ( j==2_i4b ) exit
!
                qg = g
                x = f1/fn/q
                pg = one - x
!
            end do
!
            pg = g
            g = q/df2 - p/df1 + ( q - half )/( df1 + df2 )
            x = c4_m2*g + ( f2 + sixth ) - p*( fn + third )
            g = ( one + q*qg + p*pg )/( fn + sixth )/p/q
            ff(i) = x*sqrt( g )
!
        end do
!
!$OMP END PARALLEL DO
!
        prob(:n) = probn_rv( ff(:n), upper )
!
    end if
!
!   ADJUST THE RESULT IF f=0 .
!
    if ( upper ) then
        where ( f(:n)<safmin )  prob(:n) = one
    else
        where ( f(:n)<safmin )  prob(:n) = zero
    end if
!
    probf_rv = prob(:n)
!
!
! END OF FUNCTION probf_rv
! ________________________
!
    end function probf_rv
!
! =========================================================================================
!
    function probf_rv2( f, ndf1, ndf2, upper )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom NDF1 and NDF2 
!   from F(i) to infinity if UPPER is true or from zero to F(i) if UPPER is false,
!   for i=1 to size(F). In other words, if:
!
!   - UPPER = true  :    PROBF( i ) = prob( U > F(i) )  ,
!   - UPPER = false :    PROBF( i ) = prob( U < F(i) )  ,
!   
!   , for U = Fisher(NDF1(i),NDF2(i)) and i=1 to size(F).
!
!
! Arguments
! _________
!
!   F      (INPUT) real(stnd), dimension(:)
!          On entry, upper or lower limits of integration. F(i) must be greater
!          or equal to zero for i=1 to size(F).
!
!   NDF1   (INPUT) integer(i4b), dimension(:)
!          On entry, first degree of freedom of the F distribution (numerator).
!          Any value in the array NDF1 must be greater or equal to 1.
!
!          The size of NDF1 must be  size(NDF1) = size(F) .
!
!   NDF2   (INPUT) integer(i4b), dimension(:)
!          On entry, second degree of freedom of the F distribution (denominator).
!          Any value in the array NDF2 must be greater or equal to 1.
!
!          The size of NDF2 must be  size(NDF2) = size(F) .
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  :  probability to the right of F is calculated.
!          - UPPER = false  : probability to the left of F is calculated.
!
!
! Further Details
! _______________
!
!   This function accepts only integer values of degree of freedom and uses a normal
!   approximation. This normal approximation is not accurate for small values of degrees
!   of freedom.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Peizer, D.B., and Pratt, J.W., 1968: A normal approximation for Binomial, F, Beta, and ...,
!                                (formula 2.24a), J.A.S.A., Vol. 63, 1457-1483
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, six, ten, half, third,   &
                                  sixth, tenth, c15, c21, c4_m2
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error5
    use Utilities,         only : merror, assert_eq, geop
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),       intent(in) :: f
    real(stnd), dimension(size(f))             :: probf_rv2
!
    integer(i4b), dimension(:), intent(in) :: ndf1, ndf2
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: fi, x, p, q, g, pg, qg,     &
                                             df1, df2, f1, f2, fn, safmin
    real(stnd), dimension(size(f))        :: prob 
!
    integer(i4b) :: n, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS
! __________
!
    real(stnd),  parameter, dimension(5) :: coef=(/ three, six, ten, c15, c21 /)
!
    character(len=*),  parameter :: name_proc='probf'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(f),i4b)     ,   &
                    int(size(ndf1),i4b)  ,   &
                    int(size(ndf2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(ndf1(:n)<1_i4b) .or. any(ndf2(:n)<1_i4b) .or. any(f(:n)<zero) ) then
        call merror( name_proc//prob_error5 )
    end if
!
    safmin = lamch( 's' )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,fi,x,p,q,g,pg,qg)
!
    do i = 1_i4b, n
!
        fi = f(i)
!
        if ( fi>=safmin ) then
!
            if ( ndf1(i)==1_i4b ) then
!
!               USE STUDENT T-DISTRIBUTION FOR NDF1=1 .
!
                x = sqrt( fi )
!
                prob(i) = two*probt_r( x, ndf2(i), true )
!
                if ( .not. upper ) prob(i) = one - prob(i)
!
            else if ( ndf2(i)==1_i4b ) then
!
!               USE STUDENT T-DISTRIBUTION FOR NDF2=1 .
!
                x = sqrt( one/fi )
!
                prob(i) = two*probt_r( x, ndf1(i), true )
!
                if ( upper ) prob(i) = one - prob(i)
!
            else
!
!               OTHERWISE USE PEIZER-PRATT FORMULA (GAUSSIAN APPROXIMATION).
!
                df1 = real( ndf1(i), stnd )
                df2 = real( ndf2(i), stnd )
                f1 = ( df1 - one )/two
                f2 = ( df2 - one )/two
                fn = ( df1 + df2 - two )/two
!
                p = df2/( df2 + fi*df1 )
                q = one - p
!
                x = f2/fn/p
                pg = one - x
!
                do j = 1_i4b, 2_i4b
!
                    if ( abs(pg)>tenth ) then
                        g = ( one - x*x + two*x*log(merge(x,one,x/=zero)) )/( pg*pg ) 
                    else
                        g = sum( geop(pg, pg, 5_i4b)/coef(1_i4b:5_i4b) )
                    end if
!
                    if ( j==2_i4b ) exit
!
                    qg = g
                    x = f1/fn/q
                    pg = one - x
!
                end do
!
                pg = g
                g = q/df2 - p/df1 + ( q - half )/( df1 + df2 )
                x = c4_m2*g + ( f2 + sixth ) - p*( fn + third )
                g = ( one + q*qg + p*pg )/( fn + sixth )/p/q
                p = x*sqrt( g )
!
                prob(i) = probn_r( p, upper )
!
            end if
!
        end if
!
    end do
!
!$OMP END PARALLEL DO
!
!   ADJUST THE RESULT IF f=0 .
!
    if ( upper ) then
        where ( f(:n)<safmin )  prob(:n) = one
    else
        where ( f(:n)<safmin )  prob(:n) = zero
    end if
!
    probf_rv2 = prob(:n)
!
!
! END OF FUNCTION probf_rv2
! _________________________
!
    end function probf_rv2
!
! =========================================================================================
!
    function probf_rm( f, ndf1, ndf2, upper )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom NDF1 and NDF2 
!   from F(i,j) to infinity if UPPER is true or from zero to F(i,j) if UPPER is false,
!   for i=1 to size(F,1) and j=1 to size(F,2). In other words, if:
!
!   - UPPER = true  :    PROBF( i,j ) = prob( U > F(i,j) )  ,
!   - UPPER = false :    PROBF( i,j ) = prob( U < F(i,j) )  ,
!   
!   , for U = Fisher(NDF1,NDF2) and i=1 to size(F,1) and j=1 to size(F,2).
!
!
! Arguments
! _________
!
!   F      (INPUT) real(stnd), dimension(:,:)
!          On entry, upper or lower limits of integration. F(i,j) must be greater
!          or equal to zero for i=1 to size(F,1) and j=1 to size(F,2).
!
!   NDF1   (INPUT) integer(i4b)
!          On entry, first degree of freedom of the F distribution (numerator).
!          NDF1 must be greater or equal to 1.
!
!   NDF2   (INPUT) integer(i4b)
!          On entry, second degree of freedom of the F distribution (denominator).
!          NDF2 must be greater or equal to 1.
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  :  probability to the right of F is calculated.
!          - UPPER = false  : probability to the left of F is calculated.
!
!
! Further Details
! _______________
!
!   This function accepts only integer values of degree of freedom and uses a normal
!   approximation. This normal approximation is not accurate for small values of degrees
!   of freedom.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Peizer, D.B., and Pratt, J.W., 1968: A normal approximation for Binomial, F, Beta, and ...,
!                                (formula 2.24a), J.A.S.A., Vol. 63, 1457-1483
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, six, ten, half, third, sixth, tenth,   &
                                  c15, c21, c4_m2
    use Num_Constants,     only : lamch
    use Char_Constants,    only : allocate_error, prob_error5
    use Utilities,         only : merror, geop
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                intent(in) :: f
    real(stnd), dimension(size(f,1),size(f,2))            :: probf_rm
!
    integer(i4b), intent(in) :: ndf1, ndf2
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: ffki, x, p, q, g, pg, qg,     &
                                                  df1, df2, f1, f2, fn, safmin
    real(stnd), dimension(size(f,1),size(f,2)) :: ff, prob 
    real(stnd), dimension(:,:),    allocatable :: y 
!
    integer(i4b) :: n, m, i, j, k
    integer      :: iok
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd),  parameter, dimension(5) :: coef=(/ three, six, ten, c15, c21 /)
!
    character(len=*),  parameter :: name_proc='probf'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( f, 1 )
    n = size( f, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( ndf1<1_i4b .or. ndf2<1_i4b .or. any(f(:m,:n)<zero) ) then
        call merror( name_proc//prob_error5 )
    end if
!
    safmin = lamch( 's' )
!
    where ( f(:m,:n)<safmin )
        ff(:m,:n) = one
    elsewhere
        ff(:m,:n) = f(:m,:n)
    end where
!
    if ( ndf1==1_i4b ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( y(m,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       USE STUDENT T-DISTRIBUTION FOR NDF1=1 .
!
        y(:m,:n) = sqrt( ff(:m,:n) )
        prob(:m,:n) = two*probt_rm( y(:m,:n), ndf2, true )
!
        if ( .not. upper ) prob(:m,:n) = one - prob(:m,:n)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( y )
!
    else if ( ndf2==1_i4b ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( y(m,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       USE STUDENT T-DISTRIBUTION FOR NDF2=1 .
!
        y(:m,:n) = sqrt( one/ff(:m,:n) )
!
        prob(:m,:n) = two*probt_rm( y(:m,:n), ndf1, true )
!
        if ( upper ) prob(:m,:n) = one - prob(:m,:n)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( y )
!
    else
!
!       OTHERWISE USE PEIZER-PRATT FORMULA (GAUSSIAN APPROXIMATION).
!
        df1 = real( ndf1, stnd )
        df2 = real( ndf2, stnd )
        f1 = ( df1 - one )/two
        f2 = ( df2 - one )/two
        fn = ( df1 + df2 - two )/two
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*m>=omp_limit                      .and.      &
                      n/j>=omp_limit2                     .and.      &
                      i>1_i4b                             .and.      &
                      j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,k,ffki,x,p,q,g,pg,qg)
!
        do i = 1_i4b, n
!
            do k = 1_i4b, m
!
                ffki = ff(k,i)
!
                p = df2/( df2 + ffki*df1 )
                q = one - p
!
                x = f2/fn/p
                pg = one - x
!
                do j = 1_i4b, 2_i4b
!
                    if ( abs(pg)>tenth ) then
                        g = ( one - x*x + two*x*log(merge(x,one,x/=zero)) )/( pg*pg ) 
                    else
                        g = sum( geop(pg, pg, 5_i4b)/coef(1_i4b:5_i4b) )
                    end if
!
                    if ( j==2_i4b ) exit
!
                    qg = g
                    x = f1/fn/q
                    pg = one - x
!
                end do
!
                pg = g
                g = q/df2 - p/df1 + ( q - half )/( df1 + df2 )
                x = c4_m2*g + ( f2 + sixth ) - p*( fn + third )
                g = ( one + q*qg + p*pg )/( fn + sixth )/p/q
                ff(k,i) = x*sqrt( g )
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
        prob(:m,:n) = probn_rm( ff(:m,:n), upper )
!
    end if
!
!   ADJUST THE RESULT IF f=0 .
!
    if ( upper ) then
        where ( f(:m,:n)<safmin )  prob(:m,:n) = one
    else
        where ( f(:m,:n)<safmin )  prob(:m,:n) = zero
    end if
!
    probf_rm = prob(:m,:n)
!
!
! END OF FUNCTION probf_rm
! ________________________
!
    end function probf_rm
!
! =========================================================================================
!
    function probf_rm2( f, ndf1, ndf2, upper )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom NDF1 and NDF2 
!   from F(i,j) to infinity if UPPER is true or from zero to F(i,j) if UPPER is false,
!   for i=1 to size(F,1) and j=1 to size(F,2). In other words, if:
!
!   - UPPER = true  :    PROBF( i,j ) = prob( U > F(i,j) )  ,
!   - UPPER = false :    PROBF( i,j ) = prob( U < F(i,j) )  ,
!   
!   , for U = Fisher(NDF1( i,j ),NDF2( i,j )) and i=1 to size(F,1) and j=1 to size(F,2).
!
!
! Arguments
! _________
!
!   F      (INPUT) real(stnd), dimension(:,:)
!          On entry, upper or lower limits of integration. F(i,j) must be greater
!          or equal to zero for i=1 to size(F,1) and j=1 to size(F,2).
!
!   NDF1   (INPUT) integer(i4b), dimension(:,:)
!          On entry, first degree of freedom of the F distribution (numerator).
!          Any value in the array NDF1 must be greater or equal to 1.
!
!          The shape of NDF1 must verify:
!
!          - size(NDF1,1) = size(F,1)
!          - size(NDF1,2) = size(F,2) .
!
!   NDF2   (INPUT) integer(i4b), dimension(:,:)
!          On entry, second degree of freedom of the F distribution (denominator).
!          Any value in the array NDF2 must be greater or equal to 1.
!
!          The shape of NDF2 must verify:
!
!          - size(NDF2,1) = size(F,1)
!          - size(NDF2,2) = size(F,2) .
!
!   UPPER  (INPUT) logical(lgl)
!          On entry, if:
!
!          - UPPER = true  :  probability to the right of F is calculated.
!          - UPPER = false  : probability to the left of F is calculated.
!
!
! Further Details
! _______________
!
!   This function accepts only integer values of degree of freedom and uses a normal
!   approximation. This normal approximation is not accurate for small values of degrees
!   of freedom.
!
!   This function is parallelized if OPENMP is used.
!
!   This function is adapted from
!
!   (1) Peizer, D.B., and Pratt, J.W., 1968: A normal approximation for Binomial, F, Beta, and ...,
!                                (formula 2.24a), J.A.S.A., Vol. 63, 1457-1483
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, three, six, ten, half, third,   &
                                  sixth, tenth, c15, c21, c4_m2
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error5
    use Utilities,         only : merror, assert_eq, geop
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),                intent(in) :: f
    real(stnd), dimension(size(f,1),size(f,2))            :: probf_rm2
!
    integer(i4b), dimension(:,:), intent(in) :: ndf1, ndf2
!
    logical(lgl), intent(in) :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: fki, x, p, q, g, pg, qg,     &
                                                  df1, df2, f1, f2, fn, safmin
    real(stnd), dimension(size(f,1),size(f,2)) :: prob
!
    integer(i4b) :: n, m, i, j, k
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    real(stnd),  parameter, dimension(5) :: coef=(/ three, six, ten, c15, c21 /)
!
    character(len=*),  parameter :: name_proc='probf'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m =  assert_eq( int(size(f,1),i4b)     ,   &
                    int(size(ndf1,1),i4b)  ,   &
                    int(size(ndf2,1),i4b)  ,   &
                    name_proc )
!
    n =  assert_eq( int(size(f,2),i4b)     ,   &
                    int(size(ndf1,2),i4b)  ,   &
                    int(size(ndf2,2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(ndf1(:m,:n)<1_i4b) .or. any(ndf2(:m,:n)<1_i4b) .or. any(f(:m,:n)<zero) ) then
        call merror( name_proc//prob_error5 )
    end if
!
    safmin = lamch( 's' )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC)    &
!$OMP            ,PRIVATE(i,j,k,fki,x,p,q,g,pg,qg)
!
    do i = 1_i4b, n
!
        do k = 1_i4b, m
!
            fki = f(k,i)
!
            if ( fki<safmin ) cycle
!
            if ( ndf1(k,i)==1_i4b ) then
!
!               USE STUDENT T-DISTRIBUTION FOR NDF1=1 .
!
                x = sqrt( fki )
                prob(k,i) = two*probt_r( x, ndf2(k,i), true )
                if ( .not. upper ) prob(k,i) = one - prob(k,i)
!
            else if ( ndf2(k,i)==1_i4b ) then
!
!               USE STUDENT T-DISTRIBUTION FOR NDF2=1 .
!
                x = sqrt( one/fki )
                prob(k,i) = two*probt_r( x, ndf1(k,i), true )
                if ( upper ) prob(k,i) = one - prob(k,i)
!
            else
!
!               OTHERWISE USE PEIZER-PRATT FORMULA (GAUSSIAN APPROXIMATION).
!
                df1 = real( ndf1(k,i), stnd )
                df2 = real( ndf2(k,i), stnd )
                f1 = ( df1 - one )/two
                f2 = ( df2 - one )/two
                fn = ( df1 + df2 - two )/two
!
                p = df2/( df2 + fki*df1 )
                q = one - p
!
                x = f2/fn/p
                pg = one - x
!
                do j = 1_i4b, 2_i4b
!
                    if ( abs(pg)>tenth ) then
                        g = ( one - x*x + two*x*log(merge(x,one,x/=zero)) )/( pg*pg ) 
                    else
                        g = sum( geop(pg, pg, 5_i4b)/coef(1_i4b:5_i4b) )
                    end if
!
                    if ( j==2_i4b ) exit
!
                    qg = g
                    x = f1/fn/q
                    pg = one - x
!
                end do
!
                pg = g
                g = q/df2 - p/df1 + ( q - half )/( df1 + df2 )
                x = c4_m2*g + ( f2 + sixth ) - p*( fn + third )
                g = ( one + q*qg + p*pg )/( fn + sixth )/p/q
                p = x*sqrt( g )
!
                prob(k,i) = probn_r( p, upper )
!
            end if
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
!   ADJUST THE RESULT IF f=0 .
!
    if ( upper ) then
        where ( f(:m,:n)<safmin )  prob(:m,:n) = one
    else
        where ( f(:m,:n)<safmin )  prob(:m,:n) = zero
    end if
!
    probf_rm2 = prob(:m,:n)
!
!
! END OF FUNCTION probf_rm2
! _________________________
!
    end function probf_rm2
!
! =========================================================================================
!
    function probf2_r( f, df1, df2, upper, beta, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom DF1 and DF2 
!   (integer or fractional degrees of freedom) from F to infinity if UPPER is true
!   or from zero to F if UPPER is false. In other words, if:
!
!   - UPPER = true  :    PROBF2 = prob( U > F )   ,
!   - UPPER = false :    PROBF2 = prob( U <= F )  ,
!   
!   , for U = Fisher(DF1,DF2).
!
!   For given arguments F (0<=F), DF1 (DF1>0), DF2 (DF2>0), PROBF2 returns the 
!   probability that a random variable from an F distribution having DF1 and 
!   DF2 degrees of freedom will be less than or equal to F (if UPPER is false) or
!   greater than F (if UPPER is true).
!
!
! Arguments
! _________
!
!   F       (INPUT) real(stnd)
!           On entry, upper or lower limit of integration. F must be greater
!           or equal to zero.
!
!   DF1     (INPUT) real(stnd)
!           On entry, first degree of freedom of the F distribution (numerator).
!           DF1 must be greater than zero.
!
!   DF2     (INPUT) real(stnd)
!           On entry, second degree of freedom of the F distribution (denominator).
!           DF2 must be greater than zero.
!
!   UPPER   (INPUT) logical(lgl)
!           On entry, if:
!
!           - UPPER = true  :  probability to the right of F is calculated.
!           - UPPER = false  : probability to the left of F is calculated.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(0.5 * DF1,0.5 * DF2).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result, when computing the
!           incomplete beta function. The "integrating" process for evaluating
!           the incomplete beta function is terminated when the
!           relative contribution to the integral is not greater than
!           the value of ACU. ACU  is a small strictly positive integer.
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process of the incomplete Beta integral did not converge to the desired accuracy
!           are set to -1. The algorithm fails to converge if the number of iterations
!           exceeds MAXITER.
!
!
! Further Details
! _______________
!
!   This function invoked the BETA distribution function (e.g. function PROBBETA)
!   for computing the probability associated with F and is much more accurate than PROBF
!   function.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error10, prob_error11
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)           :: f, df1, df2
    real(stnd),   intent(in), optional :: beta, acu
    real(stnd)                         :: probf2_r
!
    integer(i4b), intent(in), optional  :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(in), optional  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: x, a, b, safmin
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probf2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( df1<=zero .or. df2<=zero ) then
        call merror( name_proc//prob_error10 )
    end if
!
    if ( f<zero ) then
        call merror( name_proc//prob_error11 )
    end if
!
!   TRANSFORM VARIABLES FOR COMPATIBILITY WITH BETA PDF.
!
    if ( upper ) then
!
        x = df2/( df2 + df1*f )
        a = half*df2
        b = half*df1
!
    else
!
        safmin = sqrt( lamch( 's' ) )
!
        if ( f>=safmin ) then
            x = df1/( df1 + df2/f )
        else
            x = zero
        end if
!
        a = half*df1
        b = half*df2
!
    end if
!
    x = min( max( x, zero ), one )
!
!   INVOKE BETA PDF FOR PROBABILITY.
!
    probf2_r = probbeta_r( x, a, b, beta=beta, acu=acu, maxiter=maxiter, failure=failure )
!
!
! END OF FUNCTION probf2_r
! ________________________
!
    end function probf2_r
!
! =========================================================================================
!
    function probf2_rv( f, df1, df2, upper, beta, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom DF1 and DF2 
!   (integer or fractional degrees of freedom) from F(i) to infinity if UPPER is
!   true or from zero to F(i) if UPPER is false, for i=1 to size(F). In other words, if:
!
!   - UPPER = true  :    PROBF( i ) = prob( U > F(i) )  ,
!   - UPPER = false :    PROBF( i ) = prob( U < F(i) )  ,
!   
!   , for U = Fisher(DF1,DF2) and i=1 to size(F).
!
!   For given arguments F(:) (0<=F(:)), DF1 (DF1>0), DF2 (DF2>0), PROBF2 returns the 
!   probability that a random variable (vector) from an F distribution having DF1 and 
!   DF2 degrees of freedom will be less than or equal to F (if UPPER is false) or
!   greater than F (if UPPER is true).
!
!
! Arguments
! _________
!
!   F       (INPUT) real(stnd), dimension(:)
!           On entry, upper or lower limit of integration. F(i) must be greater
!           or equal to zero for i=1 to size(F).
!
!   DF1     (INPUT) real(stnd)
!           On entry, first degree of freedom of the F distribution (numerator).
!           DF1 must be greater than zero.
!
!   DF2     (INPUT) real(stnd)
!           On entry, second degree of freedom of the F distribution (denominator).
!           DF2 must be greater than zero.
!
!   UPPER   (INPUT) logical(lgl)
!           On entry, if:
!
!           - UPPER = true  :  probability to the right of F is calculated.
!           - UPPER = false  : probability to the left of F is calculated.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(0.5 * DF1,0.5 * DF2).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result, when computing the
!           incomplete beta function. The "integrating" process for evaluating
!           the incomplete beta function is terminated when the
!           relative contribution to the integral is not greater than
!           the value of ACU. ACU  is a small strictly positive integer.
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process of the incomplete Beta integral did not converge to the desired accuracy
!           are set to -1. The algorithm fails to converge if the number of iterations
!           exceeds MAXITER.
!
!
! Further Details
! _______________
!
!   This function invoked the BETA distribution function (e.g. function PROBBETA)
!   for computing the probability associated with F and is much more accurate than PROBF
!   function.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error10, prob_error11
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),   intent(in)           :: f
    real(stnd),                 intent(in)           :: df1, df2
    real(stnd),                 intent(in), optional :: beta, acu
    real(stnd), dimension(size(f))                   :: probf2_rv
!
    integer(i4b), intent(in), optional  :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(in), optional  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: a, b, safmin
    real(stnd), dimension(size(f)) :: x
!
    integer(i4b) :: n
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probf2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( f )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( df1<=zero .or. df2<=zero ) then
        call merror( name_proc//prob_error10 )
    end if
!
    if ( any( f(:n)<zero ) ) then
        call merror( name_proc//prob_error11 )
    end if
!
!   TRANSFORM VARIABLES FOR COMPATIBILITY WITH BETA PDF.
!
    if ( upper ) then
!
        x(:n) = df2/( df2 + df1*f(:n) )
        a     = half*df2
        b     = half*df1
!
    else
!
        safmin = sqrt( lamch( 's' ) )
!
        where ( f(:n)>=safmin )
            x(:n) = df1/( df1 + df2/f(:n) )
        elsewhere
            x(:n) = zero
        end where
!
        a     = half*df1
        b     = half*df2
!
    end if
!
    x(:n) = min( max( x(:n), zero ), one )
!
!   INVOKE BETA PDF FOR PROBABILITY.
!
    probf2_rv(:n) = probbeta_rv( x(:n), a, b, beta=beta, acu=acu,    &
                                 maxiter=maxiter, failure=failure )
!
!
! END OF FUNCTION probf2_rv
! _________________________
!
    end function probf2_rv
!
! =========================================================================================
!
    function probf2_rv2( f, df1, df2, upper, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the F distribution function with degrees of freedom DF1(i) and DF2(i) 
!   (integer or fractional degrees of freedom) from F(i) to infinity if UPPER is
!   true or from zero to F(i) if UPPER is false, for i=1 to size(F). In other words, if:
!
!   - UPPER = true  :    PROBF( i ) = prob( U > F(i) )  ,
!   - UPPER = false :    PROBF( i ) = prob( U < F(i) )  ,
!   
!   , for U = Fisher(DF1(i),DF2(i)) and i=1 to size(F).
!
!   For given arguments F(:) (0<=F(:)), DF1(:) (DF1(:)>0), DF2(:) (DF2(:)>0), PROBF2 returns
!   the probability that a random variable (vector) from an F distribution having DF1(:) and 
!   DF2(:) degrees of freedom will be less than or equal to F(:) (if UPPER is false) or
!   greater than F(:) (if UPPER is true).
!
!
! Arguments
! _________
!
!   F       (INPUT) real(stnd), dimension(:)
!           On entry, upper or lower limit of integration. F(i) must be greater
!           or equal to zero for i=1 to size(F).
!
!   DF1     (INPUT) real(stnd), dimension(:)
!           On entry, first degree of freedom of the F distribution (numerator).
!           Any value in the array DF1(:) must be greater than zero.
!
!           The size of DF1 must verify  size(DF1) = size(F) .
!
!   DF2     (INPUT) real(stnd), dimension(:)
!           On entry, second degree of freedom of the F distribution (denominator).
!           Any value in the array DF2(:) must be greater than zero.
!
!           The size of DF2 must verify  size(DF2) = size(F) .
!
!   UPPER   (INPUT) logical(lgl)
!           On entry, if:
!
!           - UPPER = true  :  probability to the right of F is calculated.
!           - UPPER = false  : probability to the left of F is calculated.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(0.5 * DF1,0.5 * DF2).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result, when computing the
!           incomplete beta function. The "integrating" process for evaluating
!           the incomplete beta function is terminated when the
!           relative contribution to the integral is not greater than
!           the value of ACU. ACU  is a small strictly positive integer.
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process of the incomplete Beta integral did not converge to the desired accuracy
!           are set to -1. The algorithm fails to converge if the number of iterations
!           exceeds MAXITER.
!
!
! Further Details
! _______________
!
!   This function invoked the BETA distribution function (e.g. function PROBBETA)
!   for computing the probability associated with F and is much more accurate than PROBF
!   function.
!
!   The function is parallelized if OPENMP is used.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error10, prob_error11
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),   intent(in)           :: f, df1, df2
    real(stnd),                 intent(in), optional :: acu
    real(stnd), dimension(size(f))                   :: probf2_rv2
!
    integer(i4b), intent(in), optional  :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(in), optional  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: x, a, b, df1i,df2i, safmin
    real(stnd), dimension(size(f)) :: prob
!
    integer(i4b) :: n, i
!
#ifdef _OPENMP
!
    integer(i4b) :: j
!
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probf2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(f),i4b)    ,   &
                    int(size(df1),i4b)  ,   &
                    int(size(df2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any( df1(:n)<zero ) .or. any( df2(:n)<zero ) ) then
        call merror( name_proc//prob_error10 )
    end if
!
    if ( any( f(:n)<zero ) ) then
        call merror( name_proc//prob_error11 )
    end if
!
    if ( .not.upper ) then
!
        safmin = sqrt( lamch( 's' ) )
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,df1i,df2i,x,a,b), SCHEDULE(STATIC)
!
    do i = 1_i4b, n
!
        x = f(i)
!
        df1i = df1(i)
        df2i = df2(i)
!
!       TRANSFORM VARIABLES FOR COMPATIBILITY WITH BETA PDF.
!
        if ( upper ) then
!
            x = df2i/( df2i + df1i*x )
            a = half*df2i
            b = half*df1i
!
        else
!
            if ( x>=safmin ) then
                x = df1i/( df1i + df2i/x )
            else
                x = zero
            end if
!
            a = half*df1i
            b = half*df2i
!
        end if
!
        x = min( max( x, zero ), one )
!
!       INVOKE BETA PDF FOR PROBABILITY.
!
        prob(i) = probbeta_r( x, a, b, acu=acu, maxiter=maxiter, failure=failure )
!
    end do
!
!$OMP END PARALLEL DO
!
    probf2_rv2(:n) = prob(:n)
!
!
! END OF FUNCTION probf2_rv2
! __________________________
!
    end function probf2_rv2
!
! =========================================================================================
!
    function probf2_rm( f, df1, df2, upper, beta, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the  F distribution function with degrees of freedom DF1 and DF2 
!   (integer or fractional degrees of freedom) from F(i,j) to infinity if UPPER is
!   true or from zero to F(i,j) if UPPER is false, for i=1 to size(F,1) and j=1 to size(F,2):
!
!   if UPPER = true  :    PROBF( i, j ) = prob( U > F(i,j) )  ,
!   if UPPER = false :    PROBF( i, j ) = prob( U < F(i,j) )  ,
!   
!   , for U = Fisher(DF1,DF2) and i=1 to size(F,1) and j=1 to size(F,2).
!
!   For given arguments F(:,:) (0<=F(:,:)), DF1 (DF1>0), DF2 (DF2>0), PROBF2 returns the 
!   probability that a random variable (matrix) from an F distribution having DF1 and 
!   DF2 degrees of freedom will be less than or equal to F (if UPPER is false) or
!   greater than F (if UPPER is true).
!
! Arguments
! _________
!
!   F       (INPUT) real(stnd), dimension(:,:)
!           On entry, upper or lower limit of integration. F(i,j) must be greater
!           or equal to zero for i=1 to size(F,1) and j=1 to size(F,2).
!
!   DF1     (INPUT) real(stnd)
!           On entry, first degree of freedom of the F distribution (numerator).
!           DF1 must be greater than zero.
!
!   DF2     (INPUT) real(stnd)
!           On entry, second degree of freedom of the F distribution (denominator).
!           DF2 must be greater than zero.
!
!   UPPER   (INPUT) logical(lgl)
!           On entry, if:
!
!           - UPPER = true  :  probability to the right of F is calculated.
!           - UPPER = false  : probability to the left of F is calculated.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(0.5 * DF1,0.5 * DF2).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result, when computing the
!           incomplete beta function. The "integrating" process for evaluating
!           the incomplete beta function is terminated when the
!           relative contribution to the integral is not greater than
!           the value of ACU. ACU  is a small strictly positive integer.
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process of the incomplete Beta integral did not converge to the desired accuracy
!           are set to -1. The algorithm fails to converge if the number of iterations
!           exceeds MAXITER.
!
!
! Further Details
! _______________
!
!   This function invoked the BETA distribution function (e.g. function PROBBETA)
!   for computing the probability associated with F and is much more accurate than PROBF
!   function.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error10, prob_error11
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: f
    real(stnd),                 intent(in)           :: df1, df2
    real(stnd),                 intent(in), optional :: beta, acu
    real(stnd), dimension(size(f,1),size(f,2))       :: probf2_rm
!
    integer(i4b), intent(in), optional  :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(in), optional  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: a, b, safmin
    real(stnd), dimension(size(f,1),size(f,2)) :: x
!
    integer(i4b) :: n, p
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probf2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( f, 1 )
    p = size( f, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. p<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( df1<=zero .or. df2<=zero ) then
        call merror( name_proc//prob_error10 )
    end if
!
    if ( any( f(:n,:p)<zero ) ) then
        call merror( name_proc//prob_error11 )
    end if
!
!   TRANSFORM VARIABLES FOR COMPATIBILITY WITH BETA PDF.
!
    if ( upper ) then
!
        x(:n,:p) = df2/( df2 + df1*f(:n,:p) )
        a        = half*df2
        b        = half*df1
!
    else
!
        safmin = sqrt( lamch( 's' ) )
!
        where ( f(:n,:p)>=safmin )
            x(:n,:p) = df1/( df1 + df2/f(:n,:p) )
        elsewhere
            x(:n,:p) = zero
        end where
!
        a        = half*df1
        b        = half*df2
!
    end if
!
    x(:n,:p) = min( max( x(:n,:p), zero ), one )
!
!   INVOKE BETA PDF FOR PROBABILITY.
!
    probf2_rm(:n,:p) = probbeta_rm( x(:n,:p), a, b, beta=beta,          &
                                    acu=acu, maxiter=maxiter, failure=failure )
!
!
! END OF FUNCTION probf2_rm
! _________________________
!
    end function probf2_rm
!
! =========================================================================================
!
    function probf2_rm2( f, df1, df2, upper, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the F distribution function with degrees of freedom DF1(i,j) and DF2(i,j) 
!   (integer or fractional degrees of freedom) from F(i,j) to infinity if UPPER is
!   true or from zero to F(i,j) if UPPER is false, for i=1 to size(F,1) and j=1 to size(F,2).
!   In other words, if:
!
!   - UPPER = true  :    PROBF( i, j ) = prob( U > F(i,j) )  ,
!   - UPPER = false :    PROBF( i, j ) = prob( U < F(i,j) )  ,
!   
!   , for U = Fisher(DF1(i,j),DF2(i,j)) and i=1 to size(F,1) and j=1 to size(F,2).
!
!   For given arguments F(:,:) (0<=F(:,:)), DF1(:,:) (DF1(:,:)>0), DF2(:,:) (DF2(:,:)>0), 
!   PROBF2 returns the probability that a random variable (matrix) from an F distribution 
!   having DF1(:,:) and DF2(:,:) degrees of freedom will be less than or equal to F(:,:) (if
!   UPPER is false) or greater than F (if UPPER is true).
!
!
! Arguments
! _________
!
!   F       (INPUT) real(stnd), dimension(:,:)
!           On entry, upper or lower limit of integration. F(i,j) must be greater
!           or equal to zero for i=1 to size(F,1) and j=1 to size(F,2).
!
!   DF1     (INPUT) real(stnd), dimension(:,:)
!           On entry, first degree of freedom of the F distribution (numerator).
!           Any value in the array DF1(:,:) must be greater than zero.
!
!           The shape of DF1 must verify:
!
!           - size(DF1,1) = size(F,1)
!           - size(DF1,2) = size(F,2) .
!
!   DF2     (INPUT) real(stnd), dimension(:,:)
!           On entry, second degree of freedom of the F distribution (denominator).
!           Any value in the array DF2(:,:) must be greater than zero.
!
!           The shape of DF2 must verify:
!
!           - size(DF2,1) = size(F,1)
!           - size(DF2,2) = size(F,2) .
!
!   UPPER   (INPUT) logical(lgl)
!           On entry, if:
!
!           - UPPER = true  :  probability to the right of F is calculated.
!           - UPPER = false  : probability to the left of F is calculated.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result, when computing the
!           incomplete beta function. The "integrating" process for evaluating
!           the incomplete beta function is terminated when the
!           relative contribution to the integral is not greater than
!           the value of ACU. ACU  is a small strictly positive integer.
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process of the incomplete Beta integral did not converge to the desired accuracy
!           are set to -1. The algorithm fails to converge if the number of iterations
!           exceeds MAXITER.
!
!
! Further Details
! _______________
!
!   This function invoked the BETA distribution function (e.g. function PROBBETA)
!   for computing the probability associated with F and is much more accurate than PROBF
!   function.
!
!   The function is parallelized if OPENMP is used.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, one
    use Num_Constants,     only : lamch
    use Char_Constants,    only : prob_error10, prob_error11
    use Utilities,         only : merror, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(in)           :: f, df1, df2
    real(stnd),                 intent(in), optional :: acu
    real(stnd), dimension(size(f,1),size(f,2))       :: probf2_rm2
!
    integer(i4b), intent(in), optional  :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(in), optional  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: x, a, b, df1ji, df2ji, safmin
    real(stnd), dimension(size(f,1),size(f,2)) :: prob
!
    integer(i4b) :: n, p, i, j
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probf2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    n =  assert_eq( int(size(f,1),i4b)    ,   &
                    int(size(df1,1),i4b)  ,   &
                    int(size(df2,1),i4b)  ,   &
                    name_proc )
!
    p =  assert_eq( int(size(f,2),i4b)    ,   &
                    int(size(df1,2),i4b)  ,   &
                    int(size(df2,2),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. p<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any( df1(:n,:p)<zero ) .or. any( df2(:n,:p)<zero ) ) then
        call merror( name_proc//prob_error10 )
    end if
!
    if ( any( f(:n,:p)<zero ) ) then
        call merror( name_proc//prob_error11 )
    end if
!
    if ( .not.upper ) then
!
        safmin = sqrt( lamch( 's' ) )
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*p>=omp_limit                      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), PRIVATE(i,j,df1ji,df2ji,x,a,b), SCHEDULE(STATIC)
!
    do i = 1_i4b, p
!
        do j = 1_i4b, n
!
            x = f(j,i)
!
            df1ji = df1(j,i)
            df2ji = df2(j,i)
!
!           TRANSFORM VARIABLES FOR COMPATIBILITY WITH BETA PDF.
!
            if ( upper ) then
!
                x = df2ji/( df2ji + df1ji*x )
                a = half*df2ji
                b = half*df1ji
!
            else
!
                if ( x>=safmin ) then
                    x = df1ji/( df1ji + df2ji/x )
                else
                    x = zero
                end if
!
                a = half*df1ji
                b = half*df2ji
!
            end if
!
            x = min( max( x, zero ), one )
!
!           INVOKE BETA PDF FOR PROBABILITY.
!
            prob(j,i) = probbeta_r( x, a, b, acu=acu, maxiter=maxiter, failure=failure )
!
        end do
!
    end do
!
!$OMP END PARALLEL DO
!
    probf2_rm2(:n,:p) = prob(:n,:p)
!
!
! END OF FUNCTION probf2_rm2
! __________________________
!
    end function probf2_rm2
!
!
! =========================================================================================
!                            INVERSE  F  DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function pinvf2_r( p, df1, df2, beta, acu, maxiter )
!
! Purpose
! _______
!
!   Evaluates the  inverse F probability distribution function with degrees of freedom
!   DF1 and DF2 (integer or fractional degrees of freedom).
!
!   For given arguments P (0<=P<=1), DF1 (DF1>0.2), DF2 (DF2>0.2), PINVF2 returns the 
!   value F such that the probability that a random variable distributed as F(DF1,DF2) is
!   less than or equal to F is P.
!
!
! Arguments
! _________
!
!   P       (INPUT) real(stnd)
!           On entry, input probability in the inclusive range (0,1).
!
!   DF1     (INPUT) real(stnd)
!           On entry, first degree of freedom of the F distribution (numerator).
!           DF1 must be greater than 0.2.
!
!   DF2     (INPUT) real(stnd)
!           On entry, second degree of freedom of the F distribution (denominator).
!           DF2 must be greater than 0.2.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(0.5 * DF1,0.5 * DF2).
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result, when computing the
!           incomplete beta function. The "integrating" process for evaluating
!           the incomplete beta function is terminated when the
!           relative contribution to the integral is not greater than
!           the value of ACU. ACU  is a small strictly positive integer.
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!          
!           See the description of the PROBBETA function for more details
!           on this argument.
!
!
! Further Details
! _______________
!
!   This function invoked the inverse BETA distribution function (e.g. function PINVBETA)
!   for computing the value F associated with the probability P.
!
!   This function is not very accurate for small values of DF1 and/or DF2 (e.g. less than 1).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, fifth, half, one
    use Num_Constants,     only : macheps, machhuge, lamch
    use Char_Constants,    only : prob_error1, prob_error23
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)           :: p, df1, df2
    real(stnd),   intent(in), optional :: beta, acu
    real(stnd)                         :: pinvf2_r
!
    integer(i4b), intent(in), optional :: maxiter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: p2, a, b, f, x, safmin
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='pinvf2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( df1<=fifth .or. df2<=fifth ) then
        call merror( name_proc//prob_error23 )
    end if
!
    if ( p<zero .or. p>one ) then
        call merror( name_proc//prob_error1 )
    end if
!
!   REPARAMETERIZE FOR BETA USAGE.
!
    a = half*df1
    b = half*df2
!
    if ( p<=half ) then
!
        x = pinvbeta_r( p, a, b, beta=beta, acu=acu, maxiter=maxiter )
!
!       MODIFY BETA OUTPUT FOR f FROM F(df1,df2) DISTRIBUTION.
!
        if ( x+macheps<one ) then
            f = df2*x/( df1*(one-x) )
        else
            f = machhuge
        end if
!
    else
!
        p2 = one - p
        x = pinvbeta_r( p2, b, a, beta=beta, acu=acu, maxiter=maxiter )
!
!       MODIFY BETA OUTPUT FOR f FROM F(df1,df2) DISTRIBUTION.
!
        safmin = lamch( 's' )
!
        if ( x>=safmin ) then
            f = ( one/x -one )*df2/df1  
        else
            f = machhuge
        end if
!
    end if
!
    pinvf2_r = f
!
!
! END OF FUNCTION pinvf2_r
! ________________________
!
    end function pinvf2_r
!
!
! =========================================================================================
!                            BINOMIAL PROBABILITY DISTRIBUTION FUNCTION
! =========================================================================================
!
!
    function probbinom_r( prob, n, k, upper, beta, acu, maxiter, failure )
!
! Purpose
! _______
!
!   Evaluates the cumulative binomial probability distribution function for a positive real 
!   argument PROB between 0 and 1, a strictly positive integer N and a positive integer K 
!   less than or equal to N.
!   
!   PROBBINOM computes the probability that an event occurring with probability PROB per trial,
!   will occur K or more times in N independent trials if UPPER is true, or will occur K or less
!   times in N independent trials if UPPER is false.
!
!   This probability is estimated with the help of the incomplete beta fonction, as computed by
!   PROBBETA function, and the optional arguments BETA, ACU, MAXITER and FAILURE are passed
!   directly to the PROBBETA function if these arguments are present.
!
!
! Arguments
! _________
!
!   PROB    (INPUT) real(stnd)
!           On entry, a positive real argument PROB which is the probability of success on each trial
!           for the given binomial probability distribution.
!           PROB must be greater or equal to zero and less than or equal to 1.
!
!   N       (INPUT) integer(i4b)
!           On entry, a strictly positive integer argument which is the total number of Bernoulli
!           trials for the given binomial probability distribution.
!           N must be greater than zero.
!
!   K       (INPUT) integer(i4b)
!           On entry, a positive integer argument which is the minimal (if UPPER is true) or maximum
!           (if UPPER is false) number of success for the N Bernoulli trials, for which we want to compute
!           the cumulative probability following the binomial probability distribution specified by PROB.
!           K must be greater or equal to zero and less or equal to N.
!
!   UPPER   (INPUT) logical(lgl)
!           On entry, if:
!
!           - UPPER = true  : the probability that the number of success is greater than or equal
!             to K in N trials is computed.
!           - UPPER = false : the probability that the number of success is less than or equal
!             to K in N trials is computed.
!
!   BETA    (INPUT, OPTIONAL) real(stnd)
!           On entry, the logarithm of the complete beta function BETA(K,N-K+1) if UPPER is true,
!           or BETA(N-K,K+1) if UPPER is false.
!           If BETA is not given, the logarithm of the beta function is
!           computed with the help of function LNGAMMA.
!
!   ACU     (INPUT, OPTIONAL) real(stnd)
!           On entry, the desired accuracy of the result when evaluating the incomplete beta fonction.
!           The "integrating" process is terminated when both the absolute and relative contributions
!           to the integral is not greater than the value of ACU.
!
!           ACU  is a small strictly positive integer. If the number of decimal digits'
!           accuracy required is r, ACU should be set to 10**-(r+1).
!          
!           The default value for ACU is epsilon( ACU ).
!
!   MAXITER (INPUT, OPTIONAL) integer(i4b)
!           On entry, MAXITER controls the maximum number of iterations when evaluating
!           the power series representation of the incomplete Beta integral.
!           The default value is 2000.
!
!   FAILURE (INPUT, OPTIONAL) logical(lgl)
!           On entry, if FAILURE is set to true, the values for which the "integrating"
!           process did not converge to the desired accuracy are set to -1. The algorithm
!           fails to converge if the number of iterations exceeds MAXITER.
!           The default value is false.
!
!
! Further Details
! _______________
!
!   The cumulative binomial probability is computed with the help of the incomplete Beta
!   function PROBBETA as follow:
!
!   If UPPER is true, the probability of an event will occur K or more times in N independant trials,
!   if its probability per trial is PROB, is computed as PROBBETA( PROB, K, N-K+1).
!
!   If UPPER is false, the probability of an event will occur K or less times in N independant trials,
!   if its probability per trial is PROB, is computed as PROBBETA( 1-PROB, N-K, K+1).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Char_Constants,    only : prob_error1, prob_error20, prob_error21
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)           :: prob
    real(stnd), intent(in), optional :: beta, acu
    real(stnd)                       :: probbinom_r
!
    integer(i4b), intent(in), optional :: n, k
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)           :: upper
    logical(lgl), intent(in), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: a, b, tmp
!   
!   
! PARAMETERS 
! __________
!
    character(len=*),  parameter :: name_proc='probbinom'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( prob<zero .or. prob>one )  then
        call merror( name_proc//prob_error1 )
    end if
!
    if ( n<=0_i4b )  then
        call merror( name_proc//prob_error20 )
    end if
!
    if ( k<0_i4b .or. k>n )  then
        call merror( name_proc//prob_error21 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( prob==zero .or. prob==one ) then
!
        if ( upper ) then
            probbinom_r = prob
        else
            probbinom_r = one - prob
        end if
!
        return
!
    end if
!
    if ( ( k==0_i4b .and. upper ) .or.  ( k==n .and. .not.upper ) ) then
!
        probbinom_r = one
!
        return
!
    end if
!
!   COMPUTE THE PROBABILITY WITH THE INCOMPLETE BETA FONCTION.
!
    if ( upper ) then
!
        a = real( k, stnd )
        b = real( n - k + 1_i4b, stnd )
        tmp = prob
!
    else
!
        a = real( n - k, stnd )
        b = real( k + 1_i4b, stnd )
        tmp = one - prob
!
    end if
!
    probbinom_r = probbeta_r( tmp, a, b, beta=beta, acu=acu, maxiter=maxiter, failure=failure )
!
!
! END OF FUNCTION probbinom_r
! ___________________________
!
    end function probbinom_r
!
!
! =========================================================================================
!                            PROBABILITY INTEGRAL OF THE NORMAL RANGE
! =========================================================================================
!
!
    function rangen_r( x, n )
!
! Purpose
! _______
!
!   Evaluates the probability of the normal range given X, the
!   upper limit of integration, and N, the sample size.
!
!   In other words, this function evaluates the probability that
!   the standardized difference between the maximum and the minimum
!   on a sample will be less than X for a normal (e.g. gaussian) sample
!   of size N.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd)
!          On entry, upper limit of integration.
!          X must be greater than zero.
!
!   N      (INPUT) integer(i4b)
!          On entry, the sample size.
!          N must be greater than 1.
!
!
! Further Details
! _______________
!
!   This function is adapted from
!
!   (1)   Barnard, J., 1978: Algorithm AS126: Probability Integral of the normal range,
!                            Applied Statistics, Vol. 27, No. 2, pp.197-198
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, two, eight
    use Logical_Constants, only : false
    use Char_Constants,    only : prob_error19
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: x
    real(stnd)             :: rangen_r
!
    integer(i4b), intent(in) :: n
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: a, b, c, xl, y 
!
    integer(i4b) :: i
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), dimension(8), parameter :: g = (/    &
                0.4947004675_stnd, 0.4722875115_stnd, 0.4328156012_stnd, 0.3777022042_stnd,  &
                0.3089381222_stnd, 0.2290083888_stnd, 0.1408017754_stnd, 0.04750625492_stnd  /)
!
    real(stnd), dimension(8), parameter :: h = (/    &
                0.01357622971_stnd, 0.03112676197_stnd, 0.04757925584_stnd, 0.06231448563_stnd,  &
                0.07479799441_stnd, 0.08457825969_stnd, 0.09130170752_stnd, 0.09472530523_stnd   /)
!
    character(len=*),  parameter :: name_proc='rangen'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( x<=zero .or. n<=1_i4b ) then
        call merror( name_proc//prob_error19 )
    end if
!
    xl = half*x
    a  = half*( eight + xl )
    b  = eight - xl
    y  = zero
!
    do  i = 1_i4b, 8_i4b
!
        c = b * g(i)
        y = y + h(i) * (risf_r(a+c) + risf_r(a-c) )
!
    end do
!
    rangen_r =  ( two*(probn_r (xl, false )-half) )**n + two*b*y*real( n, stnd )
!
!----------------------------------------------------------------------
                         contains
!----------------------------------------------------------------------
!
        function risf_r( z ) result( fn_val )
!
        real(stnd), intent(in)    :: z
        real(stnd)                :: fn_val
!
        fn_val = 0.3989422804_stnd*exp(-half*z*z)*( probn_r( z, false ) -    &
                 probn_r( z-x, false ) )**(n-1_i4b)
!
        end function risf_r
!
!----------------------------------------------------------------------
!
!
! END OF FUNCTION rangen_r
! ________________________
!
    end function rangen_r
!
! =========================================================================================
!
    function rangen_rv( x, n )
!
! Purpose
! _______
!
!   Evaluates probabilities of the normal range given X(:), the
!   upper limits of integration, and N, the sample size.
!
!   In other words, this function evaluates the probabilities that
!   the standardized difference between the maximum and the minimum
!   on a sample will be less than X(:) for a normal
!   (e.g. gaussian) sample of size N.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, upper limits of integration.
!          All elements in X(:) must be greater than zero.
!
!   N      (INPUT) integer(i4b)
!          On entry, the sample size.
!          N must be greater than 1.
!
!
! Further Details
! _______________
!
!   This function is adapted from
!
!   (1)   Barnard, J., 1978: Algorithm AS126: Probability Integral of the normal range,
!                            Applied Statistics, Vol. 27, No. 2, pp.197-198
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, half, two, eight
    use Logical_Constants, only : false
    use Char_Constants,    only : prob_error19
    use Utilities,         only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),     intent(in) :: x
    real(stnd), dimension(size(x))           :: rangen_rv
!
    integer(i4b), intent(in) :: n
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(x)) :: a, b, c, xl, y 
!
    integer(i4b) :: m, i
!   
!   
! PARAMETERS 
! __________
!
    real(stnd), dimension(8), parameter :: g = (/    &
                0.4947004675_stnd, 0.4722875115_stnd, 0.4328156012_stnd, 0.3777022042_stnd,  &
                0.3089381222_stnd, 0.2290083888_stnd, 0.1408017754_stnd, 0.04750625492_stnd  /)
!
    real(stnd), dimension(8), parameter :: h = (/    &
                0.01357622971_stnd, 0.03112676197_stnd, 0.04757925584_stnd, 0.06231448563_stnd,  &
                0.07479799441_stnd, 0.08457825969_stnd, 0.09130170752_stnd, 0.09472530523_stnd   /)
!
    character(len=*),  parameter :: name_proc='rangen'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( x )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b ) then
        return
    end if
!
!   TEST THE ARGUMENTS.
!
    if ( any(x(:m)<=zero) .or. n<=1_i4b ) then
        call merror( name_proc//prob_error19 )
    end if
!
    xl(:m) = half*x(:m)
    a(:m)  = half*( eight + xl(:m) )
    b(:m)  = eight - xl(:m)
    y(:m)  = zero
!
    do  i = 1_i4b, 8_i4b
!
        c(:m) = b(:m) * g(i)
        y(:m) = y(:m) + h(i) * (risf_rv(a(:m)+c(:m)) + risf_rv(a(:m)-c(:m)) )
!
    end do
!
    rangen_rv(:m) =  ( two*(probn_rv (xl(:m), false )-half) )**n +    &
                     two*b(:m)*y(:m)*real( n, stnd )
!
!----------------------------------------------------------------------
                         contains
!----------------------------------------------------------------------
!
        function risf_rv( z ) result( fn_val )
!
        real(stnd), dimension(:), intent(in) :: z
        real(stnd), dimension(size(z)) :: fn_val
!
        fn_val(:m) = 0.3989422804_stnd*exp(-half*z(:m)*z(:m))*( probn_rv( z(:m), false ) -    &
                     probn_rv( z(:m)-x(:m), false ) )**(n-1_i4b)
!
        end function risf_rv
!
!----------------------------------------------------------------------
!
!
! END OF FUNCTION rangen_rv
! _________________________
!
    end function rangen_rv
!
! =========================================================================================
!
!
! *******************************
! END OF MODULE Prob_Procedures *
! *******************************
!    
end module Prob_Procedures
