module Stat_Procedures
! ##############################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ##############################################################################
!                                                                              *
! ******************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR UNIVARIATE STATISTICAL        *
! COMPUTATIONS                                                                 *
!                                                                              *
! LATEST REVISION : 21/05/2018                                                 *
!                                                                              *
! ##############################################################################
!                                                                              *
! ******************************************************************************
!                                                                              *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                              *
!                                                                              *
!  _OPENMP  FOR ACTIVATING OPENMP PARALLELIZATION                              *
!  _BLAS    FOR USING BLAS WHEN POSSIBLE                                       *
!                                                                              *
! ******************************************************************************
!
!
! USED MODULES
! ============
!   
    use Select_Parameters,  only : lgl, i4b, stnd
#ifdef _OPENMP
    use Select_Parameters,  only : omp_limit
    use omp_lib,            only : omp_get_num_procs, omp_get_max_threads, omp_in_parallel
#endif
#ifdef _BLAS
    use BLAS_interfaces,    only : axpy
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!
!   use Logical_Constants
!   use Reals_Constants
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use Prob_Procedures
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   comp_unistat,   comp_unistat_miss,     &
                comp_mvs,       comp_mvs_miss,         &
                update_mvs,                            &
                comp_mvs_grp,   comp_mvs_grp_miss,     &
                update_mvs_grp, update_mvs_grp_miss,   &
                comp_anoma,     comp_anoma_miss,       &
                comp_anoma_grp, comp_anoma_grp_miss,   &
                comp_composite, comp_composite_miss,   &
                valmed
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface comp_unistat
        module procedure    comp_unistat_rv,       comp_unistat_rm,      &
                            comp_unistat_rt,       comp_unistat_rf,      &
                            comp_unistat_miss_rv,  comp_unistat_miss_rm, &
                            comp_unistat_miss_rt,  comp_unistat_miss_rf
    end interface
!
    interface comp_unistat_miss
        module procedure    comp_unistat_miss_rv,  comp_unistat_miss_rm, &
                            comp_unistat_miss_rt,  comp_unistat_miss_rf
    end interface
!
    interface comp_mvs
        module procedure    comp_mvs_rv,      comp_mvs_rm,           &   
                            comp_mvs_rt,      comp_mvs_rf,           &
                            comp_mvs_miss_rv, comp_mvs_miss_rm,      &
                            comp_mvs_miss_rt, comp_mvs_miss_rf
    end interface
!
    interface comp_mvs_miss
        module procedure    comp_mvs_miss_rv, comp_mvs_miss_rm,      &
                            comp_mvs_miss_rt, comp_mvs_miss_rf
    end interface
!
    interface update_mvs
        module procedure    update_mvs_rv, update_mvs_rm,            &
                            update_mvs_rt, update_mvs_rf
    end interface
!
    interface comp_mvs_grp
        module procedure    comp_mvs_grp_rv,      comp_mvs_grp_rm,       &     
                            comp_mvs_grp_rt,      comp_mvs_grp_rf,       &
                            comp_mvs_grp_miss_rv, comp_mvs_grp_miss_rm,  &
                            comp_mvs_grp_miss_rt, comp_mvs_grp_miss_rf
    end interface
!
    interface comp_mvs_grp_miss
        module procedure    comp_mvs_grp_miss_rv, comp_mvs_grp_miss_rm,  &
                            comp_mvs_grp_miss_rt, comp_mvs_grp_miss_rf
    end interface
!
    interface update_mvs_grp
        module procedure    update_mvs_grp_rv, update_mvs_grp_rm,           &
                            update_mvs_grp_rt, update_mvs_grp_rf,           &
                            update_mvs_grp_miss_rm, update_mvs_grp_miss_rt, &
                            update_mvs_grp_miss_rf
    end interface
!
    interface update_mvs_grp_miss
        module procedure    update_mvs_grp_rv,      update_mvs_grp_miss_rm, &
                            update_mvs_grp_miss_rt, update_mvs_grp_miss_rf
    end interface
!
    interface comp_anoma
        module procedure    comp_anoma_rv,      comp_anoma_rm,           &   
                            comp_anoma_rt
    end interface
!
    interface comp_anoma_miss
        module procedure    comp_anoma_miss_rv,  comp_anoma_miss_rm,     &   
                            comp_anoma_miss_rt
    end interface
!
    interface comp_anoma_grp
        module procedure    comp_anoma_grp_rv,  comp_anoma_grp_rm,       &   
                            comp_anoma_grp_rt
    end interface
!
    interface comp_anoma_grp_miss
        module procedure    comp_anoma_grp_miss_rv,  comp_anoma_grp_miss_rm,  &   
                            comp_anoma_grp_miss_rt
    end interface
!
    interface comp_composite
        module procedure    comp_composite_rv,      comp_composite_rm,      comp_composite_rt,      &
                            comp_composite_miss_rv, comp_composite_miss_rm, comp_composite_miss_rt
    end interface
!
    interface comp_composite_miss
        module procedure    comp_composite_miss_rv, comp_composite_miss_rm, comp_composite_miss_rt
    end interface
!
    interface valmed
        module procedure    valmed_rv, valmed_rm
    end interface
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                               UNIVARIATE STATISTICS
! =========================================================================================
!
!
    subroutine comp_unistat_rv( x, first, last, xstat, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT computes estimates of univariate statistics from a data vector.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, input subvector containing size(X) observations
!          from the vector of data for which basic univariate statistics
!          are desired. If all the data are available at once, X can be
!          the full data vector.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subvector is the first subvector 
!            of the data vector.
!          - FIRST = false the current subvector is not the first subvector
!            of the data vector.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subvector is the last subvector 
!            of the data vector.
!          - LAST = false the current subvector is not the last subvector
!            of the data vector.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(7)
!          On entry, after the first call to COMP_UNISTAT (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT. XSTAT should not be changed between calls to
!          COMP_UNISTAT.
!
!          On exit, when LAST=true, XSTAT contains the following statistics :
!
!          - XSTAT(1) contains the mean value of the data vector.
!          - XSTAT(2) contains the variance of the data vector.
!          - XSTAT(3) contains the standard deviation of the data vector.
!          - XSTAT(4) contains the coefficient of skewness of the data vector.
!          - XSTAT(5) contains the coefficient of kurtosis of the data vector.
!          - XSTAT(6) contains the minimum of the data vector.
!          - XSTAT(7) contains the maximum of the data vector.
!
!          The size of XSTAT must verify size(XSTAT) = 7.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data vector. XNOBS needs to be specified only on the
!          last call to COMP_UNISTAT (e.g. when LAST=true).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT (e.g. when LAST=true).
!          
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than four valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : nan, machhuge
    use Utilities,         only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:)           :: x
    real(stnd), intent(inout), dimension(:)           :: xstat
!
    integer(i4b), intent(out),              optional  :: xnobs
!
    logical(lgl), intent(in)                          :: first, last
    logical(lgl), intent(in),               optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cnt, cntm1, cnt2, cnt3, di, di2
!
    integer(i4b) :: i, n
!
    logical(lgl) :: comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   TEST THE ARGUMENTS.
!
    call assert( logical(size(xstat)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xstat .
!
        xstat(1_i4b:5_i4b) = zero       
        xstat(6_i4b)       = machhuge       
        xstat(7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEAN, CORRECTED SUM OF SQUARES, CORRECTED SUM OF CUBES,
!       MINIMUM AND MAXIMUM, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt = cntm1 + one
        cnt2 = cnt*cnt
        cnt3 = cnt*cnt2
!
        di = x(i) - xstat(1_i4b)
        di2 = di*di
        xstat(1_i4b) = xstat(1_i4b) + di/cnt
        xstat(2_i4b) = xstat(2_i4b) + di2*(cntm1/cnt)
        xstat(4_i4b) = xstat(4_i4b) + di*( di2*((cnt2-one)/cnt2) - (three/cnt)*xstat(2_i4b) )
        xstat(5_i4b) = xstat(5_i4b) + di*( di*( di2*((cnt3-one)/cnt3) - (six/cnt2)*xstat(2_i4b) )     &
                                           - (four/cnt)*xstat(4_i4b) )
!
        xstat(6_i4b) = min( x(i), xstat(6_i4b) )
        xstat(7_i4b) = max( x(i), xstat(7_i4b) )
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
!
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        di = nan()
!
        if ( cnt>one ) then
!
            if ( xstat(2_i4b)>zero ) then
!
                cntm1 = cnt - one
!
!               COMPUTE VARIANCE AND STANDARD-DEVIATION.
!
                xstat(2_i4b) = xstat(2_i4b)/cntm1
                xstat(3_i4b) = sqrt( xstat(2_i4b) )
!
                if ( comp_nobias ) then
!
!                   COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                    if ( cnt>two ) then
!
                        cnt2 = max( cnt - two, one )
                        di2  = cnt/(cntm1*cnt2)
!
                        xstat(4_i4b) = di2*(xstat(4_i4b)/xstat(3_i4b)**3)
!
                        if ( cnt>three ) then
!
                            cnt3         = max( cnt - three, one )
                            xstat(5_i4b) = (di2*(cnt+one)/cnt3)*(xstat(5_i4b)/xstat(2_i4b)**2)  &
                                           - three*(cntm1**2/(cnt2*cnt3))
!
                        else
!
                            xstat(5_i4b) = di
!
                        end if
!
                    else
!
                        xstat(4_i4b:5_i4b) = di
!
                    end if
!
                else
!
!                   COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                    xstat(4_i4b) = xstat(4_i4b)/( cnt*xstat(3_i4b)**3 )
                    xstat(5_i4b) = xstat(5_i4b)/( cnt*xstat(2_i4b)**2 ) - three
!
                end if
!
            else
!
                xstat(2_i4b:3_i4b) = zero
                xstat(4_i4b:5_i4b) = di
!
            end if
!
        else if ( cnt==one ) then
!
!           EXACTLY ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            xstat(2_i4b:5_i4b) = di
!
        else
!
!           LESS THAN ONE VALID OBSERVATION, SET ALL STATISTICS TO NaN CODE.
!
            xstat(1_i4b:7_i4b) = di
!
        end if
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
    else
!
!       THE CURRENT SUBVECTOR IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_rv
! _________________________________
!
    end subroutine comp_unistat_rv
!
! =========================================================================================
!
    subroutine comp_unistat_rm( x, first, last, xstat, dimvar, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT computes estimates of univariate statistics from a data matrix.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:)
!          On entry, input submatrix containing size(X,3-DIMVAR)
!          observations on size(X,DIMVAR) variables from the matrix of data
!          for which basic univariate statistics are desired. By default, 
!          DIMVAR is equal to 1. See description of optional DIMVAR
!          argument for details.
!          If all the data are available at once, X can be the full 
!          data matrix.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current submatrix is the first submatrix 
!            of the data matrix.
!          - FIRST = false the current submatrix is not the first submatrix
!            of the data matrix.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current submatrix is the last submatrix 
!            of the data matrix.
!          - LAST = false the current submatrix is not the last submatrix
!            of the data matrix.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(:,7)
!          On entry, after the first call to COMP_UNISTAT (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT. XSTAT should not be changed between calls to
!          COMP_UNISTAT.
!
!          On exit, when LAST=true, each column of XSTAT contains the following
!          statistics on all variables:
!
!          - XSTAT(:,1) contains the mean values.
!          - XSTAT(:,2) contains the variances.
!          - XSTAT(:,3) contains the standard deviations.
!          - XSTAT(:,4) contains the coefficients of skewness.
!          - XSTAT(:,5) contains the coefficients of kurtosis.
!          - XSTAT(:,6) contains the minima.
!          - XSTAT(:,7) contains the maxima.
!
!          The shape of XSTAT must verify size(XSTAT,1) = size(X,DIMVAR)
!          and size(XSTAT,2) = 7 .
!
!   DIMVAR (INPUT, OPTIONAL) integer(i4b)
!          On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!          - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!            on size(X,1) variables.
!          - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!            on size(X,2) variables.
!
!          The default is DIMVAR = 1.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data matrix. XNOBS needs to be specified only on the
!          last call to COMP_UNISTAT (e.g. when LAST=true).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than four valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : nan, machhuge
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)   :: x
    real(stnd), intent(inout), dimension(:,:)   :: xstat
!
    integer(i4b), intent(in),  optional  :: dimvar
    integer(i4b), intent(out), optional  :: xnobs
!
    logical(lgl), intent(in)             :: first, last
    logical(lgl), intent(in),  optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                             :: cnt, cntm1, cnt2, cnt3, missing, tmp
    real(stnd), dimension(size(xstat,1))   :: di, di2
!
    integer(i4b) :: i, n, m
!
    logical(lgl) :: dim_is_one, comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xstat,1),i4b),        &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    n = size( x, 3-int(i) )
!
    call assert( logical(size(xstat,2)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xstat .
!
        xstat(:m,1_i4b:5_i4b) = zero       
        xstat(:m,6_i4b)       = machhuge       
        xstat(:m,7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(1_i4b,3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, CORRECTED SUMS OF CUBES,
!       MINIMA AND MAXIMA, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt   = cntm1 + one
        cnt2  = cnt*cnt
        cnt3  = cnt*cnt2
!
        if ( dim_is_one ) then
            di2(:m) = x(:m,i)
        else
            di2(:m) = x(i,:m)
        end if
!
        xstat(:m,6_i4b) = min( di2(:m), xstat(:m,6_i4b) )
        xstat(:m,7_i4b) = max( di2(:m), xstat(:m,7_i4b) )
!
        di(:m)  = di2(:m) - xstat(:m,1_i4b)
        di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
        tmp = one/cnt
!
        call axpy( m, tmp, di(:m), 1_i4b, xstat(:m,1_i4b), 1_i4b )
!
        tmp = cntm1/cnt
!
        call axpy( m, tmp, di2(:m), 1_i4b, xstat(:m,2_i4b), 1_i4b )
#else
        xstat(:m,1_i4b) = xstat(:m,1_i4b) + di(:m)*(one/cnt)
        xstat(:m,2_i4b) = xstat(:m,2_i4b) + di2(:m)*(cntm1/cnt)
#endif
!
        xstat(:m,4_i4b) = xstat(:m,4_i4b) + di(:m)*( di2(:m)*((cnt2-one)/cnt2) - (three/cnt)*xstat(:m,2_i4b) )
        xstat(:m,5_i4b) = xstat(:m,5_i4b) + di(:m)*( di(:m)*( di2(:m)*((cnt3-one)/cnt3) - (six/cnt2)*xstat(:m,2_i4b) )   &
                          - (four/cnt)*xstat(:m,4_i4b) )
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        missing = nan()
!
        if ( cnt>one ) then
!
            cntm1 = cnt - one
!
            if ( comp_nobias ) then
!
                cnt2 = max( cnt - two, one )
                cnt3 = max( cnt - three, one )
                tmp = cnt/(cntm1*cnt2)
                cnt2 = three*(cntm1**2/(cnt2*cnt3))
                cnt3 = tmp*(cnt+one)/cnt3
!
                where ( xstat(:m,2_i4b)>zero ) 
!
!                   COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                    xstat(:m,2_i4b) = xstat(:m,2_i4b)*( one/cntm1)
                    xstat(:m,3_i4b) = sqrt( xstat(:m,2_i4b) )
!
!                   COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                    xstat(:m,4_i4b) = merge( tmp*(xstat(:m,4_i4b)/xstat(:m,3_i4b)**3), missing, cnt>two )
                    xstat(:m,5_i4b) = merge( cnt3*(xstat(:m,5_i4b)/xstat(:m,2_i4b)**2) - cnt2, missing, cnt>three )
!
                elsewhere
!
                    xstat(:m,2_i4b) = zero
                    xstat(:m,3_i4b) = zero
                    xstat(:m,4_i4b) = missing
                    xstat(:m,5_i4b) = missing
!
                end where
!
            else
!
                where ( xstat(:m,2_i4b)>zero ) 
!
!                   COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                    xstat(:m,2_i4b) = xstat(:m,2_i4b)*( one/cntm1)
                    xstat(:m,3_i4b) = sqrt( xstat(:m,2_i4b) )
!
!                   COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                    xstat(:m,4_i4b) = xstat(:m,4_i4b)/( cnt*xstat(:m,3_i4b)**3 )
                    xstat(:m,5_i4b) = xstat(:m,5_i4b)/( cnt*xstat(:m,2_i4b)**2 ) - three
!
                elsewhere
!
                    xstat(:m,2_i4b) = zero
                    xstat(:m,3_i4b) = zero
                    xstat(:m,4_i4b) = missing
                    xstat(:m,5_i4b) = missing
!
                end where
!
            end if
!
        else if ( cnt==one ) then
!
!           EXACTLY ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            xstat(:m,2_i4b:5_i4b) = missing
!
        else
!
!           LESS THAN ONE VALID OBSERVATION, SET ALL STATISTICS TO NaN CODE.
!
            xstat(:m,1_i4b:7_i4b) = missing
!
        end if
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
    else
!
!       THE CURRENT SUBMATRIX IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(1_i4b,3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_rm
! _________________________________
!
    end subroutine comp_unistat_rm
!
! =========================================================================================
!
    subroutine comp_unistat_rt( x, first, last, xstat, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT computes estimates of univariate statistics from a tridimensional 
!   data array.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:)
!          On entry, input subarray containing size(X,3) observations
!          on size(X,1) by size(X,2) variables from the array of data for which 
!          basic univariate statistics are desired.
!          If all the data are available at once, X can be the full data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!            of the data array.
!          - LAST = false the current subarray is not the last subarray
!            of the data array.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(:,:,7)
!          On entry, after the first call to COMP_UNISTAT (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT. XSTAT should not be changed between calls to
!          COMP_UNISTAT.
!
!          On exit, when LAST=true, each matrix of XSTAT contains the following
!          statistics on all variables:
!
!          - XSTAT(:,:,1) contains the mean values.
!          - XSTAT(:,:,2) contains the variances.
!          - XSTAT(:,:,3) contains the standard deviations.
!          - XSTAT(:,:,4) contains the coefficients of skewness.
!          - XSTAT(:,:,5) contains the coefficients of kurtosis.
!          - XSTAT(:,:,6) contains the minima.
!          - XSTAT(:,:,7) contains the maxima.
!
!          The shape of XSTAT must verify:
!
!          - size(XSTAT,1) = size(X,1)
!          - size(XSTAT,2) = size(X,2)
!          - size(XSTAT,3) = 7 .
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data array. XNOBS needs to be specified only on the
!          last call to COMP_UNISTAT (e.g. when LAST=true).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than four valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : nan, machhuge
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:)           :: x
    real(stnd), intent(inout), dimension(:,:,:)           :: xstat
!
    integer(i4b), intent(out), optional  :: xnobs
!
    logical(lgl), intent(in)             :: first, last
    logical(lgl), intent(in),  optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: cnt, cntm1, cnt2, cnt3, missing, tmp
    real(stnd), dimension(size(x,1)) :: di, di2
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl) :: comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 3 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),       &
                    int(size(xstat,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),       &
                    int(size(xstat,2),i4b),   &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,3)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xstat .
!
        xstat(:m,:p,1_i4b:5_i4b) = zero       
        xstat(:m,:p,6_i4b)       = machhuge       
        xstat(:m,:p,7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(1_i4b,1_i4b,3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, CORRECTED SUMS OF CUBES,
!       MINIMA AND MAXIMA, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt = cntm1 + one
        cnt2 = cnt*cnt
        cnt3 = cnt*cnt2
!
        do j = 1_i4b, p
!
            di(:m)  = x(:m,j,i) - xstat(:m,j,1_i4b)
            di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
            tmp = one/cnt
!
            call axpy( m, tmp, di(:m), 1_i4b, xstat(:m,j,1_i4b), 1_i4b )
!
            tmp = cntm1/cnt
!
            call axpy( m, tmp, di2(:m), 1_i4b, xstat(:m,j,2_i4b), 1_i4b )
#else
            xstat(:m,j,1_i4b) = xstat(:m,j,1_i4b) + di(:m)*(one/cnt)
            xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b) + di2(:m)*(cntm1/cnt)
#endif
!
            xstat(:m,j,4_i4b) = xstat(:m,j,4_i4b) + di(:m)*( di2(:m)*((cnt2-one)/cnt2) - &
                                (three/cnt)*xstat(:m,j,2_i4b) )
            xstat(:m,j,5_i4b) = xstat(:m,j,5_i4b) + di(:m)*( di(:m)*( di2(:m)*((cnt3-one)/cnt3) -  &
                                (six/cnt2)*xstat(:m,j,2_i4b) )  - (four/cnt)*xstat(:m,j,4_i4b) )
!
            xstat(:m,j,6_i4b) = min( x(:m,j,i), xstat(:m,j,6_i4b) )
            xstat(:m,j,7_i4b) = max( x(:m,j,i), xstat(:m,j,7_i4b) )
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
!
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        missing = nan()
!
        if ( cnt>one ) then
!
            cntm1 = cnt - one
!
            if ( comp_nobias ) then
!
                cnt2 = max( cnt - two, one )
                cnt3 = max( cnt - three, one )
                tmp  = cnt/(cntm1*cnt2)
                cnt2 = three*(cntm1**2/(cnt2*cnt3))
                cnt3 = tmp*(cnt+one)/cnt3
!
                where ( xstat(:m,:p,2_i4b)>zero ) 
!
!                   COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                    xstat(:m,:p,2_i4b) = xstat(:m,:p,2_i4b)*( one/cntm1 )
                    xstat(:m,:p,3_i4b) = sqrt( xstat(:m,:p,2_i4b) )
!
!                   COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                    xstat(:m,:p,4_i4b) = merge( tmp*(xstat(:m,:p,4_i4b)/xstat(:m,:p,3_i4b)**3),    &
                                                missing, cnt>two )
                    xstat(:m,:p,5_i4b) = merge( cnt3*(xstat(:m,:p,5_i4b)/xstat(:m,:p,2_i4b)**2) - cnt2,         &
                                                missing, cnt>three )
!
                elsewhere
!
                    xstat(:m,:p,2_i4b) = zero
                    xstat(:m,:p,3_i4b) = zero
                    xstat(:m,:p,4_i4b) = missing
                    xstat(:m,:p,5_i4b) = missing 
!
                end where
!
            else
!
                where ( xstat(:m,:p,2_i4b)>zero ) 
!
!                   COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                    xstat(:m,:p,2_i4b) = xstat(:m,:p,2_i4b)*( one/cntm1 )
                    xstat(:m,:p,3_i4b) = sqrt( xstat(:m,:p,2_i4b) )
!
!                   COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                    xstat(:m,:p,4_i4b) = xstat(:m,:p,4_i4b)/( cnt*xstat(:m,:p,3_i4b)**3 )
                    xstat(:m,:p,5_i4b) = xstat(:m,:p,5_i4b)/( cnt*xstat(:m,:p,2_i4b)**2 ) - three
!
                elsewhere
!
                    xstat(:m,:p,2_i4b) = zero
                    xstat(:m,:p,3_i4b) = zero
                    xstat(:m,:p,4_i4b) = missing
                    xstat(:m,:p,5_i4b) = missing
!
                end where
!
            end if
!
        else if ( cnt==one ) then
!
!           EXACTLY ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            xstat(:m,:p,2_i4b:5_i4b) = missing
!
        else
!
!           LESS THAN ONE VALID OBSERVATION, SET ALL STATISTICS TO NaN CODE.
!
            xstat(:m,:p,1_i4b:7_i4b) = missing
!
        end if
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
    else
!
!       THE CURRENT SUBARRAY IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(1_i4b,1_i4b,3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_rt
! _________________________________
!
    end subroutine comp_unistat_rt
!
! =========================================================================================
!
    subroutine comp_unistat_rf( x, first, last, xstat, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT computes estimates of univariate statistics from a fourdimensional 
!   data array.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:,:)
!          On entry, input subarray containing size(X,4) observations
!          on size(X,1) by size(X,2) by size(X,3) variables from the array
!          of data for which basic univariate statistics are desired.
!          If all the data are available at once, X can be the full data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!            of the data array.
!          - LAST = false the current subarray is not the last subarray
!            of the data array.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(:,:,:,7)
!          On entry, after the first call to COMP_UNISTAT (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT. XSTAT should not be changed between calls to
!          COMP_UNISTAT.
!
!          On exit, when LAST=true, each matrix of XSTAT contains the following
!          statistics on all variables:
!
!          - XSTAT(:,:,:,1) contains the mean values.
!          - XSTAT(:,:,:,2) contains the variances.
!          - XSTAT(:,:,:,3) contains the standard deviations.
!          - XSTAT(:,:,:,4) contains the coefficients of skewness.
!          - XSTAT(:,:,:,5) contains the coefficients of kurtosis.
!          - XSTAT(:,:,:,6) contains the minima.
!          - XSTAT(:,:,:,7) contains the maxima.
!
!          The shape of XSTAT must verify:
!
!          - size(XSTAT,1) = size(X,1),
!          - size(XSTAT,2) = size(X,2)
!          - size(XSTAT,3) = size(X,3)
!          - size(XSTAT,4) = 7 .
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data array. XNOBS needs to be specified only on the
!          last call to COMP_UNISTAT (e.g. when LAST=true).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than four valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : nan, machhuge
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:,:)           :: x
    real(stnd), intent(inout), dimension(:,:,:,:)           :: xstat
!
    integer(i4b), intent(out), optional  :: xnobs
!
    logical(lgl), intent(in)             :: first, last
    logical(lgl), intent(in),  optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: cnt, cntm1, cnt2, cnt3, missing, tmp
    real(stnd), dimension(size(x,1)) :: di, di2
!
    integer(i4b) :: i, j, k, n, m, p, q
!
    logical(lgl) :: comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 4 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),       &
                    int(size(xstat,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),       &
                    int(size(xstat,2),i4b),   &
                    name_proc )
!
    q = assert_eq(  int(size(x,3),i4b),       &
                    int(size(xstat,3),i4b),   &
                    name_proc )
!
    if ( min(m,p,q)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,4)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xstat .
!
        xstat(:m,:p,:q,1_i4b:5_i4b) = zero       
        xstat(:m,:p,:q,6_i4b)       = machhuge       
        xstat(:m,:p,:q,7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(1_i4b,1_i4b,1_i4b,3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, CORRECTED SUMS OF CUBES,
!       MINIMA AND MAXIMA, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt   = cntm1 + one
        cnt2  = cnt*cnt
        cnt3  = cnt*cnt2
!
        do k = 1_i4b, q
!
            do j = 1_i4b, p
!
                di(:m) = x(:m,j,k,i) - xstat(:m,j,k,1_i4b)
                di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
                tmp = one/cnt
!
                call axpy( m, tmp, di(:m), 1_i4b, xstat(:m,j,k,1_i4b), 1_i4b )
!
                tmp = cntm1/cnt
!
                call axpy( m, tmp, di2(:m), 1_i4b, xstat(:m,j,k,2_i4b), 1_i4b )
#else
                xstat(:m,j,k,1_i4b) = xstat(:m,j,k,1_i4b) + di(:m)*(one/cnt)
                xstat(:m,j,k,2_i4b) = xstat(:m,j,k,2_i4b) + di2(:m)*(cntm1/cnt)
#endif
!
                xstat(:m,j,k,4_i4b) = xstat(:m,j,k,4_i4b) + di(:m)*( di2(:m)*((cnt2-one)/cnt2) -     &
                                      (three/cnt)*xstat(:m,j,k,2_i4b) )
                xstat(:m,j,k,5_i4b) = xstat(:m,j,k,5_i4b) + di(:m)*( di(:m)*( di2(:m)*((cnt3-one)/cnt3) -  &
                                      (six/cnt2)*xstat(:m,j,k,2_i4b) ) - (four/cnt)*xstat(:m,j,k,4_i4b) )
!
                xstat(:m,j,k,6_i4b) = min( x(:m,j,k,i), xstat(:m,j,k,6_i4b) )
                xstat(:m,j,k,7_i4b) = max( x(:m,j,k,i), xstat(:m,j,k,7_i4b) )
!
            end do
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        missing = nan()
!
        if ( cnt>one ) then
!
            cntm1 = cnt - one
!
            if ( comp_nobias ) then
!
                cnt2 = max( cnt - two, one )
                cnt3 = max( cnt - three, one )
                tmp = cnt/(cntm1*cnt2)
                cnt2 = three*(cntm1**2/(cnt2*cnt3))
                cnt3 = tmp*(cnt+one)/cnt3
!
                where ( xstat(:m,:p,:q,2_i4b)>zero ) 
!
!                   COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                    xstat(:m,:p,:q,2_i4b) = xstat(:m,:p,:q,2_i4b)*( one/cntm1 )
                    xstat(:m,:p,:q,3_i4b) = sqrt( xstat(:m,:p,:q,2_i4b) )
!
!                   COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                    xstat(:m,:p,:q,4_i4b) =                                          &
                     merge( tmp*(xstat(:m,:p,:q,4_i4b)/xstat(:m,:p,:q,3_i4b)**3),    &
                            missing, cnt>two )
                    xstat(:m,:p,:q,5_i4b) =                                                       &
                     merge( cnt3*(xstat(:m,:p,:q,5_i4b)/xstat(:m,:p,:q,2_i4b)**2) - cnt2,         &
                            missing, cnt>three )
!
                elsewhere
!
                    xstat(:m,:p,:q,2_i4b) = zero
                    xstat(:m,:p,:q,3_i4b) = zero
                    xstat(:m,:p,:q,4_i4b) = missing
                    xstat(:m,:p,:q,5_i4b) = missing
!
                end where
!
            else
!
                where ( xstat(:m,:p,:q,2_i4b)>zero ) 
!
!                   COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                    xstat(:m,:p,:q,2_i4b) = xstat(:m,:p,:q,2_i4b)*( one/cntm1 )
                    xstat(:m,:p,:q,3_i4b) = sqrt( xstat(:m,:p,:q,2_i4b) )
!
!                   COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                    xstat(:m,:p,:q,4_i4b) = xstat(:m,:p,:q,4_i4b)/( cnt*xstat(:m,:p,:q,3_i4b)**3 )
                    xstat(:m,:p,:q,5_i4b) = xstat(:m,:p,:q,5_i4b)/( cnt*xstat(:m,:p,:q,2_i4b)**2 ) - three
!
                elsewhere
!
                    xstat(:m,:p,:q,2_i4b) = zero
                    xstat(:m,:p,:q,3_i4b) = zero
                    xstat(:m,:p,:q,4_i4b) = missing
                    xstat(:m,:p,:q,5_i4b) = missing
!
                end where
!
            end if
!
        else if ( cnt==one ) then
!
!           EXACTLY ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            xstat(:m,:p,:q,2_i4b:5_i4b) = missing
!
        else
!
!           LESS THAN ONE VALID OBSERVATION, SET ALL STATISTICS TO NaN CODE.
!
            xstat(:m,:p,:q,1_i4b:7_i4b) = missing
!
        end if
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
    else
!
!       THE CURRENT SUBARRAY IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(1_i4b,1_i4b,1_i4b,3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_rf
! _________________________________
!
    end subroutine comp_unistat_rf
!
! =========================================================================================
!
    subroutine comp_unistat_miss_rv( x, first, last, xstat, xmiss, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT_MISS computes estimates of univariate statistics from a data vector
!   possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, input subvector containing size(X) observations
!          from the vector of data for which basic univariate statistics
!          are desired. If all the data are available at once, X can be
!          the full data vector.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subvector is the first subvector 
!            of the data vector.
!          - FIRST = false the current subvector is not the first subvector
!            of the data vector.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subvector is the last subvector 
!                          of the data vector.
!          - LAST = false the current subvector is not the last subvector
!                          of the data vector.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(7)
!          On entry, after the first call to COMP_UNISTAT_MISS (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT_MISS. XSTAT should not be changed between calls to
!          COMP_UNISTAT_MISS.
!
!          On exit, when LAST=true, XSTAT contains the following statistics :
!
!          - XSTAT(1) contains the mean value of the data vector.
!          - XSTAT(2) contains the variance of the data vector.
!          - XSTAT(3) contains the standard deviation of the data vector.
!          - XSTAT(4) contains the coefficient of skewness of the data vector.
!          - XSTAT(5) contains the coefficient of kurtosis of the data vector.
!          - XSTAT(6) contains the minimum of the data vector.
!          - XSTAT(7) contains the maximum of the data vector.
!
!          The size of XSTAT must verify size(XSTAT) = 7.
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of non-missing observations in
!          the data vector. XNOBS needs to be specified only on the last call
!          to COMP_UNISTAT_MISS (e.g. when LAST=true).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!  The subroutine computes the basic statistics with only one pass through the data.
!
!  If fewer than four valid observations were present, the pertinent statistics are set
!  to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : machhuge
    use Utilities,         only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:)           :: x
    real(stnd), intent(inout), dimension(:)           :: xstat
    real(stnd), intent(in)                            :: xmiss
!
    integer(i4b), intent(out),               optional :: xnobs
!
    logical(lgl), intent(in)                          :: first, last
    logical(lgl), intent(in),               optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cnt, cntm1, cnt2, cnt3, di, di2
!
    integer(i4b) :: i, n
!
    logical(lgl) :: comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
!   TEST THE ARGUMENTS.
!
    call assert( logical(size(xstat)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!
!       INITIALIZE xstat .
!
        xstat(1_i4b:5_i4b) = zero       
        xstat(6_i4b)       = machhuge       
        xstat(7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(3_i4b)
!
    do i = 1_i4b, n
!
        if ( x(i)==xmiss ) cycle
!
!       CALCULATE PARTIAL MEAN, CORRECTED SUM OF SQUARES, CORRECTED SUM OF CUBES,
!       MINIMUM AND MAXIMUM, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt = cntm1 + one
        cnt2 = cnt*cnt
        cnt3 = cnt*cnt2
!
        di = x(i) - xstat(1_i4b)
        di2 = di*di
        xstat(1_i4b) = xstat(1_i4b) + di/cnt
        xstat(2_i4b) = xstat(2_i4b) + di2*(cntm1/cnt)
        xstat(4_i4b) = xstat(4_i4b) + di*( di2*((cnt2-one)/cnt2) - (three/cnt)*xstat(2_i4b) )
        xstat(5_i4b) = xstat(5_i4b) + di*( di*( di2*((cnt3-one)/cnt3) - (six/cnt2)*xstat(2_i4b) )     &
                                           - (four/cnt)*xstat(4_i4b) )
!
        xstat(6_i4b) = min( x(i), xstat(6_i4b) )
        xstat(7_i4b) = max( x(i), xstat(7_i4b) )
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        if ( cnt>one ) then
!
            if ( xstat(2_i4b)>zero ) then
!
                cntm1 = cnt - one
!
!               COMPUTE VARIANCE AND STANDARD-DEVIATION.
!
                xstat(2_i4b) = xstat(2_i4b)/cntm1
                xstat(3_i4b) = sqrt( xstat(2_i4b) )
!
                if ( comp_nobias ) then
!
!                   COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                    if ( cnt>two ) then
!
                        cnt2 = max( cnt - two, one )
                        di2 = cnt/(cntm1*cnt2)
!
                        xstat(4_i4b) = di2*(xstat(4_i4b)/xstat(3_i4b)**3)
!
                        if ( cnt>three ) then
                            cnt3         = max( cnt - three, one )
                            xstat(5_i4b) = (di2*(cnt+one)/cnt3)*(xstat(5_i4b)/xstat(2_i4b)**2)      &
                                           - three*(cntm1**2/(cnt2*cnt3))
                        else
                            xstat(5_i4b) = xmiss
                        end if
!
                    else
!
                        xstat(4_i4b:5_i4b) = xmiss
!
                    end if
!
                else
!
!                   COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                    xstat(4_i4b) = xstat(4_i4b)/( cnt*xstat(3_i4b)**3 )
                    xstat(5_i4b) = xstat(5_i4b)/( cnt*xstat(2_i4b)**2 ) - three
!
                end if
!
            else
!
                xstat(2_i4b:3_i4b) = zero
                xstat(4_i4b:5_i4b) = xmiss
!
            end if
!
        else if ( cnt==one ) then
!
!           EXACTLY ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            xstat(2_i4b:5_i4b) = xmiss
!
        else
!
!           LESS THAN ONE VALID OBSERVATION, SET ALL STATISTICS TO NaN CODE.
!
            xstat(1_i4b:7_i4b) = xmiss
!
        end if
!
!       OUTPUT NUMBER OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
    else
!
!       THE CURRENT SUBVECTOR IS NOT THE LAST.
!       SAVE COUNT OF NON-MISSING OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_miss_rv
! ______________________________________
!
    end subroutine comp_unistat_miss_rv
!
! =========================================================================================
!
    subroutine comp_unistat_miss_rm( x, first, last, xstat, xmiss, dimvar, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT_MISS computes estimates of univariate statistics from a data matrix
!   possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:)
!          On entry, input submatrix containing size(X,3-DIMVAR)
!          observations on size(X,DIMVAR) variables from the matrix of data
!          for which basic univariate statistics are desired. By default, 
!          DIMVAR is equal to 1. See description of optional DIMVAR
!          argument for details.
!          If all the data are available at once, X can be the full 
!          data matrix.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current submatrix is the first submatrix 
!            of the data matrix.
!          - FIRST = false the current submatrix is not the first submatrix
!            of the data matrix.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current submatrix is the last submatrix 
!                          of the data matrix.
!          - LAST = false the current submatrix is not the last submatrix
!                          of the data matrix.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(:,7)
!          On entry, after the first call to COMP_UNISTAT_MISS (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT_MISS. XSTAT should not be changed between calls to
!          COMP_UNISTAT_MISS.
!
!          On exit, when LAST=true, each column of XSTAT contains the following
!          statistics on all variables:
!
!          - XSTAT(:,1) contains the mean values.
!          - XSTAT(:,2) contains the variances.
!          - XSTAT(:,3) contains the standard deviations.
!          - XSTAT(:,4) contains the coefficients of skewness.
!          - XSTAT(:,5) contains the coefficients of kurtosis.
!          - XSTAT(:,6) contains the minima.
!          - XSTAT(:,7) contains the maxima.
!
!          The shape of XSTAT must verify:
!
!          - size(XSTAT,1) = size(X,DIMVAR)
!          - size(XSTAT,2) = 7 .
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   DIMVAR (INPUT, OPTIONAL) integer(i4b)
!          On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!          - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!            on size(X,1) variables.
!          - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!            on size(X,2) variables.
!
!          The default is DIMVAR = 1.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b), dimension(:)
!          On exit, XNOBS contains the number of non-missing observations
!          on all variables. XNOBS needs to be specified only on the
!          last call to COMP_UNISTAT_MISS (e.g. when LAST=true).
!
!          The size of XNOBS must verify size(XNOBS) = size(X,DIMVAR).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!  The subroutine computes the basic statistics with only one pass through the data.
!
!  If fewer than four valid observations were present for some variables, the pertinent 
!  statistics are set to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)           :: x
    real(stnd), intent(inout), dimension(:,:)           :: xstat
    real(stnd), intent(in)                              :: xmiss
!
    integer(i4b), intent(in),                  optional :: dimvar
    integer(i4b), intent(out), dimension(:),   optional :: xnobs
!
    logical(lgl), intent(in)             :: first, last
    logical(lgl), intent(in),  optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xstat,1))   :: di, di2, cnt, cntm1, cnt2, cnt3
!
    integer(i4b) :: i, n, m
!
    logical(lgl) :: dim_is_one, comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
     if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
   m = assert_eq(  int(size(x,int(i)),i4b),       &
                   int(size(xstat,1),i4b),        &
                   name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    n = size( x, 3-int(i) )
!
    call assert( logical(size(xstat,2)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xstat .
!
        xstat(:m,1_i4b:5_i4b) = zero       
        xstat(:m,6_i4b)       = machhuge       
        xstat(:m,7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(:m,3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, CORRECTED SUMS OF CUBES,
!       MINIMA AND MAXIMA, HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            di2(:m) = x(:m,i)
        else
            di2(:m) = x(i,:m)
        end if
!
        where ( di2(:m)/=xmiss )
!
            cntm1 = cnt
            cnt = cntm1 + one
            cnt2 = cnt*cnt
            cnt3 = cnt*cnt2
!
            xstat(:m,6_i4b) = min( di2(:m), xstat(:m,6_i4b) )
            xstat(:m,7_i4b) = max( di2(:m), xstat(:m,7_i4b) )
!
            di(:m) = di2(:m) - xstat(:m,1_i4b)
            di2(:m) = di(:m)*di(:m)
!
            xstat(:m,1_i4b) = xstat(:m,1_i4b) + di/cnt
            xstat(:m,2_i4b) = xstat(:m,2_i4b) + di2*(cntm1/cnt)
            xstat(:m,4_i4b) = xstat(:m,4_i4b) + di*( di2*((cnt2-one)/cnt2) - (three/cnt)*xstat(:m,2_i4b) )
            xstat(:m,5_i4b) = xstat(:m,5_i4b) + di*( di*( di2*((cnt3-one)/cnt3) - (six/cnt2)*xstat(:m,2_i4b) )   &
                                                     - (four/cnt)*xstat(:m,4_i4b) )
!
        end where
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        if ( comp_nobias ) then
!
            where ( xstat(:m,2_i4b)>zero ) 
!
                cntm1 = cnt - one
!
!               COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                xstat(:m,2_i4b) = xstat(:m,2_i4b)/cntm1
                xstat(:m,3_i4b) = sqrt( xstat(:m,2_i4b) )
!
!               COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                where ( cnt>two )
!
                    cnt2 = max( cnt - two, one )
                    di2  = cnt/(cntm1*cnt2)
                    xstat(:m,4_i4b) = di2*(xstat(:m,4_i4b)/xstat(:m,3_i4b)**3)
!
                    where ( cnt>three )
                        cnt3 = max( cnt - three, one )
                        xstat(:m,5_i4b) = (di2*(cnt+one)/cnt3)*(xstat(:m,5_i4b)/xstat(:m,2_i4b)**2)      &
                                           - three*(cntm1**2/(cnt2*cnt3))
                    elsewhere
                        xstat(:m,5_i4b) = xmiss
                    end where
!
                elsewhere
!
                    xstat(:m,4_i4b) = xmiss
                    xstat(:m,5_i4b) = xmiss
!
                end where
!
            elsewhere
!
                xstat(:m,2_i4b) = merge( zero, xmiss, cnt>one )
                xstat(:m,3_i4b) = xstat(:m,2_i4b)
                xstat(:m,4_i4b) = xmiss
                xstat(:m,5_i4b) = xmiss
!
            end where
!
        else
!
            where ( xstat(:m,2_i4b)>zero ) 
!
!               COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                xstat(:m,2_i4b) = xstat(:m,2_i4b)/( cnt - one )
                xstat(:m,3_i4b) = sqrt( xstat(:m,2_i4b) )
!
!               COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                xstat(:m,4_i4b) = xstat(:m,4_i4b)/( cnt*xstat(:m,3_i4b)**3 )
                xstat(:m,5_i4b) = xstat(:m,5_i4b)/( cnt*xstat(:m,2_i4b)**2 ) - three
!
            elsewhere
!
                xstat(:m,2_i4b) = merge( zero, xmiss, cnt>one )
                xstat(:m,3_i4b) = xstat(:m,2_i4b)
                xstat(:m,4_i4b) = xmiss
                xstat(:m,5_i4b) = xmiss
!
            end where
!
        end if
!
!       LESS THAN ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO xmiss .
!
        where ( cnt<one )
!
            xstat(:m,1_i4b) = xmiss
            xstat(:m,6_i4b) = xmiss
            xstat(:m,7_i4b) = xmiss
!
        end where
!
!       OUTPUT NUMBERS OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) then
!
           call assert( logical(size(xnobs)==m,lgl), name_proc )
!
           xnobs(:m) = int( cnt, i4b )
!
        end if
!
    else
!
!       THE CURRENT SUBMATRIX IS NOT THE LAST.
!       SAVE COUNTS OF NON-MISSING OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(:m,3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_miss_rm
! ______________________________________
!
    end subroutine comp_unistat_miss_rm
!
! =========================================================================================
!
    subroutine comp_unistat_miss_rt( x, first, last, xstat, xmiss, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT_MISS computes estimates of univariate statistics from a tridimensional 
!   data array possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:)
!          On entry, input subarray containing size(X,3) observations
!          on size(X,1) by size(X,2) variables from the array of data for which 
!          basic univariate statistics are desired.
!          If all the data are available at once, X can be the full tridimensional
!          data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!                          of the data array.
!          - LAST = false the current subarray is not the last subarray
!                          of the data array.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(:,:,7)
!          On entry, after the first call to COMP_UNISTAT_MISS (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT_MISS. XSTAT should not be changed between calls to
!          COMP_UNISTAT_MISS.
!
!          On exit, when LAST=true, each matrix of XSTAT contains the following
!          statistics on all variables:
!
!          - XSTAT(:,:,1) contains the mean values.
!          - XSTAT(:,:,2) contains the variances.
!          - XSTAT(:,:,3) contains the standard deviations.
!          - XSTAT(:,:,4) contains the coefficients of skewness.
!          - XSTAT(:,:,5) contains the coefficients of kurtosis.
!          - XSTAT(:,:,6) contains the minima.
!          - XSTAT(:,:,7) contains the maxima.
!
!          The shape of XSTAT must verify:
!
!          - size(XSTAT,1) = size(X,1)
!          - size(XSTAT,2) = size(X,2)
!          - size(XSTAT,3) = 7 .
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b), dimension(:,:)
!          On exit, XNOBS contains the numbers of non-missing observations
!          on all variables. XNOBS needs to be specified only on the
!          last call to COMP_UNISTAT_MISS (e.g. when LAST=true).
!
!          The shape of XNOBS must verify:
!
!          - size(XNOBS,1) = size(X,1)
!          - size(XNOBS,2) = size(X,2).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!  The subroutine computes the basic statistics with only one pass through the data.
!
!  If fewer than four valid observations were present for some variables, the pertinent 
!  statistics are set to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:)           :: x
    real(stnd), intent(inout), dimension(:,:,:)           :: xstat
    real(stnd), intent(in)                                :: xmiss
!
    integer(i4b), intent(out), dimension(:,:),   optional :: xnobs
!
    logical(lgl), intent(in)             :: first, last
    logical(lgl), intent(in),  optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(x,1),size(x,2)) :: cnt
    real(stnd), dimension(size(x,1))           :: di, di2, cntm1, cnt2, cnt3
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl) :: comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 3 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),       &
                    int(size(xstat,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),       &
                    int(size(xstat,2),i4b),   &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,3)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xstat .
!
        xstat(:m,:p,1_i4b:5_i4b) = zero       
        xstat(:m,:p,6_i4b)       = machhuge       
        xstat(:m,:p,7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(:m,:p,3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, CORRECTED SUMS OF CUBES,
!       MINIMA AND MAXIMA, HANDLING THE DATA ONCE.
!
        do j = 1_i4b, p
!
            where ( x(:m,j,i)/=xmiss )
!
                cntm1(:m) = cnt(:m,j)
                cnt(:m,j) = cntm1(:m) + one
                cnt2(:m)  = cnt(:m,j)*cnt(:m,j)
                cnt3(:m)  = cnt(:m,j)*cnt2(:m)
!
                di(:m)  = x(:m,j,i) - xstat(:m,j,1_i4b)
                di2(:m) = di(:m)*di(:m)
!
                xstat(:m,j,1_i4b) = xstat(:m,j,1_i4b) + di(:m)/cnt(:m,j)
                xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b) + di2(:m)*( cntm1(:m)/cnt(:m,j) )
                xstat(:m,j,4_i4b) = xstat(:m,j,4_i4b) + di(:m)*( di2(:m)*( (cnt2(:m)-one)/cnt2(:m) )   &
                                                        - ( three/cnt(:m,j) )*xstat(:m,j,2_i4b) )
                xstat(:m,j,5_i4b) = xstat(:m,j,5_i4b) + di(:m)*( di(:m)*( di2(:m)*( (cnt3(:m)-one)/cnt3(:m) )  &
                                                        - ( six/cnt2(:m) )*xstat(:m,j,2_i4b) )                 &
                                                        - ( four/cnt(:m,j) )*xstat(:m,j,4_i4b) )
!
                xstat(:m,j,6_i4b) = min( x(:m,j,i), xstat(:m,j,6_i4b) )
                xstat(:m,j,7_i4b) = max( x(:m,j,i), xstat(:m,j,7_i4b) )
!
            end where
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        if ( comp_nobias ) then
!
            do j = 1_i4b, p
!
                where ( xstat(:m,j,2_i4b)>zero ) 
!
                    cntm1(:m) = cnt(:m,j) - one
!
!                   wCOMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                    xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b)/cntm1(:m)
                    xstat(:m,j,3_i4b) = sqrt( xstat(:m,j,2_i4b) )
!
!                   COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                    where ( cnt(:m,j)>two )
!
                        cnt2(:m) = max( cnt(:m,j) - two, one )
                        di2(:m)  = cnt(:m,j)/(cntm1(:m)*cnt2(:m))
                        xstat(:m,j,4_i4b) = di2(:m)*(xstat(:m,j,4_i4b)/xstat(:m,j,3_i4b)**3)
!
                        where ( cnt(:m,j)>three )
                            cnt3(:m)          = max( cnt(:m,j) - three, one )
                            xstat(:m,j,5_i4b) = (di2(:m)*(cnt(:m,j)+one)/cnt3(:m))*(xstat(:m,j,5_i4b)/xstat(:m,j,2_i4b)**2)  &
                                                 - three*(cntm1(:m)**2/(cnt2(:m)*cnt3(:m)))
                        elsewhere
                            xstat(:m,j,5_i4b) = xmiss
                        end where
!
                    elsewhere
!
                        xstat(:m,j,4_i4b) = xmiss
                        xstat(:m,j,5_i4b) = xmiss
!
                    end where
!
                elsewhere
!
                    xstat(:m,j,2_i4b) = merge( zero, xmiss, cnt(:m,j)>one )
                    xstat(:m,j,3_i4b) = xstat(:m,j,2_i4b)
                    xstat(:m,j,4_i4b) = xmiss
                    xstat(:m,j,5_i4b) = xmiss
!
                end where
!
            end do
!
        else
!
            where ( xstat(:m,:p,2_i4b)>zero ) 
!
!               COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                xstat(:m,:p,2_i4b) = xstat(:m,:p,2_i4b)/( cnt - one )
                xstat(:m,:p,3_i4b) = sqrt( xstat(:m,:p,2_i4b) )
!
!               COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                xstat(:m,:p,4_i4b) = xstat(:m,:p,4_i4b)/( cnt*xstat(:m,:p,3_i4b)**3 )
                xstat(:m,:p,5_i4b) = xstat(:m,:p,5_i4b)/( cnt*xstat(:m,:p,2_i4b)**2 ) - three
!
            elsewhere
!
                xstat(:m,:p,2_i4b) = merge( zero, xmiss, cnt>one )
                xstat(:m,:p,3_i4b) = xstat(:m,:p,2_i4b)
                xstat(:m,:p,4_i4b) = xmiss
                xstat(:m,:p,5_i4b) = xmiss
!
            end where
!
        end if
!
!       LESS THAN ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO xmiss .
!
        where ( cnt<one )
!
            xstat(:m,:p,1_i4b) = xmiss
            xstat(:m,:p,6_i4b) = xmiss
            xstat(:m,:p,7_i4b) = xmiss
!
        end where
!
!       OUTPUT NUMBERS OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) then
!
           call assert( logical(size(xnobs,1)==m,lgl),   &
                        logical(size(xnobs,2)==p,lgl),   &
                        name_proc )
!
           xnobs(:m,:p) = int( cnt, i4b )
!
        end if
!
    else
!
!       THE CURRENT SUBMATRIX IS NOT THE LAST.
!       SAVE COUNTS OF NON-MISSING OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(:m,:p,3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_miss_rt
! ______________________________________
!
    end subroutine comp_unistat_miss_rt
!
! =========================================================================================
!
    subroutine comp_unistat_miss_rf( x, first, last, xstat, xmiss, xnobs, nobias )
!
! Purpose
! _______
!
!   COMP_UNISTAT_MISS computes estimates of univariate statistics from a fourdimensional 
!   data array possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:,:)
!          On entry, input subarray containing size(X,4) observations
!          on size(X,1) by size(X,2) by size(X,3) variables from the array of data for which 
!          basic univariate statistics are desired.
!          If all the data are available at once, X can be the full tridimensional
!          data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!            of the data array.
!          - LAST = false the current subarray is not the last subarray
!            of the data array.
!
!   XSTAT  (INPUT/OUTPUT) real(stnd), dimension(:,:,:,7)
!          On entry, after the first call to COMP_UNISTAT_MISS (e.g. when FIRST=true),
!          XSTAT is used as workspace to accumulate quantities on previous calls
!          to COMP_UNISTAT_MISS. XSTAT should not be changed between calls to
!          COMP_UNISTAT_MISS.
!
!          On exit, when LAST=true, each matrix of XSTAT contains the following
!          statistics on all variables:
!
!          - XSTAT(:,:,:,1) contains the mean values.
!          - XSTAT(:,:,:,2) contains the variances.
!          - XSTAT(:,:,:,3) contains the standard deviations.
!          - XSTAT(:,:,:,4) contains the coefficients of skewness.
!          - XSTAT(:,:,:,5) contains the coefficients of kurtosis.
!          - XSTAT(:,:,:,6) contains the minima.
!          - XSTAT(:,:,:,7) contains the maxima.
!
!          The shape of XSTAT must verify:
!
!          - size(XSTAT,1) = size(X,1)
!          - size(XSTAT,2) = size(X,2)
!          - size(XSTAT,3) = size(X,3)
!          - size(XSTAT,4) = 7 .
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b), dimension(:,:,:)
!          On exit, XNOBS contains the numbers of non-missing observations
!          on all variables. XNOBS needs to be specified only on the
!          last call to COMP_UNISTAT_MISS (e.g. when LAST=true).
!
!          The shape of XNOBS must verify:
!
!          - size(XNOBS,1) = size(X,1)
!          - size(XNOBS,2) = size(X,2)
!          - size(XNOBS,3) = size(X,3).
!
!   NOBIAS (INPUT, OPTIONAL) logical(lgl)
!          On entry, when LAST=true and NOBIAS=true, unbiased estimates of skewness
!          and kurtosis are computed. If NOBIAS=false or is absent, biased estimates are computed.
!
!          NOBIAS needs to be specified only on the last call to COMP_UNISTAT_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!  The subroutine computes the basic statistics with only one pass through the data.
!
!  If fewer than four valid observations were present for some variables, the pertinent 
!  statistics are set to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, two, three, four, six
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:,:)           :: x
    real(stnd), intent(inout), dimension(:,:,:,:)           :: xstat
    real(stnd), intent(in)                                  :: xmiss
!
    integer(i4b), intent(out), dimension(:,:,:),   optional :: xnobs
!
    logical(lgl), intent(in)             :: first, last
    logical(lgl), intent(in),  optional  :: nobias
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(x,1),size(x,2),size(x,3)) :: cnt
    real(stnd), dimension(size(x,1)) :: di, di2, cntm1, cnt2, cnt3
!
    integer(i4b) :: i, j, k, n, m, p, q
!
    logical(lgl) :: comp_nobias
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_unistat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 3 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),       &
                    int(size(xstat,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),       &
                    int(size(xstat,2),i4b),   &
                    name_proc )
!
    q = assert_eq(  int(size(x,3),i4b),       &
                    int(size(xstat,3),i4b),   &
                    name_proc )
!
    if ( min(m,p,q)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,4)==7,lgl), name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xstat .
!
        xstat(:m,:p,:q,1_i4b:5_i4b) = zero       
        xstat(:m,:p,:q,6_i4b)       = machhuge       
        xstat(:m,:p,:q,7_i4b)       = -machhuge      
!
    end if
!
    cnt = xstat(:m,:p,:q,3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, CORRECTED SUMS OF CUBES,
!       MINIMA AND MAXIMA, HANDLING THE DATA ONCE.
!
        do k = 1_i4b, q
!
            do j = 1_i4b, p
!
                where ( x(:m,j,k,i)/=xmiss )
!
                    cntm1(:m)   = cnt(:m,j,k)
                    cnt(:m,j,k) = cntm1(:m) + one
                    cnt2(:m)    = cnt(:m,j,k)*cnt(:m,j,k)
                    cnt3(:m)    = cnt(:m,j,k)*cnt2(:m)
!
                    di(:m)  = x(:m,j,k,i) - xstat(:m,j,k,1_i4b)
                    di2(:m) = di(:m)*di(:m)
!
                    xstat(:m,j,k,1_i4b) = xstat(:m,j,k,1_i4b) + di(:m)/cnt(:m,j,k)
                    xstat(:m,j,k,2_i4b) = xstat(:m,j,k,2_i4b) + di2(:m)*(cntm1(:m)/cnt(:m,j,k))
                    xstat(:m,j,k,4_i4b) = xstat(:m,j,k,4_i4b) + di(:m)*( di2(:m)*    &
                                          ((cnt2(:m)-one)/cnt2(:m)) -                &
                                          (three/cnt(:m,j,k))*xstat(:m,j,k,2_i4b) )
                    xstat(:m,j,k,5_i4b) = xstat(:m,j,k,5_i4b) + di(:m)*( di(:m)*     &
                                          ( di2(:m)*((cnt3(:m)-one)/cnt3(:m)) -      &
                                          (six/cnt2(:m))*xstat(:m,j,k,2_i4b) ) -     &
                                          (four/cnt(:m,j,k))*xstat(:m,j,k,4_i4b) )
!
                    xstat(:m,j,k,6_i4b) = min( x(:m,j,k,i), xstat(:m,j,k,6_i4b) )
                    xstat(:m,j,k,7_i4b) = max( x(:m,j,k,i), xstat(:m,j,k,7_i4b) )
!
                end where
!
            end do
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        comp_nobias = false
!
        if ( present(nobias) ) then
            comp_nobias = nobias
        end if
!
        if ( comp_nobias ) then
!
            do k = 1_i4b, q
!
                do j = 1_i4b, p
!
                    where ( xstat(:m,j,k,2_i4b)>zero ) 
!
                        cntm1(:m) = cnt(:m,j,k) - one
!
!                       COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                        xstat(:m,j,k,2_i4b) = xstat(:m,j,k,2_i4b)/cntm1(:m)
                        xstat(:m,j,k,3_i4b) = sqrt( xstat(:m,j,k,2_i4b) )
!
!                       COMPUTE UNBIASED SKEWNESS AND KURTOSIS.
!
                        where ( cnt(:m,j,k)>two )
!
                            cnt2(:m) = max( cnt(:m,j,k) - two, one )
                            di2(:m)  = cnt(:m,j,k)/(cntm1(:m)*cnt2(:m))
                            xstat(:m,j,k,4_i4b) = di2(:m)*(xstat(:m,j,k,4_i4b)/xstat(:m,j,k,3_i4b)**3)
!
                            where ( cnt(:m,j,k)>three )
                                cnt3(:m)            = max( cnt(:m,j,k) - three, one )
                                xstat(:m,j,k,5_i4b) = (di2(:m)*(cnt(:m,j,k)+one)/cnt3(:m))           &
                                                      *(xstat(:m,j,k,5_i4b)/xstat(:m,j,k,2_i4b)**2)  &
                                                      - three*(cntm1(:m)**2/(cnt2(:m)*cnt3(:m)))
                            elsewhere
                                xstat(:m,j,k,5_i4b) = xmiss
                            end where
!
                        elsewhere
!
                            xstat(:m,j,k,4_i4b) = xmiss
                            xstat(:m,j,k,5_i4b) = xmiss
!
                        end where
!
                    elsewhere
!
                        xstat(:m,j,k,2_i4b) = merge( zero, xmiss, cnt(:m,j,k)>one )
                        xstat(:m,j,k,3_i4b) = xstat(:m,j,k,2_i4b)
                        xstat(:m,j,k,4_i4b) = xmiss
                        xstat(:m,j,k,5_i4b) = xmiss
!
                    end where
!
                end do
!
            end do
!
        else
!
            where ( xstat(:m,:p,:q,2_i4b)>zero ) 
!
!               COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
                xstat(:m,:p,:q,2_i4b) = xstat(:m,:p,:q,2_i4b)/( cnt - one )
                xstat(:m,:p,:q,3_i4b) = sqrt( xstat(:m,:p,:q,2_i4b) )
!
!               COMPUTE BIASED SKEWNESS AND KURTOSIS.
!
                xstat(:m,:p,:q,4_i4b) = xstat(:m,:p,:q,4_i4b)/( cnt*xstat(:m,:p,:q,3_i4b)**3 )
                xstat(:m,:p,:q,5_i4b) = xstat(:m,:p,:q,5_i4b)/( cnt*xstat(:m,:p,:q,2_i4b)**2 ) - three
!
            elsewhere
!
                xstat(:m,:p,:q,2_i4b) = merge( zero, xmiss, cnt>one )
                xstat(:m,:p,:q,3_i4b) = xstat(:m,:p,:q,2_i4b)
                xstat(:m,:p,:q,4_i4b) = xmiss
                xstat(:m,:p,:q,5_i4b) = xmiss
!
            end where
!
        end if
!
!       LESS THAN ONE VALID OBSERVATION, SET THE PERTINENT STATISTICS TO xmiss .
!
        where ( cnt<one )
!
            xstat(:m,:p,:q,1_i4b) = xmiss
            xstat(:m,:p,:q,6_i4b) = xmiss
            xstat(:m,:p,:q,7_i4b) = xmiss
!
        end where
!
!       OUTPUT NUMBERS OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) then
!
           call assert( logical(size(xnobs,1)==m,lgl),   &
                        logical(size(xnobs,2)==p,lgl),   &
                        name_proc )
!
           xnobs(:m,:p,:q) = int( cnt, i4b )
!
        end if
!
    else
!
!       THE CURRENT SUBMATRIX IS NOT THE LAST.
!       SAVE COUNTS OF NON-MISSING OBSERVATIONS FOR THE NEXT CALL.
!
        xstat(:m,:p,:q,3_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_unistat_miss_rf
! ______________________________________
!
    end subroutine comp_unistat_miss_rf
!
! =========================================================================================
!
    subroutine comp_mvs_rv( x, first, last, xmean, xvar, xstd, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS computes estimates of mean, variance and standard-deviation
!   from a data vector.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, input subvector containing size(X) observations
!          from the vector of data for which basic univariate statistics
!          are desired. If all the data are available at once, X can be
!          the full data vector.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subvector is the first subvector 
!            of the data vector.
!          - FIRST = false the current subvector is not the first subvector
!            of the data vector.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subvector is the last subvector 
!            of the data vector.
!          - LAST = false the current subvector is not the last subvector
!            of the data vector.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantity on previous calls
!          to COMP_MVS. XMEAN should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XMEAN contains the mean value of the data vector.
!
!   XVAR   (INPUT/OUTPUT) real(stnd)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantity on previous calls
!          to COMP_MVS. XVAR should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XVAR contains the variance of the data vector.
!
!   XSTD   (INPUT/OUTPUT) real(stnd)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantity on previous calls
!          to COMP_MVS. XSTD should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XSTD contains the standard deviation of the data vector.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data vector. XNOBS needs to be specified only on the
!          last call to COMP_MVS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid observation is present, the statistics are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:) :: x
    real(stnd), intent(inout)               :: xmean, xvar, xstd
!
    integer(i4b), intent(out), optional :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: di, cnt, tmp
!
    integer(i4b) :: i, n
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean = zero       
        xvar = zero       
        xstd = zero       
!
    end if
!
    cnt = xstd
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEAN, CORRECTED SUM OF SQUARES, HANDLING THE DATA ONCE.
!
        cnt = cnt + one
!
        di    = x(i) - xmean
        xstd  = di/cnt
        xmean = xmean + xstd
        xvar  = xvar + di*di - di*xstd
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
        if ( cnt>=one ) then
!
!           COMPUTE VARIANCE AND STANDARD-DEVIATION.
!
            xvar = xvar/cnt
            xstd = sqrt( xvar )
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO NaN CODE.
!
            tmp   = nan()
            xmean = tmp
            xvar  = tmp
            xstd  = tmp
!
        end if
!
    else
!
!       THE CURRENT SUBVECTOR IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstd = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_rv
! _____________________________
!
    end subroutine comp_mvs_rv
!
! =========================================================================================
!
    subroutine comp_mvs_rm( x, first, last, xmean, xvar, xstd, dimvar, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS computes estimates of means, variances and standard-deviations
!   from a data matrix.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:)
!          On entry, input submatrix containing size(X,3-DIMVAR)
!          observations on size(X,DIMVAR) variables from the matrix of data
!          for which basic univariate statistics are desired. By default, 
!          DIMVAR is equal to 1. See description of optional DIMVAR
!          argument for details.
!          If all the data are available at once, X can be the full 
!          data matrix.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current submatrix is the first submatrix 
!            of the data matrix.
!          - FIRST = false the current submatrix is not the first submatrix
!            of the data matrix.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current submatrix is the last submatrix 
!            of the data matrix.
!          - LAST = false the current submatrix is not the last submatrix
!            of the data matrix.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XMEAN should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XMEAN contains the mean values.
!
!          The size of XMEAN must verify size(XMEAN) = size(X,DIMVAR).
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XVAR should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XVAR contains the variances.
!
!          The size of XVAR must verify size(XVAR) = size(X,DIMVAR).
!
!   XSTD   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XSTD should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XSTD contains the standard deviations.
!
!          The size of XSTD must verify size(XSTD) = size(X,DIMVAR).
!
!   DIMVAR (INPUT, OPTIONAL) integer(i4b)
!          On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!          - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!            on size(X,1) variables.
!          - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!            on size(X,2) variables.
!
!          The default is DIMVAR = 1.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data matrix. XNOBS needs to be specified only on the
!          last call to COMP_MVS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid observation is present, the statistics are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)  :: x
    real(stnd), intent(inout), dimension(:)    :: xmean, xvar, xstd
!
    integer(i4b), intent(in),  optional  :: dimvar
    integer(i4b), intent(out), optional  :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: cnt, cntm1, tmp
    real(stnd), dimension(size(xmean)) :: tmpvec
!
    integer(i4b) :: i, n, m
!
    logical(lgl) :: dim_is_one
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
     if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
   m = assert_eq(  int(size(x,int(i)),i4b),     &
                    int(size(xmean),i4b),       &
                    int(size(xvar),i4b),        &
                    int(size(xstd),i4b),        &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    n = size( x, 3-int(i) )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean(:m) = zero       
        xvar(:m) = zero       
        xstd(:m) = zero       
!
    end if
!
    cnt = xstd(1_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt = cntm1 + one
!
        if ( dim_is_one ) then
            tmpvec(:m) = x(:m,i) - xmean(:m)
        else
            tmpvec(:m) = x(i,:m) - xmean(:m)
        end if
!
        xstd(:m)  = tmpvec(:m)*tmpvec(:m)
!
#ifdef _BLAS
        tmp = one/cnt
!
        call axpy( m, tmp, tmpvec(:m), 1_i4b, xmean(:m), 1_i4b )
!
        tmp = cntm1/cnt
!
        call axpy( m, tmp, xstd(:m), 1_i4b, xvar(:m), 1_i4b )
#else
        xmean(:m) = xmean(:m) + tmpvec(:m)*( one/cnt )
        xvar(:m)  = xvar(:m) + xstd(:m)*( cntm1/cnt )
#endif
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
        if ( cnt>=one ) then
!
!           COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
            xvar(:m) = xvar(:m)*( one/cnt )
            xstd(:m) = sqrt( xvar(:m) )
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO NaN CODE.
!
            tmp       = nan()
            xmean(:m) = tmp
            xvar(:m)  = tmp
            xstd(:m)  = tmp
!
        end if
!
    else
!
!       THE CURRENT SUBMATRIX IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstd(1_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_rm
! _____________________________
!
    end subroutine comp_mvs_rm
!
! =========================================================================================
!
    subroutine comp_mvs_rt( x, first, last, xmean, xvar, xstd, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS computes estimates of means, variances and standard-deviations
!   from a tridimensional data array.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:)
!          On entry, input subarray containing size(X,3) observations
!          on size(X,1) by size(X,2) variables from the array of data for which 
!          basic univariate statistics are desired.
!          If all the data are available at once, X can be the full data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!            of the data array.
!          - LAST = false the current subarray is not the last subarray
!            of the data array.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XMEAN should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XMEAN contains the mean values.
!
!          The shape of XMEAN must verify:
!
!          - size(XMEAN,1) = size(X,1)
!          - size(XMEAN,2) = size(X,2).
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XVAR should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XVAR contains the variances.
!
!          The shape of XVAR must verify:
!
!          - size(XVAR,1) = size(X,1)
!          - size(XVAR,2) = size(X,2).
!
!   XSTD   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XSTD should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XSTD contains the standard deviations.
!
!          The shape of XSTD must verify:
!
!          - size(XSTD,1) = size(X,1)
!          - size(XSTD,2) = size(X,2).
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data array. XNOBS needs to be specified only on the
!          last call to COMP_MVS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid observation is present, the statistics are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:)  :: x
    real(stnd), intent(inout), dimension(:,:)    :: xmean, xvar, xstd
!
    integer(i4b), intent(out), optional  :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: cnt, cntm1, tmp
    real(stnd), dimension(size(xmean,1)) :: tmpvec
!
    integer(i4b) :: i, j, n, m, p
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 3 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),      &
                    int(size(xmean,1),i4b),  &
                    int(size(xvar,1),i4b),   &
                    int(size(xstd,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),      &
                    int(size(xmean,2),i4b),  &
                    int(size(xvar,2),i4b),   &
                    int(size(xstd,2),i4b),   &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean(:m,:p) = zero       
        xvar(:m,:p) = zero       
        xstd(:m,:p) = zero       
!
    end if
!
    cnt = xstd(1_i4b,1_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt = cntm1 + one
!
        do j = 1_i4b, p
!
            tmpvec(:m)  = x(:m,j,i) - xmean(:m,j)
!
            xstd(:m,1_i4b) = tmpvec(:m)*tmpvec(:m)
!
#ifdef _BLAS
            tmp = one/cnt
!
            call axpy( m, tmp, tmpvec(:m), 1_i4b, xmean(:m,j), 1_i4b )
!
            tmp = cntm1/cnt
!
            call axpy( m, tmp, xstd(:m,1_i4b), 1_i4b, xvar(:m,j), 1_i4b )
#else
            xmean(:m,j) = xmean(:m,j) + tmpvec(:m)*( one/cnt )
            xvar(:m,j)  = xvar(:m,j) + xstd(:m,1_i4b)*( cntm1/cnt )
#endif
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
        if ( cnt>=one ) then
!
!           COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
            xvar(:m,:p) = xvar(:m,:p)*( one/cnt )
            xstd(:m,:p) = sqrt( xvar(:m,:p) )
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO NaN CODE.
!
            tmp          = nan()
            xmean(:m,:p) = tmp
            xvar(:m,:p)  = tmp
            xstd(:m,:p)  = tmp
!
        end if
!
    else
!
!       THE CURRENT SUBARRAY IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstd(1_i4b,1_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_rt
! _____________________________
!
    end subroutine comp_mvs_rt
!
! =========================================================================================
!
    subroutine comp_mvs_rf( x, first, last, xmean, xvar, xstd, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS computes estimates of means, variances and standard-deviations
!   from a fourdimensional data array.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:,:)
!          On entry, input subarray containing size(X,4) observations
!          on size(X,1) by size(X,2) by size(X,3) variables from the array of data for which 
!          basic univariate statistics are desired.
!          If all the data are available at once, X can be the full data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!            of the data array.
!          - LAST = false the current subarray is not the last subarray
!            of the data array.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XMEAN should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XMEAN contains the mean values.
!
!          The shape of XMEAN must verify:
!
!          - size(XMEAN,1) = size(X,1)
!          - size(XMEAN,2) = size(X,2)
!          - size(XMEAN,3) = size(X,3).
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XVAR should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XVAR contains the variances.
!
!          The shape of XVAR must verify:
!
!          - size(XVAR,1) = size(X,1)
!          - size(XVAR,2) = size(X,2)
!          - size(XVAR,3) = size(X,3).
!
!   XSTD   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, after the first call to COMP_MVS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS. XSTD should not be changed between calls to
!          COMP_MVS.
!
!          On exit, when LAST=true, XSTD contains the standard deviations.
!
!          The shape of XSTD must verify: 
!
!          - size(XSTD,1) = size(X,1)
!          - size(XSTD,2) = size(X,2)
!          - size(XSTD,3) = size(X,3).
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of observations in
!          the data array. XNOBS needs to be specified only on the
!          last call to COMP_MVS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid observation is present, the statistics are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:,:)  :: x
    real(stnd), intent(inout), dimension(:,:,:)    :: xmean, xvar, xstd
!
    integer(i4b), intent(out), optional  :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: cnt, cntm1, tmp
    real(stnd), dimension(size(xmean,1)) :: tmpvec
!
    integer(i4b) :: i, j, k, n, m, p, q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 4 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),      &
                    int(size(xmean,1),i4b),  &
                    int(size(xvar,1),i4b),   &
                    int(size(xstd,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),      &
                    int(size(xmean,2),i4b),  &
                    int(size(xvar,2),i4b),   &
                    int(size(xstd,2),i4b),   &
                    name_proc )
!
    q = assert_eq(  int(size(x,3),i4b),      &
                    int(size(xmean,3),i4b),  &
                    int(size(xvar,3),i4b),   &
                    int(size(xstd,3),i4b),   &
                    name_proc )
!
    if ( min(m,p,q)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean(:m,:p,:q) = zero       
        xvar(:m,:p,:q) = zero       
        xstd(:m,:p,:q) = zero       
!
    end if
!
    cnt = xstd(1_i4b,1_i4b,1_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        cntm1 = cnt
        cnt = cntm1 + one
!
        do k = 1_i4b, q
!
            do j = 1_i4b, p
!
                tmpvec(:m) = x(:m,j,k,i) - xmean(:m,j,k)
!
                xstd(:m,1_i4b,1_i4b) = tmpvec(:m)*tmpvec(:m)
!
#ifdef _BLAS
                tmp = one/cnt
!
                call axpy( m, tmp, tmpvec(:m), 1_i4b, xmean(:m,j,k), 1_i4b )
!
                tmp = cntm1/cnt
!
                call axpy( m, tmp, xstd(:m,1_i4b,1_i4b), 1_i4b, xvar(:m,j,k), 1_i4b )
#else
                xmean(:m,j,k) = xmean(:m,j,k) + tmpvec(:m)*( one/cnt )
                xvar(:m,j,k) = xvar(:m,j,k) + xstd(:m,1_i4b,1_i4b)*( cntm1/cnt )
#endif
!
            end do
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       OUTPUT NUMBER OF OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( cnt, i4b )
!
        if ( cnt>=one ) then
!
!           COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
            xvar(:m,:p,:q) = xvar(:m,:p,:q)*( one/cnt )
            xstd(:m,:p,:q) = sqrt( xvar(:m,:p,:q) )
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO NaN CODE.
!
            tmp             = nan()
            xmean(:m,:p,:q) = tmp
            xvar(:m,:p,:q)  = tmp
            xstd(:m,:p,:q)  = tmp
!
        end if
!
    else
!
!       THE CURRENT SUBARRAY IS NOT THE LAST.
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xstd(1_i4b,1_i4b,1_i4b) = cnt
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_rf
! _____________________________
!
    end subroutine comp_mvs_rf
!
! =========================================================================================
!
    subroutine comp_mvs_miss_rv( x, first, last, xmean, xvar, xstd, xmiss, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS_MISS computes estimates of mean, variance and standard-deviation
!   from a data vector possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, input subvector containing size(X) observations
!          from the vector of data for which basic univariate statistics
!          are desired. If all the data are available at once, X can be
!          the full data vector.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subvector is the first subvector 
!            of the data vector.
!          - FIRST = false the current subvector is not the first subvector
!            of the data vector.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subvector is the last subvector 
!            of the data vector.
!          - LAST = false the current subvector is not the last subvector
!            of the data vector.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantity on previous calls
!          to COMP_MVS_MISS. XMEAN should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XMEAN contains the mean value of the data vector.
!
!   XVAR   (INPUT/OUTPUT) real(stnd)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantity on previous calls
!          to COMP_MVS_MISS. XVAR should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XVAR contains the variance of the data vector.
!
!   XSTD   (INPUT/OUTPUT) real(stnd)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantity on previous calls
!          to COMP_MVS_MISS. XSTD should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XSTD contains the standard deviation of the data vector.
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b)
!          On exit, XNOBS contains the number of non-missing observations in
!          the data vector. XNOBS needs to be specified only on the last call
!          to COMP_MVS_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid (non-missing) observation is present, the pertinent statistics
!   are set to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:) :: x
    real(stnd), intent(in)                  :: xmiss
    real(stnd), intent(inout)               :: xmean, xvar, xstd
!
    integer(i4b), intent(out), optional :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: di, dfi
!
    integer(i4b) :: i, n
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean = zero       
        xvar = zero       
        xstd = zero       
!
    end if
!
    do i = 1_i4b, n
!
        if ( x(i)==xmiss ) cycle
!
!       CALCULATE PARTIAL MEAN, CORRECTED SUM OF SQUARES, HANDLING THE DATA ONCE.
!
        xstd = xstd + one
!
        di    = x(i) - xmean
        dfi   = di/xstd
        xmean = xmean + dfi
        xvar  = xvar + di*di - di*dfi
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       OUTPUT NUMBER OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
        if ( present(xnobs) ) xnobs = int( xstd, i4b )
!
        if ( xstd>=one ) then
!
!           COMPUTE VARIANCE AND STANDARD-DEVIATION.
!
            xvar = xvar/xstd
            xstd = sqrt( xvar )
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO xmiss .
!
            xmean = xmiss
            xvar = xmiss
            xstd = xmiss
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_miss_rv
! __________________________________
!
    end subroutine comp_mvs_miss_rv
!
! =========================================================================================
!
    subroutine comp_mvs_miss_rm( x, first, last, xmean, xvar, xstd, xmiss, dimvar, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS_MISS computes estimates of means, variances and standard-deviations
!   from a data matrix possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:)
!          On entry, input submatrix containing size(X,3-DIMVAR)
!          observations on size(X,DIMVAR) variables from the matrix of data
!          for which basic univariate statistics are desired. By default, 
!          DIMVAR is equal to 1. See description of optional DIMVAR
!          argument for details.
!          If all the data are available at once, X can be the full 
!          data matrix.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current submatrix is the first submatrix 
!            of the data matrix.
!          - FIRST = false the current submatrix is not the first submatrix
!            of the data matrix.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current submatrix is the last submatrix 
!            of the data matrix.
!          - LAST = false the current submatrix is not the last submatrix
!            of the data matrix.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XMEAN should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XMEAN contains the mean values.
!
!          The size of XMEAN must verify size(XMEAN) = size(X,DIMVAR).
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XVAR should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XVAR contains the variances.
!
!          The size of XVAR must verify size(XVAR) = size(X,DIMVAR).
!
!   XSTD   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XSTD should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XSTD contains the standard deviations.
!
!          The size of XSTD must verify size(XSTD) = size(X,DIMVAR).
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   DIMVAR (INPUT, OPTIONAL) integer(i4b)
!          On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!          - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!            on size(X,1) variables.
!          - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!            on size(X,2) variables.
!
!          The default is DIMVAR = 1.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b), dimension(:)
!          On exit, XNOBS contains the numbers of non-missing observations
!          on all variables. XNOBS needs to be specified only on the
!          last call to COMP_MVS_MISS (e.g. when LAST=true).
!
!          The size of XNOBS must verify size(XNOBS) = size(X,DIMVAR).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid (non-missing) observation is present for some variables,
!   the pertinent statistics are set to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)  :: x
    real(stnd), intent(in)                     :: xmiss
    real(stnd), intent(inout), dimension(:)    :: xmean, xvar, xstd
!
    integer(i4b), intent(in),                  optional :: dimvar
    integer(i4b), intent(out), dimension(:),   optional :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xmean,1)) :: di, dfi
!
    integer(i4b) :: i, n, m
!
    logical(lgl) :: dim_is_one
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
     if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xmean),i4b),       &
                    int(size(xvar),i4b),        &
                    int(size(xstd),i4b),        &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    n = size( x, 3-int(i) )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean(:m) = zero       
        xvar(:m) = zero       
        xstd(:m) = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dfi(:m) = x(:m,i)
        else
            dfi(:m) = x(i,:m)
        end if
!
        where ( dfi(:m)/=xmiss )
!
            xstd(:m) = xstd(:m) + one
!
            di(:m)    = dfi(:m) - xmean(:m)
            dfi(:m)   = di(:m)/xstd(:m)
            xmean(:m) = xmean(:m) + dfi(:m)
            xvar(:m)  = xvar(:m) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
        end where
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        if ( present(xnobs) ) then
!
!           OUTPUT NUMBERS OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
            call assert( logical(size(xnobs)==m,lgl), name_proc )
            xnobs(:m) = int( xstd(:m), i4b )
!
        end if
!
        where ( xstd(:m)>=one )
!
!           COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
            xvar(:m) = xvar(:m)/xstd(:m)
            xstd(:m) = sqrt( xvar(:m) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO xmiss .
!
            xmean(:m) = xmiss
            xvar(:m) = xmiss
            xstd(:m) = xmiss
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_miss_rm
! __________________________________
!
    end subroutine comp_mvs_miss_rm
!
! =========================================================================================
!
    subroutine comp_mvs_miss_rt( x, first, last, xmean, xvar, xstd, xmiss, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS_MISS computes estimates of means, variances and standard-deviations
!   from a tridimensional data array possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:)
!          On entry, input subarray containing size(X,3) observations
!          on size(X,1) by size(X,2) variables from the array of data for which 
!          basic univariate statistics are desired.
!          If all the data are available at once, X can be the full data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!            of the data array.
!          - LAST = false the current subarray is not the last subarray
!            of the data array.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XMEAN should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XMEAN contains the mean values.
!
!          The shape of XMEAN must verify:
!
!          - size(XMEAN,1) = size(X,1)
!          - size(XMEAN,2) = size(X,2).
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XVAR should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XVAR contains the variances.
!
!          The shape of XVAR must verify:
!
!          - size(XVAR,1) = size(X,1)
!          - size(XVAR,2) = size(X,2).
!
!   XSTD   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XSTD should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XSTD contains the standard deviations.
!
!          The shape of XSTD must verify:
!
!          - size(XSTD,1) = size(X,1)
!          - size(XSTD,2) = size(X,2).
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b), dimension(:,:)
!          On exit, XNOBS contains the numbers of non-missing observations
!          on all variables. XNOBS needs to be specified only on the
!          last call to COMP_MVS_MISS (e.g. when LAST=true).
!
!          The shape of XNOBS must verify:
!
!          - size(XNOBS,1) = size(X,1)
!          - size(XNOBS,2) = size(X,2).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid (non-missing) observation is present for some variables,
!   the pertinent statistics are set to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:)  :: x
    real(stnd), intent(in)                       :: xmiss
    real(stnd), intent(inout), dimension(:,:)    :: xmean, xvar, xstd
!
    integer(i4b), intent(out), dimension(:,:),   optional :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(x,1),size(x,2)) :: di, dfi
!
    integer(i4b) :: i, j, n, m, p
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 3 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),      &
                    int(size(xmean,1),i4b),  &
                    int(size(xvar,1),i4b),   &
                    int(size(xstd,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),      &
                    int(size(xmean,2),i4b),  &
                    int(size(xvar,2),i4b),   &
                    int(size(xstd,2),i4b),   &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean(:m,:p) = zero       
        xvar(:m,:p) = zero       
        xstd(:m,:p) = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        do j = 1_i4b, p
!
            where ( x(:m,j,i)/=xmiss )
!
                xstd(:m,j) = xstd(:m,j) + one
!
                di(:m,j)    = x(:m,j,i) - xmean(:m,j)
                dfi(:m,j)   = di(:m,j)/xstd(:m,j)
                xmean(:m,j) = xmean(:m,j) + dfi(:m,j)
                xvar(:m,j)  = xvar(:m,j) + di(:m,j)*di(:m,j) - di(:m,j)*dfi(:m,j)
!
            end where
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        if ( present(xnobs) ) then
!
!           OUTPUT NUMBER OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
            call assert( logical(size(xnobs,1)==m,lgl),    &
                         logical(size(xnobs,2)==p,lgl),    &
                         name_proc )
!
            xnobs(:m,:p) = int( xstd(:m,:p), i4b )
!
        end if
!
        where ( xstd(:m,:p)>=one )
!
!           COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
            xvar(:m,:p) = xvar(:m,:p)/xstd(:m,:p)
            xstd(:m,:p) = sqrt( xvar(:m,:p) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO xmiss .
!
            xmean(:m,:p) = xmiss
            xvar(:m,:p) = xmiss
            xstd(:m,:p) = xmiss
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_miss_rt
! __________________________________
!
    end subroutine comp_mvs_miss_rt
!
! =========================================================================================
!
    subroutine comp_mvs_miss_rf( x, first, last, xmean, xvar, xstd, xmiss, xnobs )
!
! Purpose
! _______
!
!   COMP_MVS_MISS computes estimates of means, variances and standard-deviations
!   from a fourdimensional data array possibly containing missing values.
!
!
! Arguments
! _________
!
!   X      (INPUT) real(stnd), dimension(:,:,:,:)
!          On entry, input subarray containing size(X,4) observations
!          on size(X,1) by size(X,2) by size(X,3) variables from the array of data for which 
!          basic univariate statistics are desired.
!          If all the data are available at once, X can be the full data array.
!
!   FIRST  (INPUT) logical(lgl)
!          On entry, if:
!
!          - FIRST = true  the current subarray is the first subarray 
!            of the data array.
!          - FIRST = false the current subarray is not the first subarray
!            of the data array.
!
!   LAST   (INPUT) logical(lgl)
!          On entry, if:
!
!          - LAST = true  the current subarray is the last subarray 
!            of the data array.
!          - LAST = false the current subarray is not the last subarray
!            of the data array.
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XMEAN is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XMEAN should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XMEAN contains the mean values.
!
!          The shape of XMEAN must verify:
!
!          - size(XMEAN,1) = size(X,1)
!          - size(XMEAN,2) = size(X,2)
!          - size(XMEAN,3) = size(X,3).
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XVAR is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XVAR should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XVAR contains the variances.
!
!          The shape of XVAR must verify:
!
!          - size(XVAR,1) = size(X,1)
!          - size(XVAR,2) = size(X,2)
!          - size(XVAR,3) = size(X,3).
!
!   XSTD   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, after the first call to COMP_MVS_MISS (e.g. when FIRST=true),
!          XSTD is used as workspace to accumulate quantities on previous calls
!          to COMP_MVS_MISS. XSTD should not be changed between calls to
!          COMP_MVS_MISS.
!
!          On exit, when LAST=true, XSTD contains the standard deviations.
!
!          The shape of XSTD must verify:
!
!          - size(XSTD,1) = size(X,1)
!          - size(XSTD,2) = size(X,2)
!          - size(XSTD,3) = size(X,3).
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XNOBS  (OUTPUT, OPTIONAL) integer(i4b), dimension(:,:,:)
!          On exit, XNOBS contains the numbers of non-missing observations
!          on all variables. XNOBS needs to be specified only on the
!          last call to COMP_MVS_MISS (e.g. when LAST=true).
!
!          The shape of XNOBS must verify:
!
!          - size(XNOBS,1) = size(X,1)
!          - size(XNOBS,2) = size(X,2)
!          - size(XNOBS,3) = size(X,3).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic statistics with only one pass through the data.
!
!   If fewer than one valid (non-missing) observation is present for some variables,
!   the pertinent statistics are set to XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:,:)  :: x
    real(stnd), intent(in)                         :: xmiss
    real(stnd), intent(inout), dimension(:,:,:)    :: xmean, xvar, xstd
!
    integer(i4b), intent(out), dimension(:,:,:),   optional :: xnobs
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(x,1),size(x,2),size(x,3)) :: di, dfi
!
    integer(i4b) :: i, j, k, n, m, p, q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x, 4 )
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),      &
                    int(size(xmean,1),i4b),  &
                    int(size(xvar,1),i4b),   &
                    int(size(xstd,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),      &
                    int(size(xmean,2),i4b),  &
                    int(size(xvar,2),i4b),   &
                    int(size(xstd,2),i4b),   &
                    name_proc )
!
    q = assert_eq(  int(size(x,3),i4b),      &
                    int(size(xmean,3),i4b),  &
                    int(size(xvar,3),i4b),   &
                    int(size(xstd,3),i4b),   &
                    name_proc )
!
    if ( min(m,p,q)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean, xvar, xstd .
!
        xmean(:m,:p,:q) = zero       
        xvar(:m,:p,:q) = zero       
        xstd(:m,:p,:q) = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        do k = 1_i4b, q
!
            do j = 1_i4b, p
!
                where ( x(:m,j,k,i)/=xmiss )
!
                    xstd(:m,j,k) = xstd(:m,j,k) + one
!
                    di(:m,j,k)    = x(:m,j,k,i) - xmean(:m,j,k)
                    dfi(:m,j,k)   = di(:m,j,k)/xstd(:m,j,k)
                    xmean(:m,j,k) = xmean(:m,j,k) + dfi(:m,j,k)
                    xvar(:m,j,k)  = xvar(:m,j,k) + di(:m,j,k)*di(:m,j,k) - di(:m,j,k)*dfi(:m,j,k)
!
                end where
!
            end do
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        if ( present(xnobs) ) then
!
!           OUTPUT NUMBER OF NON-MISSING OBSERVATIONS, IF REQUIRED.
!
            call assert( logical(size(xnobs,1)==m,lgl),    &
                         logical(size(xnobs,2)==p,lgl),    &
                         logical(size(xnobs,3)==q,lgl),    &
                         name_proc )
!
            xnobs(:m,:p,:q) = int( xstd(:m,:p,:q), i4b )
!
        end if
!
        where ( xstd(:m,:p,:q)>=one )
!
!           COMPUTE VARIANCES AND STANDARD-DEVIATIONS.
!
            xvar(:m,:p,:q) = xvar(:m,:p,:q)/xstd(:m,:p,:q)
            xstd(:m,:p,:q) = sqrt( xvar(:m,:p,:q) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO xmiss .
!
            xmean(:m,:p,:q) = xmiss
            xvar(:m,:p,:q) = xmiss
            xstd(:m,:p,:q) = xmiss
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_miss_rf
! __________________________________
!
    end subroutine comp_mvs_miss_rf
!
!
! =========================================================================================
!                          UNIVARIATE STATISTICS BY GROUPS
! =========================================================================================
!
!
    subroutine comp_mvs_grp_rv( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp )
!
! Purpose
! _______
!
!   COMP_MVS_GRP computes estimates of univariate statistics by groups 
!   from a data vector.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:)
!              On entry, input subvector containing size(X) observations from
!              the vector of data for which univariate statistics by groups
!              are desired. If all the data are available at once, X can be
!              the full data vector.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subvector is the first subvector 
!                              of the data vector.
!              - FIRST = false the current subvector is not the first subvector
!                              of the data vector.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subvector is the last subvector 
!                             of the data vector.
!              - LAST = false the current subvector is not the last subvector
!                             of the data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data vector.
!
!              The size of XMEAN_GRP must verify size(XMEAN_GRP) = NGRP .
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations in the data vector.
!
!              The size of XSTD_GRP must verify size(XSTD_GRP) = NGRP .
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XN_GRP contains counts of observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of observations
!              in the NGRP groups of observations in the data vector.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one observation is present for some groups, the pertinent statistics 
!   are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
    use Utilities,         only : assert, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:)  :: x
    real(stnd), intent(inout), dimension(:)  :: xn_grp, xmean_grp, xstd_grp 
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: di, dfi
!
    integer(i4b) :: i, k, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),         &
                    int(size(ind),i4b),       &
                    name_proc )
!
    call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                 logical(size(xstd_grp)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:ngrp) = zero       
        xstd_grp(:ngrp)  = zero       
        xn_grp(:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        xn_grp(k) = xn_grp(k) + one
        di  = x(i) - xmean_grp(k)
        dfi = di/xn_grp(k)
!
        xmean_grp(k) = xmean_grp(k) + dfi
        xstd_grp(k)  = xstd_grp(k) + di*di - di*dfi
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        di = nan()
!
        where ( xn_grp(:ngrp)/=zero )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            xstd_grp(:ngrp) = sqrt( xstd_grp(:ngrp)/xn_grp(:ngrp) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            xmean_grp(:ngrp) = di
            xstd_grp(:ngrp)  = di
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_rv
! _________________________________
!
    end subroutine comp_mvs_grp_rv
!
! =========================================================================================
!
    subroutine comp_mvs_grp_rm( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp,   &
                                dimvar )
!
! Purpose
! _______
!
!   COMP_MVS_GRP computes estimates of univariate statistics by groups 
!   from a data matrix.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:)
!              On entry, input submatrix containing size(X,3-DIMVAR)
!              observations on size(X,DIMVAR) variables from the matrix of data
!              for which univariate statistics by groups are desired. By default, 
!              DIMVAR is equal to 1. See description of optional DIMVAR
!              argument for details.
!              If all the data are available at once, X can be the full 
!              data matrix.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current submatrix is the first submatrix 
!                              of the data matrix.
!              - FIRST = false the current submatrix is not the first submatrix
!                              of the data matrix.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current submatrix is the last submatrix 
!                             of the data matrix.
!              - LAST = false the current submatrix is not the last submatrix
!                             of the data matrix.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3-DIMVAR) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3-DIMVAR) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data matrix.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,DIMVAR)
!              - size(XMEAN_GRP,2) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data matrix.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,DIMVAR)
!              - size(XSTD_GRP,2) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XN_GRP contains counts of observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of observations
!              in the NGRP groups for all the variables in the data matrix.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!   DIMVAR     (INPUT, OPTIONAL) integer(i4b)
!              On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!              - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!                on size(X,1) variables.
!              - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!                on size(X,2) variables.
!
!              The default is DIMVAR = 1.
!
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one observation is present for some groups, the pertinent statistics 
!   are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),               dimension(:,:) :: x
    real(stnd), intent(inout),            dimension(:)   :: xn_grp
    real(stnd), intent(inout),            dimension(:,:) :: xmean_grp, xstd_grp
!
    integer(i4b), intent(in)                          :: ngrp
    integer(i4b), intent(in),            dimension(:) :: ind
    integer(i4b), intent(in),  optional               :: dimvar
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: xnm1, tmp
    real(stnd), dimension(size(xmean_grp,1))    :: di, di2
!
    integer(i4b) :: i, k, n, m
!
    logical(lgl) :: dim_is_one
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(ind),i4b),             &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                 logical(size(xstd_grp,2)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),       &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:m,:ngrp) = zero       
        xstd_grp(:m,:ngrp)  = zero       
        xn_grp(:ngrp)       = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        xnm1      = xn_grp(k)
        xn_grp(k) = xnm1 + one
!
        if ( dim_is_one )  then
            di(:m) = x(:m,i) - xmean_grp(:m,k)
        else
            di(:m) = x(i,:m) - xmean_grp(:m,k)
        end if
!
        di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
        tmp = one/xn_grp(k)
!
        call axpy( m, tmp, di(:m), 1_i4b, xmean_grp(:m,k), 1_i4b )
!
        tmp = xnm1/xn_grp(k)
!
        call axpy( m, tmp, di2(:m), 1_i4b, xstd_grp(:m,k), 1_i4b )
#else
        xmean_grp(:m,k) = xmean_grp(:m,k) + di(:m)*( one/xn_grp(k) )
        xstd_grp(:m,k)  = xstd_grp(:m,k) + di2(:m)*( xnm1/xn_grp(k) )
#endif
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        tmp = nan()
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        k = omp_get_max_threads()
        test_par = .not.( omp_in_parallel() )      .and.      &
                   ngrp>=2_i4b                     .and.      &
                   i>1_i4b                         .and.      &
                   k>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC)          &
!$OMP            ,PRIVATE(k)
!
        do k = 1_i4b, ngrp
!
            if ( xn_grp(k)/=zero ) then
!
!               COMPUTE STANDARD-DEVIATIONS.
!
                xstd_grp(:m,k) = sqrt( xstd_grp(:m,k)*(one/xn_grp(k)) )
!
            else
!
!               FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO MISSING.
!
                xmean_grp(:m,k) = tmp
                xstd_grp(:m,k)  = tmp
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_rm
! _________________________________
!
    end subroutine comp_mvs_grp_rm
!
! =========================================================================================
!
    subroutine comp_mvs_grp_rt( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp )
!
! Purpose
! _______
!
!   COMP_MVS_GRP computes estimates of univariate statistics by groups from a data 
!   tridimensional array.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:,:)
!              On entry, input subarray containing size(X,3) observations
!              on size(X,1) by size(X,2) variables from the array of data for 
!              which univariate statistics by groups are desired. If all the data
!              are available at once, X can be the full data array.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subarray is the first subarray 
!                              of the data array.
!              - FIRST = false the current subarray is not the first subarray
!                              of the data array.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subarray is the last subarray 
!                             of the data array.
!              - LAST = false the current subarray is not the last subarray
!                             of the data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1)
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XN_GRP contains counts of  observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of observations  
!              in the NGRP groups for all the variables in the data array.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one observation is present for some groups, the pertinent statistics 
!   are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),         dimension(:,:,:) :: x
    real(stnd), intent(inout),      dimension(:)     :: xn_grp
    real(stnd), intent(inout),      dimension(:,:,:) :: xmean_grp, xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: xnm1, tmp
    real(stnd), dimension(size(x,1)) :: di, di2
!
    integer(i4b) :: i, j, k, n, m, p
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(size(x,1),i4b),            &
                    int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),            &
                    int(size(xmean_grp,2),i4b),    &
                    int(size(xstd_grp,2),i4b),     &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                 logical(size(xstd_grp,3)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),       &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:m,:p,:ngrp) = zero       
        xstd_grp(:m,:p,:ngrp)  = zero       
        xn_grp(:ngrp)          = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        xnm1      = xn_grp(k)
        xn_grp(k) = xnm1 + one
!
        do j = 1_i4b, p
!
            di(:m) = x(:m,j,i) - xmean_grp(:m,j,k)
!
            di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
            tmp = one/xn_grp(k)
!
            call axpy( m, tmp, di(:m), 1_i4b, xmean_grp(:m,j,k), 1_i4b )
!
            tmp = xnm1/xn_grp(k)
!
            call axpy( m, tmp, di2(:m), 1_i4b, xstd_grp(:m,j,k), 1_i4b )
#else
            xmean_grp(:m,j,k) = xmean_grp(:m,j,k) + di(:m)*( one/xn_grp(k) )
            xstd_grp(:m,j,k)  = xstd_grp(:m,j,k) + di2(:m)*( xnm1/xn_grp(k) )
#endif
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        tmp = nan()
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        k = omp_get_max_threads()
        test_par = .not.( omp_in_parallel() )      .and.      &
                   ngrp>=2_i4b                     .and.      &
                   i>1_i4b                         .and.      &
                   k>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC)          &
!$OMP            ,PRIVATE(k)
!
        do k = 1_i4b, ngrp
!
            if ( xn_grp(k)/=zero ) then
!
!               COMPUTE STANDARD-DEVIATIONS.
!
                xstd_grp(:m,:p,k) = sqrt( xstd_grp(:m,:p,k)*(one/xn_grp(k)) )
!
            else
!
!               FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
                xmean_grp(:m,:p,k) = tmp
                xstd_grp(:m,:p,k)  = tmp
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_rt
! _________________________________
!
    end subroutine comp_mvs_grp_rt
!
! =========================================================================================
!
    subroutine comp_mvs_grp_rf( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp )
!
! Purpose
! _______
!
!   COMP_MVS_GRP computes estimates of univariate statistics by groups from a data 
!   fourdimensional array.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:,:,:)
!              On entry, input subarray containing size(X,4) observations
!              on size(X,1) by size(X,2)  by size(X,3) variables from the array of data for 
!              which univariate statistics by groups are desired. If all the data
!              are available at once, X can be the full data array.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subarray is the first subarray 
!                              of the data array.
!              - FIRST = false the current subarray is not the first subarray
!                              of the data array.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subarray is the last subarray 
!                             of the data array.
!              - LAST = false the current subarray is not the last subarray
!                             of the data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,4) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,4) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = size(X,3)
!              - size(XMEAN_GRP,4) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1),
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = size(X,3)
!              - size(XSTD_GRP,4) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP (e.g. when FIRST=true),
!              XN_GRP contains counts of  observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of observations  
!              in the NGRP groups for all the variables in the data array.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one observation is present for some groups, the pertinent statistics 
!   are set to Nan code.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),         dimension(:,:,:,:) :: x
    real(stnd), intent(inout),      dimension(:)       :: xn_grp
    real(stnd), intent(inout),      dimension(:,:,:,:) :: xmean_grp, xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: xnm1, tmp
    real(stnd), dimension(size(x,1)) :: di, di2
!
    integer(i4b) :: i, j, k, l, n, m, p, q
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,4),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(size(x,1),i4b),            &
                    int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),            &
                    int(size(xmean_grp,2),i4b),    &
                    int(size(xstd_grp,2),i4b),     &
                    name_proc )
!
    q = assert_eq(  int(size(x,3),i4b),            &
                    int(size(xmean_grp,3),i4b),    &
                    int(size(xstd_grp,3),i4b),     &
                    name_proc )
!
    if ( min(m,p,q)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,4)==ngrp,lgl),  &
                 logical(size(xstd_grp,4)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),       &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:m,:p,:q,:ngrp) = zero       
        xstd_grp(:m,:p,:q,:ngrp)  = zero       
        xn_grp(:ngrp)             = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        xnm1      = xn_grp(k)
        xn_grp(k) = xnm1 + one
!
        do l = 1_i4b, q
!
            do j = 1_i4b, p
!
                di(:m) = x(:m,j,l,i) - xmean_grp(:m,j,l,k)
!
                di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
                tmp = one/xn_grp(k)
!
                call axpy( m, tmp, di(:m), 1_i4b, xmean_grp(:m,j,l,k), 1_i4b )
!
                tmp = xnm1/xn_grp(k)
!
                call axpy( m, tmp, di2(:m), 1_i4b, xstd_grp(:m,j,l,k), 1_i4b )
#else
                xmean_grp(:m,j,l,k) = xmean_grp(:m,j,l,k) + di(:m)*( one/xn_grp(k) )
                xstd_grp(:m,j,l,k)  = xstd_grp(:m,j,l,k) + di2(:m)*( xnm1/xn_grp(k) )
#endif
!
            end do
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        tmp = nan()
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        k = omp_get_max_threads()
        test_par = .not.( omp_in_parallel() )      .and.      &
                   ngrp>=2_i4b                     .and.      &
                   i>1_i4b                         .and.      &
                   k>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC)          &
!$OMP            ,PRIVATE(k)
!
        do k = 1_i4b, ngrp
!
            if ( xn_grp(k)/=zero ) then
!
!               COMPUTE STANDARD-DEVIATIONS.
!
                xstd_grp(:m,:p,:q,k) = sqrt( xstd_grp(:m,:p,:q,k)*(one/xn_grp(k)) )
!
            else
!
!               FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
                xmean_grp(:m,:p,:q,k) = tmp
                xstd_grp(:m,:p,:q,k)  = tmp
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_rf
! _________________________________
!
    end subroutine comp_mvs_grp_rf
!
! =========================================================================================
!
    subroutine comp_mvs_grp_miss_rv( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp, xmiss )
!
! Purpose
! _______
!
!   COMP_MVS_GRP_MISS computes estimates of univariate statistics by groups 
!   from a data vector possibly containing missing values.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:)
!              On entry, input subvector containing size(X) observations from
!              the vector of data for which univariate statistics by groups
!              are desired. If all the data are available at once, X can be
!              the full data vector.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subvector is the first subvector 
!                              of the data vector.
!              - FIRST = false the current subvector is not the first subvector
!                              of the data vector.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subvector is the last subvector 
!                             of the data vector.
!              - LAST = false the current subvector is not the last subvector
!                             of the data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP_MISS. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data vector.
!
!              The size of XMEAN_GRP must verify size(XMEAN_GRP) = NGRP .
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP_MISS. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations in the data vector.
!
!              The size of XSTD_GRP must verify size(XSTD_GRP) = NGRP .
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XN_GRP contains counts of non-missing observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP_MISS. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP_MISS.
!
!              On exit, XN_GRP contains the number of non-missing observations
!              in the NGRP groups of observations in the data vector.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one valid observation were present for some groups
!   of observations, the pertinent statistics are set to missing (XMISS value).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Utilities,         only : assert, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                    :: xmiss
    real(stnd), intent(in),     dimension(:)  :: x
    real(stnd), intent(inout),  dimension(:)  :: xn_grp, xmean_grp, xstd_grp 
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: di, dfi
!
    integer(i4b) :: i, k, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),         &
                    int(size(ind),i4b),       &
                    name_proc )
!
    call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                 logical(size(xstd_grp)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:ngrp) = zero       
        xstd_grp(:ngrp)  = zero       
        xn_grp(:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp .or. x(i)==xmiss ) cycle
!
        xn_grp(k) = xn_grp(k) + one
        di        = x(i) - xmean_grp(k)
        dfi       = di/xn_grp(k)
!
        xmean_grp(k) = xmean_grp(k) + dfi
        xstd_grp(k)  = xstd_grp(k) + di*di - di*dfi
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
        where ( xn_grp(:ngrp)/=zero )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            xstd_grp(:ngrp) = sqrt( xstd_grp(:ngrp)/xn_grp(:ngrp) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            xmean_grp(:ngrp) = xmiss
            xstd_grp(:ngrp)  = xmiss
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_miss_rv
! ______________________________________
!
    end subroutine comp_mvs_grp_miss_rv
!
! =========================================================================================
!
    subroutine comp_mvs_grp_miss_rm( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp, xmiss,    &
                                     dimvar )
!
! Purpose
! _______
!
!   COMP_MVS_GRP_MISS computes estimates of univariate statistics by groups
!   from a data matrix possibly containing missing values.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:)
!              On entry, input submatrix containing size(X,3-DIMVAR)
!              observations on size(X,DIMVAR) variables from the matrix of data
!              for which univariate statistics by groups are desired. By default, 
!              DIMVAR is equal to 1. See description of optional DIMVAR
!              argument for details.
!              If all the data are available at once, X can be the full 
!              data matrix.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current submatrix is the first submatrix 
!                              of the data matrix.
!              - FIRST = false the current submatrix is not the first submatrix
!                              of the data matrix.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current submatrix is the last submatrix 
!                             of the data matrix.
!              - LAST = false the current submatrix is not the last submatrix
!                             of the data matrix.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3-DIMVAR) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3-DIMVAR) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP_MISS. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data matrix.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,DIMVAR)
!              - size(XMEAN_GRP,2) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP_MISS. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data matrix.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,DIMVAR)
!              - size(XSTD_GRP,2) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XN_GRP contains counts of non-missing observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP_MISS. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP_MISS.
!
!              On exit, XN_GRP contains the numbers of non-missing observations
!              in the NGRP groups for all the variables in the data matrix.
!
!              The shape of XN_GRP must verify:
!
!              - size(XN_GRP,1) = size(X,DIMVAR)
!              - size(XN_GRP,2) = NGRP.
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!   DIMVAR     (INPUT, OPTIONAL) integer(i4b)
!              On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!              - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!                on size(X,1) variables.
!              - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!                on size(X,2) variables.
!
!              The default is DIMVAR = 1.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one valid observation were present for some variables and/or groups
!   of observations, the pertinent statistics are set to missing (XMISS value).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                       :: xmiss
    real(stnd), intent(in),       dimension(:,:) :: x
    real(stnd), intent(inout),    dimension(:,:) :: xn_grp, xmean_grp, xstd_grp
!
    integer(i4b), intent(in)                          :: ngrp
    integer(i4b), intent(in),            dimension(:) :: ind
    integer(i4b), intent(in),  optional               :: dimvar
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xmean_grp,1))  :: di, dfi
!
    integer(i4b) :: i, k, n, m
!
    logical(lgl) :: dim_is_one
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(ind),i4b),             &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    int(size(xn_grp,1),i4b),       &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                 logical(size(xstd_grp,2)==ngrp,lgl),   &
                 logical(size(xn_grp,2)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:m,:ngrp) = zero       
        xstd_grp(:m,:ngrp)  = zero       
        xn_grp(:m,:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        if ( dim_is_one ) then
            dfi(:m) = x(:m,i)
        else
            dfi(:m) = x(i,:m)
        end if
!
        where ( dfi(:m)/=xmiss )
!
            xn_grp(:m,k) = xn_grp(:m,k) + one
            di(:m)       = dfi(:m) - xmean_grp(:m,k)
            dfi(:m)      = di(:m)/xn_grp(:m,k)
!
            xmean_grp(:m,k) = xmean_grp(:m,k) + dfi(:m)
            xstd_grp(:m,k)  = xstd_grp(:m,k) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
        end where
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        where ( xn_grp(:m,:ngrp)/=zero )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            xstd_grp(:m,:ngrp) = sqrt( xstd_grp(:m,:ngrp)/xn_grp(:m,:ngrp) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO MISSING.
!
            xmean_grp(:m,:ngrp) = xmiss
            xstd_grp(:m,:ngrp)  = xmiss
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_miss_rm
! ______________________________________
!
    end subroutine comp_mvs_grp_miss_rm
!
! =========================================================================================
!
    subroutine comp_mvs_grp_miss_rt( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp, xmiss )
!
! Purpose
! _______
!
!   COMP_MVS_GRP_MISS computes estimates of univariate statistics by groups from a data
!   tridimensional array possibly containing missing values.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:,:)
!              On entry, input subarray containing size(X,3) observations
!              on size(X,1) by size(X,2) variables from the array of data for 
!              which univariate statistics by groups are desired. If all the data
!              are available at once, X can be the full data array.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subarray is the first subarray 
!                              of the data array.
!              - FIRST = false the current subarray is not the first subarray
!                              of the data array.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subarray is the last subarray 
!                             of the data array.
!              - LAST = false the current subarray is not the last subarray
!                             of the data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP_MISS. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP_MISS. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1)
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XN_GRP contains counts of non-missing observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP_MISS. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of non-missing
!              observations in the NGRP groups for all the variables in the data array.
!
!              The shape of XN_GRP must verify:
!
!              - size(XN_GRP,1) = size(X,1)
!              - size(XN_GRP,2) = size(X,2)
!              - size(XN_GRP,3) = NGRP.
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one valid observation were present for some variables and/or groups
!   of observations, the pertinent statistics are set to missing (XMISS value).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                          :: xmiss
    real(stnd), intent(in),        dimension(:,:,:) :: x
    real(stnd), intent(inout),     dimension(:,:,:) :: xn_grp, xmean_grp, xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(x,1),size(x,2))    :: di, dfi
!
    integer(i4b) :: i, j, k, n, m, p
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(size(x,1),i4b),            &
                    int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    int(size(xn_grp,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),            &
                    int(size(xmean_grp,2),i4b),    &
                    int(size(xstd_grp,2),i4b),     &
                    int(size(xn_grp,2),i4b),       &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                 logical(size(xstd_grp,3)==ngrp,lgl),   &
                 logical(size(xn_grp,3)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:m,:p,:ngrp) = zero       
        xstd_grp(:m,:p,:ngrp)  = zero       
        xn_grp(:m,:p,:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        do j = 1_i4b, p
!
            where ( x(:m,j,i)/=xmiss )
!
                xn_grp(:m,j,k) = xn_grp(:m,j,k) + one
                di(:m,j)       = x(:m,j,i) - xmean_grp(:m,j,k)
                dfi(:m,j)      = di(:m,j)/xn_grp(:m,j,k)
!
                xmean_grp(:m,j,k) = xmean_grp(:m,j,k) + dfi(:m,j)
                xstd_grp(:m,j,k)  = xstd_grp(:m,j,k) + di(:m,j)*di(:m,j) - di(:m,j)*dfi(:m,j)
!
            end where
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        where ( xn_grp(:m,:p,:ngrp)/=zero )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            xstd_grp(:m,:p,:ngrp) = sqrt( xstd_grp(:m,:p,:ngrp)/xn_grp(:m,:p,:ngrp) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO MISSING.
!
            xmean_grp(:m,:p,:ngrp) = xmiss
            xstd_grp(:m,:p,:ngrp)  = xmiss
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_miss_rt
! ______________________________________
!
    end subroutine comp_mvs_grp_miss_rt
!
! =========================================================================================
!
    subroutine comp_mvs_grp_miss_rf( x, first, last, ngrp, ind, xmean_grp, xstd_grp, xn_grp, xmiss )
!
! Purpose
! _______
!
!   COMP_MVS_GRP_MISS computes estimates of univariate statistics by groups from a data
!   fourdimensional array possibly containing missing values.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:,:,:)
!              On entry, input subarray containing size(X,4) observations
!              on size(X,1) by size(X,2) by size(X,3) variables from the array of data for 
!              which univariate statistics by groups are desired. If all the data
!              are available at once, X can be the full data array.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subarray is the first subarray 
!                of the data array.
!              - FIRST = false the current subarray is not the first subarray
!                of the data array.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subarray is the last subarray 
!                             of the data array.
!              - LAST = false the current subarray is not the last subarray
!                             of the data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,4) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,4) .
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_MVS_GRP_MISS. XMEAN_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = size(X,3)
!              - size(XMEAN_GRP,4) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_MVS_GRP_MISS. XSTD_GRP should not be changed between calls 
!              to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1)
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = size(X,3)
!              - size(XSTD_GRP,4) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, after the first call to COMP_MVS_GRP_MISS (e.g. when FIRST=true),
!              XN_GRP contains counts of non-missing observations for the NGRP groups
!              from previous calls to COMP_MVS_GRP_MISS. XN_GRP should not be 
!              changed between calls to COMP_MVS_GRP_MISS.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of non-missing
!              observations in the NGRP groups for all the variables in the data array.
!
!              The shape of XN_GRP must verify:
!
!              - size(XN_GRP,1) = size(X,1)
!              - size(XN_GRP,2) = size(X,2)
!              - size(XN_GRP,3) = size(X,3)
!              - size(XN_GRP,4) = NGRP.
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one valid observation were present for some variables and/or groups
!   of observations, the pertinent statistics are set to missing (XMISS value).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero, one
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                            :: xmiss
    real(stnd), intent(in),        dimension(:,:,:,:) :: x
    real(stnd), intent(inout),     dimension(:,:,:,:) :: xn_grp, xmean_grp, xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(x,1),size(x,2),size(x,3))    :: di, dfi
!
    integer(i4b) :: i, j, k, l, n, m, p, q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,4),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(size(x,1),i4b),            &
                    int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    int(size(xn_grp,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),            &
                    int(size(xmean_grp,2),i4b),    &
                    int(size(xstd_grp,2),i4b),     &
                    int(size(xn_grp,2),i4b),       &
                    name_proc )
!
    q = assert_eq(  int(size(x,3),i4b),            &
                    int(size(xmean_grp,3),i4b),    &
                    int(size(xstd_grp,3),i4b),     &
                    int(size(xn_grp,3),i4b),       &
                    name_proc )
!
    if ( min(m,p,q)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,4)==ngrp,lgl),  &
                 logical(size(xstd_grp,4)==ngrp,lgl),   &
                 logical(size(xn_grp,4)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean_grp, xstd_grp, xn_grp .
!
        xmean_grp(:m,:p,:q,:ngrp) = zero       
        xstd_grp(:m,:p,:q,:ngrp)  = zero       
        xn_grp(:m,:p,:q,:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        do l = 1_i4b, q
!
            do j = 1_i4b, p
!
                where ( x(:m,j,l,i)/=xmiss )
!
                    xn_grp(:m,j,l,k) = xn_grp(:m,j,l,k) + one
                    di(:m,j,l)       = x(:m,j,l,i) - xmean_grp(:m,j,l,k)
                    dfi(:m,j,l)      = di(:m,j,l)/xn_grp(:m,j,l,k)
!
                    xmean_grp(:m,j,l,k) = xmean_grp(:m,j,l,k) + dfi(:m,j,l)
                    xstd_grp(:m,j,l,k)  = xstd_grp(:m,j,l,k) + di(:m,j,l)*di(:m,j,l) - di(:m,j,l)*dfi(:m,j,l)
!
                end where
!
            end do
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        where ( xn_grp(:m,:p,:q,:ngrp)/=zero )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            xstd_grp(:m,:p,:q,:ngrp) = sqrt( xstd_grp(:m,:p,:q,:ngrp)/xn_grp(:m,:p,:q,:ngrp) )
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO MISSING.
!
            xmean_grp(:m,:p,:q,:ngrp) = xmiss
            xstd_grp(:m,:p,:q,:ngrp)  = xmiss
!
        end where
!
    end if
!
!
! END OF SUBROUTINE comp_mvs_grp_miss_rf
! ______________________________________
!
    end subroutine comp_mvs_grp_miss_rf
!
!
! =========================================================================================
!                         UPDATING UNIVARIATE STATISTICS
! =========================================================================================
!
!
    subroutine update_mvs_rv( xmean, xvar, xnobs, xmean2, xvar2, xnobs2 )
!
! Purpose
! _______
!
!   UPDATE_MVS computes sample mean and corrected sum of squares
!   for a sample of size XNOBS+XNOBS2 given the means and corrected sums of squares
!   for two subsamples of size XNOBS and XNOBS2 as output by a call to COMP_MVS
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, standard-deviations for the sample of size XNOBS+XNOBS2
!   may be obtained by a call to COMP_MVS with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN  (INPUT/OUTPUT) real(stnd)
!          On entry, the sample mean of the first sample of size XNOBS.
!
!          On exit, the sample mean of the combined sample of size XNOBS+XNOBS2.
!
!   XVAR   (INPUT/OUTPUT) real(stnd)
!          On entry, the corrected sum of squares of the first sample of size XNOBS.
!
!          On exit, the corrected sum of squares of the combined sample of size XNOBS+XNOBS2.
!
!   XNOBS  (INPUT/OUTPUT) real(stnd)
!          On entry, the number of observations of the first sample.
!
!          On exit, the number of observations of the combined sample (i.e. XNOBS+XNOBS2).
!
!   XMEAN2 (INPUT) real(stnd)
!          On entry, the sample mean of the second sample of size XNOBS2.
!
!   XVAR2  (INPUT) real(stnd)
!          On entry, the corrected sum of squares of the second sample of size XNOBS2.
!
!
!   XNOBS2 (INPUT) real(stnd)
!          On entry, the number of observations of the second sample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS.
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: xmean, xvar, xnobs
    real(stnd), intent(in)    :: xmean2, xvar2, xnobs2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    if ( xnobs2/=zero ) then
!
        tmp = xnobs + xnobs2
!
!       UPDATE CORRECTED SUMS OF SQUARES.
!
        xvar  = xvar + xvar2    +   &
                ((xnobs*xnobs2)/tmp)*( xmean - xmean2 )**2
!
!       UPDATE MEANS.
!
        xmean = (xnobs/tmp)*xmean    +   &
                (xnobs2/tmp)*xmean2
!
        xnobs = tmp
!
    end if
!
!
! END OF SUBROUTINE update_mvs_rv
! _______________________________
!
    end subroutine update_mvs_rv
!
! =========================================================================================
!
    subroutine update_mvs_rm( xmean, xvar, xnobs, xmean2, xvar2, xnobs2 )
!
! Purpose
! _______
!
!   UPDATE_MVS computes sample means and corrected sums of squares by groups
!   for a sample of size XNOBS+XNOBS2 given the means and corrected sums of squares
!   for two subsamples of size XNOBS and XNOBS2 as output by a call to COMP_MVS
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, standard-deviations for the sample of size XNOBS+XNOBS2 
!   may be obtained by a call to COMP_MVS with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, the sample means of the first sample of size XNOBS.
!
!          On exit, the sample means of the combined sample of size XNOBS+XNOBS2.
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, the corrected sums of squares of the first sample of size XNOBS.
!
!          On exit, the corrected sums of squares of the combined sample of size XNOBS+XNOBS2.
!
!          The shape of XVAR must verify size(XVAR) = size(XMEAN).
!
!   XNOBS  (INPUT/OUTPUT) real(stnd)
!          On entry, the number of observations of the first sample.
!
!          On exit, the number of observations of the combined sample (i.e. XNOBS+XNOBS2).
!
!
!   XMEAN2 (INPUT) real(stnd), dimension(:)
!          On entry, the sample means of the second sample of size XNOBS2.
!
!          The shape of XMEAN2 must verify size(XMEAN2) = size(XMEAN).
!
!   XVAR2  (INPUT) real(stnd), dimension(:)
!          On entry, the corrected sum of squares of the second sample of size XNOBS2.
!
!          The shape of XVAR2 must verify size(XVAR2) = size(XMEAN).
!
!   XNOBS2 (INPUT) real(stnd)
!          On entry, the number of observations of the second sample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS.
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),   dimension(:) :: xmean, xvar
    real(stnd), intent(inout)                 :: xnobs
!
    real(stnd), intent(in),      dimension(:) :: xmean2, xvar2
    real(stnd), intent(in)                    :: xnobs2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: tmp
    real(stnd), dimension(size(xmean)) :: temp
!
    integer(i4b) :: m
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(xmean),i4b),    &
                    int(size(xvar),i4b),     &
                    int(size(xmean2),i4b),   &
                    int(size(xvar2),i4b),    &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( xnobs2/=zero ) then
!
        tmp = xnobs + xnobs2
!
!       UPDATE CORRECTED SUMS OF SQUARES.
!
        temp(:m)  = xvar(:m) + xvar2(:m)    +   &
                    ((xnobs*xnobs2)/tmp)*( xmean(:m) - xmean2(:m) )**2
        xvar(:m) = temp(:m)
!
!       UPDATE MEANS.
!
        temp(:m) = (xnobs/tmp)*xmean(:m)    +   &
                   (xnobs2/tmp)*xmean2(:m)
        xmean(:m) = temp(:m)
!
        xnobs = tmp
!
    end if
!
!
! END OF SUBROUTINE update_mvs_rm
! _______________________________
!
    end subroutine update_mvs_rm
!
! =========================================================================================
!
    subroutine update_mvs_rt( xmean, xvar, xnobs, xmean2, xvar2, xnobs2 )
!
! Purpose
! _______
!
!   UPDATE_MVS computes sample means and corrected sums of squares by groups
!   for a sample of size XNOBS+XNOBS2 given the means and corrected sums of squares
!   for two subsamples of size XNOBS and XNOBS2 as output by a call to COMP_MVS
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, standard-deviations for the sample of size XNOBS+XNOBS2
!   may be obtained by a call to COMP_MVS with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, the sample means of the first sample of size XNOBS.
!
!          On exit, the sample means of the combined sample of size XNOBS+XNOBS2.
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, the corrected sums of squares of the first sample of size XNOBS.
!
!          On exit, the corrected sums of squares of the combined sample of size XNOBS+XNOBS2
!
!          The shape of XVAR must verify:
!
!          - size(XVAR,1) = size(XMEAN,1)
!          - size(XVAR,2) = size(XMEAN,2).
!
!   XNOBS  (INPUT/OUTPUT) real(stnd)
!          On entry, the number of observations of the first sample.
!
!          On exit, the number of observations of the combined sample (i.e. XNOBS+XNOBS2).
!
!
!   XMEAN2 (INPUT) real(stnd), dimension(:,:)
!          On entry, the sample means of the second sample of size XNOBS2.
!
!          The shape of XMEAN2 must verify:
!
!          - size(XMEAN2,1) = size(XMEAN,1)
!          - size(XMEAN2,2) = size(XMEAN,2).
!
!   XVAR2  (INPUT) real(stnd), dimension(:,:)
!          On entry, the corrected sum of squares of the second sample of size XNOBS2.
!
!          The shape of XVAR2 must verify:
!
!          - size(XVAR2,1) = size(XMEAN,1)
!          - size(XVAR2,2) = size(XMEAN,2).
!
!   XNOBS2 (INPUT) real(stnd), dimension(:)
!          On entry, the number of observations of the second sample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS.
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),      dimension(:,:) :: xmean, xvar
    real(stnd), intent(inout)                      :: xnobs
!
    real(stnd), intent(in),      dimension(:,:) :: xmean2, xvar2
    real(stnd), intent(in)                      :: xnobs2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: tmp
    real(stnd), dimension(size(xmean,1)) :: temp
!
    integer(i4b) :: m, n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
!
    m = assert_eq(  int(size(xmean,1),i4b),    &
                    int(size(xvar,1),i4b),     &
                    int(size(xmean2,1),i4b),   &
                    int(size(xvar2,1),i4b),    &
                    name_proc )
!
    n = assert_eq(  int(size(xmean,2),i4b),    &
                    int(size(xvar,2),i4b),     &
                    int(size(xmean2,2),i4b),   &
                    int(size(xvar2,2),i4b),    &
                    name_proc )
!
    if ( min(m,n)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( xnobs2/=zero ) then
!
        tmp = xnobs + xnobs2
!
        do i = 1_i4b, n
!
!           UPDATE CORRECTED SUMS OF SQUARES.
!
            temp(:m)   = xvar(:m,i) + xvar2(:m,i)    +   &
                         ((xnobs*xnobs2)/tmp)*( xmean(:m,i) - xmean2(:m,i) )**2
            xvar(:m,i) = temp(:m)
!
!           UPDATE MEANS.
!
            temp(:m)    = (xnobs/tmp)*xmean(:m,i)    +   &
                          (xnobs2/tmp)*xmean2(:m,i)
            xmean(:m,i) = temp(:m)
!
        end do
!
        xnobs = tmp
!
    end if
!
!
! END OF SUBROUTINE update_mvs_rt
! _______________________________
!
    end subroutine update_mvs_rt
!
! =========================================================================================
!
    subroutine update_mvs_rf( xmean, xvar, xnobs, xmean2, xvar2, xnobs2 )
!
! Purpose
! _______
!
!   UPDATE_MVS computes sample means and corrected sums of squares by groups
!   for a sample of size XNOBS+XNOBS2 given the means and corrected sums of squares
!   for two subsamples of size XNOBS and XNOBS2 as output by a call to COMP_MVS
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, standard-deviations for the sample of size XNOBS+XNOBS2
!   may be obtained by a call to COMP_MVS with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, the sample means of the first sample of size XNOBS.
!
!          On exit, the sample means of the combined sample of size XNOBS+XNOBS2.
!
!   XVAR   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, the corrected sums of squares of the first sample of size XNOBS.
!
!          On exit, the corrected sums of squares of the combined sample of size XNOBS+XNOBS2.
!
!          The shape of XVAR must verify:
!
!          - size(XVAR,1) = size(XMEAN,1)
!          - size(XVAR,2) = size(XMEAN,2)
!          - size(XVAR,3) = size(XMEAN,3).
!
!   XNOBS  (INPUT/OUTPUT) real(stnd)
!          On entry, the number of observations of the first sample.
!
!          On exit, the number of observations of the combined sample (i.e. XNOBS+XNOBS2).
!
!
!   XMEAN2 (INPUT) real(stnd), dimension(:,:,:)
!          On entry, the sample means of the second sample of size XNOBS2.
!
!          The shape of XMEAN2 must verify:
!
!          - size(XMEAN2,1) = size(XMEAN,1)
!          - size(XMEAN2,2) = size(XMEAN,2)
!          - size(XMEAN2,3) = size(XMEAN,3).
!
!   XVAR2  (INPUT) real(stnd), dimension(:,:,:)
!          On entry, the corrected sum of squares of the second sample of size XNOBS2.
!
!          The shape of XVAR2 must verify:
!
!          - size(XVAR2,1) = size(XMEAN,1)
!          - size(XVAR2,2) = size(XMEAN,2)
!          - size(XVAR2,3) = size(XMEAN,3).
!
!   XNOBS2  (INPUT) real(stnd)
!           On entry, the number of observations of the second sample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS.
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),   dimension(:,:,:) :: xmean, xvar
    real(stnd), intent(inout)                     :: xnobs
!
    real(stnd), intent(in),      dimension(:,:,:) :: xmean2, xvar2
    real(stnd), intent(in)                        :: xnobs2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: tmp
    real(stnd), dimension(size(xmean,1)) :: temp
!
    integer(i4b) :: m, n, p, i, j
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
!
    m = assert_eq(  int(size(xmean,1),i4b),    &
                    int(size(xvar,1),i4b),     &
                    int(size(xmean2,1),i4b),   &
                    int(size(xvar2,1),i4b),    &
                    name_proc )
!
    n = assert_eq(  int(size(xmean,2),i4b),    &
                    int(size(xvar,2),i4b),     &
                    int(size(xmean2,2),i4b),   &
                    int(size(xvar2,2),i4b),    &
                    name_proc )
!
    p = assert_eq(  int(size(xmean,3),i4b),    &
                    int(size(xvar,3),i4b),     &
                    int(size(xmean2,3),i4b),   &
                    int(size(xvar2,3),i4b),    &
                    name_proc )
!
    if ( min(m,n,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( xnobs2/=zero ) then
!
        tmp = xnobs + xnobs2
!
        do j = 1_i4b, p
!
            do i = 1_i4b, n
!
!               UPDATE CORRECTED SUMS OF SQUARES.
!
                temp(:m)     = xvar(:m,i,j) + xvar2(:m,i,j)    +   &
                               ((xnobs*xnobs2)/tmp)*( xmean(:m,i,j) - xmean2(:m,i,j) )**2
                xvar(:m,i,j) = temp(:m)
!
!               UPDATE MEANS.
!
                temp(:m)      = (xnobs/tmp)*xmean(:m,i,j)    +   &
                                (xnobs2/tmp)*xmean2(:m,i,j)
                xmean(:m,i,j) = temp(:m)
!
            end do
!
        end do
!
        xnobs = tmp
!
    end if
!
!
! END OF SUBROUTINE update_mvs_rf
! _______________________________
!
    end subroutine update_mvs_rf
!
!
! =========================================================================================
!                  UPDATING UNIVARIATE STATISTICS BY GROUPS
! =========================================================================================
!
!
    subroutine update_mvs_grp_rv( xmean_grp, xstd_grp, xn_grp, xmean_grp2, xstd_grp2, xn_grp2 )
!
! Purpose
! _______
!
!   UPDATE_MVS_GRP computes sample means and corrected sums of squares by groups
!   for a sample of size sum(XN_GRP)+sum(XN_GRP2) given the means and corrected sums of squares
!   for two subsamples of size sum(XN_GRP) and sum(XN_GRP2) as output by a call to COMP_MVS_GRP
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and standard-deviations for the sample of size sum(XN_GRP)+sum(XN_GRP2) 
!   may be obtained by a call to COMP_MVS_GRP with LAST=true.
!   
!
! Arguments
! _________
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, the sample means for the groups computed on the first sample.
!
!              On exit, the sample means for the groups computed on the combined sample.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, the sample corrected sums of squares for the groups computed on the first sample.
!
!              On exit, the sample corrected sums of squares for the groups computed on the combined sample.
!
!              The shape of XSTD_GRP must verify size(XSTD_GRP) = size(XMEAN_GRP).
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the first sample.
!
!              On exit, the number of observations in each group in the combined sample.
!
!              The size of XN_GRP must verify size(XN_GRP) = size(XMEAN_GRP) .
!
!   XMEAN_GRP2 (INPUT) real(stnd), dimension(:)
!              On entry, the sample means for the groups computed on the second sample.
!
!              The shape of XMEAN_GRP2 must verify size(XMEAN_GRP2) = size(XMEAN_GRP).
!
!   XSTD_GRP2  (INPUT) real(stnd), dimension(:)
!              On entry, the sample corrected sums of squares for the groups computed on the second sample.
!
!              The shape of XSTD_GRP2 must verify  size(XSTD_GRP2) = size(XMEAN_GRP).
!
!   XN_GRP2    (INPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the second sample.
!
!              The size of XN_GRP2 must verify size(XN_GRP2) = size(XMEAN_GRP) .
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS_GRP. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS_GRP.
!
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS_GRP with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Utilities,         only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),   dimension(:) :: xmean_grp, xstd_grp, xn_grp
!
    real(stnd), intent(in),      dimension(:) :: xmean_grp2, xstd_grp2, xn_grp2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp, tmp2
!
    integer(i4b) :: ngrp, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    ngrp = assert_eq(  (/ int(size(xmean_grp),i4b),    &
                          int(size(xstd_grp),i4b),     &
                          int(size(xn_grp),i4b),       &
                          int(size(xmean_grp2),i4b),   &
                          int(size(xstd_grp2),i4b),    &
                          int(size(xn_grp2),i4b) /),   &
                     name_proc )
!
    do i = 1_i4b, ngrp
!
        if ( xn_grp2(i)==zero ) cycle
!
        tmp = xn_grp(i) + xn_grp2(i)
!
!       UPDATE CORRECTED SUMS OF SQUARES.
!
        tmp2  = xstd_grp(i) + xstd_grp2(i)    +   &
                ((xn_grp(i)*xn_grp2(i))/tmp)*( xmean_grp(i) - xmean_grp2(i) )**2
        xstd_grp(i) = tmp2
!
!       UPDATE MEANS.
!
        tmp2 = (xn_grp(i)/tmp)*xmean_grp(i)    +   &
               (xn_grp2(i)/tmp)*xmean_grp2(i)
        xmean_grp(i) = tmp2
!
        xn_grp(i) = tmp
!
    end do
!
!
! END OF SUBROUTINE update_mvs_grp_rv
! ___________________________________
!
    end subroutine update_mvs_grp_rv
!
! =========================================================================================
!
    subroutine update_mvs_grp_rm( xmean_grp, xstd_grp, xn_grp, xmean_grp2, xstd_grp2, xn_grp2 )
!
! Purpose
! _______
!
!   UPDATE_MVS_GRP computes sample means and corrected sums of squares by groups
!   for a sample of size sum(XN_GRP)+sum(XN_GRP2) given the means and corrected sums of squares
!   for two subsamples of size sum(XN_GRP) and sum(XN_GRP2) as output by a call to COMP_MVS_GRP
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and standard-deviations for the sample of size sum(XN_GRP)+sum(XN_GRP2) 
!   may be obtained by a call to COMP_MVS_GRP with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, the sample means for the groups computed on the first sample vector.
!
!              On exit, the sample means for the groups computed on the combined sample vector.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the first sample vector.
!
!              On exit, the sample corrected sums of squares for the groups computed on the combined sample vector.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP,2) = size(XMEAN_GRP,2).
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the first sample vector.
!
!              On exit, the number of observations in each group in the combined sample vector.
!
!              The size of XN_GRP must verify size(XN_GRP) = size(XMEAN_GRP,2) .
!
!   XMEAN_GRP2 (INPUT) real(stnd), dimension(:,:)
!              On entry, the sample means for the groups computed on the second sample vector.
!
!              The shape of XMEAN_GRP2 must verify:
!
!              - size(XMEAN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XMEAN_GRP2,2) = size(XMEAN_GRP,2)
!
!   XSTD_GRP2  (INPUT) real(stnd), dimension(:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the second sample vector.
!
!              The shape of XSTD_GRP2 must verify:
!
!              - size(XSTD_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP2,2) = size(XMEAN_GRP,2).
!
!   XN_GRP2    (INPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the second sample vector.
!
!              The size of XN_GRP2 must verify size(XN_GRP2) = size(XMEAN_GRP,2) .
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS_GRP. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS_GRP.
!
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS_GRP with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),      dimension(:,:) :: xmean_grp, xstd_grp
    real(stnd), intent(inout),      dimension(:)   :: xn_grp
!
    real(stnd), intent(in),      dimension(:,:) :: xmean_grp2, xstd_grp2
    real(stnd), intent(in),      dimension(:)   :: xn_grp2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                               :: tmp
    real(stnd), dimension(size(xmean_grp,1)) :: temp
!
    integer(i4b) :: m, ngrp, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    m = assert_eq(  int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    int(size(xmean_grp2,1),i4b),   &
                    int(size(xstd_grp2,1),i4b),    &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    ngrp = assert_eq(  (/ int(size(xmean_grp,2),i4b),    &
                          int(size(xstd_grp,2),i4b),     &
                          int(size(xn_grp),i4b),         &
                          int(size(xmean_grp2,2),i4b),   &
                          int(size(xstd_grp2,2),i4b),    &
                          int(size(xn_grp2),i4b) /),     &
                     name_proc )
!
    do i = 1_i4b, ngrp
!
        if ( xn_grp2(i)==zero ) cycle
!
        tmp = xn_grp(i) + xn_grp2(i)
!
!       UPDATE CORRECTED SUMS OF SQUARES.
!
        temp(:m)  = xstd_grp(:m,i) + xstd_grp2(:m,i)    +   &
                    ((xn_grp(i)*xn_grp2(i))/tmp)*( xmean_grp(:m,i) - xmean_grp2(:m,i) )**2
        xstd_grp(:m,i) = temp(:m)
!
!       UPDATE MEANS.
!
        temp(:m) = (xn_grp(i)/tmp)*xmean_grp(:m,i)    +   &
                   (xn_grp2(i)/tmp)*xmean_grp2(:m,i)
        xmean_grp(:m,i) = temp(:m)
!
        xn_grp(i) = tmp
!
    end do
!
!
! END OF SUBROUTINE update_mvs_grp_rm
! ___________________________________
!
    end subroutine update_mvs_grp_rm
!
! =========================================================================================
!
    subroutine update_mvs_grp_rt( xmean_grp, xstd_grp, xn_grp, xmean_grp2, xstd_grp2, xn_grp2 )
!
! Purpose
! _______
!
!   UPDATE_MVS_GRP computes sample means and corrected sums of squares by groups
!   for a sample of size sum(XN_GRP)+sum(XN_GRP2) given the means and corrected sums of squares
!   for two subsamples of size sum(XN_GRP) and sum(XN_GRP2) as output by a call to COMP_MVS_GRP
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and standard-deviations for the sample of size sum(XN_GRP)+sum(XN_GRP2) 
!   may be obtained by a call to COMP_MVS_GRP with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample means for the groups computed on the first sample matrix.
!
!              On exit, the sample means for the groups computed on the combined sample matrix.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the first sample matrix.
!
!              On exit, the sample corrected sums of squares for the groups computed on the combined sample matrix.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP,3) = size(XMEAN_GRP,3).
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the first sample matrix.
!
!              On exit, the number of observations in each group in the combined sample matrix.
!
!              The size of XN_GRP must verify size(XN_GRP) = size(XMEAN_GRP,3) .
!
!   XMEAN_GRP2 (INPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample means for the groups computed on the second sample matrix.
!
!              The shape of XMEAN_GRP2 must verify:
!
!              - size(XMEAN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XMEAN_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XMEAN_GRP2,3) = size(XMEAN_GRP,3).
!
!   XSTD_GRP2  (INPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the second sample matrix.
!
!              The shape of XSTD_GRP2 must verify:
!
!              - size(XSTD_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP2,3) = size(XMEAN_GRP,3).
!
!   XN_GRP2    (INPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the second sample matrix.
!
!              The size of XN_GRP2 must verify size(XN_GRP2) = size(XMEAN_GRP,3) .
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS_GRP. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS_GRP.
!
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS_GRP with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),      dimension(:,:,:) :: xmean_grp, xstd_grp
    real(stnd), intent(inout),      dimension(:)     :: xn_grp
!
    real(stnd), intent(in),      dimension(:,:,:) :: xmean_grp2, xstd_grp2
    real(stnd), intent(in),      dimension(:)     :: xn_grp2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                               :: tmp, tmp2, tmp_grp, tmp_grp2
    real(stnd), dimension(size(xmean_grp,1)) :: temp
!
    integer(i4b) :: m, n, ngrp, i, j
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    int(size(xmean_grp2,1),i4b),   &
                    int(size(xstd_grp2,1),i4b),    &
                    name_proc )
!
    n = assert_eq(  int(size(xmean_grp,2),i4b),    &
                    int(size(xstd_grp,2),i4b),     &
                    int(size(xmean_grp2,2),i4b),   &
                    int(size(xstd_grp2,2),i4b),    &
                    name_proc )
!
    if ( min(m,n)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    ngrp = assert_eq(  (/ int(size(xmean_grp,3),i4b),    &
                          int(size(xstd_grp,3),i4b),     &
                          int(size(xn_grp),i4b),         &
                          int(size(xmean_grp2,3),i4b),   &
                          int(size(xstd_grp2,3),i4b),    &
                          int(size(xn_grp2),i4b) /),     &
                     name_proc )
!
    do i = 1_i4b, ngrp
!
        if ( xn_grp2(i)==zero ) cycle
!
        tmp  = xn_grp(i) + xn_grp2(i)
        tmp2 = ( xn_grp(i)*xn_grp2(i) )/tmp
!
        tmp_grp  = xn_grp(i)/tmp
        tmp_grp2 = xn_grp2(i)/tmp
!
        do j = 1_i4b, n
!
!           UPDATE CORRECTED SUMS OF SQUARES.
!
            temp(:m)         = xstd_grp(:m,j,i) + xstd_grp2(:m,j,i)    +   &
                               tmp2*( xmean_grp(:m,j,i) - xmean_grp2(:m,j,i) )**2
            xstd_grp(:m,j,i) = temp(:m)
!
!           UPDATE MEANS.
!
            temp(:m)          = tmp_grp*xmean_grp(:m,j,i) + tmp_grp2*xmean_grp2(:m,j,i)
            xmean_grp(:m,j,i) = temp(:m)
!
        end do
!
        xn_grp(i) = tmp
!
    end do
!
!
! END OF SUBROUTINE update_mvs_grp_rt
! ___________________________________
!
    end subroutine update_mvs_grp_rt
!
! =========================================================================================
!
    subroutine update_mvs_grp_rf( xmean_grp, xstd_grp, xn_grp, xmean_grp2, xstd_grp2, xn_grp2 )
!
! Purpose
! _______
!
!   UPDATE_MVS_GRP computes sample means and corrected sums of squares by groups
!   for a sample of size sum(XN_GRP)+sum(XN_GRP2) given the means and corrected sums of squares
!   for two subsamples of size sum(XN_GRP) and sum(XN_GRP2) as output by a call to COMP_MVS_GRP
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and standard-deviations for the sample of size sum(XN_GRP)+sum(XN_GRP2) 
!   may be obtained by a call to COMP_MVS_GRP with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample means for the groups computed on the first sample array.
!
!              On exit, the sample means for the groups computed on the combined sample array.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the first sample array.
!
!              On exit, the sample corrected sums of squares for the groups computed on the combined sample array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP,3) = size(XMEAN_GRP,3)
!              - size(XSTD_GRP,4) = size(XMEAN_GRP,4).
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the first sample array.
!
!              On exit, the number of observations in each group in the combined sample array.
!
!              The size of XN_GRP must verify size(XN_GRP) = size(XMEAN_GRP,4) .
!
!   XMEAN_GRP2 (INPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample means for the groups computed on the second sample array.
!
!              The shape of XMEAN_GRP2 must verify: 
!
!              - size(XMEAN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XMEAN_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XMEAN_GRP2,3) = size(XMEAN_GRP,3)
!              - size(XMEAN_GRP2,4) = size(XMEAN_GRP,4).
!
!   XSTD_GRP2  (INPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the second sample array.
!
!              The shape of XSTD_GRP2 must verify: 
!
!              - size(XSTD_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP2,3) = size(XMEAN_GRP,3)
!              - size(XSTD_GRP2,4) = size(XMEAN_GRP,4).
!
!   XN_GRP2    (INPUT) real(stnd), dimension(:)
!              On entry, the number of observations in each group in the second sample array.
!
!              The size of XN_GRP2 must verify size(XN_GRP2) = size(XMEAN_GRP,4) .
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS_GRP. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS_GRP.
!
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS_GRP with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),      dimension(:,:,:,:) :: xmean_grp, xstd_grp
    real(stnd), intent(inout),      dimension(:)       :: xn_grp
!
    real(stnd), intent(in),      dimension(:,:,:,:) :: xmean_grp2, xstd_grp2
    real(stnd), intent(in),      dimension(:)       :: xn_grp2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                               :: tmp, tmp2, tmp_grp, tmp_grp2
    real(stnd), dimension(size(xmean_grp,1)) :: temp
!
    integer(i4b) :: m, n, p, ngrp, i, j, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(xmean_grp,1),i4b),    &
                    int(size(xstd_grp,1),i4b),     &
                    int(size(xmean_grp2,1),i4b),   &
                    int(size(xstd_grp2,1),i4b),    &
                    name_proc )
!
    n = assert_eq(  int(size(xmean_grp,2),i4b),    &
                    int(size(xstd_grp,2),i4b),     &
                    int(size(xmean_grp2,2),i4b),   &
                    int(size(xstd_grp2,2),i4b),    &
                    name_proc )
!
    p = assert_eq(  int(size(xmean_grp,3),i4b),    &
                    int(size(xstd_grp,3),i4b),     &
                    int(size(xmean_grp2,3),i4b),   &
                    int(size(xstd_grp2,3),i4b),    &
                    name_proc )
!
    if ( min(m,n,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    ngrp = assert_eq(  (/ int(size(xmean_grp,4),i4b),    &
                          int(size(xstd_grp,4),i4b),     &
                          int(size(xn_grp),i4b),         &
                          int(size(xmean_grp2,4),i4b),   &
                          int(size(xstd_grp2,4),i4b),    &
                          int(size(xn_grp2),i4b) /),     &
                     name_proc )
!
    do i = 1_i4b, ngrp
!
        if ( xn_grp2(i)==zero ) cycle
!
        tmp  = xn_grp(i) + xn_grp2(i)
        tmp2 = ( xn_grp(i)*xn_grp2(i) )/tmp
!
        tmp_grp  = xn_grp(i)/tmp
        tmp_grp2 = xn_grp2(i)/tmp
!
        do k = 1_i4b, p
!
            do j = 1_i4b, n
!
!               UPDATE CORRECTED SUMS OF SQUARES.
!
                temp(:m)           = xstd_grp(:m,j,k,i) + xstd_grp2(:m,j,k,i)    +   &
                                     tmp2*( xmean_grp(:m,j,k,i) - xmean_grp2(:m,j,k,i) )**2
                xstd_grp(:m,j,k,i) = temp(:m)
!
!               UPDATE MEANS.
!
                temp(:m)            = tmp_grp*xmean_grp(:m,j,k,i) + tmp_grp2*xmean_grp2(:m,j,k,i)
                xmean_grp(:m,j,k,i) = temp(:m)
!
            end do
!
        end do
!
        xn_grp(i) = tmp
!
    end do
!
!
! END OF SUBROUTINE update_mvs_grp_rf
! ___________________________________
!
    end subroutine update_mvs_grp_rf
!
! =========================================================================================
!
    subroutine update_mvs_grp_miss_rm( xmean_grp, xstd_grp, xn_grp, xmean_grp2, xstd_grp2, xn_grp2 )
!
! Purpose
! _______
!
!   UPDATE_MVS_GRP_MISS computes sample means and corrected sums of squares
!   by groups for a sample, possibly containing missing values, given the means
!   and corrected sums of squares for two subsamples as output by a call to COMP_MVS_GRP_MISS
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and standard-deviations for the sample
!   may be obtained by a call to COMP_MVS_GRP_MISS with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, the sample means for the groups computed on the first sample vector.
!
!              On exit, the sample means for the groups computed on the combined sample vector.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the first sample vector.
!
!              On exit, the sample corrected sums of squares for the groups computed on the combined sample vector.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP,2) = size(XMEAN_GRP,2).
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, the number of observations in each group in the first sample vector.
!
!              On exit, the number of observations in each group in the combined sample vector.
!
!              The shape of XSTD_GRP must verify: 
!
!              - size(XN_GRP,1) = size(XMEAN_GRP,1)
!              - size(XN_GRP,2) = size(XMEAN_GRP,2).
!
!   XMEAN_GRP2 (INPUT) real(stnd), dimension(:,:)
!              On entry, the sample means for the groups computed on the second sample vector.
!
!              The shape of XMEAN_GRP2 must verify: 
!
!              - size(XMEAN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XMEAN_GRP2,2) = size(XMEAN_GRP,2).
!
!   XSTD_GRP2  (INPUT) real(stnd), dimension(:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the second sample vector.
!
!              The shape of XSTD_GRP2 must verify: 
!
!              - size(XSTD_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP2,2) = size(XMEAN_GRP,2).
!
!   XN_GRP2    (INPUT) real(stnd), dimension(:,:)
!              On entry, the number of observations in each group in the second sample vector.
!
!              The shape of XSTD_GRP must verify: 
!
!              - size(XN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XN_GRP2,2) = size(XMEAN_GRP,2).
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS_GRP_MISS. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS_GRP_MISS.
!
!   The means, variances and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS_GRP_MISS with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),   dimension(:,:) :: xmean_grp, xstd_grp, xn_grp
!
    real(stnd), intent(in),      dimension(:,:) :: xmean_grp2, xstd_grp2, xn_grp2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xmean_grp,1)) :: temp, temp2
!
    integer(i4b) :: m, ngrp, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  (/    int(size(xmean_grp,1),i4b),    &
                          int(size(xstd_grp,1),i4b),     &
                          int(size(xn_grp,1),i4b),       &
                          int(size(xmean_grp2,1),i4b),   &
                          int(size(xstd_grp2,1),i4b),    &
                          int(size(xn_grp2,1),i4b) /),   &
                     name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    ngrp = assert_eq(  (/ int(size(xmean_grp,2),i4b),    &
                          int(size(xstd_grp,2),i4b),     &
                          int(size(xn_grp,2),i4b),       &
                          int(size(xmean_grp2,2),i4b),   &
                          int(size(xstd_grp2,2),i4b),    &
                          int(size(xn_grp2,2),i4b) /),   &
                     name_proc )
!
    do i = 1_i4b, ngrp
!
        where( xn_grp2(:m,i)/=zero )
!
            temp2(:m) = xn_grp(:m,i) + xn_grp2(:m,i)
!
!           UPDATE CORRECTED SUMS OF SQUARES.
!
            temp(:m)  = xstd_grp(:m,i) + xstd_grp2(:m,i)    +            &
                        ((xn_grp(:m,i)*xn_grp2(:m,i))/temp2(:m))*        &
                        ( xmean_grp(:m,i) - xmean_grp2(:m,i) )**2
            xstd_grp(:m,i) = temp(:m)
!
!           UPDATE MEANS.
!
            temp(:m) = (xn_grp(:m,i)/temp2(:m))*xmean_grp(:m,i)    +   &
                       (xn_grp2(:m,i)/temp2(:m))*xmean_grp2(:m,i)
            xmean_grp(:m,i) = temp(:m)
!
            xn_grp(:m,i) = temp2(:m)
!
        end where
!
    end do
!
!
! END OF SUBROUTINE update_mvs_grp_miss_rm
! ________________________________________
!
    end subroutine update_mvs_grp_miss_rm
!
! =========================================================================================
!
    subroutine update_mvs_grp_miss_rt( xmean_grp, xstd_grp, xn_grp, xmean_grp2, xstd_grp2, xn_grp2 )
!
! Purpose
! _______
!
!   UPDATE_MVS_GRP_MISS computes sample means and corrected sums of squares
!   by groups for a sample, possibly containing missing values, given the means
!   and corrected sums of squares for two subsamples as output by a call to COMP_MVS_GRP_MISS
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and standard-deviations for the sample
!   may be obtained by a call to COMP_MVS_GRP_MISS with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample means for the groups computed on the first sample matrix.
!
!              On exit, the sample means for the groups computed on the combined sample matrix.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the first sample matrix.
!
!              On exit, the sample corrected sums of squares for the groups computed on the combined sample matrix.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP,3) = size(XMEAN_GRP,3).
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, the number of observations in each group in the first sample matrix.
!
!              On exit, the number of observations in each group in the combined sample matrix.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XN_GRP,1) = size(XMEAN_GRP,1)
!              - size(XN_GRP,2) = size(XMEAN_GRP,2)
!              - size(XN_GRP,3) = size(XMEAN_GRP,3).
!
!   XMEAN_GRP2 (INPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample means for the groups computed on the second sample matrix.
!
!              The shape of XMEAN_GRP2 must verify:
!
!              - size(XMEAN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XMEAN_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XMEAN_GRP2,3) = size(XMEAN_GRP,3).
!
!   XSTD_GRP2  (INPUT) real(stnd), dimension(:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the second sample matrix.
!
!              The shape of XSTD_GRP2 must verify:
!
!              - size(XSTD_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP2,3) = size(XMEAN_GRP,3).
!
!   XN_GRP2    (INPUT) real(stnd), dimension(:,:,:)
!              On entry, the number of observations in each group in the second sample matrix.
!
!              The shape of XSTD_GRP must verify: 
!
!              - size(XN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XN_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XN_GRP2,3) = size(XMEAN_GRP,3).
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS_GRP_MISS. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS_GRP_MISS.
!
!   The means, variances and and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS_GRP_MISS with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:,:) :: xmean_grp, xstd_grp, xn_grp
!
    real(stnd), intent(in),    dimension(:,:,:) :: xmean_grp2, xstd_grp2, xn_grp2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xmean_grp,1)) :: temp, temp2
!
    integer(i4b) :: m, n, ngrp, i, j
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  (/    int(size(xmean_grp,1),i4b),    &
                          int(size(xstd_grp,1),i4b),     &
                          int(size(xn_grp,1),i4b),       &
                          int(size(xmean_grp2,1),i4b),   &
                          int(size(xstd_grp2,1),i4b),    &
                          int(size(xn_grp2,1),i4b) /),   &
                     name_proc )
!
    n = assert_eq(  (/    int(size(xmean_grp,2),i4b),    &
                          int(size(xstd_grp,2),i4b),     &
                          int(size(xn_grp,2),i4b),       &
                          int(size(xmean_grp2,2),i4b),   &
                          int(size(xstd_grp2,2),i4b),    &
                          int(size(xn_grp2,2),i4b) /),   &
                     name_proc )
!
    if ( min(m,n)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    ngrp = assert_eq(  (/ int(size(xmean_grp,3),i4b),    &
                          int(size(xstd_grp,3),i4b),     &
                          int(size(xn_grp,3),i4b),       &
                          int(size(xmean_grp2,3),i4b),   &
                          int(size(xstd_grp2,3),i4b),    &
                          int(size(xn_grp2,3),i4b) /),   &
                     name_proc )
!
    do i = 1_i4b, ngrp
!
        do j = 1_i4b, n
!
            where( xn_grp2(:m,j,i)/=zero )
!
                temp2(:m) = xn_grp(:m,j,i) + xn_grp2(:m,j,i)
!
!               UPDATE CORRECTED SUMS OF SQUARES.
!
                temp(:m)  = xstd_grp(:m,j,i) + xstd_grp2(:m,j,i)    +         &
                            ((xn_grp(:m,j,i)*xn_grp2(:m,j,i))/temp2(:m))*     &
                            ( xmean_grp(:m,j,i) - xmean_grp2(:m,j,i) )**2
                xstd_grp(:m,j,i) = temp(:m)
!
!               UPDATE MEANS.
!
                temp(:m) = (xn_grp(:m,j,i)/temp2(:m))*xmean_grp(:m,j,i)    +   &
                           (xn_grp2(:m,j,i)/temp2(:m))*xmean_grp2(:m,j,i)
                xmean_grp(:m,j,i) = temp(:m)
!
                xn_grp(:m,j,i) = temp2(:m)
!
            end where
!
        end do
!
    end do
!
!
! END OF SUBROUTINE update_mvs_grp_miss_rt
! ________________________________________
!
    end subroutine update_mvs_grp_miss_rt
!
! =========================================================================================
!
    subroutine update_mvs_grp_miss_rf( xmean_grp, xstd_grp, xn_grp, xmean_grp2, xstd_grp2, xn_grp2 )
!
! Purpose
! _______
!
!   UPDATE_MVS_GRP_MISS computes sample means and corrected sums of squares
!   by groups for a sample, possibly containing missing values, given the means
!   and corrected sums of squares for two subsamples as output by a call to COMP_MVS_GRP_MISS
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and standard-deviations for the sample
!   may be obtained by a call to COMP_MVS_GRP_MISS with LAST=true.
!
!
! Arguments
! _________
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample means for the groups computed on the first sample array.
!
!              On exit, the sample means for the groups computed on the combined sample array.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the first sample array.
!
!              On exit, the sample corrected sums of squares for the groups computed on the combined sample array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP,3) = size(XMEAN_GRP,3)
!              - size(XSTD_GRP,4) = size(XMEAN_GRP,4).
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the number of observations in each group in the first sample array.
!
!              On exit, the number of observations in each group in the combined sample array.
!
!              The shape of XSTD_GRP must verify: 
!
!              - size(XN_GRP,1) = size(XMEAN_GRP,1)
!              - size(XN_GRP,2) = size(XMEAN_GRP,2)
!              - size(XN_GRP,3) = size(XMEAN_GRP,3)
!              - size(XN_GRP,4) = size(XMEAN_GRP,4).
!
!   XMEAN_GRP2 (INPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample means for the groups computed on the second sample array.
!
!              The shape of XMEAN_GRP2 must verify:
!
!              - size(XMEAN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XMEAN_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XMEAN_GRP2,3) = size(XMEAN_GRP,3)
!              - size(XMEAN_GRP2,4) = size(XMEAN_GRP,4).
!
!   XSTD_GRP2  (INPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the sample corrected sums of squares for the groups computed on the second sample array.
!
!              The shape of XSTD_GRP2 must verify: 
!
!              - size(XSTD_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XSTD_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XSTD_GRP2,3) = size(XMEAN_GRP,3)
!              - size(XSTD_GRP2,4) = size(XMEAN_GRP,4).
!
!   XN_GRP2    (INPUT) real(stnd), dimension(:,:,:,:)
!              On entry, the number of observations in each group in the second sample array.
!
!              The shape of XSTD_GRP must verify: 
!
!              - size(XN_GRP2,1) = size(XMEAN_GRP,1)
!              - size(XN_GRP2,2) = size(XMEAN_GRP,2)
!              - size(XN_GRP2,3) = size(XMEAN_GRP,3).
!              - size(XN_GRP2,4) = size(XMEAN_GRP,4).
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares computed for each
!   subsample independently using COMP_MVS_GRP_MISS. The means and corrected sums of squares
!   for the original sample can then be calculated using UPDATE_MVS_GRP_MISS.
!
!   The means, variances and and standard-deviations for the original sample can 
!   be computed by a final call to COMP_MVS_GRP_MISS with LAST=true.
!
!   This subroutine is adapted from
!
!   (1) Chan, T.F., Golub, G.H., and Leveque, R.J., 1979: Updating formulae and a pairwise
!           algorithm for computing sample variances. STAN-CS-79-773, November 1979.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),   dimension(:,:,:,:) :: xmean_grp, xstd_grp, xn_grp
!
    real(stnd), intent(in),      dimension(:,:,:,:) :: xmean_grp2, xstd_grp2, xn_grp2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xmean_grp,1)) :: temp, temp2
!
    integer(i4b) :: m, n, p, ngrp, i, j, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_mvs_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  (/    int(size(xmean_grp,1),i4b),    &
                          int(size(xstd_grp,1),i4b),     &
                          int(size(xn_grp,1),i4b),       &
                          int(size(xmean_grp2,1),i4b),   &
                          int(size(xstd_grp2,1),i4b),    &
                          int(size(xn_grp2,1),i4b) /),   &
                     name_proc )
!
    n = assert_eq(  (/    int(size(xmean_grp,2),i4b),    &
                          int(size(xstd_grp,2),i4b),     &
                          int(size(xn_grp,2),i4b),       &
                          int(size(xmean_grp2,2),i4b),   &
                          int(size(xstd_grp2,2),i4b),    &
                          int(size(xn_grp2,2),i4b) /),   &
                     name_proc )
!
    p = assert_eq(  (/    int(size(xmean_grp,3),i4b),    &
                          int(size(xstd_grp,3),i4b),     &
                          int(size(xn_grp,3),i4b),       &
                          int(size(xmean_grp2,3),i4b),   &
                          int(size(xstd_grp2,3),i4b),    &
                          int(size(xn_grp2,3),i4b) /),   &
                     name_proc )
!
    if ( min(m,n,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    ngrp = assert_eq(  (/ int(size(xmean_grp,4),i4b),    &
                          int(size(xstd_grp,4),i4b),     &
                          int(size(xn_grp,4),i4b),       &
                          int(size(xmean_grp2,4),i4b),   &
                          int(size(xstd_grp2,4),i4b),    &
                          int(size(xn_grp2,4),i4b) /),   &
                     name_proc )
!
    do i = 1_i4b, ngrp
!
        do k = 1_i4b, p
!
            do j = 1_i4b, n
!
                where( xn_grp2(:m,j,k,i)/=zero )
!
                    temp2(:m) = xn_grp(:m,j,k,i) + xn_grp2(:m,j,k,i)
!
!                   UPDATE CORRECTED SUMS OF SQUARES.
!
                    temp(:m)  = xstd_grp(:m,j,k,i) + xstd_grp2(:m,j,k,i)    +         &
                                ((xn_grp(:m,j,k,i)*xn_grp2(:m,j,k,i))/temp2(:m))*     &
                                ( xmean_grp(:m,j,k,i) - xmean_grp2(:m,j,k,i) )**2
                    xstd_grp(:m,j,k,i) = temp(:m)
!
!                   UPDATE MEANS.
!
                    temp(:m) = (xn_grp(:m,j,k,i)/temp2(:m))*xmean_grp(:m,j,k,i)    +   &
                               (xn_grp2(:m,j,k,i)/temp2(:m))*xmean_grp2(:m,j,k,i)
                    xmean_grp(:m,j,k,i) = temp(:m)
!
                    xn_grp(:m,j,k,i) = temp2(:m)
!
                end where
!
            end do
!
        end do
!
    end do
!
!
! END OF SUBROUTINE update_mvs_grp_miss_rf
! ________________________________________
!
    end subroutine update_mvs_grp_miss_rf
!
!
! =========================================================================================
!                                    STANDARDIZATION
! =========================================================================================
!
!
    subroutine comp_anoma_rv( x, xmean, xstd )
!
! Purpose
! _______
!
!   COMP_ANOMA computes (standardized) anomalies from a data vector.
!   
!
! Arguments
! _________
!
!   X      (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, input subvector containing size(X) observations
!          from the vector of data for which standardization is desired.
!          If all the data are available at once, X can be the full data vector.
!
!   XMEAN  (INPUT) real(stnd)
!          On entry, XMEAN contains the mean value of the data vector.
!
!   XSTD   (INPUT, OPTIONAL) real(stnd)
!          On entry, if XSTD is present, XSTD contains the standard deviation
!          of the data vector and the anomalies are standardized.
!
!
! Further Details
! _______________
!
!   It is assumed that the argument XSTD is greater than zero.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:) :: x
    real(stnd), intent(in)                   :: xmean
    real(stnd), intent(in),       optional   :: xstd
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: invstd
!
    integer(i4b) :: n
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
    if ( n<=0_i4b ) return
!
    if ( present(xstd) ) then
!
        invstd = one/xstd
!
        x(:n) = ( x(:n) - xmean )*invstd
!
    else
!
        x(:n) = x(:n) - xmean
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_rv
! _______________________________
!
    end subroutine comp_anoma_rv
!
! =========================================================================================
!
    subroutine comp_anoma_rm( x, xmean, xstd, dimvar )
!
! Purpose
! _______
!
!   COMP_ANOMA computes (standardized) anomalies from a data matrix.
!   
!
! Arguments
! _________
!
!   X      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, input submatrix containing size(X,3-DIMVAR)
!          observations on size(X,DIMVAR) variables from the matrix of data
!          for which standardization is desired. By default, 
!          DIMVAR is equal to 1. See description of optional DIMVAR
!          argument for details.
!          If all the data are available at once, X can be the full data vector.
!
!   XMEAN  (INPUT) real(stnd), dimension(:)
!          On entry, XMEAN contains the mean values.
!
!          The size of XMEAN must verify size(XMEAN) = size(X,DIMVAR).
!
!   XSTD   (INPUT, OPTIONAL) real(stnd), dimension(:)
!          On entry, if XSTD is present, XSTD contains the standard deviations
!          and the anomalies are standardized.
!
!          The size of XSTD must verify size(XSTD) = size(X,DIMVAR).
!
!   DIMVAR (INPUT, OPTIONAL) integer(i4b)
!          On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!          - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!            on size(X,1) variables.
!          - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!            on size(X,2) variables.
!
!          The default is DIMVAR = 1.
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XSTD are greater than zero.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error5, allocate_error
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:)           :: x
    real(stnd), intent(in) ,    dimension(:)             :: xmean
    real(stnd), intent(in),     dimension(:),   optional :: xstd
!
    integer(i4b), intent(in),  optional  :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:), allocatable :: invstd
!
    integer(i4b) :: i, j, n, m
    integer      :: iok
!
#ifdef _OPENMP
    integer(i4b) :: k
!
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    m = assert_eq(  int(size(x,int(i)),i4b),     &
                    int(size(xmean),i4b),        &
                    name_proc )
!
    j = 3_i4b - i
    n = size( x, int(j) )
!
    if ( n<=0_i4b .or. m<=0_i4b ) return
!
#ifdef _OPENMP
    j = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.      &
               (m*n)>=omp_limit                .and.      &
               j>1_i4b                         .and.      &
               k>1_i4b
#endif
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd)==m,lgl), name_proc )
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m) = one/xstd(:m)
!
        if ( i==1_i4b ) then
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do j = 1_i4b, n
!
                x(:m,j) = ( x(:m,j) - xmean(:m) )*invstd(:m)
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do j = 1_i4b, m
!
                x(:n,j) = ( x(:n,j) - xmean(j) )*invstd(j)
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
        if ( i==1_i4b ) then
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do j = 1_i4b, n
!
                x(:m,j) = x(:m,j) - xmean(:m)
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do j = 1_i4b, m
!
                x(:n,j) = x(:n,j) - xmean(j)
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_rm
! _______________________________
!
    end subroutine comp_anoma_rm
!
! =========================================================================================
!
    subroutine comp_anoma_rt( x, xmean, xstd )
!
! Purpose
! _______
!
!   COMP_ANOMA computes (standardized) anomalies from a data matrix.
!   
!
! Arguments
! _________
!
!   X      (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, input subarray containing size(X,3) observations
!          on size(X,1) by size(X,2) variables from the array of data
!          for which standardization is desired.
!          If all the data are available at once, X can be the full data array.
!
!   XMEAN  (INPUT) real(stnd), dimension(:,:)
!          On entry, XMEAN contains the mean values.
!
!          The shape of XMEAN must verify:
!
!          - size(XMEAN,1) = size(X,1)
!          - size(XMEAN,2) = size(X,2).
!
!   XSTD   (INPUT, OPTIONAL) real(stnd), dimension(:,:)
!          On entry, if XSTD is present, XSTD contains the standard deviations
!          and the anomalies are standardized.
!
!          The shape of XSTD must verify:
!
!          - size(XSTD,1) = size(X,1)
!          - size(XSTD,2) = size(X,2).
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XSTD are greater than zero.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Char_Constants,    only : allocate_error
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:,:)           :: x
    real(stnd), intent(in) ,    dimension(:,:)             :: xmean
    real(stnd), intent(in),     dimension(:,:),   optional :: xstd
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:,:), allocatable :: invstd
!
    integer(i4b) :: k, i, n, m, p
    integer      :: iok
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(x,1),i4b),       &
                    int(size(xmean,1),i4b),   &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),       &
                    int(size(xmean,2),i4b),   &
                    name_proc )
!
    n = size( x, 3 )
!
    if ( n<=0_i4b .or. m<=0_i4b .or. p<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.    &
               (m*p*n)>=omp_limit              .and.    &
               i>1_i4b                         .and.    &
               k>1_i4b
#endif
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd,1)==m,lgl),     &
                     logical(size(xstd,2)==p,lgl),     &
                     name_proc )
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m,p), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m,:p) = one/xstd(:m,:p)
!
!$OMP PARALLEL DO IF(test_par)                   &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)
!
        do i=1_i4b, n
!
            do k=1_i4b, p
!
                x(:m,k,i) = ( x(:m,k,i) - xmean(:m,k) )*invstd(:m,k)
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)
!
        do i=1_i4b, n
!
            do k=1_i4b, p
!
                x(:m,k,i) = x(:m,k,i) - xmean(:m,k)
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_rt
! _______________________________
!
    end subroutine comp_anoma_rt
!
! =========================================================================================
!
    subroutine comp_anoma_miss_rv( x, xmiss, xmean, xstd )
!
! Purpose
! _______
!
!   COMP_ANOMA_MISS computes (standardized) anomalies from a data vector
!   possibly containing missing values.
!   
!
! Arguments
! _________
!
!   X      (INPUT/OUTPUT) real(stnd), dimension(:)
!          On entry, input subvector containing size(X) observations
!          from the vector of data for which standardization is desired.
!          If all the data are available at once, X can be the full data vector.
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XMEAN  (INPUT) real(stnd)
!          On entry, XMEAN contains the mean value of the data vector.
!
!   XSTD   (INPUT, OPTIONAL) real(stnd)
!          On entry, if XSTD is present, XSTD contains the standard deviation
!          of the data vector and the anomalies are standardized.
!
!
! Further Details
! _______________
!
!   It is assumed that the argument XMEAN is not missing.
!
!   It is assumed that the argument XSTD is greater than zero and is not missing.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:) :: x
    real(stnd), intent(in)                   :: xmean, xmiss
    real(stnd), intent(in),       optional   :: xstd
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: invstd
!
    integer(i4b) :: n
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = size( x )
!
    if ( n<=0_i4b ) return
!
    if ( present(xstd) ) then
!
        invstd = one/xstd
!
        where( x(:n)/=xmiss ) x(:n) = ( x(:n) - xmean )*invstd
!
    else
!
        where( x(:n)/=xmiss ) x(:n) = x(:n) - xmean
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_miss_rv
! ____________________________________
!
    end subroutine comp_anoma_miss_rv
!
! =========================================================================================
!
    subroutine comp_anoma_miss_rm( x, xmiss, xmean, xstd, dimvar )
!
! Purpose
! _______
!
!   COMP_ANOMA_MISS computes (standardized) anomalies from a data matrix
!   possibly containing missing values.
!   
!
! Arguments
! _________
!
!   X      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On entry, input submatrix containing size(X,3-DIMVAR)
!          observations on size(X,DIMVAR) variables from the matrix of data
!          for which standardization is desired. By default, 
!          DIMVAR is equal to 1. See description of optional DIMVAR
!          argument for details.
!          If all the data are available at once, X can be the full data vector.
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XMEAN  (INPUT) real(stnd), dimension(:)
!          On entry, XMEAN contains the mean values.
!
!          The size of XMEAN must verify size(XMEAN) = size(X,DIMVAR).
!
!   XSTD   (INPUT, OPTIONAL) real(stnd), dimension(:)
!          On entry, if XSTD is present, XSTD contains the standard deviations
!          and the anomalies are standardized.
!
!          The size of XSTD must verify size(XSTD) = size(X,DIMVAR).
!
!   DIMVAR (INPUT, OPTIONAL) integer(i4b)
!          On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!          - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!            on size(X,1) variables.
!          - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!            on size(X,2) variables.
!
!          The default is DIMVAR = 1.
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XMEAN are not missing.
!
!   It is assumed that elements of the array argument XSTD are greater than zero
!   and are not missing.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error5, allocate_error
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:)           :: x
    real(stnd), intent(in)                               :: xmiss
    real(stnd), intent(in) ,    dimension(:)             :: xmean
    real(stnd), intent(in),     dimension(:),   optional :: xstd
!
    integer(i4b), intent(in),  optional  :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:), allocatable :: invstd
!
    integer(i4b) :: i, n, m
    integer      :: iok
!
    logical(lgl) :: standardize, dim_is_one
!
#ifdef _OPENMP
    integer(i4b) :: k
!
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    if ( present(xstd) ) then
!
        m = assert_eq(  int(size(x,int(i)),i4b),     &
                        int(size(xmean),i4b),        &
                        int(size(xstd),i4b),         &
                        name_proc )
!
        standardize = true
!
    else
!
        m = assert_eq(  int(size(x,int(i)),i4b),     &
                        int(size(xmean),i4b),        &
                        name_proc )
!
        standardize = false
!
    end if
!
    n = size( x, 3-int(i) )
!
    if ( n<=0_i4b .or. m<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.      &
               (m*n)>=omp_limit                .and.      &
               i>1_i4b                         .and.      &
               k>1_i4b
#endif
!
    if ( standardize ) then
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m) = one/xstd(:m)
!
        if ( dim_is_one ) then
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do i=1_i4b, n
!
                where( x(:m,i)/=xmiss ) x(:m,i) = ( x(:m,i) - xmean(:m) )*invstd(:m)
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do i=1_i4b, m
!
                where( x(:n,i)/=xmiss ) x(:n,i) = ( x(:n,i) - xmean(i) )*invstd(i)
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
        if ( dim_is_one ) then
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do i=1_i4b, n
!
                where( x(:m,i)/=xmiss )  x(:m,i) = x(:m,i) - xmean(:m)
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC) 
!
            do i=1_i4b, m
!
                where( x(:n,i)/=xmiss )  x(:n,i) = x(:n,i) - xmean(i)
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_miss_rm
! ____________________________________
!
    end subroutine comp_anoma_miss_rm
!
! =========================================================================================
!
    subroutine comp_anoma_miss_rt( x, xmiss, xmean, xstd )
!
! Purpose
! _______
!
!   COMP_ANOMA_MISS computes (standardized) anomalies from a data matrix
!   possibly containing missing values.
!   
!
! Arguments
! _________
!
!   X      (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!          On entry, input subarray containing size(X,3) observations
!          on size(X,1) by size(X,2) variables from the array of data
!          for which standardization is desired.
!          If all the data are available at once, X can be the full data array.
!
!   XMISS  (INPUT) real(stnd)
!          On entry, the missing value indicator. Any value in
!          X which is equal to XMISS is assumed to be missing.
!
!   XMEAN  (INPUT) real(stnd), dimension(:,:)
!          On entry, XMEAN contains the mean values.
!
!          The shape of XMEAN must verify:
!
!          - size(XMEAN,1) = size(X,1)
!          - size(XMEAN,2) = size(X,2).
!
!   XSTD   (INPUT, OPTIONAL) real(stnd), dimension(:,:)
!          On entry, if XSTD is present, XSTD contains the standard deviations
!          and the anomalies are standardized.
!
!          The shape of XSTD must verify:
!
!          - size(XSTD,1) = size(X,1)
!          - size(XSTD,2) = size(X,2).
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XMEAN are not missing.
!
!   It is assumed that elements of the array argument XSTD are greater than zero
!   and are not missing.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : allocate_error
    use Utilities,         only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:,:)           :: x
    real(stnd), intent(in)                                 :: xmiss
    real(stnd), intent(in) ,    dimension(:,:)             :: xmean
    real(stnd), intent(in),     dimension(:,:),   optional :: xstd
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:,:), allocatable :: invstd
!
    integer(i4b) :: k, i, n, m, p
    integer      :: iok
!
    logical(lgl) :: standardize
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(xstd) ) then
!
        m = assert_eq(  int(size(x,1),i4b),       &
                        int(size(xmean,1),i4b),   &
                        int(size(xstd,1),i4b),   &
                        name_proc )
!
        p = assert_eq(  int(size(x,2),i4b),       &
                        int(size(xmean,2),i4b),   &
                        int(size(xstd,2),i4b),   &
                        name_proc )
!
        standardize = true
!
    else
!
        m = assert_eq(  int(size(x,1),i4b),       &
                        int(size(xmean,1),i4b),   &
                        name_proc )
        p = assert_eq(  int(size(x,2),i4b),       &
                        int(size(xmean,2),i4b),   &
                        name_proc )
!
        standardize = false
!
    end if
!
    n = size( x, 3 )
!
    if ( n<=0_i4b .or. m<=0_i4b .or. p<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.    &
               (m*p*n)>=omp_limit              .and.    &
               i>1_i4b                         .and.    &
               k>1_i4b
#endif
!
    if ( standardize ) then
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m,p), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m,:p) = one/xstd(:m,:p)
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)
!
        do i=1_i4b, n
!
            do k=1_i4b, p
!
                where( x(:m,k,i)/=xmiss ) x(:m,k,i) = ( x(:m,k,i) - xmean(:m,k) )*invstd(:m,k)
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
!$OMP PARALLEL DO IF(test_par)                   &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)
!
        do i=1_i4b, n
!
            do k=1_i4b, p
!
                where( x(:m,k,i)/=xmiss ) x(:m,k,i) = x(:m,k,i) - xmean(:m,k)
!
            end do
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_miss_rt
! ____________________________________
!
    end subroutine comp_anoma_miss_rt
!
! =========================================================================================
!
    subroutine comp_anoma_grp_rv( x, ngrp, ind, xmean_grp, xstd_grp )
!
! Purpose
! _______
!
!   COMP_ANOMA_GRP computes (standardized) anomalies by groups from a data vector.
!   
!
! Arguments
! _________
!
!   X          (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, input subvector containing size(X) observations
!              from the vector of data for which standardization by groups is desired.
!              If all the data are available at once, X can be the full data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group and this observation is not standardized.
!
!              The size of IND must verify size(IND) = size(X) .
!
!   XMEAN_GRP  (INPUT) real(stnd), dimension(:)
!              On entry, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data vector.
!
!              The size of XMEAN_GRP must verify size(XMEAN_GRP) = NGRP .
!
!   XSTD_GRP   (INPUT, OPTIONAL) real(stnd), dimension(:)
!              On entry, if XSTD_GRP is present, XSTD_GRP contains the standard deviations
!              for the NGRP groups of observations in the data vector and the
!              observations are standardized.
!
!              The size of XSTD_GRP must verify size(XSTD_GRP) = NGRP .
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XSTD_GRP are greater than zero.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Utilities,         only : assert, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:)            :: x
    real(stnd), intent(in),     dimension(:)            :: xmean_grp
    real(stnd), intent(in),     dimension(:),  optional :: xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(ngrp)  :: invstd
!
    integer(i4b) :: i, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = assert_eq(  int(size(x),i4b),         &
                    int(size(ind),i4b),       &
                    name_proc )
!
    if ( n<=0_i4b ) return
!
    if ( present(xstd_grp) ) then
!
        call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                     logical(size(xstd_grp)==ngrp,lgl),   &
                     name_proc )
!
        invstd(:ngrp) = one/xstd_grp(:ngrp)
!
        do i=1_i4b, ngrp
!
            where( ind(:n)==i ) x(:n) = ( x(:n) - xmean_grp(i) )*invstd(i)
!
        end do
!
    else
!
        call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                     name_proc )
!
        do i=1_i4b, ngrp
!
            where( ind(:n)==i )  x(:n) = x(:n) - xmean_grp(i)
!
        end do
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_grp_rv
! ___________________________________
!
    end subroutine comp_anoma_grp_rv
!
! =========================================================================================
!
    subroutine comp_anoma_grp_rm( x, ngrp, ind, xmean_grp, xstd_grp, dimvar )
!
! Purpose
! _______
!
!   COMP_ANOMA_GRP computes (standardized) anomalies by groups from a data matrix.
!   
!
! Arguments
! _________
!
!   X          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, input submatrix containing size(X,3-DIMVAR)
!              observations on size(X,DIMVAR) variables from the matrix of data
!              for which standardization by groups is desired. By default, 
!              DIMVAR is equal to 1. See description of optional DIMVAR
!              argument for details.
!              If all the data are available at once, X can be the full data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3-DIMVAR) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group and this observation is not standardized.
!
!              The size of IND must verify size(IND) = size(X,3-DIMVAR) .
!
!   XMEAN_GRP  (INPUT) real(stnd), dimension(:,:)
!              On entry, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data matrix.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,DIMVAR)
!              - size(XMEAN_GRP,2) = NGRP.
!
!   XSTD_GRP   (INPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On entry, if XSTD_GRP is present, XSTD_GRP contains the standard deviations
!              for the NGRP groups of observations in the data matrix and the
!              observations are standardized.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,DIMVAR)
!              - size(XSTD_GRP,2) = NGRP.
!
!   DIMVAR     (INPUT, OPTIONAL) integer(i4b)
!              On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!              - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!                on size(X,1) variables.
!              - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!                on size(X,2) variables.
!
!              The default is DIMVAR = 1.
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XSTD_GRP are greater than zero.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error5, allocate_error
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:)             :: x
    real(stnd), intent(in) ,    dimension(:,:)             :: xmean_grp
    real(stnd), intent(in),     dimension(:,:),   optional :: xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
    integer(i4b), intent(in),  optional    :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:,:), allocatable :: invstd
!
    integer(i4b) :: i, k, n, m
    integer      :: iok
!
    logical(lgl) :: standardize, dim_is_one
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(ind),i4b),             &
                    name_proc )
!
    if ( present(xstd_grp) ) then
!
        m = assert_eq(  int(size(x,int(i)),i4b),      &
                        int(size(xmean_grp,1),i4b),   &
                        int(size(xstd_grp,1),i4b),    &
                        name_proc )
!
        call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                     logical(size(xstd_grp,2)==ngrp,lgl),   &
                     name_proc )
!
        standardize = true
!
    else
!
        m = assert_eq(  int(size(x,int(i)),i4b),      &
                        int(size(xmean_grp,1),i4b),   &
                        name_proc )
!
        call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                     name_proc )
!
        standardize = false
!
    end if
!
    if ( min(n,m,ngrp)<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.      &
               (m*n)>=omp_limit                .and.      &
               i>1_i4b                         .and.      &
               k>1_i4b
#endif
!
    if ( standardize ) then
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m,ngrp), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m,:ngrp) = one / xstd_grp(:m,:ngrp)
!
        if ( dim_is_one ) then
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    x(:m,i) = ( x(:m,i) - xmean_grp(:m,k) )*invstd(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k) 
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    x(i,:m) = ( x(i,:m) - xmean_grp(:m,k) )*invstd(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
        if ( dim_is_one ) then
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k) 
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    x(:m,i) = x(:m,i) - xmean_grp(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)  
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    x(i,:m) = x(i,:m) - xmean_grp(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_grp_rm
! ___________________________________
!
    end subroutine comp_anoma_grp_rm
!
! =========================================================================================
!
    subroutine comp_anoma_grp_rt( x, ngrp, ind, xmean_grp, xstd_grp )
!
! Purpose
! _______
!
!   COMP_ANOMA_GRP computes (standardized) anomalies by groups from a data matrix
!   possibly containing missing values.
!   
!
! Arguments
! _________
!
!   X          (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, input subarray containing size(X,3) observations
!              on size(X,1) by size(X,2) variables from the array of data
!              for which standardization by groups is desired.
!              If all the data are available at once, X can be the full data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group and this observation is not standardized.
!
!              The size of IND must verify size(IND) = size(X,3) .
!
!   XMEAN_GRP  (INPUT) real(stnd), dimension(:,:,:)
!              On entry, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = NGRP.
!
!   XSTD_GRP   (INPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On entry, if XSTD_GRP is present, XSTD_GRP contains the standard deviations
!              for the NGRP groups of observations in the data array and the
!              observations are standardized.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1)
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = NGRP.
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XSTD_GRP are greater than zero.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : allocate_error
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:,:)             :: x
    real(stnd), intent(in) ,    dimension(:,:,:)             :: xmean_grp
    real(stnd), intent(in),     dimension(:,:,:),   optional :: xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:,:,:), allocatable :: invstd
!
    integer(i4b) :: i, j, k, n, m, p
    integer      :: iok
!
    logical(lgl) :: standardize

#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_grp'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    if ( present(xstd_grp) ) then
!
        m = assert_eq(  int(size(x,1),i4b),           &
                        int(size(xmean_grp,1),i4b),   &
                        int(size(xstd_grp,1),i4b),    &
                        name_proc )
!
        p = assert_eq(  int(size(x,2),i4b),           &
                        int(size(xmean_grp,2),i4b),   &
                        int(size(xstd_grp,2),i4b),    &
                        name_proc )
!
        call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                     logical(size(xstd_grp,3)==ngrp,lgl),   &
                     name_proc )
!
        standardize = true
!
    else
!
        m = assert_eq(  int(size(x,1),i4b),           &
                        int(size(xmean_grp,1),i4b),   &
                        name_proc )
        p = assert_eq(  int(size(x,2),i4b),           &
                        int(size(xmean_grp,2),i4b),   &
                        name_proc )
!
        call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                     name_proc )
!
        standardize = false
!
    end if
!
    if ( min(n,m,p,ngrp)<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.    &
               (m*p*n)>=omp_limit              .and.    &
               i>1_i4b                         .and.    &
               k>1_i4b
#endif
!
    if ( standardize ) then
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m,p,ngrp), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m,:p,:ngrp) = one/xstd_grp(:m,:p,:ngrp)
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,j,k)
!
        do i=1_i4b, n
!
            k = ind(i)
!
            if ( k>=1_i4b .and. k<=ngrp ) then
!
                do j=1_i4b, p
!
                    x(:m,j,i) = ( x(:m,j,i) - xmean_grp(:m,j,k) )*invstd(:m,j,k)
!
                end do
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
!$OMP PARALLEL DO IF(test_par)      &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,j,k)
!
        do i=1_i4b, n
!
            k = ind(i)
!
            if ( k>=1_i4b .and. k<=ngrp ) then
!
                do j=1_i4b, p
!
                    x(:m,j,i) = x(:m,j,i) - xmean_grp(:m,j,k)
!
                end do
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_grp_rt
! ___________________________________
!
    end subroutine comp_anoma_grp_rt
!
! =========================================================================================
!
    subroutine comp_anoma_grp_miss_rv( x, ngrp, ind, xmiss, xmean_grp, xstd_grp )
!
! Purpose
! _______
!
!   COMP_ANOMA_GRP_MISS computes (standardized) anomalies by groups from a data vector
!   possibly containing missing values.
!   
!
! Arguments
! _________
!
!   X          (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, input subvector containing size(X) observations
!              from the vector of data for which standardization by groups is desired.
!              If all the data are available at once, X can be the full data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group and this observation is not standardized.
!
!              The size of IND must verify size(IND) = size(X) .
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!   XMEAN_GRP  (INPUT) real(stnd), dimension(:)
!              On entry, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data vector.
!
!              The size of XMEAN_GRP must verify size(XMEAN_GRP) = NGRP .
!
!   XSTD_GRP   (INPUT, OPTIONAL) real(stnd), dimension(:)
!              On entry, if XSTD_GRP is present, XSTD_GRP contains the standard deviations
!              for the NGRP groups of observations in the data vector and the
!              observations are standardized.
!
!              The size of XSTD_GRP must verify size(XSTD_GRP) = NGRP .
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XMEAN_GRP are not missing.
!
!   It is assumed that elements of the array argument XSTD_GRP are greater than zero
!   and are not missing.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : one
    use Utilities,         only : assert, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:)            :: x
    real(stnd), intent(in)                              :: xmiss
    real(stnd), intent(in),     dimension(:)            :: xmean_grp
    real(stnd), intent(in),     dimension(:),  optional :: xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(ngrp)  :: invstd
!
    integer(i4b) :: i, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_grp_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    n = assert_eq(  int(size(x),i4b),         &
                    int(size(ind),i4b),       &
                    name_proc )
!
    if ( n<=0_i4b ) return
!
    if ( present(xstd_grp) ) then
!
        call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                     logical(size(xstd_grp)==ngrp,lgl),   &
                     name_proc )
!
        invstd(:ngrp) = one/xstd_grp(:ngrp)
!
        do i=1_i4b, ngrp
!
            where( x(:n)/=xmiss .and. ind(:n)==i ) x(:n) = ( x(:n) - xmean_grp(i) )*invstd(i)
!
        end do
!
    else
!
        call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                     name_proc )
!
        do i=1_i4b, ngrp
!
            where( x(:n)/=xmiss .and. ind(:n)==i )  x(:n) = x(:n) - xmean_grp(i)
!
        end do
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_grp_miss_rv
! ________________________________________
!
    end subroutine comp_anoma_grp_miss_rv
!
! =========================================================================================
!
    subroutine comp_anoma_grp_miss_rm( x, ngrp, ind, xmiss, xmean_grp, xstd_grp, dimvar )
!
! Purpose
! _______
!
!   COMP_ANOMA_GRP_MISS computes (standardized) anomalies by groups from a data matrix
!   possibly containing missing values.
!   
!
! Arguments
! _________
!
!   X          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, input submatrix containing size(X,3-DIMVAR)
!              observations on size(X,DIMVAR) variables from the matrix of data
!              for which standardization by groups is desired. By default, 
!              DIMVAR is equal to 1. See description of optional DIMVAR
!              argument for details.
!              If all the data are available at once, X can be the full data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3-DIMVAR) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group and this observation is not standardized.
!
!              The size of IND must verify size(IND) = size(X,3-DIMVAR) .
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!   XMEAN_GRP  (INPUT) real(stnd), dimension(:,:)
!              On entry, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data matrix.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,DIMVAR)
!              - size(XMEAN_GRP,2) = NGRP.
!
!   XSTD_GRP   (INPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On entry, if XSTD_GRP is present, XSTD_GRP contains the standard deviations
!              for the NGRP groups of observations in the data matrix and the
!              observations are standardized.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,DIMVAR)
!              - size(XSTD_GRP,2) = NGRP.
!
!   DIMVAR     (INPUT, OPTIONAL) integer(i4b)
!              On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!              - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!                on size(X,1) variables.
!              - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!                on size(X,2) variables.
!
!              The default is DIMVAR = 1.
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XMEAN_GRP are not missing.
!
!   It is assumed that elements of the array argument XSTD_GRP are greater than zero
!   and are not missing.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error5, allocate_error
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:)             :: x
    real(stnd), intent(in)                                 :: xmiss
    real(stnd), intent(in) ,    dimension(:,:)             :: xmean_grp
    real(stnd), intent(in),     dimension(:,:),   optional :: xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
    integer(i4b), intent(in),  optional    :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:,:), allocatable :: invstd
!
    integer(i4b) :: i, k, n, m
    integer      :: iok
!
    logical(lgl) :: standardize, dim_is_one
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_grp_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(ind),i4b),             &
                    name_proc )
!
    if ( present(xstd_grp) ) then
!
        m = assert_eq(  int(size(x,int(i)),i4b),      &
                        int(size(xmean_grp,1),i4b),   &
                        int(size(xstd_grp,1),i4b),    &
                        name_proc )
!
        call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                     logical(size(xstd_grp,2)==ngrp,lgl),   &
                     name_proc )
!
        standardize = true
!
    else
!
        m = assert_eq(  int(size(x,int(i)),i4b),      &
                        int(size(xmean_grp,1),i4b),   &
                        name_proc )
!
        call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                     name_proc )
!
        standardize = false
!
    end if
!
    if ( min(n,m,ngrp)<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.      &
               (m*n)>=omp_limit                .and.      &
               i>1_i4b                         .and.      &
               k>1_i4b
#endif
!
    if ( standardize ) then
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m,ngrp), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m,:ngrp) = one/xstd_grp(:m,:ngrp)
!
         if ( dim_is_one ) then
!
!$OMP PARALLEL DO IF(test_par)                   &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    where( x(:m,i)/=xmiss ) x(:m,i) = ( x(:m,i) - xmean_grp(:m,k) )*invstd(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)                   &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k) 
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    where( x(i,:m)/=xmiss ) x(i,:m) = ( x(i,:m) - xmean_grp(:m,k) )*invstd(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
         if ( dim_is_one ) then
!
!$OMP PARALLEL DO IF(test_par)                   &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k) 
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    where( x(:m,i)/=xmiss )  x(:m,i) = x(:m,i) - xmean_grp(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!$OMP PARALLEL DO IF(test_par)                   &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,k)  
!
            do i=1_i4b, n
!
                k = ind(i)
!
                if ( k>=1_i4b .and. k<=ngrp ) then
                    where( x(i,:m)/=xmiss )  x(i,:m) = x(i,:m) - xmean_grp(:m,k)
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_grp_miss_rm
! ________________________________________
!
    end subroutine comp_anoma_grp_miss_rm
!
! =========================================================================================
!
    subroutine comp_anoma_grp_miss_rt( x, ngrp, ind, xmiss, xmean_grp, xstd_grp )
!
! Purpose
! _______
!
!   COMP_ANOMA_GRP_MISS computes (standardized) anomalies by groups from a data matrix
!   possibly containing missing values.
!   
!
! Arguments
! _________
!
!   X          (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, input subarray containing size(X,3) observations
!              on size(X,1) by size(X,2) variables from the array of data
!              for which standardization by groups is desired.
!              If all the data are available at once, X can be the full data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group and this observation is not standardized.
!
!              The size of IND must verify size(IND) = size(X,3) .
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!   XMEAN_GRP  (INPUT) real(stnd), dimension(:,:,:)
!              On entry, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = NGRP.
!
!   XSTD_GRP   (INPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On entry, if XSTD_GRP is present, XSTD_GRP contains the standard deviations
!              for the NGRP groups of observations in the data array and the
!              observations are standardized.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1)
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = NGRP.
!
!
! Further Details
! _______________
!
!   It is assumed that elements of the array argument XMEAN_GRP are not missing.
!
!   It is assumed that elements of the array argument XSTD_GRP are greater than zero
!   and are not missing.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : allocate_error
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout),  dimension(:,:,:)             :: x
    real(stnd), intent(in)                                   :: xmiss
    real(stnd), intent(in) ,    dimension(:,:,:)             :: xmean_grp
    real(stnd), intent(in),     dimension(:,:,:),   optional :: xstd_grp
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),    dimension(:,:,:), allocatable :: invstd
!
    integer(i4b) :: i, j, k, n, m, p
    integer      :: iok
!
    logical(lgl) :: standardize
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_anoma_grp_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    if ( present(xstd_grp) ) then
!
        m = assert_eq(  int(size(x,1),i4b),           &
                        int(size(xmean_grp,1),i4b),   &
                        int(size(xstd_grp,1),i4b),    &
                        name_proc )
!
        p = assert_eq(  int(size(x,2),i4b),           &
                        int(size(xmean_grp,2),i4b),   &
                        int(size(xstd_grp,2),i4b),    &
                        name_proc )
!
        call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                     logical(size(xstd_grp,3)==ngrp,lgl),   &
                     name_proc )
!
        standardize = true
!
    else
!
        m = assert_eq(  int(size(x,1),i4b),           &
                        int(size(xmean_grp,1),i4b),   &
                        name_proc )
!
        p = assert_eq(  int(size(x,2),i4b),           &
                        int(size(xmean_grp,2),i4b),   &
                        name_proc )
!
        call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                     name_proc )
!
        standardize = false
!
    end if
!
    if ( min(n,m,p,ngrp)<=0_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    k = omp_get_max_threads()
    test_par = .not.( omp_in_parallel() )      .and.    &
               (m*p*n)>=omp_limit              .and.    &
               i>1_i4b                         .and.    &
               k>1_i4b
#endif
!
    if ( standardize ) then
!
!       ALLOCATE NEEDED WORKSPACE.
!
        allocate( invstd(m,p,ngrp), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        invstd(:m,:p,:ngrp) = one/xstd_grp(:m,:p,:ngrp)
!
!$OMP PARALLEL DO IF(test_par)                     &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,j,k)
!
        do i=1_i4b, n
!
            k = ind(i)
!
            if ( k>=1_i4b .and. k<=ngrp ) then
!
               do j=1_i4b, p
!
                   where( x(:m,j,i)/=xmiss ) x(:m,j,i) = ( x(:m,j,i) - xmean_grp(:m,j,k) )*invstd(:m,j,k)
!
               end do
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
!       DEALLOCATE WORKSPACE.
!
        deallocate( invstd )
!
    else
!
!$OMP PARALLEL DO IF(test_par)                     &
!$OMP            ,SCHEDULE(STATIC), PRIVATE(i,j,k)
!
        do i=1_i4b, n
!
            k = ind(i)
!
            if ( k>=1_i4b .and. k<=ngrp ) then
!
               do j=1_i4b, p
!
                   where( x(:m,j,i)/=xmiss ) x(:m,j,i) = x(:m,j,i) - xmean_grp(:m,j,k)
!
               end do
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_anoma_grp_miss_rt
! ________________________________________
!
    end subroutine comp_anoma_grp_miss_rt
!
!
! =========================================================================================
!                                 COMPOSITE ANALYSIS
! =========================================================================================
!
!
    subroutine comp_composite_rv( x, first, last, ngrp, ind, xmean, xstd, xn,        &
                                  xmean_grp, xstd_grp, xn_grp,                       &
                                  xcomp, u, prob, utest )
!
! Purpose
! _______
!
!   COMP_COMPOSITE computes a composite analysis from a data vector.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:)
!              On entry, input subvector containing size(X) observations
!              from the vector of data for which a composite analysis is
!              desired. If all the data are available at once, X can be
!              the full data vector.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subvector is the first subvector 
!                of the data vector.
!              - FIRST = false the current subvector is not the first subvector
!                of the data vector.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subvector is the last subvector 
!                             of the data vector.
!              - LAST = false the current subvector is not the last subvector
!                             of the data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups in the analysis.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X) .
!
!   XMEAN      (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XMEAN contains the mean value from previous calls to COMP_COMPOSITE.
!              XMEAN should not be changed between calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XMEAN contains the mean value of the data 
!              vector.
!
!   XSTD       (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XSTD contains adjusted sum of squares from previous calls
!              to COMP_COMPOSITE. XSTD should not be changed between
!              calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XSTD contains the standard deviation of
!              the data vector.
!
!   XN         (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XN contains count of observations from previous calls
!              to COMP_COMPOSITE. XN should not be changed between calls to
!              COMP_COMPOSITE.
!
!              On exit, when LAST=true, XN contains the number of observations 
!              in the data vector.
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_COMPOSITE. XMEAN_GRP should not be changed between calls 
!              to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data vector.
!
!              The size of XMEAN_GRP must verify size(XMEAN_GRP) = NGRP .
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_COMPOSITE. XSTD_GRP should not be changed between calls 
!              to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations in the data vector.
!
!              The size of XSTD_GRP must verify size(XSTD_GRP) = NGRP .
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XN_GRP contains counts of observations for the NGRP groups
!              from previous calls to COMP_COMPOSITE. XN_GRP should not be 
!              changed between calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of observations
!              in the NGRP groups of observations in the data vector.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!   XCOMP      (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the standardized profiles of the NGRP groups
!              of observations in the data vector.
!
!              The size of XCOMP must verify size(XCOMP) = NGRP .
!
!   U          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the U statistics for the NGRP groups of
!              observations in the data vector.
!
!              The size of U must verify size(U) = NGRP .
!
!   PROB       (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the significance probabilities of the U
!              statistics under the null hypothesis that the groups
!              have been formed by uniform random sampling without 
!              duplication in the set of all the observations.
!
!              The size of PROB must verify size(PROB) = NGRP .
!
!   UTEST      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, a probability. UTEST is the sum of the areas (equal) 
!              in both tails of the normal-distribution.
!              UTEST must verify   0. < P < 1.
!
!              On exit, the two-tail quantile of the normal-distribution, that is a
!              value X such that the probability of the absolute value of U being
!              greater than X is UTEST under the null hypothesis that the groups
!              have been formed by uniform random sampling without duplication in 
!              the set of all the observations.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one observation is present, the statistics are set to Nan code.
!
!   The optional parameters need only to be specified when LAST=true.
!
!   For more details, on the statistics and tests computed by this subroutine, see:
!
!   (1) Terray, P., Delecluse, P., Labattu, S., Terray, L., 2003: Sea Surface
!           Temperature associations with the Late Indian Summer Monsoon,
!           Climate Dynamics, vol. 21, 593-618.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error3, stat_error4, allocate_error
    use Utilities,         only : assert, assert_eq, merror
    use Prob_Procedures,   only : probn, pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),               dimension(:)  :: x
    real(stnd), intent(inout)                           :: xn, xmean, xstd
    real(stnd), intent(inout),            dimension(:)  :: xn_grp, xmean_grp, xstd_grp 
    real(stnd), intent(inout), optional                 :: utest
    real(stnd), intent(out),   optional , dimension(:)  :: xcomp, u, prob
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: di, dfi
    real(stnd), dimension(:), allocatable   :: work
!
    integer(i4b) :: i, k, n
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_composite'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),         &
                    int(size(ind),i4b),       &
                    name_proc )
!
    call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                 logical(size(xstd_grp)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xmean, xstd, xn, xmean_grp, xstd_grp, xn_grp .
!
        xmean = zero       
        xstd  = zero       
        xn    = zero       
!
        xmean_grp(:ngrp) = zero       
        xstd_grp(:ngrp)  = zero       
        xn_grp(:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        xn = xn + one
        di = x(i) - xmean
        dfi = di/xn
        xmean = xmean + dfi
        xstd = xstd + di*di - di*dfi
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        xn_grp(k) = xn_grp(k) + one
!
        di  = x(i) - xmean_grp(k)
        dfi = di/xn_grp(k)
!
        xmean_grp(k) = xmean_grp(k) + dfi
        xstd_grp(k)  = xstd_grp(k) + di*di - di*dfi
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(xcomp) ) call assert( logical(size(xcomp)==ngrp,lgl), name_proc )
!
        if ( present(u) ) call assert( logical(size(u)==ngrp,lgl), name_proc )
!
        if ( present(prob) ) call assert( logical(size(prob)==ngrp,lgl), name_proc )
!
        if ( present(utest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF NORMAL DISTRIBUTION.
!
            if ( zero>=utest .or. utest>=one ) then
                call merror( name_proc//stat_error3 )
            else
                di = utest*half
                utest = abs( pinvn( di ) )
            end if
!
        end if
!
        if ( xn>zero ) then
!
            if ( any(xn_grp(:ngrp)==zero) ) call merror( name_proc//stat_error4 )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            di = xstd/xn
            xstd = sqrt( di )
!
            xstd_grp(:ngrp) = sqrt( xstd_grp(:ngrp)/xn_grp(:ngrp) )
!
            if ( present(xcomp) .or. present(u) .or. present(prob) ) then
!
!               COMPUTE STANDARDIZED PROFILES OF THE GROUPS.
!
                allocate( work(ngrp), stat=iok )
!
                if ( iok/=0 ) call merror( name_proc//allocate_error )
!
                if ( xstd>=lamch( 's' )  )  then
                    di = one/xstd
                else
                    di = zero
                end if
!
                work(:ngrp) = (xmean_grp(:ngrp) - xmean)*di
!
                if ( present(xcomp) ) then
!
!                   OUTPUT STANDARDIZED PROFILES OF THE GROUPS.
!
                    xcomp(:ngrp) = work(:ngrp)
!
                end if
!
                if ( present(u) .or. present(prob)  ) then
!
!                   COMPUTE U STATISTICS OF THE GROUPS.
!
                    di = xn - one
                    where ( xn_grp(:ngrp)<xn )                   &
                    work(:ngrp) = sqrt( ( xn_grp(:ngrp)*di )/( xn - xn_grp(:ngrp) ) )*work(:ngrp)
!
                    if ( present(u) ) then
!
!                       OUTPUT U STATISTICS OF THE GROUPS.
!
                        u(:ngrp) = work(:ngrp)
!
                    end if
!
                    if ( present(prob) ) then
!
!                       TEST U STATISTICS OF THE GROUPS.
!
                        work(:ngrp) = abs( work(:ngrp) )
                        prob(:ngrp) = two*probn( work(:ngrp), upper=true )
!
                    end if
!
                end if
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!
            end if
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            di = nan()
!
            xmean            = di
            xstd             = di
            xmean_grp(:ngrp) = di
            xstd_grp(:ngrp)  = di
!
            if ( present(xcomp) ) xcomp(:ngrp) = di
            if ( present(u) )     u(:ngrp)     = di
            if ( present(prob) )  prob(:ngrp)  = di
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_composite_rv
! ___________________________________
!
    end subroutine comp_composite_rv
!
! =========================================================================================
!
    subroutine comp_composite_rm( x, first, last, ngrp, ind, xmean, xstd, xn,        &
                                  xmean_grp, xstd_grp, xn_grp,                       &
                                  dimvar, xcomp, u, prob, utest )
!
! Purpose
! _______
!
!   COMP_COMPOSITE computes a composite analysis from a data matrix.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:)
!              On entry, input submatrix containing size(X,3-DIMVAR)
!              observations on size(X,DIMVAR) variables from the matrix of data
!              for which a composite analysis is desired. By default, 
!              DIMVAR is equal to 1. See description of optional DIMVAR
!              argument for details.
!              If all the data are available at once, X can be the full 
!              data matrix.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current submatrix is the first submatrix 
!                of the data matrix.
!              - FIRST = false the current submatrix is not the first submatrix
!                of the data matrix.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current submatrix is the last submatrix 
!                             of the data matrix.
!              - LAST = false the current submatrix is not the last submatrix
!                             of the data matrix.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups in the analysis.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3-DIMVAR) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3-DIMVAR) .
!
!   XMEAN      (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XMEAN contains the mean values from previous calls to COMP_COMPOSITE.
!              XMEAN should not be changed between calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XMEAN contains the mean values of the data 
!              matrix.
!
!              The size of XMEAN must verify size(XMEAN) = size(X,DIMVAR).
!
!   XSTD       (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XSTD contains adjusted sums of squares from previous calls
!              to COMP_COMPOSITE. XSTD should not be changed between
!              calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XSTD contains the standard deviations of 
!              the data matrix.
!
!              The size of XSTD must verify size(XSTD) = size(X,DIMVAR).
!
!   XN         (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XN contains count of observations from previous calls
!              to COMP_COMPOSITE. XN should not be changed between calls to
!              COMP_COMPOSITE.
!
!              On exit, XN contains the number of observations in the data matrix.
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_COMPOSITE. XMEAN_GRP should not be changed between calls 
!              to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data matrix.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,DIMVAR)
!              - size(XMEAN_GRP,2) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_COMPOSITE. XSTD_GRP should not be changed between calls 
!              to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data matrix.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,DIMVAR)
!              - size(XSTD_GRP,2) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XN_GRP contains counts of observations for the NGRP groups
!              from previous calls to COMP_COMPOSITE. XN_GRP should not be 
!              changed between calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of observations
!              in the NGRP groups for all the variables in the data matrix.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!   DIMVAR     (INPUT, OPTIONAL) integer(i4b)
!              On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!              - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!                on size(X,1) variables.
!              - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!                on size(X,2) variables.
!
!              The default is DIMVAR = 1.
!
!   XCOMP      (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the standardized profiles of the NGRP groups of
!              observations in the data matrix.
!
!              The shape of XCOMP must verify:
!
!              - size(XCOMP,1) = size(X,DIMVAR)
!              - size(XCOMP,2) = NGRP.
!
!   U          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the U statistics for the NGRP groups and all
!              the variables in the data matrix.
!
!              The shape of U must verify:
!
!              - size(U,1) = size(X,DIMVAR)
!              - size(U,2) = NGRP.
!
!   PROB       (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the significance probabilities of the U
!              statistics under the null hypothesis that the groups
!              have been formed by uniform random sampling without 
!              duplication in the set of all the observations.
!
!              The shape of PROB must verify:
!
!              - size(PROB,1) = size(X,DIMVAR)
!              - size(PROB,2) = NGRP.
!
!   UTEST      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, a probability. UTEST is the sum of the areas (equal) 
!              in both tails of the normal-distribution.
!              UTEST must verify   0. < P < 1.
!
!              On exit, the two-tail quantile of the normal-distribution, that is a
!              value X such that the probability of the absolute value of U being
!              greater than X is UTEST under the null hypothesis that the groups
!              have been formed by uniform random sampling without duplication in 
!              the set of all the observations.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one observation is present, the statistics are set to Nan code.
!
!   The optional parameters, except DIMVAR, need only to be specified when LAST=true.
!
!   For more details, on the statistics and tests computed by this subroutine, see:
!
!   (1) Terray, P., Delecluse, P., Labattu, S., Terray, L., 2003: Sea Surface
!           Temperature associations with the Late Indian Summer Monsoon,
!           Climate Dynamics, vol. 21, 593-618.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error3, stat_error4, stat_error5, allocate_error
    use Utilities,         only : assert, assert_eq, merror
    use Prob_Procedures,   only : probn, pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),               dimension(:,:) :: x
    real(stnd), intent(inout)                            :: xn
    real(stnd), intent(inout),            dimension(:)   :: xmean, xstd, xn_grp
    real(stnd), intent(inout),            dimension(:,:) :: xmean_grp, xstd_grp
    real(stnd), intent(inout), optional                  :: utest
    real(stnd), intent(out),   optional , dimension(:,:) :: xcomp, u, prob
!
    integer(i4b), intent(in)                          :: ngrp
    integer(i4b), intent(in),            dimension(:) :: ind
    integer(i4b), intent(in),  optional               :: dimvar
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: xnm1, tmp
    real(stnd), dimension(size(xmean))          :: di, di2, xi
    real(stnd), dimension(:,:),   allocatable   :: work
    real(stnd), dimension(:),     allocatable   :: df
!
    integer(i4b) :: i, k, n, m
    integer      :: iok
!
    logical(lgl) :: dim_is_one
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_composite'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(ind),i4b),             &
                    name_proc )
!
    m = assert_eq(  int( (/ size(x,int(i)),               &
                            size(xmean),                  &
                            size(xstd),                   &
                            size(xmean_grp,1),            &
                            size(xstd_grp,1) /), i4b ),   &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                 logical(size(xstd_grp,2)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),       &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xstd, xn, xmean_grp, xstd_grp, xn_grp .
!
        xmean(:m) = zero       
        xstd(:m) = zero       
        xn       = zero       
!
        xmean_grp(:m,:ngrp) = zero       
        xstd_grp(:m,:ngrp)  = zero       
        xn_grp(:ngrp)       = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        xnm1 = xn
        xn = xnm1 + one
!
        if ( dim_is_one ) then
            xi(:m) = x(:m,i)
        else
            xi(:m) = x(i,:m)
        end if
!
        di(:m)  = xi(:m) - xmean(:m)
        di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
        tmp = one/xn
!
        call axpy( m, tmp, di(:m), 1_i4b, xmean(:m), 1_i4b )
!
        tmp = xnm1/xn
!
        call axpy( m, tmp, di2(:m), 1_i4b, xstd(:m), 1_i4b )
#else
        xmean(:m) = xmean(:m) + di(:m)*( one/xn )
        xstd(:m)  = xstd(:m) + di2(:m)*( xnm1/xn )
#endif
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        xnm1 = xn_grp(k)
        xn_grp(k) = xnm1 + one
!
        di(:m)  = xi(:m) - xmean_grp(:m,k)
        di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
        tmp = one/xn_grp(k)
!
        call axpy( m, tmp, di(:m), 1_i4b, xmean_grp(:m,k), 1_i4b )
!
        tmp = xnm1/xn_grp(k)
!
        call axpy( m, tmp, di2(:m), 1_i4b, xstd_grp(:m,k), 1_i4b )
#else
        xmean_grp(:m,k) = xmean_grp(:m,k) + di(:m)*( one/xn_grp(k) )
        xstd_grp(:m,k)  = xstd_grp(:m,k) + di2(:m)*( xnm1/xn_grp(k) )
#endif
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(xcomp) ) call assert( logical(size(xcomp,2)==ngrp,lgl),    &
                                           logical(size(xcomp,1)==m,lgl),       &
                                           name_proc )
!
        if ( present(u) ) call assert( logical(size(u,2)==ngrp,lgl),    &
                                       logical(size(u,1)==m,lgl),       &
                                       name_proc )
!
        if ( present(prob) ) call assert( logical(size(prob,2)==ngrp,lgl),    &
                                          logical(size(prob,1)==m,lgl),       &
                                          name_proc )
!
        if ( present(utest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF NORMAL DISTRIBUTION.
!
            if ( zero>=utest .or. utest>=one ) then
                call merror( name_proc//stat_error3 )
            else
                tmp = utest*half
                utest = abs( pinvn( tmp ) )
            end if
!
        end if
!
        if ( xn>zero ) then
!
            if ( any(xn_grp(:ngrp)==zero) ) call merror( name_proc//stat_error4 )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            di(:m) = xstd(:m)/xn
            xstd(:m) = sqrt( di(:m) )
!
            do k = 1_i4b, ngrp
!
                xstd_grp(:m,k) = sqrt( xstd_grp(:m,k)/xn_grp(k) )
!
            end do
!
            if ( present(xcomp) .or. present(u) .or. present(prob) ) then
!
!               COMPUTE STANDARDIZED PROFILES OF THE GROUPS.
!
                if ( present(u) .or. present(prob)  ) then
                    allocate( work(m,ngrp), df(ngrp), stat=iok )
                else
                    allocate( work(m,ngrp), stat=iok )
                end if
!
                if ( iok/=0 ) call merror( name_proc//allocate_error )
!
                tmp = lamch( 's' )
!
                where ( xstd(:m)>=tmp )
                    di(:m) = one/xstd(:m)
                elsewhere
                    di(:m) = zero
                end where
!
                do k = 1_i4b, ngrp
!
                    work(:m,k) = ( xmean_grp(:m,k) - xmean(:m) )*di(:m)
!
                end do
!
                if ( present(xcomp) ) then
!
!                   OUTPUT STANDARDIZED PROFILES OF THE GROUPS.
!
                    xcomp(:m,:ngrp) = work(:m,:ngrp)
!
                end if
!
                if ( present(u) .or. present(prob)  ) then
!
!                   COMPUTE U STATISTICS OF THE GROUPS.
!
                    xnm1 = xn - one
!
                    where ( xn_grp(:ngrp)<xn )
                        df(:ngrp) = sqrt( ( xn_grp(:ngrp)*xnm1 )/( xn - xn_grp(:ngrp) ) )
                    elsewhere
                        df(:ngrp) = zero
                    end where
!
                    do k = 1_i4b, ngrp
!
                        work(:m,k) = work(:m,k)*df(k)
!
                    end do
!
                    if ( present(u) ) then
!
!                       OUTPUT U STATISTICS OF THE GROUPS.
!
                        u(:m,:ngrp) = work(:m,:ngrp)
!
                    end if
!
                    if ( present(prob) ) then
!
!                       TEST U STATISTICS OF THE GROUPS.
!
                        work(:m,:ngrp) = abs( work(:m,:ngrp) )
                        prob(:m,:ngrp) = two*probn( work(:m,:ngrp), upper=true )
!
                    end if
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( work, df )
!
                else
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( work )
!
                end if
!
            end if
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            tmp = nan()
!
            xmean(:m)           = tmp
            xstd(:m)            = tmp
            xmean_grp(:m,:ngrp) = tmp
            xstd_grp(:m,:ngrp)  = tmp
!
            if ( present(xcomp) ) xcomp(:m,:ngrp) = tmp
            if ( present(u) )     u(:m,:ngrp)     = tmp
            if ( present(prob) )  prob(:m,:ngrp)  = tmp
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_composite_rm
! ___________________________________
!
    end subroutine comp_composite_rm
!
! =========================================================================================
!
    subroutine comp_composite_rt( x, first, last, ngrp, ind, xmean, xstd, xn,             &
                                  xmean_grp, xstd_grp, xn_grp,                            &
                                  xcomp, u, prob, utest )
!
! Purpose
! _______
!
!   COMP_COMPOSITE computes a composite analysis from a data tridimensional array.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:,:)
!              On entry, input subarray containing size(X,3) observations
!              on size(X,1) by size(X,2) variables from the array of data for 
!              which a composite analysis is desired. If all the data are available
!              at once, X can be the full data array.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subarray is the first subarray 
!                of the data array.
!              - FIRST = false the current subarray is not the first subarray
!                of the data array.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subarray is the last subarray 
!                             of the data array.
!              - LAST = false the current subarray is not the last subarray
!                             of the data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups in the analysis.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3) .
!
!   XMEAN      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XMEAN contains the mean values from previous calls to COMP_COMPOSITE.
!              XMEAN should not be changed between calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XMEAN contains the mean values of the data array.
!
!              The shape of XMEAN must verify:
!
!              - size(XMEAN,1) = size(X,1)
!              - size(XMEAN,2) = size(X,2).
!
!   XSTD       (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XSTD contains adjusted sums of squares from previous calls to COMP_COMPOSITE.
!              XSTD should not be changed between calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XSTD contains the standard deviations of
!              the data array.
!
!              The shape of XSTD must verify:
!
!              - size(XSTD,1) = size(X,1)
!              - size(XSTD,2) = size(X,2).
!
!   XN         (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XN contains counts of observations from previous calls
!              to COMP_COMPOSITE. XN should not be changed between calls to
!              COMP_COMPOSITE.
!
!              On exit, when LAST=true, XN contains the number of observations 
!              in the data array.
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_COMPOSITE. XMEAN_GRP should not be changed between calls 
!              to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_COMPOSITE. XSTD_GRP should not be changed between calls 
!              to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1)
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE (e.g. when FIRST=true),
!              XN_GRP contains counts of  observations for the NGRP groups
!              from previous calls to COMP_COMPOSITE. XN_GRP should not be 
!              changed between calls to COMP_COMPOSITE.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of observations  
!              in the NGRP groups for all the variables in the data array.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!   XCOMP      (OUTPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On exit, the standardized profiles of the NGRP groups of
!              observations in the data array.
!
!              The shape of XCOMP must verify:
!
!              - size(XCOMP,1) = size(X,1)
!              - size(XCOMP,2) = size(X,2)
!              - size(XCOMP,3) = NGRP.
!
!   U          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On exit, the U statistics for the NGRP groups and all
!              the variables in the data array.
!
!              The shape of U must verify:
!
!              - size(U,1) = size(X,1)
!              - size(U,2) = size(X,2)
!              - size(U,3) = NGRP.
!
!   PROB       (OUTPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On exit, the significance probabilities of the U
!              statistics under the null hypothesis that the groups
!              have been formed by uniform random sampling without
!              duplication in the set of all the observations.
!
!              The shape of PROB must verify:
!
!              - size(PROB,1) = size(X,1)
!              - size(PROB,2) = size(X,2)
!              - size(PROB,3) = NGRP.
!
!   UTEST      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, a probability. UTEST is the sum of the areas (equal) 
!              in both tails of the normal-distribution.
!              UTEST must verify   0. < P < 1.
!
!              On exit, the two-tail quantile of the normal-distribution, that is a
!              value X such that the probability of the absolute value of U being
!              greater than X is UTEST under the null hypothesis that the groups
!              have been formed by uniform random sampling without duplication in
!              the set of all the observations.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one observation is present, the statistics are set to Nan code.
!
!   The optional parameters need only to be specified when LAST=true.
!
!   For more details, on the statistics and tests computed by this subroutine, see:
!
!   (1) Terray, P., Delecluse, P., Labattu, S., Terray, L., 2003: Sea Surface
!           Temperature associations with the Late Indian Summer Monsoon,
!           Climate Dynamics, vol. 21, 593-618.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error3, stat_error4, allocate_error
    use Utilities,         only : assert, assert_eq, merror
    use Prob_Procedures,   only : probn, pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),               dimension(:,:,:) :: x
    real(stnd), intent(inout)                              :: xn
    real(stnd), intent(inout),            dimension(:)     :: xn_grp
    real(stnd), intent(inout),            dimension(:,:)   :: xmean, xstd
    real(stnd), intent(inout),            dimension(:,:,:) :: xmean_grp, xstd_grp
    real(stnd), intent(inout), optional                    :: utest
    real(stnd), intent(out),   optional , dimension(:,:,:) :: xcomp, u, prob
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                :: xnm1, xn_grpm1, tmp
    real(stnd), dimension(size(x,1))          :: di, di2
    real(stnd), dimension(:,:,:), allocatable :: work
    real(stnd), dimension(:),     allocatable :: df
!
    integer(i4b) :: i, j, k, n, m, p
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_composite'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    m = assert_eq(  int( (/ size(x,1),                    &
                            size(xmean,1),                &
                            size(xstd,1),                 &
                            size(xmean_grp,1),            &
                            size(xstd_grp,1) /), i4b),    &
                            name_proc )
!
    p = assert_eq(  int( (/ size(x,2),                    &
                            size(xmean,2),                &
                            size(xstd,2),                 &
                            size(xmean_grp,2),            &
                            size(xstd_grp,2) /), i4b),    &
                            name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                 logical(size(xstd_grp,3)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),       &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean, xstd, xn, xmean_grp, xstd_grp, xn_grp .
!
        xmean(:m,:p) = zero       
        xstd(:m,:p)  = zero       
        xn           = zero       
!
        xmean_grp(:m,:p,:ngrp) = zero       
        xstd_grp(:m,:p,:ngrp)  = zero       
        xn_grp(:ngrp)          = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        xnm1 = xn
        xn = xnm1 + one
!
        k = ind(i)
!
        if ( k>=1_i4b .and. k<=ngrp ) then
!
            xn_grpm1 = xn_grp(k)
            xn_grp(k) = xn_grpm1 + one
!
            do j = 1_i4b, p
!
                di(:m)  = x(:m,j,i) - xmean(:m,j)
                di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
                tmp = one/xn
!
                call axpy( m, tmp, di(:m), 1_i4b, xmean(:m,j), 1_i4b )
!
                tmp = xnm1/xn
!
                call axpy( m, tmp, di2(:m), 1_i4b, xstd(:m,j), 1_i4b )
#else
                xmean(:m,j) = xmean(:m,j) + di(:m)*( one/xn )
                xstd(:m,j) = xstd(:m,j) + di2(:m)*( xnm1/xn )
#endif
!
                di(:m)  = x(:m,j,i) - xmean_grp(:m,j,k)
                di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
                tmp = one/xn_grp(k)
!
                call axpy( m, tmp, di(:m), 1_i4b, xmean_grp(:m,j,k), 1_i4b )
!
                tmp = xn_grpm1/xn_grp(k)
!
                call axpy( m, tmp, di2(:m), 1_i4b, xstd_grp(:m,j,k), 1_i4b )
#else
                xmean_grp(:m,j,k) = xmean_grp(:m,j,k) + di(:m)*( one/xn_grp(k) )
                xstd_grp(:m,j,k)  = xstd_grp(:m,j,k) + di2(:m)*( xn_grpm1/xn_grp(k) )
#endif
!
            end do
!
        else
!
            do j = 1_i4b, p
!
                di(:m)  = x(:m,j,i) - xmean(:m,j)
                di2(:m) = di(:m)*di(:m)
!
#ifdef _BLAS
                tmp = one/xn
!
                call axpy( m, tmp, di(:m), 1_i4b, xmean(:m,j), 1_i4b )
!
                tmp = xnm1/xn
!
                call axpy( m, tmp, di2(:m), 1_i4b, xstd(:m,j), 1_i4b )
#else
                xmean(:m,j) = xmean(:m,j) + di(:m)*( one/xn )
                xstd(:m,j) = xstd(:m,j) + di2(:m)*( xnm1/xn )
#endif
!
            end do
!
        end if
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(xcomp) ) call assert( logical(size(xcomp,3)==ngrp,lgl),    &
                                           logical(size(xcomp,2)==p,lgl),       &
                                           logical(size(xcomp,1)==m,lgl),       &
                                           name_proc )
!
        if ( present(u) ) call assert( logical(size(u,3)==ngrp,lgl),    &
                                       logical(size(u,2)==p,lgl),       &
                                       logical(size(u,1)==m,lgl),       &
                                       name_proc )
!
        if ( present(prob) ) call assert( logical(size(prob,3)==ngrp,lgl),    &
                                          logical(size(prob,2)==p,lgl),       &
                                          logical(size(prob,1)==m,lgl),       &
                                          name_proc )
!
        if ( present(utest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF NORMAL DISTRIBUTION.
!
            if ( zero>=utest .or. utest>=one ) then
                call merror( name_proc//stat_error3 )
            else
                tmp = utest*half
                utest = abs( pinvn( tmp ) )
            end if
!
        end if
!
        if ( xn>zero ) then
!
            if ( any(xn_grp(:ngrp)==zero) ) call merror( name_proc//stat_error4 )
!
!           COMPUTE STANDARD-DEVIATIONS.
!
            do j = 1_i4b, p
!
                di(:m) = xstd(:m,j)/xn
                xstd(:m,j) = sqrt( di(:m) )
!
            end do
!
            do k = 1_i4b, ngrp
!
                tmp = one/xn_grp(k)
!
                do j = 1_i4b, p
!
                    di(:m) = xstd_grp(:m,j,k)*tmp
                    xstd_grp(:m,j,k) = sqrt( di(:m) )
!
                end do
!
            end do
!
            if ( present(xcomp) .or. present(u) .or. present(prob) ) then
!
!               COMPUTE STANDARDIZED PROFILES OF THE GROUPS.
!
                if ( present(u) .or. present(prob)  ) then
                    allocate( work(m,p,ngrp), df(ngrp), stat=iok )
                else
                    allocate( work(m,p,ngrp), stat=iok )
                end if
!
                if ( iok/=0 ) call merror( name_proc//allocate_error )
!
                tmp = lamch( 's' )
!
                do k = 1_i4b, ngrp
!
                    do j = 1_i4b, p
!
                        where ( xstd(:m,j)>=tmp )
                            work(:m,j,k) = ( xmean_grp(:m,j,k) - xmean(:m,j) )/xstd(:m,j)
                        elsewhere
                            work(:m,j,k) = zero
                        end where
!
                    end do
!
                end do
!
                if ( present(xcomp) ) then
!
!                   OUTPUT STANDARDIZED PROFILES OF THE GROUPS.
!
                    xcomp(:m,:p,:ngrp) = work(:m,:p,:ngrp)
!
                end if
!
                if ( present(u) .or. present(prob)  ) then
!
!                   COMPUTE U STATISTICS OF THE GROUPS.
!
                    xnm1 = xn - one
!
                    where ( xn_grp(:ngrp)<xn )
                        df(:ngrp) = sqrt( ( xn_grp(:ngrp)*xnm1 )/( xn - xn_grp(:ngrp) ) )
                    elsewhere
                        df(:ngrp) = zero
                    end where
!
                    do k = 1_i4b, ngrp
!
                        work(:m,:p,k) = df(k)*work(:m,:p,k)
!
                    end do
!
                    if ( present(u) ) then
!
!                       OUTPUT U STATISTICS OF THE GROUPS.
!
                        u(:m,:p,:ngrp) = work(:m,:p,:ngrp)
!
                    end if
!
                    if ( present(prob) ) then
!
!                       TEST U STATISTICS OF THE GROUPS.
!
                        work(:m,:p,:ngrp) = abs( work(:m,:p,:ngrp) )
!
                        do k = 1_i4b, ngrp
!
                            prob(:m,:p,k) = two*probn( work(:m,:p,k), upper=true )
!
                        end do
!
                    end if
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( work, df )
!
                else
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( work )
!
                end if
!
            end if
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            tmp = nan()
!
            xmean(:m,:p)           = tmp
            xstd(:m,:p)            = tmp
            xmean_grp(:m,:p,:ngrp) = tmp
            xstd_grp(:m,:p,:ngrp)  = tmp
!
            if ( present(xcomp) ) xcomp(:m,:p,:ngrp) = tmp
            if ( present(u) )     u(:m,:p,:ngrp)     = tmp
            if ( present(prob) )  prob(:m,:p,:ngrp)  = tmp
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_composite_rt
! ___________________________________
!
    end subroutine comp_composite_rt
!
! =========================================================================================
!
    subroutine comp_composite_miss_rv( x, first, last, ngrp, ind, xmean, xstd, xn,       &
                                       xmean_grp, xstd_grp, xn_grp, xmiss,               &
                                       xcomp, u, prob, utest )
!
! Purpose
! _______
!
!   COMP_COMPOSITE_MISS computes a composite analysis from a data vector
!   possibly containing missing values.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:)
!              On entry, input subvector containing size(X) observations
!              from the vector of data for which a composite analysis is
!              desired. If all the data are available at once, X can be
!              the full data vector.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subvector is the first subvector 
!                of the data vector.
!              - FIRST = false the current subvector is not the first subvector
!                of the data vector.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subvector is the last subvector 
!                             of the data vector.
!              - LAST = false the current subvector is not the last subvector
!                             of the data vector.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups in the analysis.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X) .
!
!   XMEAN      (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XMEAN contains the mean value from previous calls to COMP_COMPOSITE_MISS.
!              XMEAN should not be changed between calls to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XMEAN contains the mean value of the data 
!              vector.
!
!   XSTD       (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XSTD contains adjusted sum of squares from previous calls
!              to COMP_COMPOSITE_MISS. XSTD should not be changed between
!              calls to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XSTD contains the standard deviation of
!              the data vector.
!
!   XN         (INPUT/OUTPUT) real(stnd)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XN contains count of non-missing observations from previous calls
!              to COMP_COMPOSITE_MISS. XN should not be changed between calls to
!              COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XN contains the number of observations 
!              in the data vector.
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_COMPOSITE_MISS. XMEAN_GRP should not be changed between calls 
!              to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP
!              groups of observations in the data vector.
!
!              The size of XMEAN_GRP must verify size(XMEAN_GRP) = NGRP .
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_COMPOSITE_MISS. XSTD_GRP should not be changed between calls 
!              to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations in the data vector.
!
!              The size of XSTD_GRP must verify size(XSTD_GRP) = NGRP .
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XN_GRP contains counts of non-missing observations for the NGRP groups
!              from previous calls to COMP_COMPOSITE_MISS. XN_GRP should not be 
!              changed between calls to COMP_COMPOSITE_MISS.
!
!              On exit, XN_GRP contains the number of non-missing observations
!              in the NGRP groups of observations in the data vector.
!
!              The size of XN_GRP must verify size(XN_GRP) = NGRP .
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!   XCOMP      (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the standardized profiles of the NGRP groups.
!
!              The size of XCOMP must verify size(XCOMP) = NGRP .
!
!   U          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the U statistics for the NGRP groups of
!              observations in the data vector.
!
!              The size of U must verify size(U) = NGRP .
!
!   PROB       (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!              On exit, the significance probabilities of the U
!              statistics under the null hypothesis that the groups
!              have been formed by uniform random sampling without 
!              duplication in the set of all the observations.
!
!              The size of PROB must verify size(PROB) = NGRP .
!
!   UTEST      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, a probability. UTEST is the sum of the areas (equal) 
!              in both tails of the normal-distribution.
!              UTEST must verify   0. < P < 1.
!
!              On exit, the two-tail quantile of the normal-distribution, that is a
!              value X such that the probability of the absolute value of U being
!              greater than X is UTEST under the null hypothesis that the groups
!              have been formed by uniform random sampling without duplication in 
!              the set of all the observations.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one valid observation were present, the statistics are set to Nan code.
!
!   If fewer than one valid observation were present only for some groups
!   of observations, the pertinent statistics are set to missing (XMISS value).
!
!   The optional parameters need only to be specified when LAST=true.
!
!   For more details, on the statistics and tests computed by this subroutine, see:
!
!   (1) Terray, P., Delecluse, P., Labattu, S., Terray, L., 2003: Sea Surface
!           Temperature associations with the Late Indian Summer Monsoon,
!           Climate Dynamics, vol. 21, 593-618.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error3, allocate_error
    use Utilities,         only : assert, assert_eq, merror
    use Prob_Procedures,   only : probn, pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: xmiss
    real(stnd), intent(in),               dimension(:)  :: x
    real(stnd), intent(inout)                           :: xn, xmean, xstd
    real(stnd), intent(inout),            dimension(:)  :: xn_grp, xmean_grp, xstd_grp 
    real(stnd), intent(inout), optional                 :: utest
    real(stnd), intent(out),   optional , dimension(:)  :: xcomp, u, prob
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: di, dfi
    real(stnd), dimension(:), allocatable   :: work
!
    integer(i4b) :: i, k, n
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_composite'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),         &
                    int(size(ind),i4b),       &
                    name_proc )
!
    call assert( logical(size(xmean_grp)==ngrp,lgl),  &
                 logical(size(xstd_grp)==ngrp,lgl),   &
                 logical(size(xn_grp)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xmean, xstd, xn, xmean_grp, xstd_grp, xn_grp .
!
        xmean = zero       
        xstd  = zero       
        xn    = zero       
!
        xmean_grp(:ngrp) = zero       
        xstd_grp(:ngrp)  = zero       
        xn_grp(:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
        if ( x(i)==xmiss ) cycle
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        xn = xn + one
        di = x(i) - xmean
        dfi = di/xn
!
        xmean = xmean + dfi
        xstd = xstd + di*di - di*dfi
!
        k = ind(i)
!
        if ( k<1_i4b .or. k>ngrp ) cycle
!
        xn_grp(k) = xn_grp(k) + one
        di = x(i) - xmean_grp(k)
        dfi = di/xn_grp(k)
!
        xmean_grp(k) = xmean_grp(k) + dfi
        xstd_grp(k) = xstd_grp(k) + di*di - di*dfi
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(xcomp) ) call assert( logical(size(xcomp)==ngrp,lgl), name_proc )
!
        if ( present(u) ) call assert( logical(size(u)==ngrp,lgl), name_proc )
!
        if ( present(prob) ) call assert( logical(size(prob)==ngrp,lgl), name_proc )
!
        if ( present(utest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF NORMAL DISTRIBUTION.
!
            if ( zero>=utest .or. utest>=one ) then
                call merror( name_proc//stat_error3 )
            else
                di = utest*half
                utest = abs( pinvn( di ) )
            end if
!
        end if
!
        if ( xn>zero ) then
!
!           COMPUTE STANDARD-DEVIATIONS AND SET MISSING INDICATOR.
!
            di = xstd/xn
            xstd = sqrt( di )
!
            where ( xn_grp(:ngrp)/=zero ) 
                xstd_grp(:ngrp) = sqrt( xstd_grp(:ngrp)/xn_grp(:ngrp) )
            elsewhere
                xmean_grp(:ngrp) = xmiss
                xstd_grp(:ngrp) = xmiss
            end where
!
            if ( present(xcomp) .or. present(u) .or. present(prob) ) then
!
!               COMPUTE STANDARDIZED PROFILES OF THE GROUPS.
!
                allocate( work(ngrp), stat=iok )
!
                if ( iok/=0 ) call merror( name_proc//allocate_error )
!
                if ( xstd>=lamch( 's' )  )  then
                    di = one/xstd
                else
                    di = zero
                end if
!
                where ( xn_grp(:ngrp)/=zero ) 
                    work(:ngrp) = (xmean_grp(:ngrp) - xmean)*di
                elsewhere
                    work(:ngrp) = zero
                end where
!
                if ( present(xcomp) ) then
!
!                   OUTPUT STANDARDIZED PROFILES OF THE GROUPS.
!
                    where ( xn_grp(:ngrp)/=zero )
                        xcomp(:ngrp) = work(:ngrp)
                    elsewhere
                        xcomp(:ngrp) = xmiss
                    end where
!
                end if
!
                if ( present(u) .or. present(prob)  ) then
!
!                   COMPUTE U STATISTICS OF THE GROUPS.
!
                    di = xn - one
!
                    where ( xn_grp(:ngrp)/=zero  .and. xn_grp(:ngrp)/=xn )                   &
                    work(:ngrp) = sqrt( ( xn_grp(:ngrp)*di )/( xn - xn_grp(:ngrp) ) )*work(:ngrp)
!
                    if ( present(u) ) then
!
!                       OUTPUT U STATISTICS OF THE GROUPS.
!
                        where ( xn_grp(:ngrp)/=zero )
                            u(:ngrp) = work(:ngrp)
                        elsewhere
                            u(:ngrp) = xmiss
                        end where
!
                    end if
!
                    if ( present(prob) ) then
!
!                       TEST U STATISTICS OF THE GROUPS.
!
                        work(:ngrp) = abs( work(:ngrp) )
                        prob(:ngrp) = two*probn( work(:ngrp), upper=true )
                        where ( xn_grp(:ngrp)==zero ) prob(:ngrp) = xmiss
!
                    end if
!
                end if
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!
            end if
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            di = nan()
!
            xmean            = di
            xstd             = di
            xmean_grp(:ngrp) = di
            xstd_grp(:ngrp)  = di
!
            if ( present(xcomp) ) xcomp(:ngrp) = di
            if ( present(u) )     u(:ngrp)     = di
            if ( present(prob) )  prob(:ngrp)  = di
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_composite_miss_rv
! ________________________________________
!
    end subroutine comp_composite_miss_rv
!
! =========================================================================================
!
    subroutine comp_composite_miss_rm( x, first, last, ngrp, ind, xmean, xstd, xn,        &
                                       xmean_grp, xstd_grp, xn_grp, xmiss,                &
                                       dimvar, xcomp, u, prob, utest )
!
! Purpose
! _______
!
!   COMP_COMPOSITE_MISS computes a composite analysis from a data matrix
!   possibly containing missing values.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:)
!              On entry, input submatrix containing size(X,3-DIMVAR)
!              observations on size(X,DIMVAR) variables from the matrix of data
!              for which a composite analysis is desired. By default, 
!              DIMVAR is equal to 1. See description of optional DIMVAR
!              argument for details.
!              If all the data are available at once, X can be the full 
!              data matrix.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current submatrix is the first submatrix 
!                of the data matrix.
!              - FIRST = false the current submatrix is not the first submatrix
!                of the data matrix.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current submatrix is the last submatrix 
!                of the data matrix.
!              - LAST = false the current submatrix is not the last submatrix
!                of the data matrix.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups in the analysis.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3-DIMVAR) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3-DIMVAR) .
!
!   XMEAN      (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XMEAN contains the mean values from previous calls to COMP_COMPOSITE_MISS.
!              XMEAN should not be changed between calls to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XMEAN contains the mean values of the data 
!              matrix.
!
!              The size of XMEAN must verify size(XMEAN) = size(X,DIMVAR).
!
!   XSTD       (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XSTD contains adjusted sums of squares from previous calls
!              to COMP_COMPOSITE_MISS. XSTD should not be changed between
!              calls to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XSTD contains the standard deviations of 
!              the data matrix.
!
!              The size of XSTD must verify size(XSTD) = size(X,DIMVAR).
!
!   XN         (INPUT/OUTPUT) real(stnd), dimension(:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XN contains counts of non-missing observations from previous calls
!              to COMP_COMPOSITE_MISS. XN should not be changed between calls to
!              COMP_COMPOSITE_MISS.
!
!              On exit, XN contains the numbers of non-missing observations for
!              the variables in the data matrix.
!
!              The size of XN must verify size(XN) = size(X,DIMVAR).
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_COMPOSITE_MISS. XMEAN_GRP should not be changed between calls 
!              to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data matrix.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,DIMVAR)
!              - size(XMEAN_GRP,2) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_COMPOSITE_MISS. XSTD_GRP should not be changed between calls 
!              to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data matrix.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,DIMVAR)
!              - size(XSTD_GRP,2) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XN_GRP contains counts of non-missing observations for the NGRP groups
!              from previous calls to COMP_COMPOSITE_MISS. XN_GRP should not be 
!              changed between calls to COMP_COMPOSITE_MISS.
!
!              On exit, XN_GRP contains the numbers of non-missing observations
!              in the NGRP groups for all the variables in the data matrix.
!
!              The shape of XN_GRP must verify:
!
!              - size(XN_GRP,1) = size(X,DIMVAR)
!              - size(XN_GRP,2) = NGRP.
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!   DIMVAR     (INPUT, OPTIONAL) integer(i4b)
!              On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!              - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!                on size(X,1) variables.
!              - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!                on size(X,2) variables.
!
!              The default is DIMVAR = 1.
!
!   XCOMP      (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the standardized profiles of the NGRP groups of
!              observations in the data matrix.
!
!              The shape of XCOMP must verify:
!
!              - size(XCOMP,1) = size(X,DIMVAR)
!              - size(XCOMP,2) = NGRP.
!
!   U          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the U statistics for the NGRP groups and all
!              the variables in the data matrix.
!
!              The shape of U must verify:
!
!              - size(U,1) = size(X,DIMVAR)
!              - size(U,2) = NGRP.
!
!   PROB       (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!              On exit, the significance probabilities of the U
!              statistics under the null hypothesis that the groups
!              have been formed by uniform random sampling without 
!              duplication in the set of all the observations.
!
!              The shape of PROB must verify:
!
!              - size(PROB,1) = size(X,DIMVAR)
!              - size(PROB,2) = NGRP.
!
!   UTEST      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, a probability. UTEST is the sum of the areas (equal) 
!              in both tails of the normal-distribution.
!              UTEST must verify   0. < P < 1.
!
!              On exit, the two-tail quantile of the normal-distribution, that is a
!              value X such that the probability of the absolute value of U being
!              greater than X is UTEST under the null hypothesis that the groups
!              have been formed by uniform random sampling without duplication in 
!              the set of all the observations.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one valid observation were present for all variables, the statistics
!   are set to Nan code.
!
!   If fewer than one valid observation were present only for some variables and/or groups
!   of observations, the pertinent statistics are set to missing (XMISS value).
!
!   The optional parameters, except DIMVAR, need only to be specified when LAST=true.
!
!   For more details, on the statistics and tests computed by this subroutine, see:
!
!   (1) Terray, P., Delecluse, P., Labattu, S., Terray, L., 2003: Sea Surface
!           Temperature associations with the Late Indian Summer Monsoon,
!           Climate Dynamics, vol. 21, 593-618.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error3, stat_error5, allocate_error
    use Utilities,         only : assert, assert_eq, merror
    use Prob_Procedures,   only : probn, pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                               :: xmiss
    real(stnd), intent(in),               dimension(:,:) :: x
    real(stnd), intent(inout),            dimension(:)   :: xn, xmean, xstd
    real(stnd), intent(inout),            dimension(:,:) :: xn_grp, xmean_grp, xstd_grp
    real(stnd), intent(inout), optional                  :: utest
    real(stnd), intent(out),   optional , dimension(:,:) :: xcomp, u, prob
!
    integer(i4b), intent(in)                          :: ngrp
    integer(i4b), intent(in),            dimension(:) :: ind
    integer(i4b), intent(in),  optional               :: dimvar
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tmp
    real(stnd), dimension(size(xmean))      :: di, dfi, xi
    real(stnd), dimension(:,:), allocatable :: work
!
    integer(i4b) :: i, k, n, m
    integer      :: iok
!
    logical(lgl) :: dim_is_one
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_composite'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(ind),i4b),             &
                    name_proc )
!
    m = assert_eq(  int( (/ size(x,int(i)),              &
                            size(xn),                    &
                            size(xmean),                 &
                            size(xstd),                  &
                            size(xn_grp,1),              &
                            size(xmean_grp,1),           &
                            size(xstd_grp,1) /) ,i4b),   &
                            name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,2)==ngrp,lgl),  &
                 logical(size(xstd_grp,2)==ngrp,lgl),   &
                 logical(size(xn_grp,2)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xstd, xn, xmean_grp, xstd_grp, xn_grp .
!
        xmean(:m) = zero       
        xstd(:m)  = zero       
        xn(:m)    = zero       
!
        xmean_grp(:m,:ngrp) = zero       
        xstd_grp(:m,:ngrp)  = zero       
        xn_grp(:m,:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            xi(:m) = x(:m,i)
        else
            xi(:m) = x(i,:m)
        end if
!
        k = ind(i)
!
        if ( k>=1_i4b .and. k<=ngrp ) then
!
            where ( xi(:m)/=xmiss )
!
                xn(:m) = xn(:m) + one
                di(:m) = xi(:m) - xmean(:m)
                dfi(:m) = di(:m)/xn(:m)
                xmean(:m) = xmean(:m) + dfi(:m)
                xstd(:m) = xstd(:m) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
                xn_grp(:m,k) = xn_grp(:m,k) + one
                di(:m) = xi(:m) - xmean_grp(:m,k)
                dfi(:m) = di(:m)/xn_grp(:m,k)
                xmean_grp(:m,k) = xmean_grp(:m,k) + dfi(:m)
                xstd_grp(:m,k) = xstd_grp(:m,k) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
            end where
!
        else
!
            where ( xi(:m)/=xmiss )
!
                xn(:m) = xn(:m) + one
                di(:m) = xi(:m) - xmean(:m)
                dfi(:m) = di(:m)/xn(:m)
                xmean(:m) = xmean(:m) + dfi(:m)
                xstd(:m) = xstd(:m) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
            end where
!
        end if
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(xcomp) ) call assert( logical(size(xcomp,2)==ngrp,lgl),    &
                                           logical(size(xcomp,1)==m,lgl),       &
                                           name_proc )
!
        if ( present(u) ) call assert( logical(size(u,2)==ngrp,lgl),    &
                                       logical(size(u,1)==m,lgl),       &
                                       name_proc )
!
        if ( present(prob) ) call assert( logical(size(prob,2)==ngrp,lgl),    &
                                          logical(size(prob,1)==m,lgl),       &
                                          name_proc )
!
        if ( present(utest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF NORMAL DISTRIBUTION.
!
            if ( zero>=utest .or. utest>=one ) then
                call merror( name_proc//stat_error3 )
            else
                tmp = utest*half
                utest = abs( pinvn( tmp ) )
            end if
!
        end if
!
        if ( any(xn(:m)>zero) ) then
!
!           COMPUTE STANDARD-DEVIATIONS AND SET MISSING INDICATOR.
!
            where ( xn(:m)>zero )
                di(:m) = xstd(:m)/xn(:m)
                xstd(:m) = sqrt( di(:m) )
            elsewhere
                xmean(:m) = xmiss
                xstd(:m) = xmiss
            end where
!
            where ( xn_grp(:m,:ngrp)/=zero ) 
                xstd_grp(:m,:ngrp) = sqrt( xstd_grp(:m,:ngrp)/xn_grp(:m,:ngrp) )
            elsewhere
                xmean_grp(:m,:ngrp) = xmiss
                xstd_grp(:m,:ngrp) = xmiss
            end where
!
            if ( present(xcomp) .or. present(u) .or. present(prob) ) then
!
!               COMPUTE STANDARDIZED PROFILES OF THE GROUPS.
!
                allocate( work(m,ngrp), stat=iok )
!
                if ( iok/=0 ) call merror( name_proc//allocate_error )
!
                tmp = lamch( 's' )
!
                where ( xstd(:m)>=tmp .and. xn(:m)/=zero  )
                    di(:m) = one/xstd(:m)
                elsewhere
                    di(:m) = zero
                end where
!
                do k = 1_i4b, ngrp
!
                    where ( xn_grp(:m,k)/=zero ) 
                        work(:m,k) = ( xmean_grp(:m,k) - xmean(:m) )*di(:m)
                    elsewhere
                        work(:m,k) = zero
                    end where
!
                end do
!
                if ( present(xcomp) ) then
!
!                   OUTPUT STANDARDIZED PROFILES OF THE GROUPS.
!
                    where ( xn_grp(:m,:ngrp)/=zero ) 
                        xcomp(:m,:ngrp) = work(:m,:ngrp)
                    elsewhere
                        xcomp(:m,:ngrp) = xmiss
                    end where
!                    xcomp(:m,:ngrp) = merge( work(:m,:ngrp), xmiss, xn_grp(:m,:ngrp)/=zero )
!
                end if
!
                if ( present(u) .or. present(prob)  ) then
!
!                   COMPUTE U STATISTICS OF THE GROUPS.
!
                    where ( xn(:m)/=zero ) di(:m) = xn(:m) - one
!
                    do k = 1_i4b, ngrp
!
                        where ( xn_grp(:m,k)/=zero .and. xn_grp(:m,k)/=xn(:m) )
                            dfi(:m) = sqrt( ( xn_grp(:m,k)*di(:m) )/( xn(:m) - xn_grp(:m,k) ) )
                            work(:m,k) = dfi(:m)*work(:m,k) 
                        end where
!
                    end do
!
                    if ( present(u) ) then
!
!                       OUTPUT U STATISTICS OF THE GROUPS.
!
                        where ( xn_grp(:m,:ngrp)/=zero ) 
                            u(:m,:ngrp) = work(:m,:ngrp)
                        elsewhere
                            u(:m,:ngrp) = xmiss
                        end where
!                        u(:m,:ngrp) = merge( work(:m,:ngrp), xmiss, xn_grp(:m,:ngrp)/=zero )
!
                    end if
!
                    if ( present(prob) ) then
!
!                       TEST U STATISTICS OF THE GROUPS.
!
                        work(:m,:ngrp) = abs( work(:m,:ngrp) )
                        prob(:m,:ngrp) = two*probn( work(:m,:ngrp), upper=true )
                        where ( xn_grp(:m,:ngrp)==zero ) prob(:m,:ngrp) = xmiss
!
                    end if
!
                end if
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!
            end if
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            tmp = nan()
!
            xmean(:m)           = tmp
            xstd(:m)            = tmp
            xmean_grp(:m,:ngrp) = tmp
            xstd_grp(:m,:ngrp)  = tmp
!
            if ( present(xcomp) ) xcomp(:m,:ngrp) = tmp
            if ( present(u) )     u(:m,:ngrp)     = tmp
            if ( present(prob) )  prob(:m,:ngrp)  = tmp
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_composite_miss_rm
! ________________________________________
!
    end subroutine comp_composite_miss_rm
!
! =========================================================================================
!
    subroutine comp_composite_miss_rt( x, first, last, ngrp, ind, xmean, xstd, xn,        &
                                       xmean_grp, xstd_grp, xn_grp, xmiss,                &
                                       xcomp, u, prob, utest )
!
! Purpose
! _______
!
!   COMP_COMPOSITE_MISS computes a composite analysis from a data tridimensional
!   array possibly containing missing values.
!
!
! Arguments
! _________
!
!   X          (INPUT) real(stnd), dimension(:,:,:)
!              On entry, input subarray containing size(X,3) observations
!              on size(X,1) by size(X,2) variables from the array of data for 
!              which a composite analysis is desired. If all the data are available
!              at once, X can be the full data array.
!
!   FIRST      (INPUT) logical(lgl)
!              On entry, if:
!
!              - FIRST = true  the current subarray is the first subarray 
!                of the data array.
!              - FIRST = false the current subarray is not the first subarray
!                of the data array.
!
!   LAST       (INPUT) logical(lgl)
!              On entry, if:
!
!              - LAST = true  the current subarray is the last subarray 
!                of the data array.
!              - LAST = false the current subarray is not the last subarray
!                of the data array.
!
!   NGRP       (INPUT) integer(i4b)
!              On entry, the number of groups in the analysis.
!
!   IND        (INPUT) integer(i4b), dimension(:)
!              On entry, input subvector containing size(X,3) observations
!              which is used to classify the observations into the NGRP groups.
!              A value outside the interval 1:NGRP means that the current observation
!              does not belong to any group in the analysis.
!
!              The size of IND must verify size(IND) = size(X,3) .
!
!   XMEAN      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XMEAN contains the mean values from previous calls to COMP_COMPOSITE_MISS.
!              XMEAN should not be changed between calls to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XMEAN contains the mean values of the data array.
!
!              The shape of XMEAN must verify: 
!
!              - size(XMEAN,1) = size(X,1)
!              - size(XMEAN,2) = size(X,2).
!
!   XSTD       (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XSTD contains adjusted sums of squares from previous calls to COMP_COMPOSITE_MISS.
!              XSTD should not be changed between calls to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XSTD contains the standard deviations of 
!              the data array.
!
!              The shape of XSTD must verify:
!
!              - size(XSTD,1) = size(X,1)
!              - size(XSTD,2) = size(X,2).
!
!   XN         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XN contains counts of non-missing observations from previous calls
!              to COMP_COMPOSITE_MISS. XN should not be changed between calls to
!              COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XN contains the numbers of non-missing
!              observations for the variables in the data array.
!
!              The shape of XN must verify:
!
!              - size(XN,1) = size(X,1)
!              - size(XN,2) = size(X,2).
!
!   XMEAN_GRP  (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XMEAN_GRP contains the mean values for the NGRP groups from previous calls
!              to COMP_COMPOSITE_MISS. XMEAN_GRP should not be changed between calls 
!              to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XMEAN_GRP contains the mean values for the NGRP groups
!              of observations on all the variables in the data array.
!
!              The shape of XMEAN_GRP must verify:
!
!              - size(XMEAN_GRP,1) = size(X,1)
!              - size(XMEAN_GRP,2) = size(X,2)
!              - size(XMEAN_GRP,3) = NGRP.
!
!   XSTD_GRP   (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XSTD_GRP contains  adjusted sums of squares for the NGRP groups from previous
!              calls to COMP_COMPOSITE_MISS. XSTD_GRP should not be changed between calls 
!              to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XSTD_GRP contains the standard deviations for the NGRP
!              groups of observations on all the variables in the data array.
!
!              The shape of XSTD_GRP must verify:
!
!              - size(XSTD_GRP,1) = size(X,1)
!              - size(XSTD_GRP,2) = size(X,2)
!              - size(XSTD_GRP,3) = NGRP.
!
!   XN_GRP     (INPUT/OUTPUT) real(stnd), dimension(:,:,:)
!              On entry, after the first call to COMP_COMPOSITE_MISS (e.g. when FIRST=true),
!              XN_GRP contains counts of non-missing observations for the NGRP groups
!              from previous calls to COMP_COMPOSITE_MISS. XN_GRP should not be 
!              changed between calls to COMP_COMPOSITE_MISS.
!
!              On exit, when LAST=true, XN_GRP contains the numbers of non-missing
!              observations in the NGRP groups for all the variables in the data array.
!
!              The shape of XN_GRP must verify:
!
!              - size(XN_GRP,1) = size(X,1)
!              - size(XN_GRP,2) = size(X,2)
!              - size(XN_GRP,3) = NGRP.
!
!   XMISS      (INPUT) real(stnd)
!              On entry, the missing value indicator. Any value in
!              X which is equal to XMISS is assumed to be missing.
!
!   XCOMP      (OUTPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On exit, the standardized profiles of the NGRP groups of
!              observations in the data array.
!
!              The shape of XCOMP must verify:
!
!              - size(XCOMP,1) = size(X,1)
!              - size(XCOMP,2) = size(X,2)
!              - size(XCOMP,3) = NGRP.
!
!   U          (OUTPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On exit, the U statistics for the NGRP groups and all
!              the variables in the data array.
!
!              The shape of U must verify:
!
!              - size(U,1) = size(X,1)
!              - size(U,2) = size(X,2)
!              - size(U,3) = NGRP.
!
!   PROB       (OUTPUT, OPTIONAL) real(stnd), dimension(:,:,:)
!              On exit, the significance probabilities of the U
!              statistics under the null hypothesis that the groups
!              have been formed by uniform random sampling without
!              duplication in the set of all the observations.
!
!              The shape of PROB must verify:
!
!              - size(PROB,1) = size(X,1)
!              - size(PROB,2) = size(X,2)
!              - size(PROB,3) = NGRP.
!
!   UTEST      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!              On entry, a probability. UTEST is the sum of the areas (equal) 
!              in both tails of the normal-distribution.
!              UTEST must verify   0. < P < 1.
!
!              On exit, the two-tail quantile of the normal-distribution, that is a
!              value X such that the probability of the absolute value of U being
!              greater than X is UTEST under the null hypothesis that the groups
!              have been formed by uniform random sampling without duplication in
!              the set of all the observations.
!
!
! Further Details
! _______________
!
!   The subroutine computes all the statistics with only one pass through the data.
!
!   If fewer than one valid observation were present for all variables, the statistics
!   are set to Nan code.
!
!   If fewer than one valid observation were present only for some variables and/or groups
!   of observations, the pertinent statistics are set to missing (XMISS value).
!
!   The optional parameters need only to be specified when LAST=true.
!
!   For more details, on the statistics and tests computed by this subroutine, see:
!
!   (1) Terray, P., Delecluse, P., Labattu, S., Terray, L., 2003: Sea Surface
!           Temperature associations with the Late Indian Summer Monsoon,
!           Climate Dynamics, vol. 21, 593-618.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error3, allocate_error
    use Utilities,         only : assert, assert_eq, merror
    use Prob_Procedures,   only : probn, pinvn
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                                 :: xmiss
    real(stnd), intent(in),               dimension(:,:,:) :: x
    real(stnd), intent(inout),            dimension(:,:)   :: xn, xmean, xstd
    real(stnd), intent(inout),            dimension(:,:,:) :: xn_grp, xmean_grp, xstd_grp
    real(stnd), intent(inout), optional                    :: utest
    real(stnd), intent(out),   optional , dimension(:,:,:) :: xcomp, u, prob
!
    integer(i4b), intent(in)               :: ngrp
    integer(i4b), intent(in), dimension(:) :: ind
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                :: tmp
    real(stnd), dimension(size(x,1))          :: di, dfi
    real(stnd), dimension(:,:,:), allocatable :: work
!
    integer(i4b) :: i, j, k, n, m, p
    integer      :: iok
!
    logical(lgl), dimension(size(x,1)) :: mask
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_composite'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),       &
                    int(size(ind),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(  (/ size(x,1),                   &
                             size(xn,1),                  &
                             size(xmean,1),               &
                             size(xstd,1),                &
                             size(xn_grp,1),              &
                             size(xmean_grp,1),           &
                             size(xstd_grp,1) /), i4b) ,  &
                             name_proc )
!
    p = assert_eq(  int(  (/ size(x,2),                   &
                             size(xn,2),                  &
                             size(xmean,2),               &
                             size(xstd,2),                &
                             size(xn_grp,2),              &
                             size(xmean_grp,2),           &
                             size(xstd_grp,2) /), i4b) ,  &
                             name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xmean_grp,3)==ngrp,lgl),  &
                 logical(size(xstd_grp,3)==ngrp,lgl),   &
                 logical(size(xn_grp,3)==ngrp,lgl),     &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xmean, xstd, xn, xmean_grp, xstd_grp, xn_grp .
!
        xmean(:m,:p) = zero       
        xstd(:m,:p)  = zero       
        xn(:m,:p)    = zero       
!
        xmean_grp(:m,:p,:ngrp) = zero       
        xstd_grp(:m,:p,:ngrp)  = zero       
        xn_grp(:m,:p,:ngrp)    = zero       
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS, CORRECTED SUMS OF SQUARES, HANDLING THE DATA ONCE.
!
        k = ind(i)
!
        if ( k>=1_i4b .and. k<=ngrp ) then
!
            do j = 1_i4b, p
!
                mask(:m) = x(:m,j,i)/=xmiss
!
                where ( mask(:m) )
!
                    xn(:m,j) = xn(:m,j) + one
                    di(:m)   = x(:m,j,i) - xmean(:m,j)
                    dfi(:m)  = di(:m)/xn(:m,j)
                    xmean(:m,j) = xmean(:m,j) + dfi(:m)
                    xstd(:m,j)  = xstd(:m,j) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
                    xn_grp(:m,j,k) = xn_grp(:m,j,k) + one
                    di(:m)  = x(:m,j,i) - xmean_grp(:m,j,k)
                    dfi(:m) = di(:m)/xn_grp(:m,j,k)
                    xmean_grp(:m,j,k) = xmean_grp(:m,j,k) + dfi(:m)
                    xstd_grp(:m,j,k)  = xstd_grp(:m,j,k) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
                end where
!
            end do
!
        else
!
            do j = 1_i4b, p
!
                mask(:m) = x(:m,j,i)/=xmiss
!
                where ( mask(:m) )
!
                    xn(:m,j) = xn(:m,j) + one
                    di(:m)   = x(:m,j,i) - xmean(:m,j)
                    dfi(:m)  = di(:m)/xn(:m,j)
                    xmean(:m,j) = xmean(:m,j) + dfi(:m)
                    xstd(:m,j) = xstd(:m,j) + di(:m)*di(:m) - di(:m)*dfi(:m)
!
                end where
!
            end do
!
        end if
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(xcomp) ) call assert( logical(size(xcomp,3)==ngrp,lgl),    &
                                           logical(size(xcomp,2)==p,lgl),       &
                                           logical(size(xcomp,1)==m,lgl),       &
                                           name_proc )
!
        if ( present(u) ) call assert( logical(size(u,3)==ngrp,lgl),    &
                                       logical(size(u,2)==p,lgl),       &
                                       logical(size(u,1)==m,lgl),       &
                                       name_proc )
!
        if ( present(prob) ) call assert( logical(size(prob,3)==ngrp,lgl),    &
                                          logical(size(prob,2)==p,lgl),       &
                                          logical(size(prob,1)==m,lgl),       &
                                          name_proc )
!
        if ( present(utest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF NORMAL DISTRIBUTION.
!
            if ( zero>=utest .or. utest>=one ) then
                call merror( name_proc//stat_error3 )
            else
                tmp = utest*half
                utest = abs( pinvn( tmp ) )
            end if
!
        end if
!
        if ( any(xn(:m,:p)>zero) ) then
!
!           COMPUTE STANDARD-DEVIATIONS AND SET MISSING INDICATOR.
!
            where ( xn(:m,:p)>zero )
                xstd(:m,:p) = sqrt( xstd(:m,:p)/xn(:m,:p) )
            elsewhere
                xmean(:m,:p) = xmiss
                xstd(:m,:p) = xmiss
            end where
!
            where ( xn_grp(:m,:p,:ngrp)/=zero ) 
                xstd_grp(:m,:p,:ngrp) = sqrt( xstd_grp(:m,:p,:ngrp)/xn_grp(:m,:p,:ngrp) )
            elsewhere
                xmean_grp(:m,:p,:ngrp) = xmiss
                xstd_grp(:m,:p,:ngrp) = xmiss
            end where
!
            if ( present(xcomp) .or. present(u) .or. present(prob) ) then
!
!               COMPUTE STANDARDIZED PROFILES OF THE GROUPS.
!
                allocate( work(m,p,ngrp), stat=iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
                tmp = lamch( 's' )
!
                do k = 1_i4b, ngrp
!
                    do j = 1_i4b, p
!
                        mask(:m) = xstd(:m,j)>=tmp .and. xn(:m,j)>zero .and. xn_grp(:m,j,k)>zero
!
                        where ( mask(:m) ) 
                            work(:m,j,k) = ( xmean_grp(:m,j,k) - xmean(:m,j) )/xstd(:m,j)
                        elsewhere
                            work(:m,j,k) = zero
                        end where
!
                    end do
!
                end do
!
                if ( present(xcomp) ) then
!
!                   OUTPUT STANDARDIZED PROFILES OF THE GROUPS.
!
                    where ( xn_grp(:m,:p,:ngrp)/=zero ) 
                        xcomp(:m,:p,:ngrp) = work(:m,:p,:ngrp)
                    elsewhere
                        xcomp(:m,:p,:ngrp) = xmiss
                    end where
!
                end if
!
                if ( present(u) .or. present(prob)  ) then
!
!                   COMPUTE U STATISTICS OF THE GROUPS.
!
                    do k = 1_i4b, ngrp
!
                       do j = 1_i4b, p
!
                            where ( xn_grp(:m,j,k)/=zero .and. xn_grp(:m,j,k)/=xn(:m,j) )
!
                                di(:m)  = xn(:m,j) - one
                                dfi(:m) = sqrt( ( xn_grp(:m,j,k)*di(:m) )/( xn(:m,j) - xn_grp(:m,j,k) ) )
                                work(:m,j,k) = dfi(:m)*work(:m,j,k) 
!
                            end where
!
                        end do
!
                    end do
!
                    if ( present(u) ) then
!
!                       OUTPUT U STATISTICS OF THE GROUPS.
!
                        where ( xn_grp(:m,:p,:ngrp)/=zero ) 
                            u(:m,:p,:ngrp) = work(:m,:p,:ngrp)
                        elsewhere
                            u(:m,:p,:ngrp) = xmiss
                        end where
!
                    end if
!
                    if ( present(prob) ) then
!
!                       TEST U STATISTICS OF THE GROUPS.
!
                        work(:m,:p,:ngrp) = abs( work(:m,:p,:ngrp) )
!
                        do k = 1_i4b, ngrp
                            prob(:m,:p,k) = two*probn( work(:m,:p,k), upper=true )
                        end do
!
                        where ( xn_grp(:m,:p,:ngrp)==zero ) prob(:m,:p,:ngrp) = xmiss
!
                    end if
!
                end if
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!
            end if
!
        else
!
!           FEWER THAN ONE VALID OBSERVATION SET ALL STATISTICS TO NAN CODE.
!
            tmp = nan()
!
            xmean(:m,:p)           = tmp
            xstd(:m,:p)            = tmp
            xmean_grp(:m,:p,:ngrp) = tmp
            xstd_grp(:m,:p,:ngrp)  = tmp
!
            if ( present(xcomp) ) xcomp(:m,:p,:ngrp) = tmp
            if ( present(u) )     u(:m,:p,:ngrp)     = tmp
            if ( present(prob) )  prob(:m,:p,:ngrp)  = tmp
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_composite_miss_rt
! ________________________________________
!
    end subroutine comp_composite_miss_rt
!
!
! =========================================================================================
!                                 RANK STATISTICS
! =========================================================================================
!
!
    function valmed_rv( x )
!
!
! Purpose
! _______
!                                                                  
!   Find the median of the vector X(:).          
!
! Arguments
! _________
!                                          
!   X      (INPUT) real(stnd), dimension(:)
!          On entry, the vector of observations.
!
!
! Further Details
! _______________
!
!   This subroutine uses a modified quicksort algorithm.
!
!
! _________________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,  only : half
    use Num_Constants,    only : nan
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(in) :: x
    real(stnd)                           :: valmed_rv
!
!
! SPECIFICATIONS FOR VARIABLES
! ____________________________
!
    real(stnd)                     :: temp, xhi, xlo, xmax, xmin, xmed
    real(stnd), dimension(size(x)) :: xtmp
!
    integer(i4b) :: n, hi, lo, nby2, nby2p1, mid, i, j, k
!
    logical(lgl) :: odd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = size( x )
!
    select case (n)
!
        case(:0_i4b)
!
!           FEWER THAN ONE VALID OBSERVATION SET THE MEDIAN TO NAN CODE.
!
            xmed = nan()
!
        case(1_i4b)
!
            xmed = x(1)
!
        case(2_i4b)
!
            xmed = half*( x(1) + x(2) )
!
        case(3_i4b:)
!
            xtmp(:n) = x(:n)
!
            nby2   = n/2_i4b
            nby2p1 = nby2 + 1_i4b
            odd    = n/=2*nby2
!
!           hi AND lo ARE POSITION LIMITS ENCOMPASSING THE MEDIAN.
!
            lo = 1_i4b
            hi = n
!
            if ( odd ) then
!
!               FIND MEDIAN OF 1ST, MIDDLE AND LAST VALUES.
!
                do
!
                    mid  = ( lo + hi )/2_i4b
                    xmed = xtmp(mid)
                    xlo  = xtmp(lo)
                    xhi  = xtmp(hi)
!
!                   SWAP xhi AND xlo
!
                    if ( xhi<xlo ) then
                        temp = xhi
                        xhi  = xlo
                        xlo  = temp
                    end if
!
                    if ( xmed>xhi ) then
                        xmed = xhi
                    else if ( xmed<xlo ) then
                        xmed = xlo
                    end if
!
!                   THE BASIC QUICKSORT ALGORITHM TO MOVE ALL VALUES <= THE SORT KEY (xmed)
!                   TO THE LEFT-HAND END, AND ALL HIGHER VALUES TO THE OTHER END.
!
                    i = lo
                    j = hi
!
                    do
!
                        do
!
                            if ( xtmp(i)>=xmed ) exit
!
                            i = i + 1_i4b
!
                        end do
!
                        do
!
                            if ( xtmp(j)<=xmed ) exit
!
                            j = j - 1_i4b
!
                        end do
!
                        if ( i>=j ) exit
!
                        temp    = xtmp(i)
                        xtmp(i) = xtmp(j)
                        xtmp(j) = temp
                        i = i + 1_i4b
                        j = j - 1_i4b
!
!                       DECIDE WHICH HALF THE MEDIAN IS IN.
!
                        if ( i>j ) exit
!
                    end do    
!
                    if ( j<nby2p1 ) lo = i
                    if ( i>nby2p1 ) hi = j
!
!                   TEST WHETHER MEDIAN HAS BEEN ISOLATED.
!
                    if ( i==j .and. i==nby2p1 ) exit
!
                    if ( lo>=hi-1_i4b ) then
!
                        temp = xtmp(lo)
                        if ( temp>xtmp(hi) ) then
                            xtmp(lo) = xtmp(hi)
                            xtmp(hi) = temp
                        end if
                        xmed = xtmp(nby2p1)
!
                        exit
!
                    end if
!
                end do        
!
            else
!
!               FIND MEDIAN OF 1ST, MIDDLE AND LAST VALUES.
!
                do
!
                    mid  = ( lo + hi )/2_i4b
                    xmed = xtmp(mid)
                    xlo  = xtmp(lo)
                    xhi  = xtmp(hi)
!
!                   Swap xhi & xlo
!
                    if ( xhi<xlo ) then
                        temp = xhi
                        xhi  = xlo
                        xlo  = temp
                    end if
!
                    if ( xmed>xhi ) then
                        xmed = xhi
                    else if ( xmed<xlo ) then
                        xmed = xlo
                    end if
!
!                   THE BASIC QUICKSORT ALGORITHM TO MOVE ALL VALUES <= THE SORT KEY (xmed)
!                   TO THE LEFT-HAND END, AND ALL HIGHER VALUES TO THE OTHER END.
!
                    i = lo
                    j = hi
!
                    do
!
                        do
!
                            if ( xtmp(i)>=xmed ) exit
!
                            i = i + 1_i4b
!
                        end do
!
                        do
!
                            if ( xtmp(j)<=xmed ) exit
!
                            j = j - 1_i4b
!
                        end do
!
                        if ( i>=j ) exit
!
                        temp    = xtmp(i)
                        xtmp(i) = xtmp(j)
                        xtmp(j) = temp
                        i = i + 1_i4b
                        j = j - 1_i4b
!
!                       DECIDE WHICH HALF THE MEDIAN IS IN.
!
                        if ( i>j ) exit
!
                    end do    
!
                    if ( j==nby2 .and. i==nby2p1 ) then
!
!                       SPECIAL CASE, N EVEN, J = N/2 AND I = J + 1, SO THE MEDIAN IS
!                       BETWEEN THE TWO HALVES OF THE SERIES.   FIND MAX. OF THE FIRST
!                       HALF AND MIN. OF THE SECOND HALF, THEN AVERAGE.
!
                        xmax = xtmp(1)
!
                        do k = lo, j
                            xmax = max(xmax, xtmp(k))
                        end do
!
                        xmin = xtmp(n)
!
                        do k = i, hi
                            xmin = min(xmin, xtmp(k))
                        end do
!
                        xmed = half*( xmin + xmax )
!
                        exit
!
                    else
!
                        if ( j<nby2 )   lo = i
                        if ( i>nby2p1 ) hi = j
!
                        if ( i==j ) then
                            if ( i==nby2 )   lo = nby2
                            if ( j==nby2p1 ) hi = nby2p1
                        end if
!
                    end if
!
                    if ( lo>=hi-1_i4b ) then
!
                        xmed = half*( xtmp(nby2) + xtmp(nby2p1) )
!
                        exit
!
                    end if
!
                end do        
!
            end if
!
    end select
!
    valmed_rv = xmed
!
!
! END OF FUNCTION valmed_rv
! _________________________
!
    end function valmed_rv
!
! ===========================================================================================
!
    function valmed_rm( x )
!
!
! Purpose
! _______
!                                          
!   Find the medians of the column vectors of the matrix X(:,:).     
!
! Arguments
! _________
! 
!   X      (INPUT) real(stnd), dimension(:,:)
!          On entry, the matrix of observations.
!
!
! Further Details
! _______________
!
!   This subroutine uses a modified quicksort algorithm.
!
! _________________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,  only : half
    use Num_Constants,    only : nan
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:),       intent(in) :: x
    real(stnd), dimension(size(x,2))             :: valmed_rm
!
!
! SPECIFICATIONS FOR VARIABLES
! ____________________________
!
    real(stnd)                       :: temp, xhi, xlo, xmax, xmin, xmed
    real(stnd), dimension(size(x,1)) :: xtmp
    real(stnd), dimension(size(x,2)) :: xvalmed
!
    integer(i4b) :: n, m, hi, lo, nby2, nby2p1, mid, i, j, k, l
!
    logical(lgl) :: odd
!
#ifdef _OPENMP
    logical      :: test_par
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!  TEST THE ARGUMENTS.
!
    n = size( x, 1 )
    m = size( x, 2 )
!
    if ( m<=0_i4b ) return
!
    select case (n)
!
        case(:0_i4b)
!
!           FEWER THAN ONE VALID OBSERVATION SET THE MEDIANS TO NAN CODE.
!
            xmed        = nan()
            xvalmed(:m) = xmed
!
        case(1_i4b)
!
            xvalmed(:m) = x(1,:m)
!
        case(2_i4b)
!
            xvalmed(:m) = half*( x(1,:m) + x(2,:m) )
!
        case(3_i4b:)
!
            nby2   = n/2_i4b
            nby2p1 = nby2 + 1_i4b
            odd    = n/=2_i4b*nby2
!
#ifdef _OPENMP
            i = omp_get_num_procs()
            k = omp_get_max_threads()
            test_par = .not.( omp_in_parallel() )      .and.      &
                       (m*n)>=omp_limit                .and.      &
                       i>1_i4b                         .and.      &
                       k>1_i4b
#endif
!
            if ( odd ) then
!
!$OMP PARALLEL DO IF(test_par)                &
!$OMP            ,PRIVATE(i,j,k,l,hi,lo,mid,xtmp,temp,xhi,xlo,xmax,xmin,xmed)
!
                do l = 1_i4b, m
!
                    xtmp(:n) = x(:n,i)
!
!                   hi AND lo ARE POSITION LIMITS ENCOMPASSING THE MEDIAN.
!
                    lo = 1_i4b
                    hi = n
!
!                   FIND MEDIAN OF 1ST, MIDDLE AND LAST VALUES.
!
                    do
!
                        mid  = ( lo + hi )/2_i4b
                        xmed = xtmp(mid)
                        xlo  = xtmp(lo)
                        xhi  = xtmp(hi)
!
!                       SWAP xhi AND xlo .
!
                        if ( xhi<xlo ) then
                            temp = xhi
                            xhi  = xlo
                            xlo  = temp
                        end if
!
                        if ( xmed>xhi ) then
                            xmed = xhi
                        else if ( xmed<xlo ) then
                            xmed = xlo
                        end if
!
!                       THE BASIC QUICKSORT ALGORITHM TO MOVE ALL VALUES <= THE SORT KEY (xmed)
!                       TO THE LEFT-HAND END, AND ALL HIGHER VALUES TO THE OTHER END.
!
                        i = lo
                        j = hi
!
                        do
!
                            do
!
                                if ( xtmp(i)>=xmed ) exit
!
                                i = i + 1_i4b
!
                            end do
!
                            do
!
                                if ( xtmp(j)<=xmed ) exit
!
                                j = j - 1_i4b
!
                            end do
!
                            if ( i>=j ) exit
!
                            temp    = xtmp(i)
                            xtmp(i) = xtmp(j)
                            xtmp(j) = temp
                            i = i + 1_i4b
                            j = j - 1_i4b
!
!                           DECIDE WHICH HALF THE MEDIAN IS IN.
!
                            if ( i>j ) exit
!
                        end do        
!
                        if ( j<nby2p1 ) lo = i
                        if ( i>nby2p1 ) hi = j
!
!                       TEST WHETHER MEDIAN HAS BEEN ISOLATED.
!
                        if ( i==j .and. i==nby2p1 ) exit
!
                        if ( lo>=hi-1_i4b ) then
!
                            temp = xtmp(lo)
                            if ( temp>xtmp(hi) ) then
                                xtmp(lo) = xtmp(hi)
                                xtmp(hi) = temp
                            end if
                            xmed = xtmp(nby2p1)
!
                            exit
!
                        end if
!
                    end do        
!
                    xvalmed(l) = xmed
!
                end do    
!
!$OMP END PARALLEL DO
!
            else
!
!$OMP PARALLEL DO IF(test_par)                &
!$OMP            ,PRIVATE(i,j,k,l,hi,lo,mid,xtmp,temp,xhi,xlo,xmax,xmin,xmed)
!
                do l = 1_i4b, m
!
                    xtmp(:n) = x(:n,i)
!
!                   hi AND lo ARE POSITION LIMITS ENCOMPASSING THE MEDIAN.
!
                    lo = 1_i4b
                    hi = n
!
!                   FIND MEDIAN OF 1ST, MIDDLE AND LAST VALUES.
!
                    do
!
                        mid  = ( lo + hi )/2_i4b
                        xmed = xtmp(mid)
                        xlo  = xtmp(lo)
                        xhi  = xtmp(hi)
!
!                       SWAP xhi AND xlo .
!
                        if ( xhi<xlo ) then
                            temp = xhi
                            xhi  = xlo
                            xlo  = temp
                        end if
!
                        if ( xmed>xhi ) then
                            xmed = xhi
                        else if ( xmed<xlo ) then
                            xmed = xlo
                        end if
!
!                       THE BASIC QUICKSORT ALGORITHM TO MOVE ALL VALUES <= THE SORT KEY (xmed)
!                       TO THE LEFT-HAND END, AND ALL HIGHER VALUES TO THE OTHER END.
!
                        i = lo
                        j = hi
!
                        do
!
                            do
!
                                if ( xtmp(i)>=xmed ) exit
!
                                i = i + 1_i4b
!
                            end do
!
                            do
!
                                if ( xtmp(j)<=xmed ) exit
!
                                j = j - 1_i4b
!
                            end do
!
                            if ( i>=j ) exit
!
                            temp = xtmp(i)
                            xtmp(i) = xtmp(j)
                            xtmp(j) = temp
                            i = i + 1_i4b
                            j = j - 1_i4b
!
!                           DECIDE WHICH HALF THE MEDIAN IS IN.
!
                            if ( i>j ) exit
!
                        end do        
!
                        if ( j==nby2 .and. i==nby2p1 ) then
!
!                           SPECIAL CASE, N EVEN, J = N/2 AND I = J + 1, SO THE MEDIAN IS
!                           BETWEEN THE TWO HALVES OF THE SERIES.   FIND MAX. OF THE FIRST
!                           HALF AND MIN. OF THE SECOND HALF, THEN AVERAGE.
!
                            xmax = xtmp(1)
!
                            do k = lo, j
                                xmax = max(xmax, xtmp(k))
                            end do
!
                            xmin = xtmp(n)
!
                            do k = i, hi
                                xmin = min(xmin, xtmp(k))
                            end do
!
                            xmed = half*( xmin + xmax )
!
                            exit
!
                        else
!
                            if ( j<nby2 )   lo = i
                            if ( i>nby2p1 ) hi = j
!
                            if ( i==j ) then
                                if ( i==nby2 )   lo = nby2
                                if ( j==nby2p1 ) hi = nby2p1
                            end if
!
                        end if
!
                        if ( lo>=hi-1_i4b ) then
!
                            xmed = half*( xtmp(nby2) + xtmp(nby2p1) )
!
                            exit
!
                        end if
!
                    end do        
!
                    xvalmed(l) = xmed
!
                end do    
!
!$OMP END PARALLEL DO
!
            end if
!
    end select
!
    valmed_rm(:m) = xvalmed(:m)
!
!
! END OF FUNCTION valmed_rm
! _________________________
!
    end function valmed_rm
!
! =========================================================================================
!
!
! *******************************
! END OF MODULE Stat_Procedures *
! *******************************
!    
end module Stat_Procedures
